(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

theory Confluence
imports
  Main
  Abstract_Rewriting
  Multiset
begin

(*
definition confluent :: "'a rel \<Rightarrow> bool"
 where "confluent R \<equiv> (R\<^sup>\<up> \<subseteq> R\<^sup>\<down>)"

definition CR :: "'a rel \<Rightarrow> bool"
 where "CR R \<equiv> (R^<->* \<subseteq> R\<^sup>\<down>)"
*)

definition commute_l :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where "commute_l R S \<equiv> ((R^-1)^* O S) \<subseteq> (S O (R^-1)^*)"

definition commute_r :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where "commute_r R S \<equiv> (R^-1) O S^* \<subseteq> S^* O R^-1"

definition commute_lr :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
where "commute_lr R S \<equiv> R^-1 O S \<subseteq> S O R^-1"

lemma comp_union_tmp: assumes "(x,z) \<in> R O S" shows "(x,z) \<in> (R \<union> S)^*"
 proof - 
 from assms obtain y where "(x,y) \<in> R" and "(y,z) \<in> S" by auto
 hence "(x,y) \<in> (R \<union> S)^*" and "(y,z) \<in> (R \<union> S)^*" by auto
 thus ?thesis by auto
qed

lemma union_subset_rtrancl_composition: "R \<union> S \<subseteq> R^* O S^*" by auto
lemma composition_subset_union_rtrancl: "R O S \<subseteq> (R \<union> S)^*" by (simp add: comp_union_tmp subsetI2)
lemma rtrancl_composition_subset_union_rtrancl: "R^* O S^* \<subseteq> (R \<union> S)^*" 
 using composition_subset_union_rtrancl rtrancl_Un_rtrancl by blast

lemma CR_intermediate: assumes "R \<subseteq> S" and "S \<subseteq> R^*" and "CR(S)" shows "CR(R)"
  proof -
 have "S^* = R^*" using assms(1,2) rtrancl_mono rtrancl_idemp by blast
 thus ?thesis using assms(3) unfolding CR_defs Abstract_Rewriting.join_def rtrancl_converse by simp
qed

lemma commute_iff_meet_subset_join: "(commute R R) = (meet R \<subseteq> Abstract_Rewriting.join R)"
  using assms unfolding commute_def meet_def Abstract_Rewriting.join_def by auto

lemma CR_iff_meet_subset_join: "CR(R) = (meet R \<subseteq> Abstract_Rewriting.join R)"
proof
 assume "CR R" show "meet R \<subseteq> Abstract_Rewriting.join R" using `CR R` by auto
next
 assume "meet R \<subseteq> join R" show "CR R" using `meet R \<subseteq> join R` unfolding CR_def CR_elt_def using subsetD by auto
qed

lemma CR_iff_commute: "CR R = commute R R" using commute_iff_meet_subset_join CR_iff_meet_subset_join by auto


lemma CR_iff_rtrancl_diamond: assumes "CR(R)" shows "\<diamond>(R^*)" using CR_iff_commute assms unfolding commute_def diamond_def rtrancl_converse by blast

lemma diamond_commute: assumes "\<diamond>(R)" and "\<diamond>(S)" and "commute_lr R S" shows "\<diamond>(R \<union> S)" 
 using assms unfolding diamond_def commute_lr_def by auto

lemma commute0: assumes "CR(R)" and "CR(S)" and "commute R S" shows "CR(R^* \<union> S^*)" 
proof -
 from assms have "\<diamond>(R^*)" and  "\<diamond>(S^*)" using CR_iff_rtrancl_diamond by auto
 thus ?thesis using assms(3) diamond_commute diamond_imp_CR unfolding commute_lr_def commute_def rtrancl_converse by auto 
qed

lemma commute: assumes "CR(R)" and "CR(S)" and "commute R S" shows "CR(R \<union> S)"
 proof -
 from assms have a: "CR(R^* \<union> S^*)" using commute0 unfolding commute_def by blast
 have "(R \<union> S) \<subseteq> (R^* \<union> S^*)" and "(R^* \<union> S^*) \<subseteq> (R \<union> S)^*" using rtrancl_Un_rtrancl by auto
 thus ?thesis using a CR_intermediate by blast
qed

(* shorter/more general proofs than Abstract_Rewriting.thy *)
lemma diamond_imp_commute_r: assumes "\<diamond>(R)" shows "commute_r R R" unfolding commute_r_def
proof (rule subsetI2)
fix y z assume "(y,z) \<in> (R^-1) O R^*" 
 then obtain x where "(x,y) \<in> (R)" and "(x,z) \<in> R^*" unfolding rtrancl_converse converseI by auto
 then obtain n where "(x,z) \<in> R^^n" by auto
 with `(x,y) \<in> R` show "(y,z) \<in> R^* O (R^-1)"
 proof(induct n arbitrary: x y z)
  case 0 thus ?case by auto
 next
 case (Suc n)
  from `(x,z) \<in> R^^(Suc n)` obtain w where "(x,w) \<in> R^^n" and "(w,z) \<in> R" by auto
  with `(x,y) \<in> R` and `(x,w) \<in> R^^n` have "(y,w) \<in> R^* O R^-1" using Suc(1) by auto
  then obtain w' where "(y,w') \<in> R^*" and "(w',w) \<in> R^-1" by auto
  with `(w,z) \<in> R` and `(w',w) \<in> R^-1` obtain z' where "(z,z') \<in> R" and "(w',z') \<in> R" using assms(1)by auto
  have "(y,z') \<in> R^*" using `(y,w') \<in> R^*` and `(w',z') \<in> R` by auto
  show ?case using `(y,z') \<in> R^*` and `(z,z') \<in> R` by auto
qed
qed

lemma commute_lr_imp_commute_r: assumes "commute_lr R S" shows "commute_r R S" unfolding commute_r_def  
proof (rule subsetI2)
fix y z assume "(y,z) \<in> (R^-1) O S^*" 
 then obtain x where "(x,y) \<in> R" and "(x,z) \<in> S^*" unfolding rtrancl_converse converseI by auto
 then obtain n where "(x,z) \<in> S^^n" by auto
 with `(x,y) \<in> R` show "(y,z) \<in> S^* O (R^-1)"
 proof(induct n arbitrary: x y z)
  case 0 thus ?case by auto
 next
 case (Suc n)
  from `(x,z) \<in> S^^(Suc n)` obtain w where "(x,w) \<in> S^^n" and "(w,z) \<in> S" by auto
  with `(x,y) \<in> R` and `(x,w) \<in> S^^n` have "(y,w) \<in> S^* O R^-1" using Suc(1) by auto
  then obtain w' where "(y,w') \<in> S^*" and "(w',w) \<in> R^-1" by auto
  with `(w,z) \<in> S` and `(w',w) \<in> R^-1` obtain z' where "(z,z') \<in> R" and "(w',z') \<in> S" using assms(1) unfolding commute_lr_def by auto
  have "(y,z') \<in> S^*" using `(y,w') \<in> S^*` and `(w',z') \<in> S` by auto
  show ?case using `(y,z') \<in> S^*` and `(z,z') \<in> R` by auto
qed
qed

lemma commute_r_imp_commute: assumes "commute_r R S" shows "commute R S" unfolding commute_def 
 proof (rule subsetI2)
  fix y z assume "(y,z) \<in> (R^-1)^* O S^*"
  then obtain x where "(x,y) \<in> R^*" and "(x,z) \<in> S^*" unfolding rtrancl_converse by auto
  then obtain n where "(x,y) \<in> R^^n" by auto
  with `(x,z) \<in> S^*` show "(y,z) \<in> S^* O (R^-1)^*"
  proof(induct n arbitrary: x y z)
   case 0 thus ?case  by auto 
    next
   case (Suc n)
   from `(x,y) \<in> R^^(Suc n)` obtain w where "(x,w) \<in> R" and "(w,y) \<in> R^^n" using rel_pow_Suc_D2 by best
   from `(x,w) \<in> R` and `(x,z) \<in> S^*` have "(w,z) \<in> (R^-1 O S^*)" by auto
   hence "(w,z) \<in> S^* O (R^-1)" using assms unfolding commute_r_def by auto
   then obtain w' where "(w,w') \<in> S^*" and "(z,w') \<in> R" using assms by best
   from `(w,y) \<in> R^^n` and `(w,w') \<in> S^*` have "(y,w') \<in> S^* O (R^-1)^*" using Suc by auto
   then obtain z' where "(y,z') \<in> S^*" and "(z',w') \<in> (R^-1)^*" by auto
   from `(z,w') \<in> R` and `(z',w') \<in> (R^-1)^*` have "(z,z') \<in> R^*" unfolding rtrancl_converse by auto
   thus ?case using `(y,z') \<in> S^*`  unfolding rtrancl_converse by auto
 qed
qed

lemma diamond_imp_self_commute: assumes "\<diamond> R" shows "commute R R" using assms diamond_imp_commute_r commute_r_imp_commute by auto
lemma meet_imp_conversion: "R\<^sup>\<down> \<subseteq> R^<->*" unfolding join_def conversion_def using rtrancl_composition_subset_union_rtrancl by auto

lemma rtrancl_join_join:
 assumes "(a,b) \<in> A^*" and "(b,c) \<in> join A" shows "(a,c) \<in> join A"
proof -
from assms(2) obtain x where "(b,x) \<in> A^*" and "(c,x) \<in> (A)^*" by best
hence "(a,x) \<in> A^*" using assms(1) by auto
thus ?thesis using `(c,x) \<in> A^*` by auto 
qed

lemma newman: assumes "SN R" and "WCR R" shows "CR R"
 proof 
 fix x
 from assms(1) show "CR_elt R x"
 proof (induct)
 case (IH x) show ?case
 proof
  fix y z assume "(x,y) \<in> R^*" and "(x,z) \<in> R^*"
  then obtain n m where m: "(x,y) \<in> R^^m" and n: "(x,z) \<in> R^^n" by blast
  show "(y,z) \<in> join R" 
  proof (cases n)
   case 0 
    hence "x=z" using n m unfolding join_def by auto
    hence "(z,y) \<in> R^*" using `(x, y) \<in> R\<^sup>*` by auto
    thus ?thesis by auto
   next
   case (Suc n')
    from `(x,z) \<in> R^^n`[unfolded Suc] obtain t where "(x,t) \<in> R" and "(t,z) \<in> R^*" .. 
  show ?thesis
   proof (cases m)
    case 0 
     hence "x=y" using n m unfolding join_def by auto
     hence "(y,z) \<in> R^*" using `(x,z) \<in> R^*` by auto
     thus ?thesis by auto
    next
    case (Suc m') 
     from `(x,y) \<in> R^^(m)`[unfolded Suc] obtain s where "(x,s) \<in> R" and "(s,y) \<in> R^*" ..
     from assms(2) and `(x,t) \<in> R` and `(x,s) \<in> R` have "(t,s) \<in> join R" by auto
     then obtain u where "(t,u) \<in> R^*" and "(s,u) \<in> R^*" ..
     from IH[OF `(x,s) \<in> R`] have "CR_elt R s" .
     hence "(u,y) \<in> join R" using `(s,y) \<in> R^*` `(s,u) \<in> R^*` by auto
     then obtain v where "(u,v) \<in> R^*" and "(y,v) \<in> R^*" by auto
     from IH[OF `(x,t) \<in> R`] have "CR_elt R t" .
     have "(t,v) \<in> R^*" using `(t,u) \<in> R^*` `(u,v) \<in> R^*` by auto
     hence "(v,z) \<in> join R" using `CR_elt R t` `(t,v) \<in> R^*` `(t,z) \<in> R^*` by auto
     thus ?thesis using `(y,v) \<in> R^*` by (simp add: rtrancl_join_join) 
   qed
 qed
qed
qed
qed

(*Decreasing Diagrams*)

(*Definition 2.5(1)*)
definition downset_elt :: "('a \<times>'a) set \<Rightarrow> 'a \<Rightarrow> 'a set" ("(\<bottom>e\<^sub>_ _)" [1000] 999)
 where "downset_elt R a = {y | x y. (x,y) \<in> R \<and> y = a}"

definition downset_set :: "('a \<times> 'a) set \<Rightarrow> 'a set \<Rightarrow> 'a set" ("(\<bottom>s\<^sub>_ _)" [1000] 999)
 where "downset_set R s = {y | x y. x \<in> s \<and> y \<in> downset_elt R x}"

definition downset_multiset :: "('a \<times> 'a) set \<Rightarrow> 'a multiset \<Rightarrow> 'a set" ("(\<bottom>m\<^sub>_ _)" [1000] 999)
 where "downset_multiset R M = downset_set R (Multiset.set_of M)"

definition downset_list :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a set" ("(\<bottom>l\<^sub>_ _)" [1000] 999)
 where "downset_list R xs = downset_set R (List.set xs)"

(*Definition 2.5(2)*)

(*Lemma 2.6*)

lemma downset_list_empty_is_empty: "downset_list r [] = {}" unfolding downset_list_def downset_set_def by auto

lemma downset_elt_vs_downset_singleton_list: "downset_elt r \<alpha> = downset_list r [\<alpha>]" 
 unfolding downset_list_def unfolding downset_set_def downset_elt_def by auto


lemma lemma2_6_1': "downset_list r (xs@ys) = downset_list r xs \<union> downset_list r ys" 
 proof (induct xs)
  case Nil thus ?case unfolding downset_list_empty_is_empty by auto
 next
  case (Cons x xs) thus ?case unfolding downset_list_def downset_set_def by auto
qed

lemma lemma2_6_1'_elt: "downset_list r (\<alpha>#\<sigma>) = downset_list r \<sigma> \<union> downset_elt r \<alpha>" 
 unfolding downset_elt_vs_downset_singleton_list using lemma2_6_1'[where xs = "[\<alpha>]"] 
 unfolding append_Cons append_Nil by blast


(*Def 3.1: lexicographic maximum measure*)

definition remove :: "'a multiset \<Rightarrow> 'a set \<Rightarrow> 'a multiset" ("(_-s_)" [999] 998)
 where "M -s S = Multiset.filter (\<lambda>x. x \<notin> S) M"
definition intersect :: "'a multiset \<Rightarrow> 'a set \<Rightarrow> 'a multiset" ("(_\<inter>s_)" [999] 998)
 where "M \<inter>s S = Multiset.filter (\<lambda>x. x \<in> S) M"

fun lexmax :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a multiset" ("(_|_|)") where 
   "lexmax r [] = {#}"
 | "lexmax r (\<alpha>#\<sigma>) =  {#\<alpha>#} + ((lexmax r \<sigma>) -s (downset_elt r \<alpha>))" 

lemma lemma2_6_1_set: "downset_set r (S \<union> T) = downset_set r S \<union> downset_set r T" unfolding downset_set_def by auto
lemma lemma2_6_1_multiset: "downset_multiset r (M + N) = downset_multiset r M \<union> downset_multiset r N"
 unfolding downset_multiset_def set_of_union using lemma2_6_1_set[where S="set_of M" and T="set_of N"] by auto  


lemma lemma2_6_1_set_aux: "downset_set r (S - downset_set r T) =  (downset_set r S) - downset_set r T" sorry
lemma lemma2_6_1_multiset_aux: 
 "downset_multiset r (M -s downset_multiset r N) = downset_multiset r M - downset_multiset r N" sorry


(*lemma w: "set_of (M -s S) = set_of M - S" unfolding remove_def by auto*)
lemma y0: "\<bottom>s\<^sub>r {a} = \<bottom>e\<^sub>r a" unfolding downset_elt_def downset_set_def by auto
lemma y: "\<bottom>m\<^sub>r {#a#} = \<bottom>e\<^sub>r a" unfolding downset_multiset_def set_of_single y0 by auto 
lemma y':"\<bottom>l\<^sub>r [a] = \<bottom>e\<^sub>r a" unfolding downset_list_def set.simps(2) set.simps(1) y0 by auto 
lemma z: "downset_set r (downset_elt r a) = downset_elt r a" sorry

(*Lemma 3.2(1)*)
lemma lemma3_2_1: "downset_multiset r (lexmax r \<sigma>) = downset_list r \<sigma>" proof (induct \<sigma>)
 case Nil thus ?case unfolding downset_multiset_def downset_list_def by auto
 next
 case (Cons a \<sigma>) 
 have "\<bottom>m\<^sub>r r|a # \<sigma>| = \<bottom>m\<^sub>r ({#a#} + r|\<sigma>|-s\<bottom>e\<^sub>r a)" (is "(?L = _)") by auto 
 hence "?L = downset_multiset r ({#a#} + (r|\<sigma>|-s\<bottom>e\<^sub>r a))" (is "_ = downset_multiset r (?A + ?B)") by auto 
 hence "?L = downset_multiset r ?A \<union> downset_multiset r ?B" unfolding lemma2_6_1_multiset by auto
 hence "?L = downset_multiset r ?A \<union> (downset_multiset r (lexmax r \<sigma>) - downset_multiset r {#a#})" unfolding y[symmetric] unfolding lemma2_6_1_multiset_aux by auto 
 hence "?L = downset_multiset r ?A \<union> downset_multiset r (lexmax r \<sigma>)" using Set.Un_Diff_cancel by auto
 hence "?L = downset_multiset r ?A \<union> downset_list r \<sigma>" using Cons by auto
 hence "?L = downset_list r [a] \<union> downset_list r \<sigma>" unfolding y y' Cons  by auto
 hence "?L = downset_list r (a#\<sigma>)"
by (metis List.set.simps(1) List.set.simps(2) downset_list_def insert_is_Un lemma2_6_1_set set_of_single) 
thus ?case by auto
qed


lemma empty_union_empty_is_empty : "({#}::'a multiset) + {#} = {#}" by auto
lemma empty_subset_single: "{#} \<subset># {#x#}" unfolding less_multiset_def mult_def mult1_def by auto 
lemma empty_subseteq_single: "{#} \<subseteq># {#x#}" using empty_subset_single unfolding le_multiset_def by auto

lemma filter_subset_single: "Multiset.filter P {#x#} \<subseteq># {#x#}" 
 proof (cases "P x") 
 case True thus ?thesis by auto
 next
 case False thus ?thesis using empty_subseteq_single by auto 
qed

lemma filter_subset: "Multiset.filter P M \<subseteq># M" 
 proof (induct M )
 case empty thus ?case by auto
next 
 case (add M x) thus ?case unfolding filter_union using filter_subset_single multiset_order.add_mono by blast
qed

lemma filter: "M = Multiset.filter (\<lambda>x. True) M" unfolding filter_def by (simp add: count_inverse)
lemma obvious: "x \<notin> {} = True" by auto
lemma filter_strange: "M = Multiset.filter (\<lambda>x. x \<notin> {}) M" unfolding obvious using filter by auto 


lemma minuss_empty_neutral1: "{#} -s S = {#}" unfolding remove_def by auto 

lemma lemmaA_3_8:  "remove (M + N) S = (remove M S) + (remove N S)" proof (rule multiset_eqI)
fix x show "count ((M + N) -sS) x = count (M -s S + N -s S) x" unfolding remove_def by auto
qed

lemma lemmaA_3_9: "((remove (remove M S) T)::'a multiset) = remove M (S \<union> T)" proof (rule multiset_eqI)
fix x show "count ((M -sS) -sT) x = count (M -sS \<union> T) x" unfolding remove_def by auto
qed

lemma x1: "\<exists> M0 K. (N::'a multiset) = M0 + K" proof (induct N)
 case empty thus ?case by auto 
next
 case (add n N) thus ?case by auto
qed

lemma x2: "M \<noteq> {#} \<Longrightarrow> \<exists> M0 a. M = M0 + {#a#}" by (rule multi_nonempty_split) 

lemma xy: "N <# N+{#a#}"
 proof -
 have "\<forall>b. b \<in># {#} \<longrightarrow> (b,a) \<in> {(x,y). x < y}" (is "?A3") by auto
 hence "N = N + {#} \<and> ?A3" (is "?A2") by auto
 hence "N + {#a#} = N + {#a#} \<and> ?A2" by auto
 hence "\<exists> M0 K. N+{#a#} = M0 + {#a#} \<and> N = M0 + K \<and> (\<forall>b. b \<in># K \<longrightarrow> (b, a) \<in> {(x, y). x < y})" by blast 
 thus ?thesis unfolding less_multiset_def mult_def mult1_def by blast 
qed

lemma xz: "N <=# N + M" proof 
 (induct M arbitrary: N)  
  case empty thus ?case by auto
 next
  case (add M m) 
  have 1: "N <=# N + {#m#}" using xy multiset_order.less_imp_le by auto
  have "N+{#m#} <=# N+{#m#} + M" using add by auto
  hence 2: "N + {#m#} <=# N + M + {#m#}" unfolding add_assoc unfolding union_commute by auto 
  show ?case using 1 2 unfolding add_assoc by auto
qed

definition submultiset :: "'a multiset \<Rightarrow> 'a multiset \<Rightarrow> bool" ("_ \<subseteq>m _" [1000] 999) where
 "submultiset M N = (\<forall>x. count M x <= count N x)"

lemma yz: assumes "M \<subseteq>m N" shows "\<exists>P. M + P = N"
 using assms unfolding multiset_eq_iff submultiset_def 
 by (metis mset_le_exists_conv mset_less_eqI) 

lemma lemma6_2_2: assumes "M \<subseteq>m N" shows "M \<subseteq># N" 
 proof - 
 from assms have "\<exists>P. M + P = N" using yz by auto 
 thus ?thesis using xz by auto 
qed


lemma lemma3_2_2: "lexmax r (\<sigma>@\<tau>) = lexmax r \<sigma> + (Multiset.filter (\<lambda> x. x \<notin> downset_list r \<sigma>) (lexmax r \<tau>))"
proof(induct \<sigma>)
 case Nil thus ?case unfolding downset_list_empty_is_empty obvious using filter by auto 
 next
 case (Cons \<alpha> \<sigma>)
  have "lexmax r (\<alpha>#\<sigma>@\<tau>) = {#\<alpha>#} + (remove (lexmax r (\<sigma>@\<tau>)) (downset_elt r \<alpha>))" (is "?L = _") by auto
  hence "?L = {#\<alpha>#} + (remove (lexmax r \<sigma> + (remove (lexmax r \<tau>) (downset_list r \<sigma>))) (downset_elt r \<alpha>))" 
   using Cons unfolding remove_def by auto 
  hence "?L = {#\<alpha>#} + (remove (lexmax r \<sigma>) (downset_elt r \<alpha>)) + (remove (remove (lexmax r \<tau>) (downset_list r \<sigma>)) (downset_elt r \<alpha>))"
   unfolding remove_def add_assoc by simp (*no lemmaA_3_8 needed*)
 hence "?L = lexmax r (\<alpha>#\<sigma>) + (remove (remove (lexmax r \<tau>) (downset_list r \<sigma>)) (downset_elt r \<alpha>))" by simp
  hence tmp: "?L= lexmax r (\<alpha>#\<sigma>) + (remove (lexmax r \<tau>) (downset_list r \<sigma> \<union> downset_list r [\<alpha>]))" 
   unfolding lemmaA_3_9 downset_list_def downset_set_def by simp
 hence "?L = lexmax r (\<alpha>#\<sigma>) + (remove (lexmax r \<tau>) (downset_list r (\<alpha>#\<sigma>)))" 
  unfolding lemma2_6_1'_elt downset_list_empty_is_empty (*different lemma2_6_1 needed*)
  using Un_commute by auto
thus ?case unfolding remove_def by simp 
qed


definition mult_eq :: "('a \<times> 'a) set => ('a multiset \<times> 'a multiset) set" where
  "mult_eq r = (mult1 r)\<^sup>*"

(*Definition 3.3*)

definition decreasing :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
 "decreasing r \<tau> \<sigma> \<sigma>' \<tau>' = (
  (lexmax r (\<sigma>@\<tau>'), lexmax r \<tau> + lexmax r \<sigma>) \<in> mult_eq r \<and> 
  (lexmax r (\<tau>@\<sigma>'), lexmax r \<tau> + lexmax r \<sigma>) \<in> mult_eq r)"

lemma dec_eq1: assumes "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" 
 shows "((lexmax r \<tau>') -s (downset_list r \<sigma>), lexmax r \<tau>) \<in> mult_eq r" sorry 

lemma dec_eq2: assumes "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'"
 shows "((lexmax r \<sigma>') -s (downset_list r \<tau>), lexmax r \<sigma>) \<in> mult_eq r" sorry

lemma mul_eq: assumes "(M,N) \<in> mult_eq r" 
 shows "\<exists> X Y Z. M = Z + X \<and> N = Z + Y \<and> (\<forall>x. x \<in># X \<and> (\<exists> y. y \<in># Y \<and> (x,y) \<in> r))" 
sorry

lemma mul_eq2: assumes "\<exists> X Y Z. M = Z + X \<and> N = Z + Y \<and> (\<forall>x. x \<in># X \<and> (\<exists> y. y \<in># Y \<and> (x,y) \<in> r))" 
 shows "(M,N) \<in> mult_eq r"
sorry

lemma fool: assumes "(M,N) \<in> mult_eq r" shows "(P + M,P + N) \<in> mult_eq r" proof - 
 have "\<exists> X Y Z. M = Z + X \<and> N = Z + Y \<and> (\<forall>x. x \<in># X \<and> (\<exists> y. y \<in># Y \<and> (x,y) \<in> r))" using mul_eq[OF assms] . 
 hence "\<exists> X Y Z. P + M = P + Z + X \<and> P + N = P + Z + Y \<and> (\<forall>x. x \<in># X \<and> (\<exists> y. y \<in># Y \<and> (x,y) \<in> r))"
  by (metis union_assoc)
 hence x: "\<exists> X Y Z. P + M = Z + X \<and> P + N = Z + Y \<and> (\<forall>x. x \<in># X \<and> (\<exists> y. y \<in># Y \<and> (x,y) \<in> r))" by metis
 thus ?thesis using mul_eq2[where M="P + M" and N="P+N" and r="r"] by blast 
qed

lemma foor: assumes "(M,N) \<in> mult_eq r" shows "(M + P, N + P) \<in> mult_eq r" 
 using assms fool unfolding union_commute by auto 


(*only the case when S \<subseteq> N,M*)
lemma foo1: assumes "(N,M) \<in> mult_eq r" shows "(N -s S, M -s S) \<in> mult_eq r" sorry

lemma lemma3_5_1: 
 assumes D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "(lexmax r (\<tau> @ v @ \<sigma>''), lexmax r (\<tau> @ v) + lexmax r \<sigma>) \<in> mult_eq r" proof -
 have "lexmax r (\<tau> @ v @ \<sigma>'') = (lexmax r (\<tau> @ v) + ((lexmax r \<sigma>'') -s (downset_list r (\<tau>@v))))" (is "?L = ?R")
  by (metis append_assoc remove_def lemma3_2_2) 
 hence x: "?L = lexmax r (\<tau>@v) + (((lexmax r \<sigma>'') -s downset_list r v) -s downset_list r \<tau>)" (is "?L = ?R")
 using lemmaA_3_9 lemma2_6_1' by (metis Un_commute)
  have "((lexmax r \<sigma>'') -s (downset_list r v),lexmax r \<sigma>') \<in> mult_eq r" (is "(?l,?r) \<in> _ ") using assms dec_eq2 by blast
  hence "(?l -s (downset_list r \<tau>), ?r -s (downset_list r \<tau>)) \<in> mult_eq r" (is "(?l,?r) \<in> _")
   using foo1 by auto
  hence "(lexmax r (\<tau>@v) + ?l,lexmax r (\<tau>@v) + ?r) \<in> mult_eq r" using fool by blast 
 hence y: "(?L,lexmax r (\<tau>@v) + ((lexmax r \<sigma>') -s (downset_list r \<tau>))) \<in> mult_eq r" using x by auto
  have "((lexmax r \<sigma>') -s (downset_list r \<tau>),lexmax r \<sigma>) \<in> mult_eq r" (is "(?l,?r) \<in> _") using D1 dec_eq2 by blast
  hence "(lexmax r (\<tau>@v) + ?l, lexmax r (\<tau>@v) + ?r) \<in> mult_eq r" using fool by auto
 hence "(?L,lexmax r (\<tau>@v) + lexmax r \<sigma>) \<in> mult_eq r" using y unfolding mult_eq_def using rtrancl_trans by auto 
 thus ?thesis by auto
qed 

lemma lemmaA_3_10: "M = (M \<inter>s S) + (M -s S)" unfolding intersect_def remove_def by (metis multiset_partition) 
(*lemma lemmaA_3_11: "(M -s T) \<inter>s S = (M \<inter>s S) -s T" unfolding remove_def intersect_def sorry*)

lemma claim1: "(lexmax r (\<sigma>@\<tau>') + (((lexmax r v' -s downset_list r (\<sigma>@\<tau>')) \<inter>s downset_list r \<tau>))
                ,lexmax r \<sigma> + lexmax r \<tau>) \<in> mult_eq r" sorry
lemma claim1': "(lexmax r (\<sigma>@\<tau>') + (((lexmax r v' -s downset_list r (\<sigma>@\<tau>')) \<inter>s downset_list r \<tau>))
                  \<subseteq># lexmax r \<sigma> + lexmax r \<tau>)" sorry

lemma add_right: assumes "M <=# N" shows "M + C <=# N + C" sorry

lemma lemma3_5_2: assumes D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "(lexmax r (\<sigma> @ \<tau>' @ v'), lexmax r \<sigma> + lexmax r (\<tau>@v)) \<in> mult_eq r" 
proof -
 have "lexmax r (\<sigma> @ \<tau>' @ v') = lexmax r (\<sigma>@\<tau>') + (lexmax r v' -s downset_list r (\<sigma>@\<tau>'))" using lemma3_2_2 
  by (metis append_assoc lemma3_2_1 remove_def) 
 hence "... = lexmax r (\<sigma>@\<tau>') + ((lexmax r v' -s downset_list r (\<sigma>@\<tau>')) \<inter>s downset_list r \<tau>) + 
                               ((lexmax r v' -s downset_list r (\<sigma>@\<tau>')) -s downset_list r \<tau>)" using lemmaA_3_10 
   by (metis union_assoc) 
oops




lemma Lemma3_5: 
 assumes D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "decreasing r (\<tau>@v) \<sigma> \<sigma>'' (\<tau>'@v')"
 proof -
 from assms have 1:"(lexmax r (\<tau>@v@\<sigma>''),lexmax r (\<tau>@v) + lexmax r \<sigma>) \<in> mult_eq r" using lemma3_5_1 by blast
 from assms have 2:"(lexmax r (\<sigma>@\<tau>'@v'),lexmax r \<sigma> + lexmax r (\<tau>@v)) \<in> mult_eq r" using lemma3_5_2 by blast
 from 1 2 show ?thesis  unfolding decreasing_def by (metis append_eq_appendI union_commute) 
qed
 

lemma Lemma3_6: 
"lexmax r \<sigma> + (lexmax r (\<tau>@v)) > lexmax r \<sigma>' + lexmax r' v"
sorry

lemma theorem3_7:
"main_result"
sorry


end

