(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Linear_Poly_Carrier
imports
  Linear_Poly_Order
  Matrix_Poly 
begin

instantiation arctic :: "show" begin
fun shows_arctic :: "arctic \<Rightarrow> shows"
  where "shows_arctic(Num_arc i) = shows i"
  | "shows_arctic(MinInfty)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ai = shows_arctic ai"
definition "shows_list (as::arctic list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic: "shows_prec d (a::arctic) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_def)
lemma assoc_arctic_elt: "\<forall>(a::arctic)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic by auto
instance 
  by (intro_classes,rule assoc_arctic,unfold shows_list_arctic_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_elt])
end

instantiation arctic_rat :: "show" begin
fun shows_arctic_rat :: "arctic_rat \<Rightarrow> shows"
  where "shows_arctic_rat (Num_arc_rat i) = shows i"
  | "shows_arctic_rat(MinInfty_rat)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ari = shows_arctic_rat ari"
definition "shows_list (as::arctic_rat list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic_rat: "shows_prec d (a::arctic_rat) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_rat_def)
lemma assoc_arctic_rat_elt: "\<forall>(a::arctic_rat)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic_rat by auto
instance 
  by (intro_classes,rule assoc_arctic_rat,unfold shows_list_arctic_rat_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_rat_elt])
end

definition int_name :: "string" where "int_name = ''naturals with negative constants''"
definition rat_name :: "string" where "rat_name = ''rationals (with negative constants)''"
definition arctic_name :: "string" where "arctic_name = ''arctic semiring''"
definition arctic_rat_name :: "string" where "arctic_rat_name = ''arctic rationals''"


definition mat_mono :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a mat \<Rightarrow> bool)"
where "mat_mono sd m \<equiv> mat_monoI m sd"

definition int_components :: "(int) poly_c"
where "int_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = max0,
  ge = op \<ge>,
  gt = op >,
  mono = True,
  description = int_name,
  arc_pos = (\<lambda> _ . True),
  check_mono = int_mono,
  max = max,
  derive_complexity = (\<lambda> x. if x \<le> 1 then Some 0 else None)
  \<rparr>"

definition rat_components :: "(rat) poly_c"
where "rat_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = max0,
  ge = op \<ge>,
  gt = op >,
  mono = True,
  description = rat_name,
  arc_pos = (\<lambda> _ . True),
  check_mono = rat_mono,
  max = max,
  derive_complexity = (\<lambda> x. if x \<le> 1 then Some 0 else None)
  \<rparr>"

abbreviation (input) no_complexity_lpoly where "no_complexity_lpoly \<equiv> \<lambda> _ . None :: (nat option)"

definition arctic_components :: "(arctic) poly_c"
where "arctic_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = gt_arctic,
  mono = False,
  description = arctic_name,
  arc_pos = pos_arctic,
  check_mono = (\<lambda> _. False),
  max = op +,
  derive_complexity = no_complexity_lpoly
  \<rparr>"

definition arctic_rat_components :: "(arctic_rat) poly_c"
where "arctic_rat_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = weak_gt_arctic_rat,
  mono = False,
  description = arctic_rat_name,
  arc_pos = pos_arctic_rat,
  check_mono = (\<lambda> _. False),
  max = op +,
  derive_complexity = no_complexity_lpoly
  \<rparr>"

lemma create_mono_poly_redtriple:
  fixes cI :: "('f :: show, 'a :: show)lpoly_interL \<Rightarrow> shows check"
  assumes carrier: "\<And> I as. isOK(cI I) \<Longrightarrow> \<exists> gt bound. 
    complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound \<and> complexity_cond c bound \<and> (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"
  shows "generic_redtriple_impl (create_poly_redtriple cI c)"
proof(rule create_poly_redtriple)
  fix I as
  assume "isOK(cI I)"
  from carrier[OF this, of as]
  obtain gt bound where orient: "\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b"
    and mono: "
    complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound"
    and cpx: "complexity_cond c bound" by auto
  show "\<exists> gt. linear_poly_order_carrier 
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> 
  (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and>
  (poly_c.mono c \<longrightarrow> (\<exists> bound.
    complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound \<and> complexity_cond c bound))"
    by (intro exI, rule conjI[OF _ conjI[OF orient]], insert cpx mono mono_linear_poly_order_carrier.axioms(1)[OF complexity_linear_poly_order_carrier.axioms(1)[OF mono]], auto)
qed


abbreviation int_redtriple :: "('f :: show,int)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "int_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) int_components"

abbreviation arctic_redtriple :: "('f :: show,arctic)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) arctic_components"

abbreviation rat_redtriple :: "('f :: show,rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "rat_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) rat_components"

abbreviation arctic_rat_redtriple :: "('f :: show,arctic_rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_rat_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) arctic_rat_components"

locale complexity_poly_interpretation = 
  fixes c :: "('a :: {show, bin_max_ordered_semiring_1})poly_c"
 assumes c: "c = \<lparr>poly_c.wf = (\<lambda>x :: 'a. True), 
  zero = 0 :: 'a,
  one = 1,
  default = poly_c.default c,
  plus = op +,
  times = op *,
  max0 = poly_c.max0 c,
  ge = ge,
  gt = poly_c.gt c,
  mono = True,
  description = poly_c.description c,
  arc_pos = (\<lambda> _ :: 'a. True),
  check_mono = poly_c.check_mono c,
  max = max,
  derive_complexity = poly_c.derive_complexity c\<rparr>"
  and carrier: "\<And> I as. \<exists> gt bound. (mono_matrix_carrier gt (poly_c.default c) bound (poly_c.check_mono c)
  \<and> complexity_linear_poly_order_carrier
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c)
  (poly_c.check_mono c) 
  (poly_c.max c)
  bound \<and> complexity_cond c bound) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"

sublocale complexity_poly_interpretation \<subseteq> generic_redtriple_impl "create_poly_redtriple something c"
  by (rule create_mono_poly_redtriple, insert carrier, blast)

lemma int_complexity_poly_carrier: defines "c \<equiv> int_components" 
  shows "mono_matrix_carrier (poly_c.gt c) (poly_c.default c) nat (poly_c.check_mono c) \<and> complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) (poly_c.gt c) (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) nat" 
proof -
  have mono: "mono_matrix_carrier (op >) 1 nat int_mono" ..
  have id: "(op \<succeq> :: int \<Rightarrow> int \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_int.simps by (intro ext, simp)
  from mono_matrix_carrier.complexity_linear_poly_order_carrier[OF mono]
  interpret complexity_linear_poly_order_carrier "(\<lambda>_. True)" 0 "op +" "op *" "(\<lambda>x y. y \<le> x)"
     "(\<lambda>gt y. y < gt)" True 1 1 max0 "(\<lambda>_. True)" int_mono max nat  
    unfolding id by  simp
  show ?thesis
    by (rule conjI, unfold c_def int_components_def poly_c.simps, unfold_locales)
qed

interpretation int_complexity_poly_carrier: complexity_poly_interpretation int_components
proof (unfold_locales, simp add: int_components_def, intro ext, simp add: ge_int.simps, rule exI[of _ "poly_c.gt int_components"], rule exI[of _ nat], rule conjI, insert int_complexity_poly_carrier, auto)
  assume "complexity_linear_poly_order_carrier (poly_c.wf int_components)
      (poly_c.zero int_components) (poly_c.plus int_components)
      (poly_c.times int_components) (poly_c.ge int_components)
      (poly_c.gt int_components) (poly_c.mono int_components)
      (poly_c.one int_components) (poly_c.default int_components)
      (poly_c.max0 int_components) (poly_c.arc_pos int_components)
      (poly_c.check_mono int_components) (poly_c.max int_components) nat"
  interpret complexity_linear_poly_order_carrier "poly_c.wf int_components"
      "poly_c.zero int_components" "poly_c.plus int_components"
      "poly_c.times int_components" "poly_c.ge int_components"
      "poly_c.gt int_components" "poly_c.mono int_components"
      "poly_c.one int_components" "poly_c.default int_components"
      "poly_c.max0 int_components" "poly_c.arc_pos int_components"
      "poly_c.check_mono int_components" "poly_c.max int_components" nat by fact
  show "complexity_cond int_components nat"  
    unfolding complexity_cond_def int_components_def poly_c.simps simp_thms
  proof (intro allI impI)
    fix bc :: int and bcv :: int and bd :: int
    let ?bound = "if bc \<le> 1 then Some (0 :: nat) else None"
    assume bc: "0 \<le> bc" and bcv: "0 \<le> bcv" and bd: "0 \<le> bd" and bound: "?bound \<noteq> None"
    from bound have bound: "bc \<le> 1" by (cases "bc \<le> 1", auto)
    hence the: "the ?bound = 0" by simp
    {
      fix i n
      have "pow i n = i ^ n"
      proof (induct n arbitrary: i)
        case 0
        show ?case unfolding pow.simps by (simp add: int_components_def)
      next
        case (Suc n)
        show ?case unfolding pow.simps unfolding Suc
          by (simp add: int_components_def)
      qed
    }
    hence pow: "\<And> (i :: int) n. linear_poly_order_carrier.pow op * 1 i n = i ^ n"
      unfolding int_components_def by simp
    let ?b = "nat"
    let ?bcv = "Suc (?b (bd * bcv))"
    show "\<exists> g. g \<in> O_of (Comp_Poly (the ?bound)) \<and> (\<forall> n. nat (bd * (linear_poly_order_carrier.pow op * 1 bc n * bcv)) \<le> g n)"
      unfolding the pow
      le
    proof (rule exI[of _ "\<lambda> n. ?bcv"], rule conjI, rule O_of_polyI)
      fix n
      show "?bcv \<le> 0 * n ^ 0 + ?bcv" by simp
    next
      show "\<forall> n. ?b (bd * (bc ^ n * bcv)) \<le> ?bcv"
      proof
        fix n
        have "?b (bd * (bc ^ n * bcv)) \<le> nat (1 + bd * bcv)"
        proof (rule nat_mono)
          have n: "0 \<le> n" by simp
          show "bd * (bc ^ n * bcv) \<le> 1 + bd * bcv"
            using mult_left_mono[OF mult_right_mono[OF power_decreasing[OF n bc bound] bcv], OF bd]
            by auto
        qed
        thus "?b (bd * (bc ^ n * bcv)) \<le> ?bcv" by simp
      qed
    qed
  qed
qed

lemma int_redtriple: "generic_redtriple_impl int_redtriple" ..

lemma rat_complexity_poly_carrier: defines "c \<equiv> rat_components" 
  shows "\<exists> gt bound. (mono_matrix_carrier gt (poly_c.default c) bound (poly_c.check_mono c) \<and> complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound \<and> complexity_cond c bound) \<and> (\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)" 
proof -
  let ?ab = "filter (\<lambda> (a,b). a > b) ab"
  have "\<And> a b. (a,b) \<in> set ((1,0) # ?ab) \<Longrightarrow> a > b" by auto
  from rat_minimal_delta[OF this]
  obtain d' where d0: "d' > 0" and orient: "\<forall> a b. (a,b) \<in> set ((1,0) # ?ab) \<longrightarrow> rat_gt d' a b" by blast+
  hence orient: "\<forall> a b. (a,b) \<in> set (?ab) \<longrightarrow> rat_gt d' a b" 
    and d1: "rat_gt d' 1 0" by auto
  from d1 have d1: "d' \<le> 1" unfolding rat_gt_def by auto
  from ex_le_of_nat[of "1 / d'"] obtain dd where le: "1 / d' \<le> of_nat dd" by auto
  with d0 have dd: "dd > 0" by (cases dd, auto)
  obtain d :: rat where d: "d = 1 / of_nat dd" by auto
  from le have "1 / d' \<le> 1 / d" unfolding d by simp
  hence "1 / (1 / d') \<ge> 1 / (1 / d)"
    by (rule divide_left_mono[OF _ _ mult_pos_pos[OF divide_pos_pos divide_pos_pos]], insert d0 d dd, auto)
  hence dd': "d' \<ge> d" by auto
  from d dd have d0: "d > 0" by simp
  from d1 dd' have d1: "d \<le> 1" by simp
  from orient dd' have orient: "\<forall> (a,b) \<in> set ab. b < a \<longrightarrow> rat_gt d a b"
    unfolding rat_gt_def by force
  let ?gt = "rat_gt d"
  let ?nat = "\<lambda> x. nat (ceiling (x / d))"
  have id: "(op \<succeq> :: rat \<Rightarrow> rat \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_int.simps by (intro ext, simp)
  from d have "\<exists> n. d = 1 / of_nat n" by auto
  note mono_mat = rat_complexity[OF d0 d1 this]
  then interpret mono_matrix_carrier ?gt 1 ?nat rat_mono .
  from mono_matrix_carrier.complexity_linear_poly_order_carrier[OF mono_mat]
  have carr: "complexity_linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) (\<lambda> x y. x \<ge> y) ?gt True 1 1 max0 (\<lambda>x. True) rat_mono max  ?nat" unfolding id .
  then interpret complexity_linear_poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" "\<lambda> x y. x \<ge> y" ?gt True 1 1 max0 "\<lambda>x. True" rat_mono max ?nat .
  have rat_complexity_cond: "complexity_cond rat_components ?nat" 
    unfolding complexity_cond_def rat_components_def poly_c.simps simp_thms
  proof (intro allI impI)
    fix bc :: rat and bcv :: rat and bd :: rat
    let ?bound = "if bc \<le> 1 then Some (0 :: nat) else None"
    assume bc: "0 \<le> bc" and bcv: "0 \<le> bcv" and bd: "0 \<le> bd" and bound: "?bound \<noteq> None"
    from bound have bound: "bc \<le> 1" by (cases "bc \<le> 1", auto)
    hence the: "the ?bound = 0" by simp
    {
      fix i n
      have "pow i n = i ^ n"
        by (induct n arbitrary: i, auto)
    } note pow = this
    let ?b = "?nat"
    let ?bcv = "?b (1 + bd * bcv)"
    show "\<exists> g. g \<in> O_of (Comp_Poly (the ?bound)) \<and> (\<forall> n. ?nat (bd * (linear_poly_order_carrier.pow op * 1 bc n * bcv)) \<le> g n)"
      unfolding the pow
      le
    proof (rule exI[of _ "\<lambda> n. ?bcv"], rule conjI, rule O_of_polyI)
      fix n
      show "?bcv \<le> 0 * n ^ 0 + ?bcv" by simp
    next
      show "\<forall> n. ?b (bd * (bc ^ n * bcv)) \<le> ?bcv"
      proof
        fix n
        show "?b (bd * (bc ^ n * bcv)) \<le> ?bcv"
        proof (rule nat_mono, rule ceiling_mono, rule divide_right_mono)
          have n: "0 \<le> n" by simp
          show "bd * (bc ^ n * bcv) \<le> 1 + bd * bcv"
            using mult_left_mono[OF mult_right_mono[OF power_decreasing[OF n bc bound] bcv], OF bd] by auto
        qed (insert d0, auto)
      qed
    qed
  qed    
  show ?thesis using rat_complexity_cond unfolding c_def rat_components_def poly_c.simps simp_thms
    by (intro exI[of _ ?gt] exI[of _ ?nat], insert mono_mat carr orient, blast)
qed

interpretation rat_complexity_poly_carrier: complexity_poly_interpretation rat_components
  by (unfold_locales, simp add: rat_components_def, intro ext, simp add: ge_rat.simps, rule rat_complexity_poly_carrier)

lemma rat_redtriple: "generic_redtriple_impl rat_redtriple" ..

locale linear_poly_interpretation = 
  fixes c :: "('a :: {show, max_ordered_semiring_1}) poly_c"
 assumes c: "c = \<lparr>poly_c.wf = (\<lambda>x. True), 
  zero = 0,
  one = 1,
  default = poly_c.default c,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = poly_c.gt c,
  mono = False,
  description = poly_c.description c,
  arc_pos = poly_c.arc_pos c, 
  check_mono = (\<lambda> _. False),
  max = op +,
  derive_complexity = poly_c.derive_complexity c\<rparr>"
  and zero_leastIII: "\<And> x :: 'a. ge x 0"
  and carrier: "\<And> I as. \<exists> gt. linear_poly_order_carrier
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"

lemma (in linear_poly_interpretation) c_mono: "(a \<and> (poly_c.mono c \<longrightarrow> b)) = a"
proof -
  have "poly_c.mono c = False" using arg_cong[OF c, of "poly_c.mono"] by simp
  thus ?thesis by simp
qed

sublocale linear_poly_interpretation \<subseteq> generic_redtriple_impl "create_poly_redtriple something c"
  by (rule create_poly_redtriple, unfold c_mono, rule carrier)

interpretation arctic_SN: both_mono_matrix_carrier pos_arctic gt_arctic 1 ..
  
lemma arctic_linear_poly_carrier: defines "c \<equiv> arctic_components" 
  shows "linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) (poly_c.gt c) (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c)"
proof -
  from arctic_SN.linear_poly_order_carrier
  show ?thesis unfolding c_def arctic_components_def poly_c.simps .
qed

interpretation arctic_linear_poly_carrier: linear_poly_interpretation arctic_components
  by (unfold_locales, simp add: arctic_components_def, simp add: zero_arctic_def,
  rule exI[of _ "poly_c.gt arctic_components"], rule conjI[OF arctic_linear_poly_carrier], auto) 


lemma arctic_rat_linear_poly_carrier: defines "c \<equiv> arctic_rat_components" 
  shows "\<exists> gt. linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)" 
proof -
  let ?ab = "filter (\<lambda> (a,b). weak_gt_arctic_rat a b) ab"
  have "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> weak_gt_arctic_rat a b" by auto
  from weak_arctic_rat_SN.weak_gt_both_mono[OF this]
  obtain gt where mono: "SN_both_mono_ordered_semiring_1 1 gt pos_arctic_rat"
    and orient: "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> gt a b" by auto
  from orient have orient: "\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> weak_gt_arctic_rat a b \<longrightarrow> gt a b" by auto
  from mono interpret SN_both_mono_ordered_semiring_1 1 gt pos_arctic_rat .
  interpret both_mono_matrix_carrier pos_arctic_rat gt 1 ..
  from linear_poly_order_carrier
  have  carrier: "linear_poly_order_carrier (\<lambda>_. True) 0 op + op * op \<succeq>
     gt False 1 1 id pos_arctic_rat" .
  thus ?thesis using orient unfolding c_def arctic_rat_components_def poly_c.simps by auto
qed

interpretation arctic_rat_linear_poly_carrier: linear_poly_interpretation arctic_rat_components
  by (unfold_locales, simp add: arctic_rat_components_def, simp add: zero_arctic_rat_def,
  rule arctic_rat_linear_poly_carrier)

lemma arctic_rat_redtriple: "generic_redtriple_impl arctic_rat_redtriple" ..

definition mat_complexity :: "nat \<Rightarrow> ('a :: bin_max_ordered_semiring_1)mat \<Rightarrow> nat option"
  where "mat_complexity n M \<equiv>
     if upper_triangular M then Some (n - 1) else None"

definition mat_components :: "nat \<Rightarrow> nat \<Rightarrow> ('a :: bin_max_ordered_semiring_1) poly_c \<Rightarrow> ('a mat) poly_c"
where "mat_components n sd c \<equiv> 
  \<lparr> poly_c.wf = mat n n, 
    zero = mat0I 0 n n, 
    one = mat1I 0 1 n, 
    default = mat1I 0 (poly_c.default c) n, 
    plus = mat_plusI (op +), 
    times = mat_multI 0 (op +) (op *) n,
    max0 = mat_max0, 
    ge = mat_ge, 
    gt = mat_gtI (poly_c.gt c) sd, 
    mono = poly_c.mono c, 
    description = concat [to_string n, ''x'',to_string n,'' matrices with strict dimension '',to_string sd,'' over '',poly_c.description c], 
    arc_pos = (\<lambda>x. True), 
    check_mono = mat_mono sd (poly_c.check_mono c),
    max = mat_max,
    derive_complexity = mat_complexity n \<rparr>"

abbreviation mat_check :: "nat \<Rightarrow> nat \<Rightarrow> shows check"
where "mat_check n sd \<equiv> check (sd > 0 \<and> sd \<le> n) (shows_string ''invalid dimension'')"

lemma (in complexity_poly_interpretation) id_c: 
    "poly_c.wf c = (\<lambda>x. True) \<and> poly_c.zero c = 0 \<and> poly_c.one c = 1 \<and> poly_c.arc_pos c = (\<lambda>x. True) \<and> poly_c.plus c = op +
     \<and> poly_c.times c = op * \<and> poly_c.ge c = ge \<and> poly_c.mono c = True \<and> poly_c.max c = max" 
  using c[symmetric] by (cases c, auto)

  
  
lemma (in complexity_poly_interpretation) mat_poly_order_carrier: 
  fixes as :: "('a mat \<times> 'a mat) list"
  assumes sdpos: "0 < sd" and sd_n: "sd \<le> n" and mc: "mc = mat_components n sd c"
  shows "\<exists>gt. linear_poly_order_carrier (mat n n) (mat0 n n)
                   mat_plus (mat_mult n) mat_ge gt (poly_c.mono c) (mat1 n)
                   (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True) \<and>
                  (\<forall>(a, b)\<in>set as.
                      mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_gtI (poly_c.gt c) sd a b \<longrightarrow> gt a b) \<and>
                   (\<exists> bound. complexity_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus
                    (mat_mult n) mat_ge gt True (mat1 n)
                    (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True)
                    (mat_mono sd (poly_c.check_mono c)) mat_max bound \<and> complexity_cond mc bound)"  
proof -
  interpret weak_complexity_linear_poly_order_carrier "poly_c.gt c" "poly_c.default c" "poly_c.check_mono c"
  proof
    fix as :: "('a \<times> 'a)list"
    assume orient1: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> poly_c.gt c a b"
    from carrier[of as, simplified id_c] obtain gt bound
      where mono: "complexity_linear_poly_order_carrier (\<lambda> _. True) 0 (op +) (op *) (op \<succeq>) gt True 1 (poly_c.default c) (poly_c.max0 c) (\<lambda> _. True) (poly_c.check_mono c) max bound" and orient2: "\<forall> (a,b) \<in> set as. poly_c.gt c a b \<longrightarrow> gt a b" 
    and mono2: "mono_matrix_carrier gt (poly_c.default c) bound (poly_c.check_mono c)" by blast
    from orient1 orient2 have orient: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> gt a b" (is "?orient gt") by auto
    show "\<exists> gt bound. mono_matrix_carrier gt (poly_c.default c) bound (poly_c.check_mono c) \<and> ?orient gt"
      by (intro exI, rule conjI[OF mono2 orient])
  qed
  let ?as = "filter (\<lambda> (a,b). mat n n a \<and> mat n n b \<and> mat_gtI (poly_c.gt c) sd a b) as"
  have as: "\<And> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<Longrightarrow> mat_gtI (poly_c.gt c) sd a b" by force
  from weak_mat_gt_mono[OF sd_n, of ?as, OF this]
  obtain gt bound where carrier: "mono_matrix_carrier gt (poly_c.default c) bound (poly_c.check_mono c)"
    and "\<forall> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<longrightarrow> mat_gtI gt sd a b" by blast
  hence orient: "\<forall> (a,b) \<in> set as. mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_gtI (poly_c.gt c) sd a b \<longrightarrow> mat_gtI gt sd a b" by auto
  interpret mono_matrix_carrier gt "poly_c.default c" bound "poly_c.check_mono c" by fact
  note comp = mat_complexity_linear_poly_order_carrier[OF sd_n sdpos]
  interpret bar: complexity_linear_poly_order_carrier "(mat n n)" "(mat0 n n)" mat_plus "(mat_mult n)" mat_ge "(mat_gtI gt sd)" True "(mat1 n)" "(mat1I 0 (poly_c.default c) n)" mat_max0 "(\<lambda>x. True)" "(mat_monoI (poly_c.check_mono c) sd)" mat_max "\<lambda> a. bound (linear_norm a)" by (rule comp)
  let ?wf = "poly_c.wf mc"
  let ?ge = "poly_c.ge mc"
  let ?ze = "poly_c.zero mc"
  let ?on = "poly_c.one mc"
  let ?ap = "poly_c.arc_pos mc"
  let ?ti = "poly_c.times mc"
  let ?cpx = "poly_c.derive_complexity mc"
  let ?pow = "linear_poly_order_carrier.pow ?ti ?on"
  show ?thesis unfolding mat_mono_def
  proof (intro exI, rule conjI[OF _ conjI[OF orient exI[of _ "(\<lambda> a. bound (linear_norm a))"]]], simp add: id_c, unfold_locales,
    rule conjI[OF comp])
    show "complexity_cond mc (\<lambda> a. bound (linear_norm a))"
      unfolding complexity_cond_def
    proof (intro allI impI)
      fix M N K
      assume wfM: "?wf M" and wfN: "?wf N" and wfK: "?wf K" and M0: "?ge M ?ze" and N0: "?ge N ?ze" and K0: "?ge K ?ze"
      and cpx: "?cpx M \<noteq> None"
      have lpc: "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gtI gt sd) True (mat1 n) (mat1I 0 (poly_c.default c) n) mat_max0 (\<lambda>x. True)" ..
      from cpx have cpx: "mat_complexity n M \<noteq> None"
        by (simp add: mc mat_components_def)
      have wf: "?wf = mat n n" unfolding mc mat_components_def by simp
      have ge: "?ge = mat_ge" unfolding mc mat_components_def by simp
      have ze: "?ze = mat0 n n" unfolding mc mat_components_def by simp
      note wM = wfM[unfolded wf] 
      note wN = wfN[unfolded wf]
      note wK = wfK[unfolded wf]
      note M0 = M0[unfolded ge ze]
      note N0 = N0[unfolded ge ze]
      note K0 = K0[unfolded ge ze]
      have lpc: "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gtI gt sd) True (mat1 n) (mat1I 0 (poly_c.default c) n) mat_max0 (\<lambda>x. True)" ..
      from cpx have tri: "upper_triangular M" and cpx: "mat_complexity n M = Some (n - 1)"
        unfolding mat_complexity_def Let_def by (auto split: if_splits)
      from upper_triangular_mat_pow_bound[OF wM tri M0]
      obtain c where bound: "\<And> nn. nn > 0 \<Longrightarrow> bound (linear_norm (mat_pow n M nn)) \<le> c * nn ^ (n - Suc 0)" by blast
      note pow_simps = linear_poly_order_carrier.pow.simps[OF lpc]      
      {
        fix nn and M :: "'a mat"
        have "linear_poly_order_carrier.pow (mat_mult n) (mat1 n) M nn = mat_pow n M nn"
          by (induct nn, auto simp: pow_simps)            
      } note pow_conv = this
      let ?goal1 = "\<exists>g. g \<in> O_of (Comp_Poly (the (?cpx M))) \<and>
              (\<forall>n. bound (linear_norm (?ti K (?ti (?pow M n) N))) \<le> g n)"
      let ?goal2 = " \<exists>g. 
                  g \<in> O_of (Comp_Poly (n - 1)) \<and>
                  (\<forall>na. bound (linear_norm (mat_mult n K (mat_mult n (mat_pow n M na) N))) \<le> g na)"
      have goal: "?goal1 = ?goal2" by (simp add: mc mat_components_def pow_conv cpx)
      from linear_norm_mult_left_ex[OF wN N0] 
      obtain d where mult: "\<And> m'. mat n n m' \<Longrightarrow> mat_ge m' (mat0 n n) \<Longrightarrow> linear_norm m' * of_nat d \<succeq> linear_norm (mat_mult n m' N)"
        by blast
      from linear_norm_mult_right_ex[OF wK K0] 
      obtain e where mult': "\<And> m'. mat n n m' \<Longrightarrow> mat_ge m' (mat0 n n) \<Longrightarrow> linear_norm m' * of_nat e \<succeq> linear_norm (mat_mult n K m')"
        by blast
      show  ?goal1 unfolding goal
      proof (intro exI conjI)
        let ?b = "\<lambda> nn. bound (linear_norm (mat_mult n K (mat_mult n (mat_pow n M nn) N)))"
        {
          fix nn :: nat
          assume nn: "nn > 0"
          from mat_pow_ge_zero[OF wM M0]
          have pow0: "mat_ge (mat_pow n M nn) (mat0 n n)" . 
          from mat_mult_right_mono[OF mat_pow[OF wM] wN mat0 pow0 N0]
          have prod0: "mat_ge (mat_mult n (mat_pow n M nn) N) (mat0 n n)"
            unfolding mat0_mult_right[OF mat_pow[OF wM]] .
          from mat_mult[OF mat_pow[OF wM] wN] 
          have prod: "mat n n (mat_mult n (mat_pow n M nn) N)" .
          have "?b nn
            \<le> bound (linear_norm (mat_mult n (mat_pow n M nn) N) * of_nat e)"
            by (rule bound_mono[OF mult'[OF prod prod0]])
          also have "... \<le> bound (linear_norm (mat_mult n (mat_pow n M nn) N)) * bound (of_nat e)"
            by (rule bound_mult_of_nat)
          also have "bound (linear_norm (mat_mult n (mat_pow n M nn) N)) \<le> bound (linear_norm (mat_pow n M nn) * of_nat d)"
            by (rule bound_mono[OF mult[OF mat_pow[OF wM] pow0]])
          also have "... \<le> bound (linear_norm (mat_pow n M nn)) * bound (of_nat d)"
            by (rule bound_mult_of_nat)
          also have "... \<le> c * nn ^ (n - Suc 0) * bound (of_nat d)"
            by (rule mult_le_mono1[OF bound[OF nn]])
          finally have "?b nn \<le> (c * bound (of_nat d) * bound (of_nat e)) * nn ^ (n - Suc 0)"
            by (auto simp: field_simps)
        } note main = this
        let ?gp = "\<lambda> nn. c * bound (of_nat d) * bound (of_nat e) * nn ^ (n - Suc 0)"
        let ?g = "\<lambda> nn. ?gp nn + ?b 0"
        {
          fix n
          have "?b n \<le> ?g n"
          proof (cases n)
            case 0
            show ?thesis unfolding 0 by simp
          next
            case (Suc m)
            hence "n > 0" by simp
            from main[OF this] show ?thesis by simp
          qed
        } note main = this
        thus "\<forall> nn. ?b nn \<le> ?g nn" by auto
        show "?g \<in> O_of (Comp_Poly (n - 1))"
          by (rule O_of_polyI[of _ "c * bound (of_nat d) * bound (of_nat e)" _ "?b 0"], simp)
      qed
    qed
  qed
qed

sublocale complexity_poly_interpretation \<subseteq> mono_matrix: generic_redtriple_impl "create_poly_redtriple (\<lambda>I. mat_check n sd) (mat_components n sd c)"
proof -
  let ?c = "mat_components n sd c"
  show "generic_redtriple_impl (create_poly_redtriple (\<lambda> I. mat_check n sd) ?c)"
  proof (rule create_poly_redtriple)
    fix I as 
    assume "isOK (mat_check n sd)"
    hence "0 < sd" "sd \<le> n" by auto
    from mat_poly_order_carrier[OF this refl, of as]
    obtain gt where gt: "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge gt (poly_c.mono c) (mat1 n)
         (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True) \<and>
        (\<forall>(a, b)\<in>set as. mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_gtI (poly_c.gt c) sd a b \<longrightarrow> gt a b) \<and>
        (\<exists>bound. complexity_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge gt True (mat1 n)
                  (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True) (mat_mono sd (poly_c.check_mono c)) mat_max bound \<and>
                 complexity_cond ?c bound)" ..
    show "\<exists> gt. linear_poly_order_carrier 
  (poly_c.wf ?c)
  (poly_c.zero ?c)
  (poly_c.plus ?c)
  (poly_c.times ?c)
  (poly_c.ge ?c)
  gt
  (poly_c.mono ?c)
  (poly_c.one ?c)
  (poly_c.default ?c)
  (poly_c.max0 ?c)
  (poly_c.arc_pos ?c) \<and> 
  (\<forall>(a, b)\<in>set as. poly_c.wf ?c a \<longrightarrow> poly_c.wf ?c b \<longrightarrow> poly_c.gt ?c a b \<longrightarrow> gt a b) \<and>
  (poly_c.mono ?c \<longrightarrow>
    (\<exists> bound. complexity_linear_poly_order_carrier (poly_c.wf ?c) (poly_c.zero ?c) (poly_c.plus ?c) (poly_c.times ?c) (poly_c.ge ?c) gt (poly_c.mono ?c) (poly_c.one ?c) (poly_c.default ?c) (poly_c.max0 ?c) (poly_c.arc_pos ?c) (poly_c.check_mono ?c) (poly_c.max ?c) bound \<and> complexity_cond ?c bound))" 
      by (rule exI[of _ gt], insert gt, simp add: mat_components_def, auto)
  qed
qed

abbreviation int_mat_redtriple :: "nat \<Rightarrow> nat \<Rightarrow> ('f :: show,int mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "int_mat_redtriple n sd \<equiv> create_poly_redtriple (\<lambda>i. mat_check n sd) (mat_components n sd int_components)"

abbreviation rat_mat_redtriple :: "nat \<Rightarrow> nat \<Rightarrow> ('f :: show,rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "rat_mat_redtriple n sd \<equiv> create_poly_redtriple (\<lambda>i. mat_check n sd) (mat_components n sd rat_components)"

lemma int_mat_redtriple: "generic_redtriple_impl (int_mat_redtriple n sd)" ..
lemma rat_mat_redtriple: "generic_redtriple_impl (rat_mat_redtriple n sd)" ..


abbreviation mat_arctic_components :: "nat \<Rightarrow> ('a :: max_ordered_semiring_1) poly_c \<Rightarrow> ('a mat) poly_c"
where "mat_arctic_components n c \<equiv> 
  \<lparr> poly_c.wf = mat n n, 
    zero = mat0I 0 n n, 
    one = mat1I 0 1 n, 
    default = mat1I 0 (poly_c.default c) n, 
    plus = mat_plusI (op +), 
    times = mat_multI 0 (op +) (op *) n,
    max0 = id, 
    ge = mat_ge, 
    gt = mat_comp_all (poly_c.gt c), 
    mono = False, 
    description = concat [to_string n, ''x'',to_string n,'' matrices over '',poly_c.description c], 
    arc_pos = mat_arc_posI (poly_c.arc_pos c),
    check_mono = (\<lambda> _. False),
    max = mat_plusI (op +),
    derive_complexity = no_complexity_lpoly \<rparr>"

abbreviation mat_arctic_check :: "nat \<Rightarrow> shows check"
where "mat_arctic_check n \<equiv> check (n > 0) (shows_string ''invalid dimension'')"

context weak_SN_both_mono_ordered_semiring_1
begin

lemma weak_mat_gt_both_mono: assumes orient: "\<And> m1 m2. mat n n m1 \<and> mat n n m2 \<and> (m1,m2) \<in> set m12s \<Longrightarrow> weak_mat_gt_arc m1 m2"
   shows "\<exists> gt. SN_both_mono_ordered_semiring_1 default gt arc_pos \<and> (\<forall> m1 m2. mat n n m1 \<and> mat n n m2 \<and> (m1, m2) \<in> set m12s \<longrightarrow> mat_comp_all gt m1 m2)"
proof -
  let ?m1x = "concat (concat (map fst m12s))"
  let ?m2y = "concat (concat (map snd m12s))"
  let ?pairs = "concat (map (\<lambda> x. map (\<lambda> y. (x,y)) ?m2y) ?m1x)"
  let ?strict = "filter (\<lambda> (x,y). weak_gt x y) ?pairs"
  have "\<forall> x y. (x,y) \<in> set ?strict \<longrightarrow> weak_gt x y" by auto
  from weak_gt_both_mono[OF this] obtain gt where order: "SN_both_mono_ordered_semiring_1 default gt arc_pos" 
    and orient2: "\<And> x y. (x, y) \<in> set ?strict \<Longrightarrow> gt x y" by auto
  show ?thesis
  proof (intro exI allI conjI impI, rule order)
    fix m1 m2
    assume ass: "mat n n m1 \<and> mat n n m2 \<and> (m1,m2) \<in> set m12s"
    hence wf1: "mat n n m1" and wf2: "mat n n m2" and m1m2: "(m1,m2) \<in> set m12s" by auto
    show "mat_comp_all gt  m1 m2" unfolding mat_comp_all_index[OF wf1 wf2]
    proof (intro allI impI)
      fix i j
      assume i: "i < n" and j: "j < n"
      with orient[OF ass, unfolded mat_comp_all_index[OF wf1 wf2]] have weak_gt: "weak_gt (m1 ! i ! j) (m2 ! i ! j)" (is "weak_gt ?x ?y") by auto  
      from wf1 i j have m1ij: "m1 ! i \<in> set m1 \<and> m1 ! i ! j \<in> set (m1 ! i)" unfolding mat_def  by (auto simp: vec_def)
      from wf2 i j have m2ij: "m2 ! i \<in> set m2 \<and> m2 ! i ! j \<in> set (m2 ! i)" unfolding mat_def  by (auto simp: vec_def)
      have x: "?x \<in> set ?m1x" by (simp, rule bexI[OF _ m1m2], rule bexI[of _ "m1 ! i"], auto simp: m1ij)
      have y: "?y \<in> set ?m2y" by (simp, rule bexI[OF _ m1m2], rule bexI[of _ "m2 ! i"], auto simp: m2ij)
      from x y have "(?x,?y) \<in> set ?pairs" by force
      with weak_gt have gt: "(?x,?y) \<in> set ?strict" by simp
      show "gt ?x ?y" by (rule orient2[OF gt])
    qed
  qed
qed
end


sublocale linear_poly_interpretation \<subseteq> matrix: generic_redtriple_impl "create_poly_redtriple (\<lambda>I. mat_arctic_check n) (mat_arctic_components n c)"
proof (rule create_poly_redtriple, simp_all)
  fix as
  assume n_pos: "0 < n"
  let ?goal = "\<exists>gta. linear_poly_order_carrier (mat n n) (mat0 n n)
                mat_plus (mat_mult n) mat_ge gta False (mat1 n)
                (mat1I (0\<Colon>'a) (poly_c.default c) n) id
                (mat_arc_posI (poly_c.arc_pos c)) \<and>
               (\<forall>(a, b)\<in>set as.
                   mat n n a \<longrightarrow>
                   mat n n b \<longrightarrow> mat_comp_all (poly_c.gt c) a b \<longrightarrow> gta a b)"
  from c[symmetric] have id: 
    "poly_c.wf c = (\<lambda>x. True) \<and> poly_c.zero c = 0 \<and> poly_c.one c = 1 \<and> poly_c.max0 c = id \<and> poly_c.plus c = op +
     \<and> poly_c.times c = op * \<and> poly_c.ge c = ge \<and> poly_c.mono c = False" by (cases c, auto)
  interpret weak_SN_both_mono_ordered_semiring_1 "poly_c.gt c" "poly_c.default c" "poly_c.arc_pos c"
  proof
    fix as :: "('a \<times> 'a)list"
    assume orient1: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> poly_c.gt c a b"
    from carrier[of as, simplified id] obtain gt 
      where mono: "linear_poly_order_carrier (\<lambda> _. True) 0 (op +) (op *) (op \<succeq>) gt False 1 (poly_c.default c) id (poly_c.arc_pos c)" and orient2: "\<forall> (a,b) \<in> set as. poly_c.gt c a b \<longrightarrow> gt a b" by blast
    from orient1 orient2 have orient: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> gt a b" (is "?orient gt") by auto
    {
      interpret linear_poly_order_carrier "\<lambda> _. True" 0 "op +" "op *" "op \<succeq>" gt False 1 "poly_c.default c" id "poly_c.arc_pos c" by fact
      have SN_both: "SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c)" 
        by (unfold_locales,
        (rule compat, simp+)[1],
        (rule compat2, simp+)[1],
        (rule gt_trans, simp+)[1],
          rule default_geq_zero,
        (rule plus_gt_both_mono, simp+)[1],
        (rule times_gt_left_mono, simp+)[1],
        (rule times_gt_right_mono, simp+)[1],
        (rule zero_leastI, simp+)[1],
        (rule zero_leastII, simp+)[1],
          rule zero_leastIII, 
          rule arc_pos_one,
          rule arc_pos_default,
          (rule arc_pos_zero, simp)[1],
          rule arc_pos_plus[simplified], simp+,
          rule arc_pos_mult[simplified], simp+,
          insert not_all_ge SN[simplified zero_leastIII, simplified], auto) 
    } note SN_both = this
    thus "\<exists> gt. SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c) \<and> ?orient gt"
      using orient by blast
  qed
  let ?as = "filter (\<lambda> (a,b). mat n n a \<and> mat n n b \<and> mat_comp_all (poly_c.gt c) a b) as"
  have as: "\<And> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<Longrightarrow> mat_comp_all (poly_c.gt c) a b" by force
  from weak_mat_gt_both_mono[OF as]
  obtain gt where carrier: "SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c)"
    and "\<forall> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<longrightarrow> mat_comp_all gt a b" by blast
  hence orient: "\<forall> (a,b) \<in> set as. mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_comp_all (poly_c.gt c) a b \<longrightarrow> mat_comp_all gt a b" by auto
  interpret SN_both_mono_ordered_semiring_1 "poly_c.default c" gt "poly_c.arc_pos c" by fact
  interpret both_mono_matrix_carrier "poly_c.arc_pos c" gt "poly_c.default c" ..
  from mat_linear_poly_order_carrier[OF n_pos] 
  interpret linear_poly_order_carrier "(mat n n)" "(mat0 n n)" mat_plus "(mat_mult n)" mat_ge "(mat_comp_all gt)" False "(mat1 n)" "(mat1I 0 (poly_c.default c) n)" id "(mat_arc_posI (poly_c.arc_pos c))" .
  show ?goal
    by (rule exI[of _ "mat_comp_all gt"], intro conjI[OF _ orient], unfold_locales) 
qed
  
abbreviation arctic_mat_redtriple :: "nat \<Rightarrow> ('f :: show,arctic mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_mat_redtriple n \<equiv> create_poly_redtriple (\<lambda>i. mat_arctic_check n) (mat_arctic_components n arctic_components)"

abbreviation arctic_rat_mat_redtriple :: "nat \<Rightarrow> ('f :: show,arctic_rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_rat_mat_redtriple n \<equiv> create_poly_redtriple (\<lambda>i. mat_arctic_check n) (mat_arctic_components n arctic_rat_components)"


lemma arctic_mat_redtriple: "generic_redtriple_impl (arctic_mat_redtriple n)" ..
lemma arctic_rat_mat_redtriple: "generic_redtriple_impl (arctic_rat_mat_redtriple n)" ..


end
