{-# LANGUAGE EmptyDataDecls, RankNTypes, ScopedTypeVariables #-}

module Ceta where {

import qualified Natural;

data Nat = Zero_nat | Suc Nat;

data Numa = One | Bit0 Numa | Bit1 Numa;

data Color = R | B;

data Rbta a b = Emptya | Branch Color (Rbta a b) a b (Rbta a b);

class Orda a where {
  less_eq :: a -> a -> Bool;
  less :: a -> a -> Bool;
};

class (Orda a) => Preorder a where {
};

class (Preorder a) => Order a where {
};

class (Order a) => Linorder a where {
};

newtype (Linorder a) => Rbt a b = Rbt (Rbta a b);

newtype Rat = Frct (Integer, Integer);

data Set a = Set [a] | Coset [a];

plus_nat :: Nat -> Nat -> Nat;
plus_nat (Suc m) n = plus_nat m (Suc n);
plus_nat Zero_nat n = n;

abs_int :: Integer -> Integer;
abs_int i = (if i < 0 then negate i else i);

sgn_int :: Integer -> Integer;
sgn_int i = (if i == 0 then 0 else (if 0 < i then 1 else negate 1));

apsnd :: forall a b c. (a -> b) -> (c, a) -> (c, b);
apsnd f (x, y) = (x, f y);

divmod_int :: Integer -> Integer -> (Integer, Integer);
divmod_int k l =
  (if k == 0 then (0, 0)
    else (if l == 0 then (0, k)
           else apsnd (\ a -> sgn_int l * a)
                  (if sgn_int k == sgn_int l then divMod (abs k) (abs l)
                    else let {
                           (r, s) = divMod (abs k) (abs l);
                         } in (if s == 0 then (negate r, 0)
                                else (negate r - 1, abs_int l - s)))));

nat :: Integer -> Nat;
nat k =
  (if k <= 0 then Zero_nat
    else let {
           (l, j) = divmod_int k 2;
           n = nat l;
           la = plus_nat n n;
         } in (if j == 0 then la else Suc la));

data Term a b = Var b | Fun a [Term a b];

foldra :: forall a b. (a -> b -> b) -> [a] -> b -> b;
foldra f [] = id;
foldra f (x : xs) = f x . foldra f xs;

class Plus a where {
  plus :: a -> a -> a;
};

class Zero a where {
  zero :: a;
};

class (Plus a) => Semigroup_add a where {
};

class (Semigroup_add a, Zero a) => Monoid_add a where {
};

listsum :: forall a. (Monoid_add a) => [a] -> a;
listsum xs = foldra plus xs zero;

size_list :: forall a. [a] -> Nat;
size_list [] = Zero_nat;
size_list (a : list) = plus_nat (size_list list) (Suc Zero_nat);

instance Zero Nat where {
  zero = Zero_nat;
};

instance Plus Nat where {
  plus = plus_nat;
};

instance Semigroup_add Nat where {
};

instance Monoid_add Nat where {
};

weight :: forall a b. ((a, Nat) -> Nat) -> Nat -> Term a b -> Nat;
weight w w0 (Fun f ts) =
  plus_nat (w (f, size_list ts)) (listsum (map (weight w w0) ts));
weight w w0 (Var x) = w0;

less_nat :: Nat -> Nat -> Bool;
less_nat m (Suc n) = less_eq_nat m n;
less_nat n Zero_nat = False;

less_eq_nat :: Nat -> Nat -> Bool;
less_eq_nat (Suc m) n = less_nat m n;
less_eq_nat Zero_nat n = True;

extract :: forall a. (a -> Bool) -> [a] -> Maybe ([a], (a, [a]));
extract p [] = Nothing;
extract p (x : xs) =
  (if p x then Just ([], (x, xs))
    else (case extract p xs of {
           Nothing -> Nothing;
           Just (ys, (y, zs)) -> Just (x : ys, (y, zs));
         }));

ms_lesseq :: forall a. (Eq a) => [a] -> [a] -> Bool;
ms_lesseq [] uu = True;
ms_lesseq (x : xs) ys =
  (case extract (\ a -> x == a) ys of {
    Nothing -> False;
    Just (ys1, (_, ys2)) -> ms_lesseq xs (ys1 ++ ys2);
  });

folda :: forall a b c. (a -> b) -> (c -> [b] -> b) -> Term c a -> b;
folda var fun (Var x) = var x;
folda var fun (Fun f ts) = fun f (map (folda var fun) ts);

flipa :: forall a b c. (a -> b -> c) -> b -> a -> c;
flipa f = (\ x y -> f y x);

consta :: forall a b. a -> b -> a;
consta = (\ x _ -> x);

add_vars_term_list :: forall a b. Term a b -> [b] -> [b];
add_vars_term_list =
  folda (\ a b -> a : b) (consta (flipa (foldra (\ a b -> a . b)) id));

vars_term_list :: forall a b. Term a b -> [b];
vars_term_list t = add_vars_term_list t [];

lex_ext_unbounded ::
  forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
lex_ext_unbounded f [] [] = (False, True);
lex_ext_unbounded f (uu : uv) [] = (True, True);
lex_ext_unbounded f [] (uw : ux) = (False, False);
lex_ext_unbounded f (a : asa) (b : bs) =
  (case f a b of {
    (True, nstri) -> (True, True);
    (False, True) -> lex_ext_unbounded f asa bs;
    (False, False) -> (False, False);
  });

kbo ::
  forall a b.
    (Eq b) => ((a, Nat) -> Nat) ->
                Nat ->
                  ((a, Nat) -> Nat) ->
                    (a -> Bool) -> Term a b -> Term a b -> (Bool, Bool);
kbo w w0 prc least s t =
  (if ms_lesseq (vars_term_list t) (vars_term_list s) &&
        less_eq_nat (weight w w0 t) (weight w w0 s)
    then (if less_nat (weight w w0 t) (weight w w0 s) then (True, True)
           else (case s of {
                  Var _ ->
                    (False,
                      (case t of {
                        Var _ -> True;
                        Fun g ts -> null ts && least g;
                      }));
                  Fun f ss ->
                    (case t of {
                      Var _ -> (True, True);
                      Fun g ts ->
                        let {
                          pf = prc (f, size_list ss);
                          pg = prc (g, size_list ts);
                        } in (if less_nat pg pf then (True, True)
                               else (if less_eq_nat pg pf
                                      then lex_ext_unbounded
     (kbo w w0 prc least) ss ts
                                      else (False, False)));
                    });
                }))
    else (False, False));

hd :: forall a. [a] -> a;
hd (x : xs) = x;

tl :: forall a. [a] -> [a];
tl [] = [];
tl (x : xs) = xs;

data Order_tag = Lex | Mul;

bex :: forall a. Set a -> (a -> Bool) -> Bool;
bex (Set xs) p = any p xs;

the :: forall a. Maybe a -> a;
the (Just x) = x;

data Nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5
  | Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC
  | NibbleD | NibbleE | NibbleF;

newline :: Char;
newline = '\n';

data Xml = Xml [Char] [([Char], [Char])] [Xml] (Maybe [Char]);

is_none :: forall a. Maybe a -> Bool;
is_none (Just x) = False;
is_none Nothing = True;

class Showa a where {
  shows_prec :: Nat -> a -> [Char] -> [Char];
  shows_list :: [a] -> [Char] -> [Char];
};

shows_string :: [Char] -> [Char] -> [Char];
shows_string = (\ a b -> a ++ b);

shows_list_char :: [Char] -> [Char] -> [Char];
shows_list_char cs = shows_string cs;

shows_prec_char :: Nat -> Char -> [Char] -> [Char];
shows_prec_char d c = (\ a -> c : a);

instance Showa Char where {
  shows_prec = shows_prec_char;
  shows_list = shows_list_char;
};

replicatea :: forall a. Nat -> a -> [a];
replicatea Zero_nat x = [];
replicatea (Suc n) x = x : replicatea n x;

shows_map :: forall a. (a -> [Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_map s [] = id;
shows_map s (x : xs) = s x . shows_map s xs;

shows_prec_list :: forall a. (Showa a) => Nat -> [a] -> [Char] -> [Char];
shows_prec_list d l = shows_list l;

shows_attr :: ([Char], [Char]) -> [Char] -> [Char];
shows_attr av =
  shows_prec_list Zero_nat (fst av) .
    shows_string ['='] . shows_string ('\"' : snd av ++ ['\"']);

shows_attrs :: [([Char], [Char])] -> [Char] -> [Char];
shows_attrs asa = shows_map (\ a -> shows_string [' '] . shows_attr a) asa;

shows_XML_indent :: [Char] -> Nat -> Xml -> [Char] -> [Char];
shows_XML_indent ind i (Xml n a c t) =
  shows_string ind .
    shows_string ['<'] .
      shows_prec_list Zero_nat n .
        shows_attrs a .
          (if null c && is_none t then shows_string ['/', '>']
            else shows_string ['>'] .
                   (if not (null c)
                     then shows_map
                            (shows_XML_indent (ind ++ replicatea i ' ') i) c .
                            shows_string ind
                     else id) .
                     (if is_none t then id else shows_string (the t)) .
                       shows_string ['<', '/'] .
                         shows_prec_list Zero_nat n . shows_string ['>']);

shows_prec_xml :: Nat -> Xml -> [Char] -> [Char];
shows_prec_xml d xml = shows_XML_indent [newline] (nat 2) xml;

data Sum a b = Inl a | Inr b;

newtype Parse_result a b = Parseresult (Sum a b);

errora :: forall a b. a -> Parse_result a b;
errora x = Parseresult (Inl x);

returna :: forall a b. a -> Parse_result b a;
returna x = Parseresult (Inr x);

text :: [Char] -> Xml -> Parse_result [Char] [Char];
text tag (Xml n atts cs t) =
  (if n == tag && null atts && null cs && not (is_none t) && not (null (the t))
    then returna (the t)
    else errora
           (concat
             [['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'x', 't',
                'r', 'a', 'c', 't', ' ', 't', 'e', 'x', 't', ' ', 'f', 'o', 'r',
                ' '],
               tag, [' ', 'f', 'r', 'o', 'm', ' '], [newline],
               shows_prec_xml Zero_nat (Xml n atts cs t) []]));

takea :: forall a. Nat -> [a] -> [a];
takea n [] = [];
takea n (x : xs) = (case n of {
                     Zero_nat -> [];
                     Suc m -> x : takea m xs;
                   });

one_nat :: Nat;
one_nat = Suc Zero_nat;

bindc ::
  forall a b c. Parse_result a b -> (b -> Parse_result a c) -> Parse_result a c;
bindc (Parseresult a) f =
  (case a of {
    Inl b -> Parseresult (Inl b);
    Inr aa -> f aa;
  });

digit2int :: Char -> Parse_result [Char] Integer;
digit2int x =
  (if x == '0' then returna 0
    else (if x == '1' then returna 1
           else (if x == '2' then returna 2
                  else (if x == '3' then returna 3
                         else (if x == '4' then returna 4
                                else (if x == '5' then returna 5
                                       else (if x == '6' then returna 6
      else (if x == '7' then returna 7
             else (if x == '8' then returna 8
                    else (if x == '9' then returna 9
                           else errora
                                  ([x] ++
                                    [' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                      'a', ' ', 'd', 'i', 'g', 'i',
                                      't'])))))))))));

string2int_aux :: Integer -> [Char] -> Parse_result [Char] Integer;
string2int_aux n [] = returna n;
string2int_aux n (d : s) =
  bindc (digit2int d) (\ m -> string2int_aux (10 * n + m) s);

string2int :: [Char] -> Parse_result [Char] Integer;
string2int s =
  (if null s
    then errora
           ['c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v', 'e', 'r',
             't', ' ', 'e', 'm', 'p', 't', 'y', ' ', 's', 't', 'r', 'i', 'n',
             'g', ' ', 'i', 'n', 't', 'o', ' ', 'n', 'u', 'm', 'b', 'e', 'r']
    else (if takea one_nat s == ['-']
           then bindc (string2int_aux 0 (tl s)) (\ i -> returna (0 - i))
           else string2int_aux 0 s));

int :: [Char] -> Xml -> Parse_result [Char] Integer;
int tag x = bindc (text tag x) string2int;

nata :: [Char] -> Xml -> Parse_result [Char] Nat;
nata tag x =
  bindc (text tag x) (\ txt -> bindc (string2int txt) (\ i -> returna (nat i)));

one_int :: Integer;
one_int = 1;

of_int :: Integer -> Rat;
of_int a = Frct (a, one_int);

mod_int :: Integer -> Integer -> Integer;
mod_int a b = snd (divmod_int a b);

gcd_int :: Integer -> Integer -> Integer;
gcd_int k l =
  abs_int (if l == 0 then k else gcd_int l (mod_int (abs_int k) (abs_int l)));

div_int :: Integer -> Integer -> Integer;
div_int a b = fst (divmod_int a b);

normalize :: (Integer, Integer) -> (Integer, Integer);
normalize p =
  (if 0 < snd p
    then let {
           a = gcd_int (fst p) (snd p);
         } in (div_int (fst p) a, div_int (snd p) a)
    else (if snd p == 0 then (0, 1)
           else let {
                  a = negate (gcd_int (fst p) (snd p));
                } in (div_int (fst p) a, div_int (snd p) a)));

quotient_of :: Rat -> (Integer, Integer);
quotient_of (Frct x) = x;

divide_rat :: Rat -> Rat -> Rat;
divide_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * d, c * b));

xml_change ::
  forall a b.
    (Xml -> Parse_result [Char] a) -> (a -> b) -> Xml -> Parse_result [Char] b;
xml_change p f x = bindc (p x) (\ a -> returna (f a));

map_of :: forall a b. (Eq a) => [(a, b)] -> a -> Maybe b;
map_of ((l, v) : ps) k = (if l == k then Just v else map_of ps k);
map_of [] k = Nothing;

get_name :: Xml -> [Char];
get_name (Xml name uu uv uw) = name;

xml_options ::
  forall a.
    [([Char], Xml -> Parse_result [Char] a)] -> Xml -> Parse_result [Char] a;
xml_options ps x =
  (case map_of ps (get_name x) of {
    Nothing ->
      errora
        (concat
          [['e', 'x', 'p', 'e', 'c', 't', 'e', 'd', ' ', 'o', 'n', 'e', ' ',
             'o', 'f', ':', ' '],
            concatMap (\ p -> fst p ++ [' ']) ps, [newline],
            ['b', 'u', 't', ' ', 'f', 'o', 'u', 'n', 'd'], [newline],
            shows_prec_xml Zero_nat x []]);
    Just p -> p x;
  });

nth :: forall a. [a] -> Nat -> a;
nth (x : xs) (Suc n) = nth xs n;
nth (x : xs) Zero_nat = x;

equal_nat :: Nat -> Nat -> Bool;
equal_nat (Suc nat) Zero_nat = False;
equal_nat Zero_nat (Suc nat) = False;
equal_nat (Suc nata) (Suc nat) = equal_nat nata nat;
equal_nat Zero_nat Zero_nat = True;

no_transform :: forall a. [Char] -> Xml -> Parse_result [Char] a;
no_transform tag xml =
  errora
    (concat
      [['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 't', 'r', 'a', 'n',
         's', 'f', 'o', 'r', 'm', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
         'o', 'w', 'i', 'n', 'g', '\n', ' ', ' ', ' ', ' ', ' ', 'x', 'm', 'l',
         ' ', 'e', 'l', 'e', 'm', 'e', 'n', 't', ' ', '(', 'e', 'x', 'p', 'e',
         'c', 't', 'e', 'd', ' '],
        tag, [')'], [newline], shows_prec_xml Zero_nat xml []]);

xml2elements ::
  forall a b c.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (a -> b -> c) -> Xml -> Parse_result [Char] c;
xml2elements tag p1 p2 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) (nat 2) && is_none text
    then bindc (p1 (nth cs Zero_nat))
           (\ x1 -> bindc (p2 (nth cs one_nat)) (\ x2 -> returna (f x1 x2)))
    else no_transform tag (Xml name atts cs text));

rat :: Xml -> Parse_result [Char] Rat;
rat = xml_options
        [(['i', 'n', 't', 'e', 'g', 'e', 'r'],
           xml_change (int ['i', 'n', 't', 'e', 'g', 'e', 'r']) of_int),
          (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l'],
            xml2elements ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l']
              (int ['n', 'u', 'm', 'e', 'r', 'a', 't', 'o', 'r'])
              (int ['d', 'e', 'n', 'o', 'm', 'i', 'n', 'a', 't', 'o', 'r'])
              (\ x y -> divide_rat (of_int x) (of_int y)))];

upt :: Nat -> Nat -> [Nat];
upt i j = (if less_nat i j then i : upt (Suc i) j else []);

data Ctxt a b = Hole | More a [Term a b] (Ctxt a b) [Term a b];

string2bool :: [Char] -> Parse_result [Char] Bool;
string2bool s =
  (if s == ['t', 'r', 'u', 'e'] then returna True
    else (if s == ['f', 'a', 'l', 's', 'e'] then returna False
           else errora
                  (['c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v', 'e',
                     'r', 't', ' '] ++
                    s ++ [' ', 'i', 'n', 't', 'o', ' ', 'B', 'o', 'o', 'l', 'e',
                           'a', 'n'])));

bool :: [Char] -> Xml -> Parse_result [Char] Bool;
bool tag node = bindc (text tag node) string2bool;

dropa :: forall a. Nat -> [a] -> [a];
dropa n [] = [];
dropa n (x : xs) = (case n of {
                     Zero_nat -> x : xs;
                     Suc m -> dropa m xs;
                   });

find :: forall a. (a -> Bool) -> [a] -> Maybe a;
find uu [] = Nothing;
find p (x : xs) = (if p x then Just x else find p xs);

fold :: forall a b. (a -> b -> b) -> [a] -> b -> b;
fold f [] = id;
fold f (x : xs) = fold f xs . f x;

lasta :: forall a. [a] -> a;
lasta (x : xs) = (if null xs then x else lasta xs);

empty :: forall a b. (Linorder a) => Rbt a b;
empty = Rbt Emptya;

class Times a where {
  times :: a -> a -> a;
};

class (Times a) => Dvd a where {
};

args :: forall a b. Term a b -> [Term a b];
args (Var x) = [];
args (Fun f ts) = ts;

root :: forall a b. Term a b -> Maybe (a, Nat);
root (Var x) = Nothing;
root (Fun f ts) = Just (f, size_list ts);

class One a where {
  one :: a;
};

instance Times Integer where {
  times = (\ a b -> a * b);
};

instance Dvd Integer where {
};

instance One Integer where {
  one = one_int;
};

instance Orda Integer where {
  less_eq = (\ a b -> a <= b);
  less = (\ a b -> a < b);
};

foldla :: forall a b. (a -> b -> a) -> a -> [b] -> a;
foldla f a [] = a;
foldla f a (x : xs) = foldla f (f a x) xs;

vec :: forall a. Nat -> [a] -> Bool;
vec n x = equal_nat (size_list x) n;

mat :: forall a. Nat -> Nat -> [[a]] -> Bool;
mat nr nc m = equal_nat (size_list m) nc && all (vec nr) m;

instance Orda Nat where {
  less_eq = less_eq_nat;
  less = less_nat;
};

mapa :: forall a b. (a -> b) -> Maybe a -> Maybe b;
mapa f (Just x) = Just (f x);
mapa f Nothing = Nothing;

membera :: forall a. (Eq a) => [a] -> a -> Bool;
membera [] y = False;
membera (x : xs) y = x == y || membera xs y;

insertb :: forall a. (Eq a) => a -> [a] -> [a];
insertb x xs = (if membera xs x then xs else x : xs);

removeAll :: forall a. (Eq a) => a -> [a] -> [a];
removeAll x [] = [];
removeAll x (y : xs) = (if x == y then removeAll x xs else y : removeAll x xs);

inserta :: forall a. (Eq a) => a -> Set a -> Set a;
inserta x (Coset xs) = Coset (removeAll x xs);
inserta x (Set xs) = Set (insertb x xs);

bot_set :: forall a. Set a;
bot_set = Set [];

set :: forall a. (Eq a) => Maybe a -> Set a;
set Nothing = bot_set;
set (Just x) = inserta x bot_set;

impl_of :: forall a b. (Linorder a) => Rbt a b -> Rbta a b;
impl_of (Rbt x) = x;

paint :: forall a b. Color -> Rbta a b -> Rbta a b;
paint c Emptya = Emptya;
paint c (Branch uu l k v r) = Branch c l k v r;

balance :: forall a b. Rbta a b -> a -> b -> Rbta a b -> Rbta a b;
balance (Branch R a w x b) s t (Branch R c y z d) =
  Branch R (Branch B a w x b) s t (Branch B c y z d);
balance (Branch R (Branch R a w x b) s t c) y z Emptya =
  Branch R (Branch B a w x b) s t (Branch B c y z Emptya);
balance (Branch R (Branch R a w x b) s t c) y z (Branch B va vb vc vd) =
  Branch R (Branch B a w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch R Emptya w x (Branch R b s t c)) y z Emptya =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z Emptya);
balance (Branch R (Branch B va vb vc vd) w x (Branch R b s t c)) y z Emptya =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z Emptya);
balance (Branch R Emptya w x (Branch R b s t c)) y z (Branch B va vb vc vd) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch R (Branch B ve vf vg vh) w x (Branch R b s t c)) y z
  (Branch B va vb vc vd) =
  Branch R (Branch B (Branch B ve vf vg vh) w x b) s t
    (Branch B c y z (Branch B va vb vc vd));
balance Emptya w x (Branch R b s t (Branch R c y z d)) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z d);
balance (Branch B va vb vc vd) w x (Branch R b s t (Branch R c y z d)) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z d);
balance Emptya w x (Branch R (Branch R b s t c) y z Emptya) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z Emptya);
balance Emptya w x (Branch R (Branch R b s t c) y z (Branch B va vb vc vd)) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch B va vb vc vd) w x (Branch R (Branch R b s t c) y z Emptya) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z Emptya);
balance (Branch B va vb vc vd) w x
  (Branch R (Branch R b s t c) y z (Branch B ve vf vg vh)) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t
    (Branch B c y z (Branch B ve vf vg vh));
balance Emptya s t Emptya = Branch B Emptya s t Emptya;
balance Emptya s t (Branch B va vb vc vd) =
  Branch B Emptya s t (Branch B va vb vc vd);
balance Emptya s t (Branch v Emptya vb vc Emptya) =
  Branch B Emptya s t (Branch v Emptya vb vc Emptya);
balance Emptya s t (Branch v (Branch B ve vf vg vh) vb vc Emptya) =
  Branch B Emptya s t (Branch v (Branch B ve vf vg vh) vb vc Emptya);
balance Emptya s t (Branch v Emptya vb vc (Branch B vf vg vh vi)) =
  Branch B Emptya s t (Branch v Emptya vb vc (Branch B vf vg vh vi));
balance Emptya s t
  (Branch v (Branch B ve vj vk vl) vb vc (Branch B vf vg vh vi)) =
  Branch B Emptya s t
    (Branch v (Branch B ve vj vk vl) vb vc (Branch B vf vg vh vi));
balance (Branch B va vb vc vd) s t Emptya =
  Branch B (Branch B va vb vc vd) s t Emptya;
balance (Branch B va vb vc vd) s t (Branch B ve vf vg vh) =
  Branch B (Branch B va vb vc vd) s t (Branch B ve vf vg vh);
balance (Branch B va vb vc vd) s t (Branch v Emptya vf vg Emptya) =
  Branch B (Branch B va vb vc vd) s t (Branch v Emptya vf vg Emptya);
balance (Branch B va vb vc vd) s t
  (Branch v (Branch B vi vj vk vl) vf vg Emptya) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v (Branch B vi vj vk vl) vf vg Emptya);
balance (Branch B va vb vc vd) s t
  (Branch v Emptya vf vg (Branch B vj vk vl vm)) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v Emptya vf vg (Branch B vj vk vl vm));
balance (Branch B va vb vc vd) s t
  (Branch v (Branch B vi vn vo vp) vf vg (Branch B vj vk vl vm)) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v (Branch B vi vn vo vp) vf vg (Branch B vj vk vl vm));
balance (Branch v Emptya vb vc Emptya) s t Emptya =
  Branch B (Branch v Emptya vb vc Emptya) s t Emptya;
balance (Branch v Emptya vb vc (Branch B ve vf vg vh)) s t Emptya =
  Branch B (Branch v Emptya vb vc (Branch B ve vf vg vh)) s t Emptya;
balance (Branch v (Branch B vf vg vh vi) vb vc Emptya) s t Emptya =
  Branch B (Branch v (Branch B vf vg vh vi) vb vc Emptya) s t Emptya;
balance (Branch v (Branch B vf vg vh vi) vb vc (Branch B ve vj vk vl)) s t
  Emptya =
  Branch B (Branch v (Branch B vf vg vh vi) vb vc (Branch B ve vj vk vl)) s t
    Emptya;
balance (Branch v Emptya vf vg Emptya) s t (Branch B va vb vc vd) =
  Branch B (Branch v Emptya vf vg Emptya) s t (Branch B va vb vc vd);
balance (Branch v Emptya vf vg (Branch B vi vj vk vl)) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v Emptya vf vg (Branch B vi vj vk vl)) s t
    (Branch B va vb vc vd);
balance (Branch v (Branch B vj vk vl vm) vf vg Emptya) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v (Branch B vj vk vl vm) vf vg Emptya) s t
    (Branch B va vb vc vd);
balance (Branch v (Branch B vj vk vl vm) vf vg (Branch B vi vn vo vp)) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v (Branch B vj vk vl vm) vf vg (Branch B vi vn vo vp)) s t
    (Branch B va vb vc vd);

rbt_ins ::
  forall a b. (Orda a) => (a -> b -> b -> b) -> a -> b -> Rbta a b -> Rbta a b;
rbt_ins f k v Emptya = Branch R Emptya k v Emptya;
rbt_ins f k v (Branch B l x y r) =
  (if less k x then balance (rbt_ins f k v l) x y r
    else (if less x k then balance l x y (rbt_ins f k v r)
           else Branch B l x (f k y v) r));
rbt_ins f k v (Branch R l x y r) =
  (if less k x then Branch R (rbt_ins f k v l) x y r
    else (if less x k then Branch R l x y (rbt_ins f k v r)
           else Branch R l x (f k y v) r));

rbt_insert_with_key ::
  forall a b. (Orda a) => (a -> b -> b -> b) -> a -> b -> Rbta a b -> Rbta a b;
rbt_insert_with_key f k v t = paint B (rbt_ins f k v t);

rbt_insert :: forall a b. (Orda a) => a -> b -> Rbta a b -> Rbta a b;
rbt_insert = rbt_insert_with_key (\ _ _ nv -> nv);

insert :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
insert k v t = Rbt (rbt_insert k v (impl_of t));

rbt_lookup :: forall a b. (Orda a) => Rbta a b -> a -> Maybe b;
rbt_lookup Emptya k = Nothing;
rbt_lookup (Branch uu l x y r) k =
  (if less k x then rbt_lookup l k
    else (if less x k then rbt_lookup r k else Just y));

lookupa :: forall a b. (Linorder a) => Rbt a b -> a -> Maybe b;
lookupa t = rbt_lookup (impl_of t);

times_rat :: Rat -> Rat -> Rat;
times_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * b, c * d));

instance Times Rat where {
  times = times_rat;
};

instance Dvd Rat where {
};

one_rat :: Rat;
one_rat = Frct (one_int, one_int);

instance One Rat where {
  one = one_rat;
};

less_rat :: Rat -> Rat -> Bool;
less_rat p q =
  let {
    (a, c) = quotient_of p;
    (b, d) = quotient_of q;
  } in a * d < c * b;

less_eq_rat :: Rat -> Rat -> Bool;
less_eq_rat p q =
  let {
    (a, c) = quotient_of p;
    (b, d) = quotient_of q;
  } in a * d <= c * b;

instance Orda Rat where {
  less_eq = less_eq_rat;
  less = less_rat;
};

proja :: forall a b. Term a b -> Nat -> Term a b;
proja (Fun f ts) i = (if less_nat i (size_list ts) then nth ts i else Fun f ts);

member :: forall a. (Eq a) => a -> Set a -> Bool;
member x (Coset xs) = not (membera xs x);
member x (Set xs) = membera xs x;

shows_between ::
  ([Char] -> [Char]) ->
    ([Char] -> [Char]) -> ([Char] -> [Char]) -> [Char] -> [Char];
shows_between l p r = l . p . r;

quote :: ([Char] -> [Char]) -> [Char] -> [Char];
quote s =
  shows_between (shows_prec_char Zero_nat '\'') s
    (shows_prec_char Zero_nat '\'');

debug :: forall a. [Char] -> [Char] -> a -> a;
debug i t x = x;

fun_upd :: forall a b. (Eq a) => (a -> b) -> a -> b -> a -> b;
fun_upd f a b = (\ x -> (if x == a then b else f x));

instance Plus Integer where {
  plus = (\ a b -> a + b);
};

instance Zero Integer where {
  zero = 0;
};

class (One a, Semigroup_add a) => Numeral a where {
};

bind :: forall a b. Maybe a -> (a -> Maybe b) -> Maybe b;
bind Nothing f = Nothing;
bind (Just x) f = f x;

data Pos = Empty | PCons Nat Pos;

class (One a, Times a) => Power a where {
};

powera :: forall a. a -> (a -> a -> a) -> a -> Nat -> a;
powera one times a (Suc n) = times a (powera one times a n);
powera one times a Zero_nat = one;

power :: forall a. (Power a) => a -> Nat -> a;
power = powera one times;

prc_nat :: forall a. (a -> Nat -> Nat) -> a -> Nat -> a -> Nat -> (Bool, Bool);
prc_nat pr =
  (\ f n g m ->
    let {
      pf = pr f n;
      pg = pr g m;
    } in (less_nat pg pf, less_eq_nat pg pf));

prl_nat :: forall a. (a -> Nat -> Nat) -> a -> Bool;
prl_nat pr = (\ f -> equal_nat (pr f Zero_nat) Zero_nat);

plus_rat :: Rat -> Rat -> Rat;
plus_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * d + b * c, c * d));

instance Plus Rat where {
  plus = plus_rat;
};

zero_rat :: Rat;
zero_rat = Frct (0, one_int);

instance Zero Rat where {
  zero = zero_rat;
};

lterms ::
  forall a b. ((a, Nat) -> [(Nat, Nat)]) -> Term a b -> [(Term a b, Nat)];
lterms pia =
  (\ (Fun f ts) ->
    map (\ (i, a) -> (proja (Fun f ts) i, a)) (pia (f, size_list ts)));

is_Var :: forall a b. Term a b -> Bool;
is_Var (Var x) = True;
is_Var (Fun f ts) = False;

remdups :: forall a. (Eq a) => [a] -> [a];
remdups [] = [];
remdups (x : xs) = (if membera xs x then remdups xs else x : remdups xs);

vars_term_lista :: forall a b. (Eq b) => Term a b -> [b];
vars_term_lista = remdups . vars_term_list;

wf_rule :: forall a b. (Eq b) => (Term a b, Term a b) -> Bool;
wf_rule r =
  not (is_Var (fst r)) &&
    all (membera (vars_term_lista (fst r))) (vars_term_lista (snd r));

fun_of :: forall a b. (Eq a) => [(a, b)] -> a -> b;
fun_of vec x = the (map_of vec x);

data Lab a b = Lab (Lab a b) b | FunLab (Lab a b) [Lab a b] | UnLab a
  | Sharp (Lab a b);

newtype ProjL a = Projection [((a, Nat), Nat)];

data Join_info a =
  Guided
    [(Term a [Char],
       ([(Pos, ((Term a [Char], Term a [Char]), Term a [Char]))],
         (Term a [Char],
           [(Pos, ((Term a [Char], Term a [Char]), Term a [Char]))])))]
  | Join_NF | Join_BFS Nat;

data ArithFun = Arg Nat | Const Nat | Sum [ArithFun] | Max [ArithFun]
  | Min [ArithFun] | Prod [ArithFun];

data Sl_inter a = SL_Inter Nat [((a, Nat), ArithFun)];

data Sl_variant a b = Rootlab (Maybe (a, Nat)) | Finitelab (Sl_inter a)
  | QuasiFinitelab (Sl_inter a) b;

data Arctic = MinInfty | Num_arc Integer;

data Af_entry = Collapse Nat | AFList [Nat];

data Arctic_rat = MinInfty_rat | Num_arc_rat Rat;

data Redtriple_impl a = Int_carrier [((a, Nat), (Integer, [Integer]))]
  | Int_nl_carrier [((a, Nat), [([(Nat, Nat)], Integer)])]
  | Rat_carrier [((a, Nat), (Rat, [Rat]))]
  | Rat_nl_carrier Rat [((a, Nat), [([(Nat, Nat)], Rat)])]
  | Arctic_carrier [((a, Nat), (Arctic, [Arctic]))]
  | Arctic_rat_carrier [((a, Nat), (Arctic_rat, [Arctic_rat]))]
  | Int_mat_carrier Nat Nat [((a, Nat), ([[Integer]], [[[Integer]]]))]
  | Rat_mat_carrier Nat Nat [((a, Nat), ([[Rat]], [[[Rat]]]))]
  | Arc_mat_carrier Nat [((a, Nat), ([[Arctic]], [[[Arctic]]]))]
  | Arc_rat_mat_carrier Nat [((a, Nat), ([[Arctic_rat]], [[[Arctic_rat]]]))]
  | Rpo [((a, Nat), (Nat, Order_tag))] [((a, Nat), Af_entry)]
  | Kbo ([((a, Nat), (Nat, Nat))], Nat) [((a, Nat), Af_entry)];

data List_order_type = MS_Ext | Max_Ext | Min_Ext | Dms_Ext;

data Root_redtriple_impl a =
  Scnp List_order_type [((a, Nat), [(Nat, Nat)])] (Redtriple_impl a);

data Dp_proof a b c = PisEmpty
  | SubtermCriterionProc (ProjL (Lab a b))
      [((Term (Lab a b) c, Term (Lab a b) c),
         [(Pos, ((Term (Lab a b) c, Term (Lab a b) c), Term (Lab a b) c))])]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | RedPairProc (Sum (Root_redtriple_impl (Lab a b)) (Redtriple_impl (Lab a b)))
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | RedPairUrProc
      (Sum (Root_redtriple_impl (Lab a b)) (Redtriple_impl (Lab a b)))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | UsableRulesProc [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | DepGraphProc
      [(Maybe (Dp_proof a b c), [(Term (Lab a b) c, Term (Lab a b) c)])]
  | MonoRedPairProc (Redtriple_impl (Lab a b))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | MonoRedPairUrProc (Redtriple_impl (Lab a b))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | SizeChangeSubterm
      [((Term (Lab a b) c, Term (Lab a b) c), ([(Nat, Nat)], [(Nat, Nat)]))]
  | SizeChangeRedPair (Redtriple_impl (Lab a b))
      (Maybe [(Term (Lab a b) c, Term (Lab a b) c)])
      [((Term (Lab a b) c, Term (Lab a b) c), ([(Nat, Nat)], [(Nat, Nat)]))]
  | UncurryProc (Maybe Nat)
      (Lab a b,
        ([((Lab a b, Nat), [Lab a b])],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | FccProc (Lab a b) [Ctxt (Lab a b) c] [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | SplitProc [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c) (Dp_proof a b c)
  | SemlabProc (Sl_variant (Lab a b) c) [(Term (Lab a b) c, Term (Lab a b) c)]
      [Term (Lab a b) c] [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | SwitchInnermostProc (Join_info (Lab a b)) (Dp_proof a b c)
  | RewritingProc [(Term (Lab a b) c, Term (Lab a b) c)]
      (Term (Lab a b) c, Term (Lab a b) c) (Term (Lab a b) c, Term (Lab a b) c)
      (Term (Lab a b) c, Term (Lab a b) c) (Term (Lab a b) c, Term (Lab a b) c)
      Pos (Dp_proof a b c)
  | InstantiationProc (Term (Lab a b) c, Term (Lab a b) c)
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | ForwardInstantiationProc (Term (Lab a b) c, Term (Lab a b) c)
      [(Term (Lab a b) c, Term (Lab a b) c)]
      (Maybe [(Term (Lab a b) c, Term (Lab a b) c)]) (Dp_proof a b c)
  | NarrowingProc (Term (Lab a b) c, Term (Lab a b) c) Pos
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | AssumeFinite
      (Bool,
        (Bool,
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            ([(Term (Lab a b) c, Term (Lab a b) c)],
              ([Term (Lab a b) c],
                ([(Term (Lab a b) c, Term (Lab a b) c)],
                  [(Term (Lab a b) c, Term (Lab a b) c)]))))))
  | UnlabProc [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | QReductionProc [Term (Lab a b) c] (Dp_proof a b c);

data Boundstype = Roof | Match;

data Ta_rule a b = TA_rule b [a] a;

data Tree_automaton a b = Tree_Automaton [a] [Ta_rule a b] [(a, a)];

data Bounds_info a b =
  Bounds_Info Boundstype Nat [b] (Tree_automaton b (a, Nat));

data Trs_proof a b c =
  DPTrans [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | RuleRemoval (Redtriple_impl (Lab a b))
      [(Term (Lab a b) c, Term (Lab a b) c)] (Trs_proof a b c)
  | StringReversal (Trs_proof a b c) | Bounds (Bounds_info (Lab a b) c)
  | Uncurry
      (Lab a b,
        ([((Lab a b, Nat), [Lab a b])],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])))
      [(Term (Lab a b) c, Term (Lab a b) c)] (Trs_proof a b c)
  | Semlab (Sl_variant (Lab a b) c) [Term (Lab a b) c]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Trs_proof a b c)
  | RisEmpty
  | Fcc [Ctxt (Lab a b) c] [(Term (Lab a b) c, Term (Lab a b) c)]
      (Trs_proof a b c)
  | Split [(Term (Lab a b) c, Term (Lab a b) c)] (Trs_proof a b c)
      (Trs_proof a b c)
  | SwitchInnermost (Join_info (Lab a b)) (Trs_proof a b c)
  | DropEquality (Trs_proof a b c)
  | AssumeSN
      (Bool,
        ([Term (Lab a b) c],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])));

instance Power Integer where {
};

butlast :: forall a. [a] -> [a];
butlast [] = [];
butlast (x : xs) = (if null xs then [] else x : butlast xs);

remove1 :: forall a. (Eq a) => a -> [a] -> [a];
remove1 x [] = [];
remove1 x (y : xs) = (if x == y then xs else y : remove1 x xs);

it_add ::
  forall a b c d.
    (a -> b -> c -> c) ->
      (d -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c) -> c -> d -> c;
it_add update iti m1 m2 = iti m2 (\ _ -> True) (\ (a, b) -> update a b) m1;

vec0I :: forall a. a -> Nat -> [a];
vec0I ze n = replicatea n ze;

mat0I :: forall a. a -> Nat -> Nat -> [[a]];
mat0I ze nr nc = replicatea nc (vec0I ze nr);

minus_nat :: Nat -> Nat -> Nat;
minus_nat (Suc m) (Suc n) = minus_nat m n;
minus_nat Zero_nat n = Zero_nat;
minus_nat m Zero_nat = m;

vec1I :: forall a. a -> a -> Nat -> Nat -> [a];
vec1I ze on n i =
  replicatea i ze ++ on : replicatea (minus_nat (minus_nat n one_nat) i) ze;

mat1I :: forall a. a -> a -> Nat -> [[a]];
mat1I ze on n = map (vec1I ze on n) (upt Zero_nat n);

lm_store_acc ::
  forall a b. (Eq a) => [(a, b)] -> (a, b) -> [(a, b)] -> [(a, b)];
lm_store_acc [] kv accu = kv : accu;
lm_store_acc ((a, uu) : rec) (k, v) accu =
  (case k == a of {
    True -> accu;
    False -> lm_store_acc rec (k, v) accu;
  });

lm_store :: forall a b. (Eq a) => [(a, b)] -> (a, b) -> [(a, b)];
lm_store m kv = lm_store_acc m kv m;

l2m_store ::
  forall a b c.
    (Eq a, Eq b) => [(a, [(b, c)])] -> ((a, b), c) -> [(a, [(b, c)])];
l2m_store [] ((k1, k2), v) = [(k1, [(k2, v)])];
l2m_store ((a, kvs) : rec) ((k1, k2), v) =
  (case k1 == a of {
    True -> (a, lm_store kvs (k2, v)) : rec;
    False -> (a, kvs) : l2m_store rec ((k1, k2), v);
  });

data Memory_ext a b c d =
  Memory_ext (() -> a) (a -> b -> Maybe c) (a -> (b, c) -> a) d;

l2m_lookup ::
  forall a b c. (Eq a, Eq b) => [(a, [(b, c)])] -> (a, b) -> Maybe c;
l2m_lookup [] uu = Nothing;
l2m_lookup ((a, kvs) : rec) (k1, k2) =
  (case k1 == a of {
    True -> map_of kvs k2;
    False -> l2m_lookup rec (k1, k2);
  });

l2m :: forall a b c. (Eq a, Eq b) => Memory_ext [(a, [(b, c)])] (a, b) c ();
l2m = Memory_ext (\ _ -> []) l2m_lookup l2m_store ();

less_eq_prod_aux ::
  forall a b. (Eq a, Orda a, Orda b) => (a, b) -> (a, b) -> Bool;
less_eq_prod_aux (a1, a2) (b1, b2) = less a1 b1 || a1 == b1 && less_eq a2 b2;

less_prod ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => (a, b) -> (a, b) -> Bool;
less_prod a b = not (a == b) && less_eq_prod_aux a b;

less_eq_prod :: forall a b. (Eq a, Orda a, Orda b) => (a, b) -> (a, b) -> Bool;
less_eq_prod a b = less_eq_prod_aux a b;

instance (Eq a, Orda a, Eq b, Orda b) => Orda (a, b) where {
  less_eq = less_eq_prod;
  less = less_prod;
};

rm_lookup :: forall a b. (Linorder a) => a -> Rbt a b -> Maybe b;
rm_lookup k m = lookupa m k;

rm_update :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
rm_update = insert;

rm_update_dj :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
rm_update_dj = rm_update;

aux ::
  forall a b.
    (Eq a,
      Linorder b) => (a -> Maybe b) -> Rbt b [a] -> a -> Rbt b [a] -> Rbt b [a];
aux key ma v m =
  (case key v of {
    Nothing -> m;
    Just k ->
      (case rm_lookup k ma of {
        Nothing -> m;
        Just ws ->
          (if membera ws v
            then (case rm_lookup k m of {
                   Nothing -> rm_update_dj k [v] m;
                   Just vs -> rm_update k (insertb v vs) m;
                 })
            else m);
      });
  });

instance Eq Nat where {
  a == b = equal_nat a b;
};

instance Preorder Nat where {
};

instance Order Nat where {
};

equal_rat :: Rat -> Rat -> Bool;
equal_rat a b = quotient_of a == quotient_of b;

instance Eq Rat where {
  a == b = equal_rat a b;
};

instance Power Rat where {
};

key :: forall a b c. (a, (Term b c, Term b c)) -> Maybe (b, Nat);
key (uu, (Fun f ts, uv)) = Just (f, size_list ts);
key (uw, (Var ux, uy)) = Nothing;

class Non_strict_order a where {
  ge :: a -> a -> Bool;
};

shows_sep ::
  forall a.
    (a -> [Char] -> [Char]) -> ([Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_sep s sep [] = shows_string [];
shows_sep s sep [x] = s x;
shows_sep s sep (x : v : va) = s x . sep . shows_sep s sep (v : va);

shows_list_gen ::
  forall a.
    (a -> [Char] -> [Char]) ->
      [Char] -> [Char] -> [Char] -> [Char] -> [a] -> [Char] -> [Char];
shows_list_gen elt e l s r xs =
  (if null xs then shows_string e
    else shows_between (shows_string l) (shows_sep elt (shows_string s) xs)
           (shows_string r));

shows_list_aux :: forall a. (a -> [Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_list_aux s xs = shows_list_gen s ['[', ']'] ['['] [','] [']'] xs;

divmod_nat :: Nat -> Nat -> (Nat, Nat);
divmod_nat m n =
  (if equal_nat n Zero_nat || less_nat m n then (Zero_nat, m)
    else let {
           (q, a) = divmod_nat (minus_nat m n) n;
         } in (Suc q, a));

div_nat :: Nat -> Nat -> Nat;
div_nat m n = fst (divmod_nat m n);

mod_nat :: Nat -> Nat -> Nat;
mod_nat m n = snd (divmod_nat m n);

digit2string :: Nat -> [Char];
digit2string n =
  (if equal_nat n Zero_nat then ['0']
    else (if equal_nat n one_nat then ['1']
           else (if equal_nat n (nat 2) then ['2']
                  else (if equal_nat n (nat 3) then ['3']
                         else (if equal_nat n (nat 4) then ['4']
                                else (if equal_nat n (nat 5) then ['5']
                                       else (if equal_nat n (nat 6) then ['6']
      else (if equal_nat n (nat 7) then ['7']
             else (if equal_nat n (nat 8) then ['8']
                    else (if equal_nat n (nat 9) then ['9']
                           else ['n', 'o', 't', ' ', 'a', ' ', 'd', 'i', 'g',
                                  'i', 't']))))))))));

shows_nat :: Nat -> [Char] -> [Char];
shows_nat n =
  (if less_nat n (nat 10) then shows_string (digit2string n)
    else shows_nat (div_nat n (nat 10)) .
           shows_string (digit2string (mod_nat n (nat 10))));

shows_prec_nat :: Nat -> Nat -> [Char] -> [Char];
shows_prec_nat d n = shows_nat n;

shows_prec_int :: Nat -> Integer -> [Char] -> [Char];
shows_prec_int d i =
  (if i < 0
    then shows_prec_char Zero_nat '-' . shows_prec_nat Zero_nat (nat (0 - i))
    else shows_prec_nat Zero_nat (nat i));

shows_list_int :: [Integer] -> [Char] -> [Char];
shows_list_int is = shows_list_aux (shows_prec_int Zero_nat) is;

instance Showa Integer where {
  shows_prec = shows_prec_int;
  shows_list = shows_list_int;
};

shows_list_nat :: [Nat] -> [Char] -> [Char];
shows_list_nat ns = shows_list_aux (shows_prec_nat Zero_nat) ns;

instance Showa Nat where {
  shows_prec = shows_prec_nat;
  shows_list = shows_list_nat;
};

show_rat :: Rat -> [Char];
show_rat x =
  let {
    (den, num) = quotient_of x;
  } in (if num == 1 then shows_prec_int Zero_nat den []
         else ((shows_prec_int Zero_nat den . shows_prec_char Zero_nat '/') .
                shows_prec_int Zero_nat num)
                []);

shows_prec_rat :: Nat -> Rat -> [Char] -> [Char];
shows_prec_rat d x = shows_string (show_rat x);

shows_list_rat :: [Rat] -> [Char] -> [Char];
shows_list_rat rs = shows_list_aux (shows_prec_rat Zero_nat) rs;

instance Showa Rat where {
  shows_prec = shows_prec_rat;
  shows_list = shows_list_rat;
};

subt_at :: forall a b. Term a b -> Pos -> Term a b;
subt_at s Empty = s;
subt_at (Fun f ss) (PCons i p) = subt_at (nth ss i) p;

the_Var :: forall a b. Term a b -> b;
the_Var (Var x) = x;

map_term :: forall a b c. (a -> b) -> Term a c -> Term b c;
map_term fun (Var x) = Var x;
map_term fun (Fun f ts) = Fun (fun f) (map (map_term fun) ts);

map_rule ::
  forall a b c. (a -> b) -> (Term a c, Term a c) -> (Term b c, Term b c);
map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr));

data Gctxt a b = GCHole | GCFun a [Gctxt a b];

guard :: Bool -> Maybe ();
guard b = (if b then Just () else Nothing);

merge_lists ::
  forall a b. (Eq a) => [Gctxt a b] -> [Gctxt a b] -> Maybe [Gctxt a b];
merge_lists (c : cs) [] = Nothing;
merge_lists [] (d : ds) = Nothing;
merge_lists (GCFun f ss : cs) (GCFun g ts : ds) =
  bind (guard (f == g))
    (\ _ ->
      bind (merge_lists ss ts)
        (\ us -> bind (merge_lists cs ds) (\ es -> Just (GCFun f us : es))));
merge_lists (c : cs) (GCHole : ds) =
  bind (merge_lists cs ds) (\ es -> Just (c : es));
merge_lists (GCHole : cs) (d : ds) =
  bind (merge_lists cs ds) (\ es -> Just (d : es));
merge_lists [] [] = Just [];

mergea :: forall a b. (Eq a) => Gctxt a b -> Gctxt a b -> Maybe (Gctxt a b);
mergea c d = bind (merge_lists [c] [d]) (\ es -> Just (nth es Zero_nat));

merge_var ::
  forall a b.
    (Eq a,
      Eq b) => a -> Gctxt b a ->
                      [(Gctxt b a, a)] ->
                        Maybe ((Gctxt b a, a), [(Gctxt b a, a)]);
merge_var x c ((d, y) : ps) =
  (if x == y then bind (mergea c d) (\ e -> merge_var x e ps)
    else bind (merge_var x c ps) (\ (b, psa) -> Just (b, (d, y) : psa)));
merge_var x c [] = Just ((c, x), []);

merge_alla ::
  forall a b. (Eq a, Eq b) => [(Gctxt a b, b)] -> Maybe [(Gctxt a b, b)];
merge_alla [] = Just [];
merge_alla ((c, x) : ps) =
  bind (merge_var x c ps)
    (\ (cx, psa) -> bind (merge_alla psa) (\ psb -> Just (cx : psb)));

zip_option :: forall a b. [a] -> [b] -> Maybe [(a, b)];
zip_option [] [] = Just [];
zip_option (x : xs) (y : ys) =
  bind (zip_option xs ys) (\ zs -> Just ((x, y) : zs));
zip_option (x : xs) [] = Nothing;
zip_option [] (y : ys) = Nothing;

match_lista ::
  forall a b. (Eq a) => [(Gctxt a b, Term a b)] -> Maybe [(Gctxt a b, b)];
match_lista ((GCFun f ss, Var x) : ps) =
  bind (match_lista ps) (\ psa -> Just ((GCFun f ss, x) : psa));
match_lista ((GCFun f ss, Fun g ts) : ps) =
  bind (guard (f == g))
    (\ _ -> bind (zip_option ss ts) (\ psa -> match_lista (psa ++ ps)));
match_lista ((GCHole, t) : ps) = match_lista ps;
match_lista [] = Just [];

matchc ::
  forall a b. (Eq a, Eq b) => (Gctxt a b, Term a b) -> Maybe [(Gctxt a b, b)];
matchc (c, t) = bind (match_lista [(c, t)]) merge_alla;

matchb :: forall a b. (Eq a, Eq b) => Gctxt a b -> Term a b -> Bool;
matchb c t = not (is_none (matchc (c, t)));

capI ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Gctxt a b;
capI uu (Var uv) = GCHole;
capI r (Fun f ts) =
  let {
    h = GCFun f (map (capI r) ts);
  } in (if any (\ ra -> matchb h (fst ra)) r then GCHole else h);

distinct :: forall a. (Eq a) => [a] -> Bool;
distinct [] = True;
distinct (x : xs) = not (membera xs x) && distinct xs;

merge :: forall a b. (Linorder b) => (a -> b) -> [a] -> [a] -> [a];
merge key (a : asa) (b : bs) =
  (if less (key b) (key a) then b : merge key (a : asa) bs
    else a : merge key asa (b : bs));
merge key [] bs = bs;
merge key (v : va) [] = v : va;

merge_pairs :: forall a b. (Linorder b) => (a -> b) -> [[a]] -> [[a]];
merge_pairs key (a : b : xs) = merge key a b : merge_pairs key xs;
merge_pairs key [] = [];
merge_pairs key [v] = [v];

merge_all :: forall a b. (Linorder b) => (a -> b) -> [[a]] -> [a];
merge_all key [] = [];
merge_all key [x] = x;
merge_all key (v : vb : vc) = merge_all key (merge_pairs key (v : vb : vc));

sequences :: forall a b. (Linorder b) => (a -> b) -> [a] -> [[a]];
sequences key (a : b : xs) =
  (if less (key b) (key a) then desc key b [a] xs
    else asc key b (\ ba -> a : ba) xs);
sequences key [] = [[]];
sequences key [v] = [[v]];

desc :: forall a b. (Linorder b) => (a -> b) -> a -> [a] -> [a] -> [[a]];
desc key a asa (b : bs) =
  (if less (key b) (key a) then desc key b (a : asa) bs
    else (a : asa) : sequences key (b : bs));
desc key a asa [] = (a : asa) : sequences key [];

asc ::
  forall a b. (Linorder b) => (a -> b) -> a -> ([a] -> [a]) -> [a] -> [[a]];
asc key a f (b : bs) =
  (if not (less (key b) (key a)) then asc key b (f . (\ ba -> a : ba)) bs
    else f [a] : sequences key (b : bs));
asc key a f [] = f [a] : sequences key [];

sort_key :: forall a b. (Linorder b) => (a -> b) -> [a] -> [a];
sort_key key = merge_all key . sequences key;

vec_comp_all :: forall a. (a -> a -> Bool) -> [a] -> [a] -> Bool;
vec_comp_all r v w = all (\ (a, b) -> r a b) (zip v w);

mat_comp_all :: forall a. (a -> a -> Bool) -> [[a]] -> [[a]] -> Bool;
mat_comp_all r m1 m2 = all (\ (a, b) -> vec_comp_all r a b) (zip m1 m2);

mat_ge :: forall a. (Non_strict_order a) => [[a]] -> [[a]] -> Bool;
mat_ge = mat_comp_all ge;

times_nat :: Nat -> Nat -> Nat;
times_nat Zero_nat n = Zero_nat;
times_nat (Suc m) n = plus_nat n (times_nat m n);

maxa :: forall a. (Orda a) => a -> a -> a;
maxa a b = (if less_eq a b then b else a);

mina :: forall a. (Orda a) => a -> a -> a;
mina a b = (if less_eq a b then a else b);

minus_rat :: Rat -> Rat -> Rat;
minus_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * d - b * c, c * d));

less_eq_set :: forall a. (Eq a) => Set a -> Set a -> Bool;
less_eq_set (Coset []) (Set []) = False;
less_eq_set a (Coset ys) = all (\ y -> not (member y a)) ys;
less_eq_set (Set xs) b = all (\ x -> member x b) xs;

equal_set :: forall a. (Eq a) => Set a -> Set a -> Bool;
equal_set a b = less_eq_set a b && less_eq_set b a;

shows_list_list :: forall a. (Showa a) => [[a]] -> [Char] -> [Char];
shows_list_list ls = shows_list_aux (shows_prec_list Zero_nat) ls;

instance (Showa a) => Showa [a] where {
  shows_prec = shows_prec_list;
  shows_list = shows_list_list;
};

shows_paren :: ([Char] -> [Char]) -> [Char] -> [Char];
shows_paren p = shows_prec_char Zero_nat '(' . p . shows_prec_char Zero_nat ')';

shows_prec_prod ::
  forall a b. (Showa a, Showa b) => Nat -> (a, b) -> [Char] -> [Char];
shows_prec_prod d p =
  shows_paren
    (shows_prec Zero_nat (fst p) .
      shows_prec_char Zero_nat ',' . shows_prec Zero_nat (snd p));

shows_list_prod ::
  forall a b. (Showa a, Showa b) => [(a, b)] -> [Char] -> [Char];
shows_list_prod ps = shows_list_aux (shows_prec_prod Zero_nat) ps;

instance (Showa a, Showa b) => Showa (a, b) where {
  shows_prec = shows_prec_prod;
  shows_list = shows_list_prod;
};

shows_nl :: [Char] -> [Char];
shows_nl = shows_prec_char Zero_nat newline;

hole_pos :: forall a b. Ctxt a b -> Pos;
hole_pos Hole = Empty;
hole_pos (More f ss d ts) = PCons (size_list ss) (hole_pos d);

map_vars :: forall a b c. (a -> b) -> Term c a -> Term c b;
map_vars m (Var x) = Var (m x);
map_vars m (Fun f ts) = Fun f (map (map_vars m) ts);

num_args :: forall a b. Term a b -> Nat;
num_args (Var x) = Zero_nat;
num_args (Fun f ts) = size_list ts;

unapp :: forall a b. (Eq a) => a -> Term a b -> (Term a b, [Term a b]);
unapp a (Var x) = (Var x, []);
unapp a (Fun f ss) =
  (if f == a && equal_nat (size_list ss) (nat 2)
    then let {
           (r, ts) = unapp a (nth ss Zero_nat);
         } in (r, ts ++ [nth ss one_nat])
    else (Fun f ss, []));

max_list :: [Nat] -> Nat;
max_list [] = Zero_nat;
max_list (x : xs) = maxa x (max_list xs);

min_list :: forall a. (Linorder a) => [a] -> a;
min_list [x] = x;
min_list (x : v : va) = mina x (min_list (v : va));

xml3elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (a -> b -> c -> d) -> Xml -> Parse_result [Char] d;
xml3elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) (nat 3) && is_none text
    then bindc (p1 (nth cs Zero_nat))
           (\ x1 ->
             bindc (p2 (nth cs one_nat))
               (\ x2 ->
                 bindc (p3 (nth cs (nat 2))) (\ x3 -> returna (f x1 x2 x3))))
    else no_transform tag (Xml name atts cs text));

position :: Xml -> Parse_result [Char] Nat;
position =
  xml_change (nata ['p', 'o', 's', 'i', 't', 'i', 'o', 'n'])
    (\ n -> minus_nat n one_nat);

map_parse_result ::
  forall a b c. (a -> Parse_result b c) -> [a] -> Parse_result b [c];
map_parse_result f [] = returna [];
map_parse_result f (x : xs) =
  bindc (f x) (\ y -> bindc (map_parse_result f xs) (\ ys -> returna (y : ys)));

xml_many_elements ::
  forall a b.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        ([a] -> b) -> Xml -> Parse_result [Char] b;
xml_many_elements tag p f (Xml name atts cs text) =
  (if name == tag && null atts && is_none text
    then bindc (map_parse_result p cs) (returna . f)
    else no_transform tag (Xml name atts cs text));

afs ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [((a, Nat), Af_entry)];
afs xml2name =
  xml_many_elements
    ['a', 'r', 'g', 'u', 'm', 'e', 'n', 't', 'F', 'i', 'l', 't', 'e', 'r']
    (xml3elements
      ['a', 'r', 'g', 'u', 'm', 'e', 'n', 't', 'F', 'i', 'l', 't', 'e', 'r',
        'E', 'n', 't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (xml_options
        [(['c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'],
           xml_change (nata ['c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'])
             (\ i -> Collapse (minus_nat i one_nat))),
          (['n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'],
            xml_many_elements
              ['n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g']
              position AFList)])
      (\ f a -> (\ b -> ((f, a), b))))
    id;

list2position :: [Nat] -> Pos;
list2position [] = Empty;
list2position (n : ns) = PCons n (list2position ns);

pos :: Xml -> Parse_result [Char] Pos;
pos = xml_many_elements
        ['p', 'o', 's', 'i', 't', 'i', 'o', 'n', 'I', 'n', 'T', 'e', 'r', 'm']
        position list2position;

var :: forall a. Xml -> Parse_result [Char] (Term a [Char]);
var x = bindc (text ['v', 'a', 'r'] x) (returna . Var);

xml1element ::
  forall a b.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (a -> b) -> Xml -> Parse_result [Char] b;
xml1element tag p1 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) one_nat && is_none text
    then bindc (p1 (nth cs Zero_nat)) (\ x1 -> returna (f x1))
    else no_transform tag (Xml name atts cs text));

xml1many_elements_gen ::
  forall a b c.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (a -> Xml -> Parse_result [Char] b) ->
          (a -> [b] -> c) -> Xml -> Parse_result [Char] c;
xml1many_elements_gen tag p1 p2 f (Xml name atts cs text) =
  (if name == tag && null atts && is_none text && not (null cs)
    then bindc (p1 (hd cs))
           (\ x ->
             bindc (map_parse_result (p2 x) (tl cs)) (\ xs -> returna (f x xs)))
    else no_transform tag (Xml name atts cs text));

xml1many_elements ::
  forall a b c.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (a -> [b] -> c) -> Xml -> Parse_result [Char] c;
xml1many_elements tag p1 p2 = xml1many_elements_gen tag p1 (\ _ -> p2);

term ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Term a [Char]);
term xml2name x =
  xml_options
    [(['v', 'a', 'r'], var),
      (['f', 'u', 'n', 'a', 'p', 'p'],
        xml1many_elements ['f', 'u', 'n', 'a', 'p', 'p'] xml2name
          (xml1element ['a', 'r', 'g'] (term xml2name) id) Fun)]
    x;

rule ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Term a [Char], Term a [Char]);
rule xml2name =
  xml2elements ['r', 'u', 'l', 'e']
    (xml1element ['l', 'h', 's'] (term xml2name) id)
    (xml1element ['r', 'h', 's'] (term xml2name) id) (\ a b -> (a, b));

map_filter :: forall a b. (a -> Maybe b) -> [a] -> [b];
map_filter f [] = [];
map_filter f (x : xs) =
  (case f x of {
    Nothing -> map_filter f xs;
    Just y -> y : map_filter f xs;
  });

scg_position :: Xml -> Parse_result [Char] Nat;
scg_position = nata ['p', 'o', 's', 'i', 't', 'i', 'o', 'n'];

scg ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          ((Term a [Char], Term a [Char]), ([(Nat, Nat)], [(Nat, Nat)]));
scg xml2name =
  xml1many_elements
    ['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'G', 'r', 'a', 'p', 'h']
    (rule xml2name)
    (xml3elements ['e', 'd', 'g', 'e'] scg_position
      (bool ['s', 't', 'r', 'i', 'c', 't']) scg_position
      (\ p s q -> (s, (p, q))))
    (\ lr edges ->
      (lr, (map_filter (\ x -> (if fst x then Just (snd x) else Nothing)) edges,
             map_filter (\ x -> (if not (fst x) then Just (snd x) else Nothing))
               edges)));

capRM2 ::
  forall a b.
    (Eq a,
      Eq b) => ((a, Nat) -> [(Term a b, Term a b)]) -> Term a b -> Gctxt a b;
capRM2 uu (Var uv) = GCHole;
capRM2 rm (Fun f ts) =
  let {
    h = GCFun f (map (capRM2 rm) ts);
    n = size_list ts;
  } in (if any (\ r -> matchb h (fst r)) (rm (f, n)) then GCHole else h);

capRM ::
  forall a b.
    (Eq a,
      Eq b) => Bool ->
                 ((a, Nat) -> [(Term a b, Term a b)]) -> Term a b -> Gctxt a b;
capRM nlv rm = (if nlv then capRM2 rm else (\ _ -> GCHole));

equal_term :: forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Bool;
equal_term (Fun f list) (Var v) = False;
equal_term (Var v) (Fun f list) = False;
equal_term (Fun fa lista) (Fun f list) = fa == f && lista == list;
equal_term (Var va) (Var v) = va == v;

instance (Eq a, Eq b) => Eq (Term a b) where {
  a == b = equal_term a b;
};

varsL :: forall a b. (Eq b) => [(Term a b, Term a b)] -> b -> Bool;
varsL list x =
  foldra
    (\ eq b ->
      membera (vars_term_lista (fst eq)) x ||
        membera (vars_term_lista (snd eq)) x || b)
    list False;

decompose ::
  forall a b. (Eq a) => Term a b -> Term a b -> Maybe [(Term a b, Term a b)];
decompose s t =
  (case (s, t) of {
    (Var _, b) -> Nothing;
    (Fun f ss, Var _) -> Nothing;
    (Fun f ss, Fun g ts) -> (if f == g then zip_option ss ts else Nothing);
  });

simple_subst :: forall a b. (Eq a) => a -> Term b a -> Term b a -> Term b a;
simple_subst x t (Var y) = (if x == y then t else Var y);
simple_subst x t (Fun f ss) = Fun f (map (simple_subst x t) ss);

simple_subst_list ::
  forall a b.
    (Eq a) => a -> Term b a -> [(Term b a, Term b a)] -> [(Term b a, Term b a)];
simple_subst_list x u list =
  map (\ p -> (simple_subst x u (fst p), simple_subst x u (snd p))) list;

unify ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 [(Term a b, Term a b)] -> Maybe [(Term a b, Term a b)];
unify [] seen = Just seen;
unify ((Fun f ss, Fun g ts) : unifs) seen =
  (case decompose (Fun f ss) (Fun g ts) of {
    Nothing -> Nothing;
    Just eqs -> unify (eqs ++ unifs) seen;
  });
unify ((Fun f ss, Var x) : unifs) seen = unify ((Var x, Fun f ss) : unifs) seen;
unify ((Var x, t) : unifs) seen =
  (if equal_term t (Var x) then unify unifs seen
    else (if membera (vars_term_lista t) x then Nothing
           else (if varsL (unifs ++ seen) x
                  then unify (simple_subst_list x t (unifs ++ seen) ++
                               [(Var x, t)])
                         []
                  else unify unifs ((Var x, t) : seen))));

subst_of :: forall a b. (Eq b) => [(Term a b, Term a b)] -> b -> Term a b;
subst_of [] = Var;
subst_of ((Var x, t) : ts) = (\ y -> (if x == y then t else subst_of ts y));

mgu ::
  forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Maybe (b -> Term a b);
mgu s t =
  (case unify [(s, t)] [] of {
    Nothing -> Nothing;
    Just res -> Just (subst_of res);
  });

class_to_term_intern ::
  forall a b. (Nat -> a) -> Nat -> Term b (Sum () a) -> (Nat, Term b a);
class_to_term_intern iv i (Fun f ts) =
  let {
    (k, ss) =
      foldra
        (\ t (j, ss) ->
          let {
            (k, s) = class_to_term_intern iv j t;
          } in (k, s : ss))
        ts (i, []);
  } in (k, Fun f ss);
class_to_term_intern iv i (Var (Inl uu)) = (plus_nat i one_nat, Var (iv i));
class_to_term_intern iv i (Var (Inr x)) = (i, Var x);

class_to_term :: forall a. Char -> Term a (Sum () [Char]) -> Term a [Char];
class_to_term c t =
  snd (class_to_term_intern (\ i -> c : shows_prec_nat Zero_nat i []) Zero_nat
        t);

mgu_class ::
  forall a.
    (Eq a) => Term a (Sum () [Char]) ->
                Term a [Char] -> Maybe ([Char] -> Term a [Char]);
mgu_class cs t = mgu (class_to_term 'z' cs) (map_vars (\ a -> 'y' : a) t);

instance Semigroup_add Integer where {
};

instance Numeral Integer where {
};

partition :: forall a. (a -> Bool) -> [a] -> ([a], [a]);
partition p [] = ([], []);
partition p (x : xs) =
  let {
    (yes, no) = partition p xs;
  } in (if p x then (x : yes, no) else (yes, x : no));

sub_vec :: forall a. Nat -> [a] -> [a];
sub_vec = takea;

sub_mat :: forall a. Nat -> Nat -> [[a]] -> [[a]];
sub_mat nr nc m = map (sub_vec nr) (takea nc m);

vec_pre_gtI :: forall a. (a -> a -> Bool) -> [a] -> [a] -> Bool;
vec_pre_gtI gt v w = any (\ (a, b) -> gt a b) (zip v w);

mat_pre_gtI :: forall a. (a -> a -> Bool) -> [[a]] -> [[a]] -> Bool;
mat_pre_gtI gt m1 m2 = any (\ (a, b) -> vec_pre_gtI gt a b) (zip m1 m2);

mat_gtI ::
  forall a.
    (Non_strict_order a) => (a -> a -> Bool) -> Nat -> [[a]] -> [[a]] -> Bool;
mat_gtI gt sd m1 m2 =
  mat_ge m1 m2 && mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2);

vec_map :: forall a. (a -> a) -> [a] -> [a];
vec_map = map;

mat_map :: forall a. (a -> a) -> [[a]] -> [[a]];
mat_map f = map (vec_map f);

emptya :: forall a b c d. Memory_ext a b c d -> () -> a;
emptya (Memory_ext empty lookupa store more) = empty;

store :: forall a b c d. Memory_ext a b c d -> a -> (b, c) -> a;
store (Memory_ext empty lookupa store more) = store;

instance (Eq a, Order a, Eq b, Order b) => Preorder (a, b) where {
};

instance (Eq a, Order a, Eq b, Order b) => Order (a, b) where {
};

instance Semigroup_add Rat where {
};

instance Numeral Rat where {
};

class (Times a) => Semigroup_mult a where {
};

class (Semigroup_add a) => Ab_semigroup_add a where {
};

class (Ab_semigroup_add a, Semigroup_mult a) => Semiring a where {
};

class (Ab_semigroup_add a, Non_strict_order a) => Ordered_ab_semigroup a where {
};

class (Monoid_add a, Ordered_ab_semigroup a) => Max_ordered_monoid_add a where {
  max0 :: a -> a;
};

s_ins :: forall a b c. (a -> () -> b -> c) -> a -> b -> c;
s_ins update x s = update x () s;

sharp_term :: forall a b. (a -> a) -> Term a b -> Term a b;
sharp_term shp (Var x) = Var x;
sharp_term shp (Fun f ss) = Fun (shp f) ss;

aarity :: forall a. (a -> Nat -> [a]) -> a -> Nat -> Nat;
aarity sm f n = minus_nat (size_list (sm f n)) one_nat;

list_diff :: forall a. (Eq a) => [a] -> [a] -> [a];
list_diff [] ys = [];
list_diff (x : xs) ys =
  let {
    zs = list_diff xs ys;
  } in (if membera ys x then zs else x : zs);

expecting :: forall a b. (Showa a) => [Char] -> [a] -> Sum [Char] (b, [a]);
expecting msg ts =
  Inl (['e', 'x', 'p', 'e', 'c', 't', 'i', 'n', 'g', ':', ' '] ++
        msg ++
          [',', ' ', 'b', 'u', 't', ' ', 'f', 'o', 'u', 'n', 'd', ':', ' '] ++
            quote (shows_prec_list Zero_nat (takea (nat 30) ts)) []);

eoi :: forall a. (Showa a) => [a] -> Sum [Char] ((), [a]);
eoi [] = Inr ((), []);
eoi (v : va) =
  expecting ['e', 'n', 'd', ' ', 'o', 'f', ' ', 'i', 'n', 'p', 'u', 't']
    (v : va);

catchb ::
  forall a b c. Parse_result a b -> (a -> Parse_result c b) -> Parse_result c b;
catchb (Parseresult a) f =
  (case a of {
    Inl aa -> f aa;
    Inr aa -> Parseresult (Inr aa);
  });

xml_choice ::
  forall a.
    [Char] -> [Xml -> Parse_result [Char] a] -> Xml -> Parse_result [Char] a;
xml_choice e [] x =
  errora
    (concat
      [['e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'p', 'a', 'r', 's', 'i',
         'n', 'g', ' ', 'c', 'h', 'o', 'i', 'c', 'e', ' ', 'f', 'o', 'r', ' '],
        e, [newline], shows_prec_xml Zero_nat x []]);
xml_choice e (p : ps) x = catchb (p x) (\ _ -> xml_choice e ps x);

data Alter a b c d e f g h i j = Alter1 a | Alter2 b | Alter3 c | Alter4 d
  | Alter5 e | Alter6 f | Alter7 g | Alter8 h | Alter9 i | Alter10 j;

xml0elements :: forall a. [Char] -> a -> Xml -> Parse_result [Char] a;
xml0elements tag x (Xml name atts cs text) =
  (if name == tag && null atts && null cs && is_none text then returna x
    else no_transform tag (Xml name atts cs text));

xml4elements ::
  forall a b c d e.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (Xml -> Parse_result [Char] d) ->
              (a -> b -> c -> d -> e) -> Xml -> Parse_result [Char] e;
xml4elements tag p1 p2 p3 p4 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) (nat 4) && is_none text
    then bindc (p1 (nth cs Zero_nat))
           (\ x1 ->
             bindc (p2 (nth cs one_nat))
               (\ x2 ->
                 bindc (p3 (nth cs (nat 2)))
                   (\ x3 ->
                     bindc (p4 (nth cs (nat 3)))
                       (\ x4 -> returna (f x1 x2 x3 x4)))))
    else no_transform tag (Xml name atts cs text));

ctxt ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Ctxt a [Char]);
ctxt xml2name x =
  xml_options
    [(['b', 'o', 'x'], xml0elements ['b', 'o', 'x'] Hole),
      (['f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'],
        xml4elements ['f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'] xml2name
          (xml_many_elements ['b', 'e', 'f', 'o', 'r', 'e'] (term xml2name) id)
          (ctxt xml2name)
          (xml_many_elements ['a', 'f', 't', 'e', 'r'] (term xml2name) id)
          More)]
    x;

subst ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [([Char], Term a [Char])];
subst xml2name =
  xml_many_elements ['s', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i', 'o', 'n']
    (xml2elements ['s', 'u', 'b', 's', 't', 'E', 'n', 't', 'r', 'y']
      (text ['v', 'a', 'r']) (term xml2name) (\ a b -> (a, b)))
    id;

xml3to4elements ::
  forall a b c d e.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (Xml -> Parse_result [Char] d) ->
              (a -> b -> Maybe c -> d -> e) -> Xml -> Parse_result [Char] e;
xml3to4elements tag p1 p2 p3 p4 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             null atts &&
               less_eq_nat (nat 3) l && less_eq_nat l (nat 4) && is_none text
         then bindc (p1 (nth cs Zero_nat))
                (\ x1 ->
                  bindc (p2 (nth cs one_nat))
                    (\ x2 ->
                      (if equal_nat l (nat 4)
                        then bindc (p3 (nth cs (nat 2)))
                               (\ x3 ->
                                 bindc (p4 (nth cs (nat 3)))
                                   (\ x4 -> returna (f x1 x2 (Just x3) x4)))
                        else bindc (p4 (nth cs (nat 2)))
                               (\ x4 -> returna (f x1 x2 Nothing x4)))))
         else no_transform tag (Xml name atts cs text));

relstep ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])));
relstep xml2name =
  xml3to4elements ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 't', 'e', 'p'] pos
    (rule xml2name) (xml0elements ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e'] ())
    (term xml2name) (\ p r rel t -> (p, (r, (is_none rel, t))));

relsteps ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Term a [Char],
            [(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))]);
relsteps xml2name =
  xml1many_elements
    ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e']
    (xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'] (term xml2name)
      id)
    (relstep xml2name) (\ a b -> (a, b));

loopa ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Term a [Char],
            ([(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))],
              ([([Char], Term a [Char])], Ctxt a [Char])));
loopa xml2name =
  xml3elements ['l', 'o', 'o', 'p'] (relsteps xml2name) (subst xml2name)
    (ctxt xml2name) (\ (s, rseq) sigma c -> (s, (rseq, (sigma, c))));

proj ::
  forall a.
    (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] (ProjL a);
proj xml2name =
  xml_change (afs xml2name)
    (\ afl ->
      Projection
        (map (\ a ->
               (case a of {
                 (fa, Collapse aa) -> (fa, aa);
                 (fa, AFList _) -> (fa, snd fa);
               }))
          afl));

rstep ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Pos, ((Term a [Char], Term a [Char]), Term a [Char]));
rstep xml2name =
  xml3elements ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 't', 'e', 'p'] pos
    (rule xml2name) (term xml2name) (\ p r t -> (p, (r, t)));

start_term ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] a) ->
                Term a [Char] -> Xml -> Parse_result [Char] (Term a [Char]);
start_term xml2name t =
  xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm']
    (\ xml ->
      bindc (term xml2name xml)
        (\ s ->
          (if equal_term s t then returna t
            else errora
                   ['<', 's', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm', '>', ' ',
                     'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'm', 'a', 't',
                     'c', 'h', ' ', 'l', 'h', 's'])))
    id;

proj_term :: forall a b. ((a, Nat) -> Nat) -> Term a b -> Term a b;
proj_term p (Var x) = Var x;
proj_term p (Fun f ts) =
  let {
    n = size_list ts;
    i = p (f, n);
  } in (if less_nat i n then nth ts i else Fun f ts);

rseq ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] a) ->
                ((a, Nat) -> Nat) ->
                  (Term a [Char], Term a [Char]) ->
                    Xml ->
                      Parse_result [Char]
                        ((Term a [Char], Term a [Char]),
                          [(Pos, ((Term a [Char], Term a [Char]),
                                   Term a [Char]))]);
rseq xml2name pia r =
  xml1many_elements
    ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e']
    (start_term xml2name (proj_term pia (fst r))) (rstep xml2name)
    (\ _ -> (\ a -> (r, a)));

match_vars ::
  forall a b c.
    (Eq a, Eq b, Eq c) => Term a b -> [(c, Term a b)] -> Maybe (c -> Term a b);
match_vars w [] = Just (\ _ -> w);
match_vars w ((x, t) : ps) =
  (case map_of ps x of {
    Nothing ->
      (case match_vars w ps of {
        Nothing -> Nothing;
        Just sigma -> Just (fun_upd sigma x t);
      });
    Just s -> (if equal_term t s then match_vars w ps else Nothing);
  });

match_term_list ::
  forall a b c. (Eq a) => [(Term a b, Term a c)] -> Maybe [(c, Term a b)];
match_term_list [] = Just [];
match_term_list ((Var uu, Fun uv uw) : ps) = Nothing;
match_term_list ((Fun f ts, Fun g ls) : ps) =
  bind (guard (f == g))
    (\ _ -> bind (zip_option ts ls) (\ eqs -> match_term_list (eqs ++ ps)));
match_term_list ((t, Var x) : ps) =
  bind (match_term_list ps) (\ eqs -> Just ((x, t) : eqs));

match_list ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => Term a b -> [(Term a b, Term a c)] -> Maybe (c -> Term a b);
match_list w tls =
  (case match_term_list tls of {
    Nothing -> Nothing;
    Just a -> match_vars w a;
  });

matcha ::
  forall a b c.
    (Eq a, Eq b, Eq c) => Term a b -> Term a c -> Maybe (c -> Term a b);
matcha t l = match_list t [(t, l)];

matches :: forall a b c. (Eq a, Eq b, Eq c) => Term a b -> Term a c -> Bool;
matches t l = (case matcha t l of {
                Nothing -> False;
                Just _ -> True;
              });

supteq_list :: forall a b. Term a b -> [Term a b];
supteq_list (Var x) = [Var x];
supteq_list (Fun f ts) = Fun f ts : concatMap supteq_list ts;

rm_alpha :: forall a b. (Linorder a) => Rbt a b -> a -> Maybe b;
rm_alpha = lookupa;

rm_set_lookup :: forall a b. (Linorder a) => Rbt a [b] -> a -> [b];
rm_set_lookup rm =
  (\ a -> (case rm_alpha rm a of {
            Nothing -> [];
            Just rules -> rules;
          }));

is_NF ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Bool -> [Term a b] -> Rbt a [Term a b] -> Term a b -> Bool;
is_NF var_cond r m =
  (if var_cond then (\ _ -> False)
    else (if null r then (\ _ -> True)
           else (\ t ->
                  all (\ u ->
                        all (\ l -> not (matches u l))
                          (rm_set_lookup m (fst (the (root u)))))
                    (supteq_list t))));

instance Semigroup_mult Integer where {
};

instance Ab_semigroup_add Integer where {
};

instance Semiring Integer where {
};

label :: forall a b. Lab a b -> Nat -> Sum b [Lab a b] -> Lab a b;
label f n (Inl l) = Lab f l;
label f n (Inr l) = FunLab f l;

unlab :: forall a b. Lab a b -> Lab a b;
unlab (Lab f l) = f;
unlab (FunLab f l) = f;
unlab (UnLab v) = UnLab v;
unlab (Sharp v) = Sharp v;

mat_max0 :: forall a. (Max_ordered_monoid_add a) => [[a]] -> [[a]];
mat_max0 = mat_map max0;

lookupb :: forall a b c d. Memory_ext a b c d -> a -> b -> Maybe c;
lookupb (Memory_ext empty lookupa store more) = lookupa;

it_map_to_list ::
  forall a b.
    (a -> ([b] -> Bool) -> (b -> [b] -> [b]) -> [b] -> [b]) -> a -> [b];
it_map_to_list iti m = iti m (\ _ -> True) (\ a b -> a : b) [];

rm_reverse_iterateoi ::
  forall a b c. Rbta a b -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c;
rm_reverse_iterateoi Emptya c f sigma = sigma;
rm_reverse_iterateoi (Branch col l k v r) c f sigma =
  (if c sigma
    then let {
           sigmaa = rm_reverse_iterateoi r c f sigma;
         } in (if c sigmaa then rm_reverse_iterateoi l c f (f (k, v) sigmaa)
                else sigmaa)
    else sigma);

rm_reverse_iterateoia ::
  forall a b c.
    (Linorder a) => Rbt a b -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c;
rm_reverse_iterateoia r = rm_reverse_iterateoi (impl_of r);

rm_to_list :: forall a b. (Linorder a) => Rbt a b -> [(a, b)];
rm_to_list = it_map_to_list rm_reverse_iterateoia;

values :: forall a b. (Linorder a) => Rbt a [b] -> [b];
values m = concatMap snd (rm_to_list m);

instance Linorder Nat where {
};

data Tpoly a b = PVar a | PNum b | PSum [Tpoly a b] | PMult [Tpoly a b];

instance Semigroup_mult Rat where {
};

instance Ab_semigroup_add Rat where {
};

instance Semiring Rat where {
};

class (Times a, Zero a) => Mult_zero a where {
};

s_memb :: forall a b c. (a -> b -> Maybe c) -> a -> b -> Bool;
s_memb lookupa x s = not (is_none (lookupa x s));

ctxt_apply :: forall a b. Ctxt a b -> Term a b -> Term a b;
ctxt_apply Hole s = s;
ctxt_apply (More f ss1 c ss2) s = Fun f (ss1 ++ ctxt_apply c s : ss2);

equal_ctxt :: forall a b. (Eq a, Eq b) => Ctxt a b -> Ctxt a b -> Bool;
equal_ctxt (More f list1 ctxt list2) Hole = False;
equal_ctxt Hole (More f list1 ctxt list2) = False;
equal_ctxt (More fa list1a ctxta list2a) (More f list1 ctxt list2) =
  fa == f && list1a == list1 && equal_ctxt ctxta ctxt && list2a == list2;
equal_ctxt Hole Hole = True;

hvf_top :: forall a b. (Eq a) => a -> Nat -> Term a b -> Bool;
hvf_top a n (Fun f ts) =
  (if f == a && equal_nat (size_list ts) n then not (is_Var (hd ts)) else True);
hvf_top a n (Var uu) = False;

list_inter :: forall a. (Eq a) => [a] -> [a] -> [a];
list_inter [] bs = [];
list_inter (a : asa) bs =
  (if membera bs a then a : list_inter asa bs else list_inter asa bs);

list_union :: forall a. (Eq a) => [a] -> [a] -> [a];
list_union [] ys = ys;
list_union (x : xs) ys =
  let {
    zs = list_union xs ys;
  } in (if membera zs x then zs else x : zs);

bindb :: forall a b c. Sum a b -> (b -> Sum a c) -> Sum a c;
bindb m f = (case m of {
              Inl a -> Inl a;
              Inr a -> f a;
            });

binda ::
  forall a b c.
    ([a] -> Sum [Char] (b, [a])) ->
      (b -> [a] -> Sum [Char] (c, [a])) -> [a] -> Sum [Char] (c, [a]);
binda m f ts = bindb (m ts) (\ (a, b) -> f a b);

many :: (Char -> Bool) -> [Char] -> Sum [Char] ([Char], [Char]);
many p [] = Inr ([], []);
many p (t : ts) =
  (if p t then bindb (many p ts) (\ (rs, tsa) -> Inr (t : rs, tsa))
    else Inr ([], t : ts));

trim :: [Char] -> [Char];
trim [] = [];
trim (c : s) = (if membera [' ', newline, '\t', '\r'] c then trim s else c : s);

data Domain = Naturala Nat | Arctic | Arctic_rat | Int_mat Nat Nat | Arc_mat Nat
  | Arc_rat_mat Nat | Rationala Rat Nat | Rat_mat Nat Nat;

estep ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])));
estep xml2name =
  xml4elements ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'S', 't', 'e', 'p'] pos
    (rule xml2name)
    (xml_options
      [(['l', 'e', 'f', 't', 'R', 'i', 'g', 'h', 't'],
         xml0elements ['l', 'e', 'f', 't', 'R', 'i', 'g', 'h', 't'] True),
        (['r', 'i', 'g', 'h', 't', 'L', 'e', 'f', 't'],
          xml0elements ['r', 'i', 'g', 'h', 't', 'L', 'e', 'f', 't'] False)])
    (term xml2name) (\ p r b t -> (p, (r, (b, t))));

rules ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [(Term a [Char], Term a [Char])];
rules xml2name = xml_many_elements ['r', 'u', 'l', 'e', 's'] (rule xml2name) id;

state :: Xml -> Parse_result [Char] [Char];
state = text ['s', 't', 'a', 't', 'e'];

xml_lab :: forall a b. (Showa a, Showa b) => Lab a [b] -> Xml;
xml_lab (UnLab x) =
  Xml ['n', 'a', 'm', 'e'] [] [] (Just (shows_prec Zero_nat x []));
xml_lab (Sharp x) = Xml ['s', 'h', 'a', 'r', 'p'] [] [xml_lab x] Nothing;
xml_lab (FunLab x l) =
  Xml ['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'] []
    [xml_lab x,
      Xml ['s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l'] []
        (map xml_lab l) Nothing]
    Nothing;
xml_lab (Lab x l) =
  Xml ['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'] []
    [xml_lab x,
      Xml ['n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l'] []
        (map (\ n ->
               Xml ['n', 'u', 'm', 'b', 'e', 'r'] [] []
                 (Just (shows_prec Zero_nat n [])))
          l)
        Nothing]
    Nothing;

xml_term ::
  forall a b c. (Showa a, Showa b, Showa c) => Term (Lab a [b]) c -> Xml;
xml_term (Var x) = Xml ['v', 'a', 'r'] [] [] (Just (shows_prec Zero_nat x []));
xml_term (Fun f ts) =
  Xml ['f', 'u', 'n', 'a', 'p', 'p'] []
    (xml_lab f : map (\ t -> Xml ['a', 'r', 'g'] [] [xml_term t] Nothing) ts)
    Nothing;

data Strategy a b = No_Strategy | Innermost | Innermost_Q [Term a b];

data Ostrategy a b = Outermost | Outermost_Q [Term a b];

xml_Q ::
  forall a b c.
    (Showa a, Showa b,
      Showa c) => Sum (Strategy (Lab a [b]) c) (Ostrategy (Lab a [b]) c) ->
                    [Xml];
xml_Q (Inl No_Strategy) = [];
xml_Q (Inl Innermost) =
  [Xml ['s', 't', 'r', 'a', 't', 'e', 'g', 'y'] []
     [Xml ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'] [] [] Nothing] Nothing];
xml_Q (Inr Outermost) =
  [Xml ['s', 't', 'r', 'a', 't', 'e', 'g', 'y'] []
     [Xml ['o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't'] [] [] Nothing] Nothing];
xml_Q (Inl (Innermost_Q q)) =
  [Xml ['s', 't', 'r', 'a', 't', 'e', 'g', 'y'] []
     [Xml ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's'] []
        (map xml_term q) Nothing]
     Nothing];

data Cr_proof a b c = SN_WCR (Join_info (Lab a b)) (Trs_proof a b c)
  | Weakly_Orthogonal;

update :: forall a b c. (a -> b) -> (a, c) -> (b, c);
update p_to_p (p, info) = (p_to_p p, info);

isOK :: forall a b. Sum a b -> Bool;
isOK m = (case m of {
           Inl _ -> False;
           Inr _ -> True;
         });

mapMa :: forall a b c. (a -> Sum b c) -> [a] -> Sum b [c];
mapMa f [] = Inr [];
mapMa f (x : xs) =
  bindb (f x) (\ y -> bindb (mapMa f xs) (\ ys -> Inr (y : ys)));

instance Mult_zero Integer where {
};

list_update :: forall a. [a] -> Nat -> a -> [a];
list_update (x : xs) (Suc i) y = x : list_update xs i y;
list_update (x : xs) Zero_nat y = y : xs;
list_update [] i y = [];

base :: forall a. (a, Nat) -> a;
base (f, h) = f;

lift :: forall a. Nat -> a -> (a, Nat);
lift h f = (f, h);

roof :: forall a b. (Eq b) => (Term a b, Term a b) -> Term a b -> Bool;
roof (l, r) =
  let {
    xs = vars_term_lista r;
  } in (\ t -> let {
                 xt = vars_term_lista t;
               } in all (membera xt) xs);

mat_monoI :: forall a. (a -> Bool) -> Nat -> [[a]] -> Bool;
mat_monoI geq1 sd m = all (any geq1) (sub_mat sd sd m);

transpose :: forall a. Nat -> [[a]] -> [[a]];
transpose nr [] = replicatea nr [];
transpose nr (v : m) = map (\ (a, b) -> a : b) (zip v (transpose nr m));

scalar_prodI ::
  forall a. a -> (a -> a -> a) -> (a -> a -> a) -> [a] -> [a] -> a;
scalar_prodI ze pl ti v w = foldra (\ (x, y) -> pl (ti x y)) (zip v w) ze;

matT_vec_multI ::
  forall a. a -> (a -> a -> a) -> (a -> a -> a) -> [[a]] -> [a] -> [a];
matT_vec_multI ze pl ti m v = map (\ w -> scalar_prodI ze pl ti w v) m;

mat_multI ::
  forall a.
    a -> (a -> a -> a) -> (a -> a -> a) -> Nat -> [[a]] -> [[a]] -> [[a]];
mat_multI ze pl ti nr m1 m2 =
  map (matT_vec_multI ze pl ti (transpose nr m1)) m2;

vec_plusI :: forall a. (a -> a -> a) -> [a] -> [a] -> [a];
vec_plusI pl v w = map (\ xy -> pl (fst xy) (snd xy)) (zip v w);

mat_plusI :: forall a. (a -> a -> a) -> [[a]] -> [[a]] -> [[a]];
mat_plusI pl m1 m2 = map (\ uv -> vec_plusI pl (fst uv) (snd uv)) (zip m1 m2);

equal_color :: Color -> Color -> Bool;
equal_color B R = False;
equal_color R B = False;
equal_color B B = True;
equal_color R R = True;

equal_rbta :: forall a b. (Eq a, Eq b) => Rbta a b -> Rbta a b -> Bool;
equal_rbta (Branch color rbt1 a b rbt2) Emptya = False;
equal_rbta Emptya (Branch color rbt1 a b rbt2) = False;
equal_rbta (Branch colora rbt1a aa ba rbt2a) (Branch color rbt1 a b rbt2) =
  equal_color colora color &&
    equal_rbta rbt1a rbt1 && aa == a && ba == b && equal_rbta rbt2a rbt2;
equal_rbta Emptya Emptya = True;

equal_rbt :: forall a b. (Eq a, Linorder a, Eq b) => Rbt a b -> Rbt a b -> Bool;
equal_rbt ra r = equal_rbta (impl_of ra) (impl_of r);

instance (Eq a, Linorder a, Eq b) => Eq (Rbt a b) where {
  a == b = equal_rbt a b;
};

instance Mult_zero Rat where {
};

class (Ab_semigroup_add a, Monoid_add a) => Comm_monoid_add a where {
};

class (Comm_monoid_add a, Mult_zero a, Semiring a) => Semiring_0 a where {
};

class (One a, Zero a) => Zero_neq_one a where {
};

class (Semigroup_mult a, Power a) => Monoid_mult a where {
};

class (Monoid_mult a, Numeral a, Semiring a) => Semiring_numeral a where {
};

class (Semiring_numeral a, Semiring_0 a, Zero_neq_one a) => Semiring_1 a where {
};

fun_ofa :: forall a b. (Eq b) => a -> [(b, a)] -> b -> a;
fun_ofa def xs a = (case map_of xs a of {
                     Nothing -> def;
                     Just b -> b;
                   });

linear_term_impl :: forall a b. (Eq a) => [a] -> Term b a -> Maybe [a];
linear_term_impl xs (Var x) = (if membera xs x then Nothing else Just (x : xs));
linear_term_impl xs (Fun uu []) = Just xs;
linear_term_impl xs (Fun f (t : ts)) =
  (case linear_term_impl xs t of {
    Nothing -> Nothing;
    Just ys -> linear_term_impl ys (Fun f ts);
  });

linear_term :: forall a b. (Eq b) => Term a b -> Bool;
linear_term t =
  (case linear_term_impl [] t of {
    Nothing -> False;
    Just _ -> True;
  });

map_term_wa :: forall a b c. (a -> Nat -> b) -> Term a c -> Term b c;
map_term_wa fg (Var x) = Var x;
map_term_wa fg (Fun f ts) = Fun (fg f (size_list ts)) (map (map_term_wa fg) ts);

rq :: forall a b. Ta_rule a b -> a;
rq (TA_rule f qs q) = q;

poss_list :: forall a b. Term a b -> [Pos];
poss_list (Var x) = [Empty];
poss_list (Fun f ss) =
  Empty :
    concatMap (\ (i, a) -> map (PCons i) a)
      (zip (upt Zero_nat (size_list ss)) (map poss_list ss));

ctxt_of_pos_term :: forall a b. Pos -> Term a b -> Ctxt a b;
ctxt_of_pos_term Empty t = Hole;
ctxt_of_pos_term (PCons i ps) (Fun f ts) =
  More f (takea i ts) (ctxt_of_pos_term ps (nth ts i)) (dropa (Suc i) ts);

subst_apply_term :: forall a b c. Term a b -> (b -> Term a c) -> Term a c;
subst_apply_term (Var x) sigma = sigma x;
subst_apply_term (Fun f ss) sigma =
  Fun f (map (\ t -> subst_apply_term t sigma) ss);

rewrite ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> [Term a b];
rewrite r s =
  concatMap
    (\ (l, ra) ->
      concatMap
        (\ p ->
          (case matcha (subt_at s p) l of {
            Nothing -> [];
            Just sigma ->
              [ctxt_apply (ctxt_of_pos_term p s) (subst_apply_term ra sigma)];
          }))
        (poss_list s))
    r;

hvf_term :: forall a b. (Eq a) => a -> Term a b -> Bool;
hvf_term a t =
  (case unapp a t of {
    (Var _, ts) -> null ts;
    (Fun _ us, ts) -> all (hvf_term a) (us ++ ts);
  });

oneof_aux :: [[Char]] -> [[Char]] -> [Char] -> Sum [Char] ([Char], [Char]);
oneof_aux allowed [] ts =
  expecting (['o', 'n', 'e', ' ', 'o', 'f', ' '] ++ shows_list_list allowed [])
    ts;
oneof_aux allowed (x : xs) ts =
  (if map snd (zip x ts) == x then Inr (x, trim (dropa (size_list x) ts))
    else oneof_aux allowed xs ts);

oneof :: [[Char]] -> [Char] -> Sum [Char] ([Char], [Char]);
oneof xs = oneof_aux xs xs;

data Xmldoc = Xmldoc [Char] Xml;

xml5elements ::
  forall a b c d e f.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (Xml -> Parse_result [Char] d) ->
              (Xml -> Parse_result [Char] e) ->
                (a -> b -> c -> d -> e -> f) -> Xml -> Parse_result [Char] f;
xml5elements tag p1 p2 p3 p4 p5 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) (nat 5) && is_none text
    then bindc (p1 (nth cs Zero_nat))
           (\ x1 ->
             bindc (p2 (nth cs one_nat))
               (\ x2 ->
                 bindc (p3 (nth cs (nat 2)))
                   (\ x3 ->
                     bindc (p4 (nth cs (nat 3)))
                       (\ x4 ->
                         bindc (p5 (nth cs (nat 4)))
                           (\ x5 -> returna (f x1 x2 x3 x4 x5))))))
    else no_transform tag (Xml name atts cs text));

xml6elements ::
  forall a b c d e f g.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (Xml -> Parse_result [Char] d) ->
              (Xml -> Parse_result [Char] e) ->
                (Xml -> Parse_result [Char] f) ->
                  (a -> b -> c -> d -> e -> f -> g) ->
                    Xml -> Parse_result [Char] g;
xml6elements tag p1 p2 p3 p4 p5 p6 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && equal_nat (size_list cs) (nat 6) && is_none text
    then bindc (p1 (nth cs Zero_nat))
           (\ x1 ->
             bindc (p2 (nth cs one_nat))
               (\ x2 ->
                 bindc (p3 (nth cs (nat 2)))
                   (\ x3 ->
                     bindc (p4 (nth cs (nat 3)))
                       (\ x4 ->
                         bindc (p5 (nth cs (nat 4)))
                           (\ x5 ->
                             bindc (p6 (nth cs (nat 5)))
                               (\ x6 -> returna (f x1 x2 x3 x4 x5 x6)))))))
    else no_transform tag (Xml name atts cs text));

rsteps ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Term a [Char],
            [(Pos, ((Term a [Char], Term a [Char]), Term a [Char]))]);
rsteps xml2name =
  xml1many_elements
    ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e']
    (xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'] (term xml2name)
      id)
    (rstep xml2name) (\ a b -> (a, b));

data Non_join_info a b = Diff_NFs
  | Tcap_Non_Unif (Term a b -> Term a b -> b -> Term a b);

data Ncr_proof a b c = SN_NWCR (Trs_proof a b c)
  | Non_Join (Term (Lab a b) c)
      [(Pos, ((Term (Lab a b) c, Term (Lab a b) c), Term (Lab a b) c))]
      [(Pos, ((Term (Lab a b) c, Term (Lab a b) c), Term (Lab a b) c))]
      (Non_join_info (Lab a b) c);

missing :: [Char] -> ([Char] -> [Char]) -> [Char] -> [Char];
missing s x =
  shows_string ['t', 'h', 'e', ' '] .
    shows_string s .
      shows_string [' '] .
        x . shows_string
              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'];

instance (Eq a, Eq b) => Eq (Lab a b) where {
  a == b = equal_lab a b;
};

equal_lab :: forall a b. (Eq a, Eq b) => Lab a b -> Lab a b -> Bool;
equal_lab (Sharp lab) (UnLab f) = False;
equal_lab (UnLab f) (Sharp lab) = False;
equal_lab (Sharp laba) (FunLab lab list) = False;
equal_lab (FunLab laba list) (Sharp lab) = False;
equal_lab (UnLab f) (FunLab lab list) = False;
equal_lab (FunLab lab list) (UnLab f) = False;
equal_lab (Sharp laba) (Lab lab l) = False;
equal_lab (Lab laba l) (Sharp lab) = False;
equal_lab (UnLab f) (Lab lab l) = False;
equal_lab (Lab lab l) (UnLab f) = False;
equal_lab (FunLab laba list) (Lab lab l) = False;
equal_lab (Lab laba l) (FunLab lab list) = False;
equal_lab (Sharp laba) (Sharp lab) = equal_lab laba lab;
equal_lab (UnLab fa) (UnLab f) = fa == f;
equal_lab (FunLab laba lista) (FunLab lab list) =
  equal_lab laba lab && lista == list;
equal_lab (Lab laba la) (Lab lab l) = equal_lab laba lab && la == l;

no_decr ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => [(Term (Lab a b) c, Term (Lab a b) c)] ->
                 [(Term (Lab a b) c, Term (Lab a b) c)];
no_decr =
  filter
    (\ (l, r) ->
      not (not (equal_term l r) &&
            equal_term (map_term unlab l) (map_term unlab r)));

toomuch :: [Char] -> ([Char] -> [Char]) -> [Char] -> [Char];
toomuch s x =
  shows_string ['s', 'u', 'p', 'e', 'r', 'f', 'l', 'u', 'o', 'u', 's', ' '] .
    shows_string s . shows_string [' '] . x . x;

newtype Complexity_class = Comp_Poly Nat;

degree :: Complexity_class -> Nat;
degree (Comp_Poly d) = d;

catcha :: forall a b c. Sum a b -> (a -> Sum c b) -> Sum c b;
catcha m f = (case m of {
               Inl a -> f a;
               Inr a -> Inr a;
             });

forallm :: forall a b. (a -> Sum b ()) -> [a] -> Sum (a, b) ();
forallm f [] = Inr ();
forallm f (x : xs) =
  bindb (catcha (f x) (\ xa -> Inl (x, xa))) (\ _ -> forallm f xs);

check :: forall a. Bool -> a -> Sum a ();
check b e = (if b then Inr () else Inl e);

check_edges ::
  forall a.
    (a -> [Char] -> [Char]) ->
      ((a, a) -> Bool) -> [a] -> [a] -> Sum ([Char] -> [Char]) ();
check_edges ss g c d =
  catcha
    (forallm
      (\ n ->
        catcha
          (forallm
            (\ m ->
              check (not (g (m, n)))
                (shows_string
                   ['e', 'd', 'g', 'e', ' ', 'f', 'r', 'o', 'm', ' '] .
                  ss m . shows_string [' ', 't', 'o', ' '] . ss n))
            d)
          (\ x -> Inl (snd x)))
      c)
    (\ x -> Inl (snd x));

instance Monoid_add Integer where {
};

instance Comm_monoid_add Integer where {
};

instance Semiring_0 Integer where {
};

instance Zero_neq_one Integer where {
};

instance Monoid_mult Integer where {
};

instance Semiring_numeral Integer where {
};

instance Semiring_1 Integer where {
};

match :: forall a b. (Term a b, Term a b) -> Term a b -> Bool;
match = (\ _ _ -> True);

instance (Eq a, Linorder a, Eq b, Linorder b) => Linorder (a, b) where {
};

mapMb :: forall a b. (a -> Maybe b) -> [a] -> Maybe [b];
mapMb f [] = Just [];
mapMb f (x : xs) =
  bind (f x) (\ y -> bind (mapMb f xs) (\ ys -> Just (y : ys)));

equal_pos :: Pos -> Pos -> Bool;
equal_pos (PCons nat pos) Empty = False;
equal_pos Empty (PCons nat pos) = False;
equal_pos (PCons nata posa) (PCons nat pos) =
  equal_nat nata nat && equal_pos posa pos;
equal_pos Empty Empty = True;

instance Eq Pos where {
  a == b = equal_pos a b;
};

rm_iterateoi ::
  forall a b c. Rbta a b -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c;
rm_iterateoi Emptya c f sigma = sigma;
rm_iterateoi (Branch col l k v r) c f sigma =
  (if c sigma
    then let {
           sigmaa = rm_iterateoi l c f sigma;
         } in (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa)
                else sigmaa)
    else sigma);

rm_iterateoia ::
  forall a b c.
    (Linorder a) => Rbt a b -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c;
rm_iterateoia r = rm_iterateoi (impl_of r);

rm_iteratei ::
  forall a b c.
    (Linorder a) => Rbt a b -> (c -> Bool) -> ((a, b) -> c -> c) -> c -> c;
rm_iteratei = rm_iterateoia;

rm_add :: forall a b. (Linorder a) => Rbt a b -> Rbt a b -> Rbt a b;
rm_add = it_add rm_update rm_iteratei;

rs_ins :: forall a. (Linorder a) => a -> Rbt a () -> Rbt a ();
rs_ins = s_ins rm_update;

equal_order_tag :: Order_tag -> Order_tag -> Bool;
equal_order_tag Mul Lex = False;
equal_order_tag Lex Mul = False;
equal_order_tag Mul Mul = True;
equal_order_tag Lex Lex = True;

diag_l :: forall a b. (a, b) -> ((a, a), b);
diag_l (a, b) = ((a, a), b);

lex_ext_efficient_m ::
  forall a b.
    (a -> a -> b -> ((Bool, Bool), b)) -> [a] -> [a] -> b -> ((Bool, Bool), b);
lex_ext_efficient_m f [] [] m = ((False, True), m);
lex_ext_efficient_m f (uu : uv) [] m = ((True, True), m);
lex_ext_efficient_m f [] (uw : ux) m = ((False, False), m);
lex_ext_efficient_m f (a : asa) (b : bs) m =
  let {
    (aa, ba) = f a b m;
  } in let {
         (s, ns) = aa;
       } in (\ ma ->
              (if s then ((True, True), ma)
                else (if ns then lex_ext_efficient_m f asa bs ma
                       else ((False, False), ma))))
         ba;

filter_not_stri_efficient_m_aux ::
  forall a b. (a -> b -> ((Bool, Bool), b)) -> [a] -> [a] -> b -> ([a], b);
filter_not_stri_efficient_m_aux f [] bs m = (bs, m);
filter_not_stri_efficient_m_aux f (a : asa) bs m =
  let {
    (b, c) = f a m;
  } in let {
         (s, _) = b;
       } in (\ ma ->
              (if s then filter_not_stri_efficient_m_aux f asa bs ma
                else filter_not_stri_efficient_m_aux f asa (a : bs) ma))
         c;

filter_not_stri_efficient_m ::
  forall a b. (a -> b -> ((Bool, Bool), b)) -> [a] -> b -> ([a], b);
filter_not_stri_efficient_m f asa m =
  let {
    (ys, a) = filter_not_stri_efficient_m_aux f asa [] m;
  } in (reverse ys, a);

mul_ext_efficient_m ::
  forall a b.
    (Eq a) => (a -> a -> b -> ((Bool, Bool), b)) ->
                [a] -> [a] -> b -> ((Bool, Bool), b);
mul_ext_efficient_m f [] [] m = ((False, True), m);
mul_ext_efficient_m f [] (v : va) m = ((False, False), m);
mul_ext_efficient_m f (v : va) [] m = ((True, True), m);
mul_ext_efficient_m f (v : va) (y : ys) m =
  ex_dom_efficient_m f y (v : va) (v : va) ys m;

ex_dom_efficient_m ::
  forall a b.
    (Eq a) => (a -> a -> b -> ((Bool, Bool), b)) ->
                a -> [a] -> [a] -> [a] -> b -> ((Bool, Bool), b);
ex_dom_efficient_m f y [] uw ux m = ((False, False), m);
ex_dom_efficient_m f y (b : bs) xs ys m =
  let {
    (a, c) = f b y m;
  } in let {
         (s, ns) = a;
       } in (\ ma ->
              (if s then let {
                           (ysa, m2) = filter_not_stri_efficient_m (f b) ys ma;
                           (aa, ba) =
                             mul_ext_efficient_m f (remove1 b xs) ysa m2;
                         } in let {
                                (_, ns1) = aa;
                              } in (\ m3 ->
                                     (if ns1 then ((True, True), m3)
                                       else ex_dom_efficient_m f y bs xs ys m3))
                                ba
                else (if ns
                       then let {
                              (aa, ba) =
                                mul_ext_efficient_m f (remove1 b xs) ys ma;
                            } in let {
                                   (s1, ns1) = aa;
                                 } in (\ mb ->
(if s1 then ((True, True), mb)
  else let {
         (ab, bb) = ex_dom_efficient_m f y bs xs ys mb;
       } in let {
              (s2, ns2) = ab;
            } in (\ m3 ->
                   (if s2 then ((True, True), m3)
                     else ((False, ns1 || ns2), m3)))
              bb))
                                   ba
                       else ex_dom_efficient_m f y bs xs ys ma)))
         c;

all_stri_efficient_m ::
  forall a b. (a -> b -> ((Bool, Bool), b)) -> [a] -> b -> (Bool, b);
all_stri_efficient_m f [] m = (True, m);
all_stri_efficient_m f (a : asa) m =
  let {
    (aa, b) = f a m;
  } in let {
         (s, _) = aa;
       } in (\ ma -> (if s then all_stri_efficient_m f asa ma else (False, ma)))
         b;

any_nstri_efficient_m ::
  forall a b. (a -> b -> ((Bool, Bool), b)) -> [a] -> b -> (Bool, b);
any_nstri_efficient_m f [] m = (False, m);
any_nstri_efficient_m f (a : asa) m =
  let {
    (aa, b) = f a m;
  } in let {
         (_, ns) = aa;
       } in (\ ma ->
              (if ns then (True, ma) else any_nstri_efficient_m f asa ma))
         b;

rpo_efficient_m ::
  forall a b c.
    (Eq b,
      Eq c) => Memory_ext a (Term b c, Term b c) (Bool, Bool) () ->
                 (b -> Nat -> b -> Nat -> (Bool, Bool), b -> Bool) ->
                   (b -> Nat -> Order_tag) ->
                     Term b c -> Term b c -> a -> ((Bool, Bool), a);
rpo_efficient_m model pr tag s t m =
  (case lookupb model m (s, t) of {
    Nothing ->
      let {
        (res, res_m) =
          (case (s, t) of {
            (Var x, Var y) -> ((False, x == y), m);
            (Var x, Fun g ts) -> ((False, null ts && snd pr g), m);
            (Fun f ss, Var y) ->
              diag_l
                (any_nstri_efficient_m
                  (\ sa -> rpo_efficient_m model pr tag sa (Var y)) ss m);
            (Fun f ss, Fun g ts) ->
              (case any_nstri_efficient_m
                      (\ sa -> rpo_efficient_m model pr tag sa (Fun g ts)) ss m
                of {
                (True, ma) -> ((True, True), ma);
                (False, ma) ->
                  let {
                    maa = size_list ss;
                    n = size_list ts;
                  } in (case fst pr f maa g n of {
                         (prs, True) ->
                           (case all_stri_efficient_m
                                   (rpo_efficient_m model pr tag (Fun f ss)) ts
                                   ma
                             of {
                             (True, mb) ->
                               (if prs then ((True, True), mb)
                                 else let {
cf = tag f maa;
                                      } in
(if equal_order_tag cf (tag g n)
  then (if equal_order_tag cf Mul
         then mul_ext_efficient_m (rpo_efficient_m model pr tag) ss ts mb
         else lex_ext_efficient_m (rpo_efficient_m model pr tag) ss ts mb)
  else let {
         b = equal_nat n Zero_nat;
       } in ((not (equal_nat maa Zero_nat) && b, b), mb)));
                             (False, mb) -> ((False, False), mb);
                           });
                         (prs, False) -> ((False, False), ma);
                       });
              });
          });
      } in (res, store model res_m ((s, t), res));
    Just v -> (v, m);
  });

efficient_rpo_2 ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Nat -> a -> Nat -> (Bool, Bool), a -> Bool) ->
                 (a -> Nat -> Order_tag) ->
                   Term a b -> Term a b -> (Bool, Bool);
efficient_rpo_2 pr tag p q =
  fst (rpo_efficient_m l2m pr tag p q (emptya l2m ()));

rpo_unbounded_impl ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Nat -> a -> Nat -> (Bool, Bool), a -> Bool) ->
                 (a -> Nat -> Order_tag) ->
                   Term a b -> Term a b -> (Bool, Bool);
rpo_unbounded_impl = efficient_rpo_2;

rpo_unbounded ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Nat -> a -> Nat -> (Bool, Bool), a -> Bool) ->
                 (a -> Nat -> Order_tag) ->
                   Term a b -> Term a b -> (Bool, Bool);
rpo_unbounded = rpo_unbounded_impl;

instance Monoid_add Rat where {
};

instance Comm_monoid_add Rat where {
};

instance Semiring_0 Rat where {
};

instance Zero_neq_one Rat where {
};

instance Monoid_mult Rat where {
};

instance Semiring_numeral Rat where {
};

instance Semiring_1 Rat where {
};

equal_sum :: forall a b. (Eq a, Eq b) => Sum a b -> Sum a b -> Bool;
equal_sum (Inr b) (Inl a) = False;
equal_sum (Inl a) (Inr b) = False;
equal_sum (Inr ba) (Inr b) = ba == b;
equal_sum (Inl aa) (Inl a) = aa == a;

instance (Eq a, Eq b) => Eq (Sum a b) where {
  a == b = equal_sum a b;
};

subt_at_ctxt :: forall a b. Ctxt a b -> Pos -> Ctxt a b;
subt_at_ctxt c Empty = c;
subt_at_ctxt (More f bef c aft) (PCons i p) = subt_at_ctxt c p;

in_poss :: forall a b. Pos -> Term a b -> Bool;
in_poss Empty uu = True;
in_poss (PCons i p) (Fun f ts) =
  less_nat i (size_list ts) && in_poss p (nth ts i);
in_poss (PCons i p) (Var uv) = False;

data Complexity_measure a b = Derivational_Complexity [(a, Nat)]
  | Runtime_Complexity [(a, Nat)] [(a, Nat)];

data Redtriple_ext a b c =
  Redtriple_ext (Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((a, Nat) -> Nat -> Bool)
    ([(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ()) ([Char] -> [Char])
    (Complexity_measure a b -> Maybe Complexity_class) c;

s :: forall a b c.
       Redtriple_ext a b c -> (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
s (Redtriple_ext valid s ns nst af mono desc cpx more) = s;

instance_rule ::
  forall a b c.
    (Eq a, Eq b, Eq c) => (Term a b, Term a b) -> (Term a c, Term a c) -> Bool;
instance_rule lr st =
  not (is_none (match_list (fst lr) [(fst lr, fst st), (snd lr, snd st)]));

gen_label :: forall a b. Lab a b -> Nat -> Lab a b;
gen_label f Zero_nat = f;
gen_label f (Suc n) = FunLab (gen_label f n) [];

label_depth :: forall a b. Lab a b -> Nat;
label_depth (UnLab uu) = Zero_nat;
label_depth (Lab f uv) = Suc (label_depth f);
label_depth (FunLab f uw) = Suc (label_depth f);
label_depth (Sharp f) = Suc (label_depth f);

fmapa ::
  forall a b.
    (Eq a,
      Eq b) => Lab a b ->
                 Nat ->
                   [((Lab a b, Nat), [Lab a b])] -> Lab a b -> Nat -> Lab a b;
fmapa a nn sml =
  let {
    m = Suc (max_list (map label_depth (a : concatMap snd sml)));
  } in (\ f n -> (if (f, n) == (a, nn) then a else gen_label f m));

enum_vectors :: forall a b. [a] -> [b] -> [[(b, a)]];
enum_vectors c [] = [[]];
enum_vectors c (x : xs) =
  let {
    a = enum_vectors c xs;
  } in concatMap (\ vec -> map (\ ca -> (x, ca) : vec) c) a;

replace_impl :: forall a. (Eq a) => a -> [a] -> [a] -> [a];
replace_impl a bs m =
  (if membera m a then bs ++ filter (\ b -> not (b == a)) m else m);

xml2nd_choice ::
  forall a b c.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        [Char] ->
          (Xml -> Parse_result [Char] b) ->
            (a -> Maybe b -> c) -> Xml -> Parse_result [Char] c;
xml2nd_choice tag p1 cn p2 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag && null atts && less_eq_nat (nat 2) l && is_none text
         then bindc (p1 (nth cs Zero_nat))
                (\ x1 ->
                  bindc (bool cn (nth cs one_nat))
                    (\ b ->
                      (if b then bindc (p2 (nth cs (minus_nat l one_nat)))
                                   (\ x2 -> returna (f x1 (Just x2)))
                        else returna (f x1 Nothing))))
         else no_transform tag (Xml name atts cs text));

alter1a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> a;
alter1a (Alter1 x) = x;

alter2a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> b;
alter2a (Alter2 x) = x;

alter3a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> c;
alter3a (Alter3 x) = x;

alter4a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> d;
alter4a (Alter4 x) = x;

alter5a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> e;
alter5a (Alter5 x) = x;

alter6a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> f;
alter6a (Alter6 x) = x;

alter7a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> g;
alter7a (Alter7 x) = x;

alter8a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> h;
alter8a (Alter8 x) = x;

alter9a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> i;
alter9a (Alter9 x) = x;

pat_term ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Term a [Char],
            ([([Char], Term a [Char])], [([Char], Term a [Char])]));
pat_term xml2name =
  xml3elements ['p', 'a', 't', 't', 'e', 'r', 'n', 'T', 'e', 'r', 'm']
    (term xml2name) (subst xml2name) (subst xml2name)
    (\ t s1 s2 -> (t, (s1, s2)));

data Pat_rule_pos = Pat_Base | Pat_Pump | Pat_Close;

data Pat_eqv_prf a b = Pat_Dom_Renaming [(b, Term a b)]
  | Pat_Irrelevant [(b, Term a b)] [(b, Term a b)]
  | Pat_Simplify [(b, Term a b)] [(b, Term a b)];

data Pat_rule_prf a b = Pat_OrigRule (Term a b, Term a b) Bool
  | Pat_InitPump (Pat_rule_prf a b) [(b, Term a b)] [(b, Term a b)]
  | Pat_InitPumpCtxt (Pat_rule_prf a b) [(b, Term a b)] Pos b
  | Pat_Equiv (Pat_rule_prf a b) Bool (Pat_eqv_prf a b)
  | Pat_Narrow (Pat_rule_prf a b) (Pat_rule_prf a b) Pos
  | Pat_Inst (Pat_rule_prf a b) [(b, Term a b)] Pat_rule_pos
  | Pat_Rewr (Pat_rule_prf a b)
      (Term a b, [(Pos, ((Term a b, Term a b), Term a b))]) Pat_rule_pos b;

pat_eqv_prf ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Pat_eqv_prf a [Char]);
pat_eqv_prf xml2name =
  let {
    sub = subst xml2name;
  } in xml1element
         ['p', 'a', 't', 't', 'e', 'r', 'n', 'E', 'q', 'u', 'i', 'v', 'a', 'l',
           'e', 'n', 'c', 'e']
         (xml_options
           [(['d', 'o', 'm', 'a', 'i', 'n', 'R', 'e', 'n', 'a', 'm', 'i', 'n',
               'g'],
              xml1element
                ['d', 'o', 'm', 'a', 'i', 'n', 'R', 'e', 'n', 'a', 'm', 'i',
                  'n', 'g']
                sub Pat_Dom_Renaming),
             (['i', 'r', 'r', 'e', 'l', 'e', 'v', 'a', 'n', 't'],
               xml2elements ['i', 'r', 'r', 'e', 'l', 'e', 'v', 'a', 'n', 't']
                 sub sub Pat_Irrelevant),
             (['s', 'i', 'm', 'p', 'l', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o',
                'n'],
               xml2elements
                 ['s', 'i', 'm', 'p', 'l', 'i', 'f', 'i', 'c', 'a', 't', 'i',
                   'o', 'n']
                 sub sub Pat_Simplify)])
         id;

pat_rule_prf ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Pat_rule_prf a [Char]);
pat_rule_prf xml2name x =
  let {
    pat = pat_rule_prf xml2name;
    sub = subst xml2name;
    pt = pat_term xml2name;
    var = text ['v', 'a', 'r'];
  } in xml3elements ['p', 'a', 't', 't', 'e', 'r', 'n', 'R', 'u', 'l', 'e'] pt
         pt (xml_options
              [(['o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', 'R', 'u', 'l', 'e'],
                 xml2elements
                   ['o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', 'R', 'u', 'l', 'e']
                   (rule xml2name) (bool ['i', 's', 'P', 'a', 'i', 'r'])
                   Pat_OrigRule),
                (['i', 'n', 'i', 't', 'i', 'a', 'l', 'P', 'u', 'm', 'p', 'i',
                   'n', 'g'],
                  xml3elements
                    ['i', 'n', 'i', 't', 'i', 'a', 'l', 'P', 'u', 'm', 'p', 'i',
                      'n', 'g']
                    pat sub sub Pat_InitPump),
                (['i', 'n', 'i', 't', 'i', 'a', 'l', 'P', 'u', 'm', 'p', 'i',
                   'n', 'g', 'C', 'o', 'n', 't', 'e', 'x', 't'],
                  xml4elements
                    ['i', 'n', 'i', 't', 'i', 'a', 'l', 'P', 'u', 'm', 'p', 'i',
                      'n', 'g', 'C', 'o', 'n', 't', 'e', 'x', 't']
                    pat sub pos var Pat_InitPumpCtxt),
                (['e', 'q', 'u', 'i', 'v', 'a', 'l', 'e', 'n', 'c', 'e'],
                  xml3elements
                    ['e', 'q', 'u', 'i', 'v', 'a', 'l', 'e', 'n', 'c', 'e'] pat
                    (xml_options
                      [(['l', 'e', 'f', 't'],
                         xml0elements ['l', 'e', 'f', 't'] True),
                        (['r', 'i', 'g', 'h', 't'],
                          xml0elements ['r', 'i', 'g', 'h', 't'] False)])
                    (pat_eqv_prf xml2name) Pat_Equiv),
                (['n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g'],
                  xml3elements ['n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g'] pat
                    pat pos Pat_Narrow),
                (['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o',
                   'n'],
                  xml3elements
                    ['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o',
                      'n']
                    pat sub
                    (xml_options
                      [(['b', 'a', 's', 'e'],
                         xml0elements ['b', 'a', 's', 'e'] Pat_Base),
                        (['p', 'u', 'm', 'p', 'i', 'n', 'g'],
                          xml0elements ['p', 'u', 'm', 'p', 'i', 'n', 'g']
                            Pat_Pump),
                        (['c', 'l', 'o', 's', 'i', 'n', 'g'],
                          xml0elements ['c', 'l', 'o', 's', 'i', 'n', 'g']
                            Pat_Close)])
                    Pat_Inst),
                (['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g'],
                  xml3elements ['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g'] pat
                    (rsteps xml2name)
                    (xml_options
                      [(['b', 'a', 's', 'e'],
                         xml0elements ['b', 'a', 's', 'e'] (Pat_Base, [])),
                        (['p', 'u', 'm', 'p', 'i', 'n', 'g'],
                          xml1element ['p', 'u', 'm', 'p', 'i', 'n', 'g'] var
                            (\ a -> (Pat_Pump, a))),
                        (['c', 'l', 'o', 's', 'i', 'n', 'g'],
                          xml1element ['c', 'l', 'o', 's', 'i', 'n', 'g'] var
                            (\ a -> (Pat_Close, a)))])
                    (\ p seqa (a, b) -> Pat_Rewr p seqa a b))])
         (\ _ _ z -> z) x;

data Non_loop_prf a b =
  Non_Loop_Prf (Pat_rule_prf a b) [(b, Term a b)] [(b, Term a b)] Nat Nat Pos;

nonloop ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Non_loop_prf a [Char]);
nonloop xml2name =
  xml6elements ['n', 'o', 'n', 'L', 'o', 'o', 'p'] (pat_rule_prf xml2name)
    (subst xml2name) (subst xml2name) (nata ['n', 'a', 't', 'u', 'r', 'a', 'l'])
    (nata ['n', 'a', 't', 'u', 'r', 'a', 'l']) pos Non_Loop_Prf;

data Dpp_ops_ext a b c d =
  Dpp_ops_ext
    (a -> (Bool,
            (Bool,
              (Set (Term b c, Term b c),
                (Set (Term b c, Term b c),
                  (Set (Term b c),
                    (Set (Term b c, Term b c), Set (Term b c, Term b c))))))))
    (a -> [(Term b c, Term b c)]) (a -> [(Term b c, Term b c)])
    (a -> [(Term b c, Term b c)]) (a -> [Term b c])
    (a -> [(Term b c, Term b c)]) (a -> [(Term b c, Term b c)])
    (a -> [(Term b c, Term b c)]) (a -> Bool) (a -> Bool) (a -> Bool)
    (a -> Term b c -> Bool) (a -> Bool)
    (a -> (b, Nat) -> [(Term b c, Term b c)])
    (a -> (b, Nat) -> [(Term b c, Term b c)]) (a -> [(Term b c, Term b c)] -> a)
    (a -> (Term b c, Term b c) -> [(Term b c, Term b c)] -> a)
    (a -> [(Term b c, Term b c)] -> a)
    (a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a)
    (a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a)
    (a -> [(Term b c, Term b c)] ->
            ([(Term b c, Term b c)], [(Term b c, Term b c)]))
    (a -> [(Term b c, Term b c)] ->
            ([(Term b c, Term b c)], [(Term b c, Term b c)]))
    (Bool ->
      Bool ->
        [(Term b c, Term b c)] ->
          [(Term b c, Term b c)] ->
            [Term b c] -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a)
    (a -> Bool) (a -> Bool) (a -> Bool) d;

rules_mapb ::
  forall a b c d.
    Dpp_ops_ext a b c d -> a -> (b, Nat) -> [(Term b c, Term b c)];
rules_mapb
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = rules_map;

rules_no_left_vara :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
rules_no_left_vara
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = rules_no_left_var;

capRM_dpp ::
  forall a b c.
    (Eq b, Eq c) => Dpp_ops_ext a b c () -> a -> Term b c -> Gctxt b c;
capRM_dpp i dpp = capRM (rules_no_left_vara i dpp) (rules_mapb i dpp);

rm_empty :: forall a b. (Linorder a) => () -> Rbt a b;
rm_empty = (\ _ -> empty);

elem_list_to_rm :: forall a b. (Linorder b) => (a -> b) -> [a] -> Rbt b [a];
elem_list_to_rm key [] = rm_empty ();
elem_list_to_rm key (d : ds) =
  let {
    rm = elem_list_to_rm key ds;
    k = key d;
  } in (case rm_alpha rm k of {
         Nothing -> rm_update_dj k [d] rm;
         Just dataa -> rm_update k (d : dataa) rm;
       });

term_map :: forall a b. (Linorder a) => [Term a b] -> Rbt a [Term a b];
term_map ts = elem_list_to_rm ((fst . the) . root) ts;

height :: forall a. (a, Nat) -> Nat;
height (f, h) = h;

extracta :: forall a. (a -> Bool) -> [a] -> Maybe ([a], (a, [a]));
extracta p [] = Nothing;
extracta p (x : xs) =
  (if p x then Just ([], (x, xs))
    else (case extracta p xs of {
           Nothing -> Nothing;
           Just (ys, (y, zs)) -> Just (x : ys, (y, zs));
         }));

one_poly :: forall a b. (Semiring_1 b) => [([(a, Nat)], b)];
one_poly = [([], one)];

eq_monom :: forall a. (Eq a) => [(a, Nat)] -> [(a, Nat)] -> Bool;
eq_monom [] n = null n;
eq_monom ((x, p) : m) n =
  (case extracta (\ yq -> fst yq == x) n of {
    Nothing -> False;
    Just (n1, ((_, q), n2)) -> equal_nat p q && eq_monom m (n1 ++ n2);
  });

poly_add ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => [([(a, Nat)], b)] ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
poly_add [] q = q;
poly_add ((m, c) : p) q =
  (case extracta (\ mc -> eq_monom (fst mc) m) q of {
    Nothing -> (m, c) : poly_add p q;
    Just (q1, ((_, d), q2)) ->
      (if plus c d == zero then poly_add p (q1 ++ q2)
        else (m, plus c d) : poly_add p (q1 ++ q2));
  });

monom_mult :: forall a. (Eq a) => [(a, Nat)] -> [(a, Nat)] -> [(a, Nat)];
monom_mult [] n = n;
monom_mult ((x, p) : m) n =
  (case extracta (\ yq -> fst yq == x) n of {
    Nothing -> (x, p) : monom_mult m n;
    Just (n1, ((_, q), n2)) -> (x, plus_nat p q) : monom_mult m (n1 ++ n2);
  });

monom_mult_poly ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => ([(a, Nat)], b) ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
monom_mult_poly uu [] = [];
monom_mult_poly (ma, c) ((m, d) : p) =
  (if times c d == zero then monom_mult_poly (ma, c) p
    else (monom_mult ma m, times c d) : monom_mult_poly (ma, c) p);

poly_mult ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => [([(a, Nat)], b)] ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
poly_mult [] q = [];
poly_mult (mc : p) q = poly_add (monom_mult_poly mc q) (poly_mult p q);

zero_poly :: forall a b. [([(a, Nat)], b)];
zero_poly = [];

class (Semigroup_mult a) => Ab_semigroup_mult a where {
};

class (Ab_semigroup_mult a, Semiring a) => Comm_semiring a where {
};

class (Comm_semiring a, Semiring_0 a) => Comm_semiring_0 a where {
};

class (Ab_semigroup_mult a, Monoid_mult a) => Comm_monoid_mult a where {
};

class (Comm_monoid_mult a, Comm_semiring_0 a, Dvd a,
        Semiring_1 a) => Comm_semiring_1 a where {
};

poly_of ::
  forall a b. (Eq a, Eq b, Comm_semiring_1 b) => Tpoly a b -> [([(a, Nat)], b)];
poly_of (PNum i) = (if i == zero then [] else [([], i)]);
poly_of (PVar x) = [([(x, one_nat)], one)];
poly_of (PSum []) = zero_poly;
poly_of (PSum (p : ps)) = poly_add (poly_of p) (poly_of (PSum ps));
poly_of (PMult []) = one_poly;
poly_of (PMult (p : ps)) = poly_mult (poly_of p) (poly_of (PMult ps));

rs_memb :: forall a. (Linorder a) => a -> Rbt a () -> Bool;
rs_memb = s_memb rm_lookup;

shows_rpo_repr ::
  forall a. (Showa a) => [((a, Nat), (Nat, Order_tag))] -> [Char] -> [Char];
shows_rpo_repr prs =
  shows_string
    ['R', 'P', 'O', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
      'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e',
      'n', 'c', 'e'] .
    shows_nl .
      shows_map
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ (pr, _) ->
                 shows_string
                   ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '('] .
                   shows_prec Zero_nat f .
                     shows_string ['['] .
                       shows_prec_nat Zero_nat n .
                         shows_string [']', ')', ' ', '=', ' '] .
                           shows_prec_nat Zero_nat pr . shows_nl)
            b)
        prs .
        shows_nl .
          shows_string
            ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '(', '_', ')',
              ' ', '=', ' ', '0'] .
            shows_nl .
              shows_string
                ['a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                  'o', 'w', 'i', 'n', 'g', ' ', 's', 't', 'a', 't', 'u', 's'] .
                shows_nl .
                  shows_map
                    (\ (a, b) ->
                      let {
                        (f, n) = a;
                      } in (\ (_, s) ->
                             shows_string ['s', 't', 'a', 't', 'u', 's', '('] .
                               shows_prec Zero_nat f .
                                 shows_string ['['] .
                                   shows_prec_nat Zero_nat n .
                                     shows_string [']', ')', ' ', '=', ' '] .
                                       shows_string
 (case s of {
   Lex -> ['l', 'e', 'x'];
   Mul -> ['m', 'u', 'l'];
 }) .
 shows_nl)
                        b)
                    prs .
                    shows_nl .
                      shows_string
                        ['s', 't', 'a', 't', 'u', 's', '(', '_', ')', ' ', '=',
                          ' ', 'l', 'e', 'x'] .
                        shows_nl;

scnp_af_to_af ::
  forall a.
    ((a, Nat) -> [(Nat, Nat)]) ->
      ((a, Nat) -> Nat -> Bool) -> (a, Nat) -> Nat -> Bool;
scnp_af_to_af piaa pia =
  (\ (f, n) ->
    let {
      is = map fst (piaa (f, n));
    } in (if any (less_eq_nat n) is
           then let {
                  isa = pia (f, n);
                } in (\ i -> isa i || membera is i)
           else membera is));

mk_subst :: forall a b. (Eq a) => [(a, Term b a)] -> a -> Term b a;
mk_subst xts =
  (\ x -> (case map_of xts x of {
            Nothing -> Var x;
            Just t -> t;
          }));

shows_term ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> Term a b -> [Char] -> [Char];
shows_term fun var (Var x) = var x;
shows_term fun var (Fun f ts) =
  fun f .
    shows_list_gen id [] ['('] [',', ' '] [')'] (map (shows_term fun var) ts);

shows_prec_term ::
  forall a b. (Showa a, Showa b) => Nat -> Term a b -> [Char] -> [Char];
shows_prec_term d t = shows_term (shows_prec Zero_nat) (shows_prec Zero_nat) t;

shows_list_term ::
  forall a b. (Showa a, Showa b) => [Term a b] -> [Char] -> [Char];
shows_list_term is = shows_list_aux (shows_prec_term Zero_nat) is;

instance (Showa a, Showa b) => Showa (Term a b) where {
  shows_prec = shows_prec_term;
  shows_list = shows_list_term;
};

mk_subst_domain ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> [(a, Term b a)];
mk_subst_domain sigma =
  let {
    tau = mk_subst sigma;
  } in remdups
         (filter (\ (x, t) -> not (equal_term (Var x) t))
           (map (\ (x, _) -> (x, tau x)) sigma));

subst_eq ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> [(a, Term b a)] -> Bool;
subst_eq sigma tau =
  let {
    sigmaa = mk_subst_domain sigma;
    taua = mk_subst_domain tau;
  } in equal_set (Set sigmaa) (Set taua);

full_af :: forall a. (a, Nat) -> Nat -> Bool;
full_af fn i = True;

data Root_redtriple_ext a b c =
  Root_redtriple_ext (Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ())
    ((a, Nat) -> Nat -> Bool) ((a, Nat) -> Nat -> Bool) ([Char] -> [Char]) c;

af :: forall a b c. Root_redtriple_ext a b c -> (a, Nat) -> Nat -> Bool;
af (Root_redtriple_ext valid s ns nst af aft desc more) = af;

ns :: forall a b c.
        Redtriple_ext a b c ->
          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
ns (Redtriple_ext valid s ns nst af mono desc cpx more) = ns;

sa :: forall a b c.
        Root_redtriple_ext a b c ->
          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
sa (Root_redtriple_ext valid s ns nst af aft desc more) = s;

defs_list :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
defs_list [] = [];
defs_list ((Fun f ss, r) : ts) =
  let {
    fs = defs_list ts;
    n = size_list ss;
  } in (if membera fs (f, n) then fs else (f, n) : fs);
defs_list ((Var x, r) : ts) = defs_list ts;

shows_pos :: Pos -> [Char] -> [Char];
shows_pos (PCons i Empty) = shows_prec_nat Zero_nat (Suc i);
shows_pos (PCons i (PCons v va)) =
  shows_prec_nat Zero_nat (Suc i) . shows_pos (PCons v va);
shows_pos Empty = shows_string ['e', 'p', 's', 'i', 'l', 'o', 'n'];

shows_rule ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) ->
        [Char] -> (Term a b, Term a b) -> [Char] -> [Char];
shows_rule fun var arr (l, r) =
  shows_term fun var l . shows_string arr . shows_term fun var r;

shows_rules ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) ->
        [Char] -> [(Term a b, Term a b)] -> [Char] -> [Char];
shows_rules fun var arr trs =
  shows_list_gen (shows_rule fun var arr) [] [] [newline] [] trs . shows_nl;

shows_trs ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) ->
        [Char] -> [Char] -> [(Term a b, Term a b)] -> [Char] -> [Char];
shows_trs fun var name arr r =
  shows_string name . shows_nl . shows_nl . shows_rules fun var arr r;

apply_args :: forall a b. a -> Term a b -> [Term a b] -> Term a b;
apply_args a t [] = t;
apply_args a t (s : ss) = apply_args a (Fun a [t, s]) ss;

get_symbol :: forall a. (a -> Nat -> [a]) -> a -> Nat -> Nat -> a;
get_symbol sm f n i = nth (sm f n) i;

letters :: [Char];
letters =
  ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '_', '0', '1', '2', '3', '4', '5', '6',
    '7', '8', '9', '&', ';', ':'];

alter10a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> j;
alter10a (Alter10 x) = x;

sum_lpoly :: forall a. (a -> a -> a) -> [a] -> [a] -> [a];
sum_lpoly uu [] ys = ys;
sum_lpoly uv (v : va) [] = v : va;
sum_lpoly pl (x : xs) (y : ys) = pl x y : sum_lpoly pl xs ys;

lpoly_of ::
  forall a.
    (Eq a) => a -> a -> (a -> a -> a) ->
                          (a -> a -> a) ->
                            Tpoly Nat a -> Parse_result [Char] (a, [a]);
lpoly_of uu uv uw ux (PNum i) = returna (i, []);
lpoly_of ze on pl ti (PVar x) = returna (ze, replicatea x ze ++ [on]);
lpoly_of ze on pl ti (PSum []) = returna (ze, []);
lpoly_of ze on pl ti (PSum (p : ps)) =
  bindc (lpoly_of ze on pl ti p)
    (\ (cp, ncp) ->
      bindc (lpoly_of ze on pl ti (PSum ps))
        (\ (cq, ncq) -> returna (pl cp cq, sum_lpoly pl ncp ncq)));
lpoly_of ze on pl ti (PMult []) = returna (on, []);
lpoly_of ze on pl ti (PMult (p : ps)) =
  bindc (lpoly_of ze on pl ti p)
    (\ (cp, ncp) ->
      bindc (lpoly_of ze on pl ti (PMult ps))
        (\ (cq, ncq) ->
          (if all (\ a -> ze == a) ncp then returna (ti cp cq, map (ti cp) ncq)
            else (if all (\ a -> ze == a) ncq
                   then returna (ti cp cq, map (\ x -> ti x cq) ncp)
                   else errora
                          ['c', 'a', 'n', 'n', 'o', 't', ' ', 't', 'r', 'a',
                            'n', 's', 'f', 'o', 'r', 'm', ' ', 'n', 'o', 'n',
                            '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p', 'o',
                            'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 't',
                            'o', ' ', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p',
                            'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']))));

innermostLhss ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [Term a [Char]];
innermostLhss xml2name =
  xml_many_elements
    ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's']
    (term xml2name) id;

strategy ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Strategy a [Char]);
strategy xml2name =
  xml1element ['s', 't', 'r', 'a', 't', 'e', 'g', 'y']
    (xml_options
      [(['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'],
         xml0elements ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'] Innermost),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's'],
          xml_change (innermostLhss xml2name) Innermost_Q)])
    id;

icap_impl ::
  forall a.
    (Eq a) => Bool ->
                (Term a [Char] -> Bool) ->
                  [Term a [Char]] ->
                    [Term a [Char]] ->
                      ([Char] -> Bool) ->
                        Term a [Char] -> Term a (Sum () [Char]);
icap_impl nf isQnf ls s sx (Var x) =
  (if nf && sx x then Var (Inr x) else Var (Inl ()));
icap_impl nf isQnf ls s sx (Fun f ts) =
  let {
    t = Fun f (map (icap_impl nf isQnf ls s sx) ts);
  } in (if any (\ l ->
                 (case mgu_class t l of {
                   Nothing -> False;
                   Just mu ->
                     all (\ u ->
                           isQnf (subst_apply_term (map_vars (\ a -> 'y' : a) u)
                                   mu))
                       (args l) &&
                       all (\ u -> isQnf (subst_apply_term u mu)) s;
                 }))
             ls
         then Var (Inl ()) else t);

nF_subset :: forall a b. (Eq a, Eq b) => [Term a b] -> [Term a b] -> Bool;
nF_subset ls q =
  all (\ l -> any (\ t -> any (\ u -> matches u t) (supteq_list l)) q) ls;

kbo_strict ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ((a, Nat) -> Nat) ->
                    ((a, Nat) -> Nat) ->
                      Nat ->
                        (a -> Bool) ->
                          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
kbo_strict pr w w0 least =
  (\ (s, t) ->
    check (fst (kbo w w0 pr least s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', 'K', 'B', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

less_list :: forall a. (Eq a, Order a) => [a] -> [a] -> Bool;
less_list (x : xs) (y : ys) = less x y || x == y && less_list xs ys;
less_list [] (x : xs) = True;
less_list xs [] = False;

less_eq_list :: forall a. (Eq a, Order a) => [a] -> [a] -> Bool;
less_eq_list (x : xs) (y : ys) = less x y || x == y && less_eq_list xs ys;
less_eq_list [] xs = True;
less_eq_list (x : xs) [] = False;

instance (Eq a, Order a) => Orda [a] where {
  less_eq = less_eq_list;
  less = less_list;
};

mat_arc_posI :: forall a. (a -> Bool) -> [[a]] -> Bool;
mat_arc_posI ap m = ap (nth (nth m Zero_nat) Zero_nat);

poly_var :: forall a b. (Eq a) => [([(a, Nat)], b)] -> a -> Bool;
poly_var p v = membera (concatMap (map fst . fst) p) v;

data Cert_result = Certified [Char] | Unsupported [Char] | Error [Char];

rs_empty :: forall a. (Linorder a) => () -> Rbt a ();
rs_empty = rm_empty;

rs_union :: forall a. (Linorder a) => Rbt a () -> Rbt a () -> Rbt a ();
rs_union = rm_add;

rules_with ::
  forall a b c.
    (Eq b,
      Linorder b) => (a -> Bool) ->
                       Rbt (b, Nat) [(a, (Term b c, Term b c))] ->
                         [(Term b c, Term b c)];
rules_with p m =
  map_filter (\ x -> (if (p . fst) x then Just (snd x) else Nothing))
    (values m);

scnp_desc ::
  forall a.
    (Showa a) => [((a, Nat), [(Nat, Nat)])] -> [Char] -> [Char] -> [Char];
scnp_desc af mu =
  shows_prec_list Zero_nat
    ['S', 'C', 'N', 'P', '-', 'v', 'e', 'r', 's', 'i', 'o', 'n', ' ', 'w', 'i',
      't', 'h', ' ', 'm', 'u', ' ', '=', ' '] .
    shows_prec_list Zero_nat mu .
      shows_prec_list Zero_nat
        [' ', 'a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'l', 'e', 'v', 'e', 'l',
          ' ', 'm', 'a', 'p', 'p', 'i', 'n', 'g', ' ', 'd', 'e', 'f', 'i', 'n',
          'e', 'd', ' ', 'b', 'y', ' '] .
        shows_nl .
          shows_sep
            (\ (a, b) ->
              let {
                (f, n) = a;
              } in (\ asa ->
                     shows_prec_list Zero_nat ['p', 'i', '('] .
                       shows_prec Zero_nat f .
                         shows_prec_list Zero_nat [')', ' ', '=', ' '] .
                           shows_list_aux
                             (\ (p, l) ->
                               shows_prec_list Zero_nat ['('] .
                                 (if less_nat p n
                                   then shows_prec_nat Zero_nat (Suc p)
                                   else shows_prec_list Zero_nat
  ['e', 'p', 's', 'i', 'l', 'o', 'n']) .
                                   shows_prec_list Zero_nat [','] .
                                     shows_prec_nat Zero_nat l .
                                       shows_prec_list Zero_nat [')'])
                             asa)
                b)
            shows_nl af .
            shows_nl;

s_iteratei ::
  forall a b c.
    (a -> (b -> Bool) -> ((c, ()) -> b -> b) -> b -> b) ->
      a -> (b -> Bool) -> (c -> b -> b) -> b -> b;
s_iteratei it s c f = it s c (\ x -> f (fst x));

data Tp_ops_ext a b c d =
  Tp_ops_ext
    (a -> (Bool,
            (Set (Term b c),
              (Set (Term b c, Term b c), Set (Term b c, Term b c)))))
    (a -> [Term b c]) (a -> [(Term b c, Term b c)])
    (a -> [(Term b c, Term b c)]) (a -> [(Term b c, Term b c)]) (a -> Bool)
    (a -> Term b c -> Bool) (a -> Bool)
    (a -> (b, Nat) -> [(Term b c, Term b c)])
    (a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a)
    (a -> [(Term b c, Term b c)] ->
            ([(Term b c, Term b c)], [(Term b c, Term b c)]))
    (Bool ->
      [Term b c] -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a)
    (a -> Bool) d;

qa :: forall a b c d. Tp_ops_ext a b c d -> a -> [Term b c];
qa (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
     delete_R_Rw split_rules mk nfs more)
  = q;

mka ::
  forall a b c d.
    Tp_ops_ext a b c d ->
      Bool ->
        [Term b c] -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a;
mka (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules
      rules_map delete_R_Rw split_rules mk nfs more)
  = mk;

nfsa :: forall a b c d. Tp_ops_ext a b c d -> a -> Bool;
nfsa (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules
       rules_map delete_R_Rw split_rules mk nfs more)
  = nfs;

delete_R_Rwa ::
  forall a b c d.
    Tp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a;
delete_R_Rwa
  (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs more)
  = delete_R_Rw;

split_rulesa ::
  forall a b c d.
    Tp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] ->
             ([(Term b c, Term b c)], [(Term b c, Term b c)]);
split_rulesa
  (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs more)
  = split_rules;

split_tt ::
  forall a b c.
    (Showa b,
      Showa c) => Tp_ops_ext a b c () -> a -> [(Term b c, Term b c)] -> (a, a);
split_tt i tp r_remove =
  let {
    (r, rw) = split_rulesa i tp r_remove;
    tp1 = mka i (nfsa i tp) (qa i tp) r rw;
    a = delete_R_Rwa i tp r_remove r_remove;
  } in (tp1, a);

supt_impl :: forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Bool;
supt_impl (Var x) t = False;
supt_impl (Fun f ss) t = membera ss t || any (\ s -> supt_impl s t) ss;

afa :: forall a b c. Redtriple_ext a b c -> (a, Nat) -> Nat -> Bool;
afa (Redtriple_ext valid s ns nst af mono desc cpx more) = af;

aft :: forall a b c. Root_redtriple_ext a b c -> (a, Nat) -> Nat -> Bool;
aft (Root_redtriple_ext valid s ns nst af aft desc more) = aft;

cpx ::
  forall a b c.
    Redtriple_ext a b c -> Complexity_measure a b -> Maybe Complexity_class;
cpx (Redtriple_ext valid s ns nst af mono desc cpx more) = cpx;

nsa ::
  forall a b c.
    Root_redtriple_ext a b c ->
      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
nsa (Root_redtriple_ext valid s ns nst af aft desc more) = ns;

nst ::
  forall a b c.
    Redtriple_ext a b c -> (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
nst (Redtriple_ext valid s ns nst af mono desc cpx more) = nst;

aarity_term ::
  forall a b. (Eq a) => a -> (a -> Nat -> [a]) -> Term a b -> Maybe Nat;
aarity_term a sm t =
  (case unapp a t of {
    (Var _, ts) -> Nothing;
    (Fun f ss, ts) ->
      Just (minus_nat (aarity sm f (size_list ss)) (size_list ts));
  });

uncurry_top ::
  forall a b. (Eq a) => a -> Nat -> (a -> Nat -> [a]) -> Term a b -> Term a b;
uncurry_top a n sm (Fun f ts) =
  let {
    mt = map (map_term_wa (\ fa na -> get_symbol sm fa na Zero_nat));
    (h, m) = the (root (hd ts));
  } in (if f == a &&
             equal_nat (size_list ts) n &&
               not (is_Var (hd ts)) && not (equal_nat (aarity sm h m) Zero_nat)
         then let {
                (Fun g ss) = hd ts;
              } in Fun (get_symbol sm g (size_list ss) one_nat)
                     (mt (ss ++ tl ts))
         else Fun (get_symbol sm f (size_list ts) Zero_nat) (mt ts));
uncurry_top a n sm (Var x) = Var x;

xml1to2elements ::
  forall a b c.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (a -> Maybe b -> c) -> Xml -> Parse_result [Char] c;
xml1to2elements tag p1 p2 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             null atts &&
               less_eq_nat one_nat l && less_eq_nat l (nat 2) && is_none text
         then bindc (p1 (nth cs Zero_nat))
                (\ x1 ->
                  (if equal_nat l (nat 2)
                    then bindc (p2 (nth cs one_nat))
                           (\ x2 -> returna (f x1 (Just x2)))
                    else returna (f x1 Nothing)))
         else no_transform tag (Xml name atts cs text));

xml2to3elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (a -> b -> Maybe c -> d) -> Xml -> Parse_result [Char] d;
xml2to3elements tag p1 p2 p3 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             null atts &&
               less_eq_nat (nat 2) l && less_eq_nat l (nat 3) && is_none text
         then bindc (p1 (nth cs Zero_nat))
                (\ x1 ->
                  bindc (p2 (nth cs one_nat))
                    (\ x2 ->
                      (if equal_nat l (nat 3)
                        then bindc (p3 (nth cs (nat 2)))
                               (\ x3 -> returna (f x1 x2 (Just x3)))
                        else returna (f x1 x2 Nothing))))
         else no_transform tag (Xml name atts cs text));

arith_fun :: Xml -> Parse_result [Char] ArithFun;
arith_fun xml =
  xml1element ['a', 'r', 'i', 't', 'h', 'F', 'u', 'n', 'c', 't', 'i', 'o', 'n']
    (xml_options
      [(['n', 'a', 't', 'u', 'r', 'a', 'l'],
         xml_change (nata ['n', 'a', 't', 'u', 'r', 'a', 'l']) Const),
        (['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'],
          xml_change (nata ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'])
            (\ n -> Arg (minus_nat n one_nat))),
        (['s', 'u', 'm'], xml_many_elements ['s', 'u', 'm'] arith_fun Sum),
        (['p', 'r', 'o', 'd', 'u', 'c', 't'],
          xml_many_elements ['p', 'r', 'o', 'd', 'u', 'c', 't'] arith_fun Prod),
        (['m', 'i', 'n'], xml_many_elements ['m', 'i', 'n'] arith_fun Min),
        (['m', 'a', 'x'], xml_many_elements ['m', 'a', 'x'] arith_fun Max)])
    id xml;

int_coeff :: Xml -> Parse_result [Char] Integer;
int_coeff = int ['i', 'n', 't', 'e', 'g', 'e', 'r'];

vec_coeff ::
  forall a. (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] [a];
vec_coeff xml2coeff =
  xml_many_elements ['v', 'e', 'c', 't', 'o', 'r']
    (xml1element ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't']
      xml2coeff id)
    id;

mat_coeff ::
  forall a.
    a -> (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] [[a]];
mat_coeff ze xml2coeff =
  xml_options
    [(['m', 'a', 't', 'r', 'i', 'x'],
       xml_many_elements ['m', 'a', 't', 'r', 'i', 'x'] (vec_coeff xml2coeff)
         id),
      (['v', 'e', 'c', 't', 'o', 'r'],
        xml_change (vec_coeff xml2coeff)
          (\ v ->
            v : replicatea (minus_nat (size_list v) one_nat)
                  (replicatea (size_list v) ze)))];

status_precedence ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [((a, Nat), (Nat, Order_tag))];
status_precedence xml2name =
  xml_many_elements
    ['s', 't', 'a', 't', 'u', 's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c',
      'e']
    (xml4elements
      ['s', 't', 'a', 't', 'u', 's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
        'c', 'e', 'E', 'n', 't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (nata ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'])
      (xml_options
        [(['l', 'e', 'x'], xml0elements ['l', 'e', 'x'] Lex),
          (['m', 'u', 'l'], xml0elements ['m', 'u', 'l'] Mul)])
      (\ f a p s -> ((f, a), (p, s))))
    id;

path_order ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Redtriple_impl a);
path_order xml2name =
  xml1to2elements ['p', 'a', 't', 'h', 'O', 'r', 'd', 'e', 'r']
    (status_precedence xml2name) (afs xml2name)
    (\ prec_tau afo ->
      let {
        a = (case afo of {
              Nothing -> [];
              Just af -> af;
            });
      } in Rpo prec_tau a);

data Poly_c_ext a b =
  Poly_c_ext (a -> Bool) a a a (a -> a -> a) (a -> a -> a) (a -> a)
    (a -> a -> Bool) (a -> a -> Bool) Bool [Char] (a -> Bool) (a -> Bool)
    (a -> a -> a) (a -> Maybe Nat) b;

gt :: forall a b. Poly_c_ext a b -> a -> a -> Bool;
gt (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
     arc_pos check_mono maxa derive_complexity more)
  = gt;

fit_length :: forall a. a -> Nat -> [a] -> [a];
fit_length uu Zero_nat uv = [];
fit_length a (Suc v) [] = replicatea (Suc v) a;
fit_length a (Suc n) (b : asa) = b : fit_length a n asa;

polynomial ::
  forall a.
    (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] (Tpoly Nat a);
polynomial xml2coeff x =
  xml1element ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']
    (xml_options
      [(['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'],
         xml1element ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't']
           xml2coeff PNum),
        (['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'],
          xml_change (nata ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'])
            (\ n -> PVar (minus_nat n one_nat))),
        (['s', 'u', 'm'],
          xml_many_elements ['s', 'u', 'm'] (polynomial xml2coeff) PSum),
        (['p', 'r', 'o', 'd', 'u', 'c', 't'],
          xml_many_elements ['p', 'r', 'o', 'd', 'u', 'c', 't']
            (polynomial xml2coeff) PMult)])
    id x;

onea :: forall a b. Poly_c_ext a b -> a;
onea (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
       arc_pos check_mono maxa derive_complexity more)
  = one;

instance Ab_semigroup_mult Integer where {
};

instance Comm_semiring Integer where {
};

instance Comm_semiring_0 Integer where {
};

instance Comm_monoid_mult Integer where {
};

instance Comm_semiring_1 Integer where {
};

plusa :: forall a b. Poly_c_ext a b -> a -> a -> a;
plusa (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
        arc_pos check_mono maxa derive_complexity more)
  = plus;

zeroa :: forall a b. Poly_c_ext a b -> a;
zeroa (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
        arc_pos check_mono maxa derive_complexity more)
  = zero;

instance Ab_semigroup_mult Rat where {
};

instance Comm_semiring Rat where {
};

instance Comm_semiring_0 Rat where {
};

instance Comm_monoid_mult Rat where {
};

instance Comm_semiring_1 Rat where {
};

arctic_coeff :: Xml -> Parse_result [Char] Arctic;
arctic_coeff =
  xml_options
    [(['i', 'n', 't', 'e', 'g', 'e', 'r'],
       xml_change (int ['i', 'n', 't', 'e', 'g', 'e', 'r']) Num_arc),
      (['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'],
        xml0elements
          ['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y']
          MinInfty)];

timesa :: forall a b. Poly_c_ext a b -> a -> a -> a;
timesa
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = times;

arc_pos :: forall a b. Poly_c_ext a b -> a -> Bool;
arc_pos
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = arc_pos;

defaulta :: forall a b. Poly_c_ext a b -> a;
defaulta
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = defaulta;

arctic_rat_coeff :: Xml -> Parse_result [Char] Arctic_rat;
arctic_rat_coeff =
  xml_options
    [(['i', 'n', 't', 'e', 'g', 'e', 'r'], xml_change rat Num_arc_rat),
      (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l'], xml_change rat Num_arc_rat),
      (['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'],
        xml0elements
          ['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y']
          MinInfty_rat)];

one_arctic :: Arctic;
one_arctic = Num_arc 0;

equal_arctic :: Arctic -> Arctic -> Bool;
equal_arctic (Num_arc int) MinInfty = False;
equal_arctic MinInfty (Num_arc int) = False;
equal_arctic (Num_arc inta) (Num_arc int) = inta == int;
equal_arctic MinInfty MinInfty = True;

instance Eq Arctic where {
  a == b = equal_arctic a b;
};

plus_arctic :: Arctic -> Arctic -> Arctic;
plus_arctic MinInfty y = y;
plus_arctic (Num_arc v) MinInfty = Num_arc v;
plus_arctic (Num_arc x) (Num_arc y) = Num_arc (maxa x y);

zero_arctic :: Arctic;
zero_arctic = MinInfty;

description :: forall a b. Poly_c_ext a b -> [Char];
description
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = description;

times_arctic :: Arctic -> Arctic -> Arctic;
times_arctic MinInfty y = MinInfty;
times_arctic (Num_arc v) MinInfty = MinInfty;
times_arctic (Num_arc x) (Num_arc y) = Num_arc (x + y);

rat_domain :: Xml -> Parse_result [Char] Rat;
rat_domain =
  xml1element ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's']
    (xml1element ['d', 'e', 'l', 't', 'a'] rat id) id;

basic_domain :: Xml -> Parse_result [Char] (Nat -> Domain);
basic_domain =
  xml_options
    [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'],
       xml0elements ['n', 'a', 't', 'u', 'r', 'a', 'l', 's'] Naturala),
      (['a', 'r', 'c', 't', 'i', 'c'],
        xml1element ['a', 'r', 'c', 't', 'i', 'c']
          (xml1element ['d', 'o', 'm', 'a', 'i', 'n']
            (xml_options
              [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'],
                 xml0elements ['n', 'a', 't', 'u', 'r', 'a', 'l', 's']
                   (\ _ -> Arctic)),
                (['i', 'n', 't', 'e', 'g', 'e', 'r', 's'],
                  xml0elements ['i', 'n', 't', 'e', 'g', 'e', 'r', 's']
                    (\ _ -> Arctic)),
                (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'],
                  xml_change rat_domain (\ _ _ -> Arctic_rat))])
            id)
          id),
      (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'],
        xml_change rat_domain Rationala)];

interpretation_type :: Xml -> Parse_result [Char] Domain;
interpretation_type =
  xml1element ['t', 'y', 'p', 'e']
    (xml_options
      [(['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l'],
         xml2elements ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']
           (xml1element ['d', 'o', 'm', 'a', 'i', 'n']
             (xml_options
               [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'], basic_domain),
                 (['a', 'r', 'c', 't', 'i', 'c'], basic_domain),
                 (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'], basic_domain),
                 (['m', 'a', 't', 'r', 'i', 'c', 'e', 's'],
                   xml3elements ['m', 'a', 't', 'r', 'i', 'c', 'e', 's']
                     (nata ['d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'])
                     (nata ['s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e',
                             'n', 's', 'i', 'o', 'n'])
                     (xml1element ['d', 'o', 'm', 'a', 'i', 'n'] basic_domain
                       id)
                     (\ di sd domain d ->
                       (case domain d of {
                         Naturala _ -> Int_mat di sd;
                         Arctic -> Arc_mat di;
                         Arctic_rat -> Arc_rat_mat di;
                         Rationala _ _ -> Rat_mat di sd;
                       })))])
             id)
           (nata ['d', 'e', 'g', 'r', 'e', 'e']) (\ typea -> typea)),
        (['m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p', 'r', 'e',
           't', 'a', 't', 'i', 'o', 'n'],
          xml3elements
            ['m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p', 'r',
              'e', 't', 'a', 't', 'i', 'o', 'n']
            (xml1element ['d', 'o', 'm', 'a', 'i', 'n'] basic_domain id)
            (nata ['d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'])
            (nata ['s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's',
                    'i', 'o', 'n'])
            (\ domain di sd ->
              (case domain Zero_nat of {
                Naturala _ -> Int_mat di sd;
                Arctic -> Arc_mat di;
                Arctic_rat -> Arc_rat_mat di;
                Rationala _ _ -> Rat_mat di sd;
              })))])
    id;

one_arctic_rat :: Arctic_rat;
one_arctic_rat = Num_arc_rat zero_rat;

equal_arctic_rat :: Arctic_rat -> Arctic_rat -> Bool;
equal_arctic_rat (Num_arc_rat rat) MinInfty_rat = False;
equal_arctic_rat MinInfty_rat (Num_arc_rat rat) = False;
equal_arctic_rat (Num_arc_rat rata) (Num_arc_rat rat) = equal_rat rata rat;
equal_arctic_rat MinInfty_rat MinInfty_rat = True;

instance Eq Arctic_rat where {
  a == b = equal_arctic_rat a b;
};

plus_arctic_rat :: Arctic_rat -> Arctic_rat -> Arctic_rat;
plus_arctic_rat MinInfty_rat y = y;
plus_arctic_rat (Num_arc_rat v) MinInfty_rat = Num_arc_rat v;
plus_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = Num_arc_rat (maxa x y);

zero_arctic_rat :: Arctic_rat;
zero_arctic_rat = MinInfty_rat;

times_arctic_rat :: Arctic_rat -> Arctic_rat -> Arctic_rat;
times_arctic_rat MinInfty_rat y = MinInfty_rat;
times_arctic_rat (Num_arc_rat v) MinInfty_rat = MinInfty_rat;
times_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = Num_arc_rat (plus_rat x y);

int_mono :: Integer -> Bool;
int_mono x = 1 <= x;

max0_int :: Integer -> Integer;
max0_int x = maxa 0 x;

int_name :: [Char];
int_name =
  ['n', 'a', 't', 'u', 'r', 'a', 'l', 's', ' ', 'w', 'i', 't', 'h', ' ', 'n',
    'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't', 'a', 'n',
    't', 's'];

int_components :: Poly_c_ext Integer ();
int_components =
  Poly_c_ext (\ _ -> True) 0 1 1 (\ a b -> a + b) (\ a b -> a * b) max0_int
    (\ x y -> y <= x) (\ x y -> y < x) True int_name (\ _ -> True) int_mono maxa
    (\ x -> (if x <= 1 then Just Zero_nat else Nothing)) ();

monoa :: forall a b. Poly_c_ext a b -> Bool;
monoa (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
        arc_pos check_mono maxa derive_complexity more)
  = mono;

mat_mono :: forall a. Nat -> (a -> Bool) -> [[a]] -> Bool;
mat_mono sd m = mat_monoI m sd;

check_mono :: forall a b. Poly_c_ext a b -> a -> Bool;
check_mono
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = check_mono;

upper_triangularI ::
  forall a. (Eq a) => a -> a -> (a -> a -> Bool) -> [[a]] -> Bool;
upper_triangularI ze on g [] = True;
upper_triangularI ze on g ((a : asa) : m) =
  g on a && all (\ b -> b == ze) asa && upper_triangularI ze on g (map tl m);
upper_triangularI ze on g ([] : m) = False;

class (Semiring_0 a, Ordered_ab_semigroup a) => Ordered_semiring_0 a where {
};

class (Semiring_1 a, Ordered_semiring_0 a) => Ordered_semiring_1 a where {
};

class (Max_ordered_monoid_add a,
        Ordered_semiring_1 a) => Max_ordered_semiring_1 a where {
};

class (Comm_semiring_1 a,
        Ordered_semiring_1 a) => Large_ordered_semiring_1 a where {
};

class (Monoid_add a, Orda a,
        Ordered_ab_semigroup a) => Max_ordered_ab_semigroup a where {
};

class (Large_ordered_semiring_1 a, Max_ordered_ab_semigroup a,
        Max_ordered_semiring_1 a) => Bin_max_ordered_semiring_1 a where {
};

mat_complexity ::
  forall a. (Eq a, Bin_max_ordered_semiring_1 a) => Nat -> [[a]] -> Maybe Nat;
mat_complexity n m =
  (if upper_triangularI zero one ge m then Just (minus_nat n one_nat)
    else Nothing);

mat_components ::
  forall a.
    (Eq a,
      Bin_max_ordered_semiring_1 a) => Nat ->
 Nat -> Poly_c_ext a () -> Poly_c_ext [[a]] ();
mat_components n sd c =
  Poly_c_ext (mat n n) (mat0I zero n n) (mat1I zero one n)
    (mat1I zero (defaulta c) n) (mat_plusI plus) (mat_multI zero plus times n)
    mat_max0 mat_ge (mat_gtI (gt c) sd) (monoa c)
    (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h',
          ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's',
          'i', 'o', 'n', ' '],
        shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
        description c])
    (\ _ -> True) (mat_mono sd (check_mono c)) (mat_plusI maxa)
    (mat_complexity n) ();

max0_rat :: Rat -> Rat;
max0_rat x = maxa zero_rat x;

rat_mono :: Rat -> Bool;
rat_mono x = less_eq_rat one_rat x;

rat_name :: [Char];
rat_name =
  ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's', ' ', '(', 'w', 'i', 't', 'h',
    ' ', 'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't',
    'a', 'n', 't', 's', ')'];

rat_components :: Poly_c_ext Rat ();
rat_components =
  Poly_c_ext (\ _ -> True) zero_rat one_rat one_rat plus_rat times_rat max0_rat
    (\ x y -> less_eq_rat y x) (\ x y -> less_rat y x) True rat_name
    (\ _ -> True) rat_mono maxa
    (\ x -> (if less_eq_rat x one_rat then Just Zero_nat else Nothing)) ();

ge_arctic :: Arctic -> Arctic -> Bool;
ge_arctic x MinInfty = True;
ge_arctic MinInfty (Num_arc uu) = False;
ge_arctic (Num_arc x) (Num_arc y) = y <= x;

gt_arctic :: Arctic -> Arctic -> Bool;
gt_arctic uu MinInfty = True;
gt_arctic MinInfty (Num_arc uv) = False;
gt_arctic (Num_arc x) (Num_arc y) = y < x;

pos_arctic :: Arctic -> Bool;
pos_arctic MinInfty = False;
pos_arctic (Num_arc n) = 0 <= n;

arctic_name :: [Char];
arctic_name =
  ['a', 'r', 'c', 't', 'i', 'c', ' ', 's', 'e', 'm', 'i', 'r', 'i', 'n', 'g'];

arctic_components :: Poly_c_ext Arctic ();
arctic_components =
  Poly_c_ext (\ _ -> True) zero_arctic one_arctic one_arctic plus_arctic
    times_arctic id ge_arctic gt_arctic False arctic_name pos_arctic
    (\ _ -> False) plus_arctic (\ _ -> Nothing) ();

instance Non_strict_order Arctic where {
  ge = ge_arctic;
};

ge_arctic_rat :: Arctic_rat -> Arctic_rat -> Bool;
ge_arctic_rat x MinInfty_rat = True;
ge_arctic_rat MinInfty_rat (Num_arc_rat uu) = False;
ge_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = less_eq_rat y x;

pos_arctic_rat :: Arctic_rat -> Bool;
pos_arctic_rat MinInfty_rat = False;
pos_arctic_rat (Num_arc_rat n) = less_eq_rat zero_rat n;

arctic_rat_name :: [Char];
arctic_rat_name =
  ['a', 'r', 'c', 't', 'i', 'c', ' ', 'r', 'a', 't', 'i', 'o', 'n', 'a', 'l',
    's'];

weak_gt_arctic_rat :: Arctic_rat -> Arctic_rat -> Bool;
weak_gt_arctic_rat uu MinInfty_rat = True;
weak_gt_arctic_rat MinInfty_rat (Num_arc_rat uv) = False;
weak_gt_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = less_rat y x;

arctic_rat_components :: Poly_c_ext Arctic_rat ();
arctic_rat_components =
  Poly_c_ext (\ _ -> True) zero_arctic_rat one_arctic_rat one_arctic_rat
    plus_arctic_rat times_arctic_rat id ge_arctic_rat weak_gt_arctic_rat False
    arctic_rat_name pos_arctic_rat (\ _ -> False) plus_arctic_rat
    (\ _ -> Nothing) ();

instance Non_strict_order Arctic_rat where {
  ge = ge_arctic_rat;
};

ge_int :: Integer -> Integer -> Bool;
ge_int x y = y <= x;

instance Non_strict_order Integer where {
  ge = ge_int;
};

instance Ordered_ab_semigroup Integer where {
};

instance Ordered_semiring_0 Integer where {
};

instance Ordered_semiring_1 Integer where {
};

instance Max_ordered_monoid_add Integer where {
  max0 = max0_int;
};

instance Max_ordered_semiring_1 Integer where {
};

instance Large_ordered_semiring_1 Integer where {
};

instance Max_ordered_ab_semigroup Integer where {
};

instance Bin_max_ordered_semiring_1 Integer where {
};

ge_rat :: Rat -> Rat -> Bool;
ge_rat x y = less_eq_rat y x;

instance Non_strict_order Rat where {
  ge = ge_rat;
};

instance Ordered_ab_semigroup Rat where {
};

instance Ordered_semiring_0 Rat where {
};

instance Ordered_semiring_1 Rat where {
};

instance Max_ordered_monoid_add Rat where {
  max0 = max0_rat;
};

instance Max_ordered_semiring_1 Rat where {
};

instance Large_ordered_semiring_1 Rat where {
};

instance Max_ordered_ab_semigroup Rat where {
};

instance Bin_max_ordered_semiring_1 Rat where {
};

interpretation ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Redtriple_impl a);
interpretation xml2name =
  xml1many_elements_gen
    ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n']
    interpretation_type
    (\ a ->
      (case a of {
        Naturala deg ->
          (if less_eq_nat deg one_nat
            then xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     bindc (polynomial int_coeff x)
                       (lpoly_of (zeroa int_components) (onea int_components)
                         (plusa int_components) (timesa int_components)))
                   (\ f aa poly ->
                     Alter1
                       ((f, aa),
                         (fst poly,
                           fit_length (zeroa int_components) aa (snd poly))))
            else xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     bindc (polynomial int_coeff x)
                       (\ p -> returna (poly_of p)))
                   (\ f aa poly -> Alter9 ((f, aa), poly)));
        Arctic ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial arctic_coeff x)
                (lpoly_of (zeroa arctic_components) (onea arctic_components)
                  (plusa arctic_components) (timesa arctic_components)))
            (\ f aa poly ->
              Alter3
                ((f, aa),
                  (fst poly,
                    fit_length (zeroa arctic_components) aa (snd poly))));
        Arctic_rat ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial arctic_rat_coeff x)
                (lpoly_of (zeroa arctic_rat_components)
                  (onea arctic_rat_components) (plusa arctic_rat_components)
                  (timesa arctic_rat_components)))
            (\ f aa poly ->
              Alter7
                ((f, aa),
                  (fst poly,
                    fit_length (zeroa arctic_rat_components) aa (snd poly))));
        Int_mat n sd ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial (mat_coeff 0 int_coeff) x)
                (lpoly_of (zeroa (mat_components n sd int_components))
                  (onea (mat_components n sd int_components))
                  (plusa (mat_components n sd int_components))
                  (timesa (mat_components n sd int_components))))
            (\ f aa poly ->
              Alter4
                ((f, aa),
                  (fst poly,
                    fit_length (zeroa (mat_components n sd int_components)) aa
                      (snd poly))));
        Arc_mat n ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial (mat_coeff MinInfty arctic_coeff) x)
                (lpoly_of
                  (zeroa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
                      (mat1I zero_arctic one_arctic n)
                      (mat1I zero_arctic (defaulta arctic_components) n)
                      (mat_plusI plus_arctic)
                      (mat_multI zero_arctic plus_arctic times_arctic n) id
                      mat_ge (mat_comp_all (gt arctic_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_components])
                      (mat_arc_posI (arc_pos arctic_components)) (\ _ -> False)
                      (mat_plusI plus_arctic) (\ _ -> Nothing) ()))
                  (onea (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
                          (mat1I zero_arctic one_arctic n)
                          (mat1I zero_arctic (defaulta arctic_components) n)
                          (mat_plusI plus_arctic)
                          (mat_multI zero_arctic plus_arctic times_arctic n) id
                          mat_ge (mat_comp_all (gt arctic_components)) False
                          (concat
                            [shows_prec_nat Zero_nat n [], ['x'],
                              shows_prec_nat Zero_nat n [],
                              [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                                'o', 'v', 'e', 'r', ' '],
                              description arctic_components])
                          (mat_arc_posI (arc_pos arctic_components))
                          (\ _ -> False) (mat_plusI plus_arctic)
                          (\ _ -> Nothing) ()))
                  (plusa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
                      (mat1I zero_arctic one_arctic n)
                      (mat1I zero_arctic (defaulta arctic_components) n)
                      (mat_plusI plus_arctic)
                      (mat_multI zero_arctic plus_arctic times_arctic n) id
                      mat_ge (mat_comp_all (gt arctic_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_components])
                      (mat_arc_posI (arc_pos arctic_components)) (\ _ -> False)
                      (mat_plusI plus_arctic) (\ _ -> Nothing) ()))
                  (timesa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
                      (mat1I zero_arctic one_arctic n)
                      (mat1I zero_arctic (defaulta arctic_components) n)
                      (mat_plusI plus_arctic)
                      (mat_multI zero_arctic plus_arctic times_arctic n) id
                      mat_ge (mat_comp_all (gt arctic_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_components])
                      (mat_arc_posI (arc_pos arctic_components)) (\ _ -> False)
                      (mat_plusI plus_arctic) (\ _ -> Nothing) ()))))
            (\ f aa poly ->
              Alter6
                ((f, aa),
                  (fst poly,
                    fit_length
                      (zeroa
                        (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
                          (mat1I zero_arctic one_arctic n)
                          (mat1I zero_arctic (defaulta arctic_components) n)
                          (mat_plusI plus_arctic)
                          (mat_multI zero_arctic plus_arctic times_arctic n) id
                          mat_ge (mat_comp_all (gt arctic_components)) False
                          (concat
                            [shows_prec_nat Zero_nat n [], ['x'],
                              shows_prec_nat Zero_nat n [],
                              [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                                'o', 'v', 'e', 'r', ' '],
                              description arctic_components])
                          (mat_arc_posI (arc_pos arctic_components))
                          (\ _ -> False) (mat_plusI plus_arctic)
                          (\ _ -> Nothing) ()))
                      aa (snd poly))));
        Arc_rat_mat n ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial (mat_coeff MinInfty_rat arctic_rat_coeff) x)
                (lpoly_of
                  (zeroa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
                      (mat1I zero_arctic_rat one_arctic_rat n)
                      (mat1I zero_arctic_rat (defaulta arctic_rat_components) n)
                      (mat_plusI plus_arctic_rat)
                      (mat_multI zero_arctic_rat plus_arctic_rat
                        times_arctic_rat n)
                      id mat_ge (mat_comp_all (gt arctic_rat_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_rat_components])
                      (mat_arc_posI (arc_pos arctic_rat_components))
                      (\ _ -> False) (mat_plusI plus_arctic_rat)
                      (\ _ -> Nothing) ()))
                  (onea (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
                          (mat1I zero_arctic_rat one_arctic_rat n)
                          (mat1I zero_arctic_rat
                            (defaulta arctic_rat_components) n)
                          (mat_plusI plus_arctic_rat)
                          (mat_multI zero_arctic_rat plus_arctic_rat
                            times_arctic_rat n)
                          id mat_ge (mat_comp_all (gt arctic_rat_components))
                          False
                          (concat
                            [shows_prec_nat Zero_nat n [], ['x'],
                              shows_prec_nat Zero_nat n [],
                              [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                                'o', 'v', 'e', 'r', ' '],
                              description arctic_rat_components])
                          (mat_arc_posI (arc_pos arctic_rat_components))
                          (\ _ -> False) (mat_plusI plus_arctic_rat)
                          (\ _ -> Nothing) ()))
                  (plusa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
                      (mat1I zero_arctic_rat one_arctic_rat n)
                      (mat1I zero_arctic_rat (defaulta arctic_rat_components) n)
                      (mat_plusI plus_arctic_rat)
                      (mat_multI zero_arctic_rat plus_arctic_rat
                        times_arctic_rat n)
                      id mat_ge (mat_comp_all (gt arctic_rat_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_rat_components])
                      (mat_arc_posI (arc_pos arctic_rat_components))
                      (\ _ -> False) (mat_plusI plus_arctic_rat)
                      (\ _ -> Nothing) ()))
                  (timesa
                    (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
                      (mat1I zero_arctic_rat one_arctic_rat n)
                      (mat1I zero_arctic_rat (defaulta arctic_rat_components) n)
                      (mat_plusI plus_arctic_rat)
                      (mat_multI zero_arctic_rat plus_arctic_rat
                        times_arctic_rat n)
                      id mat_ge (mat_comp_all (gt arctic_rat_components)) False
                      (concat
                        [shows_prec_nat Zero_nat n [], ['x'],
                          shows_prec_nat Zero_nat n [],
                          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                            'o', 'v', 'e', 'r', ' '],
                          description arctic_rat_components])
                      (mat_arc_posI (arc_pos arctic_rat_components))
                      (\ _ -> False) (mat_plusI plus_arctic_rat)
                      (\ _ -> Nothing) ()))))
            (\ f aa poly ->
              Alter8
                ((f, aa),
                  (fst poly,
                    fit_length
                      (zeroa
                        (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
                          (mat1I zero_arctic_rat one_arctic_rat n)
                          (mat1I zero_arctic_rat
                            (defaulta arctic_rat_components) n)
                          (mat_plusI plus_arctic_rat)
                          (mat_multI zero_arctic_rat plus_arctic_rat
                            times_arctic_rat n)
                          id mat_ge (mat_comp_all (gt arctic_rat_components))
                          False
                          (concat
                            [shows_prec_nat Zero_nat n [], ['x'],
                              shows_prec_nat Zero_nat n [],
                              [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                                'o', 'v', 'e', 'r', ' '],
                              description arctic_rat_components])
                          (mat_arc_posI (arc_pos arctic_rat_components))
                          (\ _ -> False) (mat_plusI plus_arctic_rat)
                          (\ _ -> Nothing) ()))
                      aa (snd poly))));
        Rationala _ deg ->
          (if less_eq_nat deg one_nat
            then xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     bindc (polynomial rat x)
                       (lpoly_of (zeroa rat_components) (onea rat_components)
                         (plusa rat_components) (timesa rat_components)))
                   (\ f aa poly ->
                     Alter2
                       ((f, aa),
                         (fst poly,
                           fit_length (zeroa rat_components) aa (snd poly))))
            else xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     bindc (polynomial rat x) (\ p -> returna (poly_of p)))
                   (\ f aa poly -> Alter10 ((f, aa), poly)));
        Rat_mat n sd ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              bindc (polynomial (mat_coeff zero_rat rat) x)
                (lpoly_of (zeroa (mat_components n sd rat_components))
                  (onea (mat_components n sd rat_components))
                  (plusa (mat_components n sd rat_components))
                  (timesa (mat_components n sd rat_components))))
            (\ f aa poly ->
              Alter5
                ((f, aa),
                  (fst poly,
                    fit_length (zeroa (mat_components n sd rat_components)) aa
                      (snd poly))));
      }))
    (\ typea pia ->
      (case typea of {
        Naturala deg ->
          (if less_eq_nat deg one_nat then Int_carrier (map alter1a pia)
            else Int_nl_carrier (map alter9a pia));
        Arctic -> Arctic_carrier (map alter3a pia);
        Arctic_rat -> Arctic_rat_carrier (map alter7a pia);
        Int_mat n sd -> Int_mat_carrier n sd (map alter4a pia);
        Arc_mat n -> Arc_mat_carrier n (map alter6a pia);
        Arc_rat_mat n -> Arc_rat_mat_carrier n (map alter8a pia);
        Rationala d deg ->
          (if less_eq_nat deg one_nat then Rat_carrier (map alter2a pia)
            else Rat_nl_carrier d (map alter10a pia));
        Rat_mat n sd -> Rat_mat_carrier n sd (map alter5a pia);
      }));

precedence_weight ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Nat -> ([((a, Nat), (Nat, Nat))], Nat));
precedence_weight xml2name =
  xml_many_elements
    ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', 'W', 'e', 'i', 'g', 'h',
      't']
    (xml4elements
      ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', 'W', 'e', 'i', 'g',
        'h', 't', 'E', 'n', 't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (nata ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'])
      (nata ['w', 'e', 'i', 'g', 'h', 't']) (\ f a p w -> ((f, a), (p, w))))
    (\ a b -> (a, b));

knuth_bendix_order ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Redtriple_impl a);
knuth_bendix_order xml2name =
  xml2to3elements
    ['k', 'n', 'u', 't', 'h', 'B', 'e', 'n', 'd', 'i', 'x', 'O', 'r', 'd', 'e',
      'r']
    (nata ['w', '0']) (precedence_weight xml2name) (afs xml2name)
    (\ w0 prw afo ->
      let {
        a = (case afo of {
              Nothing -> [];
              Just af -> af;
            });
      } in Kbo (prw w0) a);

redtriple ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Redtriple_impl a);
redtriple xml2name =
  xml1element ['r', 'e', 'd', 'P', 'a', 'i', 'r']
    (xml_options
      [(['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'],
         interpretation xml2name),
        (['p', 'a', 't', 'h', 'O', 'r', 'd', 'e', 'r'], path_order xml2name),
        (['k', 'n', 'u', 't', 'h', 'B', 'e', 'n', 'd', 'i', 'x', 'O', 'r', 'd',
           'e', 'r'],
          knuth_bendix_order xml2name)])
    id;

signature ::
  forall a.
    (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] [(a, Nat)];
signature xml2name =
  xml_many_elements ['s', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e']
    (xml2elements ['s', 'y', 'm', 'b', 'o', 'l'] xml2name
      (nata ['a', 'r', 'i', 't', 'y']) (\ a b -> (a, b)))
    id;

joinable_critical_pairs ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          [(Term a [Char],
             ([(Pos, ((Term a [Char], Term a [Char]), Term a [Char]))],
               (Term a [Char],
                 [(Pos, ((Term a [Char], Term a [Char]), Term a [Char]))])))];
joinable_critical_pairs xml2name =
  let {
    rew = rsteps xml2name;
  } in xml_many_elements
         ['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i', 'c',
           'a', 'l', 'P', 'a', 'i', 'r', 's']
         (xml2elements
           ['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i',
             'c', 'a', 'l', 'P', 'a', 'i', 'r']
           rew rew (\ (s, sseq) (t, tseq) -> (s, (sseq, (t, tseq)))))
         id;

wcr_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) -> Xml -> Parse_result [Char] (Join_info a);
wcr_proof xml2name =
  xml1element ['w', 'c', 'r', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i', 'c',
          'a', 'l', 'P', 'a', 'i', 'r', 's'],
         xml_change (joinable_critical_pairs xml2name) Guided),
        (['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i', 'c',
           'a', 'l', 'P', 'a', 'i', 'r', 's', 'B', 'F', 'S'],
          xml_change
            (nata ['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't',
                    'i', 'c', 'a', 'l', 'P', 'a', 'i', 'r', 's', 'B', 'F', 'S'])
            Join_BFS),
        (['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i', 'c',
           'a', 'l', 'P', 'a', 'i', 'r', 's', 'A', 'u', 't', 'o'],
          xml0elements
            ['j', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i', 't', 'i',
              'c', 'a', 'l', 'P', 'a', 'i', 'r', 's', 'A', 'u', 't', 'o']
            Join_NF)])
    id;

xml2dp_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Bool,
            ([(Term a [Char], Term a [Char])],
              (Strategy a [Char], [(Term a [Char], Term a [Char])])));
xml2dp_input xml2name =
  xml_choice ['d', 'p', 'I', 'n', 'p', 'u', 't']
    [xml4elements ['d', 'p', 'I', 'n', 'p', 'u', 't']
       (xml1element ['t', 'r', 's'] (rules xml2name) id)
       (xml1element ['d', 'p', 's'] (rules xml2name) id) (strategy xml2name)
       (bool ['m', 'i', 'n', 'i', 'm', 'a', 'l'])
       (\ r p s m -> (m, (p, (s, r)))),
      xml3elements ['d', 'p', 'I', 'n', 'p', 'u', 't']
        (xml1element ['t', 'r', 's'] (rules xml2name) id)
        (xml1element ['d', 'p', 's'] (rules xml2name) id)
        (bool ['m', 'i', 'n', 'i', 'm', 'a', 'l'])
        (\ r p m -> (m, (p, (No_Strategy, r))))];

i_o_strategy ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Sum (Strategy a [Char]) (Ostrategy a [Char]));
i_o_strategy xml2name =
  xml1element ['s', 't', 'r', 'a', 't', 'e', 'g', 'y']
    (xml_options
      [(['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't'],
         xml0elements ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't']
           (Inl Innermost)),
        (['o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't'],
          xml0elements ['o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't']
            (Inr Outermost)),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's'],
          xml_change (innermostLhss xml2name) (\ q -> Inl (Innermost_Q q)))])
    id;

xml2i_o_trs_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Sum (Strategy a [Char]) (Ostrategy a [Char]),
            ([(Term a [Char], Term a [Char])],
              Maybe [(Term a [Char], Term a [Char])]));
xml2i_o_trs_input xml2name =
  (\ x ->
    let {
      e = no_transform ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] x;
    } in (case x of {
           Xml tag [] [] option -> e;
           Xml tag [] (c : cs) Nothing ->
             (if less_eq_nat (size_list cs) (nat 2) &&
                   tag == ['t', 'r', 's', 'I', 'n', 'p', 'u', 't']
               then bindc (xml1element ['t', 'r', 's'] (rules xml2name) id c)
                      (\ r ->
                        bindc (if null cs ||
                                    not (get_name (hd cs) ==
  ['s', 't', 'r', 'a', 't', 'e', 'g', 'y'])
                                then returna (Inl No_Strategy, cs)
                                else bindc (i_o_strategy xml2name (hd cs))
                                       (\ s -> returna (s, tl cs)))
                          (\ (str, csa) ->
                            bindc (if null csa ||
not (get_name (hd csa) ==
      ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'])
                                    then returna (Nothing, csa)
                                    else bindc
   (xml1element
     ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's']
     (rules xml2name) id (hd csa))
   (\ s -> returna (Just s, tl csa)))
                              (\ (s_opt, csb) ->
                                (if null csb then returna (str, (r, s_opt))
                                  else e))))
               else e);
           Xml tag [] (c : cs) (Just _) -> e;
           Xml tag (_ : _) list3 option -> e;
         }));

xml2trs_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Strategy a [Char],
            ([(Term a [Char], Term a [Char])],
              Maybe [(Term a [Char], Term a [Char])]));
xml2trs_input xml2name x =
  bindc (xml2i_o_trs_input xml2name x)
    (\ a ->
      (case a of {
        (Inl is, (r, s)) -> returna (is, (r, s));
        (Inr _, (r, s)) ->
          errora
            ['o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't', ' ', 's', 't', 'r',
              'a', 't', 'e', 'g', 'y', ' ', 'n', 'o', 't', ' ', 's', 'u', 'p',
              'p', 'o', 'r', 't', 'e', 'd', ' ', 'a', 't', ' ', 't', 'h', 'i',
              's', ' ', 'p', 'o', 'i', 'n', 't'];
      }));

xml2otrs_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Ostrategy a [Char], [(Term a [Char], Term a [Char])]);
xml2otrs_input xml2name x =
  bindc (xml2i_o_trs_input xml2name x)
    (\ (ios, (r, s)) ->
      (case (ios, s) of {
        (Inl _, b) ->
          errora
            ['n', 'o', 't', ' ', 'a', 'n', ' ', 'o', 'u', 't', 'e', 'r', 'm',
              'o', 's', 't', ' ', 's', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ',
              'o', 'r', ' ', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'r',
              'u', 'l', 'e', 's', ' ', 'p', 'r', 'e', 's', 'e', 'n', 't'];
        (Inr os, Nothing) -> returna (os, r);
        (Inr os, Just _) ->
          errora
            ['n', 'o', 't', ' ', 'a', 'n', ' ', 'o', 'u', 't', 'e', 'r', 'm',
              'o', 's', 't', ' ', 's', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ',
              'o', 'r', ' ', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'r',
              'u', 'l', 'e', 's', ' ', 'p', 'r', 'e', 's', 'e', 'n', 't'];
      }));

xml2completion_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          ([(Term a [Char], Term a [Char])], [(Term a [Char], Term a [Char])]);
xml2completion_input xml2name =
  xml2elements
    ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'I', 'n', 'p', 'u', 't']
    (xml1element ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] (rules xml2name)
      id)
    (xml1element ['t', 'r', 's'] (rules xml2name) id) (\ a b -> (a, b));

complexity_class :: Xml -> Parse_result [Char] Complexity_class;
complexity_class =
  xml_change (nata ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l'])
    Comp_Poly;

complexity_measure ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Complexity_measure a [Char]);
complexity_measure xml2name =
  xml_options
    [(['d', 'e', 'r', 'i', 'v', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'C', 'o',
        'm', 'p', 'l', 'e', 'x', 'i', 't', 'y'],
       xml1element
         ['d', 'e', 'r', 'i', 'v', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'C', 'o',
           'm', 'p', 'l', 'e', 'x', 'i', 't', 'y']
         (signature xml2name) Derivational_Complexity),
      (['r', 'u', 'n', 't', 'i', 'm', 'e', 'C', 'o', 'm', 'p', 'l', 'e', 'x',
         'i', 't', 'y'],
        xml2elements
          ['r', 'u', 'n', 't', 'i', 'm', 'e', 'C', 'o', 'm', 'p', 'l', 'e', 'x',
            'i', 't', 'y']
          (signature xml2name) (signature xml2name) Runtime_Complexity)];

xml2complexity_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Strategy a [Char],
            ([(Term a [Char], Term a [Char])],
              (Maybe [(Term a [Char], Term a [Char])],
                (Complexity_measure a [Char], Complexity_class))));
xml2complexity_input xml2name =
  xml3elements
    ['c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', 'I', 'n', 'p', 'u', 't']
    (xml2trs_input xml2name) (complexity_measure xml2name) complexity_class
    (\ (q, (r, s)) cm cc -> (q, (r, (s, (cm, cc)))));

xml2equational_input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          ([(Term a [Char], Term a [Char])], (Term a [Char], Term a [Char]));
xml2equational_input xml2name =
  xml2elements
    ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'R', 'e', 'a', 's', 'o',
      'n', 'i', 'n', 'g', 'I', 'n', 'p', 'u', 't']
    (xml1element ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] (rules xml2name)
      id)
    (xml2elements ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n'] (term xml2name)
      (term xml2name) (\ a b -> (a, b)))
    (\ a b -> (a, b));

xml2input ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (Sum (Bool,
                 ([(Term a [Char], Term a [Char])],
                   (Strategy a [Char], [(Term a [Char], Term a [Char])])))
            (Sum (Strategy a [Char],
                   ([(Term a [Char], Term a [Char])],
                     Maybe [(Term a [Char], Term a [Char])]))
              (Sum ([(Term a [Char], Term a [Char])],
                     [(Term a [Char], Term a [Char])])
                (Sum ([(Term a [Char], Term a [Char])],
                       (Term a [Char], Term a [Char]))
                  (Sum (Strategy a [Char],
                         ([(Term a [Char], Term a [Char])],
                           (Maybe [(Term a [Char], Term a [Char])],
                             (Complexity_measure a [Char], Complexity_class))))
                    (Ostrategy a [Char], [(Term a [Char], Term a [Char])]))))));
xml2input xml2name =
  xml1element ['i', 'n', 'p', 'u', 't']
    (xml_choice ['i', 'n', 'p', 'u', 't']
      [xml_change (xml2dp_input xml2name) Inl,
        xml_change (xml2trs_input xml2name) (Inr . Inl),
        xml_change (xml2completion_input xml2name) ((Inr . Inr) . Inl),
        xml_change (xml2equational_input xml2name) (((Inr . Inr) . Inr) . Inl),
        xml_change (xml2complexity_input xml2name)
          ((((Inr . Inr) . Inr) . Inr) . Inl),
        xml_change (xml2otrs_input xml2name)
          ((((Inr . Inr) . Inr) . Inr) . Inr)])
    id;

xml_rules ::
  forall a b c.
    (Showa a, Showa b,
      Showa c) => [Char] -> [(Term (Lab a [b]) c, Term (Lab a [b]) c)] -> Xml;
xml_rules tag rls =
  Xml tag []
    [Xml ['r', 'u', 'l', 'e', 's'] []
       (map (\ (l, r) ->
              Xml ['r', 'u', 'l', 'e'] []
                [Xml ['l', 'h', 's'] [] [xml_term l] Nothing,
                  Xml ['r', 'h', 's'] [] [xml_term r] Nothing]
                Nothing)
         rls)
       Nothing]
    Nothing;

gen_list_to_set_aux :: forall a b. (a -> b -> b) -> b -> [a] -> b;
gen_list_to_set_aux ins accs [] = accs;
gen_list_to_set_aux ins accs (x : l) = gen_list_to_set_aux ins (ins x accs) l;

gen_list_to_set :: forall a b. (() -> a) -> (b -> a -> a) -> [b] -> a;
gen_list_to_set empt ins = gen_list_to_set_aux ins (empt ());

list_to_rs :: forall a. (Linorder a) => [a] -> Rbt a ();
list_to_rs = gen_list_to_set rs_empty rs_ins;

instance (Eq a, Order a) => Preorder [a] where {
};

instance (Eq a, Order a) => Order [a] where {
};

instance (Eq a, Linorder a) => Linorder [a] where {
};

instance Orda Char where {
  less_eq = (\ a b -> a <= b);
  less = (\ a b -> a < b);
};

instance Preorder Char where {
};

instance Order Char where {
};

instance Linorder Char where {
};

icap_impla ::
  forall a.
    (Eq a,
      Linorder a) => [(Term a [Char], Term a [Char])] ->
                       [Term a [Char]] ->
                         [Term a [Char]] ->
                           Term a [Char] -> Term a (Sum () [Char]);
icap_impla r q =
  let {
    ls = map fst r;
    isnf = is_NF (any is_Var q) q (term_map q);
    nf = nF_subset ls q;
    ic = icap_impl nf isnf ls;
  } in (\ s ->
         let {
           sa = map (map_vars (\ a -> 'x' : a)) s;
           sxs = list_to_rs (concatMap vars_term_lista sa);
           sx = (\ x -> rs_memb x sxs);
         } in (\ t -> ic sa sx (map_vars (\ a -> 'x' : a) t)));

kbo_nstrict ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ((a, Nat) -> Nat) ->
                    ((a, Nat) -> Nat) ->
                      Nat ->
                        (a -> Bool) ->
                          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
kbo_nstrict pr w w0 least =
  (\ (s, t) ->
    check (snd (kbo w w0 pr least s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', '=', 'K', 'B', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

lexord_code :: forall a. (Eq a) => (a -> a -> Bool) -> [a] -> [a] -> Bool;
lexord_code lt uu [] = False;
lexord_code lt [] (v : va) = True;
lexord_code lt (x : xs) (y : ys) = lt x y || x == y && lexord_code lt xs ys;

less_lab ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Lab a b -> Lab a b -> Bool;
less_lab (UnLab f) (UnLab g) = less f g;
less_lab (UnLab f) (Lab v va) = True;
less_lab (UnLab f) (FunLab v va) = True;
less_lab (UnLab f) (Sharp v) = True;
less_lab (Lab f la) (Lab g l) = less la l || la == l && less_lab f g;
less_lab (Lab f l) (UnLab uv) = False;
less_lab (Lab f l) (FunLab v va) = True;
less_lab (Lab f l) (Sharp v) = True;
less_lab (Sharp f) (Sharp g) = less_lab f g;
less_lab (Sharp f) (UnLab ux) = False;
less_lab (Sharp f) (Lab uy uz) = False;
less_lab (Sharp f) (FunLab v vb) = True;
less_lab (FunLab f fs) (FunLab g gs) =
  less_lab f g || equal_lab f g && lexord_code less_lab fs gs;
less_lab (FunLab f fs) (Lab v va) = False;
less_lab (FunLab f fs) (UnLab v) = False;
less_lab (FunLab f fs) (Sharp v) = False;

less_eq_lab ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Lab a b -> Lab a b -> Bool;
less_eq_lab xs ys = less_lab xs ys || equal_lab xs ys;

instance (Eq a, Orda a, Eq b, Orda b) => Orda (Lab a b) where {
  less_eq = less_eq_lab;
  less = less_lab;
};

gea :: forall a b. Poly_c_ext a b -> a -> a -> Bool;
gea (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
      arc_pos check_mono maxa derive_complexity more)
  = ge;

wf :: forall a b. Poly_c_ext a b -> a -> Bool;
wf (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
     arc_pos check_mono maxa derive_complexity more)
  = wf;

class (Comm_semiring_1 a, Max_ordered_semiring_1 a) => Poly_carrier a where {
};

default_I :: forall a. (Poly_carrier a) => a -> Nat -> [([(Nat, Nat)], a)];
default_I def n =
  ([], def) : map (\ i -> ([(i, one_nat)], one)) (upt Zero_nat n);

poly_power ::
  forall a b.
    (Eq a, Eq b,
      Comm_semiring_1 b) => [([(a, Nat)], b)] -> Nat -> [([(a, Nat)], b)];
poly_power uu Zero_nat = one_poly;
poly_power p (Suc n) = poly_mult p (poly_power p n);

monom_subst ::
  forall a b c.
    (Eq b, Eq c,
      Comm_semiring_1 c) => (a -> [([(b, Nat)], c)]) ->
                              [(a, Nat)] -> [([(b, Nat)], c)];
monom_subst sigma [] = one_poly;
monom_subst sigma ((x, p) : m) =
  poly_mult (poly_power (sigma x) p) (monom_subst sigma m);

poly_subst ::
  forall a b c.
    (Eq b, Eq c,
      Comm_semiring_1 c) => (a -> [([(b, Nat)], c)]) ->
                              [([(a, Nat)], c)] -> [([(b, Nat)], c)];
poly_subst sigma [] = zero_poly;
poly_subst sigma ((m, c) : p) =
  poly_add (poly_mult [([], c)] (monom_subst sigma m)) (poly_subst sigma p);

eval_term ::
  forall a b c.
    (Eq b, Poly_carrier b,
      Eq c) => ((a, Nat) -> [([(Nat, Nat)], b)]) ->
                 Term a c -> [([(c, Nat)], b)];
eval_term uu (Var x) = [([(x, one_nat)], one)];
eval_term i (Fun f ts) =
  let {
    ps = map (eval_term i) ts;
    n = size_list ts;
  } in poly_subst (\ ia -> (if less_nat ia n then nth ps ia else zero_poly))
         (i (f, n));

eval_monom :: forall a b. (Comm_semiring_1 b) => (a -> b) -> [(a, Nat)] -> b;
eval_monom alpha [] = one;
eval_monom alpha ((x, p) : m) = times (eval_monom alpha m) (power (alpha x) p);

eval_poly ::
  forall a b. (Comm_semiring_1 b) => (a -> b) -> [([(a, Nat)], b)] -> b;
eval_poly alpha [] = zero;
eval_poly alpha (mc : p) =
  plus (times (eval_monom alpha (fst mc)) (snd mc)) (eval_poly alpha p);

scnp_arity :: forall a. [((a, Nat), [(Nat, Nat)])] -> Nat;
scnp_arity af = max_list (map (\ (_, a) -> size_list a) af);

it_set_to_list ::
  forall a b.
    (a -> ([b] -> Bool) -> (b -> [b] -> [b]) -> [b] -> [b]) -> a -> [b];
it_set_to_list iti s = iti s (\ _ -> True) (\ a b -> a : b) [];

check_supt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_supt s t =
  check (supt_impl s t)
    (shows_prec_term Zero_nat t .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'r', 'o', 'p',
          'e', 'r', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f',
          ' '] .
        shows_prec_term Zero_nat s);

shows_ctxt ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> Ctxt a b -> [Char] -> [Char];
shows_ctxt fun var Hole = shows_string ['[', ']'];
shows_ctxt fun var (More f ss1 d ss2) =
  fun f .
    shows_string ['('] .
      shows_list_gen (shows_term fun var) [] [] [',', ' '] [',', ' '] ss1 .
        shows_ctxt fun var d .
          shows_list_gen (shows_term fun var) [')'] [',', ' '] [',', ' '] [')']
            ss2;

desca :: forall a b c. Redtriple_ext a b c -> [Char] -> [Char];
desca (Redtriple_ext valid s ns nst af mono desc cpx more) = desc;

mono ::
  forall a b c.
    Redtriple_ext a b c -> [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
mono (Redtriple_ext valid s ns nst af mono desc cpx more) = mono;

nsta ::
  forall a b c.
    Root_redtriple_ext a b c ->
      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
nsta (Root_redtriple_ext valid s ns nst af aft desc more) = nst;

eq_rule_mod_vars ::
  forall a b.
    (Eq a, Eq b) => (Term a b, Term a b) -> (Term a b, Term a b) -> Bool;
eq_rule_mod_vars lr st = instance_rule lr st && instance_rule st lr;

generate_var :: Nat -> [Char];
generate_var i = 'x' : shows_prec_nat Zero_nat i [];

uncurry_term ::
  forall a b. (Eq a) => a -> (a -> Nat -> [a]) -> Term a b -> Term a b;
uncurry_term a sm t =
  (case unapp a t of {
    (Var x, ts) -> apply_args a (Var x) (map (uncurry_term a sm) ts);
    (Fun f ss, ts) ->
      let {
        n = size_list ss;
        uss = map (uncurry_term a sm) ss;
        uts = map (uncurry_term a sm) ts;
        aa = aarity sm f n;
        m = mina (size_list ts) aa;
        fm = get_symbol sm f n m;
      } in apply_args a (Fun fm (uss ++ takea m uts)) (dropa m uts);
  });

parse_name :: [Char] -> Sum [Char] ([Char], [Char]);
parse_name =
  binda (many (\ c ->
                'a' <= c && c <= 'z' ||
                  'A' <= c && c <= 'Z' ||
                    '0' <= c && c <= '9' || membera ['_', '&', ';', ':'] c))
    (\ n ->
      binda (binda (many (membera [' ', newline, '\t', '\r']))
              (\ _ ts -> Inr ((), ts)))
        (\ _ ->
          (if null n then (\ _ -> Inl letters) else (\ ts -> Inr (n, ts)))));

parse_text :: [Char] -> Sum [Char] (Maybe [Char], [Char]);
parse_text =
  binda (many (\ y -> not ('<' == y)))
    (\ ts ->
      let {
        text = trim ts;
      } in (if null text then (\ tsa -> Inr (Nothing, tsa))
             else (\ tsa -> Inr (Just (reverse (trim (reverse text))), tsa))));

parse_attribute_value :: [Char] -> Sum [Char] ([Char], [Char]);
parse_attribute_value =
  binda (oneof [['\"']])
    (\ _ ->
      binda (many (\ y -> not ('\"' == y)))
        (\ v -> binda (oneof [['\"']]) (\ _ ts -> Inr (v, ts))));

parse_attributes :: [Char] -> Sum [Char] ([([Char], [Char])], [Char]);
parse_attributes [] = Inr ([], []);
parse_attributes (c : s) =
  (if membera ['/', '>'] c then Inr ([], c : s)
    else binda parse_name
           (\ k ->
             binda (oneof [['=']])
               (\ _ ->
                 binda parse_attribute_value
                   (\ v ->
                     binda parse_attributes
                       (\ atts ts -> Inr ((k, v) : atts, ts)))))
           (c : s));

parse_nodes :: [Char] -> Sum [Char] ([Xml], [Char]);
parse_nodes ts =
  (if null ts || takea (nat 2) ts == ['<', '/'] then Inr ([], ts)
    else binda (oneof [['<']])
           (\ _ ->
             binda parse_name
               (\ n ->
                 binda parse_attributes
                   (\ atts ->
                     binda (oneof [['/', '>'], ['>']])
                       (\ e tsa ->
                         (if e == ['/', '>']
                           then binda parse_nodes
                                  (\ cs tsb ->
                                    Inr (Xml n atts [] Nothing : cs, tsb))
                                  tsa
                           else binda parse_text
                                  (\ t ->
                                    binda parse_nodes
                                      (\ cs ->
binda (oneof [['<', '/']])
  (\ _ ->
    binda (oneof [n])
      (\ _ ->
        binda (binda (many (membera [' ', newline, '\t', '\r']))
                (\ _ tsb -> Inr ((), tsb)))
          (\ _ ->
            binda (oneof [['>']])
              (\ _ ->
                binda parse_nodes
                  (\ ns tsb -> Inr (Xml n atts cs t : ns, tsb))))))))
                                  tsa)))))
           ts);

parse_node :: [Char] -> Sum [Char] (Xml, [Char]);
parse_node =
  binda (oneof [['<']])
    (\ _ ->
      binda parse_name
        (\ n ->
          binda parse_attributes
            (\ atts ->
              binda (oneof [['/', '>'], ['>']])
                (\ e ->
                  (if e == ['/', '>']
                    then (\ ts -> Inr (Xml n atts [] Nothing, ts))
                    else binda parse_text
                           (\ t ->
                             binda parse_nodes
                               (\ cs ->
                                 binda (oneof [['<', '/'] ++ n ++ ['>']])
                                   (\ _ ts -> Inr (Xml n atts cs t, ts)))))))));

scan_upto :: [Char] -> [Char] -> Sum [Char] ([Char], [Char]);
scan_upto end [] =
  Inl (['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'e',
         'n', 'd', '-', 'm', 'a', 'r', 'k', 'e', 'r', ' '] ++
        quote (shows_prec_list Zero_nat end) []);
scan_upto end (t : ts) =
  (if map snd (zip end (t : ts)) == end
    then Inr (end, dropa (size_list end) (t : ts))
    else bindb (scan_upto end ts) (\ (res, tsa) -> Inr (t : res, tsa)));

parse_header :: [Char] -> Sum [Char] ([Char], [Char]);
parse_header ts =
  (if takea (nat 2) (trim ts) == ['<', '?']
    then binda (scan_upto ['?', '>'])
           (\ h -> binda parse_header (\ hs tsa -> Inr (h ++ hs, tsa))) ts
    else binda (binda (many (membera [' ', newline, '\t', '\r']))
                 (\ _ tsa -> Inr ((), tsa)))
           (\ _ tsa -> Inr ([], tsa)) ts);

update_tokens :: forall a. ([a] -> [a]) -> [a] -> Sum [Char] ([a], [a]);
update_tokens f ts = Inr (ts, f ts);

remove_comments_aux :: Bool -> [Char] -> [Char];
remove_comments_aux False (c : cs) =
  (if c == '<' && takea (nat 3) cs == ['!', '-', '-']
    then remove_comments_aux True (tl cs)
    else c : remove_comments_aux False cs);
remove_comments_aux True (c : cs) =
  (if c == '-' && takea (nat 2) cs == ['-', '>']
    then remove_comments_aux False (dropa (nat 2) cs)
    else remove_comments_aux True cs);
remove_comments_aux uu [] = [];

remove_comments_aux_acc :: [Char] -> Bool -> [Char] -> [Char];
remove_comments_aux_acc a False (c : cs) =
  (if c == '<' && takea (nat 3) cs == ['!', '-', '-']
    then remove_comments_aux_acc a True (tl cs)
    else remove_comments_aux_acc (c : a) False cs);
remove_comments_aux_acc a True (c : cs) =
  (if c == '-' && takea (nat 2) cs == ['-', '>']
    then remove_comments_aux_acc a False (dropa (nat 2) cs)
    else remove_comments_aux_acc a True cs);
remove_comments_aux_acc a uu [] = a;

remove_comments :: [Char] -> [Char];
remove_comments =
  (if False then reverse . remove_comments_aux_acc [] False
    else remove_comments_aux False);

parse_doc :: [Char] -> Sum [Char] (Xmldoc, [Char]);
parse_doc =
  debug ['0']
    ['p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 't', 'x', 't', ' ', 't', 'o', ' ',
      'x', 'm', 'l']
    (binda (update_tokens remove_comments)
      (\ _ ->
        binda parse_header
          (\ h ->
            binda parse_node
              (\ xml -> binda eoi (\ _ ts -> Inr (Xmldoc h xml, ts))))));

data Filtered a = FPair a Nat;

conversion ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          [(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))];
conversion xml2name =
  xml1many_elements ['c', 'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n']
    (xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'] (term xml2name)
      id)
    (estep xml2name) (\ _ x -> x);

plain_name :: Xml -> Parse_result [Char] (Lab [Char] [Nat]);
plain_name x =
  xml_options
    [(['n', 'a', 'm', 'e'], xml_change (text ['n', 'a', 'm', 'e']) UnLab),
      (['s', 'h', 'a', 'r', 'p'],
        xml1element ['s', 'h', 'a', 'r', 'p'] plain_name Sharp),
      (['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'],
        xml2elements
          ['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l']
          plain_name
          (xml_options
            [(['n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l'],
               xml_many_elements
                 ['n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l']
                 (nata ['n', 'u', 'm', 'b', 'e', 'r']) Inl),
              (['s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l'],
                xml_many_elements
                  ['s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l']
                  plain_name Inr)])
          (\ f a ->
            (case a of {
              Inl aa -> Lab f aa;
              Inr aa -> FunLab f aa;
            })))]
    x;

xml1or2many_elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (a -> Maybe b -> [c] -> d) -> Xml -> Parse_result [Char] d;
xml1or2many_elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && is_none text && less_eq_nat (nat 2) (size_list cs)
    then bindc (p1 (nth cs Zero_nat))
           (\ x ->
             catchb
               (bindc (p2 (nth cs one_nat))
                 (\ y ->
                   bindc (map_parse_result p3 (tl (tl cs)))
                     (\ xs -> returna (f x (Just y) xs))))
               (\ _ ->
                 bindc (map_parse_result p3 (tl cs))
                   (\ xs -> returna (f x Nothing xs))))
    else no_transform tag (Xml name atts cs text));

sl_variant ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Sl_variant a [Char]);
sl_variant xml2name =
  xml1element ['m', 'o', 'd', 'e', 'l']
    (xml_options
      [(['f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e', 'l'],
         xml1or2many_elements
           ['f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e', 'l']
           (xml_change
             (nata ['c', 'a', 'r', 'r', 'i', 'e', 'r', 'S', 'i', 'z', 'e'])
             (\ n -> minus_nat n one_nat))
           (xml1element ['t', 'u', 'p', 'l', 'e', 'O', 'r', 'd', 'e', 'r']
             (xml0elements ['p', 'o', 'i', 'n', 't', 'W', 'i', 's', 'e'] ()) id)
           (xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
             (nata ['a', 'r', 'i', 't', 'y']) arith_fun
             (\ f a -> (\ b -> ((f, a), b))))
           (\ n model xs ->
             (if is_none model then Finitelab (SL_Inter n xs)
               else QuasiFinitelab (SL_Inter n xs) []))),
        (['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'],
          xml_choice
            ['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g']
            [xml0elements
               ['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g']
               (Rootlab Nothing),
              xml1element
                ['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g']
                xml2name (\ f -> Rootlab (Just (f, one_nat)))])])
    id;

xml2many_elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (a -> b -> [c] -> d) -> Xml -> Parse_result [Char] d;
xml2many_elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        null atts && is_none text && less_eq_nat (nat 2) (size_list cs)
    then bindc (p1 (nth cs Zero_nat))
           (\ x ->
             bindc (p2 (nth cs one_nat))
               (\ y ->
                 bindc (map_parse_result p3 (tl (tl cs)))
                   (\ xs -> returna (f x y xs))))
    else no_transform tag (Xml name atts cs text));

transition_lhs ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Sum ((a, Nat), [[Char]]) [Char]);
transition_lhs xml2name x =
  (if let {
        (Xml _ _ cs _) = x;
      } in equal_nat (size_list cs) one_nat
    then xml1element ['l', 'h', 's'] state Inr x
    else xml2many_elements ['l', 'h', 's'] xml2name
           (nata ['h', 'e', 'i', 'g', 'h', 't']) state
           (\ f h qs -> Inl ((f, h), qs)) x);

transition ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char] (Sum (Ta_rule [Char] (a, Nat)) ([Char], [Char]));
transition xml2name =
  xml2elements ['t', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n']
    (transition_lhs xml2name) (xml1element ['r', 'h', 's'] state id)
    (\ lhs q ->
      (case lhs of {
        Inl (f, qs) -> Inl (TA_rule f qs q);
        Inr qa -> Inr (qa, q);
      }));

less_term ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Term a b -> Term a b -> Bool;
less_term (Var f) (Var g) = less f g;
less_term (Var f) (Fun v va) = True;
less_term (Fun f fs) (Fun g gs) =
  less f g || f == g && lexord_code less_term fs gs;
less_term (Fun f fs) (Var v) = False;

less_eq_term ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Term a b -> Term a b -> Bool;
less_eq_term xs ys = less_term xs ys || equal_term xs ys;

instance (Eq a, Orda a, Eq b, Orda b) => Orda (Term a b) where {
  less_eq = less_eq_term;
  less = less_term;
};

maxb :: forall a b. Poly_c_ext a b -> a -> a -> a;
maxb (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
       arc_pos check_mono maxa derive_complexity more)
  = maxa;

all_interval_nat :: (Nat -> Bool) -> Nat -> Nat -> Bool;
all_interval_nat p i j = less_eq_nat j i || p i && all_interval_nat p (Suc i) j;

gen_list_to_map_aux :: forall a b c. (a -> b -> c -> c) -> c -> [(a, b)] -> c;
gen_list_to_map_aux upd accs [] = accs;
gen_list_to_map_aux upd accs ((k, v) : l) =
  gen_list_to_map_aux upd (upd k v accs) l;

gen_list_to_map ::
  forall a b c. (() -> a) -> (b -> c -> a -> a) -> [(b, c)] -> a;
gen_list_to_map empt upd l = gen_list_to_map_aux upd (empt ()) (reverse l);

delete_value ::
  forall a b.
    (Eq a, Linorder b) => (a -> Maybe b) -> a -> Rbt b [a] -> Rbt b [a];
delete_value key v m =
  (case key v of {
    Nothing -> m;
    Just k ->
      (case rm_lookup k m of {
        Nothing -> m;
        Just vs -> rm_update k (removeAll v vs) m;
      });
  });

insert_value ::
  forall a b.
    (Eq a, Linorder b) => (a -> Maybe b) -> a -> Rbt b [a] -> Rbt b [a];
insert_value key v m =
  (case key v of {
    Nothing -> m;
    Just k ->
      (case rm_lookup k m of {
        Nothing -> rm_update_dj k [v] m;
        Just vs -> rm_update k (insertb v vs) m;
      });
  });

shows_monom :: forall a. (Showa a) => [(a, Nat)] -> [Char] -> [Char];
shows_monom [(x, p)] =
  (if equal_nat p one_nat then shows_prec Zero_nat x
    else shows_prec Zero_nat x .
           shows_string ['^'] . shows_prec_nat Zero_nat p);
shows_monom ((x, p) : v : va) =
  (if equal_nat p one_nat then shows_prec Zero_nat x
    else shows_prec Zero_nat x .
           shows_string ['^'] . shows_prec_nat Zero_nat p) .
    shows_string ['*'] . shows_monom (v : va);
shows_monom [] = shows_string ['1'];

shows_poly ::
  forall a b.
    (Showa a, Eq b, Poly_carrier b,
      Showa b) => [([(a, Nat)], b)] -> [Char] -> [Char];
shows_poly [] = shows_string ['0'];
shows_poly ((m, c) : p) =
  (if c == one then shows_monom m
    else (if null m then shows_prec Zero_nat c
           else shows_prec Zero_nat c . shows_string ['*'] . shows_monom m)) .
    (if null p then shows_string []
      else shows_string [' ', '+', ' '] . shows_poly p);

poly_split ::
  forall a b.
    (Eq a, Zero b) => [(a, Nat)] -> [([(a, Nat)], b)] -> (b, [([(a, Nat)], b)]);
poly_split m p =
  (case extracta (\ (n, _) -> eq_monom m n) p of {
    Nothing -> (zero, p);
    Just (p1, ((_, c), p2)) -> (c, p1 ++ p2);
  });

data Dp_disproof a b =
  DP_Loop (Term a b) [(Pos, ((Term a b, Term a b), (Bool, Term a b)))]
    [(b, Term a b)] (Ctxt a b)
  | DP_Nonloop (Non_loop_prf a b)
  | DP_Rule_Removal (Maybe [(Term a b, Term a b)])
      (Maybe [(Term a b, Term a b)]) (Dp_disproof a b)
  | DP_Q_Increase [Term a b] (Dp_disproof a b)
  | DP_Q_Reduction [Term a b] (Dp_disproof a b)
  | DP_Termination_Switch (Join_info a) (Dp_disproof a b)
  | DP_Instantiation [(Term a b, Term a b)] (Dp_disproof a b);

data Trs_disproof a b =
  TRS_Loop (Term a b) [(Pos, ((Term a b, Term a b), Term a b))] [(b, Term a b)]
    (Ctxt a b)
  | TRS_Not_Well_Formed
  | TRS_Rule_Removal [(Term a b, Term a b)] (Trs_disproof a b)
  | TRS_String_Reversal (Trs_disproof a b)
  | TRS_DP_Trans [(Term a b, Term a b)] (Dp_disproof a b)
  | TRS_Nonloop (Non_loop_prf a b)
  | TRS_Q_Increase [Term a b] (Trs_disproof a b);

data Complexity_proof a b =
  Rule_Shift_Complexity (Redtriple_impl a) [(Term a b, Term a b)]
    (Complexity_proof a b)
  | RisEmpty_Complexity;

data Otrs_disproof a b =
  OTRS_Loop (Term a b) [(Pos, ((Term a b, Term a b), Term a b))] [(b, Term a b)]
    (Ctxt a b)
  | OTRS_Rule_Removal [(Term a b, Term a b)] (Otrs_disproof a b);

data Reltrs_disproof a b =
  Rel_Loop (Term a b) [(Pos, ((Term a b, Term a b), (Bool, Term a b)))]
    [(b, Term a b)] (Ctxt a b)
  | Rel_Not_Well_Formed
  | Rel_Rule_Removal (Maybe [(Term a b, Term a b)])
      (Maybe [(Term a b, Term a b)]) (Reltrs_disproof a b)
  | Rel_R_Not_SN (Trs_disproof a b);

data Completion_proof a b c =
  SN_WCR_Eq (Join_info (Lab a b)) (Trs_proof a b c)
    [((Term (Lab a b) c, Term (Lab a b) c),
       [(Pos, ((Term (Lab a b) c, Term (Lab a b) c),
                (Bool, Term (Lab a b) c)))])]
    (Maybe [((Term (Lab a b) c, Term (Lab a b) c),
              [(Pos, ((Term (Lab a b) c, Term (Lab a b) c),
                       (Bool, Term (Lab a b) c)))])]);

data Eq_proof a b = Refl (Term a b) | Sym (Eq_proof a b)
  | Trans (Eq_proof a b) (Eq_proof a b)
  | Assm (Term a b, Term a b) (b -> Term a b) | Cong a [Eq_proof a b];

data Equational_proof a b c = Equational_Proof_Tree (Eq_proof (Lab a b) c)
  | Completion_and_Normalization [(Term (Lab a b) c, Term (Lab a b) c)]
      (Completion_proof a b c)
  | Conversion
      [(Pos, ((Term (Lab a b) c, Term (Lab a b) c), (Bool, Term (Lab a b) c)))];

data Equational_disproof a b c =
  Completion_and_Normalization_Different [(Term (Lab a b) c, Term (Lab a b) c)]
    (Completion_proof a b c);

data Cert_problem a b c =
  TRS_Proof (Strategy (Lab a b) c) [(Term (Lab a b) c, Term (Lab a b) c)]
    (Maybe [(Term (Lab a b) c, Term (Lab a b) c)]) (Trs_proof a b c)
  | Complexity_Proof (Strategy (Lab a b) c)
      [(Term (Lab a b) c, Term (Lab a b) c)]
      (Maybe [(Term (Lab a b) c, Term (Lab a b) c)])
      (Complexity_measure (Lab a b) c) Complexity_class
      (Complexity_proof (Lab a b) c)
  | DP_Proof Bool [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Strategy (Lab a b) c)
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Dp_proof a b c)
  | DP_Disproof Bool [(Term (Lab a b) c, Term (Lab a b) c)]
      (Strategy (Lab a b) c) [(Term (Lab a b) c, Term (Lab a b) c)]
      (Dp_disproof (Lab a b) c)
  | TRS_Disproof (Strategy (Lab a b) c) [(Term (Lab a b) c, Term (Lab a b) c)]
      (Trs_disproof (Lab a b) c)
  | OTRS_Disproof (Ostrategy (Lab a b) c) [(Term (Lab a b) c, Term (Lab a b) c)]
      (Otrs_disproof (Lab a b) c)
  | Rel_TRS_Disproof (Strategy (Lab a b) c)
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Reltrs_disproof (Lab a b) c)
  | TRS_Confluence [(Term (Lab a b) c, Term (Lab a b) c)] (Cr_proof a b c)
  | TRS_Non_Confluence [(Term (Lab a b) c, Term (Lab a b) c)] (Ncr_proof a b c)
  | Completion_Proof [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (Completion_proof a b c)
  | Equational_Proof [(Term (Lab a b) c, Term (Lab a b) c)]
      (Term (Lab a b) c, Term (Lab a b) c) (Equational_proof a b c)
  | Equational_Disproof [(Term (Lab a b) c, Term (Lab a b) c)]
      (Term (Lab a b) c, Term (Lab a b) c) (Equational_disproof a b c);

xml_tag :: forall a b c. Cert_problem a b c -> [Char];
xml_tag (TRS_Proof uu uv (Just uw) ux) =
  ['<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n',
    'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (TRS_Proof uy uz Nothing va) =
  ['<', 't', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n',
    'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (Complexity_Proof vb vc vd ve vf vg) =
  ['<', 'c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', 'P', 'r', 'o', 'o',
    'f', '>'];
xml_tag (TRS_Disproof vh vi vj) =
  ['<', 't', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
    'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (OTRS_Disproof vk vl vm) =
  ['<', 't', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
    'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (Rel_TRS_Disproof vn vo vp vq) =
  ['<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e', 'r',
    'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (DP_Proof vr vs vt vu vv vw vx) =
  ['<', 'd', 'p', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (DP_Disproof vy vz wa wb wc) =
  ['<', 'd', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
    'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (TRS_Confluence wd we) = ['<', 'c', 'r', 'P', 'r', 'o', 'o', 'f', '>'];
xml_tag (TRS_Non_Confluence wf wg) =
  ['<', 'c', 'r', 'D', 'i', 's', 'p', 'r', 'o', 'o', 'f', '>'];
xml_tag (Completion_Proof wh wi wj) =
  ['<', 'c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
    'f', '>'];
xml_tag (Equational_Proof wk wl wm) =
  ['<', 'e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'P', 'r', 'o', 'o',
    'f', '>'];
xml_tag (Equational_Disproof wn wo wp) =
  ['<', 'e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'D', 'i', 's', 'p',
    'r', 'o', 'o', 'f', '>'];

list_to_rm :: forall a b. (Linorder a) => [(a, b)] -> Rbt a b;
list_to_rm = gen_list_to_map rm_empty rm_update;

rm_isEmpty :: forall a b. (Eq a, Linorder a, Eq b) => Rbt a b -> Bool;
rm_isEmpty m = m == empty;

rs_isEmpty :: forall a. (Eq a, Linorder a) => Rbt a () -> Bool;
rs_isEmpty = rm_isEmpty;

rs_reverse_iterateoi ::
  forall a b.
    (Linorder a) => Rbt a () -> (b -> Bool) -> (a -> b -> b) -> b -> b;
rs_reverse_iterateoi = s_iteratei rm_reverse_iterateoia;

rs_to_list :: forall a. (Linorder a) => Rbt a () -> [a];
rs_to_list = it_set_to_list rs_reverse_iterateoi;

delete_values ::
  forall a b.
    (Eq a, Linorder b) => (a -> Maybe b) -> [a] -> Rbt b [a] -> Rbt b [a];
delete_values uu [] m = m;
delete_values key (v : vs) m = delete_value key v (delete_values key vs m);

delete_rules ::
  forall a b c.
    (Eq a, Eq b, Linorder b,
      Eq c) => a -> [(Term b c, Term b c)] ->
                      Rbt (b, Nat) [(a, (Term b c, Term b c))] ->
                        Rbt (b, Nat) [(a, (Term b c, Term b c))];
delete_rules a rs = delete_values key (map (\ b -> (a, b)) rs);

insert_values ::
  forall a b.
    (Eq a, Linorder b) => (a -> Maybe b) -> [a] -> Rbt b [a] -> Rbt b [a];
insert_values uu [] m = m;
insert_values key (v : vs) m = insert_value key v (insert_values key vs m);

insert_rules ::
  forall a b c.
    (Eq a, Eq b, Linorder b,
      Eq c) => a -> [(Term b c, Term b c)] ->
                      Rbt (b, Nat) [(a, (Term b c, Term b c))] ->
                        Rbt (b, Nat) [(a, (Term b c, Term b c))];
insert_rules a rs = insert_values key (map (\ b -> (a, b)) rs);

qc :: forall a b c d. Dpp_ops_ext a b c d -> a -> [Term b c];
qc (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
     rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
     intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
     split_pairs split_rules mk minimal nfs wwf_rules more)
  = q;

mkc ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      Bool ->
        Bool ->
          [(Term b c, Term b c)] ->
            [(Term b c, Term b c)] ->
              [Term b c] ->
                [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a;
mkc (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
      rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map
      reverse_rules_map intersect_pairs replace_pair intersect_rules delete_P_Pw
      delete_R_Rw split_pairs split_rules mk minimal nfs wwf_rules more)
  = mk;

nfsc :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
nfsc (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
       rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map
       reverse_rules_map intersect_pairs replace_pair intersect_rules
       delete_P_Pw delete_R_Rw split_pairs split_rules mk minimal nfs wwf_rules
       more)
  = nfs;

minimal :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
minimal
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = minimal;

delete_P_Pwa ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a;
delete_P_Pwa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = delete_P_Pw;

delete_R_Rwc ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] -> [(Term b c, Term b c)] -> a;
delete_R_Rwc
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = delete_R_Rw;

split_pairsa ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] ->
             ([(Term b c, Term b c)], [(Term b c, Term b c)]);
split_pairsa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = split_pairs;

split_rulesc ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      a -> [(Term b c, Term b c)] ->
             ([(Term b c, Term b c)], [(Term b c, Term b c)]);
split_rulesc
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = split_rules;

split_proc ::
  forall a b c.
    (Showa b,
      Showa c) => Dpp_ops_ext a b c () ->
                    a -> [(Term b c, Term b c)] ->
                           [(Term b c, Term b c)] -> (a, a);
split_proc i d p_remove r_remove =
  let {
    (p, pw) = split_pairsa i d p_remove;
    (r, rw) = split_rulesc i d r_remove;
    dpp1 = mkc i (nfsc i d) (minimal i d) p pw (qc i d) r rw;
    a = delete_R_Rwc i (delete_P_Pwa i d p_remove p_remove) r_remove r_remove;
  } in (dpp1, a);

map_subst_domain ::
  forall a b c d. (a -> b) -> (b -> Term c d) -> a -> Term c d;
map_subst_domain m sigma = (\ v -> sigma (m v));

subst_apply_ctxt :: forall a b c. Ctxt a b -> (b -> Term a c) -> Ctxt a c;
subst_apply_ctxt Hole sigma = Hole;
subst_apply_ctxt (More f ss1 d ss2) sigma =
  More f (map (\ t -> subst_apply_term t sigma) ss1) (subst_apply_ctxt d sigma)
    (map (\ t -> subst_apply_term t sigma) ss2);

prefix_list :: Pos -> [Pos];
prefix_list Empty = [];
prefix_list (PCons i p) = Empty : map (PCons i) (prefix_list p);

descaa :: forall a b c. Root_redtriple_ext a b c -> [Char] -> [Char];
descaa (Root_redtriple_ext valid s ns nst af aft desc more) = desc;

valid :: forall a b c. Redtriple_ext a b c -> Sum ([Char] -> [Char]) ();
valid (Redtriple_ext valid s ns nst af mono desc cpx more) = valid;

check_subseteq :: forall a. (Eq a) => [a] -> [a] -> Sum a ();
check_subseteq xs ys =
  catcha (forallm (\ x -> (if membera ys x then Inr () else Inl x)) xs)
    (\ x -> Inl (snd x));

check_wf_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_wf_trs r =
  catcha
    (catcha
      (forallm
        (\ rule ->
          catcha
            (bindb
              (check (not (is_Var (fst rule)))
                (shows_string
                  ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'l', 'e', 'f',
                    't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', ' ',
                    'i', 'n']))
              (\ _ ->
                catcha
                  (check_subseteq (vars_term_lista (snd rule))
                    (vars_term_lista (fst rule)))
                  (\ x ->
                    Inl (shows_string
                           ['f', 'r', 'e', 'e', ' ', 'v', 'a', 'r', 'i', 'a',
                             'b', 'l', 'e', ' '] .
                          shows_prec Zero_nat x .
                            shows_string
                              [' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't', '-',
                                'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e',
                                ' ', 'o', 'f']))))
            (\ x ->
              Inl (x . shows_string [' ', 'r', 'u', 'l', 'e', ' '] .
                         shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                           [' ', '-', '>', ' '] rule .
                           shows_nl)))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (shows_string
             ['t', 'h', 'e', ' ', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'n', 'o',
               't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e',
               'd'] .
            shows_nl . x));

map_rules_wa ::
  forall a b c.
    (a -> Nat -> b) -> [(Term a c, Term a c)] -> [(Term b c, Term b c)];
map_rules_wa fg r = map (\ (l, ra) -> (map_term_wa fg l, map_term_wa fg ra)) r;

generate_f_xs :: forall a. a -> Nat -> Term a [Char];
generate_f_xs f n = Fun f (map (\ i -> Var (generate_var i)) (upt Zero_nat n));

bounds_type :: Xml -> Parse_result [Char] Boundstype;
bounds_type =
  xml1element ['t', 'y', 'p', 'e']
    (xml_options
      [(['r', 'o', 'o', 'f'], xml0elements ['r', 'o', 'o', 'f'] Roof),
        (['m', 'a', 't', 'c', 'h'],
          xml0elements ['m', 'a', 't', 'c', 'h'] Match)])
    id;

bounds_bound :: Xml -> Parse_result [Char] Nat;
bounds_bound = nata ['b', 'o', 'u', 'n', 'd'];

final_states :: Xml -> Parse_result [Char] [[Char]];
final_states =
  xml_many_elements ['f', 'i', 'n', 'a', 'l', 'S', 't', 'a', 't', 'e', 's']
    state id;

transitions ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char] ([Ta_rule [Char] (a, Nat)], [([Char], [Char])]);
transitions xml2name =
  xml_many_elements ['t', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n', 's']
    (transition xml2name)
    (\ rls ->
      let {
        (rules, eps) =
          partition (\ a -> (case a of {
                              Inl _ -> True;
                              Inr _ -> False;
                            }))
            rls;
        ruls = map (\ (Inl r) -> r) rules;
        a = map (\ (Inr e) -> e) eps;
      } in (ruls, a));

bounds_tree_automaton ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Tree_automaton [Char] (a, Nat));
bounds_tree_automaton xml2name =
  xml2elements ['t', 'r', 'e', 'e', 'A', 'u', 't', 'o', 'm', 'a', 't', 'o', 'n']
    final_states (transitions xml2name) (\ f (a, b) -> Tree_Automaton f a b);

bounds_info ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Bounds_info a [Char]);
bounds_info xml2name =
  xml4elements ['b', 'o', 'u', 'n', 'd', 's'] bounds_type bounds_bound
    final_states (bounds_tree_automaton xml2name) Bounds_Info;

unlab_of_trs ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => [(Term (Lab a b) c, Term (Lab a b) c)] ->
                 [(Term (Lab a b) c, Term (Lab a b) c)];
unlab_of_trs r = map (map_rule unlab) (no_decr r);

label_decomp :: forall a b. Lab a b -> (Lab a b, Sum b [Lab a b]);
label_decomp (Lab f l) = (f, Inl l);
label_decomp (FunLab f l) = (f, Inr l);

shows_lab ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> Lab a b -> [Char] -> [Char];
shows_lab fun lab (UnLab f) = fun f;
shows_lab fun lab (Lab f l) =
  shows_lab fun lab f . shows_string ['['] . lab l . shows_string [']'];
shows_lab fun lab (Sharp f) = shows_lab fun lab f . shows_string ['#'];
shows_lab fun lab (FunLab f l) =
  shows_lab fun lab f .
    shows_string ['['] . shows_map (shows_lab fun lab) l . shows_string [']'];

shows_prec_lab ::
  forall a b. (Showa a, Showa b) => Nat -> Lab a b -> [Char] -> [Char];
shows_prec_lab d l = shows_lab (shows_prec Zero_nat) (shows_prec Zero_nat) l;

shows_list_lab ::
  forall a b. (Showa a, Showa b) => [Lab a b] -> [Char] -> [Char];
shows_list_lab fs = shows_list_aux (shows_prec_lab Zero_nat) fs;

instance (Showa a, Showa b) => Showa (Lab a b) where {
  shows_prec = shows_prec_lab;
  shows_list = shows_list_lab;
};

instance (Eq a, Order a, Eq b, Order b) => Preorder (Lab a b) where {
};

instance (Eq a, Order a, Eq b, Order b) => Order (Lab a b) where {
};

max0a :: forall a b. Poly_c_ext a b -> a -> a;
max0a (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
        arc_pos check_mono maxa derive_complexity more)
  = max0;

concat_lists :: forall a. [[a]] -> [[a]];
concat_lists [] = [[]];
concat_lists (asa : xs) =
  let {
    a = concat_lists xs;
  } in concatMap (\ vec -> map (\ aa -> aa : vec) asa) a;

rs_iteratei ::
  forall a b.
    (Linorder a) => Rbt a () -> (b -> Bool) -> (a -> b -> b) -> b -> b;
rs_iteratei = s_iteratei rm_iteratei;

set_ext :: forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
set_ext s_ns =
  (\ asa bs ->
    (not (null asa) && all (\ b -> any (\ a -> fst (s_ns a b)) asa) bs,
      all (\ b -> any (\ a -> snd (s_ns a b)) asa) bs));

check_no_var ::
  forall a b. (Showa a, Showa b) => Term a b -> Sum ([Char] -> [Char]) ();
check_no_var t =
  check (not (is_Var t))
    (shows_string
       ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'f', 'o', 'u', 'n', 'd'] .
      shows_nl);

check_supteq ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_supteq s t =
  check (equal_term s t || supt_impl s t)
    (shows_prec_term Zero_nat t .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 't',
          'e', 'r', 'm', ' ', 'o', 'f', ' '] .
        shows_prec_term Zero_nat s);

valida :: forall a b c. Root_redtriple_ext a b c -> Sum ([Char] -> [Char]) ();
valida (Root_redtriple_ext valid s ns nst af aft desc more) = valid;

first_rewrite ::
  forall a b.
    (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Maybe (Term a b);
first_rewrite r s = (case rewrite r s of {
                      [] -> Nothing;
                      t : _ -> Just t;
                    });

compute_NF :: forall a. (a -> Maybe a) -> a -> Maybe a;
compute_NF f a =
  (case f a of {
    Nothing -> Just a;
    Just aa -> compute_NF f aa;
  });

compute_rstep_NF ::
  forall a b.
    (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Maybe (Term a b);
compute_rstep_NF r s = compute_NF (first_rewrite r) s;

check_join_NF ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_join_NF r s t =
  (case (compute_rstep_NF r s, compute_rstep_NF r t) of {
    (Nothing, b) ->
      Inl (shows_prec_list Zero_nat
            ['s', 't', 'r', 'a', 'n', 'g', 'e', ' ', 'e', 'r', 'r', 'o', 'r',
              ' ', 'i', 'n', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o',
              'r', 'm', ' ', 'c', 'o', 'm', 'p', 'u', 't', 'a', 't', 'i', 'o',
              'n']);
    (Just sa, Nothing) ->
      Inl (shows_prec_list Zero_nat
            ['s', 't', 'r', 'a', 'n', 'g', 'e', ' ', 'e', 'r', 'r', 'o', 'r',
              ' ', 'i', 'n', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o',
              'r', 'm', ' ', 'c', 'o', 'm', 'p', 'u', 't', 'a', 't', 'i', 'o',
              'n']);
    (Just sa, Just ta) ->
      check (equal_term sa ta)
        (shows_string
           ['t', 'h', 'e', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o',
             'r', 'm', ' '] .
          shows_prec_term Zero_nat sa .
            shows_prec_list Zero_nat [' ', 'o', 'f', ' '] .
              shows_prec_term Zero_nat s .
                shows_prec_list Zero_nat
                  [' ', 'd', 'i', 'f', 'f', 'e', 'r', 's', ' ', 'f', 'r', 'o',
                    'm', ' '] .
                  shows_nl .
                    shows_prec_list Zero_nat
                      ['t', 'h', 'e', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ',
                        'f', 'o', 'r', 'm', ' '] .
                      shows_prec_term Zero_nat ta .
                        shows_prec_list Zero_nat [' ', 'o', 'f', ' '] .
                          shows_prec_term Zero_nat t);
  });

reverse_rules :: forall a b. [(Term a b, Term a b)] -> [(Term a b, Term a b)];
reverse_rules rs = map (\ lr -> (snd lr, fst lr)) rs;

vars_rule_lista :: forall a b. (Term a b, Term a b) -> [b];
vars_rule_lista r = vars_term_list (fst r) ++ vars_term_list (snd r);

vars_trs_list :: forall a b. (Eq b) => [(Term a b, Term a b)] -> [b];
vars_trs_list trs = remdups (concatMap vars_rule_lista trs);

check_pairwise :: forall a b. (a -> a -> Sum b ()) -> [a] -> Sum b ();
check_pairwise c [] = Inr ();
check_pairwise c (x : xs) =
  bindb (catcha (forallm (c x) xs) (\ xa -> Inl (snd xa)))
    (\ _ -> check_pairwise c xs);

check_disjoint :: forall a. (Eq a) => [a] -> [a] -> Sum a ();
check_disjoint xs ys =
  catcha (forallm (\ x -> (if not (membera ys x) then Inr () else Inl x)) xs)
    (\ x -> Inl (snd x));

check_partition :: forall a. (Eq a) => [[a]] -> Sum a ();
check_partition xss = check_pairwise check_disjoint xss;

check_inj ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Lab a b ->
                    Nat ->
                      [((Lab a b, Nat), [Lab a b])] ->
                        Sum ([Char] -> [Char]) ();
check_inj a nn sml =
  let {
    symbols =
      map (\ (aa, b) ->
            let {
              (_, n) = aa;
            } in (\ fs ->
                   map (\ (g, i) ->
                         (g, plus_nat n (times_nat i (minus_nat nn one_nat))))
                     (zip fs (upt Zero_nat (size_list fs))))
              b)
        sml;
    fsymbols = concat symbols;
  } in bindb (catcha (check_partition symbols)
               (\ x ->
                 Inl (shows_prec_list Zero_nat
                        ['s', 'y', 'm', 'b', 'o', 'l', ' '] .
                       shows_prec_prod Zero_nat x .
                         shows_prec_list Zero_nat
                           [' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w',
                             'i', 'c', 'e', ')'])))
         (\ _ ->
           bindb (check (not (membera fsymbols (a, nn)))
                   (shows_prec_list Zero_nat
                      ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n',
                        ' ', 's', 'y', 'm', 'b', 'o', 'l'] .
                     shows_prec_lab Zero_nat a .
                       shows_prec_list Zero_nat
                         [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'o',
                           'c', 'c', 'u', 'r', ' ', 'a', 's', ' ', 'n', 'e',
                           'w', ' ', 's', 'y', 'm', 'b', 'o', 'l']))
             (\ _ ->
               bindb (check (not (membera (map fst sml) (a, nn)))
                       (shows_prec_list Zero_nat
                          ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o',
                            'n', ' ', 's', 'y', 'm', 'b', 'o', 'l'] .
                         shows_prec_lab Zero_nat a .
                           shows_prec_list Zero_nat
                             [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ',
                               'b', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'i',
                               'e', 'd']))
                 (\ _ ->
                   (if less_eq_nat nn one_nat
                     then catcha
                            (forallm
                              (check_pairwise
                                (\ gn1 gn2 ->
                                  check (not (gn1 == gn2))
                                    (shows_prec_list Zero_nat
                                       ['s', 'y', 'm', 'b', 'o', 'l', ' '] .
                                      shows_prec_prod Zero_nat gn1 .
shows_prec_list Zero_nat
  [' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w', 'i', 'c', 'e'])))
                              symbols)
                            (\ x -> Inl (snd x))
                     else Inr ()))));

uncurry_info ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          (a, ([((a, Nat), [a])],
                ([(Term a [Char], Term a [Char])],
                  [(Term a [Char], Term a [Char])])));
uncurry_info xml2name =
  xml4elements
    ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'I', 'n', 'f', 'o', 'r', 'm', 'a', 't',
      'i', 'o', 'n']
    xml2name
    (xml_many_elements
      ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm', 'b', 'o',
        'l', 's']
      (xml2many_elements
        ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm', 'b', 'o',
          'l', 'E', 'n', 't', 'r', 'y']
        xml2name (nata ['a', 'r', 'i', 't', 'y']) xml2name
        (\ f n -> (\ a -> ((f, n), a))))
      id)
    (xml1element ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'R', 'u', 'l', 'e', 's']
      (rules xml2name) id)
    (xml1element ['e', 't', 'a', 'R', 'u', 'l', 'e', 's'] (rules xml2name) id)
    (\ a sml u e -> (a, (sml, (u, e))));

flat_contexts ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [Ctxt a [Char]];
flat_contexts xml2name =
  xml_many_elements ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 's']
    (ctxt xml2name) id;

projected_rseq ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] a) ->
                ((a, Nat) -> Nat) ->
                  Xml ->
                    Parse_result [Char]
                      ((Term a [Char], Term a [Char]),
                        [(Pos, ((Term a [Char], Term a [Char]),
                                 Term a [Char]))]);
projected_rseq xml2name pia (Xml name atts cs text) =
  let {
    tag = ['p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', 'R', 'e', 'w', 'r', 'i',
            't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e'];
  } in (if name == tag &&
             null atts && is_none text && equal_nat (size_list cs) (nat 2)
         then bindc (rule xml2name (nth cs Zero_nat))
                (\ r -> rseq xml2name pia r (nth cs one_nat))
         else no_transform tag (Xml name atts cs text));

xml1many2elements_gen ::
  forall a b c d e.
    [Char] ->
      (Xml -> Parse_result [Char] a) ->
        (a -> Xml -> Parse_result [Char] b) ->
          (Xml -> Parse_result [Char] c) ->
            (Xml -> Parse_result [Char] d) ->
              (a -> [b] -> c -> d -> e) -> Xml -> Parse_result [Char] e;
xml1many2elements_gen tag p1 p2 p3 p4 f (Xml name atts cs text) =
  let {
    ds = reverse cs;
    l = size_list cs;
  } in (if name == tag && null atts && is_none text && less_eq_nat (nat 3) l
         then bindc (p1 (nth cs Zero_nat))
                (\ x ->
                  bindc (map_parse_result (p2 x)
                          (tl (takea (minus_nat l (nat 2)) cs)))
                    (\ xs ->
                      bindc (p3 (nth ds one_nat))
                        (\ y ->
                          bindc (p4 (nth ds Zero_nat))
                            (\ z -> returna (f x xs y z)))))
         else no_transform tag (Xml name atts cs text));

strategy_to_Q ::
  forall a b.
    (Eq a, Eq b) => Strategy a b -> [(Term a b, Term a b)] -> [Term a b];
strategy_to_Q No_Strategy uu = [];
strategy_to_Q Innermost r = remdups (map fst r);
strategy_to_Q (Innermost_Q q) uv = q;

create_proj :: forall a. (Eq a) => ProjL a -> (a, Nat) -> Nat;
create_proj (Projection p) f =
  (case map_of p f of {
    Nothing -> Zero_nat;
    Just n -> n;
  });

ordering_constraint_proofa ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Redtriple_impl a);
ordering_constraint_proofa xml2name =
  xml1element
    ['o', 'r', 'd', 'e', 'r', 'i', 'n', 'g', 'C', 'o', 'n', 's', 't', 'r', 'a',
      'i', 'n', 't', 'P', 'r', 'o', 'o', 'f']
    (redtriple xml2name) id;

level_mapping ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] [((a, Nat), [(Nat, Nat)])];
level_mapping xml2name =
  xml_many_elements ['l', 'e', 'v', 'e', 'l', 'M', 'a', 'p', 'p', 'i', 'n', 'g']
    (xml2many_elements
      ['l', 'e', 'v', 'e', 'l', 'M', 'a', 'p', 'p', 'i', 'n', 'g', 'E', 'n',
        't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (xml2elements
        ['p', 'o', 's', 'i', 't', 'i', 'o', 'n', 'L', 'e', 'v', 'e', 'l', 'E',
          'n', 't', 'r', 'y']
        (nata ['p', 'o', 's', 'i', 't', 'i', 'o', 'n'])
        (nata ['l', 'e', 'v', 'e', 'l']) (\ a b -> (a, b)))
      (\ f a ps ->
        ((f, a),
          map (\ (p, b) ->
                ((if equal_nat p Zero_nat then a else minus_nat p one_nat), b))
            ps)))
    id;

root_redtriple ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Root_redtriple_impl a);
root_redtriple xml2name =
  xml1element ['r', 'e', 'd', 'P', 'a', 'i', 'r']
    (xml3elements ['s', 'c', 'n', 'p']
      (xml1element ['s', 't', 'a', 't', 'u', 's']
        (xml_options
          [(['m', 's'], xml0elements ['m', 's'] MS_Ext),
            (['m', 'i', 'n'], xml0elements ['m', 'i', 'n'] Min_Ext),
            (['d', 'm', 's'], xml0elements ['d', 'm', 's'] Dms_Ext),
            (['m', 'a', 'x'], xml0elements ['m', 'a', 'x'] Max_Ext)])
        id)
      (level_mapping xml2name) (redtriple xml2name) Scnp)
    id;

root_ordering_constraint_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Root_redtriple_impl a);
root_ordering_constraint_proof xml2name =
  xml1element
    ['o', 'r', 'd', 'e', 'r', 'i', 'n', 'g', 'C', 'o', 'n', 's', 't', 'r', 'a',
      'i', 'n', 't', 'P', 'r', 'o', 'o', 'f']
    (root_redtriple xml2name) id;

ordering_constraint_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char] (Sum (Root_redtriple_impl a) (Redtriple_impl a));
ordering_constraint_proof xml2name =
  xml_choice
    ['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r']
    [xml_change (ordering_constraint_proofa xml2name) Inr,
      xml_change (root_ordering_constraint_proof xml2name) Inl];

xml2dp_proof ::
  forall a b.
    (Eq a,
      Eq b) => (Xml -> Parse_result [Char] (Lab a b)) ->
                 Xml -> Parse_result [Char] (Dp_proof a b [Char]);
xml2dp_proof xml2name x =
  xml1element ['d', 'p', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['p', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['p', 'I', 's', 'E', 'm', 'p', 't', 'y'] PisEmpty),
        (['d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c'],
          xml_many_elements
            ['d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c']
            (xml2nd_choice ['c', 'o', 'm', 'p', 'o', 'n', 'e', 'n', 't']
              (xml1element ['d', 'p', 's'] (rules xml2name) id)
              ['r', 'e', 'a', 'l', 'S', 'c', 'c'] (xml2dp_proof xml2name)
              (\ dps prfOpt -> (prfOpt, dps)))
            DepGraphProc),
        (['r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'],
          xml3elements ['r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) RedPairProc),
        (['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's', 'P', 'r', 'o',
           'c'],
          xml2elements
            ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's', 'P', 'r',
              'o', 'c']
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) UsableRulesProc),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's', 'R',
           'e', 'm', 'o', 'v', 'a', 'l', 'P', 'r', 'o', 'c'],
          xml2elements
            ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's',
              'R', 'e', 'm', 'o', 'v', 'a', 'l', 'P', 'r', 'o', 'c']
            (innermostLhss xml2name) (xml2dp_proof xml2name) QReductionProc),
        (['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o', 'c'],
          xml_choice
            ['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o', 'c']
            [xml4elements
               ['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o', 'c']
               (rule xml2name) (rstep xml2name)
               (xml1element
                 ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
                 (rules xml2name) id)
               (xml2dp_proof xml2name)
               (\ (s, t) (p, (lr, ta)) u ->
                 RewritingProc u (s, t) (s, ta) (s, ta) lr p),
              xml5elements
                ['r', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o',
                  'c']
                (rule xml2name) (rstep xml2name) (rule xml2name)
                (xml1element
                  ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
                  (rules xml2name) id)
                (xml2dp_proof xml2name)
                (\ (s, t) (p, (lr, ta)) st u ->
                  RewritingProc u (s, t) (s, ta) st lr p)]),
        (['n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 'P', 'r', 'o', 'c'],
          xml4elements
            ['n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 'P', 'r', 'o', 'c']
            (rule xml2name) pos
            (xml1element ['n', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) NarrowingProc),
        (['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 'P',
           'r', 'o', 'c'],
          xml3elements
            ['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n',
              'P', 'r', 'o', 'c']
            (rule xml2name)
            (xml1element
              ['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n',
                's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) InstantiationProc),
        (['f', 'o', 'r', 'w', 'a', 'r', 'd', 'I', 'n', 's', 't', 'a', 'n', 't',
           'i', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c'],
          xml3to4elements
            ['f', 'o', 'r', 'w', 'a', 'r', 'd', 'I', 'n', 's', 't', 'a', 'n',
              't', 'i', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c']
            (rule xml2name)
            (xml1element
              ['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n',
                's']
              (rules xml2name) id)
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) ForwardInstantiationProc),
        (['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'],
          xml_choice ['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
            [xml4elements ['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
               (sl_variant xml2name)
               (xml1element ['d', 'p', 's'] (rules xml2name) id)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml2dp_proof xml2name) (\ sli lp -> SemlabProc sli lp []),
              xml5elements ['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
                (sl_variant xml2name)
                (xml1element ['d', 'p', 's'] (rules xml2name) id)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (innermostLhss xml2name) (xml2dp_proof xml2name)
                (\ sli lp lr lq -> SemlabProc sli lp lq lr)]),
        (['s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c'],
          xml1many2elements_gen
            ['s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c']
            (proj xml2name) (\ pia -> projected_rseq xml2name (create_proj pia))
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) SubtermCriterionProc),
        (['r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'],
          xml4elements
            ['r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) RedPairUrProc),
        (['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P',
           'r', 'o', 'c'],
          xml5elements
            ['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r',
              'P', 'r', 'o', 'c']
            (ordering_constraint_proofa xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dp_proof xml2name) MonoRedPairUrProc),
        (['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
           'c'],
          xml4elements
            ['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r',
              'o', 'c']
            (ordering_constraint_proofa xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) MonoRedPairProc),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c'],
          xml_choice ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
            [xml4elements
               ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
               (uncurry_info xml2name)
               (xml1element ['d', 'p', 's'] (rules xml2name) id)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml2dp_proof xml2name) (UncurryProc Nothing),
              xml5elements
                ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
                (nata ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'v', 'e',
                        'T', 'o', 'p'])
                (uncurry_info xml2name)
                (xml1element ['d', 'p', 's'] (rules xml2name) id)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (xml2dp_proof xml2name) (\ n -> UncurryProc (Just n))]),
        (['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
           's', 'u', 'r', 'e', 'P', 'r', 'o', 'c'],
          xml5elements
            ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l',
              'o', 's', 'u', 'r', 'e', 'P', 'r', 'o', 'c']
            (xml1element ['f', 'r', 'e', 's', 'h', 'S', 'y', 'm', 'b', 'o', 'l']
              xml2name id)
            (flat_contexts xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) FccProc),
        (['s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's',
           't', 'P', 'r', 'o', 'c'],
          xml2elements
            ['s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o',
              's', 't', 'P', 'r', 'o', 'c']
            (wcr_proof xml2name) (xml2dp_proof xml2name) SwitchInnermostProc),
        (['s', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c'],
          xml4elements ['s', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) (xml2dp_proof xml2name) SplitProc),
        (['f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', 'A', 's', 's', 'u',
           'm', 'p', 't', 'i', 'o', 'n'],
          xml1element
            ['f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', 'A', 's', 's',
              'u', 'm', 'p', 't', 'i', 'o', 'n']
            (xml2dp_input xml2name)
            (\ (m, (p, (q, r))) ->
              AssumeFinite
                (False, (m, (p, ([], (strategy_to_Q q r, ([], r)))))))),
        (['u', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c'],
          xml3elements ['u', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dp_proof xml2name) UnlabProc),
        (['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'],
          xml1many_elements
            ['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o',
              'c']
            (xml_options
              [(['s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i', 't', 'e',
                  'r', 'i', 'o', 'n'],
                 xml0elements
                   ['s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i', 't', 'e',
                     'r', 'i', 'o', 'n']
                   Nothing),
                (['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P', 'a', 'i',
                   'r'],
                  xml1to2elements
                    ['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P', 'a', 'i',
                      'r']
                    (ordering_constraint_proofa xml2name)
                    (xml1element
                      ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
                      (rules xml2name) id)
                    (\ redp ur -> Just (redp, ur)))])
            (scg xml2name)
            (\ a ->
              (case a of {
                Nothing -> SizeChangeSubterm;
                Just redp_ur -> SizeChangeRedPair (fst redp_ur) (snd redp_ur);
              })))])
    id x;

xml2trs_proof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Trs_proof a [Nat] [Char]);
xml2trs_proof xml2name x =
  xml1element
    ['t', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P',
      'r', 'o', 'o', 'f']
    (xml_options
      [(['r', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['r', 'I', 's', 'E', 'm', 'p', 't', 'y'] RisEmpty),
        (['s', 'e', 'm', 'l', 'a', 'b'],
          xml_choice ['s', 'e', 'm', 'l', 'a', 'b']
            [xml3elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_variant xml2name)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml2trs_proof xml2name) (\ sli -> Semlab sli []),
              xml4elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_variant xml2name)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (innermostLhss xml2name) (xml2trs_proof xml2name)
                (\ sli lr lq -> Semlab sli lq lr)]),
        (['s', 'p', 'l', 'i', 't'],
          xml3elements ['s', 'p', 'l', 'i', 't']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_proof xml2name) (xml2trs_proof xml2name) Split),
        (['d', 'p', 'T', 'r', 'a', 'n', 's'],
          xml3elements ['d', 'p', 'T', 'r', 'a', 'n', 's']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (bool ['m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                    's'])
            (xml2dp_proof xml2name) (\ dps _ -> DPTrans dps)),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml3elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (ordering_constraint_proofa xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_proof xml2name) RuleRemoval),
        (['b', 'o', 'u', 'n', 'd', 's'],
          (\ xml -> bindc (bounds_info xml2name xml) (returna . Bounds))),
        (['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'],
          xml2elements
            ['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a',
              'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_proof xml2name) (\ _ -> StringReversal)),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y'],
          xml3elements ['u', 'n', 'c', 'u', 'r', 'r', 'y']
            (uncurry_info xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_proof xml2name) Uncurry),
        (['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
           's', 'u', 'r', 'e'],
          xml3elements
            ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l',
              'o', 's', 'u', 'r', 'e']
            (flat_contexts xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_proof xml2name) Fcc),
        (['s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's',
           't'],
          xml2elements
            ['s', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o',
              's', 't']
            (wcr_proof xml2name) (xml2trs_proof xml2name) SwitchInnermost),
        (['t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'A', 's', 's',
           'u', 'm', 'p', 't', 'i', 'o', 'n'],
          xml1element
            ['t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'A', 's',
              's', 'u', 'm', 'p', 't', 'i', 'o', 'n']
            (xml2trs_input xml2name)
            (\ (q, (r, so)) ->
              AssumeSN
                (False,
                  (strategy_to_Q q r,
                    (r, (case so of {
                          Nothing -> [];
                          Just s -> s;
                        }))))))])
    id x;

xml2cr_proof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Cr_proof a [Nat] [Char]);
xml2cr_proof xml2name x =
  xml1element ['c', 'r', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['w', 'c', 'r', 'A', 'n', 'd', 'S', 'N'],
         xml2elements ['w', 'c', 'r', 'A', 'n', 'd', 'S', 'N']
           (wcr_proof xml2name) (xml2trs_proof xml2name) SN_WCR),
        (['o', 'r', 't', 'h', 'o', 'g', 'o', 'n', 'a', 'l'],
          xml0elements ['o', 'r', 't', 'h', 'o', 'g', 'o', 'n', 'a', 'l']
            Weakly_Orthogonal)])
    id x;

xml2eq_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Eq_proof a [Char]);
xml2eq_proof xml2name x =
  xml_options
    [(['r', 'e', 'f', 'l'],
       xml1element ['r', 'e', 'f', 'l'] (term xml2name) Refl),
      (['s', 'y', 'm'],
        xml1element ['s', 'y', 'm'] (xml2eq_proof xml2name) Sym),
      (['t', 'r', 'a', 'n', 's'],
        xml2elements ['t', 'r', 'a', 'n', 's'] (xml2eq_proof xml2name)
          (xml2eq_proof xml2name) Trans),
      (['a', 's', 's', 'm'],
        xml2elements ['a', 's', 's', 'm'] (rule xml2name) (subst xml2name)
          (\ r s -> Assm r (mk_subst s))),
      (['c', 'o', 'n', 'g'],
        xml1many_elements ['c', 'o', 'n', 'g'] xml2name (xml2eq_proof xml2name)
          Cong)]
    x;

rulesd :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
rulesd
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = rules;

is_QNFc :: forall a b c d. Dpp_ops_ext a b c d -> a -> Term b c -> Bool;
is_QNFc
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = is_QNF;

nFQ_subset_NF_rulesb :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
nFQ_subset_NF_rulesb
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = nFQ_subset_NF_rules;

icap_impl_dpp ::
  forall a b.
    (Eq b) => Dpp_ops_ext a b [Char] () ->
                a -> [Term b [Char]] -> Term b [Char] -> Term b (Sum () [Char]);
icap_impl_dpp i d =
  let {
    qr = nFQ_subset_NF_rulesb i d;
    qnf = is_QNFc i d;
    r = rulesd i d;
    ic = icap_impl qr qnf (map fst r);
  } in (\ s ->
         let {
           sx = remdups (concatMap vars_term_lista s);
         } in ic s (membera sx));

shows_kbo_repr ::
  forall a. (Showa a) => ([((a, Nat), (Nat, Nat))], Nat) -> [Char] -> [Char];
shows_kbo_repr (prs, w0) =
  shows_string
    ['K', 'B', 'O', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
      'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e',
      'n', 'c', 'e', ' ', 'a', 'n', 'd', ' ', 'w', 'e', 'i', 'g', 'h', 't', ' ',
      'f', 'u', 'n', 'c', 't', 'i', 'o', 'n'] .
    shows_nl .
      shows_map
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ (pr, _) ->
                 shows_string
                   ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '('] .
                   shows_prec Zero_nat f .
                     shows_string ['['] .
                       shows_prec_nat Zero_nat n .
                         shows_string [']', ')', ' ', '=', ' '] .
                           shows_prec_nat Zero_nat pr . shows_nl)
            b)
        prs .
        shows_nl .
          shows_string
            ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '(', '_', ')',
              ' ', '=', ' ', '0'] .
            shows_nl .
              shows_string
                ['a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                  'o', 'w', 'i', 'n', 'g', ' ', 'w', 'e', 'i', 'g', 'h', 't'] .
                shows_nl .
                  shows_map
                    (\ (a, b) ->
                      let {
                        (f, n) = a;
                      } in (\ (_, w) ->
                             shows_string ['w', 'e', 'i', 'g', 'h', 't', '('] .
                               shows_prec Zero_nat f .
                                 shows_string ['['] .
                                   shows_prec_nat Zero_nat n .
                                     shows_string [']', ')', ' ', '=', ' '] .
                                       shows_prec_nat Zero_nat w . shows_nl)
                        b)
                    prs .
                    shows_nl .
                      shows_string
                        ['w', 'e', 'i', 'g', 'h', 't', '(', '_', ')', ' ', '=',
                          ' '] .
                        shows_prec_nat Zero_nat (Suc w0) .
                          shows_nl .
                            shows_prec_list Zero_nat ['w', '0', ' ', '=', ' '] .
                              shows_prec_nat Zero_nat w0 . shows_nl;

instance (Eq a, Order a, Eq b, Order b) => Preorder (Term a b) where {
};

instance (Eq a, Order a, Eq b, Order b) => Order (Term a b) where {
};

data L_poly a b = LPoly b [(a, b)];

sym_collect :: forall a b. (Term a b -> Bool) -> Term a b -> [a];
sym_collect p (Var x) = [];
sym_collect p (Fun f ts) =
  (if p (Fun f ts) then [f] else []) ++ concatMap (sym_collect p) ts;

fresh_strings_list :: [Char] -> Nat -> [[Char]] -> Nat -> [[Char]];
fresh_strings_list name offset used n =
  takea n
    (filter (\ s -> not (membera (remdups used) s))
      (map (\ i -> name ++ shows_prec_nat Zero_nat (plus_nat i offset) [])
        (upt Zero_nat (plus_nat n (size_list (remdups used))))));

mk_rtrancl_list_main ::
  forall a. (a -> a -> Bool) -> (a -> [a]) -> [a] -> [a] -> [a];
mk_rtrancl_list_main subsumes r todo fin =
  (case todo of {
    [] -> fin;
    a : tod ->
      (if any (\ b -> subsumes b a) fin
        then mk_rtrancl_list_main subsumes r tod fin
        else mk_rtrancl_list_main subsumes r (r a ++ tod) (a : fin));
  });

mk_rtrancl_list :: forall a. (a -> a -> Bool) -> (a -> [a]) -> [a] -> [a];
mk_rtrancl_list subsumes r inita = mk_rtrancl_list_main subsumes r inita [];

rat_gt :: Rat -> Rat -> Rat -> Bool;
rat_gt delta = (\ x y -> less_eq_rat delta (minus_rat x y));

subst_compose_impl ::
  forall a b.
    (Eq a, Eq b) => [(a, Term b a)] -> [(a, Term b a)] -> [(a, Term b a)];
subst_compose_impl sigma tau =
  let {
    sigmaa = mk_subst_domain sigma;
    taua = mk_subst_domain tau;
    d_sigma = map fst sigmaa;
  } in map (\ (x, t) -> (x, subst_apply_term t (mk_subst taua))) sigmaa ++
         filter (\ (x, _) -> not (membera d_sigma x)) taua;

commutes_impl ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> [(a, Term b a)] -> Bool;
commutes_impl sigma mu =
  subst_eq (subst_compose_impl sigma mu) (subst_compose_impl mu sigma);

mk_subst_case ::
  forall a b.
    (Eq a,
      Eq b) => [a] -> (a -> Term b a) -> [(a, Term b a)] -> [(a, Term b a)];
mk_subst_case xs sigma tau =
  subst_compose_impl (map (\ x -> (x, sigma x)) xs) tau;

funas_term_list :: forall a b. Term a b -> [(a, Nat)];
funas_term_list (Var uu) = [];
funas_term_list (Fun f ts) = (f, size_list ts) : concatMap funas_term_list ts;

funas_rule_list :: forall a b. (Term a b, Term a b) -> [(a, Nat)];
funas_rule_list r = funas_term_list (fst r) ++ funas_term_list (snd r);

funas_trs_list :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_trs_list r = remdups (concatMap funas_rule_list r);

add_funs_term_list :: forall a b. Term a b -> [a] -> [a];
add_funs_term_list =
  folda (consta id) (\ f xs -> (\ a -> f : a) . foldra (\ a b -> a . b) xs id);

funs_term_list :: forall a b. Term a b -> [a];
funs_term_list t = add_funs_term_list t [];

funs_rule_lista :: forall a b. (Term a b, Term a b) -> [a];
funs_rule_lista r = funs_term_list (fst r) ++ funs_term_list (snd r);

funs_rule_list :: forall a b. (Eq a) => (Term a b, Term a b) -> [a];
funs_rule_list r = remdups (funs_rule_lista r);

shows_prec_pos :: Nat -> Pos -> [Char] -> [Char];
shows_prec_pos d p = shows_pos p;

vars_rule_list :: forall a b. (Eq b) => (Term a b, Term a b) -> [b];
vars_rule_list r = remdups (vars_rule_lista r);

ra :: forall a b c d. Tp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
ra (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
     delete_R_Rw split_rules mk nfs more)
  = r;

check_CS_subseteq ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 [(Term a b, Term a b)] -> Sum (Term a b, Term a b) ();
check_CS_subseteq r s =
  catcha
    (forallm (\ (l, ra) -> check (any (instance_rule (l, ra)) s) (l, ra)) r)
    (\ x -> Inl (snd x));

uncurry_rules ::
  forall a b.
    (Eq a) => a -> (a -> Nat -> [a]) ->
                     [(Term a b, Term a b)] -> [(Term a b, Term a b)];
uncurry_rules a sm =
  map (\ (l, r) -> (uncurry_term a sm l, uncurry_term a sm r));

rwa :: forall a b c d. Tp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
rwa (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules
      rules_map delete_R_Rw split_rules mk nfs more)
  = rw;

eta_closed_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> (a -> Nat -> [a]) ->
                         [(Term a b, Term a b)] ->
                           [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
eta_closed_rules a sm ra r =
  catcha
    (forallm
      (\ (l, raa) ->
        (case aarity_term a sm l of {
          Nothing -> Inr ();
          Just Zero_nat -> Inr ();
          Just (Suc _) ->
            check (any (\ (lll, rrr) ->
                         (case (lll, rrr) of {
                           (Var _, b) -> False;
                           (Fun f [], b) -> False;
                           (Fun f [ll], b) -> False;
                           (Fun f [ll, Var x], Var _) -> False;
                           (Fun f [ll, Var x], Fun g []) -> False;
                           (Fun f [ll, Var x], Fun g [rr]) -> False;
                           (Fun f [ll, Var x], Fun g [rr, Var y]) ->
                             f == a &&
                               g == a &&
                                 x == y &&
                                   not (membera (vars_rule_list (ll, rr)) x) &&
                                     instance_rule (l, raa) (ll, rr);
                           (Fun f [ll, Var x], Fun g (rr : Var y : _ : _)) ->
                             False;
                           (Fun f [ll, Var x], Fun g (rr : Fun _ _ : list)) ->
                             False;
                           (Fun f (ll : Var x : _ : _), b) -> False;
                           (Fun f (ll : Fun _ _ : list), b) -> False;
                         }))
                    r)
              (shows_prec_list Zero_nat
                 ['e', 't', 'a', ' ', 'e', 'x', 'p', 'a', 'n', 's', 'i', 'o',
                   'n', ' ', 'o', 'f', ' '] .
                shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                  [' ', '-', '>', ' '] (l, raa) .
                  shows_prec_list Zero_nat
                    [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);
        }))
      ra)
    (\ x -> Inl (snd x));

uncurry_eta_split ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
uncurry_eta_split eboth rtest =
  let {
    test =
      (\ (l, r) ->
        less_nat Zero_nat (num_args l) &&
          less_nat Zero_nat (num_args r) &&
            any (eq_rule_mod_vars (hd (args l), hd (args r))) rtest);
  } in partition test eboth;

sig_list_to_sig_map ::
  forall a.
    (Eq a) => a -> [((a, Nat), [a])] ->
                     ([((a, Nat), [a])] -> a -> Nat -> a) -> a -> Nat -> [a];
sig_list_to_sig_map a sml fmapa =
  let {
    fm = fmapa sml;
  } in (\ f n ->
         (case map_of sml (f, n) of {
           Nothing -> [fm f n];
           Just xs -> (if null xs then [fm f n] else xs);
         }));

uncurry_of_sig_list ::
  forall a.
    a -> [((a, Nat), [a])] ->
           (a -> Nat -> [a]) -> [(Term a [Char], Term a [Char])];
uncurry_of_sig_list a sml sm =
  concatMap
    (\ (b, c) ->
      let {
        (f, n) = b;
      } in (\ _ ->
             let {
               g = get_symbol sm f n;
             } in map (\ i ->
                        (Fun a [generate_f_xs (g i) (plus_nat n i),
                                 Var (generate_var (plus_nat n i))],
                          generate_f_xs (g (Suc i)) (plus_nat n (Suc i))))
                    (upt Zero_nat (aarity sm f n)))
        c)
    sml;

uncurry_tt ::
  forall a b.
    (Eq b,
      Showa b) => Tp_ops_ext a b [Char] () ->
                    (b, ([((b, Nat), [b])],
                          ([(Term b [Char], Term b [Char])],
                            [(Term b [Char], Term b [Char])]))) ->
                      [(Term b [Char], Term b [Char])] ->
                        a -> Sum ([Char] -> [Char]) a;
uncurry_tt i info r tp =
  let {
    (a, (sml, (u, eb))) = info;
    raa = ra i tp;
    rw = rwa i tp;
    (e, ew) = uncurry_eta_split eb raa;
    r_eta = e ++ raa;
    rw_eta = ew ++ rw;
    rb_eta = r_eta ++ rw_eta;
    fmapa = (\ _ f _ -> f);
    sm = sig_list_to_sig_map a sml fmapa;
    uR = uncurry_rules a sm r_eta;
    uRw = uncurry_rules a sm rw_eta;
  } in (case let {
               s = uncurry_of_sig_list a sml sm;
             } in bindb (catcha
                          (forallm
                            (\ (l, _) ->
                              check (hvf_term a l)
                                (shows_prec_list Zero_nat
                                   ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i',
                                     'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ',
                                     'l', 'h', 's', ' '] .
                                  shows_term (shows_prec Zero_nat)
                                    (shows_prec_list Zero_nat) l .
                                    shows_prec_list Zero_nat
                                      [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l',
'o', 'w', 'e', 'd']))
                            r_eta)
                          (\ x -> Inl (snd x)))
                    (\ _ ->
                      bindb (catcha
                              (forallm
                                (\ (l, _) ->
                                  check (hvf_term a l)
                                    (shows_prec_list Zero_nat
                                       ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r',
 'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                                      shows_term (shows_prec Zero_nat)
(shows_prec_list Zero_nat) l .
shows_prec_list Zero_nat
  [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']))
                                rw_eta)
                              (\ x -> Inl (snd x)))
                        (\ _ ->
                          bindb (eta_closed_rules a sm r_eta r_eta)
                            (\ _ ->
                              bindb (eta_closed_rules a sm rb_eta rb_eta)
                                (\ _ ->
                                  bindb (catcha
  (forallm
    (\ (l, rb) ->
      check (not (is_Var l))
        (shows_prec_list Zero_nat
           ['l', 'h', 's', ' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ',
             'b', 'e', ' ', 'a', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
             ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] (l, rb)))
    rw_eta)
  (\ x -> Inl (snd x)))
                                    (\ _ ->
                                      bindb
(catcha (check_subseteq uR r)
  (\ x ->
    Inl (shows_prec_list Zero_nat
           ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l',
             'e', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] x .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
(\ _ ->
  bindb (catcha (check_subseteq uRw r)
          (\ x ->
            Inl (shows_prec_list Zero_nat
                   ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u',
                     'l', 'e', ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                    [' ', '-', '>', ' '] x .
                    shows_prec_list Zero_nat
                      [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
    (\ _ ->
      bindb (catcha (check_CS_subseteq s u)
              (\ x ->
                Inl (shows_prec_list Zero_nat
                       ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l',
                         'e', ' '] .
                      shows_rule (shows_prec Zero_nat)
                        (shows_prec_list Zero_nat) [' ', '-', '>', ' '] x .
                        shows_prec_list Zero_nat
                          [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                            'g'])))
        (\ _ ->
          catcha (check_subseteq u r)
            (\ x ->
              Inl (shows_prec_list Zero_nat
                     ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l',
                       'e', ' '] .
                    shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                      [' ', '-', '>', ' '] x .
                      shows_prec_list Zero_nat
                        [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g',
                          ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'T', 'R',
                          'S']))))))))))
         of {
         Inl aa -> Inl aa;
         Inr _ -> Inr (mka i (nfsa i tp) [] uR (uRw ++ u));
       });

apply_af_entry :: forall a b. a -> Af_entry -> [Term a b] -> Term a b;
apply_af_entry uu (Collapse i) ts = nth ts i;
apply_af_entry f (AFList is) ts = Fun f (map (nth ts) is);

afs_term ::
  forall a b. (a -> Nat -> Af_entry) -> Term a b -> Term (Filtered a) b;
afs_term pia (Fun f ts) =
  let {
    l = size_list ts;
  } in apply_af_entry (FPair f l) (pia f l) (map (afs_term pia) ts);
afs_term pia (Var x) = Var x;

afs_rule ::
  forall a b.
    (a -> Nat -> Af_entry) ->
      (Term a b, Term a b) -> (Term (Filtered a) b, Term (Filtered a) b);
afs_rule pia lr = (afs_term pia (fst lr), afs_term pia (snd lr));

xml2nwcr_proof :: Xml -> Parse_result [Char] ();
xml2nwcr_proof =
  xml1element ['w', 'c', 'r', 'D', 'i', 's', 'p', 'r', 'o', 'o', 'f']
    (xml0elements
      ['n', 'o', 'n', 'J', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'C', 'r', 'i',
        't', 'i', 'c', 'a', 'l', 'P', 'a', 'i', 'r', 's', 'A', 'u', 't', 'o']
      ())
    (\ _ -> ());

default_grd_fun ::
  forall a.
    (Eq a) => Term (Lab [Char] a) [Char] ->
                Term (Lab [Char] a) [Char] ->
                  [Char] -> Term (Lab [Char] a) [Char];
default_grd_fun s t =
  let {
    f = funs_rule_list (s, t);
    m = fold (\ fa m ->
               (case fa of {
                 Lab _ _ -> m;
                 FunLab _ _ -> m;
                 UnLab _ -> m;
                 Sharp (Lab _ _) -> m;
                 Sharp (FunLab _ _) -> m;
                 Sharp (UnLab g) -> maxa (size_list g) m;
                 Sharp (Sharp _) -> m;
               }))
          f Zero_nat;
    suffix = replicatea (Suc m) 'a';
  } in (\ x -> Fun (Sharp (UnLab (x ++ suffix))) []);

xml2ncr_proof ::
  (Xml -> Parse_result [Char] (Lab [Char] [Nat])) ->
    Xml -> Parse_result [Char] (Ncr_proof [Char] [Nat] [Char]);
xml2ncr_proof xml2name x =
  let {
    rew = rsteps xml2name;
  } in xml1element ['c', 'r', 'D', 'i', 's', 'p', 'r', 'o', 'o', 'f']
         (xml_options
           [(['n', 'o', 'n', 'W', 'c', 'r', 'A', 'n', 'd', 'S', 'N'],
              xml2elements
                ['n', 'o', 'n', 'W', 'c', 'r', 'A', 'n', 'd', 'S', 'N']
                xml2nwcr_proof (xml2trs_proof xml2name) (\ _ -> SN_NWCR)),
             (['n', 'o', 'n', 'J', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'F', 'o',
                'r', 'k'],
               xml3elements
                 ['n', 'o', 'n', 'J', 'o', 'i', 'n', 'a', 'b', 'l', 'e', 'F',
                   'o', 'r', 'k']
                 rew rew
                 (xml_options
                   [(['d', 'i', 's', 't', 'i', 'n', 'c', 't', 'N', 'o', 'r',
                       'm', 'a', 'l', 'F', 'o', 'r', 'm', 's'],
                      xml0elements
                        ['d', 'i', 's', 't', 'i', 'n', 'c', 't', 'N', 'o', 'r',
                          'm', 'a', 'l', 'F', 'o', 'r', 'm', 's']
                        Diff_NFs),
                     (['c', 'a', 'p', 'N', 'o', 't', 'U', 'n', 'i', 'f'],
                       xml0elements
                         ['c', 'a', 'p', 'N', 'o', 't', 'U', 'n', 'i', 'f']
                         (Tcap_Non_Unif default_grd_fun))])
                 (\ (s, seq1) (_, a) -> Non_Join s seq1 a))])
         id x;

xml_signature :: forall a b. (Showa a, Showa b) => [(Lab a [b], Nat)] -> Xml;
xml_signature fs =
  Xml ['s', 'i', 'g', 'n', 'a', 't', 'u', 'r', 'e'] []
    (map (\ (f, n) ->
           Xml ['s', 'y', 'm', 'b', 'o', 'l'] []
             [xml_lab f,
               Xml ['a', 'r', 'i', 't', 'y'] [] []
                 (Just (shows_prec_nat Zero_nat n []))]
             Nothing)
      fs)
    Nothing;

xml_trs_input ::
  forall a b c.
    (Showa a, Showa b,
      Showa c) => Sum (Strategy (Lab a [b]) c) (Ostrategy (Lab a [b]) c) ->
                    [(Term (Lab a [b]) c, Term (Lab a [b]) c)] ->
                      Maybe [(Term (Lab a [b]) c, Term (Lab a [b]) c)] -> Xml;
xml_trs_input q r Nothing =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    ([xml_rules ['t', 'r', 's'] r] ++ xml_Q q) Nothing;
xml_trs_input q r (Just s) =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    (xml_rules ['t', 'r', 's'] r :
      xml_Q q ++
        [xml_rules
           ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'] s])
    Nothing;

search_len :: [Char] -> [Char] -> Nat -> Maybe ([Char], [Char]);
search_len t s n =
  (if takea n t == s then Just ([], dropa n t)
    else (case t of {
           [] -> Nothing;
           h : tt ->
             (case search_len tt s n of {
               Nothing -> Nothing;
               Just (bef, aft) -> Just (h : bef, aft);
             });
         }));

instance (Eq a, Linorder a, Eq b, Linorder b) => Linorder (Lab a b) where {
};

indent :: ([Char] -> [Char]) -> [Char] -> [Char];
indent p s =
  concatMap (\ c -> (if c == newline then [newline, ' '] else [c])) (p s);

shows_terms ::
  forall a b.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> [Char] -> [Term a b] -> [Char] -> [Char];
shows_terms fun var name ts =
  shows_string name .
    shows_nl .
      shows_list_gen (shows_term fun var) [] [] [newline] [] ts . shows_nl;

shows_tp ::
  forall a b c.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> Tp_ops_ext c a b () -> c -> [Char] -> [Char];
shows_tp fun var i t =
  let {
    r = ra i t;
    rw = rwa i t;
    q = qa i t;
  } in shows_trs fun var ['r', 'u', 'l', 'e', 's', ':'] [' ', '-', '>', ' '] r .
         (if null rw then id
           else shows_trs fun var
                  ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'r', 'u', 'l',
                    'e', 's', ':']
                  [' ', '-', '>', '=', ' '] rw) .
           (if null q then id
             else shows_terms fun var
                    ['Q', '-', 'c', 'o', 'm', 'p', 'o', 'n', 'e', 'n', 't', ':']
                    q);

nF_terms_list ::
  forall a b c. (Eq a, Eq b, Eq c) => [Term a b] -> Term a c -> Bool;
nF_terms_list ls t =
  all (\ s -> all (\ l -> not (matches s l)) ls) (supteq_list t);

check_NF_terms_subset ::
  forall a b. (Eq a, Eq b) => [Term a b] -> [Term a b] -> Sum (Term a b) ();
check_NF_terms_subset qa q =
  catcha
    (forallm (\ x -> (if not (nF_terms_list qa x) then Inr () else Inl x)) q)
    (\ x -> Inl (snd x));

check_tp_subsumes ::
  forall a b c d.
    (Eq b, Showa b, Eq c, Showa c, Eq d,
      Showa d) => Tp_ops_ext a (Lab b c) d () ->
                    (Bool,
                      ([Term (Lab b c) d],
                        ([(Term (Lab b c) d, Term (Lab b c) d)],
                          [(Term (Lab b c) d, Term (Lab b c) d)]))) ->
                      a -> Sum ([Char] -> [Char]) ();
check_tp_subsumes i (nfs, (q, (r, rw))) tp =
  catcha
    (let {
       qaa = qa i tp;
       raa = ra i tp;
       rwaa = rwa i tp;
       rb = r ++ rw;
     } in bindb (catcha (check_NF_terms_subset qaa q)
                  (\ x ->
                    Inl (shows_prec_list Zero_nat
                           ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                             't', 'h', ' ', 'i', 'n', 'n', 'e', 'r', 'm', 'o',
                             's', 't', ' ', 's', 't', 'r', 'a', 't', 'e', 'g',
                             'y', ' ', 'd', 'u', 'e', ' ', 't', 'o', ' ', 't',
                             'e', 'r', 'm', ' '] .
                          shows_term (shows_prec_lab Zero_nat)
                            (shows_prec Zero_nat) x)))
            (\ _ ->
              bindb (catcha (check_subseteq raa r)
                      (\ x ->
                        Inl (toomuch ['r', 'u', 'l', 'e']
                              (shows_rule (shows_prec_lab Zero_nat)
                                (shows_prec Zero_nat) [' ', '-', '>', ' '] x))))
                (\ _ ->
                  bindb (catcha (check_subseteq rwaa rb)
                          (\ x ->
                            Inl (toomuch
                                  ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ',
                                    'r', 'u', 'l', 'e']
                                  (shows_rule (shows_prec_lab Zero_nat)
                                    (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                    x))))
                    (\ _ -> Inr ()))))
    (\ x ->
      Inl (shows_string
             ['t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'o',
               'f', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'b', 'l', 'e',
               'm'] .
            shows_nl .
              shows_tp (shows_prec_lab Zero_nat) (shows_prec Zero_nat) i tp .
                shows_nl .
                  shows_string
                    ['m', 'a', 'y', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ', 'c',
                      'o', 'n', 'c', 'l', 'u', 'd', 'e', 'd', ' ', 'f', 'r',
                      'o', 'm', ' ', 'a', 's', 's', 'u', 'm', 'i', 'n', 'g',
                      ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
                      'n', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'p', 'r',
                      'o', 'b', 'l', 'e', 'm'] .
                    shows_nl .
                      shows_tp (shows_prec_lab Zero_nat) (shows_prec Zero_nat) i
                        (mka i nfs q r rw) .
                        shows_nl . x . shows_nl));

data Ta_rule_impl a b = TA_rule_impl b [a] a (Rbt a ());

data Ta_impl_ext a b c =
  Ta_impl_ext (Rbt a ()) (Rbt (b, Nat) [Ta_rule_impl a b]) [a] (Rbt a ())
    [(a, a)] (a -> Rbt a ()) (a -> Rbt a ()) c;

ta_rqs_set :: forall a b c. Ta_impl_ext a b c -> Rbt a ();
ta_rqs_set
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_rqs_set;

rsym_impl :: forall a b. Ta_rule_impl a b -> (b, Nat);
rsym_impl (TA_rule_impl f qsa q qs) = (f, size_list qsa);

conv_ta_rule :: forall a b. (a -> Rbt a ()) -> Ta_rule a b -> Ta_rule_impl a b;
conv_ta_rule eps (TA_rule f qs q) = TA_rule_impl f qs q (eps q);

relpow_impl ::
  forall a b.
    ([a] -> [a]) -> ([a] -> b -> b) -> (a -> b -> Bool) -> [a] -> b -> Nat -> b;
relpow_impl succa un memb new have Zero_nat = un new have;
relpow_impl succa un memb new have (Suc m) =
  (if null new then have
    else let {
           maybea = succa new;
           havea = un new have;
           newa = filter (\ n -> not (memb n havea)) maybea;
         } in relpow_impl succa un memb newa havea m);

rtrancl_impl ::
  forall a b.
    ([(a, a)] -> [a] -> [a]) ->
      ([a] -> b -> b) -> (a -> b -> Bool) -> b -> [(a, a)] -> [a] -> b;
rtrancl_impl gen_succ un memb emp rel =
  let {
    succa = gen_succ rel;
    n = size_list rel;
  } in (\ asa -> relpow_impl succa un memb asa emp n);

rtrancl_rbt_impl :: forall a. (Linorder a) => [(a, a)] -> [a] -> Rbt a ();
rtrancl_rbt_impl =
  rtrancl_impl
    (\ r ->
      let {
        rm = elem_list_to_rm fst r;
      } in (\ asa ->
             rs_to_list
               (foldla
                 (\ a x ->
                   rs_union a (list_to_rs (map snd (rm_set_lookup rm x))))
                 (rs_empty ()) asa)))
    (\ asa bs -> rs_union bs (list_to_rs asa)) rs_memb (rs_empty ());

memo_rbt_rtrancl :: forall a. (Linorder a) => [(a, a)] -> a -> Rbt a ();
memo_rbt_rtrancl r =
  let {
    tr = rtrancl_rbt_impl r;
    rm = list_to_rm
           (map (\ a -> (a, tr [a])) (((rs_to_list . list_to_rs) . map fst) r));
  } in (\ a ->
         (case rm_lookup a rm of {
           Nothing -> list_to_rs [a];
           Just asa -> asa;
         }));

generate_ta ::
  forall a b.
    (Linorder a, Eq b, Linorder b) => Tree_automaton a b -> Ta_impl_ext a b ();
generate_ta (Tree_Automaton fin rules eps) =
  let {
    ep = memo_rbt_rtrancl eps;
    epr = memo_rbt_rtrancl (map (\ (q, qa) -> (qa, q)) eps);
    rqs_rs = foldla (\ a x -> rs_union a (ep (rq x))) (rs_empty ()) rules;
    rrules = map (conv_ta_rule ep) rules;
  } in Ta_impl_ext (list_to_rs fin) (elem_list_to_rm rsym_impl rrules)
         (rs_to_list rqs_rs) rqs_rs eps ep epr ();

check_left_linear_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_left_linear_trs trs =
  catcha
    (catcha
      (forallm (\ x -> (if linear_term (fst x) then Inr () else Inl x)) trs)
      (\ x -> Inl (snd x)))
    (\ _ ->
      Inl (shows_prec_list Zero_nat trs .
            shows_nl .
              shows_string
                ['i', 's', ' ', 'n', 'o', 't', ' ', 'l', 'e', 'f', 't', '-',
                  'l', 'i', 'n', 'e', 'a', 'r'] .
                shows_nl));

boundstype_fun ::
  forall a b. (Eq b) => Boundstype -> (Term a b, Term a b) -> Term a b -> Bool;
boundstype_fun Roof = roof;
boundstype_fun Match = match;

ta_rules_impl ::
  forall a b c. Ta_impl_ext a b c -> Rbt (b, Nat) [Ta_rule_impl a b];
ta_rules_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_rules_impl;

bounds_condition ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => Boundstype ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
bounds_condition Roof uu = Inr ();
bounds_condition Match r =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (_, a) = x;
              } in linear_term a
            then Inr () else Inl x))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (let {
             (l, ra) = x;
           } in shows_prec_list Zero_nat ['r', 'u', 'l', 'e', ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '-', '>', ' '] (l, ra) .
                    shows_prec_list Zero_nat
                      [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'r', 'i', 'g',
                        'h', 't', '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ']));

rqs_impl :: forall a b. Ta_rule_impl a b -> [a];
rqs_impl (TA_rule_impl f qsa q qs) = qsa;

rqss_impl :: forall a b. Ta_rule_impl a b -> Rbt a ();
rqss_impl (TA_rule_impl f qsa q qs) = qs;

ta_epss_impl :: forall a b c. Ta_impl_ext a b c -> a -> Rbt a ();
ta_epss_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_epss_impl;

ta_final_impl :: forall a b c. Ta_impl_ext a b c -> Rbt a ();
ta_final_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_final_impl;

ta_contains_impl ::
  forall a b.
    (Eq a, Linorder a, Eq b,
      Linorder b) => [(a, Nat)] ->
                       Ta_impl_ext b a () -> [b] -> Sum (Sum b (Term a b)) ();
ta_contains_impl f ta qs =
  let {
    fin = ta_final_impl ta;
    look = rm_set_lookup (ta_rules_impl ta);
    _ = ta_epss_impl ta;
  } in bindb (catcha
               (catcha (forallm (\ q -> check (rs_memb q fin) q) qs)
                 (\ x -> Inl (snd x)))
               (\ x -> Inl (Inl x)))
         (\ _ ->
           catcha
             (catcha
               (forallm
                 (\ (fa, n) ->
                   let {
                     rules = look (fa, n);
                   } in catcha
                          (catcha
                            (forallm
                              (\ x ->
                                (if any (\ rule ->
  rqs_impl rule == x &&
    let {
      qq = rqss_impl rule;
    } in any (\ q -> rs_memb q qq) qs)
                                      rules
                                  then Inr () else Inl x))
                              (concat_lists (replicatea n qs)))
                            (\ x -> Inl (snd x)))
                          (\ x -> Inl (Fun fa (map Var x))))
                 f)
               (\ x -> Inl (snd x)))
             (\ x -> Inl (Inr x)));

flatten_term_enum_filter ::
  forall a b. (Term a b -> Bool) -> Term [a] b -> [Term a b];
flatten_term_enum_filter f (Var x) =
  let {
    tx = Var x;
  } in (if f tx then [tx] else []);
flatten_term_enum_filter f (Fun fs ts) =
  let {
    lts = map (flatten_term_enum_filter f) ts;
  } in (if any null lts then []
         else let {
                ss = concat_lists lts;
              } in filter f (concatMap (\ fa -> map (Fun fa) ss) fs));

inverse_base_term_filter ::
  forall a b. (Term (a, Nat) b -> Bool) -> Term a b -> Nat -> [Term (a, Nat) b];
inverse_base_term_filter filt l c =
  let {
    hs = upt Zero_nat (Suc c);
  } in flatten_term_enum_filter filt
         (map_term (\ f -> map (\ h -> lift h f) hs) l);

cover_bound_list_filter ::
  forall a b.
    (Term (a, Nat) b -> Bool) ->
      ((Term a b, Term a b) -> Term a b -> Bool) ->
        Nat -> [(Term a b, Term a b)] -> [(Term (a, Nat) b, Term (a, Nat) b)];
cover_bound_list_filter filt ff c r =
  concatMap
    (\ (l, ra) ->
      let {
        ee = ff (l, ra);
      } in map (\ la ->
                 (la, map_term
                        (lift (Suc (min_list
                                     (map height
                                       (sym_collect
 (\ t -> ee (map_term base t)) la)))))
                        ra))
             (inverse_base_term_filter filt l c))
    r;

ta_res_impl ::
  forall a b.
    (Eq a, Linorder a,
      Linorder b) => Rbt (a, Nat) [Ta_rule_impl b a] ->
                       (b -> Rbt b ()) -> Term a b -> Rbt b ();
ta_res_impl ta eps (Var q) = eps q;
ta_res_impl ta eps (Fun f ts) =
  let {
    rec = map (ta_res_impl ta eps) ts;
    n = size_list ts;
    rules = rm_set_lookup ta (f, n);
    arules =
      filter
        (\ rule ->
          let {
            qs = rqs_impl rule;
          } in all_interval_nat (\ i -> rs_memb (nth qs i) (nth rec i)) Zero_nat
                 n)
        rules;
    a = map rqss_impl arules;
  } in foldla rs_union (rs_empty ()) a;

ta_rqs_impl :: forall a b c. Ta_impl_ext a b c -> [a];
ta_rqs_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_rqs_impl;

rs_subset :: forall a. (Linorder a) => Rbt a () -> Rbt a () -> Maybe a;
rs_subset asa bs =
  rs_iteratei asa (\ a -> (case a of {
                            Nothing -> True;
                            Just _ -> False;
                          }))
    (\ a _ -> (if rs_memb a bs then Nothing else Just a)) Nothing;

ta_epsrs_impl :: forall a b c. Ta_impl_ext a b c -> a -> Rbt a ();
ta_epsrs_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    ta_epss_impl ta_epsrs_impl more)
  = ta_epsrs_impl;

ta_match_impl ::
  forall a b c.
    (Eq a, Linorder a, Eq b, Linorder b, Eq c,
      Linorder c) => Rbt (a, Nat) [Ta_rule_impl b a] ->
                       (b -> Rbt b ()) -> Term a c -> b -> Rbt [(c, b)] ();
ta_match_impl ta eps (Var x) q =
  list_to_rs (map (\ qa -> [(x, qa)]) (rs_to_list (eps q)));
ta_match_impl ta eps (Fun f ts) q =
  let {
    n = size_list ts;
    rules = rm_set_lookup ta (f, n);
    ep = eps q;
    fa = (\ rule ->
           list_to_rs
             (let {
                (TA_rule_impl _ qs qa _) = rule;
              } in (if rs_memb qa ep
                     then let {
                            rec = map (\ (tsi, qsi) ->
rs_to_list (ta_match_impl ta eps tsi qsi))
                                    (zip ts qs);
                          } in map concat (concat_lists rec)
                     else [])));
  } in foldla (\ a x -> rs_union a (fa x)) (rs_empty ()) rules;

ta_rule_compatible_eff_list ::
  forall a b c.
    (Eq a, Linorder a, Eq b, Linorder b, Eq c,
      Linorder c) => Ta_impl_ext a b () ->
                       (Term b c, Term b c) -> Sum ((Term b a, Term b a), a) ();
ta_rule_compatible_eff_list ta (Fun f ls, r) =
  let {
    rm = ta_rules_impl ta;
    eps = ta_epss_impl ta;
    epsa = ta_epsrs_impl ta;
    n = size_list ls;
    ta_res = ta_res_impl rm eps;
  } in catcha
         (forallm
           (\ (TA_rule_impl fa _ q _) ->
             catcha
               (forallm
                 (\ sigma ->
                   let {
                     sigmaa = fun_of sigma;
                     l_sigma = map_vars sigmaa (Fun fa ls);
                     r_sigma = map_vars sigmaa r;
                     qsl = ta_res l_sigma;
                     qsr = ta_res r_sigma;
                   } in (case rs_subset qsl qsr of {
                          Nothing -> Inr ();
                          Just qa -> Inl ((l_sigma, r_sigma), qa);
                        }))
                 (rs_to_list (ta_match_impl rm epsa (Fun fa ls) q)))
               (\ x -> Inl (snd x)))
           (rm_set_lookup rm (f, n)))
         (\ x -> Inl (snd x));
ta_rule_compatible_eff_list ta (Var x, r) =
  let {
    rm = ta_rules_impl ta;
    eps = ta_epss_impl ta;
    ta_res = ta_res_impl rm eps;
  } in catcha
         (forallm
           (\ q ->
             let {
               r_sigma = map_vars (\ _ -> q) r;
             } in check (rs_memb q (ta_res r_sigma)) ((Var q, r_sigma), q))
           (ta_rqs_impl ta))
         (\ xa -> Inl (snd xa));

ta_trs_compatible_eff_list ::
  forall a b c.
    (Eq a, Linorder a, Eq b, Linorder b, Eq c,
      Linorder c) => Ta_impl_ext a b () ->
                       [(Term b c, Term b c)] ->
                         Sum ((Term b c, Term b c), ((Term b a, Term b a), a))
                           ();
ta_trs_compatible_eff_list ta r =
  let {
    check = ta_rule_compatible_eff_list ta;
  } in catcha (forallm (\ lr -> catcha (check lr) (\ x -> Inl (lr, x))) r)
         (\ x -> Inl (snd x));

ta_res_impl_all ::
  forall a b c.
    (Linorder a, Eq b,
      Linorder b) => Rbt a () ->
                       Rbt (b, Nat) [Ta_rule_impl a b] -> Term b c -> Rbt a ();
ta_res_impl_all q ta (Var uu) = q;
ta_res_impl_all q ta (Fun f ts) =
  let {
    rec = map (ta_res_impl_all q ta) ts;
    n = size_list ts;
    rules = rm_set_lookup ta (f, n);
    arules =
      filter
        (\ rule ->
          let {
            qs = rqs_impl rule;
          } in all_interval_nat (\ i -> rs_memb (nth qs i) (nth rec i)) Zero_nat
                 n)
        rules;
    a = map rqss_impl arules;
  } in foldla rs_union (rs_empty ()) a;

ta_rule_compatible_heuristic ::
  forall a b c.
    (Eq a, Linorder a, Eq b, Linorder b,
      Linorder c) => Ta_impl_ext a b () -> Term b c -> Bool;
ta_rule_compatible_heuristic ta l =
  rs_isEmpty (ta_res_impl_all (ta_rqs_set ta) (ta_rules_impl ta) l);

check_bounds ::
  forall a b.
    (Eq a, Linorder a, Showa a, Eq b, Linorder b,
      Showa b) => Bounds_info a b ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_bounds (Bounds_Info typea c qfin preTA) r =
  let {
    ta = generate_ta preTA;
  } in bindb (check_wf_trs r)
         (\ _ ->
           bindb (check_left_linear_trs r)
             (\ _ ->
               bindb (bounds_condition typea r)
                 (\ _ ->
                   bindb (check (any (\ q -> rs_memb q (ta_rqs_set ta)) qfin)
                           (shows_prec_list Zero_nat
                             ['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                               'n', 'd', ' ', 'm', 'e', 'n', 't', 'i', 'o', 'n',
                               'e', 'd', ' ', 'f', 'i', 'n', 'a', 'l', ' ', 's',
                               't', 'a', 't', 'e', ' ', 'i', 'n', ' ', 'T',
                               'A']))
                     (\ _ ->
                       bindb (catcha
                               (catcha
                                 (forallm
                                   (\ x ->
                                     (if let {
   (f, _) = x;
 } in less_eq_nat (height f) c
                                       then Inr () else Inl x))
                                   (map fst (rm_to_list (ta_rules_impl ta))))
                                 (\ x -> Inl (snd x)))
                               (\ x ->
                                 Inl (let {
(f, _) = x;
                                      } in
shows_prec_prod Zero_nat f .
  shows_prec_list Zero_nat
    [' ', 'i', 's', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 'T',
      'A', ' ', 'w', 'i', 't', 'h', ' ', 'h', 'e', 'i', 'g', 'h', 't', ' ', 'l',
      'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ', 'c', ' ', '=',
      ' '] .
    shows_prec_nat Zero_nat c)))
                         (\ _ ->
                           bindb (catcha
                                   (ta_contains_impl
                                     (map (\ (f, a) -> (lift Zero_nat f, a))
                                       (funas_trs_list r))
                                     ta qfin)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['i', 't', ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'b', 'e',
      ' ', 'g', 'u', 'a', 'r', 'a', 'n', 't', 'e', 'e', 'd', ' ', 't', 'h', 'a',
      't', ' ', 'l', 'i', 'f', 't', '0', '(', 'T', '(', 'S', 'i', 'g', 'm', 'a',
      ')', ')', ' ', 'i', 's', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd', ' ',
      'b', 'y', ' ', 'T', 'A'] .
   shows_nl .
     (case x of {
       Inl q ->
         shows_prec_list Zero_nat ['s', 't', 'a', 't', 'e', ' '] .
           shows_prec Zero_nat q .
             shows_prec_list Zero_nat
               [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'a',
                 'l'];
       Inr fqs ->
         shows_prec_list Zero_nat
           ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o', ' ', 't',
             'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n', ' ', 'f', 'r', 'o',
             'm', ' '] .
           shows_prec_term Zero_nat fqs .
             shows_prec_list Zero_nat
               [' ', 't', 'o', ' ', 'a', ' ', 'f', 'i', 'n', 'a', 'l', ' ', 's',
                 't', 'a', 't', 'e'];
     }))))
                             (\ _ ->
                               catcha
                                 (ta_trs_compatible_eff_list ta
                                   (cover_bound_list_filter
                                     (\ l ->
                                       not (ta_rule_compatible_heuristic ta l))
                                     (boundstype_fun typea) c r))
                                 (\ x ->
                                   Inl (let {
  (lr, (lrq, q)) = x;
} in shows_prec_list Zero_nat
       ['T', 'A', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm', 'p',
         'a', 't', 'i', 'b', 'l', 'e', ' ', 'w', 'i', 't', 'h', ' ', 'R'] .
       shows_nl .
         shows_prec_list Zero_nat
           ['f', 'o', 'r', ' ', 'r', 'u', 'l', 'e', ' '] .
           shows_rule (shows_prec_prod Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] lr .
             shows_nl .
               shows_prec_list Zero_nat
                 ['w', 'h', 'i', 'c', 'h', ' ', 'i', 's', ' ', 'i', 'n', 's',
                   't', 'a', 'n', 't', 'i', 'a', 't', 'e', 'd', ' ', 'b', 'y',
                   ' ', 's', 't', 'a', 't', 'e', 's', ' ', 't', 'o', ' '] .
                 shows_rule (shows_prec_prod Zero_nat) (shows_prec Zero_nat)
                   [' ', '-', '>', ' '] lrq .
                   shows_nl .
                     shows_prec_list Zero_nat
                       ['t', 'h', 'e', ' ', 's', 't', 'a', 't', 'e', ' '] .
                       shows_prec Zero_nat q .
                         shows_prec_list Zero_nat
                           [' ', 'i', 's', ' ', 'o', 'n', 'l', 'y', ' ', 'r',
                             'e', 'a', 'c', 'h', 'a', 'b', 'l', 'e', ' ', 'f',
                             'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'l', 'h',
                             's'] .
                           shows_nl))))))));

rulesb :: forall a b c d. Tp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
rulesb
  (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs more)
  = rules;

bounds_tt ::
  forall a b c.
    (Eq b, Linorder b, Showa b, Eq c, Linorder c,
      Showa c) => Tp_ops_ext a b c () ->
                    Bounds_info b c -> a -> Sum ([Char] -> [Char]) ();
bounds_tt i info tp = let {
                        a = rulesb i tp;
                      } in check_bounds info a;

shows_prec_ctxt ::
  forall a b. (Showa a, Showa b) => Nat -> Ctxt a b -> [Char] -> [Char];
shows_prec_ctxt d c = shows_ctxt (shows_prec Zero_nat) (shows_prec Zero_nat) c;

check_flat_ctxt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [a] -> Ctxt b a -> Sum ([Char] -> [Char]) ();
check_flat_ctxt vs c =
  (case c of {
    Hole ->
      Inl (shows_prec_ctxt Zero_nat c .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a',
                't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
              shows_nl);
    More _ ss1 Hole ss2 ->
      let {
        ss = ss1 ++ ss2;
      } in bindb (check (distinct ss)
                   (shows_prec_ctxt Zero_nat c .
                     shows_string
                       [' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 's', ' ', 'd',
                         'u', 'p', 'l', 'i', 'c', 'a', 't', 'e', ' ', 'v', 'a',
                         'r', 'i', 'a', 'b', 'l', 'e', 's'] .
                       shows_nl))
             (\ _ ->
               bindb (check (all is_Var ss)
                       (shows_prec_ctxt Zero_nat c .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'l',
                             'a', 't', ',', ' ', 'i', '.', 'e', '.', ',', ' ',
                             'h', 'a', 's', ' ', 'd', 'e', 'p', 't', 'h', ' ',
                             'g', 'r', 'e', 'a', 't', 'e', 'r', ' ', 't', 'h',
                             'a', 'n', ' ', 'o', 'n', 'e'] .
                           shows_nl))
                 (\ _ ->
                   check (all (\ t -> not (membera vs (the_Var t)))
                           (ss1 ++ ss2))
                     (shows_prec_ctxt Zero_nat c .
                       shows_string
                         [' ', 'h', 'a', 's', ' ', 't', 'o', ' ', 'c', 'o', 'n',
                           't', 'a', 'i', 'n', ' ', 'o', 'n', 'l', 'y', ' ',
                           'f', 'r', 'e', 's', 'h', ' ', 'v', 'a', 'r', 'i',
                           'a', 'b', 'l', 'e', 's'] .
                         shows_nl)));
    More _ ss1 (More _ _ _ _) ss2 ->
      Inl (shows_prec_ctxt Zero_nat c .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a',
                't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
              shows_nl);
  });

partition_rules ::
  forall a b.
    (Eq a,
      Eq b) => [Ctxt a b] ->
                 [(Term a b, Term a b)] ->
                   [(Term a b, Term a b)] ->
                     ([(Term a b, Term a b)], [(Term a b, Term a b)]);
partition_rules cs r =
  partition
    (\ lr ->
      any (\ (u, v) ->
            any (\ c -> lr == (ctxt_apply c u, ctxt_apply c v)) (Hole : cs))
        r);

is_flat_ctxt_list ::
  forall a b. (Eq a, Eq b) => [a] -> [(b, Nat)] -> Ctxt b a -> Bool;
is_flat_ctxt_list vs fas (More f ss1 Hole ss2) =
  let {
    ss = ss1 ++ ss2;
  } in membera fas (f, Suc (size_list ss)) &&
         all is_Var ss && distinct ss && null (list_inter (map the_Var ss) vs);
is_flat_ctxt_list vs fas Hole = False;
is_flat_ctxt_list vs fas (More v va (More vd ve vf vg) vc) = False;

check_is_flat_ctxt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [a] -> [(b, Nat)] -> Ctxt b a -> Sum ([Char] -> [Char]) ();
check_is_flat_ctxt vs fas c =
  check (is_flat_ctxt_list vs fas c)
    (shows_prec_ctxt Zero_nat c .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a', 't',
          ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
        shows_nl);

check_rule_preserving ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [Ctxt a b] ->
                    [(Term a b, Term a b)] ->
                      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_rule_preserving fcs rs rule =
  check (membera rs rule ||
          all (\ c ->
                membera rs (ctxt_apply c (fst rule), ctxt_apply c (snd rule)))
            fcs)
    (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
      shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
        [' ', '-', '>', ' '] rule .
        shows_string
          [' ', 'i', 's', ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
            'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 't', 'h',
            'e', ' ', 'r', 'e', 's', 'u', 'l', 't', 'i', 'n', 'g', ' ', 's',
            'e', 't', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'n',
            'o', 'r', ' ', 'c', 'l', 'o', 's', 'e', 'd', ' ', 'u', 'n', 'd',
            'e', 'r', ' ', 'a', 'l', 'l', ' ', 'f', 'l', 'a', 't', ' ', 'c',
            'o', 'n', 't', 'e', 'x', 't', 's'] .
          shows_nl);

hole_at :: forall a b. (Eq a) => Nat -> Nat -> a -> Ctxt a b -> Bool;
hole_at n i f (More g ss1 Hole ss2) =
  g == f &&
    equal_nat (size_list ss1) i &&
      equal_nat (size_list ss2) (minus_nat (minus_nat n i) one_nat);
hole_at n i f Hole = False;
hole_at n i f (More v va (More vd ve vf vg) vc) = False;

check_flat_ctxt_complete ::
  forall a b.
    (Eq a, Showa a,
      Showa b) => [Ctxt a b] -> (a, Nat) -> Sum ([Char] -> [Char]) ();
check_flat_ctxt_complete fcs fa =
  check (all_interval_nat (\ i -> any (hole_at (snd fa) i (fst fa)) fcs)
          Zero_nat (snd fa))
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'i', 's', 't', ' ', 'o', 'f', ' ', 'f', 'l',
         'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's', ' ', 'i', 's',
         ' ', 'i', 'n', 'c', 'o', 'm', 'p', 'l', 'e', 't', 'e'] .
      shows_nl);

fcc_tt ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Tp_ops_ext a b c () ->
                    [Ctxt b c] ->
                      [(Term b c, Term b c)] -> a -> Sum ([Char] -> [Char]) a;
fcc_tt i fcs cRb tp =
  let {
    r = ra i tp;
    rw = rwa i tp;
    nfs = nfsa i tp;
    rb = r ++ rw;
    (cR, cRw) = partition_rules fcs r cRb;
    _ = qa i tp;
    vs = vars_trs_list rb;
    fas = funas_trs_list rb;
  } in (case bindb (check (not (null fcs))
                     (shows_string
                        ['a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n',
                          'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't',
                          'e', 'x', 't', ' ', 'i', 's', ' ', 'r', 'e', 'q', 'u',
                          'i', 'r', 'e', 'd', ' ', 'f', 'o', 'r', ' ', 'f', 'l',
                          'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ',
                          'c', 'l', 'o', 's', 'u', 'r', 'e'] .
                       shows_nl))
               (\ _ ->
                 bindb (catcha (forallm (check_flat_ctxt vs) fcs)
                         (\ x -> Inl (snd x)))
                   (\ _ ->
                     bindb (catcha (forallm (check_is_flat_ctxt vs fas) fcs)
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         bindb (catcha
                                 (forallm (check_flat_ctxt_complete fcs) fas)
                                 (\ x -> Inl (snd x)))
                           (\ _ ->
                             bindb (catcha
                                     (forallm (check_rule_preserving fcs cR) r)
                                     (\ x -> Inl (snd x)))
                               (\ _ ->
                                 catcha
                                   (forallm (check_rule_preserving fcs cRb) rw)
                                   (\ x -> Inl (snd x)))))))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mka i nfs [] cR cRw);
       });

get_fcc_option ::
  forall a b c.
    Dp_proof a b c ->
      Maybe (Lab a b,
              ([Ctxt (Lab a b) c],
                ([(Term (Lab a b) c, Term (Lab a b) c)],
                  ([(Term (Lab a b) c, Term (Lab a b) c)], Dp_proof a b c))));
get_fcc_option (FccProc f fcs pb rb prf) = Just (f, (fcs, (pb, (rb, prf))));
get_fcc_option PisEmpty = Nothing;
get_fcc_option (SubtermCriterionProc v va vb vc) = Nothing;
get_fcc_option (RedPairProc v va vb) = Nothing;
get_fcc_option (RedPairUrProc v va vb vc) = Nothing;
get_fcc_option (UsableRulesProc v va) = Nothing;
get_fcc_option (DepGraphProc v) = Nothing;
get_fcc_option (MonoRedPairProc v va vb vc) = Nothing;
get_fcc_option (MonoRedPairUrProc v va vb vc vd) = Nothing;
get_fcc_option (SizeChangeSubterm v) = Nothing;
get_fcc_option (SizeChangeRedPair v va vb) = Nothing;
get_fcc_option (UncurryProc v va vb vc vd) = Nothing;
get_fcc_option (SplitProc v va vb vc) = Nothing;
get_fcc_option (SemlabProc v va vb vc vd) = Nothing;
get_fcc_option (SwitchInnermostProc v va) = Nothing;
get_fcc_option (RewritingProc v va vb vc vd ve vf) = Nothing;
get_fcc_option (InstantiationProc v va vb) = Nothing;
get_fcc_option (ForwardInstantiationProc v va vb vc) = Nothing;
get_fcc_option (NarrowingProc v va vb vc) = Nothing;
get_fcc_option (AssumeFinite v) = Nothing;
get_fcc_option (UnlabProc v va vb) = Nothing;
get_fcc_option (QReductionProc v va) = Nothing;

forallm_index_aux ::
  forall a b. (a -> Nat -> Sum b ()) -> Nat -> [a] -> Sum ((a, Nat), b) ();
forallm_index_aux p i [] = Inr ();
forallm_index_aux p i (x : xs) =
  bindb (catcha (p x i) (\ xa -> Inl ((x, i), xa)))
    (\ _ -> forallm_index_aux p (Suc i) xs);

forallm_index ::
  forall a b. (a -> Nat -> Sum b ()) -> [a] -> Sum ((a, Nat), b) ();
forallm_index p xs = forallm_index_aux p Zero_nat xs;

pa :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
pa (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
     rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
     intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
     split_pairs split_rules mk minimal nfs wwf_rules more)
  = p;

rc :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
rc (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
     rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
     intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
     split_pairs split_rules mk minimal nfs wwf_rules more)
  = r;

pwa :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
pwa (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
      rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map
      reverse_rules_map intersect_pairs replace_pair intersect_rules delete_P_Pw
      delete_R_Rw split_pairs split_rules mk minimal nfs wwf_rules more)
  = pw;

rwc :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
rwc (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
      rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map
      reverse_rules_map intersect_pairs replace_pair intersect_rules delete_P_Pw
      delete_R_Rw split_pairs split_rules mk minimal nfs wwf_rules more)
  = rw;

shows_dpp ::
  forall a b c.
    (a -> [Char] -> [Char]) ->
      (b -> [Char] -> [Char]) -> Dpp_ops_ext c a b () -> c -> [Char] -> [Char];
shows_dpp fun var i d =
  let {
    m = minimal i d;
    nfs = nfsc i d;
    p = pa i d;
    pw = pwa i d;
    r = rc i d;
    rw = rwc i d;
    q = qc i d;
  } in shows_trs fun var ['p', 'a', 'i', 'r', 's', ':'] [' ', '-', '>', ' '] p .
         (if null pw then id
           else shows_trs fun var
                  ['w', 'e', 'a', 'k', ' ', 'p', 'a', 'i', 'r', 's', ':']
                  [' ', '-', '>', '=', ' '] pw) .
           (if null r then id
             else shows_trs fun var
                    ['s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e', 's',
                      ':']
                    [' ', '-', '>', '!', ' '] r) .
             shows_trs fun var ['r', 'u', 'l', 'e', 's', ':']
               [' ', '-', '>', ' '] rw .
               (if null q then id
                 else shows_terms fun var
                        ['Q', '-', 'c', 'o', 'm', 'p', 'o', 'n', 'e', 'n', 't',
                          ':']
                        q) .
                 (if m then shows_nl .
                              shows_prec_list Zero_nat
                                ['(', 'm', 'i', 'n', 'i', 'm', 'a', 'l', ')']
                   else id) .
                   (if nfs && not (null q)
                     then shows_nl .
                            shows_prec_list Zero_nat
                              ['(', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o',
                                'r', 'm', ' ', 's', 'u', 'b', 's', 't', 'i',
                                't', 'u', 't', 'i', 'o', 'n', 's', ')']
                     else id);

check_NF_terms_eq ::
  forall a b. (Eq a, Eq b) => [Term a b] -> [Term a b] -> Sum (Term a b) ();
check_NF_terms_eq qa q =
  bindb (check_NF_terms_subset qa q) (\ _ -> check_NF_terms_subset q qa);

check_dpp_subsumes ::
  forall a b c d.
    (Eq b, Showa b, Eq c, Showa c, Eq d,
      Showa d) => Dpp_ops_ext a (Lab b c) d () ->
                    (Bool,
                      (Bool,
                        ([(Term (Lab b c) d, Term (Lab b c) d)],
                          ([(Term (Lab b c) d, Term (Lab b c) d)],
                            ([Term (Lab b c) d],
                              ([(Term (Lab b c) d, Term (Lab b c) d)],
                                [(Term (Lab b c) d, Term (Lab b c) d)])))))) ->
                      a -> Sum ([Char] -> [Char]) ();
check_dpp_subsumes i (nfs, (m, (p, (pw, (q, (r, rw)))))) d =
  catcha
    (let {
       paa = pa i d;
       pwaa = pwa i d;
       qa = qc i d;
       ra = rc i d;
       rwa = rwc i d;
       _ = nfsc i d;
       pb = p ++ pw;
       rb = ra ++ rwa;
       rba = r ++ rw;
     } in bindb (catcha (check_subseteq paa p)
                  (\ x ->
                    Inl (toomuch ['p', 'a', 'i', 'r']
                          (shows_rule (shows_prec_lab Zero_nat)
                            (shows_prec Zero_nat) [' ', '-', '>', ' '] x))))
            (\ _ ->
              bindb (catcha (check_subseteq pwaa pb)
                      (\ x ->
                        Inl (toomuch
                              ['w', 'e', 'a', 'k', ' ', 'p', 'a', 'i', 'r']
                              (shows_rule (shows_prec_lab Zero_nat)
                                (shows_prec Zero_nat) [' ', '-', '>', ' '] x))))
                (\ _ ->
                  bindb (catcha (check_NF_terms_eq qa q)
                          (\ x ->
                            Inl (shows_prec_list Zero_nat
                                   ['N', 'F', '(', 'Q', ')', ' ', 'd', 'i', 'f',
                                     'f', 'e', 'r', 's', ' ', 'd', 'u', 'e',
                                     ' ', 't', 'o', ' ', 't', 'e', 'r', 'm',
                                     ' '] .
                                  shows_term (shows_prec_lab Zero_nat)
                                    (shows_prec Zero_nat) x)))
                    (\ _ ->
                      bindb (catcha (check_subseteq ra r)
                              (\ x ->
                                Inl (toomuch
                                      ['s', 't', 'r', 'i', 'c', 't', ' ', 'r',
'u', 'l', 'e']
                                      (shows_rule (shows_prec_lab Zero_nat)
(shows_prec Zero_nat) [' ', '-', '>', ' '] x))))
                        (\ _ ->
                          bindb (catcha (check_subseteq rb rba)
                                  (\ x ->
                                    Inl (toomuch
  ['s', 't', 'r', 'i', 'c', 't', '/', 'w', 'e', 'a', 'k', ' ', 'r', 'u', 'l',
    'e']
  (shows_rule (shows_prec_lab Zero_nat) (shows_prec Zero_nat)
    [' ', '-', '>', ' '] x))))
                            (\ _ ->
                              bindb (catcha (check_subseteq rba rb)
                                      (\ x ->
Inl (missing
      ['s', 't', 'r', 'i', 'c', 't', '/', 'w', 'e', 'a', 'k', ' ', 'r', 'u',
        'l', 'e']
      (shows_rule (shows_prec_lab Zero_nat) (shows_prec Zero_nat)
        [' ', '-', '>', ' '] x))))
                                (\ _ -> Inr ())))))))
    (\ x ->
      Inl (shows_string
             ['f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', ' ', 'o', 'f',
               ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
            shows_nl .
              shows_dpp (shows_prec_lab Zero_nat) (shows_prec Zero_nat) i d .
                shows_nl .
                  shows_string
                    ['m', 'a', 'y', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ', 'c',
                      'o', 'n', 'c', 'l', 'u', 'd', 'e', 'd', ' ', 'f', 'r',
                      'o', 'm', ' ', 'a', 's', 's', 'u', 'm', 'i', 'n', 'g',
                      ' ', 'f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's',
                      ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o',
                      'b', 'l', 'e', 'm'] .
                    shows_nl .
                      shows_dpp (shows_prec_lab Zero_nat) (shows_prec Zero_nat)
                        i (mkc i nfs m p pw q r rw) .
                        shows_nl . x . shows_nl));

icap_impl_dpp_mv ::
  forall a b.
    (Eq b) => Dpp_ops_ext a b [Char] () ->
                a -> [Term b [Char]] -> Term b [Char] -> Term b (Sum () [Char]);
icap_impl_dpp_mv i d =
  let {
    qr = nFQ_subset_NF_rulesb i d;
    qnf = is_QNFc i d;
    r = rulesd i d;
    ic = icap_impl qr qnf (map fst r);
  } in (\ s ->
         let {
           sa = map (map_vars (\ a -> 'x' : a)) s;
           sx = remdups (concatMap vars_term_lista sa);
         } in (\ t -> ic sa (membera sx) (map_vars (\ a -> 'x' : a) t)));

option_to_list :: forall a. Maybe a -> [a];
option_to_list (Just a) = [a];
option_to_list Nothing = [];

mgu_var_disjoint_generic ::
  forall a b c d.
    (Eq b,
      Eq d) => (a -> b) ->
                 (c -> b) ->
                   Term d a -> Term d c -> Maybe (a -> Term d b, c -> Term d b);
mgu_var_disjoint_generic vu wu s t =
  (case mgu (map_vars vu s) (map_vars wu t) of {
    Nothing -> Nothing;
    Just gamma -> Just (map_subst_domain vu gamma, map_subst_domain wu gamma);
  });

nF_subst_impl ::
  forall a b.
    (Eq b) => (Term a b -> Bool) ->
                Bool -> (Term a b, Term a b) -> (b -> Term a b) -> Bool;
nF_subst_impl nf nfs r sigma =
  (if nfs then all (\ x -> nf (sigma x)) (vars_rule_list r) else True);

qnarrows_impl ::
  forall a.
    (Eq a) => (Term a [Char] -> Bool) ->
                Bool ->
                  [(Term a [Char], Term a [Char])] ->
                    Term a [Char] -> [(Term a [Char], [Char] -> Term a [Char])];
qnarrows_impl isnf nfs r t =
  concatMap
    (\ p ->
      let {
        tp = subt_at t p;
      } in (if not (is_Var tp)
             then concatMap
                    (\ (l, ra) ->
                      concatMap
                        (\ (mu_1, mu_2) ->
                          (if nF_subst_impl isnf nfs (l, ra) mu_2
                            then (if all isnf (args (subst_apply_term l mu_2))
                                   then [(ctxt_apply
    (ctxt_of_pos_term p (subst_apply_term t mu_1)) (subst_apply_term ra mu_2),
   mu_1)]
                                   else [])
                            else []))
                        (option_to_list
                          (mgu_var_disjoint_generic (\ a -> 'x' : a)
                            (\ a -> 'y' : a) tp l)))
                    r
             else []))
    (poss_list t);

pairsa :: forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)];
pairsa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = pairs;

wwf_rulesa :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
wwf_rulesa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = wwf_rules;

ur_term_impl ::
  forall a b c.
    (Eq a,
      Eq c) => (Term a [Char] -> Bool) ->
                 (Term a b -> Term a (Sum () [Char])) ->
                   [(Term a [Char], c)] ->
                     ((a, Nat) -> Nat -> Bool) ->
                       [Term a [Char]] -> Term a b -> [(Term a [Char], c)];
ur_term_impl nfq e_cap r pia s (Var x) = [];
ur_term_impl nfq e_cap r pia s (Fun f ts) =
  let {
    n = size_list ts;
    rec = map (ur_term_impl nfq e_cap r pia s) ts;
  } in remdups
         (concatMap (\ (i, urs) -> (if pia (f, n) i then urs else []))
            (zip (upt Zero_nat n) rec) ++
           filter
             (\ (l, _) ->
               (case mgu_class (Fun f (map e_cap ts)) l of {
                 Nothing -> False;
                 Just mu ->
                   all (\ u -> nfq (subst_apply_term u mu))
                     (s ++ map (map_vars (\ a -> 'y' : a)) (args l));
               }))
             r);

usable_rules_calculator ::
  forall a.
    (Eq a) => (Term a [Char] -> Bool) ->
                ([Term a [Char]] -> Term a [Char] -> Term a (Sum () [Char])) ->
                  [(Term a [Char], Term a [Char])] ->
                    ((a, Nat) -> Nat -> Bool) ->
                      (Term a [Char], Term a [Char]) ->
                        [(Term a [Char], Term a [Char])];
usable_rules_calculator nfq e_cap r pia =
  let {
    urt = (\ s ->
            let {
              sa = map (map_vars (\ a -> 'x' : a)) s;
            } in ur_term_impl nfq (e_cap sa) r pia sa);
    urules =
      map (\ (l, ra) -> ((l, ra), urt (args l) (map_vars (\ a -> 'x' : a) ra)))
        r;
    ufun =
      (\ lr -> (case map_of urules lr of {
                 Nothing -> [];
                 Just ur -> ur;
               }));
    rtran = mk_rtrancl_list (\ a b -> a == b) ufun;
  } in (\ (s, t) -> rtran (urt [s] (map_vars (\ a -> 'x' : a) t)));

get_ur_calc_impl_dpp ::
  forall a b.
    (Eq b) => Dpp_ops_ext a b [Char] () ->
                a -> ((b, Nat) -> Nat -> Bool) ->
                       (Term b [Char], Term b [Char]) ->
                         [(Term b [Char], Term b [Char])];
get_ur_calc_impl_dpp i d pia =
  let {
    ic = icap_impl_dpp i d;
    qnf = is_QNFc i d;
    r = rulesd i d;
  } in usable_rules_calculator qnf ic r pia;

inn_usable_rules_pair ::
  forall a b.
    (Eq b) => Dpp_ops_ext a b [Char] () ->
                a -> (Term b [Char], Term b [Char]) ->
                       [(Term b [Char], Term b [Char])];
inn_usable_rules_pair i d =
  let {
    inn = nFQ_subset_NF_rulesb i d;
    r = rulesd i d;
    _ = is_QNFc i d;
    calc = get_ur_calc_impl_dpp i d full_af;
    nfs = nfsc i d;
    wwf = wwf_rulesa i d;
    m = minimal i d;
  } in (\ (s, t) ->
         (if inn &&
               (nfs || all (membera (vars_term_lista s)) (vars_term_lista t)) &&
                 (nfs || m || wwf)
           then calc (s, t) else r));

is_iedg_edge_dpp ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> (Term b [Char], Term b [Char]) ->
                           Term b [Char] -> Bool;
is_iedg_edge_dpp i d =
  let {
    qnf = is_QNFc i d;
    ic = icap_impl_dpp_mv i d;
    _ = rulesd i d;
    urules = inn_usable_rules_pair i d;
  } in (\ (s, t) ->
         let {
           cst = ic [s] t;
           urls = reverse_rules (urules (s, t));
           ica = icap_impla urls [];
         } in (\ u ->
                (case mgu_class cst u of {
                  Nothing -> False;
                  Just mu ->
                    qnf (subst_apply_term (map_vars (\ a -> 'x' : a) s) mu) &&
                      qnf (subst_apply_term (map_vars (\ a -> 'y' : a) u) mu);
                }) &&
                  let {
                    cu = ica [] u;
                  } in (case mgu_class cu t of {
                         Nothing -> False;
                         Just mu ->
                           qnf (subst_apply_term (map_vars (\ a -> 'y' : a) s)
                                 mu);
                       })));

replace_paira ::
  forall a b c d.
    Dpp_ops_ext a b c d ->
      a -> (Term b c, Term b c) -> [(Term b c, Term b c)] -> a;
replace_paira
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = replace_pair;

narrowing_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    (Term b [Char], Term b [Char]) ->
                      Pos ->
                        [(Term b [Char], Term b [Char])] ->
                          a -> Sum ([Char] -> [Char]) a;
narrowing_proc i st p sts dpp =
  (case let {
          (s, t) = st;
          q = qc i dpp;
        } in bindb (check
                     (nFQ_subset_NF_rulesb i dpp || null q && linear_term t)
                     (shows_prec_list Zero_nat
                       ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'o',
                         'r', ' ', 'f', 'u', 'l', 'l', ' ', 'r', 'e', 'w', 'r',
                         'i', 't', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i',
                         'r', 'e', 'd', ' ', '(', 'a', 'n', 'd', ' ', 'l', 'i',
                         'n', 'e', 'a', 'r', 'i', 't', 'y', ' ', 'o', 'f', ' ',
                         't', ' ', 'i', 'n', ' ', 'f', 'u', 'l', 'l', ' ', 'r',
                         'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'c', 'a',
                         's', 'e', ')']))
               (\ _ ->
                 let {
                   ic = icap_impl_dpp_mv i dpp;
                   isnf = is_QNFc i dpp;
                   pairs = pairsa i dpp;
                 } in bindb (check (membera (poss_list (ic [s] t)) p)
                              (shows_prec_list Zero_nat
                                 ['p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
                                   'n', 'o', 't', ' ', 'c', 'o', 'n', 't', 'a',
                                   'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 'c',
                                   'a', 'p', 'p', 'e', 'd', ' ', 't', 'e', 'r',
                                   'm', ' ', 'o', 'f', ' '] .
                                shows_prec_term Zero_nat t))
                        (\ _ ->
                          let {
                            tp = subt_at t p;
                            nfs = nfsc i dpp;
                            narrows = qnarrows_impl isnf nfs (rulesd i dpp) tp;
                            stsa =
                              filter (\ (smu, _) -> isnf smu)
                                (map (\ (ta, mu) ->
                                       (subst_apply_term s mu,
 ctxt_apply (ctxt_of_pos_term p (subst_apply_term t mu)) ta))
                                  narrows);
                          } in bindb (catcha
                                       (forallm
 (\ new ->
   check (any (\ sta ->
                instance_rule new sta && (not nfs || null q || wf_rule sta))
           sts)
     (shows_prec_list Zero_nat
        ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd',
          ' ', 'n', 'a', 'r', 'r', 'o', 'w', 'e', 'd', ' ', 'p', 'a', 'i', 'r',
          ' '] .
       shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
         [' ', '-', '>', ' '] new))
 stsa)
                                       (\ x -> Inl (snd x)))
                                 (\ _ ->
                                   let {
                                     iedg = is_iedg_edge_dpp i dpp (s, t);
                                   } in bindb
  (catcha (check_subseteq (vars_term_lista tp) (vars_term_lista s))
    (\ x ->
      Inl (shows_prec_list Zero_nat
            (['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' '] ++
              x ++ [' ', 'o', 'n', 'l', 'y', ' ', 'o', 'c', 'c', 'u', 'r', 's',
                     ' ', 'o', 'n', ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'p',
                     'a', 'i', 'r']))))
  (\ _ ->
    bindb (check (membera (pa i dpp) st || null (rc i dpp))
            (shows_prec_list Zero_nat
              ['s', 't', 'r', 'i', 'c', 't', ' ', 'D', 'P', ' ', 'o', 'r', ' ',
                'n', 'o', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l',
                'e', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
      (\ _ ->
        (if isnf tp
          then catcha
                 (forallm
                   (\ (u, v) ->
                     bindb (check (membera (poss_list u) p)
                             (shows_prec_list Zero_nat
                                ['p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ',
                                  'n', 'o', 't', ' ', 'c', 'o', 'n', 't', 'a',
                                  'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 'l',
                                  'h', 's', ' ', 'o', 'f', ' ', 'p', 'a', 'i',
                                  'r', ' '] .
                               shows_rule (shows_prec Zero_nat)
                                 (shows_prec_list Zero_nat) [' ', '-', '>', ' ']
                                 (u, v)))
                       (\ _ ->
                         (case mgu_var_disjoint_generic (\ a -> 'x' : a)
                                 (\ a -> 'y' : a) tp (subt_at u p)
                           of {
                           Nothing -> Inr ();
                           Just (mu_1, mu_2) ->
                             check (not (isnf (subst_apply_term s mu_1)) ||
                                     not (isnf (subst_apply_term u mu_2)))
                               (shows_prec_list Zero_nat
                                  ['t', ' ', '|', '_', ' ', 'p', ' ', 'a', 'n',
                                    'd', ' ', 'u', ' ', '|', '_', ' ', 'p', ' ',
                                    'u', 'n', 'i', 'f', 'y', ' ', 'a', 'n', 'd',
                                    ' ', 's', 'a', 't', 'i', 's', 'f', 'y', ' ',
                                    'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ',
                                    'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n',
                                    ' ', 'f', 'o', 'r', ' ', 'p', 'a', 'i', 'r',
                                    ' ', '(', 'u', ',', 'v', ')', ' ', '=',
                                    ' '] .
                                 shows_rule (shows_prec Zero_nat)
                                   (shows_prec_list Zero_nat)
                                   [' ', '-', '>', ' '] (u, v));
                         })))
                   (filter (\ (u, _) -> iedg u) pairs))
                 (\ x -> Inl (snd x))
          else Inr ()))))))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (replace_paira i dpp st sts);
  });

critical_pairs_impl ::
  forall a.
    (Eq a) => [(Term a [Char], Term a [Char])] ->
                [(Term a [Char], Term a [Char])] ->
                  [(Bool, (Term a [Char], Term a [Char]))];
critical_pairs_impl p r =
  concatMap
    (\ (l, ra) ->
      concatMap
        (\ pa ->
          let {
            c = ctxt_of_pos_term pa l;
            la = subt_at l pa;
            b = equal_ctxt c Hole;
          } in (if is_Var la then []
                 else concatMap
                        (\ (laa, rb) ->
                          (case mgu_var_disjoint_generic (\ a -> 'x' : a)
                                  (\ a -> 'y' : a) la laa
                            of {
                            Nothing -> [];
                            Just (sigma, tau) ->
                              [(b, (subst_apply_term ra sigma,
                                     ctxt_apply (subst_apply_ctxt c sigma)
                                       (subst_apply_term rb tau)))];
                          }))
                        r))
        (poss_list l))
    p;

check_prop_rstep_rule ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Bool ->
                    (Term a b -> Sum ([Char] -> [Char]) ()) ->
                      Pos ->
                        (Term a b, Term a b) ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_prop_rstep_rule nfs pa p rule s t =
  bindb (check (in_poss p s)
          (shows_prec_pos Zero_nat p .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'o', 's',
                'i', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
              shows_prec_term Zero_nat s . shows_nl))
    (\ _ ->
      bindb (check (in_poss p t)
              (shows_prec_pos Zero_nat p .
                shows_string
                  [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'o',
                    's', 'i', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                  shows_prec_term Zero_nat t . shows_nl))
        (\ _ ->
          let {
            c = ctxt_of_pos_term p s;
            d = ctxt_of_pos_term p t;
            u = subt_at s p;
            v = subt_at t p;
          } in (case match_list u [(u, fst rule), (v, snd rule)] of {
                 Nothing ->
                   Inl (shows_string
                          ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                         shows_prec_term Zero_nat t .
                           shows_string
                             [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ',
                               'r', 'e', 's', 'u', 'l', 't', ' ', 'f', 'r', 'o',
                               'm', ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r',
                               ' ', 'a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                               'o', 'n', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e',
                               ' '] .
                             shows_nl .
                               shows_rule (shows_prec Zero_nat)
                                 (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                 rule .
                                 shows_string
                                   [' ', 'a', 't', ' ', 'p', 'o', 's', 'i', 't',
                                     'i', 'o', 'n', ' '] .
                                   shows_prec_pos Zero_nat p . shows_nl);
                 Just sigma ->
                   bindb (catcha
                           (forallm pa
                             (args u ++
                               (if nfs then map sigma (vars_rule_list rule)
                                 else [])))
                           (\ x -> Inl (snd x)))
                     (\ _ ->
                       check (equal_ctxt c d)
                         (shows_string
                            ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                           shows_prec_term Zero_nat t .
                             shows_string
                               [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't',
                                 ' ', 'r', 'e', 's', 'u', 'l', 't', ' ', 'f',
                                 'r', 'o', 'm', ' ', 'a', ' ', 'p', 'r', 'o',
                                 'p', 'e', 'r', ' ', 'a', 'p', 'p', 'l', 'i',
                                 'c', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                                 ' ', 'r', 'u', 'l', 'e', ' '] .
                               shows_nl .
                                 shows_rule (shows_prec Zero_nat)
                                   (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                   rule .
                                   shows_string
                                     [' ', 'a', 't', ' ', 'p', 'o', 's', 'i',
                                       't', 'i', 'o', 'n', ' '] .
                                     shows_prec_pos Zero_nat p . shows_nl));
               })));

check_prop_rstepa ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Bool ->
                    (Term a b -> Sum ([Char] -> [Char]) ()) ->
                      [(Term a b, Term a b)] ->
                        Pos ->
                          (Term a b, Term a b) ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_prop_rstepa nfs pa r p rule s t =
  bindb (check (membera r rule)
          (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] rule .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'r', 'u', 'l',
                'e', ' ', 'o', 'f'] .
              shows_nl .
                shows_trs (shows_prec Zero_nat) (shows_prec Zero_nat)
                  ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't',
                    'e', 'm', ':']
                  [' ', '-', '>', ' '] r .
                  shows_nl))
    (\ _ -> check_prop_rstep_rule nfs pa p rule s t);

check_rstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Pos ->
                      (Term a b, Term a b) ->
                        Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rstep = check_prop_rstepa False (\ _ -> Inr ());

show_crit_pair ::
  forall a b. (Showa a, Showa b) => (Term a b, Term a b) -> [Char] -> [Char];
show_crit_pair lr =
  shows_prec_list Zero_nat ['('] .
    shows_prec_term Zero_nat (fst lr) .
      shows_prec_list Zero_nat [',', ' '] .
        shows_prec_term Zero_nat (snd lr) . shows_prec_list Zero_nat [')'];

critical_pairs_top_impl ::
  forall a.
    (Eq a) => [(Term a [Char], Term a [Char])] ->
                [(Term a [Char], Term a [Char])] ->
                  [(Term a [Char], Term a [Char])];
critical_pairs_top_impl p r =
  concatMap
    (\ (l, ra) ->
      (if is_Var l then []
        else concatMap
               (\ (la, rb) ->
                 (case mgu_var_disjoint_generic (\ a -> 'x' : a)
                         (\ a -> 'y' : a) l la
                   of {
                   Nothing -> [];
                   Just (sigma, tau) ->
                     [(subst_apply_term ra sigma, subst_apply_term rb tau)];
                 }))
               r))
    p;

check_critical_pairs_innermost ::
  forall a.
    (Eq a,
      Showa a) => [(Term a [Char], Term a [Char])] -> Sum ([Char] -> [Char]) ();
check_critical_pairs_innermost r =
  catcha
    (forallm
      (\ (l, ra) ->
        check (equal_term l ra)
          (shows_prec_list Zero_nat
             ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'a', ' ', 'n', 'o',
               'n', '-', 't', 'r', 'i', 'v', 'i', 'a', 'l', ' ', 'c', 'r', 'i',
               't', 'i', 'c', 'a', 'l', ' ', 'p', 'a', 'i', 'r', ' '] .
            show_crit_pair (l, ra)))
      (critical_pairs_top_impl r r))
    (\ x -> Inl (snd x));

is_ur_closed_term_impl ::
  forall a b c.
    (Eq a,
      Eq c) => (Term a [Char] -> Bool) ->
                 (Term a b -> Term a (Sum () [Char])) ->
                   [(Term a [Char], c)] ->
                     [(Term a [Char], c)] ->
                       [Term a [Char]] -> Term a b -> Bool;
is_ur_closed_term_impl nfq e_cap r u s (Var x) = True;
is_ur_closed_term_impl nfq e_cap r u s (Fun f ts) =
  all (is_ur_closed_term_impl nfq e_cap r u s) ts &&
    all (\ (l, ra) ->
          membera u (l, ra) ||
            (case mgu_class (Fun f (map e_cap ts)) l of {
              Nothing -> True;
              Just mu ->
                not (all (\ ua ->
                           nfq (subst_apply_term (map_vars (\ a -> 'y' : a) ua)
                                 mu))
                      (args l)) ||
                  not (all (\ ua -> nfq (subst_apply_term ua mu)) s);
            }))
      r;

is_ur_closed_impl_dpp_mv ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> [(Term b [Char], Term b [Char])] ->
                           [Term b [Char]] -> Term b [Char] -> Bool;
is_ur_closed_impl_dpp_mv i d =
  let {
    ic = icap_impl_dpp i d;
    qnf = is_QNFc i d;
    r = rulesd i d;
    urc = (\ s -> is_ur_closed_term_impl qnf (ic s) r);
  } in (\ u s ->
         let {
           sa = map (map_vars (\ a -> 'x' : a)) s;
         } in (\ t -> urc sa u sa (map_vars (\ a -> 'x' : a) t)));

rewriting_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    [(Term b [Char], Term b [Char])] ->
                      (Term b [Char], Term b [Char]) ->
                        (Term b [Char], Term b [Char]) ->
                          (Term b [Char], Term b [Char]) ->
                            (Term b [Char], Term b [Char]) ->
                              Pos -> a -> Sum ([Char] -> [Char]) a;
rewriting_proc i u stb sta st lr p dpp =
  (case bindb (check (nFQ_subset_NF_rulesb i dpp)
                (shows_prec_list Zero_nat
                  ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r', 'e',
                    'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u',
                    'i', 'r', 'e', 'd']))
          (\ _ ->
            bindb (check (eq_rule_mod_vars sta st)
                    (shows_prec_list Zero_nat
                       ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
                      shows_rule (shows_prec Zero_nat)
                        (shows_prec_list Zero_nat) [' ', '-', '>', ' '] sta .
                        shows_prec_list Zero_nat
                          [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                            'r', 'e', 'n', 'a', 'm', 'e', 'd', ' ', 'v', 'a',
                            'r', 'i', 'a', 'n', 't', ' ', 'o', 'f', ' '] .
                          shows_rule (shows_prec Zero_nat)
                            (shows_prec_list Zero_nat) [' ', '-', '>', ' '] st))
              (\ _ ->
                let {
                  r = rulesd i dpp;
                  s = fst stb;
                  t = snd stb;
                  ta = snd sta;
                } in bindb (check (equal_term s (fst sta))
                             (shows_prec_list Zero_nat
                               ['l', 'e', 'f', 't', '-', 'h', 'a', 'n', 'd',
                                 ' ', 's', 'i', 'd', 'e', 's', ' ', 'o', 'f',
                                 ' ', 'o', 'l', 'd', ' ', 'a', 'n', 'd', ' ',
                                 'n', 'e', 'w', ' ', 'p', 'a', 'i', 'r', ' ',
                                 'd', 'i', 'f', 'f', 'e', 'r']))
                       (\ _ ->
                         bindb (check_rstep r p lr t ta)
                           (\ _ ->
                             bindb (catcha (check_subseteq u r)
                                     (\ x ->
                                       Inl
 (shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] x .
   shows_prec_list Zero_nat
     [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'r', 'u', 'l', 'e', ' ',
       'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
       ' ', 's', 'y', 's', 't', 'e', 'm', ' '])))
                               (\ _ ->
                                 bindb (check
 (membera (pa i dpp) stb || null (rc i dpp))
 (shows_prec_list Zero_nat
   ['s', 't', 'r', 'i', 'c', 't', ' ', 'D', 'P', ' ', 'o', 'r', ' ', 'n', 'o',
     ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'r',
     'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                                   (\ _ ->
                                     let {
                                       tp = subt_at t p;
                                       urc = is_ur_closed_impl_dpp_mv i dpp u;
                                       check_urc =
 (\ sa tb ->
   check (urc sa tb)
     (shows_prec_list Zero_nat ['t', 'e', 'r', 'm', ' '] .
       shows_prec_term Zero_nat tb .
         shows_prec_list Zero_nat
           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'l', 'o', 's', 'e',
             'd', ' ', 'u', 'n', 'd', 'e', 'r', ' ', 'u', 's', 'a', 'b', 'l',
             'e', ' ', 'r', 'u', 'l', 'e', 's']));
                                       nfs = nfsc i dpp;
                                     } in bindb
    (catcha
      (check_subseteq (vars_term_lista (snd lr)) (vars_term_lista (fst lr)))
      (\ _ ->
        Inl (shows_prec_list Zero_nat
              ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n', 'd',
                'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e',
                ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 'd'])))
    (\ _ ->
      bindb (if nfs then Inr ()
              else catcha
                     (check_subseteq (vars_term_lista tp) (vars_term_lista s))
                     (\ _ ->
                       Inl (shows_prec_list Zero_nat
                             ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c',
                               'o', 'n', 'd', 'i', 't', 'i', 'o', 'n', ' ', 'i',
                               'n', ' ', 'p', 'a', 'i', 'r', ' ', 'v', 'i', 'o',
                               'l', 'a', 't', 'e', 'd'])))
        (\ _ ->
          bindb (check (nfs || wwf_rulesa i dpp)
                  (shows_prec_list Zero_nat
                    ['w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd', ' ',
                      'r', 'u', 'l', 'e', 's', ' ', 'o', 'r', ' ', 'n', 'o',
                      'r', 'm', 'a', 'l', ' ', 's', 'u', 'b', 's', 't', '.',
                      ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
            (\ _ ->
              bindb (catcha
                      (forallm
                        (\ (l, _) ->
                          check (not (is_Var l))
                            (shows_prec_list Zero_nat
                              ['l', 'h', 's', 's', ' ', 'm', 'u', 's', 't', ' ',
                                'n', 'o', 't', ' ', 'b', 'e', ' ', 'v', 'a',
                                'r', 'i', 'a', 'b', 'l', 'e', 's']))
                        u)
                      (\ x -> Inl (snd x)))
                (\ _ ->
                  bindb (check_urc [s] tp)
                    (\ _ ->
                      bindb (catcha
                              (forallm (\ (l, a) -> check_urc (args l) a) u)
                              (\ x -> Inl (snd x)))
                        (\ _ ->
                          bindb (catcha (check_critical_pairs_innermost u)
                                  (\ x ->
                                    Inl (shows_prec_list Zero_nat
   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 's', 'h', 'o', 'w',
     'i', 'n', 'g', ' ', 'U', 'N', 'F', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b',
     'l', 'e', ' ', 'r', 'u', 'l', 'e', 's'] .
  shows_nl . x)))
                            (\ _ ->
                              catcha
                                (forallm
                                  (\ (_, (sa, tb)) ->
                                    check (equal_term sa tb)
                                      (shows_prec_list Zero_nat
['n', 'o', 'n', '-', 't', 'r', 'i', 'v', 'i', 'a', 'l', ' ', 'c', 'r', 'i', 't',
  'i', 'c', 'a', 'l', ' ', 'p', 'a', 'i', 'r', ' ', 'b', 'e', 't', 'w', 'e',
  'e', 'n', ' ', 'r', 'u', 'l', 'e', ' ', 't', 'o', ' ', 'r', 'e', 'w', 'r',
  'i', 't', 'e', ' ', 'a', 'n', 'd', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
  'r', 'u', 'l', 'e', 's']))
                                  (critical_pairs_impl [lr] u))
                                (\ x -> Inl (snd x)))))))))))))))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (replace_paira i dpp stb [st]);
  });

only_eta_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
only_eta_rules e r_eta =
  catcha
    (forallm
      (\ (l, r) ->
        check (case (l, r) of {
                (Var _, b) -> False;
                (Fun f ls, Var _) -> False;
                (Fun f ls, Fun g rs) ->
                  f == g &&
                    equal_nat (size_list ls) (nat 2) &&
                      equal_nat (size_list rs) (nat 2) &&
                        equal_term (nth ls one_nat) (nth rs one_nat) &&
                          any (\ (la, ra) ->
                                instance_rule (hd ls, hd rs) (la, ra))
                            r_eta;
              })
          (shows_prec_list Zero_nat ['r', 'u', 'l', 'e', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] (l, r) .
              shows_prec_list Zero_nat
                [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', '(',
                  'e', 't', 'a', '-', 'e', 'x', 'p', 'a', 'n', 'd', 'e', 'd',
                  ')', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                  'u', 'l', 'e']))
      e)
    (\ x -> Inl (snd x));

uncurry_proc ::
  forall a b.
    (Eq b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    (b, ([((b, Nat), [b])],
                          ([(Term b [Char], Term b [Char])],
                            [(Term b [Char], Term b [Char])]))) ->
                      ([((b, Nat), [b])] -> b -> Nat -> b) ->
                        (b -> Nat ->
                                [((b, Nat), [b])] ->
                                  Sum ([Char] -> [Char]) ()) ->
                          [(Term b [Char], Term b [Char])] ->
                            [(Term b [Char], Term b [Char])] ->
                              a -> Sum ([Char] -> [Char]) a;
uncurry_proc i info fmapa check_inj p r dpp =
  let {
    (a, (sml, (u, eb))) = info;
    paa = pa i dpp;
    pw = pwa i dpp;
    ra = rc i dpp;
    rw = rwc i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    (e, ew) = uncurry_eta_split eb ra;
    sm = sig_list_to_sig_map a sml fmapa;
    uP = uncurry_rules a sm paa;
    uPw = uncurry_rules a sm pw;
    r_eta = e ++ ra;
    rw_eta = ew ++ rw;
    uR = uncurry_rules a sm r_eta;
    uRw = uncurry_rules a sm rw_eta;
  } in (case let {
               s = uncurry_of_sig_list a sml sm;
             } in bindb (check (null (qc i dpp))
                          (shows_prec_list Zero_nat
                            ['s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'n',
                              'o', 't', ' ', 's', 'u', 'p', 'p', 'o', 'r', 't',
                              'e', 'd', ' ', 'f', 'o', 'r', ' ', 'u', 'n', 'c',
                              'u', 'r', 'r', 'y', 'i', 'n', 'g']))
                    (\ _ ->
                      bindb (only_eta_rules e r_eta)
                        (\ _ ->
                          bindb (only_eta_rules ew rw_eta)
                            (\ _ ->
                              bindb (check_inj a (nat 2) sml)
                                (\ _ ->
                                  bindb (catcha
  (forallm
    (\ (l, _) ->
      check (not (is_Var l))
        (shows_prec_list Zero_nat
          ['l', 'h', 's', ' ', 'a', 's', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
            'e', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
            'w', 'e', 'd']))
    (ra ++ rw))
  (\ x -> Inl (snd x)))
                                    (\ _ ->
                                      bindb
(catcha
  (forallm
    (\ (l, _) ->
      check (hvf_term a l)
        (shows_prec_list Zero_nat
           ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
             ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
          shows_term (shows_prec Zero_nat) (shows_prec_list Zero_nat) l .
            shows_prec_list Zero_nat
              [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']))
    paa)
  (\ x -> Inl (snd x)))
(\ _ ->
  bindb (catcha
          (forallm
            (\ (l, _) ->
              check (hvf_term a l)
                (shows_prec_list Zero_nat
                   ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                     'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                  shows_term (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                    l .
                    shows_prec_list Zero_nat
                      [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                        'd']))
            pw)
          (\ x -> Inl (snd x)))
    (\ _ ->
      bindb (catcha
              (forallm
                (\ (l, _) ->
                  check (hvf_term a l)
                    (shows_prec_list Zero_nat
                       ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b',
                         'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                      shows_term (shows_prec Zero_nat)
                        (shows_prec_list Zero_nat) l .
                        shows_prec_list Zero_nat
                          [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                            'e', 'd']))
                r_eta)
              (\ x -> Inl (snd x)))
        (\ _ ->
          bindb (catcha
                  (forallm
                    (\ (l, _) ->
                      check (hvf_term a l)
                        (shows_prec_list Zero_nat
                           ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a',
                             'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's',
                             ' '] .
                          shows_term (shows_prec Zero_nat)
                            (shows_prec_list Zero_nat) l .
                            shows_prec_list Zero_nat
                              [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                                'e', 'd']))
                    rw_eta)
                  (\ x -> Inl (snd x)))
            (\ _ ->
              bindb (eta_closed_rules a sm r_eta r_eta)
                (\ _ ->
                  bindb (eta_closed_rules a sm rw_eta rw_eta)
                    (\ _ ->
                      bindb (catcha (check_subseteq uP p)
                              (\ x ->
                                Inl (shows_prec_list Zero_nat
                                       ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e',
 'd', ' ', 'p', 'a', 'i', 'r', ' '] .
                                      shows_rule (shows_prec Zero_nat)
(shows_prec_list Zero_nat) [' ', '-', '>', ' '] x .
shows_prec_list Zero_nat
  [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                        (\ _ ->
                          bindb (catcha (check_subseteq uPw p)
                                  (\ x ->
                                    Inl (shows_prec_list Zero_nat
   ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a', 'i', 'r', ' '] .
  shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] x .
    shows_prec_list Zero_nat
      [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                            (\ _ ->
                              bindb (catcha (check_subseteq uR r)
                                      (\ x ->
Inl (shows_prec_list Zero_nat
       ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
         ' '] .
      shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
        [' ', '-', '>', ' '] x .
        shows_prec_list Zero_nat
          [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                (\ _ ->
                                  bindb (catcha (check_subseteq uRw r)
  (\ x ->
    Inl (shows_prec_list Zero_nat
           ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l',
             'e', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] x .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                    (\ _ ->
                                      bindb
(catcha (check_CS_subseteq s u)
  (\ x ->
    Inl (shows_prec_list Zero_nat
           ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] x .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
(\ _ ->
  bindb (catcha (check_CS_subseteq u s)
          (\ x ->
            Inl (shows_prec_list Zero_nat ['r', 'u', 'l', 'e', ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                    [' ', '-', '>', ' '] x .
                    shows_prec_list Zero_nat
                      [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ',
                        'u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l',
                        'e'])))
    (\ _ ->
      catcha (check_subseteq u r)
        (\ x ->
          Inl (shows_prec_list Zero_nat
                 ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e',
                   ' '] .
                shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                  [' ', '-', '>', ' '] x .
                  shows_prec_list Zero_nat
                    [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ',
                      'i', 'n', ' ', 'n', 'e', 'w', ' ', 'T', 'R',
                      'S'])))))))))))))))))))
         of {
         Inl aa -> Inl aa;
         Inr _ -> Inr (mkc i nfs m uP uPw [] uR (uRw ++ u));
       });

equal_gctxt :: forall a b. (Eq a) => Gctxt a b -> Gctxt a b -> Bool;
equal_gctxt (GCFun f list) GCHole = False;
equal_gctxt GCHole (GCFun f list) = False;
equal_gctxt (GCFun fa lista) (GCFun f list) = fa == f && lista == list;
equal_gctxt GCHole GCHole = True;

instance (Eq a) => Eq (Gctxt a b) where {
  a == b = equal_gctxt a b;
};

uncurry_top_rules ::
  forall a b.
    (Eq a) => a -> Nat ->
                     (a -> Nat -> [a]) ->
                       [(Term a b, Term a b)] -> [(Term a b, Term a b)];
uncurry_top_rules a n sm =
  map (\ (l, r) -> (uncurry_top a n sm l, uncurry_top a n sm r));

eta_closed_top_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> Nat ->
                         (a -> Nat -> [a]) ->
                           [(Term a b, Term a b)] ->
                             [(Term a b, Term a b)] ->
                               Sum ([Char] -> [Char]) ();
eta_closed_top_rules a n sm r p =
  catcha
    (forallm
      (\ (l, ra) ->
        (case l of {
          Var _ -> Inr ();
          Fun ff ls ->
            check (equal_nat (aarity sm ff (size_list ls)) Zero_nat ||
                    any (\ (lll, rrr) ->
                          (case (lll, rrr) of {
                            (Var _, b) -> False;
                            (Fun f [], b) -> False;
                            (Fun f (ll : yy), Var _) -> False;
                            (Fun f (ll : yy), Fun g []) -> False;
                            (Fun f (ll : yy), Fun g (rr : zz)) ->
                              f == a &&
                                g == a &&
                                  zz == yy &&
                                    equal_nat (size_list yy)
                                      (minus_nat n one_nat) &&
                                      distinct yy &&
all is_Var yy &&
  null (list_inter (map the_Var yy) (vars_rule_list (ll, rr))) &&
    instance_rule (l, ra) (ll, rr);
                          }))
                      p)
              (shows_prec_list Zero_nat
                 ['e', 't', 'a', ' ', 'e', 'x', 'p', 'a', 'n', 's', 'i', 'o',
                   'n', ' ', 'o', 'f', ' '] .
                shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                  [' ', '-', '>', ' '] (l, ra) .
                  shows_prec_list Zero_nat
                    [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);
        }))
      r)
    (\ x -> Inl (snd x));

uncurry_of_top_sig_list ::
  forall a.
    a -> Nat ->
           [((a, Nat), [a])] ->
             (a -> Nat -> [a]) -> [(Term a [Char], Term a [Char])];
uncurry_of_top_sig_list a m sml sm =
  concatMap
    (\ (b, c) ->
      let {
        (f, n) = b;
      } in (\ _ ->
             let {
               g = get_symbol sm f n;
             } in map (\ i ->
                        (Fun a (generate_f_xs (g i) (plus_nat n i) :
                                 map (\ ia -> Var (generate_var ia))
                                   (upt (plus_nat n i)
                                     (plus_nat (plus_nat n i)
                                       (minus_nat m one_nat)))),
                          generate_f_xs (g (Suc i))
                            (plus_nat (plus_nat n i) (minus_nat m one_nat))))
                    (upt Zero_nat (aarity sm f n)))
        c)
    sml;

uncurry_top_proc ::
  forall a b.
    (Eq b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    (b, ([((b, Nat), [b])],
                          ([(Term b [Char], Term b [Char])],
                            [(Term b [Char], Term b [Char])]))) ->
                      Nat ->
                        ([((b, Nat), [b])] -> b -> Nat -> b) ->
                          (b -> Nat ->
                                  [((b, Nat), [b])] ->
                                    Sum ([Char] -> [Char]) ()) ->
                            [(Term b [Char], Term b [Char])] ->
                              [(Term b [Char], Term b [Char])] ->
                                a -> Sum ([Char] -> [Char]) a;
uncurry_top_proc i info n fmapa check_inj p r dpp =
  let {
    (a, (sml, (u, eb))) = info;
    paa = pa i dpp;
    pw = pwa i dpp;
    ra = rc i dpp;
    rw = rwc i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    (e, ew) = uncurry_eta_split eb ra;
    sm = sig_list_to_sig_map a sml fmapa;
    p_eta = e ++ paa;
    pw_eta = ew ++ pw;
    uP = uncurry_top_rules a n sm p_eta;
    uPw = uncurry_top_rules a n sm pw_eta;
    uR = map_rules_wa (\ f na -> get_symbol sm f na Zero_nat) ra;
    uRw = map_rules_wa (\ f na -> get_symbol sm f na Zero_nat) rw;
  } in (case bindb (check (null (qc i dpp))
                     (shows_prec_list Zero_nat
                       ['s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'c', 'u',
                         'r', 'r', 'e', 'n', 't', 'l', 'y', ' ', 'u', 'n', 's',
                         'u', 'p', 'p', 'o', 'r', 't', 'e', 'd']))
               (\ _ ->
                 bindb (check (not (equal_nat n Zero_nat))
                         (shows_prec_list Zero_nat
                           ['t', 'h', 'e', ' ', 'a', 'r', 'i', 't', 'y', ' ',
                             'o', 'f', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c',
                             'u', 'r', 'r', 'i', 'e', 'd', ' ', 's', 'y', 'm',
                             'b', 'o', 'l', ' ', 'm', 'u', 's', 't', ' ', 'b',
                             'e', ' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't',
                             ' ', '1']))
                   (\ _ ->
                     bindb (check_inj a n sml)
                       (\ _ ->
                         let {
                           pb = pairsa i dpp;
                           is_def = (\ fn -> not (null (rules_mapb i dpp fn)));
                           rm = rules_mapb i dpp;
                         } in bindb (catcha
                                      (forallm
(\ (l, _) ->
  check (not (is_Var l))
    (shows_prec_list Zero_nat
      ['l', 'h', 's', ' ', 'a', 's', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
        'e', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
        'e', 'd']))
(ra ++ rw))
                                      (\ x -> Inl (snd x)))
                                (\ _ ->
                                  bindb (catcha
  (forallm
    (\ (l, rb) ->
      bindb (check (hvf_top a n l)
              (shows_prec_list Zero_nat
                 ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
                   'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                shows_term (shows_prec Zero_nat) (shows_prec_list Zero_nat) l .
                  shows_prec_list Zero_nat
                    [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
                      'd']))
        (\ _ -> check_no_var rb))
    (pw_eta ++ p_eta))
  (\ x -> Inl (snd x)))
                                    (\ _ ->
                                      bindb
(catcha
  (forallm
    (\ (_, rb) ->
      check (not (is_def (the (root rb))))
        (shows_prec_list Zero_nat ['r', 'o', 'o', 't', ' ', 'o', 'f', ' '] .
          shows_term (shows_prec Zero_nat) (shows_prec_list Zero_nat) rb .
            shows_prec_list Zero_nat
              [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ',
                'd', 'e', 'f', 'i', 'n', 'e', 'd']))
    pb)
  (\ x -> Inl (snd x)))
(\ _ ->
  bindb (check (not (is_def (a, n)))
          (shows_prec_list Zero_nat
             ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 's',
               'y', 'm', 'b', 'o', 'l', ' '] .
            shows_prec Zero_nat a .
              shows_prec_list Zero_nat
                [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b', 'e',
                  ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ',
                  'R']))
    (\ _ ->
      bindb (if any (\ (_, rb) ->
                      the (root rb) == (a, n) &&
                        equal_gctxt (capRM2 rm (hd (args rb))) GCHole)
                  pb
              then bindb (catcha
                           (check_CS_subseteq
                             (uncurry_of_top_sig_list a n sml sm) u)
                           (\ x ->
                             Inl (shows_prec_list Zero_nat
                                    ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'i',
                                      'n', 'g', ' ', 'p', 'a', 'i', 'r', ' '] .
                                   shows_rule (shows_prec Zero_nat)
                                     (shows_prec_list Zero_nat)
                                     [' ', '-', '>', ' '] x .
                                     shows_prec_list Zero_nat
                                       [' ', 'i', 's', ' ', 'm', 'i', 's', 's',
 'i', 'n', 'g', ' ', 'i', 'n'] .
                                       shows_nl .
 shows_rules (shows_prec Zero_nat) (shows_prec_list Zero_nat)
   [' ', '-', '>', ' '] u)))
                     (\ _ ->
                       bindb (eta_closed_top_rules a n sm ra p_eta)
                         (\ _ -> eta_closed_top_rules a n sm rw pw_eta))
              else Inr ())
        (\ _ ->
          bindb (catcha (check_subseteq uP p)
                  (\ x ->
                    Inl (shows_prec_list Zero_nat
                           ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ',
                             'p', 'a', 'i', 'r', ' '] .
                          shows_rule (shows_prec Zero_nat)
                            (shows_prec_list Zero_nat) [' ', '-', '>', ' '] x .
                            shows_prec_list Zero_nat
                              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                                'g'])))
            (\ _ ->
              bindb (catcha (check_subseteq uPw p)
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd',
                                 ' ', 'p', 'a', 'i', 'r', ' '] .
                              shows_rule (shows_prec Zero_nat)
                                (shows_prec_list Zero_nat) [' ', '-', '>', ' ']
                                x .
                                shows_prec_list Zero_nat
                                  [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                    'n', 'g'])))
                (\ _ ->
                  bindb (catcha (check_subseteq u p)
                          (\ x ->
                            Inl (shows_prec_list Zero_nat
                                   ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n',
                                     'g', ' ', 'p', 'a', 'i', 'r', ' '] .
                                  shows_rule (shows_prec Zero_nat)
                                    (shows_prec_list Zero_nat)
                                    [' ', '-', '>', ' '] x .
                                    shows_prec_list Zero_nat
                                      [' ', 'i', 's', ' ', 'm', 'i', 's', 's',
'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n', 'e', 'w', ' ', 'p', 'a', 'i', 'r',
's'])))
                    (\ _ ->
                      bindb (catcha (check_subseteq uR r)
                              (\ x ->
                                Inl (shows_prec_list Zero_nat
                                       ['r', 'u', 'l', 'e', ' '] .
                                      shows_rule (shows_prec Zero_nat)
(shows_prec_list Zero_nat) [' ', '-', '>', ' '] x .
shows_prec_list Zero_nat
  [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ',
    'n', 'e', 'w', ' ', 'r', 'u', 'l', 'e', 's'])))
                        (\ _ ->
                          catcha (check_subseteq uRw r)
                            (\ x ->
                              Inl (shows_prec_list Zero_nat
                                     ['r', 'u', 'l', 'e', ' '] .
                                    shows_rule (shows_prec Zero_nat)
                                      (shows_prec_list Zero_nat)
                                      [' ', '-', '>', ' '] x .
                                      shows_prec_list Zero_nat
[' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ', 'n',
  'e', 'w', ' ', 'r', 'u', 'l', 'e', 's']))))))))))))))
         of {
         Inl aa -> Inl aa;
         Inr _ -> Inr (mkc i nfs m uP (uPw ++ u) [] uR uRw);
       });

uncurry_proc_both ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a (Lab b c) [Char] () ->
                    Maybe Nat ->
                      (Lab b c,
                        ([((Lab b c, Nat), [Lab b c])],
                          ([(Term (Lab b c) [Char], Term (Lab b c) [Char])],
                            [(Term (Lab b c) [Char],
                               Term (Lab b c) [Char])]))) ->
                        [(Term (Lab b c) [Char], Term (Lab b c) [Char])] ->
                          [(Term (Lab b c) [Char], Term (Lab b c) [Char])] ->
                            a -> Sum ([Char] -> [Char]) a;
uncurry_proc_both i Nothing (a, (sml, (u, eb))) =
  uncurry_proc i (a, (sml, (u, eb))) (fmapa a (nat 2)) check_inj;
uncurry_proc_both i (Just n) (a, (sml, (u, eb))) =
  uncurry_top_proc i (a, (sml, (u, eb))) n (fmapa a n) check_inj;

q_reduction_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    [Term b c] -> a -> Sum ([Char] -> [Char]) a;
q_reduction_proc i q dpp =
  let {
    pb = pairsa i dpp;
    rb = rulesd i dpp;
    f = map Just (funas_trs_list (pb ++ rb));
    qa = qc i dpp;
    isnf = is_QNFc i dpp;
    qq = filter (\ qb -> all isnf (args qb)) qa;
    rQ = filter (\ qb -> membera f (root qb)) qq;
  } in (case catcha
               (bindb
                 (check (nFQ_subset_NF_rulesb i dpp)
                   (shows_prec_list Zero_nat
                     ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r',
                       'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e',
                       'q', 'u', 'i', 'r', 'e', 'd']))
                 (\ _ ->
                   bindb (check_wf_trs rb)
                     (\ _ ->
                       bindb (catcha (check_NF_terms_subset q rQ)
                               (\ x ->
                                 Inl (shows_prec_list Zero_nat
['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                                       shows_prec_term Zero_nat x .
 shows_prec_list Zero_nat
   [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i', 'n', ' ',
     'Q', '\'', ' '])))
                         (\ _ ->
                           bindb (catcha (check_NF_terms_subset qq q)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
   shows_prec_term Zero_nat x .
     shows_prec_list Zero_nat
       [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e',
         'd', ' ', 'i', 'n', ' ', 'Q', '\'', ' '])))
                             (\ _ ->
                               catcha
                                 (forallm
                                   (\ lr ->
                                     catcha
                                       (check_subseteq
 (vars_term_lista (snd lr)) (vars_term_lista (fst lr)))
                                       (\ _ ->
 Inl (shows_prec_list Zero_nat
       ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n', 'd', 'i',
         't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 'r', 'u', 'l', 'e', ' ', 'v',
         'i', 'o', 'l', 'a', 't', 'e', 'd'])))
                                   pb)
                                 (\ x -> Inl (snd x)))))))
               (\ x ->
                 Inl (shows_prec_list Zero_nat
                        ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                          'n', ' ', 'r', 'e', 'd', 'u', 'c', 'i', 'n', 'g', ' ',
                          'Q', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                          ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                       shows_nl .
                         shows_dpp (shows_prec Zero_nat) (shows_prec Zero_nat) i
                           dpp .
                           shows_nl .
                             shows_prec_list Zero_nat
                               ['t', 'o', ' ', 't', 'h', 'e', ' ', 's', 'e',
                                 't', ' '] .
                               shows_nl .
                                 shows_terms (shows_prec Zero_nat)
                                   (shows_prec Zero_nat) ['Q', '\'', ':'] q .
                                   shows_nl . x))
         of {
         Inl a -> Inl a;
         Inr _ ->
           Inr (mkc i (nfsc i dpp) (minimal i dpp) (pa i dpp) (pwa i dpp) q
                 (rc i dpp) (rwc i dpp));
       });

funas_args_term_list :: forall a b. Term a b -> [(a, Nat)];
funas_args_term_list t = concatMap funas_term_list (args t);

funas_args_rule_list :: forall a b. (Term a b, Term a b) -> [(a, Nat)];
funas_args_rule_list r =
  funas_args_term_list (fst r) ++ funas_args_term_list (snd r);

funas_args_trs_list ::
  forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_args_trs_list r = remdups (concatMap funas_args_rule_list r);

check_rule_reflecting ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [Ctxt a b] ->
                    [(Term a b, Term a b)] ->
                      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_rule_reflecting fcs rs rule =
  check (any (\ (l, r) ->
               any (\ c ->
                     equal_term (fst rule) (ctxt_apply c l) &&
                       equal_term (snd rule) (ctxt_apply c r))
                 (Hole : fcs))
          rs)
    (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
      shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
        [' ', '-', '>', ' '] rule .
        shows_string
          [' ', 'i', 's', ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
            'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 't', 'h',
            'e', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 's', 'e',
            't', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'n', 'o',
            'r', ' ', 'o', 'b', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'b', 'y',
            ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ', 'a', ' ', 'f',
            'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
          shows_nl);

block_term :: forall a b. a -> Term a b -> Term a b;
block_term f (Var x) = Var x;
block_term f (Fun g ts) = Fun g (map (\ t -> Fun f [t]) ts);

block_rule :: forall a b. a -> (Term a b, Term a b) -> (Term a b, Term a b);
block_rule f r = (block_term f (fst r), block_term f (snd r));

check_superset_of_blocked ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> [(Term a b, Term a b)] ->
                         [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_superset_of_blocked f pa p =
  catcha
    (catcha
      (catcha
        (forallm
          (\ x -> (if membera pa (block_rule f x) then Inr () else Inl x)) p)
        (\ x -> Inl (snd x)))
      (\ x ->
        Inl (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '-', '>', ' '] (block_rule f x) .
                shows_string
                  [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'] .
                  shows_nl)))
    (\ x ->
      Inl (shows_trs (shows_prec Zero_nat) (shows_prec Zero_nat)
             ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e',
               'm', ':']
             [' ', '-', '>', ' '] p .
            shows_string
              ['i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 's',
                'e', 't', ' ', 'o', 'f'] .
              shows_trs (shows_prec Zero_nat) (shows_prec Zero_nat)
                ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't',
                  'e', 'm', ':']
                [' ', '-', '>', ' '] pa .
                x . shows_nl));

check_no_defined_root_defs ::
  forall a b.
    (Eq a, Showa a,
      Showa b) => [(a, Nat)] -> Term a b -> Sum ([Char] -> [Char]) ();
check_no_defined_root_defs f t =
  check (not (membera f (the (root t))))
    (shows_string ['t', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' '] .
      shows_prec_term Zero_nat t .
        shows_string [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd']);

fcc_proc_cond ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    b -> [Ctxt b c] ->
                           [(Term b c, Term b c)] ->
                             [(Term b c, Term b c)] ->
                               [(Term b c, Term b c)] ->
                                 [(Term b c, Term b c)] ->
                                   a -> Sum ([Char] -> [Char]) a;
fcc_proc_cond i f fcs p pw r rw dpp =
  let {
    paa = pa i dpp;
    pwaa = pwa i dpp;
    ra = rc i dpp;
    rwa = rwc i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    new_dpp = mkc i nfs m p pw [] r rw;
  } in (case catcha
               (let {
                  pb = list_union paa pwaa;
                  rb = list_union ra rwa;
                  rba = list_union r rw;
                  fa = (f, one_nat);
                  cf = More f [] Hole [];
                  fcsa = cf : fcs;
                  vs = vars_trs_list rb;
                  fs = list_union (funas_trs_list rb) (funas_args_trs_list pb);
                  fas = fa : fs;
                  ds = defs_list rb;
                } in bindb (check (not (membera ds fa))
                             (shows_prec Zero_nat f .
                               shows_string
                                 ['i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'r',
                                   'e', 's', 'h'] .
                                 shows_nl))
                       (\ _ ->
                         bindb (check_wf_trs rb)
                           (\ _ ->
                             bindb (catcha
                                     (forallm
                                       (\ rd ->
 bindb (check_no_var (fst rd))
   (\ _ ->
     bindb (check_no_var (snd rd))
       (\ _ -> check_no_defined_root_defs ds (snd rd))))
                                       pb)
                                     (\ x -> Inl (snd x)))
                               (\ _ ->
                                 bindb (catcha
 (forallm (check_flat_ctxt vs) fcsa) (\ x -> Inl (snd x)))
                                   (\ _ ->
                                     bindb (catcha
     (forallm (check_is_flat_ctxt vs fas) fcsa) (\ x -> Inl (snd x)))
                                       (\ _ ->
 bindb (catcha (forallm (check_flat_ctxt_complete fcsa) fas)
         (\ x -> Inl (snd x)))
   (\ _ ->
     bindb (catcha (forallm (check_rule_preserving fcsa r) ra)
             (\ x -> Inl (snd x)))
       (\ _ ->
         bindb (catcha (forallm (check_rule_preserving fcsa rba) rwa)
                 (\ x -> Inl (snd x)))
           (\ _ ->
             bindb (catcha (forallm (check_rule_reflecting fcsa rb) rba)
                     (\ x -> Inl (snd x)))
               (\ _ ->
                 bindb (check_superset_of_blocked f p paa)
                   (\ _ -> check_superset_of_blocked f pw pwaa)))))))))))
               (\ x ->
                 Inl (shows_prec_list Zero_nat
                        ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                          'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ',
                          'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
                          't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'c',
                          'o', 'n', 'd', 'i', 't', 'i', 'o', 'n', 's', ' ', 't',
                          'o', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 'f', 'r',
                          'o', 'm', ' '] .
                       shows_nl .
                         shows_dpp (shows_prec Zero_nat) (shows_prec Zero_nat) i
                           dpp .
                           shows_nl .
                             shows_prec_list Zero_nat
                               ['t', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                 ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                               shows_nl .
                                 shows_dpp (shows_prec Zero_nat)
                                   (shows_prec Zero_nat) i new_dpp .
                                   shows_nl . x))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr new_dpp;
       });

has_unary_root :: forall a b. (Eq a) => a -> Term a b -> Bool;
has_unary_root f (Fun g [t]) = f == g;
has_unary_root f (Var v) = False;
has_unary_root f (Fun v []) = False;
has_unary_root f (Fun v (vb : vd : ve)) = False;

strip_unary_root :: forall a b. (Eq a) => a -> Term a b -> Term a b;
strip_unary_root f (Fun g [t]) = (if f == g then t else Fun g [t]);
strip_unary_root f (Var v) = Var v;
strip_unary_root f (Fun v []) = Fun v [];
strip_unary_root f (Fun v (vb : vd : ve)) = Fun v (vb : vd : ve);

unblock_term :: forall a b. (Eq a) => a -> Term a b -> Term a b;
unblock_term f (Fun g ts) =
  (if all (has_unary_root f) ts then Fun g (map (strip_unary_root f) ts)
    else Fun g ts);
unblock_term f (Var v) = Var v;

unblock_rule ::
  forall a b. (Eq a) => a -> (Term a b, Term a b) -> (Term a b, Term a b);
unblock_rule f r = (unblock_term f (fst r), unblock_term f (snd r));

partition_pairs ::
  forall a b.
    (Eq a,
      Eq b) => a -> [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        ([(Term a b, Term a b)], [(Term a b, Term a b)]);
partition_pairs f p = partition (\ r -> membera p (unblock_rule f r));

fcc_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    b -> [Ctxt b c] ->
                           [(Term b c, Term b c)] ->
                             [(Term b c, Term b c)] ->
                               a -> Sum ([Char] -> [Char]) a;
fcc_proc i f fcs pb rw dpp =
  let {
    p = pa i dpp;
    q = qc i dpp;
    r = rc i dpp;
    (pc, pw) = partition_pairs f p pb;
  } in bindb (check (null q)
               (shows_string
                  ['Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p',
                    't', 'y'] .
                 shows_nl))
         (\ _ ->
           bindb (check (null r)
                   (shows_prec_list Zero_nat
                     ['s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u', 'l', 'e',
                       's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                       'e', 'd']))
             (\ _ ->
               bindb (check_left_linear_trs (rwc i dpp))
                 (\ _ -> fcc_proc_cond i f fcs pc pw [] rw dpp)));

check_graph_decomp ::
  forall a.
    (a -> [Char] -> [Char]) ->
      ((a, a) -> Bool) -> [(Bool, [a])] -> Sum ([Char] -> [Char]) ();
check_graph_decomp ss g rcs =
  bindb (check_pairwise (check_edges ss g) (map snd rcs))
    (\ _ ->
      catcha
        (forallm (\ c -> check_edges ss g c c)
          (map_filter (\ x -> (if not (fst x) then Just (snd x) else Nothing))
            rcs))
        (\ x -> Inl (snd x)));

reverse_rules_mapa ::
  forall a b c d.
    Dpp_ops_ext a b c d -> a -> (b, Nat) -> [(Term b c, Term b c)];
reverse_rules_mapa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = reverse_rules_map;

rules_non_collapsinga :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
rules_non_collapsinga
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = rules_non_collapsing;

reverse_capRM_dpp ::
  forall a b c.
    (Eq b, Eq c) => Dpp_ops_ext a b c () -> a -> Term b c -> Gctxt b c;
reverse_capRM_dpp i dpp =
  capRM (rules_non_collapsinga i dpp) (reverse_rules_mapa i dpp);

check_dep_graph_proc ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> [(Maybe c, [(Term b [Char], Term b [Char])])] ->
                           Sum ([Char] -> [Char]) ();
check_dep_graph_proc i dpp dps =
  let {
    c = capRM_dpp i dpp;
    rc = reverse_capRM_dpp i dpp;
    _ = nFQ_subset_NF_rulesb i dpp;
    _ = is_QNFc i dpp;
    _ = map fst (rulesd i dpp);
    iedg = is_iedg_edge_dpp i dpp;
    p = pairsa i dpp;
  } in bindb (catcha (check_subseteq p (concatMap snd dps))
               (\ x ->
                 Inl (shows_string
                        ['D', 'e', 'p', 'e', 'n', 'd', 'e', 'n', 'c', 'y', ' ',
                          'P', 'a', 'i', 'r', ' '] .
                       shows_rule (shows_prec Zero_nat)
                         (shows_prec_list Zero_nat) [' ', '-', '>', ' '] x .
                         shows_string
                           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                             'g', ' ', 'i', 'n', ' ', 'd', 'e', 'c', 'o', 'm',
                             'p', 'o', 's', 'i', 't', 'i', 'o', 'n'] .
                           shows_nl)))
         (\ _ ->
           catcha
             (check_graph_decomp (shows_prec_prod Zero_nat . fst)
               (\ (a, b) ->
                 let {
                   (aa, ba) = a;
                 } in let {
                        (_, t) = aa;
                      } in (\ (_, (ct, ict)) (ab, bb) ->
                             let {
                               (u, _) = ab;
                             } in (\ (cu, (_, _)) ->
                                    matchb ct u && matchb cu t && ict u)
                               bb)
                        ba
                   b)
               (map (\ (real, cs) ->
                      (not (is_none real),
                        map (\ (s, t) -> ((s, t), (rc s, (c t, iedg (s, t)))))
                          cs))
                 dps))
             (\ x ->
               Inl (shows_string
                      ['o', 'u', 'r', ' ', 'e', 's', 't', 'i', 'm', 'a', 't',
                        'i', 'o', 'n', ' ', '(', 'E', 'D', 'G', '*', '*', '*',
                        ' ', '+', ' ', 'I', 'E', 'D', 'G', '*', '*', '*', ')',
                        ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                        's', 'h', 'o', 'w', ' ', 't', 'h', 'a', 't', ' ', 'y',
                        'o', 'u', ' ', 'h', 'a', 'v', 'e', ' ', 'a', ' ', 'v',
                        'a', 'l', 'i', 'd', ' ', 'd', 'e', 'c', 'o', 'm', 'p',
                        'o', 's', 'i', 't', 'i', 'o', 'n', ' '] .
                     shows_string
                       ['d', 'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ',
                         'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'r',
                         'e', 'a', 's', 'o', 'n'] .
                       shows_nl . x)));

intersect_pairsa ::
  forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)] -> a;
intersect_pairsa
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = intersect_pairs;

dep_graph_proc ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> [(Maybe c, [(Term b [Char], Term b [Char])])] ->
                           Sum ([Char] -> [Char]) [(c, a)];
dep_graph_proc i d dps =
  (case check_dep_graph_proc i d dps of {
    Inl a -> Inl a;
    Inr _ ->
      Inr (map_filter
            (\ x ->
              (if not (is_none (fst x))
                then Just (the (fst x), intersect_pairsa i d (snd x))
                else Nothing))
            dps);
  });

instantiation_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    (Term b [Char], Term b [Char]) ->
                      [(Term b [Char], Term b [Char])] ->
                        a -> Sum ([Char] -> [Char]) a;
instantiation_proc i st sts dpp =
  (case let {
          ic = icap_impl_dpp_mv i dpp;
          isnf = is_QNFc i dpp;
          (s, t) = st;
          sy = map_vars (\ a -> 'y' : a) s;
          ty = map_vars (\ a -> 'y' : a) t;
          iedg = is_iedg_edge_dpp i dpp;
        } in catcha
               (forallm
                 (\ (u, v) ->
                   (case mgu_class (ic [u] v) s of {
                     Nothing -> Inr ();
                     Just mu ->
                       check (not (isnf (subst_apply_term sy mu)) ||
                               not (isnf (subst_apply_term
   (map_vars (\ a -> 'x' : a) u) mu)) ||
                                 any (\ sta ->
                                       instance_rule sta st &&
 instance_rule (subst_apply_term sy mu, subst_apply_term ty mu) sta)
                                   sts)
                         (shows_prec_list Zero_nat
                            ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                              'f', 'i', 'n', 'd', ' ', 'i', 'n', 's', 't', 'a',
                              'n', 'c', 'e', ' ', 'o', 'f', ' ', 'p', 'a', 'i',
                              'r', ' '] .
                           shows_rule (shows_prec Zero_nat)
                             (shows_prec_list Zero_nat) [' ', '-', '>', ' ']
                             (subst_apply_term sy mu, subst_apply_term ty mu) .
                             shows_nl .
                               shows_prec_list Zero_nat
                                 ['w', 'h', 'i', 'c', 'h', ' ', 'r', 'e', 's',
                                   'u', 'l', 't', 'e', 'd', ' ', 'f', 'r', 'o',
                                   'm', ' ', 'D', 'P', ' '] .
                                 shows_rule (shows_prec Zero_nat)
                                   (shows_prec_list Zero_nat)
                                   [' ', '-', '>', ' '] (u, v));
                   }))
                 (filter (\ (u, v) -> iedg (u, v) s) (pairsa i dpp)))
               (\ x -> Inl (snd x))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (replace_paira i dpp st sts);
  });

mono_redpair_proc ::
  forall a b c.
    (Showa b,
      Showa c) => Dpp_ops_ext a b c () ->
                    Redtriple_ext b c () ->
                      [(Term b c, Term b c)] ->
                        [(Term b c, Term b c)] -> a -> Sum ([Char] -> [Char]) a;
mono_redpair_proc i rp premove rremove dpp =
  (case catcha
          (let {
             (ps, pns) = split_pairsa i dpp premove;
             (rs, rns) = split_rulesc i dpp rremove;
           } in bindb (valid rp)
                  (\ _ ->
                    bindb (catcha (mono rp (ps ++ rs ++ pns ++ rns))
                            (\ x ->
                              Inl (shows_string
                                     ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                       'w', 'i', 't', 'h', ' ', 'm', 'o', 'n',
                                       'o', 't', 'o', 'n', 'i', 'c', 'i', 't',
                                       'y', ' ', 'o', 'f', ' ', 's', 't', 'r',
                                       'i', 'c', 't', ' ', 'o', 'r', 'd', 'e',
                                       'r'] .
                                    shows_nl . x)))
                      (\ _ ->
                        bindb (catcha
                                (catcha (forallm (ns rp) rns)
                                  (\ x -> Inl (snd x)))
                                (\ x ->
                                  Inl (shows_string
 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
   'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'T', 'R', 'S'] .
shows_nl . x)))
                          (\ _ ->
                            bindb (catcha
                                    (catcha (forallm (s rp) rs)
                                      (\ x -> Inl (snd x)))
                                    (\ x ->
                                      Inl
(shows_string
   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
     'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'T', 'R', 'S'] .
  shows_nl . x)))
                              (\ _ ->
                                bindb (catcha
(catcha (forallm (ns rp) pns) (\ x -> Inl (snd x)))
(\ x ->
  Inl (shows_string
         ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
           'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
        shows_nl . x)))
                                  (\ _ ->
                                    catcha
                                      (catcha (forallm (s rp) ps)
(\ x -> Inl (snd x)))
                                      (\ x ->
Inl (shows_string
       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
         'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
      shows_nl . x))))))))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                     'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
                     'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p',
                     'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't',
                     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w',
                     'i', 'n', 'g'] .
                  shows_nl . desca rp . shows_nl . x))
    of {
    Inl a -> Inl a;
    Inr _ ->
      Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove rremove);
  });

create_rseq_map ::
  forall a b.
    (Eq a,
      Eq b) => [((Term a b, Term a b),
                  [(Pos, ((Term a b, Term a b), Term a b))])] ->
                 (Term a b, Term a b) ->
                   Maybe [(Pos, ((Term a b, Term a b), Term a b))];
create_rseq_map rseqs r = map_of rseqs r;

funas_term_lista :: forall a b. (Eq a) => Term a b -> [(a, Nat)];
funas_term_lista = remdups . funas_term_list;

matchCapRMBelow ::
  forall a b.
    (Eq a,
      Eq b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                 Term a b -> Term a b -> Bool;
matchCapRMBelow rm l (Fun f ts) = matchb (GCFun f (map (capRM2 rm) ts)) l;

check_ur_closed_term_rm_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                    [(Term a b, Term a b)] ->
                      ((a, Nat) -> Nat -> Bool) ->
                        Term a b -> Sum ([Char] -> [Char]) ();
check_ur_closed_term_rm_af uu uv uw (Var x) = Inr ();
check_ur_closed_term_rm_af rm ur pia (Fun f ts) =
  let {
    n = size_list ts;
    piaa = pia (f, n);
  } in bindb (catcha
               (forallm_index
                 (\ t i ->
                   (if piaa i then check_ur_closed_term_rm_af rm ur pia t
                     else Inr ()))
                 ts)
               (\ x -> Inl (snd x)))
         (\ _ ->
           catcha
             (forallm
               (\ lr ->
                 check (membera ur lr ||
                         not (matchCapRMBelow rm (fst lr) (Fun f ts)))
                   (shows_string
                      ['d', 'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ',
                        's', 'u', 'b', 't', 'e', 'r', 'm', ' '] .
                     shows_prec_term Zero_nat (Fun f ts) .
                       shows_string
                         [' ', 'o', 'f', ' ', 's', 'o', 'm', 'e', ' ', 'u', 's',
                           'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',', ' ',
                           'r', 'u', 'l', 'e', ' '] .
                         shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                           [' ', '-', '>', ' '] lr .
                           shows_string
                             [' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.']))
               (rm (f, n)))
             (\ x -> Inl (snd x)));

check_ur_P_closed_rm_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        ((a, Nat) -> Nat -> Bool) -> Sum ([Char] -> [Char]) ();
check_ur_P_closed_rm_af rm ur p pia =
  bindb (catcha
          (catcha
            (forallm (\ lr -> check_ur_closed_term_rm_af rm ur pia (snd lr)) ur)
            (\ x -> Inl (snd x)))
          (\ x ->
            Inl (shows_string
                   ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
                     'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l', 'o', 's',
                     'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r', 't', 'i',
                     'e', 's', ' ', 'o', 'f', ' ', 'r', 'h', 's', ' ', 'o', 'f',
                     ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                     's'] .
                  shows_nl . x)))
    (\ _ ->
      catcha
        (catcha
          (forallm (\ st -> check_ur_closed_term_rm_af rm ur pia (snd st)) p)
          (\ x -> Inl (snd x)))
        (\ x ->
          Inl (shows_string
                 ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
                   'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l', 'o', 's',
                   'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r', 't', 'i',
                   'e', 's', ' ', 'o', 'f', ' ', 'r', 'h', 's', ' ', 'o', 'f',
                   ' ', 'D', 'P', 's'] .
                shows_nl . x)));

mono_ur_redpair_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    Redtriple_ext b c () ->
                      [(Term b c, Term b c)] ->
                        [(Term b c, Term b c)] ->
                          [(Term b c, Term b c)] ->
                            a -> Sum ([Char] -> [Char]) a;
mono_ur_redpair_proc i rp premove rremove ur dpp =
  (case catcha
          (bindb
            (check (minimal i dpp)
              (shows_prec_list Zero_nat
                ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y', ' ', 'r',
                  'e', 'q', 'u', 'i', 'r', 'e', 'd']))
            (\ _ ->
              bindb (check (if nfsc i dpp then wwf_rulesa i dpp else True)
                      (shows_prec_list Zero_nat
                        ['w', 'e', 'l', 'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd',
                          'n', 'e', 's', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                          'e', 'd']))
                (\ _ ->
                  let {
                    p = pairsa i dpp;
                    us = concatMap (funas_term_lista . snd) (p ++ ur);
                    filt =
                      (\ lr -> all (membera us) (funas_term_lista (fst lr)));
                    (pms, pns) = split_pairsa i dpp premove;
                    (ps, pnwf) = partition filt pms;
                    (urms, urns) = partition (membera rremove) ur;
                    (urs, urnwf) = partition filt urms;
                    rm = rules_mapb i dpp;
                  } in bindb (valid rp)
                         (\ _ ->
                           bindb (mono rp
                                   (ps ++ urs ++ urns ++ urnwf ++ pns ++ pnwf))
                             (\ _ ->
                               bindb (check_ur_P_closed_rm_af rm ur p full_af)
                                 (\ _ ->
                                   bindb (catcha
   (forallm
     (\ (l, _) ->
       check (not (is_Var l))
         (shows_prec_list Zero_nat
           ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 's', ' ', 'a', 's', ' ',
             'l', 'h', 's', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
             'w', 'e', 'd']))
     (rulesd i dpp))
   (\ x -> Inl (snd x)))
                                     (\ _ ->
                                       bindb
 (catcha (catcha (forallm (ns rp) (urns ++ urnwf)) (\ x -> Inl (snd x)))
   (\ x ->
     Inl (shows_string
            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
              'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's', 'a',
              'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's'] .
           shows_nl . x)))
 (\ _ ->
   bindb (catcha (catcha (forallm (s rp) urs) (\ x -> Inl (snd x)))
           (\ x ->
             Inl (shows_string
                    ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                      ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ',
                      'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                      's'] .
                   shows_nl . x)))
     (\ _ ->
       bindb (catcha
               (catcha (forallm (ns rp) (pns ++ pnwf)) (\ x -> Inl (snd x)))
               (\ x ->
                 Inl (shows_string
                        ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                          'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                          ' ', 'D', 'P', 's'] .
                       shows_nl . x)))
         (\ _ ->
           catcha (catcha (forallm (s rp) ps) (\ x -> Inl (snd x)))
             (\ x ->
               Inl (shows_string
                      ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e',
                        'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g',
                        ' ', 'D', 'P', 's'] .
                     shows_nl . x))))))))))))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                     'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
                     'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p',
                     'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't',
                     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w',
                     'i', 'n', 'g'] .
                  shows_nl . desca rp . shows_nl . x))
    of {
    Inl a -> Inl a;
    Inr _ ->
      Inr (delete_R_Rwc i (delete_P_Pwa i dpp premove premove) rremove rremove);
  });

fcc_split_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    b -> [Ctxt b c] ->
                           [(Term b c, Term b c)] ->
                             [(Term b c, Term b c)] ->
                               [(Term b c, Term b c)] ->
                                 [(Term b c, Term b c)] ->
                                   a -> Sum ([Char] -> [Char]) (a, a);
fcc_split_proc i f fcs pb rb ps rs dpp =
  let {
    p = pa i dpp;
    pw = pwa i dpp;
    r = rc i dpp;
    rw = rwc i dpp;
    q = qc i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    pba = list_union p pw;
    rba = rw;
    pns = list_diff pba ps;
    rns = list_diff rba rs;
    (paa, pwaa) = partition_pairs f ps pb;
    (ra, rwa) = partition_rules (More f [] Hole [] : fcs) rs rb;
    two = mkc i nfs m (list_diff p ps) (list_diff pw ps) [] [] rns;
    dpp_mid = mkc i nfs m ps pns [] rs rns;
  } in bindb (bindb
               (catcha (check_subseteq ps pba)
                 (\ x ->
                   Inl (shows_prec_list Zero_nat ['p', 'a', 'i', 'r', ' '] .
                         shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                           [' ', '-', '>', ' '] x .
                           shows_prec_list Zero_nat
                             [' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'd', 'e', 'l', 'e', 't', 'e', 'd', ' ', 'b',
                               'u', 't', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                               'p', 'r', 'e', 's', 'e', 'n', 't'])))
               (\ _ ->
                 bindb (catcha (check_subseteq rs rba)
                         (\ x ->
                           Inl (shows_prec_list Zero_nat
                                  ['r', 'u', 'l', 'e', ' '] .
                                 shows_rule (shows_prec Zero_nat)
                                   (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                   x .
                                   shows_prec_list Zero_nat
                                     [' ', 's', 'h', 'o', 'u', 'l', 'd', ' ',
                                       'b', 'e', ' ', 'd', 'e', 'l', 'e', 't',
                                       'e', 'd', ' ', 'b', 'u', 't', ' ', 'i',
                                       's', ' ', 'n', 'o', 't', ' ', 'p', 'r',
                                       'e', 's', 'e', 'n', 't'])))
                   (\ _ ->
                     bindb (check (null q)
                             (shows_string
                                ['Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                                  'e', 'm', 'p', 't', 'y'] .
                               shows_nl))
                       (\ _ ->
                         bindb (check (null r)
                                 (shows_prec_list Zero_nat
                                   ['s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u',
                                     'l', 'e', 's', ' ', 'n', 'o', 't', ' ',
                                     'a', 'l', 'l', 'o', 'w', 'e', 'd']))
                           (\ _ ->
                             bindb (check_left_linear_trs rw)
                               (\ _ ->
                                 fcc_proc_cond i f fcs paa pwaa ra rwa
                                   dpp_mid))))))
         (\ one -> Inr (one, two));

data Slm_ops_ext a b c d =
  Slm_ops_ext (a -> [b] -> c) (a -> [b] -> b) [b] b (a -> [b] -> c) d;

rl_slm ::
  forall a b.
    (Eq a,
      Eq b) => Maybe (Lab a b, Nat) ->
                 [(Lab a b, Nat)] ->
                   [(Lab a b, Nat)] ->
                     Sum ([Char] -> [Char])
                       (Slm_ops_ext (Lab a b) (Lab a b) (Sum b [Lab a b]) ());
rl_slm delt_opt pre_fs g =
  let {
    fs = (if is_none delt_opt then pre_fs
           else filter (\ f -> not (f == the delt_opt)) pre_fs);
  } in bindb (check (not (null fs))
               (shows_string
                  ['r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                    'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
                    't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ',
                    'f', 'u', 'n', 'c', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm',
                    'b', 'o', 'l', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 's',
                    'i', 'g', 'n', 'a', 't', 'u', 'r', 'e'] .
                 shows_nl))
         (\ _ ->
           let {
             f = fst (hd fs);
           } in Inr (Slm_ops_ext (\ _ -> Inr)
                      (\ ga cs ->
                        (if membera fs (ga, size_list cs) then ga else f))
                      (map fst fs) f
                      (if is_none delt_opt then (\ _ -> Inr)
                        else (\ _ gs ->
                               Inr (replicatea (size_list gs)
                                     (fst (the delt_opt)))))
                      ()));

data Sl_ops_ext a b c d e =
  Sl_ops_ext (a -> [b] -> c) (a -> Nat -> c -> Bool) (a -> [b] -> b) [b] b
    ([(Term a d, Term a d)] -> Sum ([Char] -> [Char]) ()) (a -> [b] -> c)
    (a -> Nat -> c -> Bool) (c -> [c]) (a -> Nat -> [c]) e;

qmodel_cge :: Nat -> Nat -> Bool;
qmodel_cge = (\ x y -> less_eq_nat y x);

pointwise_ext ::
  forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
pointwise_ext f [] [] = (False, True);
pointwise_ext f (uu : uv) [] = (False, False);
pointwise_ext f [] (uw : ux) = (False, False);
pointwise_ext f (a : asa) (b : bs) =
  (case f a b of {
    (stri, True) ->
      (case pointwise_ext f asa bs of {
        (strir, True) -> (stri || strir, True);
        (strir, False) -> (False, False);
      });
    (stri, False) -> (False, False);
  });

qmodel_lge :: forall a b c d. a -> b -> Sum [Nat] c -> Sum [Nat] d -> Bool;
qmodel_lge f n =
  (\ l r ->
    (case (l, r) of {
      (Inl cs1, Inl cs2) ->
        snd (pointwise_ext (\ x y -> (less_nat y x, less_eq_nat y x)) cs1 cs2);
      (Inl cs1, Inr _) -> False;
      (Inr _, b) -> False;
    }));

qmodel_L ::
  forall a b c d.
    (Eq a, Eq b) => [(Lab a b, Nat)] -> Lab a b -> [c] -> Sum [c] d;
qmodel_L sig =
  (\ f cs -> (if membera sig (f, size_list cs) then Inl cs else Inl []));

enum_vectors_nat :: forall a. [a] -> Nat -> [[a]];
enum_vectors_nat c Zero_nat = [[]];
enum_vectors_nat c (Suc n) =
  let {
    a = enum_vectors_nat c n;
  } in concatMap (\ vec -> map (\ ca -> ca : vec) c) a;

qmodel_LS_gen ::
  forall a b c. (Eq a) => [(a, Nat)] -> [b] -> a -> Nat -> [Sum [b] c];
qmodel_LS_gen sig ls =
  (\ f n ->
    (if membera sig (f, n) then map Inl (enum_vectors_nat ls n) else [Inl []]));

qmodel_LS ::
  forall a b c.
    (Eq a, Eq b, Eq c) => [(a, Nat)] -> [b] -> a -> Nat -> Sum [b] c -> Bool;
qmodel_LS sig ls = (\ f n -> membera (qmodel_LS_gen sig ls f n));

qmodel_LSa :: forall a. Lab a [Nat] -> Nat -> Sum [Nat] [Lab a [Nat]] -> Bool;
qmodel_LSa = (\ _ _ a -> (case a of {
                           Inl _ -> True;
                           Inr _ -> False;
                         }));

pointwise_lgen :: [Nat] -> [[Nat]];
pointwise_lgen ns = concat_lists (map (\ n -> upt Zero_nat (Suc n)) ns);

qmodel_lgen :: forall a. Sum [Nat] a -> [Sum [Nat] a];
qmodel_lgen l =
  (case l of {
    Inl ns -> map Inl (pointwise_lgen ns);
    Inr _ -> [];
  });

check_decr_present_aux_1 ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a (b, Nat), Term a (b, Nat))] ->
                 b -> a -> a -> Nat ->
                                  Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present_aux_1 r v f1 f2 n =
  let {
    vs = map (\ na -> Var (v, na)) (upt Zero_nat n);
    rule = (Fun f1 vs, Fun f2 vs);
  } in check (not (is_none (find (instance_rule rule) r))) rule;

check_decr_present_aux_2 ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 b -> [(a, (a, Nat))] ->
                        Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present_aux_2 r v req =
  let {
    add_nats = map_vars (\ va -> (va, Zero_nat));
    ra = map (\ (l, ra) -> (add_nats l, add_nats ra)) r;
  } in catcha
         (forallm (\ (f1, (a, b)) -> check_decr_present_aux_1 ra v f1 a b) req)
         (\ x -> Inl (snd x));

check_decr_present ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Nat)] ->
                 (a -> [Nat] -> a) ->
                   b -> Nat ->
                          [(Term a b, Term a b)] ->
                            Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present sig l v c r =
  let {
    ca = upt Zero_nat (Suc c);
    ls = (\ (f, n) ->
           concatMap
             (\ cs ->
               concatMap
                 (\ i ->
                   let {
                     ci = nth cs i;
                   } in (if less_nat ci c
                          then [(l f (list_update cs i (Suc ci)), (l f cs, n))]
                          else []))
                 (upt Zero_nat n))
             (enum_vectors_nat ca n));
  } in check_decr_present_aux_2 r v (concatMap ls sig);

qmodel_check_decr ::
  forall a b.
    (Eq a, Showa a,
      Eq b) => [(Lab a [Nat], Nat)] ->
                 b -> Nat ->
                        [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                          Sum ([Char] -> [Char]) ();
qmodel_check_decr sig v c =
  (\ lR ->
    catcha (check_decr_present sig Lab v c lR)
      (\ x ->
        Inl (let {
               display =
                 map_vars
                   (\ (_, n) ->
                     (shows_prec_list Zero_nat ['x'] .
                       shows_prec_nat Zero_nat n)
                       []);
             } in shows_prec_list Zero_nat
                    ['d', 'e', 'c', 'r', 'e', 'a', 's', 'i', 'n', 'g', ' ', 'r',
                      'u', 'l', 'e', ' '] .
                    shows_rule (shows_prec_lab Zero_nat)
                      (shows_prec_list Zero_nat) [' ', '-', '>', ' ']
                      (display (fst x), display (snd x)) .
                      shows_prec_list Zero_nat
                        [' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])));

take_default :: forall a. a -> [a] -> Nat -> a;
take_default def [] uu = def;
take_default uv (x : xs) Zero_nat = x;
take_default def (uw : xs) (Suc i) = take_default def xs i;

eval_arithFun_unbound :: [Nat] -> ArithFun -> Nat;
eval_arithFun_unbound nats (Arg i) = take_default Zero_nat nats i;
eval_arithFun_unbound nats (Const n) = n;
eval_arithFun_unbound nats (Sum []) = Zero_nat;
eval_arithFun_unbound nats (Sum (f : fs)) =
  plus_nat (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Sum fs));
eval_arithFun_unbound nats (Prod []) = one_nat;
eval_arithFun_unbound nats (Prod (f : fs)) =
  times_nat (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Prod fs));
eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (Max (f : v : va)) =
  maxa (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Max (v : va)));
eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (Min (f : v : va)) =
  mina (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Min (v : va)));

eval_arithFun :: Nat -> ArithFun -> [Nat] -> Nat;
eval_arithFun c f nats = mod_nat (eval_arithFun_unbound nats f) c;

sl_inter_to_inter :: forall a. (Eq a) => Sl_inter a -> a -> [Nat] -> Nat;
sl_inter_to_inter (SL_Inter c ls) fl cs =
  (case map_of ls (fl, size_list cs) of {
    Nothing -> Zero_nat;
    Just f -> eval_arithFun (Suc c) f cs;
  });

get_largest_element :: forall a. Sl_inter a -> Nat;
get_largest_element (SL_Inter n uu) = n;

qsli_to_sl_unsafe ::
  forall a b.
    (Eq a, Eq b,
      Showa b) => a -> [(Lab b [Nat], Nat)] ->
                         [(Lab b [Nat], Nat)] ->
                           Sl_inter (Lab b [Nat]) ->
                             Sl_ops_ext (Lab b [Nat]) Nat
                               (Sum [Nat] [Lab b [Nat]]) a ();
qsli_to_sl_unsafe v f g sli =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
  } in Sl_ops_ext (qmodel_L f) (qmodel_LS f ca) (sl_inter_to_inter sli) ca c
         (qmodel_check_decr f v c) (qmodel_L g) qmodel_LSa qmodel_lgen
         (qmodel_LS_gen f ca) ();

qmodel_check_interpretation ::
  ArithFun -> Nat -> Nat -> Sum ([Char] -> [Char]) ();
qmodel_check_interpretation f n c =
  let {
    ca = upt Zero_nat (Suc c);
    css = enum_vectors_nat ca n;
  } in catcha
         (forallm
           (\ cs ->
             catcha
               (forallm
                 (\ i ->
                   catcha
                     (forallm
                       (\ l ->
                         check (less_eq_nat (eval_arithFun (Suc c) f cs)
                                 (eval_arithFun (Suc c) f (list_update cs i l)))
                           (shows_prec_list Zero_nat
                              ['n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o',
                                'n', 'e', ' ', 'i', 'n', ' '] .
                             shows_prec_nat Zero_nat (Suc i) .
                               shows_prec_list Zero_nat
                                 ['.', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                                   't']))
                       (upt (nth cs i) (Suc c)))
                     (\ x -> Inl (snd x)))
                 (upt Zero_nat n))
               (\ x -> Inl (snd x)))
           css)
         (\ x -> Inl (snd x));

qmodel_check_valid ::
  forall a. (Showa a) => Sl_inter a -> Sum ([Char] -> [Char]) ();
qmodel_check_valid (SL_Inter c ls) =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ g ->
               catcha (qmodel_check_interpretation g n c)
                 (\ x ->
                   Inl (shows_prec_list Zero_nat
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n',
                            ' ', 'w', 'e', 'a', 'k', '-', 'm', 'o', 'n', 'o',
                            't', 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                            'f', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e',
                            't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                         shows_prec Zero_nat f . shows_nl . x)))
          b)
      ls)
    (\ x -> Inl (snd x));

qsli_to_sl ::
  forall a b.
    (Eq a, Eq b,
      Showa b) => a -> [(Lab b [Nat], Nat)] ->
                         [(Lab b [Nat], Nat)] ->
                           Sl_inter (Lab b [Nat]) ->
                             Sum ([Char] -> [Char])
                               (Sl_ops_ext (Lab b [Nat]) Nat
                                 (Sum [Nat] [Lab b [Nat]]) a ());
qsli_to_sl v f g sli =
  bindb (qmodel_check_valid sli) (\ _ -> Inr (qsli_to_sl_unsafe v f g sli));

sli_to_slm ::
  forall a.
    (Eq a) => Sl_inter (Lab a [Nat]) ->
                Slm_ops_ext (Lab a [Nat]) Nat (Sum [Nat] [Lab a [Nat]]) ();
sli_to_slm sli =
  let {
    c = get_largest_element sli;
  } in Slm_ops_ext (\ _ -> Inl) (sl_inter_to_inter sli) (upt Zero_nat (Suc c)) c
         (\ _ -> Inl) ();

sl_C :: forall a b c d e. Sl_ops_ext a b c d e -> [b];
sl_C (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
       sl_LS_gen more)
  = sl_C;

sl_I :: forall a b c d e. Sl_ops_ext a b c d e -> a -> [b] -> b;
sl_I (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
       sl_LS_gen more)
  = sl_I;

sl_L :: forall a b c d e. Sl_ops_ext a b c d e -> a -> [b] -> c;
sl_L (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
       sl_LS_gen more)
  = sl_La;

eval_lab ::
  forall a b c d e.
    (a -> [b] -> b) ->
      (a -> [b] -> c) ->
        (a -> Nat -> c -> d) -> (e -> b) -> Term a e -> (b, Term d e);
eval_lab i l lc alpha (Var x) = (alpha x, Var x);
eval_lab i l lc alpha (Fun f ts) =
  let {
    clts = map (eval_lab i l lc alpha) ts;
    cs = map fst clts;
    c = i f cs;
    lts = map snd clts;
  } in (c, Fun (lc f (size_list ts) (l f cs)) lts);

lab_lhs_list ::
  forall a b c d e.
    (Eq e) => (a -> [b] -> b) ->
                (a -> [b] -> c) ->
                  (a -> Nat -> c -> d) -> [b] -> Term a e -> [Term d e];
lab_lhs_list i l lc c t =
  map (\ alpha -> snd (eval_lab i l lc alpha t))
    (map fun_of (enum_vectors c (vars_term_lista t)));

lab_lhss_list ::
  forall a b c d e.
    (Eq e) => (a -> [b] -> b) ->
                (a -> [b] -> c) ->
                  (a -> Nat -> c -> d) -> [b] -> [Term a e] -> [Term d e];
lab_lhss_list i l lc c q = concatMap (lab_lhs_list i l lc c) q;

check_NF_vars_subset ::
  forall a b. (Eq a, Eq b) => [Term a b] -> [Term a b] -> Sum (Term a b) ();
check_NF_vars_subset qa q =
  catcha (forallm (\ qaa -> check (any (matches qaa) q) qaa) qa)
    (\ x -> Inl (snd x));

check_sl_Q ::
  forall a b c d.
    (Eq a, Showa a, Eq d,
      Showa d) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> Nat -> c -> a) ->
                        [b] ->
                          [Term a d] -> [Term a d] -> Sum ([Char] -> [Char]) ();
check_sl_Q i l lc c lQ q =
  bindb (check (not (null c))
          (shows_prec_list Zero_nat
            ['c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', 'm', 'u', 's', 't', ' ',
              'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
    (\ _ ->
      catcha (check_NF_vars_subset (lab_lhss_list i l lc c q) lQ)
        (\ x ->
          Inl (shows_prec_list Zero_nat
                 ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 't', 'e', 'r', 'm',
                   ' '] .
                shows_term (shows_prec Zero_nat) (shows_prec Zero_nat) x .
                  shows_prec_list Zero_nat
                    [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])));

lab_rule_ass ::
  forall a b c d e.
    (a -> [b] -> b) ->
      (a -> [b] -> c) ->
        (a -> Nat -> c -> d) ->
          (e -> b) -> (Term a e, Term a e) -> (Term d e, Term d e);
lab_rule_ass i l lc alpha rule =
  (snd (eval_lab i l lc alpha (fst rule)),
    snd (eval_lab i l lc alpha (snd rule)));

lab_rule_list ::
  forall a b c d e.
    (Eq e) => (a -> [b] -> b) ->
                (a -> [b] -> c) ->
                  (a -> Nat -> c -> d) ->
                    [b] -> (Term a e, Term a e) -> [(Term d e, Term d e)];
lab_rule_list i l lc c lr =
  map (\ alpha -> lab_rule_ass i l lc alpha lr)
    (map fun_of (enum_vectors c (vars_rule_list lr)));

lab_trs_list ::
  forall a b c d e.
    (Eq e) => (a -> [b] -> b) ->
                (a -> [b] -> c) ->
                  (a -> Nat -> c -> d) ->
                    [b] -> [(Term a e, Term a e)] -> [(Term d e, Term d e)];
lab_trs_list i l lc c r = concatMap (lab_rule_list i l lc c) r;

check_sl_lab ::
  forall a b c d.
    (Eq a, Showa a, Eq d,
      Showa d) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> Nat -> c -> a) ->
                        [b] ->
                          [(Term a d, Term a d)] ->
                            [(Term a d, Term a d)] -> Sum ([Char] -> [Char]) ();
check_sl_lab i l lc c lR r =
  bindb (check (not (null c))
          (shows_prec_list Zero_nat
            ['c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', 'm', 'u', 's', 't', ' ',
              'b', 'e', ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
    (\ _ ->
      catcha (check_subseteq lR (lab_trs_list i l lc c r))
        (\ x ->
          Inl (shows_prec_list Zero_nat
                 ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                   ' '] .
                shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                  [' ', '-', '>', ' '] x .
                  shows_prec_list Zero_nat
                    [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                      'w', 'e', 'd'])));

check_sl_Qa ::
  forall a b c.
    (Eq a, Showa a, Eq c,
      Showa c) => (a -> (a, b)) ->
                    [Term a c] -> [Term a c] -> Sum ([Char] -> [Char]) ();
check_sl_Qa ld lQ q =
  let {
    u = (\ l -> fst (ld l));
  } in catcha
         (forallm
           (\ lq ->
             check (let {
                      mlq = map_term u lq;
                    } in any (\ qa -> matches mlq qa && matches qa mlq) q)
               (shows_prec_list Zero_nat
                  ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' '] .
                 shows_term (shows_prec Zero_nat) (shows_prec Zero_nat) lq .
                   shows_prec_list Zero_nat
                     [' ', 'y', 'i', 'e', 'l', 'd', 's', ' ', 'a', ' ', 't',
                       'e', 'r', 'm', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ',
                       'Q']))
           lQ)
         (\ x -> Inl (snd x));

model_splitter ::
  forall a b c.
    (Eq a,
      Eq c) => (a -> (a, b)) ->
                 [(Term a c, Term a c)] ->
                   [(Term a c, Term a c)] ->
                     ([(Term a c, Term a c)],
                       ([(Term a c, Term a c)], [(Term a c, Term a c)]));
model_splitter ld lAll uRw =
  let {
    unlab = (\ lf -> fst (ld lf));
    la = map (\ r -> (r, map_rule unlab r)) lAll;
    (rw, r) = partition (\ (_, a) -> membera uRw a) la;
  } in (map fst r, (map fst rw, []));

sem_lab_proc ::
  forall a b c d.
    (Eq a, Showa a, Eq d,
      Showa d) => (a -> (a, b)) ->
                    Dpp_ops_ext c a d () ->
                      Sum ([Char] -> [Char]) () ->
                        ([Term a d] ->
                          [Term a d] -> Sum ([Char] -> [Char]) ()) ->
                          ([(Term a d, Term a d)] ->
                            [(Term a d, Term a d)] ->
                              Sum ([Char] -> [Char]) ()) ->
                            ([(Term a d, Term a d)] ->
                              [(Term a d, Term a d)] ->
                                Sum ([Char] -> [Char]) ()) ->
                              ([(Term a d, Term a d)] ->
                                [(Term a d, Term a d)] ->
                                  Sum ([Char] -> [Char]) ()) ->
                                [(Term a d, Term a d)] ->
                                  [Term a d] ->
                                    [(Term a d, Term a d)] ->
                                      c -> Sum ([Char] -> [Char]) c;
sem_lab_proc ld i valid check_Q check_laba check_lab check_model_lab lPAll lQ
  lRAll dpp =
  let {
    r = rc i dpp;
    rw = rwc i dpp;
    pw = pwa i dpp;
    p = pa i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    (lP, (lPw, _)) = model_splitter ld lPAll pw;
    (lR, (lRw, _)) = model_splitter ld lRAll rw;
  } in (case bindb valid
               (\ _ ->
                 let {
                   q = qc i dpp;
                 } in catcha
                        (bindb
                          (check (if nfs then wwf_rulesa i dpp else True)
                            (shows_prec_list Zero_nat
                              ['w', 'e', 'l', 'l', ' ', 'f', 'o', 'r', 'm', 'e',
                                'd', 'n', 'e', 's', 's', ' ', 'r', 'e', 'q',
                                'u', 'i', 'r', 'e', 'd']))
                          (\ _ ->
                            bindb (check_Q lQ q)
                              (\ _ ->
                                bindb (check_sl_Qa ld lQ q)
                                  (\ _ ->
                                    bindb (check_laba lP p)
                                      (\ _ ->
bindb (check_laba lPw pw)
  (\ _ ->
    bindb (check_model_lab lR r)
      (\ _ ->
        bindb (check_model_lab lRw rw)
          (\ _ -> bindb (check_lab lR r) (\ _ -> check_lab lRw rw)))))))))
                        (\ x ->
                          Inl (shows_string
                                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd',
                                   'u', 'r', 'i', 'n', 'g', ' ', 'l', 'a', 'b',
                                   'e', 'l', 'i', 'n', 'g', ':'] .
                                shows_nl . x)))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc i nfs m lP lPw lQ lR lRw);
       });

check_sl_rule_ass ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => Bool ->
                    (a -> [b] -> b) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          (b -> b -> Bool) ->
                            [(Term d e, Term d e)] ->
                              (e -> b) ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_ass mc i la lc cge lR alpha (l, r) =
  let {
    cl_ll = eval_lab i la lc alpha l;
    cr_lr = eval_lab i la lc alpha r;
  } in bindb (check (if mc then cge (fst cl_ll) (fst cr_lr) else True)
               (shows_string ['r', 'u', 'l', 'e', ' '] .
                 shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                   [' ', '-', '>', ' '] (l, r) .
                   shows_string
                     [' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 't',
                       'h', 'e', ' ', 'm', 'o', 'd', 'e', 'l', ' ', 'c', 'o',
                       'n', 'd', 'i', 't', 'i', 'o', 'n', ',', ' ', '[', 'l',
                       'h', 's', ']', ' ', '=', ' '] .
                     shows_prec Zero_nat (fst cl_ll) .
                       shows_string
                         [',', ' ', '[', 'r', 'h', 's', ']', ' ', '=', ' '] .
                         shows_prec Zero_nat (fst cr_lr)))
         (\ _ ->
           check (membera lR (snd cl_ll, snd cr_lr))
             (shows_string
                ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                  ' '] .
               shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                 [' ', '-', '>', ' '] (snd cl_ll, snd cr_lr) .
                 shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']));

check_sl_rule ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> Nat -> c -> d) ->
                        [b] ->
                          (b -> b -> Bool) ->
                            Bool ->
                              [(Term d e, Term d e)] ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule i l lc c cge mc lR lr =
  catcha
    (forallm (\ alpha -> check_sl_rule_ass mc i l lc cge lR alpha lr)
      (map fun_of (enum_vectors c (vars_rule_list lr))))
    (\ x -> Inl (snd x));

check_sl_lab_trs ::
  forall a b c d.
    (Eq a, Showa a, Showa b, Eq d,
      Showa d) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      [b] ->
                        (b -> b -> Bool) ->
                          (a -> Nat -> c -> a) ->
                            [(Term a d, Term a d)] ->
                              [(Term a d, Term a d)] ->
                                Sum ([Char] -> [Char]) ();
check_sl_lab_trs i l c cge label lP p =
  catcha (forallm (check_sl_rule i l label c cge False lP) p)
    (\ x -> Inl (snd x));

check_sl_model_lab_trs ::
  forall a b c d.
    (Eq a, Showa a, Showa b, Eq d,
      Showa d) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      [b] ->
                        (b -> b -> Bool) ->
                          (a -> Nat -> c -> a) ->
                            [(Term a d, Term a d)] ->
                              [(Term a d, Term a d)] ->
                                Sum ([Char] -> [Char]) ();
check_sl_model_lab_trs i l c cge label lR r =
  catcha (forallm (check_sl_rule i l label c cge True lR) r)
    (\ x -> Inl (snd x));

sem_lab_fin_proc ::
  forall a b c d e.
    (Eq a, Showa a, Eq d, Showa d, Eq e,
      Showa e) => (a -> Nat -> b -> a) ->
                    (a -> (a, b)) ->
                      Dpp_ops_ext c a d () ->
                        ([(a, Nat)] ->
                          [(a, Nat)] ->
                            Sum ([Char] -> [Char]) (Sl_ops_ext a e b d ())) ->
                          [(Term a d, Term a d)] ->
                            [Term a d] ->
                              [(Term a d, Term a d)] ->
                                c -> Sum ([Char] -> [Char]) c;
sem_lab_fin_proc lc ld i gen lPAll lQ lRAll dp =
  bindb (gen (list_union (funas_trs_list (rulesd i dp))
               (funas_args_trs_list (pairsa i dp)))
          [])
    (\ ops ->
      let {
        check_q = check_sl_Q (sl_I ops) (sl_L ops) lc (sl_C ops);
        check_ml =
          check_sl_model_lab_trs (sl_I ops) (sl_L ops) (sl_C ops)
            (\ a b -> a == b) lc;
        check_l =
          check_sl_lab_trs (sl_I ops) (sl_L ops) (sl_C ops) (\ a b -> a == b)
            lc;
        check_la = check_sl_lab (sl_I ops) (sl_L ops) lc (sl_C ops);
      } in sem_lab_proc ld i (Inr ()) check_q check_l check_la check_ml lPAll lQ
             lRAll dp);

slm_C :: forall a b c d. Slm_ops_ext a b c d -> [b];
slm_C (Slm_ops_ext slm_La slm_I slm_C slm_c slm_L more) = slm_C;

slm_I :: forall a b c d. Slm_ops_ext a b c d -> a -> [b] -> b;
slm_I (Slm_ops_ext slm_La slm_I slm_C slm_c slm_L more) = slm_I;

slm_L :: forall a b c d. Slm_ops_ext a b c d -> a -> [b] -> c;
slm_L (Slm_ops_ext slm_La slm_I slm_C slm_c slm_L more) = slm_La;

slm_c :: forall a b c d. Slm_ops_ext a b c d -> b;
slm_c (Slm_ops_ext slm_La slm_I slm_C slm_c slm_L more) = slm_c;

slm_La :: forall a b c d. Slm_ops_ext a b c d -> a -> [b] -> c;
slm_La (Slm_ops_ext slm_La slm_I slm_C slm_c slm_L more) = slm_L;

slm_to_sl :: forall a b c d. Slm_ops_ext a b c () -> Sl_ops_ext a b c d ();
slm_to_sl ops =
  Sl_ops_ext (slm_L ops) (\ _ _ _ -> True) (slm_I ops) (slm_C ops) (slm_c ops)
    (\ _ -> Inr ()) (slm_La ops) (\ _ _ _ -> True) (\ l -> [l]) (\ _ _ -> [])
    ();

slm_gen_to_sl_gen ::
  forall a b c d.
    ([(a, Nat)] ->
      [(a, Nat)] -> Sum ([Char] -> [Char]) (Slm_ops_ext a b c ())) ->
      [(a, Nat)] ->
        [(a, Nat)] -> Sum ([Char] -> [Char]) (Sl_ops_ext a b c d ());
slm_gen_to_sl_gen gen =
  (\ f g -> bindb (gen f g) (\ ops -> Inr (slm_to_sl ops)));

sl_La :: forall a b c d e. Sl_ops_ext a b c d e -> a -> [b] -> c;
sl_La (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
        sl_LS_gen more)
  = sl_L;

funas_root_rule_list ::
  forall a b. (Eq a) => (Term a b, Term a b) -> [(a, Nat)];
funas_root_rule_list r =
  remdups (option_to_list (root (fst r)) ++ option_to_list (root (snd r)));

funas_root_rules_list ::
  forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_root_rules_list r = remdups (concatMap funas_root_rule_list r);

check_no_defined_root ::
  forall a b.
    (Showa a,
      Showa b) => ((a, Nat) -> Bool) -> Term a b -> Sum ([Char] -> [Char]) ();
check_no_defined_root isdef t =
  check (not (isdef (the (root t))))
    (shows_string ['t', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' '] .
      shows_prec_term Zero_nat t .
        shows_string [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd']);

sem_lab_root_proc ::
  forall a b c d.
    (Eq a, Showa a, Eq d,
      Showa d) => (a -> (a, b)) ->
                    Dpp_ops_ext c a d () ->
                      Sum ([Char] -> [Char]) () ->
                        ([Term a d] ->
                          [Term a d] -> Sum ([Char] -> [Char]) ()) ->
                          ([(Term a d, Term a d)] ->
                            [(Term a d, Term a d)] ->
                              Sum ([Char] -> [Char]) ()) ->
                            ([(Term a d, Term a d)] ->
                              [(Term a d, Term a d)] ->
                                Sum ([Char] -> [Char]) ()) ->
                              ([(Term a d, Term a d)] ->
                                [(Term a d, Term a d)] ->
                                  Sum ([Char] -> [Char]) ()) ->
                                [(Term a d, Term a d)] ->
                                  [Term a d] ->
                                    [(Term a d, Term a d)] ->
                                      c -> Sum ([Char] -> [Char]) c;
sem_lab_root_proc ld i valid check_Q check_laba check_lab check_model_lab lPAll
  lQ lRAll dpp =
  let {
    r = rc i dpp;
    rw = rwc i dpp;
    pw = pwa i dpp;
    p = pa i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    (lP, (lPw, _)) = model_splitter ld lPAll pw;
    (lR, (lRw, _)) = model_splitter ld lRAll rw;
  } in (case bindb valid
               (\ _ ->
                 bindb (catcha
                         (forallm
                           (\ (l, ra) ->
                             bindb (check_no_var l)
                               (\ _ ->
                                 bindb (check_no_var ra)
                                   (\ _ ->
                                     check_no_defined_root
                                       (\ fn ->
 not (null (rules_mapb i dpp fn)))
                                       ra)))
                           (pairsa i dpp))
                         (\ x -> Inl (snd x)))
                   (\ _ ->
                     bindb (catcha
                             (forallm (\ (l, _) -> check_no_var l)
                               (rulesd i dpp))
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         let {
                           q = qc i dpp;
                         } in catcha
                                (bindb
                                  (check
                                    (if nfs then wwf_rulesa i dpp else True)
                                    (shows_prec_list Zero_nat
                                      ['w', 'e', 'l', 'l', ' ', 'f', 'o', 'r',
'm', 'e', 'd', 'n', 'e', 's', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e',
'd']))
                                  (\ _ ->
                                    bindb (check_Q lQ q)
                                      (\ _ ->
bindb (check_sl_Qa ld lQ q)
  (\ _ ->
    bindb (check_laba lP p)
      (\ _ ->
        bindb (check_laba lPw pw)
          (\ _ ->
            bindb (check_model_lab lR r)
              (\ _ ->
                bindb (check_model_lab lRw rw)
                  (\ _ ->
                    bindb (check_lab lR r) (\ _ -> check_lab lRw rw)))))))))
                                (\ x ->
                                  Inl (shows_string
 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd', 'u', 'r', 'i', 'n', 'g', ' ',
   'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ':'] .
shows_nl . x)))))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc i nfs m lP lPw lQ lR lRw);
       });

lab_root ::
  forall a b c d e.
    (a -> [b] -> b) ->
      (a -> [b] -> c) ->
        (a -> [b] -> c) ->
          (a -> Nat -> c -> d) -> (e -> b) -> Term a e -> Term d e;
lab_root i la l lc alpha (Fun f ts) =
  let {
    clts = map (eval_lab i la lc alpha) ts;
    cs = map fst clts;
    a = map snd clts;
  } in Fun (lc f (size_list ts) (l f cs)) a;
lab_root uu uv uw ux uy (Var x) = Var x;

check_sl_rule_root ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          [b] ->
                            [(Term d e, Term d e)] ->
                              (Term a e, Term a e) -> Sum ([Char] -> [Char]) ();
check_sl_rule_root i la l lc c lR lr =
  catcha
    (forallm
      (\ alpha ->
        let {
          laa = lab_root i la l lc alpha;
          lb = laa (fst lr);
          r = laa (snd lr);
        } in check (membera lR (lb, r))
               (shows_prec_list Zero_nat
                  ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                    ' '] .
                 shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                   [' ', '-', '>', ' '] (lb, r) .
                   shows_prec_list Zero_nat
                     [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g']))
      (map fun_of (enum_vectors c (vars_rule_list lr))))
    (\ x -> Inl (snd x));

check_sl_lab_root_trs ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        [b] ->
                          (a -> Nat -> c -> d) ->
                            [(Term d e, Term d e)] ->
                              [(Term a e, Term a e)] ->
                                Sum ([Char] -> [Char]) ();
check_sl_lab_root_trs i la l c label lP p =
  catcha (forallm (check_sl_rule_root i la l label c lP) p)
    (\ x -> Inl (snd x));

sem_lab_fin_root_proc ::
  forall a b c d e.
    (Eq a, Showa a, Eq d, Showa d, Eq e,
      Showa e) => (a -> Nat -> b -> a) ->
                    (a -> (a, b)) ->
                      Dpp_ops_ext c a d () ->
                        ([(a, Nat)] ->
                          [(a, Nat)] ->
                            Sum ([Char] -> [Char]) (Sl_ops_ext a e b d ())) ->
                          [(Term a d, Term a d)] ->
                            [Term a d] ->
                              [(Term a d, Term a d)] ->
                                c -> Sum ([Char] -> [Char]) c;
sem_lab_fin_root_proc lc ld i gen lPAll lQ lRAll dp =
  let {
    pairs = pairsa i dp;
  } in bindb (gen (list_union (funas_trs_list (rulesd i dp))
                    (funas_args_trs_list pairs))
               (funas_root_rules_list pairs))
         (\ ops ->
           let {
             check_q = check_sl_Q (sl_I ops) (sl_L ops) lc (sl_C ops);
             check_ml =
               check_sl_model_lab_trs (sl_I ops) (sl_L ops) (sl_C ops)
                 (\ a b -> a == b) lc;
             check_l =
               check_sl_lab_root_trs (sl_I ops) (sl_L ops) (sl_La ops)
                 (sl_C ops) lc;
             check_la = check_sl_lab (sl_I ops) (sl_L ops) lc (sl_C ops);
           } in sem_lab_root_proc ld i (Inr ()) check_q check_l check_la
                  check_ml lPAll lQ lRAll dp);

sl_LS :: forall a b c d e. Sl_ops_ext a b c d e -> a -> Nat -> c -> Bool;
sl_LS (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
        sl_LS_gen more)
  = sl_LSa;

sl_lgen :: forall a b c d e. Sl_ops_ext a b c d e -> c -> [c];
sl_lgen
  (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
    sl_LS_gen more)
  = sl_lgen;

sl_LS_gen :: forall a b c d e. Sl_ops_ext a b c d e -> a -> Nat -> [c];
sl_LS_gen
  (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
    sl_LS_gen more)
  = sl_LS_gen;

lge_to_lgr ::
  forall a b.
    (Eq b) => (a -> Nat -> b -> b -> Bool) ->
                (a -> Nat -> b -> Bool) -> a -> Nat -> b -> b -> Bool;
lge_to_lgr lge ls f n =
  let {
    lSfn = ls f n;
    lgefn = lge f n;
  } in (\ l la -> not (l == la) && lSfn l && lSfn la && lgefn l la);

check_sl_decr_rule ::
  forall a b c d.
    (Eq a, Eq b, Eq c,
      Eq d) => (a -> Nat -> b -> c) ->
                 (c -> (a, b)) ->
                   (a -> Nat -> b -> Bool) ->
                     (a -> Nat -> b -> b -> Bool) ->
                       (Term c d, Term c d) -> Bool;
check_sl_decr_rule lc ld ls lge (Fun lf ts, Fun lg us) =
  let {
    (f, l1) = ld lf;
    (g, l2) = ld lg;
    n = size_list ts;
  } in f == g &&
         ts == us &&
           lf == lc f n l1 &&
             lg == lc f n l2 &&
               ls f n l1 && ls f n l2 && lge_to_lgr lge ls f n l1 l2;
check_sl_decr_rule uu uv uw ux (Var vb, va) = False;
check_sl_decr_rule uu uv uw ux (v, Var vb) = False;

check_sl_decr ::
  forall a b c d.
    (Eq a, Eq b, Eq c, Showa c, Eq d,
      Showa d) => (a -> Nat -> b -> c) ->
                    (c -> (a, b)) ->
                      (a -> Nat -> b -> Bool) ->
                        (a -> Nat -> b -> b -> Bool) ->
                          [(Term c d, Term c d)] -> Sum ([Char] -> [Char]) ();
check_sl_decr lc ld ls lge d =
  catcha
    (forallm
      (\ lr ->
        check (check_sl_decr_rule lc ld ls lge lr)
          (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] lr .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'd', 'e', 'c',
                'r', 'e', 'a', 's', 'i', 'n', 'g', ' ', 'r', 'u', 'l', 'e']))
      d)
    (\ x -> Inl (snd x));

sl_check_decr ::
  forall a b c d e.
    Sl_ops_ext a b c d e -> [(Term a d, Term a d)] -> Sum ([Char] -> [Char]) ();
sl_check_decr
  (Sl_ops_ext sl_La sl_LSa sl_I sl_C sl_c sl_check_decr sl_L sl_LS sl_lgen
    sl_LS_gen more)
  = sl_check_decr;

check_wf_sym_F_all ::
  forall a b c.
    (Eq c,
      Showa c) => (a -> Nat -> b -> c) ->
                    (c -> (a, b)) ->
                      (a -> Nat -> b -> Bool) ->
                        (c, Nat) -> Sum ([Char] -> [Char]) ();
check_wf_sym_F_all lc ld ls =
  (\ (lf, n) ->
    let {
      (f, l) = ld lf;
    } in check (ls f n l && lf == lc f n l)
           (shows_prec_list Zero_nat
              ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 's', 'y', 'm', 'b', 'o',
                'l', ' '] .
             shows_prec Zero_nat lf .
               shows_prec_list Zero_nat
                 [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']));

check_wf_terms_F_all ::
  forall a b c d.
    (Eq c,
      Showa c) => (a -> Nat -> b -> c) ->
                    (c -> (a, b)) ->
                      (a -> Nat -> b -> Bool) ->
                        Term c d -> Sum ([Char] -> [Char]) ();
check_wf_terms_F_all lc ld ls lt =
  let {
    lfs = funas_term_lista lt;
  } in catcha (forallm (check_wf_sym_F_all lc ld ls) lfs) (\ x -> Inl (snd x));

check_Lab_all_trs ::
  forall a b c.
    (Eq a, Showa a, Eq c,
      Showa c) => (a -> Nat -> b -> a) ->
                    (a -> (a, b)) ->
                      (a -> Nat -> b -> Bool) ->
                        [(Term a c, Term a c)] ->
                          [(Term a c, Term a c)] -> Sum ([Char] -> [Char]) ();
check_Lab_all_trs lc ld ls lR r =
  catcha
    (forallm
      (\ (l, ra) ->
        catcha
          (bindb (check_wf_terms_F_all lc ld ls ra)
            (\ _ ->
              check (membera r (map_rule (\ lf -> fst (ld lf)) (l, ra)))
                (shows_prec_list Zero_nat
                  ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'o',
                    'f', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 'd',
                    'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'y', 'i', 'e', 'l',
                    'd', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                    'u', 'l', 'e'])))
          (\ x ->
            Inl (shows_prec_list Zero_nat
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                     ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l',
                     'e'] .
                  shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '-', '>', ' '] (l, ra) .
                    shows_nl . x)))
      lR)
    (\ x -> Inl (snd x));

check_sl_rule_all_ass ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          (c -> [c]) ->
                            [(Term d e, Term d e)] ->
                              (e -> b) ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_all_ass i lb la lc gen_smaller lR alpha (l, Fun f ts) =
  let {
    ll = lab_root i lb la lc alpha l;
    clts = map (eval_lab i lb lc alpha) ts;
    lts = map snd clts;
    ld = la f (map fst clts);
    n = size_list ts;
    small = gen_smaller ld;
  } in catcha
         (forallm
           (\ le ->
             check (membera lR (ll, Fun (lc f n le) lts))
               (shows_string
                  ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                    ' '] .
                 shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                   [' ', '-', '>', ' '] (ll, Fun (lc f n le) lts) .
                   shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']))
           small)
         (\ x -> Inl (snd x));
check_sl_rule_all_ass i lb la lc gen_smaller lR alpha (l, Var x) =
  let {
    ll = lab_root i lb la lc alpha l;
    lr = lab_root i lb la lc alpha (Var x);
  } in check (membera lR (ll, lr))
         (shows_string
            ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e', ' '] .
           shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] (ll, lr) .
             shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);

check_sl_rule_all ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          [b] ->
                            (c -> [c]) ->
                              [(Term d e, Term d e)] ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_all i la l lc c gen_smaller lR lr =
  catcha
    (forallm
      (\ alpha -> check_sl_rule_all_ass i la l lc gen_smaller lR alpha lr)
      (map fun_of (enum_vectors c (vars_rule_list lr))))
    (\ x -> Inl (snd x));

check_sl_lab_all_trs ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        [b] ->
                          (c -> [c]) ->
                            (a -> Nat -> c -> d) ->
                              [(Term d e, Term d e)] ->
                                [(Term a e, Term a e)] ->
                                  Sum ([Char] -> [Char]) ();
check_sl_lab_all_trs i la l c gen label lP p =
  catcha (forallm (check_sl_rule_all i la l label c gen lP) p)
    (\ x -> Inl (snd x));

flatten_term_enum :: forall a b. Term [a] b -> [Term a b];
flatten_term_enum (Var x) = [Var x];
flatten_term_enum (Fun fs ts) =
  let {
    lts = map flatten_term_enum ts;
    ss = concat_lists lts;
  } in concatMap (\ f -> map (Fun f) ss) fs;

lab_lhss_more_impl ::
  forall a b c d.
    (a -> Nat -> b -> c) -> (a -> Nat -> [b]) -> [Term a d] -> [Term c d];
lab_lhss_more_impl lc lS_gen q =
  let {
    f_all = (\ f n -> map (lc f n) (lS_gen f n));
  } in concatMap (\ qa -> flatten_term_enum (map_term_wa f_all qa)) q;

check_sl_lab_lhss_more ::
  forall a b c.
    (Eq a, Showa a, Eq c,
      Showa c) => (a -> Nat -> b -> a) ->
                    (a -> Nat -> [b]) ->
                      [Term a c] -> [Term a c] -> Sum ([Char] -> [Char]) ();
check_sl_lab_lhss_more lc lS_gen lQ q =
  catcha (check_NF_vars_subset (lab_lhss_more_impl lc lS_gen q) lQ)
    (\ x ->
      Inl (shows_term (shows_prec Zero_nat) (shows_prec Zero_nat) x .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i',
                'n', ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'Q']));

quasi_splitter ::
  forall a b c.
    (Eq a,
      Eq c) => (a -> (a, b)) ->
                 [(Term a c, Term a c)] ->
                   [(Term a c, Term a c)] ->
                     ([(Term a c, Term a c)],
                       ([(Term a c, Term a c)], [(Term a c, Term a c)]));
quasi_splitter ld lAll uRw =
  let {
    unlab = (\ lf -> fst (ld lf));
    la = map (\ r -> (r, map_rule unlab r)) lAll;
    (d, nD) =
      partition
        (\ (r, ur) ->
          equal_term (fst ur) (snd ur) && not (equal_term (fst r) (snd r)))
        la;
    (rw, r) = partition (\ (_, a) -> membera uRw a) nD;
  } in (map fst r, (map fst rw, map fst d));

sem_lab_quasi_root_proc ::
  forall a b c d.
    (Eq a, Showa a, Eq d,
      Showa d) => (a -> (a, b)) ->
                    Dpp_ops_ext c a d () ->
                      Sum ([Char] -> [Char]) () ->
                        ([(Term a d, Term a d)] -> Sum ([Char] -> [Char]) ()) ->
                          ([(Term a d, Term a d)] ->
                            Sum ([Char] -> [Char]) ()) ->
                            ([Term a d] ->
                              [Term a d] -> Sum ([Char] -> [Char]) ()) ->
                              ([(Term a d, Term a d)] ->
                                [(Term a d, Term a d)] ->
                                  Sum ([Char] -> [Char]) ()) ->
                                ([(Term a d, Term a d)] ->
                                  [(Term a d, Term a d)] ->
                                    Sum ([Char] -> [Char]) ()) ->
                                  ([(Term a d, Term a d)] ->
                                    [(Term a d, Term a d)] ->
                                      Sum ([Char] -> [Char]) ()) ->
                                    [(Term a d, Term a d)] ->
                                      [Term a d] ->
[(Term a d, Term a d)] -> c -> Sum ([Char] -> [Char]) c;
sem_lab_quasi_root_proc ld i valid check_decra check_decr check_lhss_more
  check_lab_all check_lab_all_trs check_model_lab lPAll lQ lRAll dpp =
  let {
    r = rc i dpp;
    rw = rwc i dpp;
    pw = pwa i dpp;
    p = pa i dpp;
    nfs = nfsc i dpp;
    m = minimal i dpp;
    (lP, (lPw, _)) = model_splitter ld lPAll pw;
    (lR, (lRw, d)) = quasi_splitter ld lRAll rw;
  } in (case bindb valid
               (\ _ ->
                 bindb (check (if nfs then wwf_rulesa i dpp else True)
                         (shows_prec_list Zero_nat
                           ['w', 'e', 'l', 'l', ' ', 'f', 'o', 'r', 'm', 'e',
                             'd', 'n', 'e', 's', 's', ' ', 'r', 'e', 'q', 'u',
                             'i', 'r', 'e', 'd']))
                   (\ _ ->
                     bindb (catcha
                             (forallm
                               (\ (l, ra) ->
                                 bindb (check_no_var l)
                                   (\ _ ->
                                     bindb (check_no_var ra)
                                       (\ _ ->
 check_no_defined_root (\ fn -> not (null (rules_mapb i dpp fn))) ra)))
                               (pairsa i dpp))
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         bindb (catcha
                                 (forallm (\ (l, _) -> check_no_var l)
                                   (rulesd i dpp))
                                 (\ x -> Inl (snd x)))
                           (\ _ ->
                             let {
                               q = qc i dpp;
                             } in bindb (if nfs then check_wf_trs d else Inr ())
                                    (\ _ ->
                                      bindb (check_decra d)
(\ _ ->
  bindb (check_decr d)
    (\ _ ->
      bindb (catcha
              (forallm
                (\ qa ->
                  check (linear_term qa)
                    (shows_prec_list Zero_nat
                      ['Q', ' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ',
                        'c', 'o', 'n', 't', 'a', 'i', 'n', ' ', 'n', 'o', 'n',
                        '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 't', 'e', 'r',
                        'm', 's']))
                q)
              (\ x -> Inl (snd x)))
        (\ _ ->
          catcha
            (bindb (check_lhss_more lQ q)
              (\ _ ->
                bindb (check_sl_Qa ld lQ q)
                  (\ _ ->
                    bindb (check_lab_all lP p)
                      (\ _ ->
                        bindb (check_lab_all lPw pw)
                          (\ _ ->
                            bindb (check_model_lab lR r)
                              (\ _ ->
                                bindb (check_model_lab lRw rw)
                                  (\ _ ->
                                    bindb (check_lab_all_trs lR r)
                                      (\ _ -> check_lab_all_trs lRw rw))))))))
            (\ x ->
              Inl (shows_string
                     ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'd', 'u', 'r',
                       'i', 'n', 'g', ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                       'g', ':'] .
                    shows_nl . x))))))))))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc i nfs m lP lPw lQ lR (lRw ++ d));
       });

sem_lab_fin_quasi_root_proc ::
  forall a b c d e.
    (Eq a, Showa a, Eq b, Showa c, Eq e,
      Showa e) => (a -> Nat -> b -> a) ->
                    (a -> (a, b)) ->
                      (c -> c -> Bool) ->
                        (a -> Nat -> b -> b -> Bool) ->
                          Dpp_ops_ext d a e () ->
                            ([(a, Nat)] ->
                              [(a, Nat)] ->
                                Sum ([Char] -> [Char])
                                  (Sl_ops_ext a c b e ())) ->
                              [(Term a e, Term a e)] ->
                                [Term a e] ->
                                  [(Term a e, Term a e)] ->
                                    d -> Sum ([Char] -> [Char]) d;
sem_lab_fin_quasi_root_proc lc ld cge lge i gen lPAll lQ lRAll dp =
  let {
    pairs = pairsa i dp;
  } in bindb (gen (list_union (funas_trs_list (rulesd i dp))
                    (funas_args_trs_list pairs))
               (funas_root_rules_list pairs))
         (\ ops ->
           let {
             check_d = sl_check_decr ops;
             check_da = check_sl_decr lc ld (sl_LS ops) lge;
             check_q = check_sl_lab_lhss_more lc (sl_LS_gen ops);
             check_ml =
               check_sl_model_lab_trs (sl_I ops) (sl_L ops) (sl_C ops) cge lc;
             check_l =
               check_sl_lab_all_trs (sl_I ops) (sl_L ops) (sl_La ops) (sl_C ops)
                 (sl_lgen ops) lc;
             check_la = check_Lab_all_trs lc ld (sl_LS ops);
           } in sem_lab_quasi_root_proc ld i (Inr ()) check_d check_da check_q
                  check_l check_la check_ml lPAll lQ lRAll dp);

semlab_fin_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a (Lab b [Nat]) c () ->
                    Sl_variant (Lab b [Nat]) c ->
                      [(Term (Lab b [Nat]) c, Term (Lab b [Nat]) c)] ->
                        [Term (Lab b [Nat]) c] ->
                          [(Term (Lab b [Nat]) c, Term (Lab b [Nat]) c)] ->
                            a -> Sum ([Char] -> [Char]) a;
semlab_fin_proc j (Rootlab Nothing) =
  sem_lab_fin_proc label label_decomp j (slm_gen_to_sl_gen (rl_slm Nothing));
semlab_fin_proc j (Rootlab (Just d)) =
  sem_lab_fin_root_proc label label_decomp j
    (slm_gen_to_sl_gen (rl_slm (Just d)));
semlab_fin_proc j (Finitelab sli) =
  sem_lab_fin_proc label label_decomp j
    (slm_gen_to_sl_gen (\ _ _ -> Inr (sli_to_slm sli)));
semlab_fin_proc j (QuasiFinitelab sli v) =
  sem_lab_fin_quasi_root_proc label label_decomp qmodel_cge qmodel_lge j
    (\ f g -> qsli_to_sl v f g sli);

check_critical_pairs_NF ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Bool, (Term a b, Term a b))] -> Sum ([Char] -> [Char]) ();
check_critical_pairs_NF r cp =
  catcha
    (forallm
      (\ (_, (s, t)) ->
        (if equal_term s t then Inr ()
          else catcha (check_join_NF r s t)
                 (\ x ->
                   Inl (shows_prec_list Zero_nat
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h',
                            'e', 'n', ' ', 'j', 'o', 'i', 'n', 'i', 'n', 'g',
                            ' ', 'c', 'r', 'i', 't', 'i', 'c', 'a', 'l', ' ',
                            'p', 'a', 'i', 'r', ' '] .
                         show_crit_pair (s, t) . shows_nl . x))))
      cp)
    (\ x -> Inl (snd x));

reachable_terms ::
  forall a b.
    (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Nat -> [Term a b];
reachable_terms r s Zero_nat = [s];
reachable_terms r s (Suc n) =
  let {
    ts = reachable_terms r s n;
  } in remdups (ts ++ concatMap (rewrite r) ts);

iterative_join_search_main ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 Term a b -> Term a b -> Nat -> Nat -> Bool;
iterative_join_search_main r s t i n =
  (if less_eq_nat i n
    then not (null (list_inter (reachable_terms r s i)
                     (reachable_terms r t i))) ||
           iterative_join_search_main r s t (Suc i) n
    else False);

iterative_join_search ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] -> Term a b -> Term a b -> Nat -> Bool;
iterative_join_search r s t n = iterative_join_search_main r s t Zero_nat n;

check_join_BFS_limit ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Nat ->
                    [(Term a b, Term a b)] ->
                      Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_join_BFS_limit n r s t =
  check (iterative_join_search r s t n)
    (shows_prec_list Zero_nat
       ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd',
         ' ', 'a', ' ', 'j', 'o', 'i', 'n', 'i', 'n', 'g', ' ', 's', 'e', 'q',
         'u', 'e', 'n', 'c', 'e', ' ', 'o', 'f', ' ', 'l', 'e', 'n', 'g', 't',
         'h', ' ', 'a', 't', ' ', 'm', 'o', 's', 't', ' '] .
      shows_prec_nat Zero_nat n .
        shows_prec_list Zero_nat
          [' ', 'f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 't', 'e', 'r', 'm', 's',
            ' '] .
          shows_prec_term Zero_nat s .
            shows_prec_list Zero_nat [' ', 'a', 'n', 'd', ' '] .
              shows_prec_term Zero_nat t . shows_nl);

check_critical_pairs_BFS ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Nat ->
                    [(Term a b, Term a b)] ->
                      [(Bool, (Term a b, Term a b))] ->
                        Sum ([Char] -> [Char]) ();
check_critical_pairs_BFS n r cp =
  catcha
    (forallm
      (\ (_, (s, t)) ->
        (if equal_term s t then Inr ()
          else catcha (check_join_BFS_limit n r s t)
                 (\ x ->
                   Inl (shows_prec_list Zero_nat
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h',
                            'e', 'n', ' ', 'j', 'o', 'i', 'n', 'i', 'n', 'g',
                            ' ', 'c', 'r', 'i', 't', 'i', 'c', 'a', 'l', ' ',
                            'p', 'a', 'i', 'r', ' '] .
                         show_crit_pair (s, t) . shows_nl . x))))
      cp)
    (\ x -> Inl (snd x));

rseq_last ::
  forall a b. Term a b -> [(Pos, ((Term a b, Term a b), Term a b))] -> Term a b;
rseq_last s steps = lasta (s : map (\ (_, (_, sa)) -> sa) steps);

check_prop_rstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Bool ->
                    (Term a b -> Sum ([Char] -> [Char]) ()) ->
                      [(Term a b, Term a b)] ->
                        Pos ->
                          (Term a b, Term a b) ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_prop_rstep nfs pa r p rule s t =
  check (any (\ ra ->
               eq_rule_mod_vars rule ra &&
                 isOK (check_prop_rstep_rule nfs pa p ra s t))
          r)
    (shows_prec_list Zero_nat
       ['t', 'h', 'e', ' ', 's', 't', 'e', 'p', ' ', 'f', 'r', 'o', 'm', ' '] .
      shows_prec_term Zero_nat s .
        shows_prec_list Zero_nat [' ', 't', 'o', ' '] .
          shows_prec_term Zero_nat t .
            shows_prec_list Zero_nat
              [' ', 'v', 'i', 'a', ' ', 'r', 'u', 'l', 'e', ' '] .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '-', '>', ' '] rule .
                shows_prec_list Zero_nat
                  [' ', 'a', 't', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n',
                    ' '] .
                  shows_prec_pos Zero_nat p .
                    shows_prec_list Zero_nat
                      [' ', 'i', 's', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm',
                        'a', 't', 'i', 'c'] .
                      shows_nl);

check_qrstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        Pos ->
                          (Term a b, Term a b) ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qrstep nf nfs =
  check_prop_rstep nfs
    (\ t ->
      check (nf t)
        (shows_prec_term Zero_nat t .
          shows_prec_list Zero_nat
            [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ', 'Q', '-',
              'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm']));

check_rqrstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        (Term a b, Term a b) ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rqrstep nf nfs r rule s t = check_qrstep nf nfs r Empty rule s t;

check_qsteps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qsteps nf nfs p r [] s u =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
         ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e'] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_nl .
            shows_string
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'] .
              shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_qsteps nf nfs p ra ((uu, (r, (True, t))) : prts) s u =
  bindb (check_rqrstep nf nfs p r s t)
    (\ _ -> check_qsteps nf nfs p ra prts t u);
check_qsteps nf nfs pa ra ((p, (r, (False, t))) : prts) s u =
  bindb (check_qrstep nf nfs ra p r s t)
    (\ _ -> check_qsteps nf nfs pa ra prts t u);

check_qrsteps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        [(Pos, ((Term a b, Term a b), Term a b))] ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qrsteps nf nfs r prts s u =
  check_qsteps nf nfs [] r (map (\ (p, (ra, t)) -> (p, (ra, (False, t)))) prts)
    s u;

check_rsteps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Pos, ((Term a b, Term a b), Term a b))] ->
                      Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rsteps = check_qrsteps (\ _ -> True) False;

check_rsteps_last ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Term a b ->
                      [(Pos, ((Term a b, Term a b), Term a b))] ->
                        Sum ([Char] -> [Char]) ();
check_rsteps_last = (\ r s steps -> check_rsteps r steps s (rseq_last s steps));

check_join ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Term a b ->
                      [(Pos, ((Term a b, Term a b), Term a b))] ->
                        Term a b ->
                          [(Pos, ((Term a b, Term a b), Term a b))] ->
                            Sum ([Char] -> [Char]) ();
check_join r s sseq t tseq =
  let {
    chk = check_rsteps_last r;
  } in bindb (chk s sseq)
         (\ _ ->
           bindb (chk t tseq)
             (\ _ ->
               let {
                 u = rseq_last s sseq;
                 v = rseq_last t tseq;
               } in check (equal_term u v)
                      (shows_prec_list Zero_nat
                         ['t', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
                           ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e', 's',
                           ' ', 'e', 'n', 'd', ' ', 'i', 'n', ' ', 'd', 'i',
                           'f', 'f', 'e', 'r', 'e', 'n', 't', ' ', 't', 'e',
                           'r', 'm', 's', ' '] .
                        shows_prec_term Zero_nat u .
                          shows_prec_list Zero_nat [' ', 'a', 'n', 'd', ' '] .
                            shows_prec_term Zero_nat v)));

check_critical_pairs_guided ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Bool, (Term a b, Term a b))] ->
                      [(Term a b,
                         ([(Pos, ((Term a b, Term a b), Term a b))],
                           (Term a b,
                             [(Pos, ((Term a b, Term a b), Term a b))])))] ->
                        Sum ([Char] -> [Char]) ();
check_critical_pairs_guided r cp joins =
  bindb (catcha
          (forallm (\ (s, (sseq, (a, b))) -> check_join r s sseq a b) joins)
          (\ x -> Inl (snd x)))
    (\ _ ->
      let {
        cpa = map (\ (s, (_, (t, _))) -> (s, t)) joins;
      } in catcha
             (forallm
               (\ (_, (s, t)) ->
                 check (equal_term s t ||
                         any (\ st ->
                               instance_rule (s, t) st ||
                                 instance_rule (t, s) st)
                           cpa)
                   (shows_prec_list Zero_nat
                      ['c', 'r', 'i', 't', 'i', 'c', 'a', 'l', ' ', 'p', 'a',
                        'i', 'r', ' '] .
                     show_crit_pair (s, t) .
                       shows_prec_list Zero_nat
                         [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']))
               cp)
             (\ x -> Inl (snd x)));

check_critical_pairs ::
  forall a.
    (Eq a,
      Showa a) => [(Term a [Char], Term a [Char])] ->
                    [(Bool, (Term a [Char], Term a [Char]))] ->
                      Join_info a -> Sum ([Char] -> [Char]) ();
check_critical_pairs r cp join_info =
  (case join_info of {
    Guided a -> check_critical_pairs_guided r cp a;
    Join_NF -> check_critical_pairs_NF r cp;
    Join_BFS n -> check_critical_pairs_BFS n r cp;
  });

switch_innermost_proc ::
  forall a b.
    (Eq b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    Join_info b -> a -> Sum ([Char] -> [Char]) a;
switch_innermost_proc i joins_i dpp =
  let {
    r = rwc i dpp;
    p = pa i dpp;
    pw = pwa i dpp;
    nfs = nfsc i dpp;
  } in (case bindb (catcha
                     (forallm
                       (\ (l, _) ->
                         check (not (is_Var l))
                           (shows_prec_list Zero_nat
                             ['l', 'e', 'f', 't', ' ', 'v', 'a', 'r', 'i', 'a',
                               'b', 'l', 'e', 's', ' ', 'i', 'n', ' ', 'R', ' ',
                               'f', 'o', 'r', 'b', 'i', 'd', 'd', 'e', 'n']))
                       (if nfs then r else []))
                     (\ x -> Inl (snd x)))
               (\ _ ->
                 bindb (check (minimal i dpp)
                         (shows_prec_list Zero_nat
                           ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y',
                             ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                   (\ _ ->
                     bindb (check (null (qc i dpp))
                             (shows_prec_list Zero_nat
                               ['n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y',
                                 ' ', 'Q', ' ', 'n', 'o', 't', ' ', 'y', 'e',
                                 't', ' ', 's', 'u', 'p', 'p', 'o', 'r', 't',
                                 'e', 'd']))
                       (\ _ ->
                         bindb (check (null (rc i dpp))
                                 (shows_prec_list Zero_nat
                                   ['s', 't', 'r', 'i', 'c', 't', ' ', 'r', 'u',
                                     'l', 'e', 's', ' ', 'n', 'o', 't', ' ',
                                     'a', 'l', 'l', 'o', 'w', 'e', 'd']))
                           (\ _ ->
                             bindb (check
                                     (null (critical_pairs_impl (p ++ pw) r))
                                     (shows_prec_list Zero_nat
                                       ['o', 'v', 'e', 'r', 'l', 'a', 'p', 's',
 ' ', 'b', 'e', 't', 'w', 'e', 'e', 'n', ' ', 'P', ' ', 'a', 'n', 'd', ' ', 'R',
 ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']))
                               (\ _ ->
                                 check_critical_pairs r
                                   (critical_pairs_impl r r) joins_i)))))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc i nfs True p pw (map fst r) [] r);
       });

rpo_strict_unbounded ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat -> a -> Nat -> (Bool, Bool), a -> Bool) ->
                    (a -> Nat -> Order_tag) ->
                      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
rpo_strict_unbounded pr c =
  (\ (s, t) ->
    check (fst (rpo_unbounded pr c s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', 'R', 'P', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

rpo_nstrict_unbounded ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat -> a -> Nat -> (Bool, Bool), a -> Bool) ->
                    (a -> Nat -> Order_tag) ->
                      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
rpo_nstrict_unbounded pr c =
  (\ (s, t) ->
    check (snd (rpo_unbounded pr c s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', '=', 'R', 'P', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

create_RPO_redtriple ::
  forall a b c.
    (Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([((a, Nat), (Nat, Order_tag))] ->
                    (b -> Nat -> Nat, b -> Nat -> Order_tag)) ->
                    [((a, Nat), (Nat, Order_tag))] -> Redtriple_ext b c ();
create_RPO_redtriple prec_repr_to_pr pr =
  let {
    (p, tau) = prec_repr_to_pr pr;
    ns = rpo_nstrict_unbounded (prc_nat p, prl_nat p) tau;
  } in Redtriple_ext (Inr ()) (rpo_strict_unbounded (prc_nat p, prl_nat p) tau)
         ns ns full_af (\ _ -> Inr ()) (shows_rpo_repr pr) (\ _ -> Nothing) ();

filtered_fun :: forall a. Filtered a -> a;
filtered_fun (FPair f n) = f;

shows_prec_filtered ::
  forall a. (Showa a) => Nat -> Filtered a -> [Char] -> [Char];
shows_prec_filtered d ff = shows_prec Zero_nat (filtered_fun ff);

shows_list_filtered :: forall a. (Showa a) => [Filtered a] -> [Char] -> [Char];
shows_list_filtered ffs = shows_list_aux (shows_prec_filtered Zero_nat) ffs;

instance (Showa a) => Showa (Filtered a) where {
  shows_prec = shows_prec_filtered;
  shows_list = shows_list_filtered;
};

afs_check ::
  forall a b.
    (Showa a,
      Showa b) => ([Char] -> [Char]) ->
                    (a -> Nat -> Af_entry) ->
                      ((Term (Filtered a) b, Term (Filtered a) b) ->
                        Sum ([Char] -> [Char]) ()) ->
                        (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
afs_check r pia g lr =
  let {
    pl = afs_term pia (fst lr);
    pr = afs_term pia (snd lr);
  } in catcha (g (pl, pr))
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                    'i', 'e', 'n', 't', ' '] .
                 shows_prec_term Zero_nat (fst lr) .
                   shows_string [' '] .
                     r . shows_string [' '] .
                           shows_prec_term Zero_nat (snd lr) .
                             shows_nl .
                               shows_string ['p', 'i', '(', ' '] .
                                 shows_prec_term Zero_nat (fst lr) .
                                   shows_string [' ', ')', ' ', '=', ' '] .
                                     shows_prec_term Zero_nat pl .
                                       shows_nl .
 shows_string ['p', 'i', '(', ' '] .
   shows_prec_term Zero_nat (snd lr) .
     shows_string [' ', ')', ' ', '=', ' '] .
       shows_prec_term Zero_nat pr . shows_nl . x));

shows_afs :: forall a. (Showa a) => [((a, Nat), Af_entry)] -> [Char] -> [Char];
shows_afs af =
  shows_map
    (\ (a, b) ->
      let {
        (f, n) = a;
      } in (\ e ->
             shows_string ['p', 'i', '('] .
               shows_prec Zero_nat f .
                 shows_string ['/'] .
                   shows_prec_nat Zero_nat n .
                     shows_string [')', ' ', '=', ' '] .
                       (case e of {
                         Collapse i -> shows_prec_nat Zero_nat (Suc i);
                         AFList ids -> shows_prec_list Zero_nat (map Suc ids);
                       }) .
                         shows_nl)
        b)
    af;

check_wf_afs ::
  forall a. (Showa a) => [((a, Nat), Af_entry)] -> Sum ([Char] -> [Char]) ();
check_wf_afs pia =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ e ->
               catcha
                 (case e of {
                   Collapse i -> check (less_nat i n) i;
                   AFList ids ->
                     catcha
                       (forallm
                         (\ x -> (if less_nat x n then Inr () else Inl x)) ids)
                       (\ x -> Inl (snd x));
                 })
                 (\ x ->
                   Inl (shows_string ['i', 'n', 'd', 'e', 'x', ' '] .
                         shows_prec_nat Zero_nat x .
                           shows_string
                             [' ', 't', 'o', 'o', ' ', 'h', 'i', 'g', 'h', ' ',
                               'f', 'o', 'r', ' ', 'A', 'F', '-', 'e', 'n', 't',
                               'r', 'y', ' ', 'o', 'f'] .
                             shows_prec_prod Zero_nat (f, n) . shows_nl)))
          b)
      pia)
    (\ x -> Inl (snd x));

afs_list_to_af ::
  forall a. (Eq a) => [((a, Nat), Af_entry)] -> (a, Nat) -> Nat -> Bool;
afs_list_to_af pia fn =
  (case map_of pia fn of {
    Nothing -> (\ _ -> True);
    Just (Collapse a) -> equal_nat a;
    Just (AFList a) -> membera a;
  });

mono_af_entry :: Nat -> Af_entry -> Bool;
mono_af_entry n (Collapse i) = less_eq_nat n one_nat;
mono_af_entry n (AFList ids) = all_interval_nat (membera ids) Zero_nat n;

check_mono_afs_help ::
  forall a b c. (Showa a) => ((a, b), c) -> [Char] -> [Char];
check_mono_afs_help ((f, a), e) =
  shows_string ['e', 'n', 't', 'r', 'y', ' ', 'f', 'o', 'r', ' '] .
    shows_prec Zero_nat f .
      shows_string
        [' ', 'n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i', 'c'] .
        shows_nl;

check_mono_afs ::
  forall a. (Showa a) => [((a, Nat), Af_entry)] -> Sum ([Char] -> [Char]) ();
check_mono_afs pia =
  catcha
    (catcha
      (forallm
        (\ x -> (if mono_af_entry (snd (fst x)) (snd x) then Inr () else Inl x))
        pia)
      (\ x -> Inl (snd x)))
    (\ x -> Inl (check_mono_afs_help x));

default_af_entry :: Nat -> Af_entry;
default_af_entry n = AFList (upt Zero_nat n);

afs_list_to_afs ::
  forall a. (Eq a) => [((a, Nat), Af_entry)] -> a -> Nat -> Af_entry;
afs_list_to_afs pia f n =
  (case map_of pia (f, n) of {
    Nothing -> default_af_entry n;
    Just entry -> entry;
  });

af_redtriple ::
  forall a b.
    (Eq a, Showa a,
      Showa b) => [((a, Nat), Af_entry)] ->
                    Redtriple_ext (Filtered a) b () -> Redtriple_ext a b ();
af_redtriple pia rp =
  let {
    af = afs_list_to_af pia;
    afs = afs_list_to_afs pia;
  } in Redtriple_ext (bindb (check_wf_afs pia) (\ _ -> valid rp))
         (afs_check (shows_string ['>']) afs (s rp))
         (afs_check (shows_string ['>', '=']) afs (ns rp))
         (afs_check (shows_string ['>', '=']) afs (nst rp)) af
         (\ cs ->
           bindb (check_mono_afs pia) (\ _ -> mono rp (map (afs_rule afs) cs)))
         (shows_string
            ['A', 'r', 'g', 'u', 'm', 'e', 'n', 't', ' ', 'F', 'i', 'l', 't',
              'e', 'r', ':', ' '] .
           shows_nl . shows_afs pia . shows_nl . desca rp)
         (\ _ -> Nothing) ();

equal_filtered :: forall a. (Eq a) => Filtered a -> Filtered a -> Bool;
equal_filtered (FPair fa nata) (FPair f nat) = fa == f && equal_nat nata nat;

instance (Eq a) => Eq (Filtered a) where {
  a == b = equal_filtered a b;
};

prec_weight_repr_to_prec_weight ::
  forall a.
    (Eq a,
      Showa a) => ([((a, Nat), (Nat, Nat))], Nat) ->
                    (Sum ([Char] -> [Char]) (),
                      ((a, Nat) -> Nat, ((a, Nat) -> Nat, (Nat, a -> Bool))));
prec_weight_repr_to_prec_weight prw_w0 =
  let {
    (prw, w0) = prw_w0;
    w_fun =
      (\ fn ->
        (case map_of prw fn of {
          Nothing -> Suc w0;
          Just (_, w) -> w;
        }));
    p_fun =
      (\ fn ->
        (case map_of prw fn of {
          Nothing -> Zero_nat;
          Just (p, _) -> p;
        }));
    fs = map fst prw;
    p_max = max_list (map p_fun fs);
    cw_okay =
      catcha
        (forallm
          (\ fn ->
            check (if equal_nat (snd fn) Zero_nat then less_eq_nat w0 (w_fun fn)
                    else True)
              (shows_prec_list Zero_nat
                 ['w', 'e', 'i', 'g', 'h', 't', ' ', 'o', 'f', ' ', 'c', 'o',
                   'n', 's', 't', 'a', 'n', 't', ' '] .
                shows_prec Zero_nat (fst fn) .
                  shows_prec_list Zero_nat
                    [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'a', 't', ' ',
                      'l', 'e', 'a', 's', 't', ' ', 'w', '0']))
          (map fst prw))
        (\ x -> Inl (snd x));
    adm = catcha
            (forallm
              (\ fn ->
                check (if equal_nat (snd fn) one_nat
                        then (if equal_nat (w_fun fn) Zero_nat
                               then equal_nat (p_fun fn) p_max else True)
                        else True)
                  (shows_prec_list Zero_nat
                     ['u', 'n', 'a', 'r', 'y', ' ', 's', 'y', 'm', 'b', 'o',
                       'l', ' '] .
                    shows_prec Zero_nat (fst fn) .
                      shows_prec_list Zero_nat
                        [' ', 'w', 'i', 't', 'h', ' ', 'w', 'e', 'i', 'g', 'h',
                          't', ' ', '0', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
                          't', ' ', 'h', 'a', 'v', 'e', ' ', 'm', 'a', 'x', 'i',
                          'm', 'a', 'l', ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e',
                          'n', 'c', 'e']))
              (map fst prw))
            (\ x -> Inl (snd x));
    cs = map_filter
           (\ x ->
             (if equal_nat (snd x) Zero_nat && equal_nat (w_fun x) w0
               then Just (fst x) else Nothing))
           fs;
    p_min = min_list (map (\ c -> p_fun (c, Zero_nat)) cs);
    lcs = filter (\ c -> equal_nat (p_fun (c, Zero_nat)) p_min) cs;
    least = membera lcs;
    ok = bindb (check (less_nat Zero_nat w0)
                 (shows_prec_list Zero_nat
                   ['w', '0', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'l',
                     'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ',
                     '0']))
           (\ _ -> bindb adm (\ _ -> cw_okay));
  } in (ok, (p_fun, (w_fun, (w0, least))));

create_KBO_redtriple ::
  forall a b c.
    (Showa a, Eq b, Showa b, Eq c,
      Showa c) => (([((a, Nat), (Nat, Nat))], Nat) ->
                    ([((b, Nat), (Nat, Nat))], Nat)) ->
                    ([((a, Nat), (Nat, Nat))], Nat) -> Redtriple_ext b c ();
create_KBO_redtriple f_to_g pr =
  let {
    (ch, (p, (w, (w0, least)))) = prec_weight_repr_to_prec_weight (f_to_g pr);
    ns = kbo_nstrict p w w0 least;
    s = kbo_strict p w w0 least;
  } in Redtriple_ext ch s ns ns full_af (\ _ -> Inr ()) (shows_kbo_repr pr)
         (\ _ -> Nothing) ();

shows_arctic :: Arctic -> [Char] -> [Char];
shows_arctic (Num_arc i) = shows_prec_int Zero_nat i;
shows_arctic MinInfty = shows_string ['-', 'i', 'n', 'f'];

shows_prec_arctic :: Nat -> Arctic -> [Char] -> [Char];
shows_prec_arctic d ai = shows_arctic ai;

shows_list_arctic :: [Arctic] -> [Char] -> [Char];
shows_list_arctic asa = shows_list_aux (shows_prec_arctic Zero_nat) asa;

instance Showa Arctic where {
  shows_prec = shows_prec_arctic;
  shows_list = shows_list_arctic;
};

instance Poly_carrier Integer where {
};

instance Poly_carrier Rat where {
};

shows_arctic_rat :: Arctic_rat -> [Char] -> [Char];
shows_arctic_rat (Num_arc_rat i) = shows_prec_rat Zero_nat i;
shows_arctic_rat MinInfty_rat = shows_string ['-', 'i', 'n', 'f'];

shows_prec_arctic_rat :: Nat -> Arctic_rat -> [Char] -> [Char];
shows_prec_arctic_rat d ari = shows_arctic_rat ari;

shows_list_arctic_rat :: [Arctic_rat] -> [Char] -> [Char];
shows_list_arctic_rat asa = shows_list_aux (shows_prec_arctic_rat Zero_nat) asa;

instance Showa Arctic_rat where {
  shows_prec = shows_prec_arctic_rat;
  shows_list = shows_list_arctic_rat;
};

check_poly_ge ::
  forall a b.
    (Eq a,
      Ordered_semiring_0 b) => [([(a, Nat)], b)] -> [([(a, Nat)], b)] -> Bool;
check_poly_ge [] q = all (\ (_, a) -> ge zero a) q;
check_poly_ge ((m, c) : p) q =
  (case extracta (\ nd -> eq_monom (fst nd) m) q of {
    Nothing -> ge c zero && check_poly_ge p q;
    Just (q1, ((_, d), q2)) -> ge c d && check_poly_ge p (q1 ++ q2);
  });

check_poly_gt ::
  forall a b.
    (Ordered_semiring_0 a,
      Eq b) => (a -> a -> Bool) ->
                 [([(b, Nat)], a)] -> [([(b, Nat)], a)] -> Bool;
check_poly_gt gt p q =
  let {
    (a1, p1) = poly_split [] p;
    (b1, q1) = poly_split [] q;
  } in gt a1 b1 && check_poly_ge p1 q1;

poly_inter_to_af ::
  forall a b. ((a, Nat) -> [([(Nat, Nat)], b)]) -> (a, Nat) -> Nat -> Bool;
poly_inter_to_af i fn = poly_var (i fn);

poly_vars_list :: forall a b. (Eq a) => [([(a, Nat)], b)] -> [a];
poly_vars_list p = remdups (concatMap (map fst . fst) p);

check_poly_weak_mono_all ::
  forall a b. (Ordered_semiring_0 b) => [([(a, Nat)], b)] -> Bool;
check_poly_weak_mono_all p = all (\ (_, c) -> ge c zero) p;

check_poly_weak_mono_discrete ::
  forall a b. (Eq a, Eq b, Poly_carrier b) => [([(a, Nat)], b)] -> a -> Bool;
check_poly_weak_mono_discrete p v =
  check_poly_ge
    (poly_subst
      (\ w -> poly_of (if w == v then PSum [PNum one, PVar v] else PVar w)) p)
    p;

check_poly_weak_mono_and_pos ::
  forall a b. (Eq a, Eq b, Poly_carrier b) => Bool -> [([(a, Nat)], b)] -> Bool;
check_poly_weak_mono_and_pos discrete p =
  (if discrete
    then all (check_poly_weak_mono_discrete p) (poly_vars_list p) &&
           ge (eval_poly (\ _ -> zero) p) zero
    else check_poly_weak_mono_all p);

check_poly_inter_list ::
  forall a b.
    (Eq b,
      Poly_carrier b) => Bool ->
                           [((a, Nat), [([(Nat, Nat)], b)])] ->
                             Sum (a, [([(Nat, Nat)], b)]) ();
check_poly_inter_list discrete i =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (_, a) = x;
              } in check_poly_weak_mono_and_pos discrete a
            then Inr () else Inl x))
        i)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (let {
             (a, b) = x;
           } in let {
                  (f, _) = a;
                } in (\ aa -> (f, aa))
                  b));

poly_inter_list_to_inter ::
  forall a b.
    (Poly_carrier a,
      Eq b) => a -> [((b, Nat), [([(Nat, Nat)], a)])] ->
                      (b, Nat) -> [([(Nat, Nat)], a)];
poly_inter_list_to_inter def i =
  (\ (f, n) ->
    (case map_of i (f, n) of {
      Nothing -> default_I def n;
      Just p -> p;
    }));

check_monom_strict_mono :: forall a. (Eq a) => Bool -> [(a, Nat)] -> a -> Bool;
check_monom_strict_mono pm m v =
  not (null m) &&
    null (tl m) &&
      fst (hd m) == v &&
        (if pm then less_eq_nat one_nat (snd (hd m))
          else equal_nat (snd (hd m)) one_nat);

check_poly_strict_mono ::
  forall a b. (Eq a, Poly_carrier b) => Bool -> [([(a, Nat)], b)] -> a -> Bool;
check_poly_strict_mono pm p v =
  any (\ (m, c) -> ge c one && check_monom_strict_mono pm m v) p;

check_poly_strict_mono_discrete ::
  forall a b.
    (Eq a, Poly_carrier a,
      Eq b) => (a -> a -> Bool) -> [([(b, Nat)], a)] -> b -> Bool;
check_poly_strict_mono_discrete gt p v =
  check_poly_gt gt
    (poly_subst
      (\ w -> poly_of (if w == v then PSum [PNum one, PVar v] else PVar w)) p)
    p;

check_poly_strict_mono_smart ::
  forall a b.
    (Eq a, Poly_carrier a,
      Eq b) => Bool ->
                 Bool -> (a -> a -> Bool) -> [([(b, Nat)], a)] -> b -> Bool;
check_poly_strict_mono_smart discrete pm gt p v =
  (if discrete then check_poly_strict_mono_discrete gt p v
    else check_poly_strict_mono pm p v);

create_nlpoly_redtriple ::
  forall a b c.
    (Eq a, Poly_carrier a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => a -> (a -> a -> Bool) ->
                         Bool ->
                           Bool ->
                             [((b, Nat), [([(Nat, Nat)], a)])] ->
                               Redtriple_ext b c ();
create_nlpoly_redtriple def gt power_mono discrete i =
  let {
    j = poly_inter_list_to_inter def i;
    x = poly_subst
          (\ n -> poly_of (PVar (['x', '_'] ++ shows_prec_nat Zero_nat n [])));
    ns = (\ (s, t) ->
           let {
             p = eval_term j s;
             q = eval_term j t;
           } in check (check_poly_ge p q)
                  (shows_string
                     ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e',
                       'n', 's', 'u', 'r', 'e', ' '] .
                    shows_poly p .
                      shows_string [' ', '>', '=', ' '] . shows_poly q));
  } in Redtriple_ext
         (catcha (check_poly_inter_list discrete i)
           (\ xa ->
             Inl (let {
                    (f, p) = xa;
                  } in shows_string
                         ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
                           'i', 'o', 'n', ' '] .
                         shows_poly (x p) .
                           shows_string [' ', 'o', 'f', ' '] .
                             shows_prec Zero_nat f .
                               shows_string
                                 [' ', 'i', 'n', 'v', 'a', 'l', 'i', 'd',
                                   ' '])))
         (\ (s, t) ->
           let {
             p = eval_term j s;
             q = eval_term j t;
           } in check (check_poly_gt gt p q)
                  (shows_string
                     ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e',
                       'n', 's', 'u', 'r', 'e', ' '] .
                    shows_poly p . shows_string [' ', '>', ' '] . shows_poly q))
         ns ns (poly_inter_to_af j)
         (\ _ ->
           catcha
             (catcha
               (forallm
                 (\ xa ->
                   (if let {
                         (a, b) = xa;
                       } in let {
                              (_, n) = a;
                            } in (\ p ->
                                   all_interval_nat
                                     (check_poly_strict_mono_smart discrete
                                       power_mono gt p)
                                     Zero_nat n)
                              b
                     then Inr () else Inl xa))
                 i)
               (\ xa -> Inl (snd xa)))
             (\ xa ->
               Inl (let {
                      (a, b) = xa;
                    } in let {
                           (f, _) = a;
                         } in (\ p ->
                                shows_string
                                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                    ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ', 'm',
                                    'o', 'n', 'o', 't', 'o', 'n', 'i', 'c', 't',
                                    'y', ' ', 'o', 'f', ' '] .
                                  shows_poly (x p) .
                                    shows_string
                                      [' ', 'a', 's', ' ', 'i', 'n', 't', 'e',
'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                                      shows_prec Zero_nat f)
                           b)))
         (shows_string
            ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i', 'n',
              't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'] .
           shows_nl .
             shows_sep
               (\ (a, b) ->
                 let {
                   (f, n) = a;
                 } in (\ p ->
                        shows_string ['P', 'o', 'l', '('] .
                          shows_prec Zero_nat f .
                            shows_string ['/'] .
                              shows_prec_nat Zero_nat n .
                                shows_string [')', ' ', '=', ' '] .
                                  shows_poly (x p))
                   b)
               shows_nl i)
         (\ _ -> Nothing) ();

create_afI ::
  forall a b.
    (Eq a, Eq b) => a -> [((b, Nat), (a, [a]))] -> (b, Nat) -> Nat -> Bool;
create_afI ze i fn =
  (case map_of i fn of {
    Nothing -> (\ _ -> True);
    Just (_, coeffs) ->
      let {
        len = size_list coeffs;
      } in (\ ia -> less_nat ia len && not (nth coeffs ia == ze));
  });

list_sumI :: forall a. a -> (a -> a -> a) -> [a] -> a;
list_sumI ze pl [] = ze;
list_sumI ze pl (x : xs) = pl x (list_sumI ze pl xs);

mul_pvarsI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> [(b, a)] -> [(b, a)];
mul_pvarsI uu uv a [] = [];
mul_pvarsI ze ti a ((x, b) : vas) =
  let {
    p = ti a b;
    res = mul_pvarsI ze ti a vas;
  } in (if p == ze then res else (x, p) : res);

mul_lpolyI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> L_poly b a -> L_poly b a;
mul_lpolyI ze ti a (LPoly b vas) = LPoly (ti a b) (mul_pvarsI ze ti a vas);

add_varI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> b -> a -> [(b, a)] -> [(b, a)];
add_varI uu uv x a [] = [(x, a)];
add_varI ze pl x a ((y, b) : vas) =
  (if x == y then let {
                    s = pl a b;
                  } in (if s == ze then vas else (x, s) : vas)
    else (y, b) : add_varI ze pl x a vas);

sum_pvarsI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> [(b, a)] -> [(b, a)] -> [(b, a)];
sum_pvarsI uu uv [] vbs = vbs;
sum_pvarsI ze pl ((x, a) : vas) vbs =
  (if a == ze then sum_pvarsI ze pl vas vbs
    else sum_pvarsI ze pl vas (add_varI ze pl x a vbs));

sum_lpolyI ::
  forall a b.
    (Eq a,
      Eq b) => a -> (a -> a -> a) -> L_poly b a -> L_poly b a -> L_poly b a;
sum_lpolyI ze pl (LPoly a vas) (LPoly b vbs) =
  LPoly (pl a b) (sum_pvarsI ze pl vas vbs);

pleftI ::
  forall a b c.
    (Eq a,
      Eq c) => a -> a -> (a -> a -> a) ->
                           (a -> a -> a) ->
                             (a -> a) ->
                               ((b, Nat) -> (a, [a])) -> Term b c -> L_poly c a;
pleftI ze on pl ti mz pia (Var x) = LPoly ze [(x, on)];
pleftI ze on pl ti mz pia (Fun f ts) =
  let {
    (c, asa) = pia (f, size_list ts);
  } in (case sum_lpolyI ze pl (LPoly c [])
               (list_sumI (LPoly ze []) (sum_lpolyI ze pl)
                 (map (\ at ->
                        mul_lpolyI ze ti (fst at)
                          (pleftI ze on pl ti mz pia (snd at)))
                   (zip asa ts)))
         of {
         LPoly d [] -> LPoly (mz d) [];
         LPoly d (ab : lista) -> LPoly d (ab : lista);
       });

prightI ::
  forall a b c.
    (Eq a,
      Eq c) => a -> a -> (a -> a -> a) ->
                           (a -> a -> a) ->
                             (a -> a) ->
                               ((b, Nat) -> (a, [a])) -> Term b c -> L_poly c a;
prightI ze on pl ti mz pia (Var x) = LPoly ze [(x, on)];
prightI ze on pl ti mz pia (Fun f ts) =
  let {
    (c, asa) = pia (f, size_list ts);
    (LPoly d a) =
      sum_lpolyI ze pl (LPoly c [])
        (list_sumI (LPoly ze []) (sum_lpolyI ze pl)
          (map (\ at ->
                 mul_lpolyI ze ti (fst at)
                   (prightI ze on pl ti mz pia (snd at)))
            (zip asa ts)));
  } in LPoly (mz d) a;

show_pvars :: forall a b. (Eq a, Showa a, Showa b) => a -> [(b, a)] -> [[Char]];
show_pvars on [] = [];
show_pvars on ((x, c) : vas) =
  ((if on == c then [] else shows_prec Zero_nat c []) ++
    shows_prec Zero_nat x []) :
    show_pvars on vas;

shows_lpoly ::
  forall a b.
    (Eq a, Showa a, Showa b) => a -> a -> L_poly b a -> [Char] -> [Char];
shows_lpoly ze on (LPoly c cs) =
  (case show_pvars on cs of {
    [] -> shows_prec Zero_nat c;
    a : list ->
      (if c == ze then id
        else shows_prec Zero_nat c . shows_string [' ', '+', ' ']) .
        shows_list_gen (shows_prec_list Zero_nat) [] [] [' ', '+', ' '] []
          (a : list);
  });

lookup_rest :: forall a b. (Eq a) => a -> [(a, b)] -> Maybe (b, [(a, b)]);
lookup_rest x [] = Nothing;
lookup_rest x ((y, c) : ycs) =
  (if x == y then Just (c, ycs)
    else (case lookup_rest x ycs of {
           Nothing -> Nothing;
           Just (d, yccs) -> Just (d, (y, c) : yccs);
         }));

check_pvarsI ::
  forall a b.
    (Eq b) => a -> (a -> a -> Bool) -> [(b, a)] -> [(b, a)] -> Sum b ();
check_pvarsI ze geq vas [] =
  catcha (forallm (\ va -> check (geq (snd va) ze) (fst va)) vas)
    (\ x -> Inl (snd x));
check_pvarsI ze geq vas ((x, b) : vbs) =
  let {
    (a, vasa) =
      (case lookup_rest x vas of {
        Nothing -> (ze, vas);
        Just (a, ba) -> (a, ba);
      });
  } in bindb (check (geq a b) x) (\ _ -> check_pvarsI ze geq vasa vbs);

check_lpoly_sI ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Bool ->
                    a -> a -> (a -> a -> Bool) ->
                                (a -> a -> Bool) ->
                                  L_poly b a ->
                                    L_poly b a -> Sum ([Char] -> [Char]) ();
check_lpoly_sI mode ze on geq gt (LPoly a vas) (LPoly b vbs) =
  catcha
    (bindb
      (check (gt a b)
        (shows_string
          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
            'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c', 'o', 'n', 's',
            't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't']))
      (\ _ ->
        catcha (check_pvarsI ze (if mode then geq else gt) vas vbs)
          (\ x ->
            Inl (shows_string
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                     ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                     'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                     'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                     ' '] .
                  shows_prec Zero_nat x))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
               'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' '] .
            shows_lpoly ze on (LPoly a vas) .
              shows_string [' ', '>', ' '] .
                shows_lpoly ze on (LPoly b vbs) . shows_nl . x));

check_polo_sI ::
  forall a b c.
    (Eq a, Showa a, Showa b, Eq c,
      Showa c) => Bool ->
                    a -> a -> (a -> a -> a) ->
                                (a -> a -> a) ->
                                  (a -> a) ->
                                    (a -> a -> Bool) ->
                                      (a -> a -> Bool) ->
((b, Nat) -> (a, [a])) -> (Term b c, Term b c) -> Sum ([Char] -> [Char]) ();
check_polo_sI mode ze on pl ti mz geg gt pia (s, t) =
  let {
    left = pleftI ze on pl ti mz pia s;
    right = prightI ze on pl ti mz pia t;
  } in catcha (check_lpoly_sI mode ze on geg gt left right)
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'n',
                    's', 'u', 'r', 'e', ' '] .
                 shows_prec_term Zero_nat s .
                   shows_string [' ', '>', ' '] .
                     shows_prec_term Zero_nat t . shows_nl . x));

check_lpoly_nsI ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> a -> (a -> a -> Bool) ->
                              L_poly b a ->
                                L_poly b a -> Sum ([Char] -> [Char]) ();
check_lpoly_nsI ze on geq (LPoly a vas) (LPoly b vbs) =
  catcha
    (bindb
      (check (geq a b)
        (shows_string
          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
            'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c', 'o', 'n', 's',
            't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't', 's']))
      (\ _ ->
        catcha (check_pvarsI ze geq vas vbs)
          (\ x ->
            Inl (shows_string
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                     ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                     'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                     'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                     ' '] .
                  shows_prec Zero_nat x))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
               'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' '] .
            shows_lpoly ze on (LPoly a vas) .
              shows_string [' ', '>', '=', ' '] .
                shows_lpoly ze on (LPoly b vbs) . shows_nl . x));

check_polo_nsI ::
  forall a b c.
    (Eq a, Showa a, Showa b, Eq c,
      Showa c) => a -> a -> (a -> a -> a) ->
                              (a -> a -> a) ->
                                (a -> a) ->
                                  (a -> a -> Bool) ->
                                    ((b, Nat) -> (a, [a])) ->
                                      (Term b c, Term b c) ->
Sum ([Char] -> [Char]) ();
check_polo_nsI ze on pl ti mz geq pia (s, t) =
  let {
    left = pleftI ze on pl ti mz pia s;
    right = prightI ze on pl ti mz pia t;
  } in catcha (check_lpoly_nsI ze on geq left right)
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'n',
                    's', 'u', 'r', 'e', ' '] .
                 shows_prec_term Zero_nat s .
                   shows_string [' ', '>', '=', ' '] .
                     shows_prec_term Zero_nat t . shows_nl . x));

to_lpoly_inter ::
  forall a b.
    (Eq b) => a -> a -> [((b, Nat), (a, [a]))] -> (b, Nat) -> (a, [a]);
to_lpoly_inter d on i fn =
  (case map_of i fn of {
    Nothing -> (d, replicatea (snd fn) on);
    Just cas -> cas;
  });

check_poly_monoI ::
  forall a b.
    (Showa a,
      Showa b) => a -> (a -> a -> Bool) ->
                         (a -> Bool) ->
                           [((b, Nat), (a, [a]))] -> Sum ([Char] -> [Char]) ();
check_poly_monoI ze geq cm =
  (\ xs ->
    catcha
      (forallm
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ (c, cs) ->
                 catcha
                   (bindb
                     (check (geq c ze)
                       (shows_string
                          ['c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p',
                            'a', 'r', 't', ' '] .
                         shows_prec Zero_nat c .
                           shows_string
                             [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'a',
                               't', ' ', 'l', 'e', 'a', 's', 't', ' '] .
                             shows_prec Zero_nat ze . shows_nl))
                     (\ _ ->
                       bindb (check (less_eq_nat n (size_list cs))
                               (shows_string
                                  ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ',
                                    'a', 'r', 'g', 'u', 'm', 'e', 'n', 't', ' ',
                                    'i', 's', ' ', 'i', 'g', 'n', 'o', 'r', 'e',
                                    'd'] .
                                 shows_nl))
                         (\ _ ->
                           catcha
                             (forallm
                               (\ d ->
                                 check (cm d)
                                   (shows_string
                                      ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i',
'e', 'n', 't', ' '] .
                                     shows_prec Zero_nat d .
                                       shows_string
 [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd'] .
 shows_nl))
                               cs)
                             (\ x -> Inl (snd x)))))
                   (\ x ->
                     Inl (shows_string
                            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                              't', 'h', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n',
                              'i', 'c', 'i', 't', 'y', ' ', 'd', 'u', 'e', ' ',
                              't', 'o', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r',
                              'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                              ' '] .
                           shows_prec Zero_nat f .
                             shows_string ['/'] .
                               shows_prec_nat Zero_nat n . shows_nl . x)))
            b)
        xs)
      (\ x -> Inl (snd x)));

create_lpoly_repr ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Poly_c_ext a () -> [((b, Nat), (a, [a]))] -> [Char] -> [Char];
create_lpoly_repr c i =
  let {
    pia = to_lpoly_inter (defaulta c) (onea c) i;
  } in shows_string
         ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i', 'n', 't',
           'e', 'r', 'p', 'r', 'e', 't', 'r', 'a', 't', 'i', 'o', 'n', ' ', 'o',
           'v', 'e', 'r', ' '] .
         shows_string (description c) .
           shows_nl .
             shows_sep
               (\ (f, n) ->
                 let {
                   t = Fun f (map Var
                               (fresh_strings_list ['x', '_'] one_nat [] n));
                 } in shows_string ['P', 'o', 'l', '('] .
                        shows_prec_term Zero_nat t .
                          shows_string [')', ' ', '=', ' '] .
                            shows_lpoly (zeroa c) (onea c)
                              (pleftI (zeroa c) (onea c) (plusa c) (timesa c)
                                (max0a c) pia t))
               shows_nl (remdups (map fst i));

check_lpoly_coeffs ::
  forall a b.
    (Showa a,
      Showa b) => (a -> Bool) ->
                    (a -> Bool) ->
                      (a -> Bool) ->
                        [((b, Nat), (a, [a]))] -> Sum ([Char] -> [Char]) ();
check_lpoly_coeffs wfe apos ch i =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ (c, cs) ->
               catcha
                 (bindb
                   (check (wfe c)
                     (shows_string
                        ['c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a',
                          'r', 't', ' '] .
                       shows_prec Zero_nat c .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e',
                             'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd'] .
                           shows_nl))
                   (\ _ ->
                     bindb (check (apos c || any apos (takea n cs))
                             (shows_string
                                ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                  ' ', 'f', 'i', 'n', 'd', ' ', 'p', 'o', 's',
                                  'i', 't', 'i', 'v', 'e', ' ', 'e', 'n', 't',
                                  'r', 'y', ' ', 'w', 'h', 'i', 'c', 'h', ' ',
                                  'i', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                                  'e', 'd', ' ', 'f', 'o', 'r', ' ', 'a', 'r',
                                  'c', 't', 'i', 'c', ' ', 'i', 'n', 't', 'e',
                                  'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o',
                                  'n', 's'] .
                               shows_nl))
                       (\ _ ->
                         catcha
                           (forallm
                             (\ aa ->
                               check (ch aa && wfe aa)
                                 (shows_string
                                    ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i',
                                      'e', 'n', 't', ' '] .
                                   shows_prec Zero_nat aa .
                                     shows_string
                                       [' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
 'a', 'l', 'l', 'o', 'w', 'e', 'd'] .
                                       shows_nl))
                             cs)
                           (\ x -> Inl (snd x)))))
                 (\ x ->
                   Inl (shows_string
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                            't', 'h', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r',
                            'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                            ' '] .
                         shows_prec Zero_nat f .
                           shows_string ['/'] .
                             shows_prec_nat Zero_nat n . shows_nl . x)))
          b)
      i)
    (\ x -> Inl (snd x));

check_poly_mono_npsmI ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> [(b, Nat)] ->
                         [((b, Nat), (a, [a]))] -> Sum ([Char] -> [Char]) ();
check_poly_mono_npsmI ze f pia =
  bindb (catcha
          (forallm
            (\ (a, b) ->
              let {
                (fa, n) = a;
              } in (\ (c, cs) ->
                     catcha
                       (bindb
                         (check
                           (if equal_nat n (Suc Zero_nat) then c == ze
                             else True)
                           (shows_string
                              ['c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p',
                                'a', 'r', 't', ' '] .
                             shows_prec Zero_nat c .
                               shows_string
                                 [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ',
                                   'z', 'e', 'r', 'o', ' '] .
                                 shows_nl))
                         (\ _ ->
                           bindb (check (equal_nat n (size_list cs))
                                   (shows_string
                                      ['t', 'h', 'e', ' ', 'a', 'r', 'i', 't',
'y', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 't', 'h', 'e', ' ', 's', 'a', 'm',
'e', ' ', 'a', 's', ' ', 't', 'h', 'e', ' ', 'n', 'u', 'm', 'b', 'e', 'r', ' ',
'o', 'f', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n', 't', 's'] .
                                     shows_nl))
                             (\ _ ->
                               check (less_eq_nat n (Suc Zero_nat))
                                 (shows_string
                                    ['s', 'y', 'm', 'b', 'o', 'l', ' ', 'h',
                                      'a', 's', ' ', 'a', 'r', 'i', 't', 'y',
                                      ' ', 'l', 'a', 'r', 'g', 'e', 'r', ' ',
                                      't', 'h', 'a', 'n', ' ', '1'] .
                                   shows_nl))))
                       (\ x ->
                         Inl (shows_string
                                ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                  'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                                  'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'd',
                                  'u', 'e', ' ', 't', 'o', ' ', 'i', 'n', 't',
                                  'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i',
                                  'o', 'n', ' ', 'o', 'f', ' '] .
                               shows_prec Zero_nat fa .
                                 shows_string ['/'] .
                                   shows_prec_nat Zero_nat n . shows_nl . x)))
                b)
            pia)
          (\ x -> Inl (snd x)))
    (\ _ ->
      catcha (check_subseteq f (map fst pia))
        (\ x ->
          Inl (shows_string
                 ['u', 'n', 'k', 'n', 'o', 'w', 'n', ' ', 'i', 'n', 't', 'e',
                   'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'f',
                   'o', 'r', ' '] .
                shows_prec_prod Zero_nat x . shows_nl)));

derive_complexity :: forall a b. Poly_c_ext a b -> a -> Maybe Nat;
derive_complexity
  (Poly_c_ext wf zero one defaulta plus times max0 ge gt mono description
    arc_pos check_mono maxa derive_complexity more)
  = derive_complexity;

smaller_one_complexity :: forall a b. Poly_c_ext a b -> a -> Maybe Nat;
smaller_one_complexity c bc =
  (if gea c (onea c) bc then Just Zero_nat else derive_complexity c bc);

poly_c_max_inter_bcoeff ::
  forall a b.
    a -> a -> (a -> a -> Bool) ->
                (a -> a -> a) -> [(b, Nat)] -> ((b, Nat) -> (a, [a])) -> a;
poly_c_max_inter_bcoeff ze on geq ma f pia =
  foldra ma
    (concatMap (\ fn -> filter (\ b -> not (geq on b)) (snd (pia fn))) f) ze;

convert_lpoly_complexity ::
  forall a b c d.
    Poly_c_ext a b ->
      ((c, Nat) -> (a, [a])) ->
        Complexity_measure c d -> Maybe Complexity_class;
convert_lpoly_complexity c pia cc =
  let {
    f = (case cc of {
          Derivational_Complexity f -> f;
          Runtime_Complexity ca _ -> ca;
        });
    bc = poly_c_max_inter_bcoeff (zeroa c) (onea c) (gea c) (maxb c) f pia;
  } in mapa (\ n -> Comp_Poly (Suc n)) (smaller_one_complexity c bc);

create_poly_redtriple ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([((a, Nat), (b, [b]))] -> Sum ([Char] -> [Char]) ()) ->
                    Poly_c_ext b () ->
                      [((a, Nat), (b, [b]))] -> Redtriple_ext a c ();
create_poly_redtriple cI c i =
  let {
    pia = to_lpoly_inter (defaulta c) (onea c) i;
    ns = check_polo_nsI (zeroa c) (onea c) (plusa c) (timesa c) (max0a c)
           (gea c) pia;
  } in Redtriple_ext
         (bindb (cI i)
           (\ _ ->
             check_lpoly_coeffs (wf c) (arc_pos c) (\ x -> gea c x (zeroa c))
               i))
         (check_polo_sI (monoa c) (zeroa c) (onea c) (plusa c) (timesa c)
           (max0a c) (gea c) (gt c) pia)
         ns ns (create_afI (zeroa c) i)
         (\ s_ns_nst ->
           (if monoa c then check_poly_monoI (zeroa c) (gea c) (check_mono c) i
             else check_poly_mono_npsmI (zeroa c) (funas_trs_list s_ns_nst) i))
         (create_lpoly_repr c i)
         (if monoa c then convert_lpoly_complexity c pia else (\ _ -> Nothing))
         ();

prec_repr_to_pr ::
  forall a.
    (Eq a) => [((a, Nat), (Nat, Order_tag))] -> Filtered a -> Nat -> Nat;
prec_repr_to_pr prs (FPair f a) uu =
  (case map_of prs (f, a) of {
    Nothing -> Zero_nat;
    Just aa -> fst aa;
  });

faulty_redtriple ::
  forall a b. (Showa a, Showa b) => [Char] -> Redtriple_ext a b ();
faulty_redtriple s =
  Redtriple_ext (Inl (shows_string s)) (\ _ -> Inr ()) (\ _ -> Inr ())
    (\ _ -> Inr ()) full_af (\ _ -> Inr ()) id (\ _ -> Nothing) ();

prec_repr_to_status ::
  forall a.
    (Eq a) => [((a, Nat), (Nat, Order_tag))] -> Filtered a -> Nat -> Order_tag;
prec_repr_to_status prs (FPair f a) uu =
  (case map_of prs (f, a) of {
    Nothing -> Lex;
    Just aa -> snd aa;
  });

filter_prec_weight_repr ::
  forall a.
    (a -> Nat -> Af_entry) ->
      ([((a, Nat), (Nat, Nat))], Nat) ->
        ([((Filtered a, Nat), (Nat, Nat))], Nat);
filter_prec_weight_repr pia (prw, w0) =
  let {
    fprw =
      filter
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ _ ->
                 (case pia f n of {
                   Collapse _ -> False;
                   AFList _ -> True;
                 }))
            b)
        prw;
    mprw =
      map (\ (a, b) ->
            let {
              (f, n) = a;
            } in (\ aa ->
                   ((FPair f n,
                      (case pia f n of {
                        Collapse _ -> Zero_nat;
                        AFList ab -> size_list ab;
                      })),
                     aa))
              b)
        fprw;
  } in (mprw, w0);

get_redtriple ::
  forall a b.
    (Eq a, Showa a, Eq b, Showa b) => Redtriple_impl a -> Redtriple_ext a b ();
get_redtriple (Int_carrier i) =
  create_poly_redtriple (\ _ -> Inr ()) int_components i;
get_redtriple (Int_nl_carrier i) =
  create_nlpoly_redtriple 1 (\ x y -> y < x) True True i;
get_redtriple (Rat_carrier i) =
  create_poly_redtriple (\ _ -> Inr ()) rat_components i;
get_redtriple (Rat_nl_carrier d i) =
  (if less_rat zero_rat d
    then create_nlpoly_redtriple d (rat_gt d) (less_eq_rat one_rat d) False i
    else faulty_redtriple
           ['d', 'e', 'l', 't', 'a', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
             ' ', 'l', 'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ',
             'z', 'e', 'r', 'o']);
get_redtriple (Arctic_carrier i) =
  create_poly_redtriple (\ _ -> Inr ()) arctic_components i;
get_redtriple (Arctic_rat_carrier i) =
  create_poly_redtriple (\ _ -> Inr ()) arctic_rat_components i;
get_redtriple (Int_mat_carrier n sd i) =
  create_poly_redtriple
    (\ _ ->
      check (less_nat Zero_nat sd && less_eq_nat sd n)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's',
            'i', 'o', 'n']))
    (mat_components n sd int_components) i;
get_redtriple (Rat_mat_carrier n sd i) =
  create_poly_redtriple
    (\ _ ->
      check (less_nat Zero_nat sd && less_eq_nat sd n)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's',
            'i', 'o', 'n']))
    (mat_components n sd rat_components) i;
get_redtriple (Arc_mat_carrier n i) =
  create_poly_redtriple
    (\ _ ->
      check (less_nat Zero_nat n)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's',
            'i', 'o', 'n']))
    (Poly_c_ext (mat n n) (mat0I zero_arctic n n)
      (mat1I zero_arctic one_arctic n)
      (mat1I zero_arctic (defaulta arctic_components) n) (mat_plusI plus_arctic)
      (mat_multI zero_arctic plus_arctic times_arctic n) id mat_ge
      (mat_comp_all (gt arctic_components)) False
      (concat
        [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
            ' '],
          description arctic_components])
      (mat_arc_posI (arc_pos arctic_components)) (\ _ -> False)
      (mat_plusI plus_arctic) (\ _ -> Nothing) ())
    i;
get_redtriple (Arc_rat_mat_carrier n i) =
  create_poly_redtriple
    (\ _ ->
      check (less_nat Zero_nat n)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n', 's',
            'i', 'o', 'n']))
    (Poly_c_ext (mat n n) (mat0I zero_arctic_rat n n)
      (mat1I zero_arctic_rat one_arctic_rat n)
      (mat1I zero_arctic_rat (defaulta arctic_rat_components) n)
      (mat_plusI plus_arctic_rat)
      (mat_multI zero_arctic_rat plus_arctic_rat times_arctic_rat n) id mat_ge
      (mat_comp_all (gt arctic_rat_components)) False
      (concat
        [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
          [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
            ' '],
          description arctic_rat_components])
      (mat_arc_posI (arc_pos arctic_rat_components)) (\ _ -> False)
      (mat_plusI plus_arctic_rat) (\ _ -> Nothing) ())
    i;
get_redtriple (Rpo prec_tau pia) =
  af_redtriple pia
    (create_RPO_redtriple (\ pr -> (prec_repr_to_pr pr, prec_repr_to_status pr))
      prec_tau);
get_redtriple (Kbo precw pia) =
  af_redtriple pia
    (create_KBO_redtriple (filter_prec_weight_repr (afs_list_to_afs pia))
      precw);

intersect_rulesb ::
  forall a b c d. Dpp_ops_ext a b c d -> a -> [(Term b c, Term b c)] -> a;
intersect_rulesb
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = intersect_rules;

usable_rules_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    [(Term b [Char], Term b [Char])] ->
                      a -> Sum ([Char] -> [Char]) a;
usable_rules_proc i u dpp =
  (case bindb (check (nFQ_subset_NF_rulesb i dpp)
                (shows_prec_list Zero_nat
                  ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r', 'e',
                    'w', 'r', 'i', 't', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u',
                    'i', 'r', 'e', 'd']))
          (\ _ ->
            bindb (check (nfsc i dpp || minimal i dpp || wwf_rulesa i dpp)
                    (shows_prec_list Zero_nat
                      ['n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm',
                        ' ', 's', 'u', 'b', 's', 't', ',', ' ', 'm', 'i', 'n',
                        'i', 'm', 'a', 'l', 'i', 't', 'y', ' ', 'o', 'r', ' ',
                        'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd',
                        'n', 'e', 's', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                        'e', 'd']))
              (\ _ ->
                let {
                  p = pairsa i dpp;
                  urc = is_ur_closed_impl_dpp_mv i dpp u;
                  check_urc =
                    (\ s t ->
                      check (urc s t)
                        (shows_prec_list Zero_nat ['t', 'e', 'r', 'm', ' '] .
                          shows_prec_term Zero_nat t .
                            shows_prec_list Zero_nat
                              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'l',
                                'o', 's', 'e', 'd', ' ', 'u', 'n', 'd', 'e',
                                'r', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                                'r', 'u', 'l', 'e', 's']));
                } in bindb (catcha
                             (forallm
                               (\ (l, r) ->
                                 bindb (catcha
 (check_subseteq (vars_term_lista r) (vars_term_lista l))
 (\ _ ->
   Inl (shows_prec_list Zero_nat
         ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o', 'n', 'd', 'i',
           't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 'P', ' ', 'v', 'i', 'o', 'l',
           'a', 't', 'e', 'd'])))
                                   (\ _ -> check_urc [l] r))
                               p)
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         catcha (forallm (\ (l, a) -> check_urc (args l) a) u)
                           (\ x -> Inl (snd x)))))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (intersect_rulesb i dpp u);
  });

forward_instantiation_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    (Term b [Char], Term b [Char]) ->
                      [(Term b [Char], Term b [Char])] ->
                        Maybe [(Term b [Char], Term b [Char])] ->
                          a -> Sum ([Char] -> [Char]) a;
forward_instantiation_proc i st sts u_opt dpp =
  (case let {
          isnf = is_QNFc i dpp;
          (s, t) = st;
          iedg = is_iedg_edge_dpp i dpp (s, t);
          sy = map_vars (\ a -> 'y' : a) s;
          ty = map_vars (\ a -> 'y' : a) t;
          u = (case u_opt of {
                Nothing -> rulesd i dpp;
                Just u -> u;
              });
        } in bindb (if is_none u_opt then Inr ()
                     else let {
                            urc = is_ur_closed_impl_dpp_mv i dpp u;
                            check_urc =
                              (\ sa ta ->
                                check (urc sa ta)
                                  (shows_prec_list Zero_nat
                                     ['t', 'e', 'r', 'm', ' '] .
                                    shows_prec_term Zero_nat ta .
                                      shows_prec_list Zero_nat
[' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'l', 'o', 's', 'e', 'd', ' ', 'u',
  'n', 'd', 'e', 'r', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l',
  'e', 's']));
                          } in bindb (check (nfsc i dpp || minimal i dpp)
                                       (shows_prec_list Zero_nat
 ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y', ' ', 'o', 'r', ' ', 'n',
   'o', 'r', 'm', 'a', 'l', ' ', 's', 'u', 'b', 's', 't', ' ', 'r', 'e', 'q',
   'u', 'i', 'r', 'e', 'd']))
                                 (\ _ ->
                                   bindb (check (nFQ_subset_NF_rulesb i dpp)
   (shows_prec_list Zero_nat
     ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', ' ', 'r', 'e', 'w', 'r', 'i',
       't', 'i', 'n', 'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                                     (\ _ ->
                                       bindb
 (catcha (forallm (\ (l, a) -> check_urc (args l) a) u) (\ x -> Inl (snd x)))
 (\ _ ->
   bindb (check_urc [s] t)
     (\ _ ->
       (if nfsc i dpp then Inr ()
         else catcha (check_subseteq (vars_term_lista t) (vars_term_lista s))
                (\ _ ->
                  Inl (shows_prec_list Zero_nat
                        ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'c', 'o',
                          'n', 'd', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ',
                          'p', 'a', 'i', 'r', ' ', 'v', 'i', 'o', 'l', 'a', 't',
                          'e', 'd']))))))))
               (\ _ ->
                 let {
                   ur = map (\ (l, r) -> (r, l)) u;
                   ic = icap_impla ur [] [];
                 } in catcha
                        (forallm
                          (\ (ua, v) ->
                            (case mgu_class (ic ua) t of {
                              Nothing -> Inr ();
                              Just mu ->
                                check (not (isnf (subst_apply_term sy mu)) ||
not (isnf (subst_apply_term (map_vars (\ a -> 'x' : a) ua) mu)) ||
  any (\ sta ->
        instance_rule sta st &&
          instance_rule (subst_apply_term sy mu, subst_apply_term ty mu) sta)
    sts)
                                  (shows_prec_list Zero_nat
                                     ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
                                       't', ' ', 'f', 'i', 'n', 'd', ' ', 'i',
                                       'n', 's', 't', 'a', 'n', 'c', 'e', ' ',
                                       'o', 'f', ' ', 'p', 'a', 'i', 'r', ' '] .
                                    shows_rule (shows_prec Zero_nat)
                                      (shows_prec_list Zero_nat)
                                      [' ', '-', '>', ' ']
                                      (subst_apply_term sy mu,
subst_apply_term ty mu) .
                                      shows_nl .
shows_prec_list Zero_nat
  ['w', 'h', 'i', 'c', 'h', ' ', 'r', 'e', 's', 'u', 'l', 't', 'e', 'd', ' ',
    'f', 'r', 'o', 'm', ' ', 'D', 'P', ' '] .
  shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] (ua, v));
                            }))
                          (filter (\ (ua, _) -> iedg ua) (pairsa i dpp)))
                        (\ x -> Inl (snd x)))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (replace_paira i dpp st sts);
  });

check_weak ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ((a, Nat) -> Nat) ->
                    (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_weak p r =
  catcha
    (check (equal_term (proj_term p (fst r)) (proj_term p (snd r)))
      (shows_string
         ['t', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
           'l', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'q', 'u',
           'a', 'l', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'j',
           'e', 'c', 't', 'e', 'd', ' ', 'r', 'h', 's'] .
        shows_nl))
    (\ x ->
      Inl (shows_string
             ['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
               'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] r .
              shows_string [',', ' ', 's', 'i', 'n', 'c', 'e'] .
                shows_nl .
                  shows_prec_term Zero_nat (proj_term p (fst r)) .
                    shows_string [' ', '!', '=', ' '] .
                      shows_prec_term Zero_nat (proj_term p (snd r)) .
                        shows_nl . x));

q_emptyc :: forall a b c d. Dpp_ops_ext a b c d -> a -> Bool;
q_emptyc
  (Dpp_ops_ext dpp p pw pairs q r rw rules q_empty rules_no_left_var
    rules_non_collapsing is_QNF nFQ_subset_NF_rules rules_map reverse_rules_map
    intersect_pairs replace_pair intersect_rules delete_P_Pw delete_R_Rw
    split_pairs split_rules mk minimal nfs wwf_rules more)
  = q_empty;

check_strict_rstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    ((Term a b, Term a b) ->
                      Maybe [(Pos, ((Term a b, Term a b), Term a b))]) ->
                      ((a, Nat) -> Nat) ->
                        (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_strict_rstep ra rseqm p r =
  let {
    s = proj_term p (fst r);
    t = proj_term p (snd r);
  } in (case rseqm r of {
         Nothing -> check_supt s t;
         Just rseq ->
           (if equal_nat (size_list rseq) Zero_nat then check_supt s t
             else bindb (check_rsteps_last ra s rseq)
                    (\ _ -> check_supteq (rseq_last s rseq) t));
       });

check_strict_qrstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        ((Term a b, Term a b) ->
                          Maybe [(Pos, ((Term a b, Term a b), Term a b))]) ->
                          ((a, Nat) -> Nat) ->
                            (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_strict_qrstep nf nfs ra rseqm p r =
  let {
    s = proj_term p (fst r);
    t = proj_term p (snd r);
  } in (case rseqm r of {
         Nothing -> check_supt s t;
         Just [] ->
           Inl (shows_string
                 ['m', 'o', 'r', 'e', ' ', 't', 'h', 'a', 'n', ' ', 'a', ' ',
                   's', 'i', 'n', 'g', 'l', 'e', ' ', 'Q', '-', 'r', 'e', 'w',
                   'r', 'i', 't', 'e', ' ', 's', 't', 'e', 'p', ' ', 'i', 's',
                   ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']);
         Just [(pos, (rule, u))] ->
           bindb (check_qrstep nf nfs ra pos rule s u)
             (\ _ -> check_supteq u t);
         Just ((pos, (rule, u)) : _ : _) ->
           Inl (shows_string
                 ['m', 'o', 'r', 'e', ' ', 't', 'h', 'a', 'n', ' ', 'a', ' ',
                   's', 'i', 'n', 'g', 'l', 'e', ' ', 'Q', '-', 'r', 'e', 'w',
                   'r', 'i', 't', 'e', ' ', 's', 't', 'e', 'p', ' ', 'i', 's',
                   ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']);
       });

subterm_criterion_proc ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Dpp_ops_ext a b c () ->
                    ((b, Nat) -> Nat) ->
                      ((Term b c, Term b c) ->
                        Maybe [(Pos, ((Term b c, Term b c), Term b c))]) ->
                        [(Term b c, Term b c)] -> a -> Sum ([Char] -> [Char]) a;
subterm_criterion_proc i p rseqm prm dpp =
  (case let {
          pa = pairsa i dpp;
          nfs = nfsc i dpp;
          r = rulesd i dpp;
          pb = snd (split_pairsa i dpp prm);
        } in bindb (catcha
                     (forallm
                       (\ (l, ra) ->
                         bindb (check_no_var l)
                           (\ _ ->
                             bindb (check_no_var ra)
                               (\ _ ->
                                 check_no_defined_root
                                   (\ fn -> not (null (rules_mapb i dpp fn)))
                                   ra)))
                       pa)
                     (\ x -> Inl (snd x)))
               (\ _ ->
                 bindb (check (minimal i dpp)
                         (shows_prec_list Zero_nat
                           ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y',
                             ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                   (\ _ ->
                     bindb (check (if nfs then wwf_rulesa i dpp else True)
                             (shows_prec_list Zero_nat
                               ['r', 'e', 'q', 'u', 'i', 'r', 'e', ' ', 'w',
                                 'e', 'a', 'k', 'l', 'y', ' ', 'w', 'e', 'l',
                                 'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd', ' ',
                                 'q', 't', 'r', 's']))
                       (\ _ ->
                         bindb (if q_emptyc i dpp
                                 then catcha
(forallm (check_strict_rstep r rseqm p) prm) (\ x -> Inl (snd x))
                                 else catcha
(forallm (check_strict_qrstep (is_QNFc i dpp) nfs r rseqm p) prm)
(\ x -> Inl (snd x)))
                           (\ _ ->
                             catcha (forallm (check_weak p) pb)
                               (\ x -> Inl (snd x))))))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (delete_P_Pwa i dpp prm prm);
  });

label_s_ns_impl ::
  forall a b.
    (Showa a,
      Showa b) => ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                      (Term a b, Nat) -> (Term a b, Nat) -> (Bool, Bool);
label_s_ns_impl cS cNS s t =
  let {
    (sa, i) = s;
    (ta, j) = t;
  } in (if isOK (cS (sa, ta)) then (True, True)
         else (if isOK (cNS (sa, ta)) then (less_nat j i, less_eq_nat j i)
                else (False, False)));

s_label_mul_impl ::
  forall a b.
    (Showa a,
      Showa b) => (((Term a b, Nat) -> (Term a b, Nat) -> (Bool, Bool)) ->
                    [(Term a b, Nat)] -> [(Term a b, Nat)] -> (Bool, Bool)) ->
                    ((a, Nat) -> [(Nat, Nat)]) ->
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
s_label_mul_impl list_ext af cS cNS st =
  (case st of {
    (Var _, b) ->
      Inl (shows_prec_list Zero_nat
             ['r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] st .
              shows_prec_list Zero_nat
                [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o', 'n',
                  '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']);
    (Fun f ss, Var _) ->
      Inl (shows_prec_list Zero_nat
             ['r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] st .
              shows_prec_list Zero_nat
                [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o', 'n',
                  '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']);
    (Fun f ss, Fun g ts) ->
      check (fst (list_ext (label_s_ns_impl cS cNS) (lterms af (Fun f ss))
                   (lterms af (Fun g ts))))
        (shows_prec_list Zero_nat
           ['c', 'a', 'n', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e', 'n', 't',
             ' ', 'p', 'a', 'i', 'r', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
            [' ', '-', '>', ' '] st .
            shows_prec_list Zero_nat
              [' ', 's', 't', 'r', 'i', 'c', 't', 'l', 'y', ':'] .
              shows_nl .
                shows_prec_list Zero_nat (lterms af (Fun f ss)) .
                  shows_prec_list Zero_nat [' ', '>', 'm', 'u', ' '] .
                    shows_prec_list Zero_nat (lterms af (Fun g ts)) .
                      shows_prec_list Zero_nat
                        [' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                          'b', 'e', ' ', 'e', 'n', 's', 'u', 'r', 'e', 'd']);
  });

nST_label_mul_impl ::
  forall a b.
    (Showa a,
      Showa b) => (((Term a b, Nat) -> (Term a b, Nat) -> (Bool, Bool)) ->
                    [(Term a b, Nat)] -> [(Term a b, Nat)] -> (Bool, Bool)) ->
                    ((a, Nat) -> [(Nat, Nat)]) ->
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) ()) ->
                          (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
nST_label_mul_impl list_ext af cS cNS st =
  (case st of {
    (Var _, b) ->
      Inl (shows_prec_list Zero_nat
             ['r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] st .
              shows_prec_list Zero_nat
                [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o', 'n',
                  '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']);
    (Fun f ss, Var _) ->
      Inl (shows_prec_list Zero_nat
             ['r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] st .
              shows_prec_list Zero_nat
                [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o', 'n',
                  '-', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']);
    (Fun f ss, Fun g ts) ->
      check (snd (list_ext (label_s_ns_impl cS cNS) (lterms af (Fun f ss))
                   (lterms af (Fun g ts))))
        (shows_prec_list Zero_nat
           ['c', 'a', 'n', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e', 'n', 't',
             ' ', 'p', 'a', 'i', 'r', ' '] .
          shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
            [' ', '-', '>', ' '] st .
            shows_prec_list Zero_nat [' ', 'w', 'e', 'a', 'k', 'l', 'y', ':'] .
              shows_nl .
                shows_prec_list Zero_nat (lterms af (Fun f ss)) .
                  shows_prec_list Zero_nat [' ', '>', '=', 'm', 'u', ' '] .
                    shows_prec_list Zero_nat (lterms af (Fun g ts)) .
                      shows_prec_list Zero_nat
                        [' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ',
                          'b', 'e', ' ', 'e', 'n', 's', 'u', 'r', 'e', 'd']);
  });

generate_scnp_rp ::
  forall a b c.
    (Eq a, Showa a,
      Showa b) => (((Term a b, Nat) -> (Term a b, Nat) -> (Bool, Bool)) ->
                    [(Term a b, Nat)] -> [(Term a b, Nat)] -> (Bool, Bool)) ->
                    [Char] ->
                      [((a, Nat), [(Nat, Nat)])] ->
                        (c -> Redtriple_ext a b ()) ->
                          c -> Root_redtriple_ext a b ();
generate_scnp_rp list_ext list_ext_name af rti x =
  let {
    rt = rti x;
    afaa = fun_ofa [] af;
    pia = afa rt;
    cS = s rt;
    cNS = ns rt;
  } in Root_redtriple_ext (valid rt) (s_label_mul_impl list_ext afaa cS cNS)
         (ns rt) (nST_label_mul_impl list_ext afaa cS cNS) pia
         (scnp_af_to_af afaa pia) (scnp_desc af list_ext_name . desca rt) ();

min_set_ext :: forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
min_set_ext s_ns =
  (\ asa bs ->
    (not (null bs) && all (\ a -> any (\ b -> fst (s_ns a b)) bs) asa,
      all (\ a -> any (\ b -> snd (s_ns a b)) bs) asa));

dms_convert ::
  forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> [[(Nat, (Bool, Bool))]];
dms_convert f asa bs =
  let {
    jbs = zip (upt Zero_nat (size_list bs)) bs;
  } in map (\ a -> map (\ (j, b) -> (j, f a b)) jbs) asa;

dms_simplify ::
  Bool -> [Nat] -> [[(Nat, (Bool, Bool))]] -> [[(Nat, (Bool, Bool))]];
dms_simplify stri is p = (if any (\ i -> null (nth p i)) is then [[]] else p);

dms_select :: Bool -> [[(Nat, (Bool, Bool))]] -> Nat;
dms_select stri p =
  snd (hd (sort_key fst
            (filter (\ (l, _) -> less_nat one_nat l)
              (zip (map size_list p) (upt Zero_nat (size_list p))))));

dms_decide_singletons :: Bool -> Nat -> [(Nat, (Bool, Bool))] -> Bool;
dms_decide_singletons stri n p =
  all_interval_nat
    (\ i ->
      let {
        (j, (s, ns)) = nth p i;
      } in s && not (membera (dropa (Suc i) p) (j, (False, True))) ||
             ns && not (membera (map fst (dropa (Suc i) p)) j))
    Zero_nat (size_list p) &&
    (if stri
      then not (all_interval_nat
                 (not . (\ j -> not (membera p (j, (False, True))))) Zero_nat n)
      else True);

dms_solve_or_select :: Bool -> Nat -> [[(Nat, (Bool, Bool))]] -> Sum Bool Nat;
dms_solve_or_select stri n p =
  (if all (\ jsns -> less_eq_nat (size_list jsns) one_nat) p
    then Inl (if membera p [] then False
               else dms_decide_singletons stri n (map hd p))
    else Inr (dms_select stri p));

dms_solve :: Bool -> Nat -> [[(Nat, (Bool, Bool))]] -> Bool;
dms_solve stri n p =
  (case dms_solve_or_select stri n p of {
    Inl res -> res;
    Inr k ->
      let {
        ksns = nth p k;
      } in dms_solve stri n
             (dms_simplify stri [k] (list_update p k [hd ksns])) ||
             dms_solve stri n
               (dms_simplify stri [k] (list_update p k (tl ksns)));
  });

dms_preprocess :: [[(Nat, (Bool, Bool))]] -> [[(Nat, (Bool, Bool))]];
dms_preprocess p = map (filter (\ (_, (a, b)) -> a || b)) p;

dms_bool_ex_idx_impl :: Bool -> Nat -> [[(Nat, (Bool, Bool))]] -> Bool;
dms_bool_ex_idx_impl stri n p =
  dms_solve stri n
    (dms_simplify stri (upt Zero_nat (size_list p)) (dms_preprocess p));

dms_bool_ex_idx :: Bool -> Nat -> [[(Nat, (Bool, Bool))]] -> Bool;
dms_bool_ex_idx = dms_bool_ex_idx_impl;

dms_order_ext ::
  forall a. Nat -> (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
dms_order_ext n f asa bs =
  let {
    p = dms_convert f asa bs;
    lts = size_list bs;
    len = less_eq_nat lts n || equal_nat (size_list asa) lts;
  } in (len && dms_bool_ex_idx True lts p, len && dms_bool_ex_idx False lts p);

or2 :: (Bool, Bool) -> (Bool, Bool) -> (Bool, Bool);
or2 (b1, b2) (b3, b4) = (b1 || b3, b2 || b4);

lifta :: (Bool, Bool) -> (Bool, Bool);
lifta (uu, b) = (b, b);

mul_ext_impl ::
  forall a. (Eq a) => (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
mul_ext_impl f [] [] = (False, True);
mul_ext_impl f [] (v : va) = (False, False);
mul_ext_impl f (v : va) [] = (True, True);
mul_ext_impl f (v : va) (y : ys) = ex_dominating f y (v : va) (v : va) ys;

ex_dominating ::
  forall a.
    (Eq a) => (a -> a -> (Bool, Bool)) ->
                a -> [a] -> [a] -> [a] -> (Bool, Bool);
ex_dominating f y [] uw ux = (False, False);
ex_dominating f y (b : bs) xs ys =
  (case f b y of {
    (True, ns) ->
      or2 (lifta
            (mul_ext_impl f (remove1 b xs)
              (filter (\ ya -> not (fst (f b ya))) ys)))
        (ex_dominating f y bs xs ys);
    (False, True) ->
      or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys);
    (False, False) -> ex_dominating f y bs xs ys;
  });

list_ext ::
  forall a.
    (Eq a) => Nat ->
                List_order_type ->
                  (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
list_ext uu MS_Ext = mul_ext_impl;
list_ext uv Max_Ext = set_ext;
list_ext uw Min_Ext = min_set_ext;
list_ext n Dms_Ext = dms_order_ext n;

list_ext_name :: List_order_type -> [Char];
list_ext_name MS_Ext = ['M', 'S'];
list_ext_name Dms_Ext = ['D', 'M', 'S'];
list_ext_name Min_Ext = ['M', 'I', 'N'];
list_ext_name Max_Ext = ['M', 'A', 'X'];

get_root_redtriple ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Root_redtriple_impl a -> Root_redtriple_ext a b ();
get_root_redtriple (Scnp typea af rp) =
  generate_scnp_rp (list_ext (scnp_arity af) typea) (list_ext_name typea) af
    get_redtriple rp;

data Scg a b = Null | Scg a a [(b, b)] [(b, b)];

combine :: forall a b. (Eq b) => Scg a b -> Scg a b -> Scg a b;
combine (Scg pa qa stra wka) (Scg p q str wk) =
  Scg pa qa (remdups (stra ++ str)) (remdups (wka ++ wk));
combine Null s = Null;
combine (Scg v va vb vc) Null = Null;

in_situ :: forall a b. (Eq b) => Scg a b -> Bool;
in_situ Null = True;
in_situ (Scg p q str wk) = any (\ (a, b) -> a == b) str;

comp :: forall a. (Eq a) => [(a, a)] -> [(a, a)] -> [(a, a)];
comp esa es =
  remdups
    (concatMap
      (\ (x, y) ->
        concatMap (\ (ya, z) -> (if y == ya then [(x, z)] else [])) es)
      esa);

subtract_list :: forall a. (Eq a) => [a] -> [a] -> [a];
subtract_list xs ys = filter (\ x -> not (membera ys x)) xs;

scg_comp ::
  forall a b. (Eq b) => (a -> a -> Bool) -> Scg a b -> Scg a b -> Scg a b;
scg_comp conn (Scg pa qa stra wka) (Scg p q str wk) =
  (if not (conn qa p) then Null
    else let {
           strs = remdups (comp stra str ++ comp stra wk ++ comp wka str);
           a = subtract_list (remdups (comp wka wk)) strs;
         } in Scg pa q strs a);
scg_comp conn Null g = Null;
scg_comp conn (Scg v va vb vc) Null = Null;

subsumes :: forall a b. (Eq a, Eq b) => Scg a b -> Scg a b -> Bool;
subsumes (Scg pa qa stra wka) (Scg p q str wk) =
  pa == p && qa == q && all (membera str) stra && all (membera (str ++ wk)) wka;
subsumes g Null = True;
subsumes Null (Scg v va vb vc) = False;

sagiv :: forall a b. (Eq a, Eq b) => (a -> a -> Bool) -> Scg a b -> Bool;
sagiv conn g =
  (if in_situ g then True
    else let {
           gg = scg_comp conn g g;
         } in (if subsumes gg g then False else sagiv conn (combine g gg)));

equal_scg :: forall a b. (Eq a, Eq b) => Scg a b -> Scg a b -> Bool;
equal_scg (Scg pp1 pp2 list1 list2) Null = False;
equal_scg Null (Scg pp1 pp2 list1 list2) = False;
equal_scg (Scg pp1a pp2a list1a list2a) (Scg pp1 pp2 list1 list2) =
  pp1a == pp1 && pp2a == pp2 && list1a == list1 && list2a == list2;
equal_scg Null Null = True;

generate_scgs ::
  forall a b.
    (Eq a, Eq b) => (a -> a -> Bool) -> [Scg a b] -> Scg a b -> [Scg a b];
generate_scgs conn base g =
  filter (\ ga -> not (equal_scg ga Null)) (map (scg_comp conn g) base);

check_SCT :: forall a b. (Eq a, Eq b) => (a -> a -> Bool) -> [Scg a b] -> Bool;
check_SCT conn gs =
  let {
    a = mk_rtrancl_list subsumes (generate_scgs conn gs) gs;
  } in all (sagiv conn) a;

get_arg :: forall a b. Term a b -> Nat -> Term a b;
get_arg t Zero_nat = t;
get_arg t (Suc n) = nth (args t) n;

check_sct_entry ::
  forall a b c d e f.
    (Showa a,
      Showa b) => ((a, Nat) -> Bool) ->
                    ((Term a b, Term a b) -> Sum c d) ->
                      ((Term a b, Term a b) -> Sum e f) ->
                        Term a b ->
                          Term a b ->
                            [(Nat, Nat)] ->
                              [(Nat, Nat)] -> Sum ([Char] -> [Char]) ();
check_sct_entry is_def sa nst s t stri nstri =
  catcha
    (bindb (check_no_var s)
      (\ _ ->
        bindb (check_no_var t)
          (\ _ ->
            bindb (check_no_defined_root is_def t)
              (\ _ ->
                let {
                  m = num_args t;
                  n = num_args s;
                } in bindb (catcha
                             (forallm
                               (\ i ->
                                 check (less_eq_nat i n)
                                   (shows_string
                                      ['l', 'e', 'f', 't', '-', 'i', 'n', 'd',
'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r', 'g', 'e'] .
                                     shows_prec_nat Zero_nat i . shows_nl))
                               (remdups (map fst (stri ++ nstri))))
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         bindb (catcha
                                 (forallm
                                   (\ j ->
                                     check (less_eq_nat j m)
                                       (shows_string
  ['r', 'i', 'g', 'h', 't', '-', 'i', 'n', 'd', 'e', 'x', ' ', 't', 'o', ' ',
    'l', 'a', 'r', 'g', 'e', ' ', 'o', 'r', ' ', 'a', 'r', 'g', 'u', 'm', 'e',
    'n', 't', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 'u', 's', 'a',
    'b', 'l', 'e', '-', 'r', 'u', 'l', 'e', 's', ' ', 'c', 'o', 'n', 'd', 'i',
    't', 'i', 'o', 'n'] .
 shows_prec_nat Zero_nat j))
                                   (remdups (map snd (stri ++ nstri))))
                                 (\ x -> Inl (snd x)))
                           (\ _ ->
                             let {
                               _ = args s;
                               _ = args t;
                             } in bindb (catcha
  (forallm
    (\ (i, j) ->
      check (isOK (sa (get_arg s i, get_arg t j)))
        (shows_string
           ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
             'e', 'd', 'g', 'e', ' '] .
          shows_prec_nat Zero_nat i .
            shows_string [' ', '-', 'S', '-', '>', ' '] .
              shows_prec_nat Zero_nat j))
    stri)
  (\ x -> Inl (snd x)))
                                    (\ _ ->
                                      catcha
(forallm
  (\ (i, j) ->
    check (isOK (nst (get_arg s i, get_arg t j)))
      (shows_string
         ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'e',
           'd', 'g', 'e', ' '] .
        shows_prec_nat Zero_nat i .
          shows_string [' ', '-', 'N', 'S', '-', '>', ' '] .
            shows_prec_nat Zero_nat j))
  nstri)
(\ x -> Inl (snd x)))))))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w', 'i', 't', 'h',
               ' ', 'D', 'P', ' '] .
            shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
              [' ', '-', '>', ' '] (s, t) .
              shows_nl . x));

sct_entry_to_sts ::
  forall a b c d. a -> Term b c -> [(d, Nat)] -> [(d, Nat)] -> [(a, Term b c)];
sct_entry_to_sts s t stri nstri =
  let {
    a = remdups (map snd (stri ++ nstri));
  } in map (\ j -> (s, get_arg t j)) a;

is_ur_closed_term_af_impl ::
  forall a b c.
    (Eq a,
      Eq c) => (Term a [Char] -> Bool) ->
                 (Term a b -> Term a (Sum () [Char])) ->
                   ((a, Nat) -> Nat -> Bool) ->
                     [(Term a [Char], c)] ->
                       [(Term a [Char], c)] ->
                         [Term a [Char]] -> Term a b -> Bool;
is_ur_closed_term_af_impl nfq e_cap pia r u s (Var x) = True;
is_ur_closed_term_af_impl nfq e_cap pia r u s (Fun f ts) =
  let {
    n = size_list ts;
    pi_f = pia (f, n);
  } in all (\ (i, t) ->
             (if pi_f i then is_ur_closed_term_af_impl nfq e_cap pia r u s t
               else True))
         (zip (upt Zero_nat n) ts) &&
         all (\ (l, ra) ->
               membera u (l, ra) ||
                 (case mgu_class (Fun f (map e_cap ts)) l of {
                   Nothing -> True;
                   Just mu ->
                     not (all (\ ua ->
                                nfq (subst_apply_term
                                      (map_vars (\ a -> 'y' : a) ua) mu))
                           (args l)) ||
                       not (all (\ ua -> nfq (subst_apply_term ua mu)) s);
                 }))
           r;

is_ur_closed_af_impl_dpp_mv ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> ((b, Nat) -> Nat -> Bool) ->
                           [(Term b [Char], Term b [Char])] ->
                             [Term b [Char]] -> Term b [Char] -> Bool;
is_ur_closed_af_impl_dpp_mv i d pia =
  let {
    ic = icap_impl_dpp i d;
    qnf = is_QNFc i d;
    r = rulesd i d;
    urc = (\ s -> is_ur_closed_term_af_impl qnf (ic s) pia r);
  } in (\ u s ->
         let {
           sa = map (map_vars (\ a -> 'x' : a)) s;
         } in (\ t -> urc sa u sa (map_vars (\ a -> 'x' : a) t)));

smart_usable_rules_checker_impl ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    a -> ((b, Nat) -> Nat -> Bool) ->
                           Maybe [(Term b [Char], Term b [Char])] ->
                             [(Term b [Char], Term b [Char])] ->
                               Sum ([Char] -> [Char])
                                 [(Term b [Char], Term b [Char])];
smart_usable_rules_checker_impl i d pia u_opt sts =
  let {
    nfs = nfsc i d;
    m = minimal i d;
    wwf = wwf_rulesa i d;
  } in (case u_opt of {
         Nothing -> Inr (rulesd i d);
         Just u ->
           (if nFQ_subset_NF_rulesb i d &&
                 (nfs ||
                   all (\ (s, t) ->
                         isOK (check_subseteq (vars_term_lista t)
                                (vars_term_lista s)))
                     sts) &&
                   (nfs || m || wwf)
             then let {
                    urc = is_ur_closed_af_impl_dpp_mv i d pia u;
                    check_urc =
                      (\ s t ->
                        check (urc s t)
                          (shows_prec_list Zero_nat ['t', 'e', 'r', 'm', ' '] .
                            shows_prec_term Zero_nat t .
                              shows_prec_list Zero_nat
                                [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c',
                                  'l', 'o', 's', 'e', 'd', ' ', 'u', 'n', 'd',
                                  'e', 'r', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                                  ' ', 'r', 'u', 'l', 'e', 's']));
                  } in bindb (catcha (forallm (\ (s, a) -> check_urc [s] a) sts)
                               (\ x -> Inl (snd x)))
                         (\ _ ->
                           bindb (catcha
                                   (forallm (\ (l, a) -> check_urc (args l) a)
                                     u)
                                   (\ x -> Inl (snd x)))
                             (\ _ -> Inr u))
             else bindb (check (m && (if nfs then wwf else True))
                          (shows_prec_list Zero_nat
                            ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y',
                              ' ', 'a', 'n', 'd', ' ', 'w', 'e', 'l', 'l', ' ',
                              'f', 'o', 'r', 'm', 'e', 'd', 'n', 'e', 's', 's',
                              ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                    (\ _ ->
                      bindb (catcha
                              (forallm
                                (\ (l, _) ->
                                  check (not (is_Var l))
                                    (shows_prec_list Zero_nat
                                      ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
's', ' ', 'a', 's', ' ', 'l', 'h', 's', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l',
'l', 'o', 'w', 'e', 'd']))
                                (rulesd i d))
                              (\ x -> Inl (snd x)))
                        (\ _ ->
                          let {
                            rm = rules_mapb i d;
                          } in bindb (check_ur_P_closed_rm_af rm u sts pia)
                                 (\ _ -> Inr u))));
       });

sct_ur_af_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    Redtriple_ext b [Char] () ->
                      [((Term b [Char], Term b [Char]),
                         ([(Nat, Nat)], [(Nat, Nat)]))] ->
                        Maybe [(Term b [Char], Term b [Char])] ->
                          a -> Sum ([Char] -> [Char]) ();
sct_ur_af_proc i rp gs u_opt dpp =
  catcha
    (bindb (valid rp)
      (\ _ ->
        let {
          is_def = (\ fn -> not (null (rules_mapb i dpp fn)));
          pia = afa rp;
          sa = s rp;
          nsa = ns rp;
          nsta = nst rp;
        } in bindb (catcha (forallm (\ (l, _) -> check_no_var l) (rulesd i dpp))
                     (\ x -> Inl (snd x)))
               (\ _ ->
                 bindb (catcha
                         (forallm
                           (\ (a, b) ->
                             let {
                               (saa, t) = a;
                             } in (\ (aa, ba) ->
                                    check_sct_entry is_def sa nsta saa t aa ba)
                               b)
                           gs)
                         (\ x -> Inl (snd x)))
                   (\ _ ->
                     let {
                       sts = concatMap
                               (\ (a, b) ->
                                 let {
                                   (sb, t) = a;
                                 } in (\ (aa, ba) ->
sct_entry_to_sts sb t aa ba)
                                   b)
                               gs;
                     } in bindb (smart_usable_rules_checker_impl i dpp pia u_opt
                                  sts)
                            (\ u ->
                              bindb (catcha
                                      (catcha (forallm nsa u)
(\ x -> Inl (snd x)))
                                      (\ x ->
Inl (shows_string
       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
         'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's', 'a', 'b', 'l',
         'e', ' ', 'r', 'u', 'l', 'e', 's'] .
      shows_nl . x)))
                                (\ _ ->
                                  let {
                                    p = pairsa i dpp;
                                    eidg = is_iedg_edge_dpp i dpp;
                                  } in bindb
 (catcha (check_subseteq p (map fst gs))
   (\ x ->
     Inl (shows_string
            ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o', ' ', 's',
              'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'g', 'r',
              'a', 'p', 'h', ' ', 'f', 'o', 'r', ' ', 'D', 'P', ' '] .
           shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
             [' ', '-', '>', ' '] x)))
 (\ _ ->
   check (check_SCT (\ (_, (_, succs)) (ua, (v, _)) -> membera succs (ua, v))
           (map (\ (a, b) ->
                  let {
                    (sb, t) = a;
                  } in (\ (stri, nstri) ->
                         let {
                           eidg_st = eidg (sb, t);
                           e = (sb, (t, filter (eidg_st . fst) p));
                         } in Scg e e stri nstri)
                    b)
             gs))
     (shows_string
        ['s', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'a', 'n',
          'a', 'l', 'y', 's', 'i', 's', ' ', 'f', 'a', 'i', 'l', 'e', 'd'] .
       shows_nl))))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e', '-', 'c',
               'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
               'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f',
               'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' '] .
            shows_nl .
              desca rp .
                shows_nl .
                  shows_string
                    ['f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                      'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a', 's', 'o',
                      'n'] .
                    shows_nl . x));

sct_subterm_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    [((Term b [Char], Term b [Char]),
                       ([(Nat, Nat)], [(Nat, Nat)]))] ->
                      a -> Sum ([Char] -> [Char]) ();
sct_subterm_proc i gs dpp =
  catcha
    (let {
       p = pairsa i dpp;
       is_def = (\ fn -> not (null (rules_mapb i dpp fn)));
       eidg = is_iedg_edge_dpp i dpp;
     } in bindb (catcha (check_subseteq p (map fst gs))
                  (\ x ->
                    Inl (shows_string
                           ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n',
                             'o', ' ', 's', 'i', 'z', 'e', '-', 'c', 'h', 'a',
                             'n', 'g', 'e', ' ', 'g', 'r', 'a', 'p', 'h', ' ',
                             'f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                             'i', 'r', ' '] .
                          shows_rule (shows_prec Zero_nat)
                            (shows_prec_list Zero_nat) [' ', '-', '>', ' '] x)))
            (\ _ ->
              bindb (check (minimal i dpp)
                      (shows_prec_list Zero_nat
                        ['m', 'i', 'n', 'i', 'm', 'a', 'l', 'i', 't', 'y', ' ',
                          'r', 'e', 'q', 'u', 'i', 'r', 'e', 'd']))
                (\ _ ->
                  bindb (catcha
                          (forallm (\ (l, _) -> check_no_var l) (rulesd i dpp))
                          (\ x -> Inl (snd x)))
                    (\ _ ->
                      bindb (catcha
                              (forallm
                                (\ (a, b) ->
                                  let {
                                    (s, t) = a;
                                  } in (\ (stri, nstri) ->
 catcha
   (bindb (check_no_var s)
     (\ _ ->
       bindb (check_no_var t)
         (\ _ ->
           bindb (check_no_defined_root is_def t)
             (\ _ ->
               let {
                 m = num_args t;
                 n = num_args s;
               } in bindb (catcha
                            (forallm
                              (\ (ia, j) ->
                                check (less_eq_nat ia n &&
less_eq_nat j m && isOK (check_supt (get_arg s ia) (get_arg t j)))
                                  (shows_string
                                     ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                       'w', 'i', 't', 'h', ' ', 'e', 'd', 'g',
                                       'e', ' '] .
                                    shows_prec_nat Zero_nat ia .
                                      shows_string [' ', '|', '>', ' '] .
shows_prec_nat Zero_nat j . shows_nl))
                              stri)
                            (\ x -> Inl (snd x)))
                      (\ _ ->
                        catcha
                          (forallm
                            (\ (ia, j) ->
                              check (less_eq_nat ia n &&
                                      less_eq_nat j m &&
isOK (check_supteq (get_arg s ia) (get_arg t j)))
                                (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'i', 't', 'h', ' ', 'e', 'd', 'g', 'e',
                                     ' '] .
                                  shows_prec_nat Zero_nat ia .
                                    shows_string [' ', '|', '>', '=', ' '] .
                                      shows_prec_nat Zero_nat j . shows_nl))
                            nstri)
                          (\ x -> Inl (snd x)))))))
   (\ x ->
     Inl (shows_string
            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
              'p', 'a', 'i', 'r', ' '] .
           shows_rule (shows_prec Zero_nat) (shows_prec_list Zero_nat)
             [' ', '-', '>', ' '] (s, t) .
             shows_nl . x)))
                                    b)
                                gs)
                              (\ x -> Inl (snd x)))
                        (\ _ ->
                          check (check_SCT
                                  (\ (_, (_, succs)) (u, (v, _)) ->
                                    membera succs (u, v))
                                  (map (\ (a, b) ->
 let {
   (s, t) = a;
 } in (\ (stri, nstri) ->
        let {
          eidg_st = eidg (s, t);
          e = (s, (t, filter (eidg_st . fst) p));
        } in Scg e e stri nstri)
   b)
                                    gs))
                            (shows_string
                               ['s', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n',
                                 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's',
                                 'i', 's', ' ', 'f', 'a', 'i', 'l', 'e', 'd'] .
                              shows_nl))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e', '-', 'c',
               'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
               'o', 'r', ' ', 'b', 'a', 's', 'e', 'd', ' ', 'o', 'n', ' ', 't',
               'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', '-', 'r', 'e',
               'l', 'a', 't', 'i', 'o', 'n'] .
            shows_nl . x));

generic_ur_af_redtriple_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    Redtriple_ext b [Char] () ->
                      Maybe [(Term b [Char], Term b [Char])] ->
                        [(Term b [Char], Term b [Char])] ->
                          a -> Sum ([Char] -> [Char]) a;
generic_ur_af_redtriple_proc i rp u_opt premove dpp =
  (case catcha
          (bindb (valid rp)
            (\ _ ->
              let {
                (ps, pns) = split_pairsa i dpp premove;
                p = pairsa i dpp;
              } in bindb (smart_usable_rules_checker_impl i dpp (afa rp) u_opt
                           p)
                     (\ u ->
                       bindb (catcha
                               (catcha (forallm (ns rp) u) (\ x -> Inl (snd x)))
                               (\ x ->
                                 Inl (shows_string
['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i',
  'e', 'n', 't', 'i', 'n', 'g', ' ', '(', 'u', 's', 'a', 'b', 'l', 'e', ')',
  ' ', 'r', 'u', 'l', 'e', 's'] .
                                       shows_nl . x)))
                         (\ _ ->
                           bindb (catcha
                                   (catcha (forallm (nst rp) pns)
                                     (\ x -> Inl (snd x)))
                                   (\ x ->
                                     Inl (shows_string
    ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
   shows_nl . x)))
                             (\ _ ->
                               catcha
                                 (catcha (forallm (s rp) ps)
                                   (\ x -> Inl (snd x)))
                                 (\ x ->
                                   Inl (shows_string
  ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
    'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
 shows_nl . x)))))))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                     'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'g', 'e', 'n', 'e',
                     'r', 'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                     'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
                     's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h',
                     'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
                  shows_nl . desca rp . shows_nl . x))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (delete_P_Pwa i dpp premove premove);
  });

root_aft_to_entry ::
  forall a b c. a -> Term b c -> ((b, Nat) -> Nat -> Bool) -> [(a, Term b c)];
root_aft_to_entry s t pia =
  let {
    rt = the (root t);
    pi_t = pia rt;
    ts = args t;
  } in map_filter (\ x -> (if pi_t x then Just (s, nth ts x) else Nothing))
         (upt Zero_nat (snd rt));

generic_ur_af_root_redtriple_proc ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    Root_redtriple_ext b [Char] () ->
                      Maybe [(Term b [Char], Term b [Char])] ->
                        [(Term b [Char], Term b [Char])] ->
                          a -> Sum ([Char] -> [Char]) a;
generic_ur_af_root_redtriple_proc i rp u_opt premove dpp =
  (case catcha
          (bindb (valida rp)
            (\ _ ->
              let {
                (ps, pns) = split_pairsa i dpp premove;
                p = pairsa i dpp;
                pia = af rp;
                piaa = aft rp;
                is_def = (\ fn -> not (null (rules_mapb i dpp fn)));
              } in bindb (catcha
                           (forallm
                             (\ (l, r) ->
                               bindb (check_no_var l)
                                 (\ _ ->
                                   bindb (check_no_var r)
                                     (\ _ -> check_no_defined_root is_def r)))
                             p)
                           (\ x -> Inl (snd x)))
                     (\ _ ->
                       bindb (catcha
                               (forallm (\ (l, _) -> check_no_var l)
                                 (rulesd i dpp))
                               (\ x -> Inl (snd x)))
                         (\ _ ->
                           bindb (smart_usable_rules_checker_impl i dpp pia
                                   u_opt
                                   (concatMap
                                     (\ (s, t) -> root_aft_to_entry s t piaa)
                                     p))
                             (\ u ->
                               bindb (catcha
                                       (catcha (forallm (nsa rp) u)
 (\ x -> Inl (snd x)))
                                       (\ x ->
 Inl (shows_string
        ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
          'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', '(', 'u', 's', 'a', 'b',
          'l', 'e', ')', ' ', 'r', 'u', 'l', 'e', 's'] .
       shows_nl . x)))
                                 (\ _ ->
                                   bindb (catcha
   (catcha (forallm (nsta rp) pns) (\ x -> Inl (snd x)))
   (\ x ->
     Inl (shows_string
            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
              'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
           shows_nl . x)))
                                     (\ _ ->
                                       catcha
 (catcha (forallm (sa rp) ps) (\ x -> Inl (snd x)))
 (\ x ->
   Inl (shows_string
          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
            'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
         shows_nl . x)))))))))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                     'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'g', 'e', 'n', 'e',
                     'r', 'i', 'c', ' ', 'r', 'o', 'o', 't', ' ', 'r', 'e', 'd',
                     'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ',
                     'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i',
                     't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o',
                     'w', 'i', 'n', 'g'] .
                  shows_nl . descaa rp . shows_nl . x))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (delete_P_Pwa i dpp premove premove);
  });

check_dp_proof ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Dpp_ops_ext a (Lab b [Nat]) [Char] () ->
                        a -> Dp_proof b [Nat] [Char] ->
                               Sum ([Char] -> [Char]) ();
check_dp_proof assms ia i dpp PisEmpty =
  debug (ia []) ['P', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (if null (pa i dpp) && (null (pwa i dpp) || null (rc i dpp)) then Inr ()
      else Inl (ia . shows_string
                       [':', ' ', 'P', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                         'e', 'm', 'p', 't', 'y', ' ', 'i', 'n', ' ', 't', 'h',
                         'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                         ' ', 'D', 'P', '-', 'p', 'r', 'o', 'b', 'l', 'e', 'm',
                         ' '] .
                       shows_nl .
                         shows_dpp (shows_prec_lab Zero_nat)
                           (shows_prec_list Zero_nat) i dpp));
check_dp_proof assms ia i dpp (SubtermCriterionProc p rseq rP prf) =
  debug (ia [])
    ['S', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i', 't', 'e', 'r', 'i', 'o',
      'n', 'P', 'r', 'o', 'c']
    (let {
       pa = pairsa i dpp;
       pr = list_diff pa rP;
     } in bindb (catcha
                  (subterm_criterion_proc i (create_proj p)
                    (create_rseq_map rseq) pr dpp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 's',
                                  'u', 'b', 't', 'e', 'r', 'm', ' ', 'c', 'r',
                                  'i', 't', 'e', 'r', 'i', 'o', 'n', ' ', 't',
                                  'o', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ',
                                  'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                                shows_nl .
                                  shows_dpp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl .
                                      shows_string
['a', 'n', 'd', ' ', 't', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'o', ' ', 'r', 'e',
  'm', 'o', 'v', 'e', ' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] pr .
    shows_nl . x)))
            (\ dppa ->
              catcha
                (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 's',
                                'u', 'b', 't', 'e', 'r', 'm', ' ', 'c', 'r',
                                'i', 't', 'e', 'r', 'i', 'o', 'n'] .
                              shows_nl . indent x))));
check_dp_proof assms ia i dpp (RedPairProc redp rP prf) =
  debug (ia []) ['R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
    (let {
       p = pairsa i dpp;
       pr = list_diff p rP;
       proc =
         (case redp of {
           Inl rrp ->
             generic_ur_af_root_redtriple_proc i (get_root_redtriple rrp)
               Nothing;
           Inr rp -> generic_ur_af_redtriple_proc i (get_redtriple rp) Nothing;
         });
     } in bindb (catcha (proc pr dpp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r',
                                  'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                                  'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                                  'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ',
                                  'r', 'e', 'm', 'o', 'v', 'e', ' ', 'f', 'r',
                                  'o', 'm', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                  ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                                shows_nl .
                                  shows_dpp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl .
                                      shows_string
[' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] pr .
    shows_nl . x)))
            (\ dppa ->
              catcha
                (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r',
                                'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                                'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                                'e', 's', 's', 'o', 'r'] .
                              shows_nl . indent x))));
check_dp_proof assms ia i dpp (UsableRulesProc u prf) =
  debug (ia [])
    ['U', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (usable_rules_proc i u dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                        'r', 'u', 'l', 'e', 's', ' ', 'p', 'r', 'o', 'c', 'e',
                        's', 's', 'o', 'r', ' ', 't', 'o', ' ', 'r', 'e', 's',
                        't', 'r', 'i', 'c', 't', ' ', 't', 'h', 'e', ' ', 'D',
                        'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl .
                            shows_string
                              [' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'u', 's',
                                'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                                's', ' '] .
                              shows_nl .
                                shows_rules (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  [' ', '-', '>', ' '] u .
                                  shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 's', 'a', 'b',
                          'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'p', 'r',
                          'o', 'c', 'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (QReductionProc q prf) =
  debug (ia [])
    ['Q', 'R', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (q_reduction_proc i q dpp)
        (\ x ->
          Inl (ia . shows_prec_list Zero_nat
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'Q', '-', 'r', 'e', 'd', 'u', 'c',
                        't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c', 'e', 's',
                        's', 'o', 'r'] .
                      shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'Q', '-', 'r', 'e',
                          'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o',
                          'c', 'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (MonoRedPairProc redp rP rR prf) =
  debug (ia [])
    ['M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
    (let {
       p = pairsa i dpp;
       pr = list_diff p rP;
       r = rulesd i dpp;
       rr = list_diff r rR;
     } in bindb (catcha (mono_redpair_proc i (get_redtriple redp) pr rr dpp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm',
                                  'o', 'n', 'o', ' ', 'r', 'e', 'd', 'u', 'c',
                                  't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                                  ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                                  'r', ' ', 't', 'o', ' ', 'r', 'e', 'm', 'o',
                                  'v', 'e', ' ', 'f', 'r', 'o', 'm', ' ', 't',
                                  'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o',
                                  'b', 'l', 'e', 'm'] .
                                shows_nl .
                                  shows_dpp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl .
                                      shows_string
[' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] pr .
    shows_nl .
      shows_string
        [' ', 'a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', 's',
          ' '] .
        shows_nl .
          shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] rr .
            shows_nl . x)))
            (\ dppa ->
              catcha
                (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'm',
                                'o', 'n', 'o', ' ', 'r', 'e', 'd', 'u', 'c',
                                't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                                ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                                'r'] .
                              shows_nl . indent x))));
check_dp_proof assms ia i dpp (DepGraphProc edpts) =
  debug (ia []) ['D', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (dep_graph_proc i dpp edpts)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i',
                        'l', 'e', ' ', 't', 'r', 'y', 'i', 'n', 'g', ' ', 't',
                        'o', ' ', 'p', 'e', 'r', 'f', 'o', 'r', 'm', ' ', 'S',
                        'C', 'C', '-', 'd', 'e', 'c', 'o', 'm', 'p', 'o', 's',
                        'i', 't', 'i', 'o', 'n', ' ', ' ', 'o', 'n', ' '] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . x)))
      (\ pdpps ->
        catcha
          (catcha
            (forallm_index
              (\ (prof, dppa) j ->
                check_dp_proof assms
                  (ia . shows_string ['.'] . shows_prec_nat Zero_nat (Suc j)) i
                  dppa prof)
              pdpps)
            (\ x -> Inl (snd x)))
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'd', 'e', 'p', 'e',
                          'n', 'd', 'e', 'n', 'c', 'y', ' ', 'g', 'r', 'a', 'p',
                          'h', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                          'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (RedPairUrProc redp rP ur prf) =
  debug (ia [])
    ['R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c']
    (let {
       p = pairsa i dpp;
       pr = list_diff p rP;
       proc =
         (case redp of {
           Inl rrp ->
             generic_ur_af_root_redtriple_proc i (get_root_redtriple rrp)
               (Just ur);
           Inr rp ->
             generic_ur_af_redtriple_proc i (get_redtriple rp) (Just ur);
         });
     } in bindb (catcha (proc pr dpp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r',
                                  'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                                  'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                                  'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't',
                                  'h', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                                  'r', 'u', 'l', 'e', 's', ' ', 't', 'o', ' ',
                                  'r', 'e', 'm', 'o', 'v', 'e', ' ', 'f', 'r',
                                  'o', 'm', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                  ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                                shows_nl .
                                  shows_dpp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl .
                                      shows_string
[' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] pr .
    shows_nl . x)))
            (\ dppa ->
              catcha
                (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r',
                                'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ',
                                'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                                'e', 's', 's', 'o', 'r'] .
                              shows_nl . indent x))));
check_dp_proof assms ia i dpp (MonoRedPairUrProc redp rP rR ur prf) =
  debug (ia [])
    ['M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r',
      'o', 'c']
    (let {
       p = pairsa i dpp;
       pr = list_diff p rP;
       r = rulesd i dpp;
       rr = list_diff r rR;
     } in bindb (catcha
                  (mono_ur_redpair_proc i (get_redtriple redp) pr rr ur dpp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'm',
                                  'o', 'n', 'o', ' ', 'r', 'e', 'd', 'u', 'c',
                                  't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                                  ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                                  'r', ' ', 'w', 'i', 't', 'h', ' ', 'u', 's',
                                  'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                                  's', ' ', 't', 'o', ' ', 'r', 'e', 'm', 'o',
                                  'v', 'e', ' ', 'f', 'r', 'o', 'm', ' ', 't',
                                  'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r', 'o',
                                  'b', 'l', 'e', 'm'] .
                                shows_nl .
                                  shows_dpp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl .
                                      shows_string
[' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] pr .
    shows_nl .
      shows_string
        [' ', 'a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', 's',
          ' '] .
        shows_nl .
          shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
            [' ', '-', '>', ' '] rr .
            shows_nl . x)))
            (\ dppa ->
              catcha
                (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'm',
                                'o', 'n', 'o', ' ', 'r', 'e', 'd', 'u', 'c',
                                't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r',
                                ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                                'r', ' ', 'w', 'i', 't', 'h', ' ', 'u', 's',
                                'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                                's'] .
                              shows_nl . indent x))));
check_dp_proof assms ia i dpp (UncurryProc mode u_info p r prf) =
  debug (ia []) ['U', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (uncurry_proc_both i mode u_info p r dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y',
                        'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                        'o', 'r', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ', 'D',
                        'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u',
                          'r', 'r', 'y', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
                          'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (SizeChangeSubterm graphs) =
  debug (ia [])
    ['S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'S', 'u', 'b', 't', 'e',
      'r', 'm']
    (sct_subterm_proc i graphs dpp);
check_dp_proof assms ia i dpp (SizeChangeRedPair redp u_opt graphs) =
  debug (ia [])
    ['S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'R', 'e', 'd', 'P', 'a',
      'i', 'r']
    (sct_ur_af_proc i (get_redtriple redp) graphs u_opt dpp);
check_dp_proof assms ia i dpp (FccProc f fcs pb r prf) =
  debug (ia []) ['F', 'c', 'c', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (fcc_proc i f fcs pb r dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o',
                        'n', 't', 'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u',
                        'r', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                        'r', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                        ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'f', 'l', 'a', 't',
                          ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', ' ', 'c', 'l',
                          'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o', 'c', 'e',
                          's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (SplitProc prem rrem prf1 prf2) =
  debug (ia []) ['S', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c']
    (case get_fcc_option prf1 of {
      Nothing ->
        let {
          (dpp1, dpp2) = split_proc i dpp prem rrem;
        } in bindb (catcha
                     (check_dp_proof assms (ia . shows_string ['.', '1']) i dpp1
                       prf1)
                     (\ x ->
                       Inl (ia . shows_string
                                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                     'e', 'l', 'o', 'w', ' ', 't', 'h', 'e',
                                     ' ', 's', 'p', 'l', 'i', 't', ' ', 'p',
                                     'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                                   shows_nl . indent x)))
               (\ _ ->
                 catcha
                   (check_dp_proof assms (ia . shows_string ['.', '2']) i dpp2
                     prf2)
                   (\ x ->
                     Inl (ia . shows_string
                                 [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                   'e', 'l', 'o', 'w', ' ', 't', 'h', 'e', ' ',
                                   's', 'p', 'l', 'i', 't', ' ', 'p', 'r', 'o',
                                   'c', 'e', 's', 's', 'o', 'r'] .
                                 shows_nl . indent x)));
      Just (f, (fcs, (pb, (rb, prf1a)))) ->
        debug (ia [])
          ['S', 'p', 'l', 'i', 't', 'P', 'r', 'o', 'c', 'F', 'c', 'c']
          (bindb (fcc_split_proc i f fcs pb rb prem rrem dpp)
            (\ (dpp1, dpp2) ->
              bindb (catcha
                      (check_dp_proof assms
                        (ia . shows_string ['.', '1', '.', '1']) i dpp1 prf1a)
                      (\ x ->
                        Inl (ia . shows_string
                                    [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
                                      'b', 'e', 'l', 'o', 'w', ' ', 't', 'h',
                                      'e', ' ', 's', 'p', 'l', 'i', 't', ' ',
                                      'a', 'n', 'd', ' ', 'f', 'c', 'c', ' ',
                                      'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                                      'r'] .
                                    shows_nl . indent x)))
                (\ _ ->
                  catcha
                    (check_dp_proof assms (ia . shows_string ['.', '2']) i dpp2
                      prf2)
                    (\ x ->
                      Inl (ia . shows_string
                                  [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                    'e', 'l', 'o', 'w', ' ', 't', 'h', 'e', ' ',
                                    's', 'p', 'l', 'i', 't', ' ', 'p', 'r', 'o',
                                    'c', 'e', 's', 's', 'o', 'r'] .
                                  shows_nl . indent x)))));
    });
check_dp_proof assms ia i dpp (SemlabProc sli lP lQ lR prf) =
  debug (ia []) ['S', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (semlab_fin_proc i sli lP lQ lR dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 's', 'e', 'm', 'l', 'a', 'b', ' ',
                        'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'o',
                        'n', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ', 'p', 'r',
                        'o', 'b', 'l', 'e', 'm'] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'e', 'm', 'l',
                          'a', 'b', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                          'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (SwitchInnermostProc joins prf) =
  debug (ia [])
    ['S', 'w', 'i', 't', 'c', 'h', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't',
      'P', 'r', 'o', 'c']
    (bindb
      (catcha (switch_innermost_proc i joins dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                        'o', 'r', ' ', 't', 'o', ' ', 's', 'w', 'i', 't', 'c',
                        'h', ' ', 't', 'o', ' ', 'i', 'n', 'n', 'e', 'r', 'm',
                        'o', 's', 't', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ',
                        'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                      shows_nl .
                        shows_dpp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'w', 'i', 't',
                          'c', 'h', ' ', 't', 'o', ' ', 'i', 'n', 'n', 'e', 'r',
                          'm', 'o', 's', 't', ' ', 'p', 'r', 'o', 'c', 'e', 's',
                          's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dppa (AssumeFinite dpp) =
  debug (ia [])
    ['F', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', ' ', 'A', 's', 's', 'u',
      'm', 'p', 't', 'i', 'o', 'n']
    (if assms
      then catcha (check_dpp_subsumes i dpp dppa)
             (\ x ->
               Inl (ia . shows_string
                           [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n',
                             ' ', 'f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's',
                             's', ' ', 'a', 's', 's', 'u', 'm', 'p', 't', 'i',
                             'o', 'n'] .
                           shows_nl . x . shows_nl))
      else Inl (ia . shows_string
                       [':', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'o', 'f',
                         ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 's', ' ', 'a',
                         ' ', 'l', 'o', 'c', 'a', 'l', ' ', 'f', 'i', 'n', 'i',
                         't', 'e', 'n', 'e', 's', 's', ' ', 'a', 's', 's', 'u',
                         'm', 'p', 't', 'i', 'o', 'n'] .
                       shows_nl));
check_dp_proof assms ia i dpp (RewritingProc u stb sta st lr p prf) =
  debug (ia [])
    ['R', 'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (rewriting_proc i u stb sta st lr p dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'i',
                        'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                        'r', ' ', 't', 'o', ' ', 'r', 'e', 'w', 'r', 'i', 't',
                        'e', ' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', ' '] .
                      shows_nl .
                        shows_rule (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) [' ', '-', '>', ' '] stb .
                          shows_nl .
                            shows_string
                              [' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                                'i', 'r', ' '] .
                              shows_nl .
                                shows_rule (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  [' ', '-', '>', ' '] st .
                                  shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r',
                          'i', 't', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e',
                          's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (NarrowingProc st p sts prf) =
  debug (ia [])
    ['N', 'a', 'r', 'r', 'o', 'w', 'i', 'n', 'g', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (narrowing_proc i st p sts dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'n', 'a', 'r', 'r', 'o', 'w', 'i',
                        'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                        'r', ' ', 't', 'o', ' ', 'n', 'a', 'r', 'r', 'o', 'w',
                        ' ', 't', 'h', 'e', ' ', 'p', 'a', 'i', 'r', ' '] .
                      shows_nl .
                        shows_rule (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) [' ', '-', '>', ' '] st .
                          shows_nl .
                            shows_string
                              [' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                                'i', 'r', 's', ' '] .
                              shows_nl .
                                shows_trs (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
                                    'y', 's', 't', 'e', 'm', ':']
                                  [' ', '-', '>', ' '] sts .
                                  shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'n', 'a', 'r', 'r',
                          'o', 'w', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e',
                          's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (InstantiationProc st sts prf) =
  debug (ia [])
    ['I', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 'P', 'r',
      'o', 'c']
    (bindb
      (catcha (instantiation_proc i st sts dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'i', 'n', 's', 't', 'a', 'n', 't',
                        'i', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c',
                        'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ', 'i', 'n',
                        's', 't', 'a', 'n', 't', 'i', 'a', 't', 'e', ' ', 't',
                        'h', 'e', ' ', 'p', 'a', 'i', 'r', ' '] .
                      shows_nl .
                        shows_rule (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) [' ', '-', '>', ' '] st .
                          shows_nl .
                            shows_string
                              [' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                                'i', 'r', 's', ' '] .
                              shows_nl .
                                shows_trs (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
                                    'y', 's', 't', 'e', 'm', ':']
                                  [' ', '-', '>', ' '] sts .
                                  shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'i', 'n', 's', 't',
                          'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', ' ', 'p',
                          'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (ForwardInstantiationProc st sts u_opt prf) =
  debug (ia [])
    ['F', 'o', 'r', 'w', 'a', 'r', 'd', 'I', 'n', 's', 't', 'a', 'n', 't', 'i',
      'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'c']
    (bindb
      (catcha (forward_instantiation_proc i st sts u_opt dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'f', 'o', 'r', 'w', 'a', 'r', 'd',
                        '_', 'i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                        'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                        'o', 'r', ' ', 't', 'o', ' ', 'i', 'n', 's', 't', 'a',
                        'n', 't', 'i', 'a', 't', 'e', ' ', 't', 'h', 'e', ' ',
                        'p', 'a', 'i', 'r', ' '] .
                      shows_nl .
                        shows_rule (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) [' ', '-', '>', ' '] st .
                          shows_nl .
                            shows_string
                              [' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                                'i', 'r', 's', ' '] .
                              shows_nl .
                                shows_trs (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's',
                                    'y', 's', 't', 'e', 'm', ':']
                                  [' ', '-', '>', ' '] sts .
                                  shows_nl . x)))
      (\ dppa ->
        catcha (check_dp_proof assms (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'r', 'w',
                          'a', 'r', 'd', ' ', 'i', 'n', 's', 't', 'a', 'n', 't',
                          'i', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o', 'c',
                          'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_dp_proof assms ia i dpp (UnlabProc p r prf) =
  debug (ia []) ['U', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
    (Inl (shows_prec_list Zero_nat
           ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r',
             'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'n', 'o', 't', ' ', 's',
             'u', 'p', 'p', 'o', 'r', 't', 'e', 'd']));

unlab_to_split_proc ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => Dp_proof a b c ->
                 (Dp_proof a b c,
                   [([(Term (Lab a b) c, Term (Lab a b) c)],
                      ([(Term (Lab a b) c, Term (Lab a b) c)],
                        Dp_proof a b c))]);
unlab_to_split_proc PisEmpty = (PisEmpty, []);
unlab_to_split_proc (DepGraphProc ps) =
  (DepGraphProc
     (map (\ (po, a) -> (mapa (fst . unlab_to_split_proc) po, a)) ps),
    []);
unlab_to_split_proc (SubtermCriterionProc a1 a2 a3 p) =
  update (SubtermCriterionProc a1 a2 a3) (unlab_to_split_proc p);
unlab_to_split_proc (RedPairProc a1 a2 p) =
  update (RedPairProc a1 a2) (unlab_to_split_proc p);
unlab_to_split_proc (RedPairUrProc a1 a2 a3 p) =
  update (RedPairUrProc a1 a2 a3) (unlab_to_split_proc p);
unlab_to_split_proc (UsableRulesProc a1 p) =
  update (UsableRulesProc a1) (unlab_to_split_proc p);
unlab_to_split_proc (QReductionProc a1 p) =
  update (QReductionProc a1) (unlab_to_split_proc p);
unlab_to_split_proc (MonoRedPairProc a1 a2 a3 p) =
  update (MonoRedPairProc a1 a2 a3) (unlab_to_split_proc p);
unlab_to_split_proc (MonoRedPairUrProc a1 a2 a3 a4 p) =
  update (MonoRedPairUrProc a1 a2 a3 a4) (unlab_to_split_proc p);
unlab_to_split_proc (SizeChangeSubterm a1) = (SizeChangeSubterm a1, []);
unlab_to_split_proc (SizeChangeRedPair a1 a2 a3) =
  (SizeChangeRedPair a1 a2 a3, []);
unlab_to_split_proc (UncurryProc a1 a2 a3 a4 p) =
  update (UncurryProc a1 a2 a3 a4) (unlab_to_split_proc p);
unlab_to_split_proc (FccProc a1 a2 a3 a4 p) =
  update (FccProc a1 a2 a3 a4) (unlab_to_split_proc p);
unlab_to_split_proc (SwitchInnermostProc a1 p) =
  update (SwitchInnermostProc a1) (unlab_to_split_proc p);
unlab_to_split_proc (RewritingProc a1 a2 a3 a4 a5 a6 p) =
  update (RewritingProc a1 a2 a3 a4 a5 a6) (unlab_to_split_proc p);
unlab_to_split_proc (NarrowingProc a1 a2 a3 p) =
  update (NarrowingProc a1 a2 a3) (unlab_to_split_proc p);
unlab_to_split_proc (InstantiationProc a1 a2 p) =
  update (InstantiationProc a1 a2) (unlab_to_split_proc p);
unlab_to_split_proc (ForwardInstantiationProc a1 a2 a3 p) =
  update (ForwardInstantiationProc a1 a2 a3) (unlab_to_split_proc p);
unlab_to_split_proc (AssumeFinite d) = (AssumeFinite d, []);
unlab_to_split_proc (UnlabProc pa r p) = (PisEmpty, [(pa, (r, p))]);
unlab_to_split_proc (SplitProc a1 a2 p q) =
  (SplitProc a1 a2 (fst (unlab_to_split_proc p)) (fst (unlab_to_split_proc q)),
    []);
unlab_to_split_proc (SemlabProc a1 lP a2 lR p) =
  let {
    (pa, list) = unlab_to_split_proc p;
    sl = SemlabProc a1 lP a2 lR pa;
  } in (case list of {
         [] -> (sl, list);
         (pb, (r, prof)) : lista ->
           let {
             ulP = unlab_of_trs lP;
             ulR = unlab_of_trs lR;
             pr = list_diff ulP pb;
             rr = list_diff ulR r;
           } in (SplitProc pr rr sl prof, lista);
       });

check_dp_proof_with_unlab ::
  forall a b.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Dpp_ops_ext a (Lab b [Nat]) [Char] () ->
                        a -> Dp_proof b [Nat] [Char] ->
                               Sum ([Char] -> [Char]) ();
check_dp_proof_with_unlab a ia i dpp prf =
  check_dp_proof a ia i dpp (fst (unlab_to_split_proc prf));

rule_removal_tt ::
  forall a b c.
    (Showa b,
      Showa c) => Tp_ops_ext a b c () ->
                    Redtriple_ext b c () ->
                      [(Term b c, Term b c)] -> a -> Sum ([Char] -> [Char]) a;
rule_removal_tt i rp rremove trs =
  (case catcha
          (let {
             (rs, rns) = split_rulesa i trs rremove;
           } in bindb (valid rp)
                  (\ _ ->
                    bindb (catcha (mono rp (rs ++ rns))
                            (\ x ->
                              Inl (shows_string
                                     ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                       'w', 'i', 't', 'h', ' ', 'm', 'o', 'n',
                                       'o', 't', 'o', 'n', 'i', 'c', 'i', 't',
                                       'y', ' ', 'o', 'f', ' ', 's', 't', 'r',
                                       'i', 'c', 't', ' ', 'o', 'r', 'd', 'e',
                                       'r'] .
                                    shows_nl . x)))
                      (\ _ ->
                        bindb (catcha
                                (catcha (forallm (ns rp) rns)
                                  (\ x -> Inl (snd x)))
                                (\ x ->
                                  Inl (shows_string
 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
   'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'T', 'R', 'S'] .
shows_nl . x)))
                          (\ _ ->
                            catcha
                              (catcha (forallm (s rp) rs) (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'T',
 'R', 'S'] .
                                      shows_nl . x))))))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                     'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
                     'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p',
                     'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't',
                     'h', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w',
                     'i', 'n', 'g'] .
                  shows_nl . desca rp . shows_nl . x))
    of {
    Inl a -> Inl a;
    Inr _ -> Inr (delete_R_Rwa i trs rremove rremove);
  });

sem_lab_rel_tt ::
  forall a b c d.
    (Eq a, Showa a, Eq b,
      Showa b) => ([(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      ([(Term a b, Term a b)],
                        ([(Term a b, Term a b)], [(Term a b, Term a b)]))) ->
                    (a -> (a, c)) ->
                      Tp_ops_ext d a b () ->
                        Sum ([Char] -> [Char]) () ->
                          ([(Term a b, Term a b)] ->
                            Sum ([Char] -> [Char]) ()) ->
                            ([(Term a b, Term a b)] ->
                              [(Term a b, Term a b)] ->
                                Sum ([Char] -> [Char]) ()) ->
                              [Term a b] ->
                                [(Term a b, Term a b)] ->
                                  d -> Sum ([Char] -> [Char]) d;
sem_lab_rel_tt splitter ld i valid check_decr check_model_lab lQ lAll tp =
  let {
    r = ra i tp;
    rw = rwa i tp;
    nfs = nfsa i tp;
    (lR, (lRw, d)) = splitter lAll rw;
  } in (case bindb valid
               (\ _ ->
                 let {
                   q = qa i tp;
                 } in catcha
                        (bindb (if nfs then check_wf_trs d else Inr ())
                          (\ _ ->
                            bindb (check_decr d)
                              (\ _ ->
                                bindb (check_sl_Qa ld lQ q)
                                  (\ _ ->
                                    bindb (check_model_lab lR r)
                                      (\ _ -> check_model_lab lRw rw)))))
                        (\ x ->
                          Inl (shows_string
                                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                   'i', 't', 'h', ' ', 'l', 'a', 'b', 'e', 'l',
                                   'e', 'd', ' ', 'T', 'R', 'S', ':'] .
                                shows_nl . x)))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mka i nfs lQ lR (lRw ++ d));
       });

sem_lab_fin_tt ::
  forall a b c d e.
    (Eq a, Showa a, Eq b, Showa b,
      Showa d) => ([(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      ([(Term a b, Term a b)],
                        ([(Term a b, Term a b)], [(Term a b, Term a b)]))) ->
                    (a -> Nat -> c -> a) ->
                      (a -> (a, c)) ->
                        (d -> d -> Bool) ->
                          Tp_ops_ext e a b () ->
                            ([(a, Nat)] ->
                              [(a, Nat)] ->
                                Sum ([Char] -> [Char])
                                  (Sl_ops_ext a d c b ())) ->
                              [Term a b] ->
                                [(Term a b, Term a b)] ->
                                  e -> Sum ([Char] -> [Char]) e;
sem_lab_fin_tt splitter lc ld cge i gen lQ lAll tp =
  bindb (gen (funas_trs_list (rulesb i tp)) [])
    (\ ops ->
      let {
        check_ml =
          check_sl_model_lab_trs (sl_I ops) (sl_L ops) (sl_C ops) cge lc;
        check_d = sl_check_decr ops;
      } in sem_lab_rel_tt splitter ld i (Inr ()) check_d check_ml lQ lAll tp);

semlab_fin_tt ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Tp_ops_ext a (Lab b [Nat]) c () ->
                    Sl_variant (Lab b [Nat]) c ->
                      [Term (Lab b [Nat]) c] ->
                        [(Term (Lab b [Nat]) c, Term (Lab b [Nat]) c)] ->
                          a -> Sum ([Char] -> [Char]) a;
semlab_fin_tt j (Rootlab uu) =
  sem_lab_fin_tt (model_splitter label_decomp) label label_decomp equal_lab j
    (slm_gen_to_sl_gen (rl_slm Nothing));
semlab_fin_tt j (Finitelab sli) =
  sem_lab_fin_tt (model_splitter label_decomp) label label_decomp equal_nat j
    (slm_gen_to_sl_gen (\ _ _ -> Inr (sli_to_slm sli)));
semlab_fin_tt j (QuasiFinitelab sli v) =
  sem_lab_fin_tt (quasi_splitter label_decomp) label label_decomp qmodel_cge j
    (\ f g -> qsli_to_sl v f g sli);

rev_term :: forall a b. Ctxt a b -> Term a b -> Term a b;
rev_term c (Var x) = ctxt_apply c (Var x);
rev_term c (Fun f [t]) = rev_term (More f [] c []) t;

rev_rule :: forall a b. (Term a b, Term a b) -> (Term a b, Term a b);
rev_rule (l, r) = (rev_term Hole l, rev_term Hole r);

unary_term :: forall a b. Term a b -> Bool;
unary_term (Var x) = True;
unary_term (Fun f [t]) = unary_term t;
unary_term (Fun v []) = False;
unary_term (Fun v (vb : vd : ve)) = False;

check_unary_signature ::
  forall a b.
    (Showa a, Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_unary_signature r =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (l, ra) = x;
              } in unary_term l && unary_term ra
            then Inr () else Inl x))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
            shows_nl .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '-', '>', ' '] x .
                shows_nl .
                  shows_string
                    ['v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 't', 'h', 'e',
                      ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n', ' ',
                      't', 'h', 'a', 't', ' ', 'a', 'l', 'l', ' ', 'f', 'u',
                      'n', 'c', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm', 'b',
                      'o', 'l', 's'] .
                    shows_nl .
                      shows_prec_list Zero_nat
                        ['h', 'a', 'v', 'e', ' ', 't', 'o', ' ', 'b', 'e', ' ',
                          'u', 'n', 'a', 'r', 'y']));

string_reversal_tt ::
  forall a b c.
    (Showa b, Showa c) => Tp_ops_ext a b c () -> a -> Sum ([Char] -> [Char]) a;
string_reversal_tt i trs =
  let {
    rs = rulesb i trs;
    r = ra i trs;
    s = rwa i trs;
  } in (case check_unary_signature rs of {
         Inl a -> Inl a;
         Inr _ -> Inr (mka i True [] (map rev_rule r) (map rev_rule s));
       });

switch_innermost_tt ::
  forall a b.
    (Eq b,
      Showa b) => Tp_ops_ext a b [Char] () ->
                    Join_info b -> a -> Sum ([Char] -> [Char]) a;
switch_innermost_tt i joins_i trs =
  let {
    r = rulesb i trs;
  } in (case let {
               cp = critical_pairs_impl r r;
             } in bindb (catcha
                          (forallm
                            (\ (b, _) ->
                              check b
                                (shows_prec_list Zero_nat
                                  ['r', 'u', 'l', 'e', 's', ' ', 'a', 'r', 'e',
                                    ' ', 'n', 'o', 't', ' ', 'o', 'v', 'e', 'r',
                                    'l', 'a', 'y']))
                            cp)
                          (\ x -> Inl (snd x)))
                    (\ _ ->
                      bindb (check_critical_pairs r cp joins_i)
                        (\ _ -> check_wf_trs r))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mka i True (map fst r) r []);
       });

instance (Eq a, Linorder a, Eq b, Linorder b) => Linorder (Term a b) where {
};

dP_list ::
  forall a b.
    (Eq a,
      Eq b) => (a -> a) ->
                 [(Term a b, Term a b)] -> [(a, Nat)] -> [(Term a b, Term a b)];
dP_list shp r d_list =
  concatMap
    (\ lr ->
      let {
        l = fst lr;
        s = sharp_term shp l;
      } in map_filter
             (\ x ->
               (if not (supt_impl l x) &&
                     not (is_Var x) && membera d_list (the (root x))
                 then Just (s, sharp_term shp x) else Nothing))
             (supteq_list (snd lr)))
    r;

is_QNFa :: forall a b c d. Tp_ops_ext a b c d -> a -> Term b c -> Bool;
is_QNFa
  (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs more)
  = is_QNF;

sorted_list_subset :: forall a. (Eq a, Linorder a) => [a] -> [a] -> Maybe a;
sorted_list_subset (a : asa) (b : bs) =
  (if a == b then sorted_list_subset asa (b : bs)
    else (if less b a then sorted_list_subset (a : asa) bs else Just a));
sorted_list_subset [] uu = Nothing;
sorted_list_subset (a : uv) [] = Just a;

check_subseteq_sorted :: forall a. (Eq a, Linorder a) => [a] -> [a] -> Sum a ();
check_subseteq_sorted xs ys =
  (case sorted_list_subset (sort_key (\ x -> x) xs) (sort_key (\ x -> x) ys) of
    {
    Nothing -> Inr ();
    Just a -> Inl a;
  });

applicable_rule_impla ::
  forall a b. (Term a b -> Bool) -> (Term a b, Term a b) -> Bool;
applicable_rule_impla q lr = all q (args (fst lr));

dependency_pairs_tt ::
  forall a b c d.
    (Eq b, Linorder b, Showa b, Eq c, Linorder c,
      Showa c) => Tp_ops_ext a b c () ->
                    Dpp_ops_ext d b c () ->
                      a -> (b -> b) ->
                             [(Term b c, Term b c)] -> Sum ([Char] -> [Char]) d;
dependency_pairs_tt i j tp shp p =
  let {
    r = rulesb i tp;
    q = qa i tp;
    iQ = is_QNFa i tp;
    u = filter (applicable_rule_impla iQ) r;
  } in (case catcha
               (bindb (check_wf_trs u)
                 (\ _ ->
                   bindb (catcha (forallm check_no_var q) (\ x -> Inl (snd x)))
                     (\ _ ->
                       let {
                         qr = map (\ (Fun f ss) -> (f, size_list ss)) q;
                         d = defs_list u;
                       } in bindb (catcha
                                    (forallm
                                      (\ (f, n) ->
check (not (membera d (shp f, n)))
  (shows_string ['s', 'h', 'a', 'r', 'p', 'i', 'n', 'g', ' '] .
    shows_prec Zero_nat f .
      shows_string
        [' ', 'y', 'i', 'e', 'l', 'd', 's', ' ', 't', 'h', 'e', ' ', 'd', 'e',
          'f', 'i', 'n', 'e', 'd', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' '] .
        shows_prec Zero_nat (shp f)))
                                      d)
                                    (\ x -> Inl (snd x)))
                              (\ _ ->
                                bindb (catcha
(forallm
  (\ (f, n) ->
    check (not (membera qr (shp f, n)))
      (shows_string ['s', 'h', 'a', 'r', 'p', 'i', 'n', 'g', ' '] .
        shows_prec Zero_nat f .
          shows_string
            [' ', 'y', 'i', 'e', 'l', 'd', 's', ' ', 't', 'h', 'e', ' ', 's',
              'y', 'm', 'b', 'o', 'l', ' '] .
            shows_prec Zero_nat (shp f) .
              shows_string
                [' ', 'w', 'h', 'i', 'c', 'h', ' ', 'i', 's', ' ', 'a', ' ',
                  'r', 'o', 'o', 't', ' ', 'o', 'f', ' ', 'Q']))
  d)
(\ x -> Inl (snd x)))
                                  (\ _ ->
                                    catcha
                                      (check_subseteq_sorted (dP_list shp u d)
p)
                                      (\ x ->
Inl (shows_string ['t', 'h', 'e', ' ', 'D', 'P', ' '] .
      shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
        [' ', '-', '>', ' '] x .
        shows_string
          [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p', 'e',
            'a', 'r', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ',
            'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
          shows_nl)))))))
               (\ x ->
                 Inl (shows_string
                        ['t', 'h', 'e', ' ', 'D', 'P', '-', 't', 'r', 'a', 'n',
                          's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o', 'n', ' ',
                          'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p', 'l',
                          'i', 'e', 'd', ' ', 'c', 'o', 'r', 'r', 'e', 'c', 't',
                          'l', 'y', '.'] .
                       shows_nl . x))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc j True True p [] q [] r);
       });

check_trs_proof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          a -> Trs_proof b [Nat] [Char] ->
                                 Sum ([Char] -> [Char]) ();
check_trs_proof a ia i j tp RisEmpty =
  debug (ia []) ['R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (if null (ra i tp) then Inr ()
      else Inl (ia . shows_string
                       [':', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
                         'e', 'm', 'p', 't', 'y', ' ', 'i', 'n', ' ', 't', 'h',
                         'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                         ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
                         'n', '-', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                       shows_nl .
                         shows_tp (shows_prec_lab Zero_nat)
                           (shows_prec_list Zero_nat) i tp));
check_trs_proof a ia i j tp (RuleRemoval redp rR prf) =
  debug (ia []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (let {
       r = rulesb i tp;
       rr = list_diff r rR;
     } in bindb (catcha (rule_removal_tt i (get_redtriple redp) rr tp)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 't', 'h', 'e', ' ', 'r',
                                  'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                                  'a', 'l', ' ', 't', 'e', 'c', 'h', 'n', 'i',
                                  'q', 'u', 'e', ' ', 'o', 'n', ' '] .
                                shows_nl .
                                  shows_tp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i tp .
                                    shows_nl .
                                      shows_string
[' ', 't', 'o', ' ', 'r', 'e', 'm', 'o', 'v', 'e', ' ', 't', 'h', 'e', ' ', 'r',
  'u', 'l', 'e', 's', ' '] .
shows_nl .
  shows_rules (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat)
    [' ', '-', '>', ' '] rr .
    shows_nl . x)))
            (\ tpa ->
              catcha
                (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r',
                                'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v',
                                'a', 'l', ' ', 't', 'e', 'c', 'h', 'n', 'i',
                                'q', 'u', 'e'] .
                              shows_nl . indent x))));
check_trs_proof a ia i j tp (DPTrans p prf) =
  debug (ia []) ['D', 'P', ' ', 't', 'r', 'a', 'n', 's']
    (bindb
      (catcha (dependency_pairs_tt i j tp Sharp p)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g',
                        ' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T',
                        'R', 'S'] .
                      shows_nl .
                        shows_tp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i tp .
                          shows_nl .
                            shows_string
                              ['t', 'o', ' ', 't', 'h', 'e', ' ', 'i', 'n', 'i',
                                't', 'i', 'a', 'l', ' ', 'D', 'P', ' ', 'p',
                                'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                                't', 'h', ' ', 'p', 'a', 'i', 'r', 's', ' '] .
                              shows_nl .
                                shows_rules (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  [' ', '-', '>', ' '] p .
                                  shows_nl . x)))
      (\ dpp ->
        catcha
          (check_dp_proof_with_unlab a (ia . shows_string ['.', '1']) j dpp prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 't',
                          'o', ' ', 'd', 'e', 'p', 'e', 'n', 'd', 'e', 'n', 'c',
                          'y', ' ', 'p', 'a', 'i', 'r', 's'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (StringReversal prf) =
  debug (ia [])
    ['S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l']
    (bindb
      (catcha (string_reversal_tt i tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e',
                        'r', 's', 'a', 'l', ' ', 'o', 'n', ' '] .
                      shows_nl .
                        shows_tp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i tp .
                          shows_nl . x)))
      (\ tpa ->
        catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 't', 'r', 'i',
                          'n', 'g', ' ', 'r', 'e', 'v', 'e', 'r', 's', 'a', 'l',
                          ' ', 't', 'e', 'c', 'h', 'n', 'i', 'q', 'u', 'e'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (Semlab sli lQ lAll prf) =
  debug (ia []) ['S', 'e', 'm', 'l', 'a', 'b']
    (bindb
      (catcha (semlab_fin_tt i sli lQ lAll tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a',
                        'b', 'e', 'l', 'l', 'i', 'n', 'g', ' ', 'o', 'n', ' '] .
                      shows_nl .
                        shows_tp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i tp .
                          shows_nl . x)))
      (\ tpa ->
        catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'e', 'm', '.',
                          'l', 'a', 'b', ' ', 't', 'e', 'c', 'h', 'n', 'i', 'q',
                          'u', 'e'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (Bounds info) =
  debug (ia []) ['B', 'o', 'u', 'n', 'd', 's']
    (catcha (bounds_tt i info tp)
      (\ x ->
        Inl (ia . shows_string
                    [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n',
                      ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ', 'b',
                      'o', 'u', 'n', 'd', 's', ' ', 'o', 'n', ' ', 't', 'h',
                      'e', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
                      'o', 'n', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                    shows_nl .
                      shows_tp (shows_prec_lab Zero_nat)
                        (shows_prec_list Zero_nat) i tp .
                        shows_nl . x)));
check_trs_proof a ia i j tp (Uncurry u_info rR prf) =
  debug (ia []) ['U', 'n', 'c', 'u', 'r', 'r', 'y']
    (bindb
      (catcha (uncurry_tt i u_info rR tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'r',
                        'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o',
                        'n'] .
                      shows_nl . x)))
      (\ tpa ->
        catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u',
                          'r', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'e', 'c', 'h',
                          'n', 'i', 'q', 'u', 'e'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (Fcc fcs r prf) =
  debug (ia []) ['F', 'c', 'c']
    (bindb
      (catcha (fcc_tt i fcs r tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
                        't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e'] .
                      shows_nl . x)))
      (\ tpa ->
        catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n',
                          't', 'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r',
                          'e'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (Split rrem prf1 prf2) =
  debug (ia []) ['S', 'p', 'l', 'i', 't']
    (let {
       (tp1, tp2) = split_tt i tp rrem;
     } in bindb (catcha
                  (check_trs_proof a (ia . shows_string ['.', '1']) i j tp1
                    prf1)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                  'e', 'l', 'o', 'w', ' ', 't', 'h', 'e', ' ',
                                  's', 'p', 'l', 'i', 't', 't', 'i', 'n', 'g'] .
                                shows_nl . indent x)))
            (\ _ ->
              catcha
                (check_trs_proof a (ia . shows_string ['.', '2']) i j tp2 prf2)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 's',
                                'p', 'l', 'i', 't', 't', 'i', 'n', 'g'] .
                              shows_nl . indent x))));
check_trs_proof a ia i j tp (SwitchInnermost joins prf) =
  debug (ia [])
    ['S', 'w', 'i', 't', 'c', 'h', ' ', 'I', 'n', 'n', 'e', 'r', 'm', 'o', 's',
      't']
    (bindb
      (catcha (switch_innermost_tt i joins tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g',
                        ' ', 't', 'o', ' ', 'i', 'n', 'n', 'e', 'r', 'm', 'o',
                        's', 't', ' ', 'o', 'n', ' '] .
                      shows_nl .
                        shows_tp (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) i tp .
                          shows_nl . x)))
      (\ tpa ->
        catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'w', 'i', 't',
                          'c', 'h', ' ', 't', 'o', ' ', 'i', 'n', 'n', 'e', 'r',
                          'm', 'o', 's', 't'] .
                        shows_nl . indent x))));
check_trs_proof a ia i j tp (DropEquality prf) =
  debug (ia [])
    ['D', 'r', 'o', 'p', ' ', 'E', 'q', 'u', 'a', 'l', 'i', 't', 'y']
    (let {
       tpa = mka i (nfsa i tp) (qa i tp) (ra i tp)
               (filter (\ (l, r) -> not (equal_term l r)) (rwa i tp));
     } in catcha (check_trs_proof a (ia . shows_string ['.', '1']) i j tpa prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 'd', 'r', 'o', 'p', 'p', 'i',
                            'n', 'g', ' ', 'e', 'q', 'u', 'a', 'l', 'i', 't',
                            'y', ' ', 'r', 'u', 'l', 'e', 's'] .
                          shows_nl . indent x)));
check_trs_proof a ia i j tpa (AssumeSN tp) =
  debug (ia [])
    ['T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'A', 's', 's',
      'u', 'm', 'p', 't', 'i', 'o', 'n']
    (if a then catcha (check_tp_subsumes i tp tpa)
                 (\ x ->
                   Inl (ia . shows_string
                               [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i',
                                 'n', ' ', 't', 'e', 'r', 'm', 'i', 'n', 'a',
                                 't', 'i', 'o', 'n', ' ', 'a', 's', 's', 'u',
                                 'm', 'p', 't', 'i', 'o', 'n'] .
                               shows_nl . x . shows_nl))
      else Inl (ia . shows_string
                       [':', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'o', 'f',
                         ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 's', ' ', 'a',
                         ' ', 'l', 'o', 'c', 'a', 'l', ' ', 't', 'e', 'r', 'm',
                         'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'a', 's', 's',
                         'u', 'm', 'p', 't', 'i', 'o', 'n'] .
                       shows_nl));

check_cr_proof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          [(Term (Lab b [Nat]) [Char],
                             Term (Lab b [Nat]) [Char])] ->
                            Cr_proof b [Nat] [Char] ->
                              Sum ([Char] -> [Char]) ();
check_cr_proof a ia i j r (SN_WCR joins_i prf) =
  debug (ia []) ['S', 'N', '_', 'W', 'C', 'R']
    (let {
       tp = mka i False [] r [];
     } in bindb (catcha
                  (check_trs_proof a (ia . shows_string ['.', '1']) i j tp prf)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                  'e', 'l', 'o', 'w', ' ', 's', 't', 'r', 'o',
                                  'n', 'g', ' ', 'n', 'o', 'r', 'm', 'a', 'l',
                                  'i', 'z', 'a', 't', 'i', 'o', 'n', ' ', '+',
                                  ' ', 'w', 'c', 'r'] .
                                shows_nl . indent x)))
            (\ _ ->
              catcha (check_critical_pairs r (critical_pairs_impl r r) joins_i)
                (\ x ->
                  Inl (ia . shows_prec_list Zero_nat
                              ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n',
                                ' ', 'p', 'r', 'o', 'v', 'i', 'n', 'g', ' ',
                                'l', 'o', 'c', 'a', 'l', ' ', 'c', 'o', 'n',
                                'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'o',
                                'f', ' '] .
                              shows_tp (shows_prec_lab Zero_nat)
                                (shows_prec_list Zero_nat) i tp .
                                shows_nl . indent x))));
check_cr_proof a ia i j r Weakly_Orthogonal =
  debug (ia [])
    ['W', 'e', 'a', 'k', 'l', 'y', ' ', 'O', 'r', 't', 'h', 'o', 'g', 'o', 'n',
      'a', 'l']
    (catcha
      (bindb (check_left_linear_trs r)
        (\ _ ->
          bindb (catcha
                  (forallm
                    (\ (l, _) ->
                      check (not (is_Var l))
                        (shows_prec_list Zero_nat
                          ['t', 'h', 'e', ' ', 'T', 'R', 'S', ' ', 'h', 'a',
                            's', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                            's', ' ', 'a', 's', ' ', 'l', 'e', 'f', 't', '-',
                            'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', 's']))
                    r)
                  (\ x -> Inl (snd x)))
            (\ _ ->
              catcha
                (forallm
                  (\ (_, (s, t)) ->
                    check (equal_term s t)
                      (shows_prec_list Zero_nat
                         ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'a', ' ',
                           'n', 'o', 'n', '-', 't', 'r', 'i', 'v', 'i', 'a',
                           'l', ' ', 'c', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                           ' ', 'p', 'a', 'i', 'r', 's', ':', ' '] .
                        shows_term (shows_prec_lab Zero_nat)
                          (shows_prec_list Zero_nat) s .
                          shows_prec_list Zero_nat
                            [' ', '<', '-', ' ', '.', ' ', '-', '>', ' '] .
                            shows_term (shows_prec_lab Zero_nat)
                              (shows_prec_list Zero_nat) t))
                  (critical_pairs_impl r r))
                (\ x -> Inl (snd x)))))
      (\ x ->
        Inl (ia . shows_prec_list Zero_nat
                    [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'c',
                      'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'w', 'e', 'a',
                      'k', 'l', 'y', ' ', 'o', 'r', 't', 'h', 'o', 'g', 'o',
                      'n', 'a', 'l', ' ', 's', 'y', 's', 't', 'e', 'm', ' ',
                      'f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S'] .
                    shows_nl .
                      shows_trs (shows_prec_lab Zero_nat)
                        (shows_prec_list Zero_nat)
                        ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's',
                          't', 'e', 'm', ':']
                        [' ', '-', '>', ' '] r .
                        indent x)));

pat_dv_impl ::
  forall a b.
    (Eq a, Eq b) => (Term a b, ([(b, Term a b)], [(b, Term a b)])) -> [b];
pat_dv_impl p =
  let {
    (_, (sigma, mu)) = p;
  } in remdups (map fst (mk_subst_domain sigma ++ mk_subst_domain mu));

intersect_values ::
  forall a b.
    (Eq a, Linorder b) => (a -> Maybe b) -> [a] -> Rbt b [a] -> Rbt b [a];
intersect_values key vs m = foldra (aux key m) vs (rm_empty ());

intersect_rules ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => [(Term a b, Term a b)] ->
                 Rbt (a, Nat) [(Bool, (Term a b, Term a b))] ->
                   Rbt (a, Nat) [(Bool, (Term a b, Term a b))];
intersect_rules rs =
  intersect_values key
    (map (\ a -> (True, a)) rs ++ map (\ a -> (False, a)) rs);

doc_of_string :: [Char] -> Sum [Char] Xmldoc;
doc_of_string s = bindb (parse_doc s) (\ (doc, _) -> Inr doc);

starts_with :: [Char] -> [Char] -> Bool;
starts_with t s =
  (if not (null t) && membera [' ', newline, '\t', '\r'] (hd t)
    then starts_with (tl t) s
    else (case search_len t s (size_list s) of {
           Nothing -> False;
           Just _ -> True;
         }));

check_non_join ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Term a b ->
                      [(Pos, ((Term a b, Term a b), Term a b))] ->
                        [(Pos, ((Term a b, Term a b), Term a b))] ->
                          Non_join_info a b -> Sum ([Char] -> [Char]) ();
check_non_join r s seq1 seq2 reason =
  let {
    chk = check_rsteps_last r s;
  } in bindb (chk seq1)
         (\ _ ->
           bindb (chk seq2)
             (\ _ ->
               let {
                 u = rseq_last s seq1;
                 v = rseq_last s seq2;
               } in (case reason of {
                      Diff_NFs ->
                        bindb (check (not (equal_term u v))
                                (shows_prec_list Zero_nat
                                   ['t', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i',
                                     't', 'e', ' ', 's', 'e', 'q', 'u', 'e',
                                     'n', 'c', 'e', 's', ' ', 'e', 'n', 'd',
                                     ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ',
                                     's', 'a', 'm', 'e', ' ', 't', 'e', 'r',
                                     'm', ' '] .
                                  shows_prec_term Zero_nat u))
                          (\ _ ->
                            let {
                              chknf =
                                (\ sa ->
                                  check (nF_terms_list (map fst r) sa)
                                    (shows_prec_list Zero_nat
                                       ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm',
 ' '] .
                                      shows_term (shows_prec Zero_nat)
(shows_prec Zero_nat) sa .
shows_prec_list Zero_nat
  [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ', 'n', 'o', 'r', 'm',
    'a', 'l', ' ', 'f', 'o', 'r', 'm']));
                            } in bindb (chknf u) (\ _ -> chknf v));
                      Tcap_Non_Unif grd_subst ->
                        let {
                          sigma = grd_subst u v;
                          cu = capI r (subst_apply_term u sigma);
                          cv = capI r (subst_apply_term v sigma);
                        } in check (is_none (mergea cu cv))
                               (shows_prec_list Zero_nat
                                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                    ' ', 'i', 'n', 'f', 'e', 'r', ' ', 't', 'h',
                                    'a', 't', ' '] .
                                 shows_prec_term Zero_nat u .
                                   shows_prec_list Zero_nat
                                     [' ', 'a', 'n', 'd', ' '] .
                                     shows_prec_term Zero_nat v .
                                       shows_prec_list Zero_nat
 [' ', 'a', 'r', 'e', ' ', 'n', 'o', 't', ' ', 'j', 'o', 'i', 'n', 'a', 'b',
   'l', 'e']);
                    })));

check_ncr_proof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          [(Term (Lab b [Nat]) [Char],
                             Term (Lab b [Nat]) [Char])] ->
                            Ncr_proof b [Nat] [Char] ->
                              Sum ([Char] -> [Char]) ();
check_ncr_proof a ia i j r (SN_NWCR prf) =
  debug (ia []) ['S', 'N', '_', 'N', 'W', 'C', 'R']
    (let {
       tp = mka i False [] r [];
     } in bindb (catcha
                  (check_trs_proof a (ia . shows_string ['.', '1']) i j tp prf)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                  'e', 'l', 'o', 'w', ' ', 's', 't', 'r', 'o',
                                  'n', 'g', ' ', 'n', 'o', 'r', 'm', 'a', 'l',
                                  'i', 'z', 'a', 't', 'i', 'o', 'n', ' ', '+',
                                  ' ', 'w', 'c', 'r'] .
                                shows_nl . indent x)))
            (\ _ ->
              catcha
                (check
                  (not (isOK (check_critical_pairs_NF r
                               (critical_pairs_impl r r))))
                  (shows_prec_list Zero_nat
                    ['a', 'l', 'l', ' ', 'c', 'r', 'i', 't', 'i', 'c', 'a', 'l',
                      ' ', 'p', 'a', 'i', 'r', 's', ' ', 'a', 'r', 'e', ' ',
                      'j', 'o', 'i', 'n', 'a', 'b', 'l', 'e']))
                (\ x ->
                  Inl (ia . shows_prec_list Zero_nat
                              ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n',
                                ' ', 'd', 'i', 's', 'p', 'r', 'o', 'v', 'i',
                                'n', 'g', ' ', 'l', 'o', 'c', 'a', 'l', ' ',
                                'c', 'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c',
                                'e', ' ', 'o', 'f', ' '] .
                              shows_tp (shows_prec_lab Zero_nat)
                                (shows_prec_list Zero_nat) i tp .
                                shows_nl . indent x))));
check_ncr_proof a ia i j r (Non_Join s seq1 seq2 prf) =
  debug (ia []) ['N', 'o', 'n', '_', 'J', 'o', 'i', 'n']
    (catcha (check_non_join r s seq1 seq2 prf)
      (\ x ->
        Inl (ia . shows_prec_list Zero_nat
                    ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'd',
                      'i', 's', 'p', 'r', 'o', 'v', 'i', 'n', 'g', ' ', 'C',
                      'R', ' ', 'o', 'f', ' '] .
                    shows_trs (shows_prec_lab Zero_nat)
                      (shows_prec_list Zero_nat)
                      ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's',
                        't', 'e', 'm', ':']
                      [' ', '-', '>', ' '] r .
                      shows_nl . indent x)));

conflicts ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Term b a) ->
                 (Term b a, (Term b a, Nat)) -> [(Term b a, (Term b a, Nat))];
conflicts mu (Fun f ts, (Fun g ss, n)) =
  (if f == g && equal_nat (size_list ts) (size_list ss)
    then concatMap (\ (s, t) -> conflicts mu (s, (t, n))) (zip ts ss)
    else [(Fun f ts, (Fun g ss, n))]);
conflicts mu (Var x, (Fun g ss, n)) = [(Var x, (Fun g ss, n))];
conflicts mu (Fun f ts, (Var y, Zero_nat)) = [(Var y, (Fun f ts, Zero_nat))];
conflicts mu (Var x, (Var y, Zero_nat)) =
  (if x == y then [] else [(Var x, (Var y, Zero_nat))]);
conflicts mu (s, (Var x, Suc n)) = conflicts mu (s, (mu x, n));

vars_subst_impl :: forall a b. (Eq a, Eq b) => [(a, Term b a)] -> [a];
vars_subst_impl sigma =
  let {
    sigmaa = mk_subst_domain sigma;
  } in map fst sigmaa ++ concatMap (vars_term_lista . snd) sigmaa;

xml2dp_disproof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Dp_disproof a [Char]);
xml2dp_disproof xml2name x =
  xml1element
    ['d', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
      'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
         xml_choice
           ['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
           [xml3elements
              ['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
              (xml1element ['d', 'p', 's'] (rules xml2name) Just)
              (xml1element ['t', 'r', 's'] (rules xml2name) Just)
              (xml2dp_disproof xml2name) DP_Rule_Removal,
             xml2elements
               ['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
               (xml1element ['t', 'r', 's'] (rules xml2name) Just)
               (xml2dp_disproof xml2name) (DP_Rule_Removal Nothing),
             xml2elements
               ['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
               (xml1element ['d', 'p', 's'] (rules xml2name) Just)
               (xml2dp_disproof xml2name) (\ p -> DP_Rule_Removal p Nothing)]),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) -> DP_Loop s rseq a b)),
        (['n', 'o', 'n', 'L', 'o', 'o', 'p'],
          xml_change (nonloop xml2name) DP_Nonloop),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's', 'R',
           'e', 'm', 'o', 'v', 'a', 'l', 'P', 'r', 'o', 'c'],
          xml2elements
            ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's',
              'R', 'e', 'm', 'o', 'v', 'a', 'l', 'P', 'r', 'o', 'c']
            (innermostLhss xml2name) (xml2dp_disproof xml2name) DP_Q_Reduction),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's', 'I',
           'n', 'c', 'r', 'e', 'a', 's', 'e', 'P', 'r', 'o', 'c'],
          xml2elements
            ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's',
              'I', 'n', 'c', 'r', 'e', 'a', 's', 'e', 'P', 'r', 'o', 'c']
            (innermostLhss xml2name) (xml2dp_disproof xml2name) DP_Q_Increase),
        (['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n', 'P',
           'r', 'o', 'c'],
          xml2elements
            ['i', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n',
              'P', 'r', 'o', 'c']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml2dp_disproof xml2name) DP_Instantiation),
        (['s', 'w', 'i', 't', 'c', 'h', 'F', 'u', 'l', 'l', 'S', 't', 'r', 'a',
           't', 'e', 'g', 'y', 'P', 'r', 'o', 'c'],
          xml2elements
            ['s', 'w', 'i', 't', 'c', 'h', 'F', 'u', 'l', 'l', 'S', 't', 'r',
              'a', 't', 'e', 'g', 'y', 'P', 'r', 'o', 'c']
            (wcr_proof xml2name) (xml2dp_disproof xml2name)
            DP_Termination_Switch)])
    id x;

eqWhiteSpace :: [Char] -> [Char] -> Bool;
eqWhiteSpace s1 s2 =
  debug ['0']
    ['c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'w', 'h', 'e', 't', 'h',
      'e', 'r', ' ', 'p', 'a', 'r', 's', 'e', 'd', ' ', 'i', 'n', 'p', 'u', 't',
      ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', 's', ' ', 't', 'o',
      ' ', 'r', 'e', 'a', 'l', ' ', 'i', 'n', 'p', 'u', 't']
    (filter (\ c -> not (membera [' ', newline, '\t', '\r'] c)) s1 == s2);

show_pat_term ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b, ([(b, Term a b)], [(b, Term a b)])) ->
                    [Char] -> [Char];
show_pat_term p =
  let {
    (s, (sigma, tau)) = p;
  } in shows_prec_prod Zero_nat
         (s, (mk_subst_domain sigma, mk_subst_domain tau));

show_pat_rule ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ((Term a b, ([(b, Term a b)], [(b, Term a b)])),
                    ((Term a b, ([(b, Term a b)], [(b, Term a b)])), Bool)) ->
                    [Char] -> [Char];
show_pat_rule pr =
  let {
    (p1, (p2, _)) = pr;
  } in show_pat_term p1 .
         shows_string [' ', '-', '-', '>', ' '] . show_pat_term p2;

rel_rules_of ::
  forall a b. Maybe [(Term a b, Term a b)] -> [(Term a b, Term a b)];
rel_rules_of Nothing = [];
rel_rules_of (Just r) = r;

subst_power_impl ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> Nat -> [(a, Term b a)];
subst_power_impl sigma Zero_nat = [];
subst_power_impl sigma (Suc n) =
  subst_compose_impl sigma (subst_power_impl sigma n);

newtype (Linorder a) => Tp a b = Tp
  (Bool,
    ([Term a b],
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            Rbt (a, Nat) [(Bool, (Term a b, Term a b))])))));

q_impl ::
  forall a b.
    (Bool,
      ([Term a b],
        (Bool,
          ([(Term a b, Term a b)],
            ([(Term a b, Term a b)],
              Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
      [Term a b];
q_impl (uu, (q, uv)) = q;

impl_ofa ::
  forall a b.
    (Linorder a) => Tp a b ->
                      (Bool,
                        ([Term a b],
                          (Bool,
                            ([(Term a b, Term a b)],
                              ([(Term a b, Term a b)],
                                Rbt (a, Nat)
                                  [(Bool, (Term a b, Term a b))])))));
impl_ofa (Tp x) = x;

q :: forall a b. (Linorder a) => Tp a b -> [Term a b];
q tp = q_impl (impl_ofa tp);

r_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat)
                                 [(Bool, (Term a b, Term a b))]))))) ->
                       [(Term a b, Term a b)];
r_impl (uu, (uv, (uw, (vR, (ux, m))))) = vR ++ rules_with id m;

r :: forall a b. (Eq a, Linorder a) => Tp a b -> [(Term a b, Term a b)];
r tp = r_impl (impl_ofa tp);

xml2reltrs_proof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Trs_proof a [Nat] [Char]);
xml2reltrs_proof xml2name x =
  xml1element
    ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a',
      't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['r', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['r', 'I', 's', 'E', 'm', 'p', 't', 'y'] RisEmpty),
        (['s', 'I', 's', 'E', 'm', 'p', 't', 'y'],
          xml1element ['s', 'I', 's', 'E', 'm', 'p', 't', 'y']
            (xml2trs_proof xml2name) id),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml4elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (ordering_constraint_proofa xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2reltrs_proof xml2name)
            (\ rp one two -> RuleRemoval rp (one ++ two))),
        (['s', 'e', 'm', 'l', 'a', 'b'],
          xml_choice ['s', 'e', 'm', 'l', 'a', 'b']
            [xml4elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_variant xml2name)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml2reltrs_proof xml2name)
               (\ sli lr lrw -> Semlab sli [] (lr ++ lrw)),
              xml5elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_variant xml2name)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (innermostLhss xml2name) (xml2reltrs_proof xml2name)
                (\ sli lr lrw lq -> Semlab sli lq (lr ++ lrw))]),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y'],
          xml4elements ['u', 'n', 'c', 'u', 'r', 'r', 'y']
            (uncurry_info xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2reltrs_proof xml2name) (\ i r s -> Uncurry i (r ++ s))),
        (['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'],
          xml3elements
            ['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a',
              'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2reltrs_proof xml2name) (\ _ _ -> StringReversal)),
        (['e', 'q', 'u', 'a', 'l', 'i', 't', 'y', 'R', 'e', 'm', 'o', 'v', 'a',
           'l'],
          xml1element
            ['e', 'q', 'u', 'a', 'l', 'i', 't', 'y', 'R', 'e', 'm', 'o', 'v',
              'a', 'l']
            (xml2reltrs_proof xml2name) DropEquality),
        (['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n',
           'a', 't', 'i', 'o', 'n', 'A', 's', 's', 'u', 'm', 'p', 't', 'i', 'o',
           'n'],
          xml1element
            ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i',
              'n', 'a', 't', 'i', 'o', 'n', 'A', 's', 's', 'u', 'm', 'p', 't',
              'i', 'o', 'n']
            (xml1to2elements ['t', 'r', 's', 'I', 'n', 'p', 'u', 't']
              (xml1element ['t', 'r', 's'] (rules xml2name) id)
              (xml1element
                ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e',
                  's']
                (rules xml2name) id)
              (\ r a ->
                (case a of {
                  Nothing -> AssumeSN (False, ([], (r, [])));
                  Just rw -> AssumeSN (False, ([], (r, rw)));
                })))
            id)])
    id x;

xml2trs_disproof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Trs_disproof a [Char]);
xml2trs_disproof xml2name x =
  xml1element
    ['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
      'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
          'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'],
         xml0elements
           ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i',
             't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd']
           TRS_Not_Well_Formed),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) ->
              TRS_Loop s (map (\ (xa, (y, (_, z))) -> (xa, (y, z))) rseq) a b)),
        (['n', 'o', 'n', 'L', 'o', 'o', 'p'],
          xml_change (nonloop xml2name) TRS_Nonloop),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml2elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_disproof xml2name) TRS_Rule_Removal),
        (['d', 'p', 'T', 'r', 'a', 'n', 's'],
          xml3elements ['d', 'p', 'T', 'r', 'a', 'n', 's']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (bool ['m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                    's'])
            (xml2dp_disproof xml2name) (\ p _ -> TRS_DP_Trans p)),
        (['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'],
          xml2elements
            ['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a',
              'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trs_disproof xml2name) (\ _ -> TRS_String_Reversal)),
        (['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's', 'I',
           'n', 'c', 'r', 'e', 'a', 's', 'e'],
          xml2elements
            ['i', 'n', 'n', 'e', 'r', 'm', 'o', 's', 't', 'L', 'h', 's', 's',
              'I', 'n', 'c', 'r', 'e', 'a', 's', 'e']
            (innermostLhss xml2name) (xml2trs_disproof xml2name)
            TRS_Q_Increase)])
    id x;

xml2otrs_disproof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Otrs_disproof a [Char]);
xml2otrs_disproof xml2name x =
  xml1element
    ['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
      'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['l', 'o', 'o', 'p'],
         xml_change (loopa xml2name)
           (\ (s, (rseq, (a, b))) ->
             OTRS_Loop s (map (\ (xa, (y, (_, z))) -> (xa, (y, z))) rseq) a b)),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml2elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2otrs_disproof xml2name) OTRS_Rule_Removal)])
    id x;

xml2reltrs_disproof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Reltrs_disproof a [Char]);
xml2reltrs_disproof xml2name x =
  xml1element
    ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e', 'r', 'm',
      'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
          'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'],
         xml0elements
           ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i',
             't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd']
           Rel_Not_Well_Formed),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) -> Rel_Loop s rseq a b)),
        (['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
           'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'],
          xml_change (xml2trs_disproof xml2name) Rel_R_Not_SN),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml3elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) Just)
            (xml1element ['t', 'r', 's'] (rules xml2name) Just)
            (xml2reltrs_disproof xml2name) Rel_Rule_Removal)])
    id x;

subsumption_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml ->
        Parse_result [Char]
          [((Term a [Char], Term a [Char]),
             [(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))])];
subsumption_proof xml2name =
  xml_many_elements
    ['s', 'u', 'b', 's', 'u', 'm', 'p', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
      'f']
    (xml2elements
      ['r', 'u', 'l', 'e', 'S', 'u', 'b', 's', 'u', 'm', 'p', 't', 'i', 'o',
        'n', 'P', 'r', 'o', 'o', 'f']
      (rule xml2name) (conversion xml2name) (\ a b -> (a, b)))
    id;

xml2completion_proof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Completion_proof a [Nat] [Char]);
xml2completion_proof xml2name =
  xml3elements
    ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (wcr_proof xml2name) (xml2trs_proof xml2name)
    (xml1to2elements
      ['e', 'q', 'u', 'i', 'v', 'a', 'l', 'e', 'n', 'c', 'e', 'P', 'r', 'o',
        'o', 'f']
      (subsumption_proof xml2name) (subsumption_proof xml2name)
      (\ a b -> (a, b)))
    (\ w t (a, b) -> SN_WCR_Eq w t a b);

xml2complexity_proof ::
  forall a.
    (Xml -> Parse_result [Char] a) ->
      Xml -> Parse_result [Char] (Complexity_proof a [Char]);
xml2complexity_proof xml2name x =
  xml1element
    ['c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['r', 'u', 'l', 'e', 'S', 'h', 'i', 'f', 't', 'i', 'n', 'g'],
         xml3elements
           ['r', 'u', 'l', 'e', 'S', 'h', 'i', 'f', 't', 'i', 'n', 'g']
           (ordering_constraint_proofa xml2name)
           (xml1element ['t', 'r', 's'] (rules xml2name) id)
           (xml2complexity_proof xml2name) Rule_Shift_Complexity),
        (['r', 'I', 's', 'E', 'm', 'p', 't', 'y'],
          xml0elements ['r', 'I', 's', 'E', 'm', 'p', 't', 'y']
            RisEmpty_Complexity)])
    id x;

xml2equational_proof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Equational_proof a [Nat] [Char]);
xml2equational_proof xml2name =
  xml1element
    ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'P', 'r', 'o', 'o',
          'f', 'T', 'r', 'e', 'e'],
         xml1element
           ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'P', 'r', 'o',
             'o', 'f', 'T', 'r', 'e', 'e']
           (xml2eq_proof xml2name) Equational_Proof_Tree),
        (['c', 'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n'],
          xml_change (conversion xml2name) Conversion),
        (['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'A', 'n', 'd', 'N',
           'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n'],
          xml2elements
            ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'A', 'n', 'd',
              'N', 'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2completion_proof xml2name) Completion_and_Normalization)])
    id;

xml2equational_disproof ::
  forall a.
    (Eq a) => (Xml -> Parse_result [Char] (Lab a [Nat])) ->
                Xml -> Parse_result [Char] (Equational_disproof a [Nat] [Char]);
xml2equational_disproof xml2name =
  xml1element
    ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'D', 'i', 's', 'p', 'r',
      'o', 'o', 'f']
    (xml_options
      [(['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'A', 'n', 'd', 'N',
          'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n'],
         xml2elements
           ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'A', 'n', 'd',
             'N', 'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n']
           (xml1element ['t', 'r', 's'] (rules xml2name) id)
           (xml2completion_proof xml2name)
           Completion_and_Normalization_Different)])
    id;

xml2cert_problem ::
  (Xml -> Parse_result [Char] [Char]) ->
    (Xml -> Parse_result [Char] (Lab [Char] [Nat])) ->
      Xml -> Parse_result [Char] (Cert_problem [Char] [Nat] [Char]);
xml2cert_problem xml2uname xml2name (Xml name egal cs texta) =
  (if name ==
        ['c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o', 'n', 'P',
          'r', 'o', 'b', 'l', 'e', 'm'] &&
        is_none texta && less_eq_nat (nat 3) (size_list cs)
    then bindc (xml2input xml2name (nth cs Zero_nat))
           (\ inp ->
             bindc (text ['c', 'p', 'f', 'V', 'e', 'r', 's', 'i', 'o', 'n']
                     (nth cs one_nat))
               (\ _ ->
                 (case inp of {
                   Inl (m, (p, (q, r))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['d', 'p', 'P', 'r', 'o', 'o', 'f'],
                            xml_change (xml2dp_proof xml2name)
                              (DP_Proof m p [] q [] r)),
                           (['d', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i',
                              'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o',
                              'f'],
                             xml_change (xml2dp_disproof xml2name)
                               (DP_Disproof m p q r))])
                       id (nth cs (nat 2));
                   Inr (Inl (q, (r, Nothing))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['t', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a',
                             't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'],
                            xml_change (xml2trs_proof xml2name)
                              (TRS_Proof q r Nothing)),
                           (['c', 'r', 'P', 'r', 'o', 'o', 'f'],
                             (if null (strategy_to_Q q r)
                               then xml_change (xml2cr_proof xml2name)
                                      (TRS_Confluence r)
                               else (\ _ ->
                                      errora
['s', 't', 'r', 'a', 't', 'e', 'g', 'i', 'e', 's', ' ', 'f', 'o', 'r', ' ', 'c',
  'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'u', 'n', 's', 'u', 'p',
  'p', 'o', 'r', 't', 'e', 'd']))),
                           (['c', 'r', 'D', 'i', 's', 'p', 'r', 'o', 'o', 'f'],
                             (if null (strategy_to_Q q r)
                               then xml_change (xml2ncr_proof xml2name)
                                      (TRS_Non_Confluence r)
                               else (\ _ ->
                                      errora
['s', 't', 'r', 'a', 't', 'e', 'g', 'i', 'e', 's', ' ', 'f', 'o', 'r', ' ', 'c',
  'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'u', 'n', 's', 'u', 'p',
  'p', 'o', 'r', 't', 'e', 'd']))),
                           (['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm',
                              'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o',
                              'o', 'f'],
                             xml_change (xml2trs_disproof xml2name)
                               (TRS_Disproof q r))])
                       id (nth cs (nat 2));
                   Inr (Inl (q, (r, Just s))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e',
                             'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P',
                             'r', 'o', 'o', 'f'],
                            xml_change (xml2reltrs_proof xml2name)
                              (TRS_Proof q r (Just s))),
                           (['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o',
                              'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
                              'o', 'n', 'P', 'r', 'o', 'o', 'f'],
                             xml_change (xml2reltrs_disproof xml2name)
                               (Rel_TRS_Disproof q r s))])
                       id (nth cs (nat 2));
                   Inr (Inr (Inl (e, r))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n',
                             'P', 'r', 'o', 'o', 'f'],
                            xml_change (xml2completion_proof xml2name)
                              (Completion_Proof e r))])
                       id (nth cs (nat 2));
                   Inr (Inr (Inr (Inl (eqs, eq)))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l',
                             'P', 'r', 'o', 'o', 'f'],
                            xml_change (xml2equational_proof xml2name)
                              (Equational_Proof eqs eq)),
                           (['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l',
                              'D', 'i', 's', 'p', 'r', 'o', 'o', 'f'],
                             xml_change (xml2equational_disproof xml2name)
                               (Equational_Disproof eqs eq))])
                       id (nth cs (nat 2));
                   Inr (Inr (Inr (Inr (Inl (q, (r, (s, (cm, cc)))))))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_change (xml2complexity_proof xml2name)
                         (Complexity_Proof q r s cm cc))
                       id (nth cs (nat 2));
                   Inr (Inr (Inr (Inr (Inr (q, r))))) ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_change (xml2otrs_disproof xml2name)
                         (OTRS_Disproof q r))
                       id (nth cs (nat 2));
                 })))
    else errora
           (['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'p', 'a', 'r',
              's', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'e', 'l', 'e', 'm', 'e',
              'n', 't', ' ', '<'] ++
             name ++ ['>']));

xml_complexity_class :: Complexity_class -> Xml;
xml_complexity_class (Comp_Poly n) =
  Xml ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l'] [] []
    (Just (shows_prec_nat Zero_nat n []));

xml_complexity_measure ::
  forall a b c. (Showa a, Showa b) => Complexity_measure (Lab a [b]) c -> Xml;
xml_complexity_measure (Derivational_Complexity f) =
  Xml ['d', 'e', 'r', 'i', 'v', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'C', 'o',
        'm', 'p', 'l', 'e', 'x', 'i', 't', 'y']
    [] [xml_signature f] Nothing;
xml_complexity_measure (Runtime_Complexity c d) =
  Xml ['r', 'u', 'n', 't', 'i', 'm', 'e', 'C', 'o', 'm', 'p', 'l', 'e', 'x',
        'i', 't', 'y']
    [] [xml_signature c, xml_signature d] Nothing;

xml_cert_problem ::
  forall a b c. (Showa a, Showa b, Showa c) => Cert_problem a [b] c -> Xml;
xml_cert_problem (TRS_Proof q r s_o uu) = xml_trs_input (Inl q) r s_o;
xml_cert_problem (TRS_Disproof q r uv) = xml_trs_input (Inl q) r Nothing;
xml_cert_problem (OTRS_Disproof q r uw) = xml_trs_input (Inr q) r Nothing;
xml_cert_problem (Rel_TRS_Disproof q r s ux) = xml_trs_input (Inl q) r (Just s);
xml_cert_problem (DP_Proof m p pw q r rw uy) =
  Xml ['d', 'p', 'I', 'n', 'p', 'u', 't'] []
    (xml_rules ['t', 'r', 's'] rw :
      xml_rules ['d', 'p', 's'] p :
        xml_Q (Inl q) ++
          [Xml ['m', 'i', 'n', 'i', 'm', 'a', 'l'] [] []
             (Just (if m then ['t', 'r', 'u', 'e']
                     else ['f', 'a', 'l', 's', 'e']))])
    Nothing;
xml_cert_problem (DP_Disproof m p q r uz) =
  Xml ['d', 'p', 'I', 'n', 'p', 'u', 't'] []
    (xml_rules ['t', 'r', 's'] r :
      xml_rules ['d', 'p', 's'] p :
        xml_Q (Inl q) ++
          [Xml ['m', 'i', 'n', 'i', 'm', 'a', 'l'] [] []
             (Just (if m then ['t', 'r', 'u', 'e']
                     else ['f', 'a', 'l', 's', 'e']))])
    Nothing;
xml_cert_problem (TRS_Confluence r va) =
  xml_trs_input (Inl No_Strategy) r Nothing;
xml_cert_problem (TRS_Non_Confluence r vb) =
  xml_trs_input (Inl No_Strategy) r Nothing;
xml_cert_problem (Completion_Proof e r vc) =
  Xml ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', 'I', 'n', 'p', 'u',
        't']
    [] [xml_rules ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] e,
         xml_rules ['t', 'r', 's'] r]
    Nothing;
xml_cert_problem (Equational_Proof e eq vd) =
  Xml ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'R', 'e', 'a', 's',
        'o', 'n', 'i', 'n', 'g', 'I', 'n', 'p', 'u', 't']
    [] [xml_rules ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] e,
         Xml ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n'] []
           [xml_term (fst eq), xml_term (snd eq)] Nothing]
    Nothing;
xml_cert_problem (Equational_Disproof e eq ve) =
  Xml ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', 'R', 'e', 'a', 's',
        'o', 'n', 'i', 'n', 'g', 'I', 'n', 'p', 'u', 't']
    [] [xml_rules ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] e,
         Xml ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n'] []
           [xml_term (fst eq), xml_term (snd eq)] Nothing]
    Nothing;
xml_cert_problem (Complexity_Proof q r s cm cc vf) =
  Xml ['c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', 'I', 'n', 'p', 'u',
        't']
    [] [xml_trs_input (Inl q) r s, xml_complexity_measure cm,
         xml_complexity_class cc]
    Nothing;

xmldoc2cert_problem ::
  Xmldoc -> Parse_result [Char] (Cert_problem [Char] [Nat] [Char]);
xmldoc2cert_problem (Xmldoc header xml) =
  debug ['0']
    ['p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'x', 'm', 'l', ' ', 't', 'o', ' ',
      'c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 'p',
      'r', 'o', 'b', 'l', 'e', 'm']
    (xml2cert_problem (text ['n', 'a', 'm', 'e']) plain_name xml);

parse_cert_problem ::
  [Char] -> Parse_result [Char] (Cert_problem [Char] [Nat] [Char]);
parse_cert_problem s =
  (case doc_of_string s of {
    Inl a -> errora a;
    Inr a -> xmldoc2cert_problem a;
  });

newtype (Linorder a) => Dpp a b = Dpp
  (Bool,
    (Bool,
      ([(Term a b, Term a b)],
        ([(Term a b, Term a b)],
          ([Term a b],
            (Bool,
              (Bool,
                ([(Term a b, Term a b)],
                  ([(Term a b, Term a b)],
                    (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        Bool)))))))))));

extract_input_proof :: [Char] -> Maybe ([Char], ([Char], [Char]));
extract_input_proof s =
  (case search_len
          (filter (\ c -> not (membera [' ', newline, '\t', '\r'] c)) s)
          ['<', 'i', 'n', 'p', 'u', 't', '>']
          (size_list ['<', 'i', 'n', 'p', 'u', 't', '>'])
    of {
    Nothing -> Nothing;
    Just (_, minput) ->
      (case search_len minput ['<', '/', 'i', 'n', 'p', 'u', 't', '>']
              (size_list ['<', '/', 'i', 'n', 'p', 'u', 't', '>'])
        of {
        Nothing -> Nothing;
        Just (the_input, sa) ->
          (case search_len sa ['<', 'p', 'r', 'o', 'o', 'f', '>']
                  (size_list ['<', 'p', 'r', 'o', 'o', 'f', '>'])
            of {
            Nothing -> Nothing;
            Just (version, prf) -> Just (the_input, (version, prf));
          });
      });
  });

ostrategy_to_Q ::
  forall a b.
    (Eq a, Eq b) => Ostrategy a b -> [(Term a b, Term a b)] -> [Term a b];
ostrategy_to_Q Outermost r = remdups (map fst r);
ostrategy_to_Q (Outermost_Q q) uu = q;

subst_replace_impl ::
  forall a b. (Eq a) => [(a, Term b a)] -> a -> Term b a -> [(a, Term b a)];
subst_replace_impl sigma x t = (x, t) : filter (\ (y, _) -> not (y == x)) sigma;

subst_compose_impla ::
  forall a b.
    (Eq a, Eq b) => [(a, Term b a)] -> (a -> Term b a) -> [(a, Term b a)];
subst_compose_impla sigma rho =
  map (\ (x, s) -> (x, subst_apply_term s rho)) (mk_subst_domain sigma);

rtrancl_list_impl :: forall a. (Eq a) => [(a, a)] -> [a] -> [a];
rtrancl_list_impl =
  rtrancl_impl
    (\ r asa ->
      remdups
        (map_filter
          (\ x ->
            (if let {
                  (a, _) = x;
                } in membera asa a
              then Just (snd x) else Nothing))
          r))
    (\ xs ys -> filter (\ x -> not (membera ys x)) xs ++ ys)
    (\ x xs -> membera xs x) [];

w_impl :: forall a b. (Eq a) => [(a, Term b a)] -> Term b a -> [a];
w_impl d =
  let {
    filt = filter (\ (x, y) -> not (x == y));
    xvs = concatMap (\ (x, t) -> map (\ a -> (x, a)) (vars_term_lista t)) d;
    rel = filt xvs;
    rtran = rtrancl_list_impl rel;
  } in (\ t -> rtran (vars_term_lista t));

var_renaming_impl :: forall a b. (Eq a, Eq b) => [(a, Term b a)] -> Bool;
var_renaming_impl sigma =
  let {
    sigmaa = map snd (mk_subst_domain sigma);
  } in all is_Var sigmaa && distinct sigmaa;

vars_pat_term_impl ::
  forall a b.
    (Eq a, Eq b) => (Term a b, ([(b, Term a b)], [(b, Term a b)])) -> [b];
vars_pat_term_impl p =
  let {
    (s, (sigma, mu)) = p;
  } in remdups
         (vars_term_lista s ++ vars_subst_impl sigma ++ vars_subst_impl mu);

pat_dom_renaming_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Term a b, ([(b, Term a b)], [(b, Term a b)])) ->
                 [(b, Term a b)] -> Bool;
pat_dom_renaming_impl p rho =
  let {
    rhoa = mk_subst_domain rho;
    xs = map Var (vars_pat_term_impl p);
  } in var_renaming_impl rho &&
         all (membera (pat_dv_impl p)) (map fst rhoa) &&
           all (\ t -> not (membera xs t)) (map snd rhoa);

inverse_var_renaming_impl ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> [(a, Term b a)];
inverse_var_renaming_impl sigma =
  let {
    a = mk_subst_domain sigma;
  } in map (\ (x, y) -> (the_Var y, Var x)) a;

check_pat_eqv_prf ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Pat_eqv_prf a b ->
                    (Term a b, ([(b, Term a b)], [(b, Term a b)])) ->
                      Sum ([Char] -> [Char])
                        (Term a b, ([(b, Term a b)], [(b, Term a b)]));
check_pat_eqv_prf (Pat_Irrelevant sigmaa mua) (t, (sigma, mu)) =
  let {
    w = w_impl (mk_subst_domain sigma) t;
    sig = mk_subst sigma;
    siga = mk_subst sigmaa;
    mub = mk_subst mu;
    muba = mk_subst mua;
  } in bindb (catcha
               (catcha
                 (forallm
                   (\ x ->
                     bindb (check (equal_term (sig x) (siga x))
                             (x, (sig x,
                                   (siga x,
                                     ['p', 'u', 'm', 'p', 'i', 'n', 'g']))))
                       (\ _ ->
                         check (equal_term (mub x) (muba x))
                           (x, (mub x,
                                 (muba x,
                                   ['c', 'l', 'o', 's', 'i', 'n', 'g'])))))
                   w)
                 (\ x -> Inl (snd x)))
               (\ x ->
                 Inl (let {
                        (xa, (tb, (ta, sub))) = x;
                      } in shows_string
                             ['e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'e',
                               'q', 'u', 'i', 'v', 'a', 'l', 'e', 'n', 'c', 'e',
                               ' ', '(', 'i', 'r', 'r', 'e', 'l', 'e', 'v', 'a',
                               'n', 't', ')', ':', ' ', 'f', 'o', 'r', ' ', 'v',
                               'a', 'r', 'i', 'a', 'b', 'l', 'e', ' '] .
                             shows_prec Zero_nat xa .
                               shows_string
                                 [' ', 'o', 'b', 't', 'a', 'i', 'n', ' ', 'd',
                                   'i', 'f', 'f', 'e', 'r', 'e', 'n', 't', ' ',
                                   'v', 'a', 'l', 'u', 'e', 's', ' ', 'f', 'o',
                                   'r', ' '] .
                                 shows_prec_list Zero_nat sub .
                                   shows_string
                                     [' ', 's', 'u', 'b', 's', 't', 'i', 't',
                                       'u', 't', 'i', 'o', 'n', ':', ' '] .
                                     shows_prec_term Zero_nat tb .
                                       shows_string [' ', '!', '=', ' '] .
 shows_prec_term Zero_nat ta)))
         (\ _ -> Inr (t, (sigmaa, mua)));
check_pat_eqv_prf (Pat_Simplify mu_1 mu_2) (t, (sigma, mu)) =
  bindb (check (subst_eq mu (subst_compose_impl mu_1 mu_2))
          (shows_prec_list Zero_nat
            ['m', 'u', ' ', '!', '=', ' ', 'm', 'u', '1', ' ', 'm', 'u', '2']))
    (\ _ ->
      bindb (check (commutes_impl mu_1 sigma)
              (shows_prec_list Zero_nat
                ['s', 'i', 'g', 'm', 'a', ' ', 'a', 'n', 'd', ' ', 'm', 'u',
                  '1', ' ', 'd', 'o', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm',
                  'm', 'u', 't', 'e']))
        (\ _ -> Inr (subst_apply_term t (mk_subst mu_1), (sigma, mu_2))));
check_pat_eqv_prf (Pat_Dom_Renaming rho) (t, (sigma, mu)) =
  bindb (check (pat_dom_renaming_impl (t, (sigma, mu)) rho)
          (shows_prec_list Zero_nat
            ['r', 'h', 'o', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
              'd', 'o', 'm', 'a', 'i', 'n', ' ', 'r', 'e', 'n', 'a', 'm', 'i',
              'n', 'g', ' ', 'f', 'o', 'r', ' ', 'p']))
    (\ _ ->
      let {
        i_rho = inverse_var_renaming_impl rho;
        sigmaa =
          mk_subst_case
            (map (the_Var . mk_subst rho) (map fst (mk_subst_domain sigma)))
            (\ x ->
              subst_apply_term
                (subst_apply_term (subst_apply_term (Var x) (mk_subst i_rho))
                  (mk_subst sigma))
                (mk_subst rho))
            [];
        mua = mk_subst_case
                (map (the_Var . mk_subst rho) (map fst (mk_subst_domain mu)))
                (\ x ->
                  subst_apply_term (subst_apply_term (Var x) (mk_subst i_rho))
                    (mk_subst mu))
                i_rho;
      } in Inr (subst_apply_term t (mk_subst rho), (sigmaa, mua)));

check_pat_rule_prf ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      Pat_rule_prf a b ->
                        Sum ([Char] -> [Char])
                          ((Term a b, ([(b, Term a b)], [(b, Term a b)])),
                            ((Term a b, ([(b, Term a b)], [(b, Term a b)])),
                              Bool));
check_pat_rule_prf ra p (Pat_OrigRule (l, r) isPair) =
  (if isPair
    then bindb (check (membera p (l, r))
                 (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '-', '>', ' '] (l, r) .
                   shows_prec_list Zero_nat
                     [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p',
                       'a', 'i', 'r']))
           (\ _ -> Inr ((l, ([], [])), ((r, ([], [])), isPair)))
    else bindb (check (membera ra (l, r))
                 (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '-', '>', ' '] (l, r) .
                   shows_prec_list Zero_nat
                     [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'r',
                       'u', 'l', 'e']))
           (\ _ -> Inr ((l, ([], [])), ((r, ([], [])), isPair))));
check_pat_rule_prf r p (Pat_InitPump pat sigma theta) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sig, tau)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (siga, taua)) = aa;
             } in (\ bb ->
                    catcha
                      (bindb
                        (check (null (sig ++ tau ++ siga ++ taua))
                          (shows_prec_list Zero_nat
                            ['s', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i',
                              'o', 'n', 's', ' ', 'm', 'u', 's', 't', ' ', 'b',
                              'e', ' ', 'e', 'm', 'p', 't', 'y']))
                        (\ _ ->
                          bindb (check
                                  (equal_term
                                    (subst_apply_term s (mk_subst theta))
                                    (subst_apply_term t (mk_subst sigma)))
                                  (shows_prec_list Zero_nat
                                    ['s', ' ', 't', 'h', 'e', 't', 'a', ' ',
                                      '!', '=', ' ', 't', ' ', 's', 'i', 'g',
                                      'm', 'a']))
                            (\ _ ->
                              bindb (check (commutes_impl theta sigma)
                                      (shows_prec_list Zero_nat
['s', 'i', 'g', 'm', 'a', ' ', 'a', 'n', 'd', ' ', 't', 'h', 'e', 't', 'a', ' ',
  'd', 'o', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm', 'm', 'u', 't', 'e']))
                                (\ _ ->
                                  Inr ((s, (sigma, [])),
((t, (theta, [])), bb))))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['i', 'n', 'i', 't', 'i', 'a', 'l', ' ', 'p',
                                  'u', 'm', 'p', 'i', 'n', 'g'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sig, tau)),
((t, (siga, taua)), bb)) .
                                      x)))
               ba)
        b);
check_pat_rule_prf r pa (Pat_InitPumpCtxt pat sigma p z) =
  bindb (check_pat_rule_prf r pa pat)
    (\ (a, b) ->
      let {
        (s, (sig, tau)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (siga, taua)) = aa;
             } in (\ bb ->
                    catcha
                      (bindb
                        (check (not bb)
                          (shows_prec_list Zero_nat
                            ['p', 'a', 'i', 'r', 's', ' ', 'n', 'o', 't', ' ',
                              'a', 'l', 'l', 'o', 'w', 'e', 'd', ' ', 'i', 'n',
                              ' ', 'i', 'n', 'i', 't', ' ', 'p', 'u', 'm', 'p',
                              ' ', 'c', 't', 'x', 't']))
                        (\ _ ->
                          bindb (check (null (sig ++ tau ++ siga ++ taua))
                                  (shows_prec_list Zero_nat
                                    ['s', 'u', 'b', 's', 't', 'i', 't', 'u',
                                      't', 'i', 'o', 'n', 's', ' ', 'm', 'u',
                                      's', 't', ' ', 'b', 'e', ' ', 'e', 'm',
                                      'p', 't', 'y']))
                            (\ _ ->
                              bindb (check (in_poss p t)
                                      (shows_prec_list Zero_nat
['p', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'v', 'a', 'l', 'i', 'd',
  ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n']))
                                (\ _ ->
                                  bindb (check
  (equal_term s (subst_apply_term (subt_at t p) (mk_subst sigma)))
  (shows_prec_list Zero_nat
    ['s', ' ', '!', '=', ' ', 't', ' ', '|', '_', ' ', 'p', ' ', 's', 'i', 'g',
      'm', 'a']))
                                    (\ _ ->
                                      bindb
(check
  (not (membera
         (vars_term_lista s ++ vars_term_lista t ++ vars_subst_impl sigma) z))
  (shows_prec_list Zero_nat
    ['z', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'r', 'e', 's', 'h']))
(\ _ ->
  let {
    tz = ctxt_apply (ctxt_of_pos_term p t) (Var z);
  } in Inr ((s, (sigma, [])),
             ((tz, ((z, tz) : sigma, [(z, subt_at t p)])), bb))))))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['i', 'n', 'i', 't', 'i', 'a', 'l', ' ', 'p',
                                  'u', 'm', 'p', 'i', 'n', 'g', ' ', '(', 'w',
                                  'i', 't', 'h', ' ', 'c', 't', 'x', 't', ')',
                                  ' '] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sig, tau)),
((t, (siga, taua)), bb)) .
                                      x)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Equiv pat left eqv) =
  bindb (check_pat_rule_prf r p pat)
    (\ (pleft, (pright, b)) ->
      catcha
        (bindb (check_pat_eqv_prf eqv (if left then pleft else pright))
          (\ pnew ->
            Inr (if left then (pnew, (pright, b)) else (pleft, (pnew, b)))))
        (\ x ->
          Inl (shows_prec_list Zero_nat
                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                   ' '] .
                shows_prec_list Zero_nat
                  ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'e', 'q', 'u', 'i',
                    'v', 'a', 'l', 'e', 'n', 'c', 'e'] .
                  shows_prec_list Zero_nat
                    [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e', 'r', 'i', 'v',
                      'i', 'n', 'g', ' ', 'c', 'o', 'r', 'r', 'e', 'c', 't',
                      ' ', 'p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'r', 'u',
                      'l', 'e', ' '] .
                    shows_nl . show_pat_rule (pleft, (pright, b)) . x)));
check_pat_rule_prf r pa (Pat_Narrow pat1 pat2 p) =
  bindb (check_pat_rule_prf r pa pat1)
    (\ (a, b) ->
      let {
        (s, (sigma, mu)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sig, mua)) = aa;
             } in (\ b1 ->
                    bindb (check_pat_rule_prf r pa pat2)
                      (\ (ab, bb) ->
                        let {
                          (u, (sig1, mu1)) = ab;
                        } in (\ (ac, bc) ->
                               let {
                                 (v, (sig2, mu2)) = ac;
                               } in (\ b2 ->
                                      catcha
(bindb
  (check
    (subst_eq sig sigma &&
      subst_eq sig1 sigma &&
        subst_eq sig2 sigma &&
          subst_eq mua mu && subst_eq mu1 mu && subst_eq mu2 mu)
    (shows_prec_list Zero_nat
      ['s', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i', 'o', 'n', 's', ' ',
        'a', 'r', 'e', ' ', 'n', 'o', 't', ' ', 'i', 'd', 'e', 'n', 't', 'i',
        'c', 'a', 'l']))
  (\ _ ->
    bindb (check (in_poss p t)
            (shows_prec_list Zero_nat
              ['p', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'v', 'a',
                'l', 'i', 'd', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n']))
      (\ _ ->
        bindb (check (equal_term (subt_at t p) u)
                (shows_prec_list Zero_nat
                  ['t', ' ', '|', '_', ' ', 'p', ' ', '!', '=', ' ', 'u']))
          (\ _ ->
            bindb (check (if b2 then equal_pos p Empty else True)
                    (shows_prec_list Zero_nat
                      ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'a', ' ',
                        'P', ' ', 's', 't', 'e', 'p', ',', ' ', 's', 'o', ' ',
                        'p', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'e',
                        'p', 's', 'i', 'l', 'o', 'n']))
              (\ _ ->
                Inr ((s, (sigma, mu)),
                      ((ctxt_apply (ctxt_of_pos_term p t) v, (sigma, mu)),
                        b1 || b2)))))))
(\ x ->
  Inl (shows_prec_list Zero_nat
         ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' '] .
        shows_prec_list Zero_nat
          ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'n', 'a', 'r', 'r', 'o', 'w',
            'i', 'n', 'g'] .
          shows_prec_list Zero_nat
            [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e', 'r', 'i', 'v', 'i',
              'n', 'g', ' ', 'c', 'o', 'r', 'r', 'e', 'c', 't', ' ', 'p', 'a',
              't', 't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e', 's', ' '] .
            shows_nl .
              show_pat_rule ((s, (sigma, mu)), ((t, (sig, mua)), b1)) .
                shows_nl .
                  shows_prec_list Zero_nat [' ', 'a', 'n', 'd', ' '] .
                    shows_nl .
                      show_pat_rule ((u, (sig1, mu1)), ((v, (sig2, mu2)), b2)) .
                        x)))
                                 bc)
                          bb))
               ba)
        b);
check_pat_rule_prf r p (Pat_Inst pat rho Pat_Base) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    catcha
                      (let {
                         xs = map fst
                                (mk_subst_domain sigma_s ++
                                  mk_subst_domain mu_s ++
                                    mk_subst_domain sigma_t ++
                                      mk_subst_domain mu_t);
                       } in bindb (catcha
                                    (forallm
                                      (\ x ->
check (not (membera xs x))
  (shows_prec_list Zero_nat
    ['d', 'o', 'm', 'a', 'i', 'n', 's', ' ', 'n', 'o', 't', ' ', 'd', 'i', 's',
      'j', 'o', 'i', 'n', 't']))
                                      (vars_subst_impl rho))
                                    (\ x -> Inl (snd x)))
                              (\ _ ->
                                let {
                                  rhoa = mk_subst rho;
                                } in Inr ((subst_apply_term s rhoa,
    (subst_compose_impla sigma_s rhoa, subst_compose_impla mu_s rhoa)),
   ((subst_apply_term t rhoa,
      (subst_compose_impla sigma_t rhoa, subst_compose_impla mu_t rhoa)),
     bb))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'i',
                                  'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                                  'i', 'o', 'n', ' ', '(', 'b', 'a', 's', 'e',
                                  ')'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb)) .
                                      x)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Inst pat rho Pat_Pump) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    catcha
                      (bindb
                        (check (commutes_impl rho sigma_s)
                          (shows_prec_list Zero_nat
                            ['r', 'h', 'o', ' ', 'd', 'o', 'e', 's', ' ', 'n',
                              'o', 't', ' ', 'c', 'o', 'm', 'm', 'u', 't', 'e',
                              ' ', 'w', 'i', 't', 'h', ' ', 's', 'i', 'g', 'm',
                              'a', '_', 's']))
                        (\ _ ->
                          bindb (check (commutes_impl rho mu_s)
                                  (shows_prec_list Zero_nat
                                    ['r', 'h', 'o', ' ', 'd', 'o', 'e', 's',
                                      ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm',
                                      'm', 'u', 't', 'e', ' ', 'w', 'i', 't',
                                      'h', ' ', 'm', 'u', '_', 's']))
                            (\ _ ->
                              bindb (check (commutes_impl rho sigma_t)
                                      (shows_prec_list Zero_nat
['r', 'h', 'o', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm',
  'm', 'u', 't', 'e', ' ', 'w', 'i', 't', 'h', ' ', 's', 'i', 'g', 'm', 'a',
  '_', 't']))
                                (\ _ ->
                                  bindb (check (commutes_impl rho mu_t)
  (shows_prec_list Zero_nat
    ['r', 'h', 'o', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o',
      'm', 'm', 'u', 't', 'e', ' ', 'w', 'i', 't', 'h', ' ', 'm', 'u', '_',
      't']))
                                    (\ _ ->
                                      Inr
((s, (subst_compose_impl sigma_s rho, mu_s)),
  ((t, (subst_compose_impl sigma_t rho, mu_t)), bb)))))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'i',
                                  'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                                  'i', 'o', 'n', ' ', '(', 'p', 'u', 'm', 'p',
                                  'i', 'n', 'g', ')'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb)) .
                                      x)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Inst pat rho Pat_Close) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    Inr ((s, (sigma_s, subst_compose_impl mu_s rho)),
                          ((t, (sigma_t, subst_compose_impl mu_t rho)), bb)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Rewr pat rewr Pat_Base uu) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    catcha
                      (let {
                         (ta, rseq) = rewr;
                         tb = lasta (ta : map (\ (_, (_, tb)) -> tb) rseq);
                       } in bindb (check (equal_term t ta)
                                    (shows_prec_list Zero_nat
                                      ['t', 'e', 'r', 'm', 's', ' ', 't', ' ',
'd', 'o', ' ', 'n', 'o', 't', ' ', 'm', 'a', 't', 'c', 'h']))
                              (\ _ ->
                                bindb (check_rsteps r rseq ta tb)
                                  (\ _ ->
                                    Inr ((s, (sigma_s, mu_s)),
  ((tb, (sigma_t, mu_t)), bb)))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'r',
                                  'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ',
                                  '(', 'b', 'a', 's', 'e', ')'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb)) .
                                      x)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Rewr pat rewr Pat_Pump x) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    catcha
                      (let {
                         (ta, rseq) = rewr;
                         tb = lasta (ta : map (\ (_, (_, tb)) -> tb) rseq);
                       } in bindb (check (equal_term (mk_subst sigma_t x) ta)
                                    (shows_prec_list Zero_nat
                                      ['s', 'i', 'g', 'm', 'a', '_', 't', ' ',
'x', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'm', 'a', 't', 'c', 'h',
' ', 's', 't', 'a', 'r', 't', 'i', 'n', 'g', ' ', 't', 'e', 'r', 'm']))
                              (\ _ ->
                                bindb (check_rsteps r rseq ta tb)
                                  (\ _ ->
                                    Inr ((s, (sigma_s, mu_s)),
  ((t, (subst_replace_impl sigma_t x tb, mu_t)), bb)))))
                      (\ xa ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'r',
                                  'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ',
                                  '(', 'p', 'u', 'm', 'p', 'i', 'n', 'g', ')'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb)) .
                                      xa)))
               ba)
        b);
check_pat_rule_prf r p (Pat_Rewr pat rewr Pat_Close x) =
  bindb (check_pat_rule_prf r p pat)
    (\ (a, b) ->
      let {
        (s, (sigma_s, mu_s)) = a;
      } in (\ (aa, ba) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ bb ->
                    catcha
                      (let {
                         (ta, rseq) = rewr;
                         tb = lasta (ta : map (\ (_, (_, tb)) -> tb) rseq);
                       } in bindb (check (equal_term (mk_subst mu_t x) ta)
                                    (shows_prec_list Zero_nat
                                      ['s', 'i', 'g', 'm', 'a', '_', 't', ' ',
'x', ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'm', 'a', 't', 'c', 'h',
' ', 's', 't', 'a', 'r', 't', 'i', 'n', 'g', ' ', 't', 'e', 'r', 'm']))
                              (\ _ ->
                                bindb (check_rsteps r rseq ta tb)
                                  (\ _ ->
                                    Inr ((s, (sigma_s, mu_s)),
  ((t, (sigma_t, subst_replace_impl mu_t x tb)), bb)))))
                      (\ xa ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['p', 'a', 't', 't', 'e', 'r', 'n', ' ', 'r',
                                  'e', 'w', 'r', 'i', 't', 'i', 'n', 'g', ' ',
                                  '(', 'c', 'l', 'o', 's', 'i', 'n', 'g', ')'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigma_s, mu_s)),
((t, (sigma_t, mu_t)), bb)) .
                                      xa)))
               ba)
        b);

check_non_loop_prf ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      Non_loop_prf a b -> Sum ([Char] -> [Char]) ();
check_non_loop_prf r pa (Non_Loop_Prf pat sigma mu m b p) =
  bindb (check_pat_rule_prf r pa pat)
    (\ (a, c) ->
      let {
        (s, (sigmaa, mua)) = a;
      } in (\ (aa, ca) ->
             let {
               (t, (sigma_t, mu_t)) = aa;
             } in (\ is_pair ->
                    catcha
                      (bindb
                        (check (if is_pair then equal_pos p Empty else True)
                          (shows_prec_list Zero_nat
                            ['p', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ',
                              'e', 'm', 'p', 't', 'y', ',', ' ', 's', 'i', 'n',
                              'c', 'e', ' ', 'p', 'a', 'i', 'r', 's', ' ', 'a',
                              'r', 'e', ' ', 'c', 'o', 'n', 't', 'a', 'i', 'n',
                              'e', 'd']))
                        (\ _ ->
                          bindb (check (commutes_impl sigmaa sigma)
                                  (shows_prec_list Zero_nat
                                    ['s', 'i', 'g', 'm', 'a', ' ', 'a', 'n',
                                      'd', ' ', 's', 'i', 'g', 'm', 'a', '\'',
                                      ' ', 'd', 'o', ' ', 'n', 'o', 't', ' ',
                                      'c', 'o', 'm', 'm', 'u', 't', 'e']))
                            (\ _ ->
                              bindb (check (commutes_impl mua sigma)
                                      (shows_prec_list Zero_nat
['m', 'u', ' ', 'a', 'n', 'd', ' ', 's', 'i', 'g', 'm', 'a', '\'', ' ', 'd',
  'o', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm', 'm', 'u', 't', 'e']))
                                (\ _ ->
                                  bindb (check
  (subst_eq sigma_t (subst_compose_impl (subst_power_impl sigmaa m) sigma))
  (shows_prec_list Zero_nat
    ['s', 'i', 'g', 'm', 'a', '_', 't', ' ', '!', '=', ' ', 's', 'i', 'g', 'm',
      'a', '^', 'm', ' ', 's', 'i', 'g', 'm', 'a', '\'', ' ']))
                                    (\ _ ->
                                      bindb
(check (subst_eq mu_t (subst_compose_impl mua mu))
  (shows_prec_list Zero_nat
    ['m', 'u', '_', 't', ' ', '!', '=', ' ', 'm', 'u', ' ', 'm', 'u', '\'',
      ' ']))
(\ _ ->
  bindb (check (in_poss p t)
          (shows_prec_list Zero_nat
            ['p', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'o',
              's', 'i', 't', 'i', 'o', 'n', ' ', 'i', 'n', ' ', 't']))
    (\ _ ->
      check (equal_term
              (subst_apply_term s (mk_subst (subst_power_impl sigmaa b)))
              (subt_at t p))
        (shows_prec_list Zero_nat
          ['s', ' ', 's', 'i', 'g', 'm', 'a', '^', 'b', ' ', '!', '=', ' ', 't',
            ' ', '|', '_', ' ', 'p']))))))))
                      (\ x ->
                        Inl (shows_prec_list Zero_nat
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' '] .
                              shows_prec_list Zero_nat
                                ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                  'o', 'n', ' ', 'c', 'o', 'n', 'd', 'i', 't',
                                  'i', 'o', 'n', ' ', 'o', 'f', ' ', 'n', 'o',
                                  'n', '-', 'l', 'o', 'o', 'p', ' ', 't', 'h',
                                  'e', 'o', 'r', 'e', 'm'] .
                                shows_prec_list Zero_nat
                                  [' ', 'a', 'f', 't', 'e', 'r', ' ', 'd', 'e',
                                    'r', 'i', 'v', 'i', 'n', 'g', ' ', 'c', 'o',
                                    'r', 'r', 'e', 'c', 't', ' ', 'p', 'a', 't',
                                    't', 'e', 'r', 'n', ' ', 'r', 'u', 'l', 'e',
                                    ' '] .
                                  shows_nl .
                                    show_pat_rule
                                      ((s, (sigmaa, mua)),
((t, (sigma_t, mu_t)), is_pair)) .
                                      x)))
               ca)
        c);

simplify_mp_impl ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Term b a) ->
                 [a] ->
                   [(Term b a, Term b a)] ->
                     [(Term b a, Term b a)] ->
                       Maybe ([(Term b a, Term b a)], Nat);
simplify_mp_impl mu v_incr mp1 solved =
  (case mp1 of {
    [] -> Just (solved, Zero_nat);
    (Var v, Var x) : mp ->
      simplify_mp_impl mu v_incr mp ((Var v, Var x) : solved);
    (Var v, Fun _ _) : mp ->
      bind (guard (membera v_incr v))
        (\ _ ->
          let {
            m = map (\ (s, a) -> (subst_apply_term s mu, a));
          } in bind (simplify_mp_impl mu v_incr (m mp1) (m solved))
                 (\ (smp, i) -> Just (smp, Suc i)));
    (Fun fa lista, Var x) : mp ->
      simplify_mp_impl mu v_incr mp ((Fun fa lista, Var x) : solved);
    (Fun fa lista, Fun f ls) : mp ->
      bind (guard (f == fa))
        (\ _ ->
          bind (zip_option lista ls)
            (\ pairs -> simplify_mp_impl mu v_incr (pairs ++ mp) solved));
  });

simplify_mp ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Term b a)] ->
                 [a] ->
                   [(Term b a, Term b a)] ->
                     [(Term b a, Term b a)] ->
                       Maybe ([(Term b a, Term b a)], Nat);
simplify_mp mu v = let {
                     mua = mk_subst mu;
                   } in simplify_mp_impl mua v;

v_incr_impl :: forall a b. (Eq a) => [(a, Term b a)] -> [a];
v_incr_impl d =
  let {
    (ixs, uxs) = partition (\ (_, t) -> not (is_Var t)) d;
    rel = map (\ (x, y) -> (the_Var y, x)) uxs;
    a = map fst ixs;
  } in rtrancl_list_impl rel a;

ident_prob_of_smp ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> [(Term a b, Term a b)];
ident_prob_of_smp [] = [];
ident_prob_of_smp ((t, l) : other) =
  map_filter
    (\ x ->
      (if let {
            (_, s) = x;
          } in equal_term s l
        then Just (let {
                     (s, _) = x;
                   } in (t, s))
        else Nothing))
    other ++
    ident_prob_of_smp other;

gmatch_to_idents_impl ::
  forall a b c d.
    (Eq c,
      Eq d) => (a -> [b] -> Maybe ([(Term c d, Term c d)], Nat)) ->
                 ((Term c d, Term c d) -> Maybe Nat) -> a -> Maybe Nat;
gmatch_to_idents_impl simp ident mp =
  bind (simp mp [])
    (\ (smp, i) ->
      let {
        ids = map ident (ident_prob_of_smp smp);
      } in (if membera ids Nothing then Nothing
             else Just (plus_nat (max_list (map the ids)) i)));

gmatch_decision ::
  forall a b c d e.
    (Eq d,
      Eq e) => a -> (b -> [c] -> Maybe ([(Term d e, Term d e)], Nat)) ->
                      ((Term d e, Term d e) -> Maybe Nat) -> b -> Bool;
gmatch_decision mu simp ident mp =
  not (is_none (gmatch_to_idents_impl simp ident mp));

match_decision ::
  forall a b c d e.
    (Eq d,
      Eq e) => a -> ([b] -> [c] -> Maybe ([(Term d e, Term d e)], Nat)) ->
                      ((Term d e, Term d e) -> Maybe Nat) -> b -> Bool;
match_decision mu simp ident mp = gmatch_decision mu simp ident [mp];

match_prob_of_rp_impl ::
  forall a b c d e.
    (Eq b,
      Eq c) => (a -> Term b c) ->
                 (Term b c -> [a]) ->
                   (Term b c, Term d e) -> [(Term b c, Term d e)];
match_prob_of_rp_impl mu w (t, Var x) = [(t, Var x)];
match_prob_of_rp_impl mu w (t, Fun v va) =
  let {
    sterms = remdups (t : map mu (w t));
    uterms = concatMap (filter (\ ta -> not (is_Var ta)) . supteq_list) sterms;
  } in map (\ u -> (u, Fun v va)) (remdups uterms);

redex_decision ::
  forall a b c d e f g h.
    (Eq b, Eq c, Eq g,
      Eq h) => (a -> Term b c) ->
                 ([(Term b c, Term d e)] ->
                   [f] -> Maybe ([(Term g h, Term g h)], Nat)) ->
                   ((Term g h, Term g h) -> Maybe Nat) ->
                     (Term b c -> [a]) -> (Term b c, Term d e) -> Bool;
redex_decision mu simp ident w rp =
  let {
    mps = match_prob_of_rp_impl mu w rp;
    solver = match_decision mu simp ident;
  } in any solver mps;

ident_solve_impla ::
  forall a b.
    (Eq a,
      Eq b) => (a -> Term b a) ->
                 [(Term b a, Term b a)] ->
                   (Term b a, (Term b a, Nat)) -> Maybe Nat;
ident_solve_impla mu cps stn =
  let {
    cp = remdups (conflicts mu stn);
  } in (if any (\ (u, (_, _)) -> not (is_Var u)) cp then Nothing
         else (if any (\ (u, (v, _)) -> membera cps (u, v)) cp then Nothing
                else bind (mapMb
                            (\ (u, (v, m)) ->
                              ident_solve_impla mu ((u, v) : cps)
                                (subst_apply_term u mu, (v, Suc m)))
                            cp)
                       (\ is -> Just (max_list (map Suc is)))));

ident_solve_impl ::
  forall a b.
    (Eq a, Eq b) => (a -> Term b a) -> (Term b a, Term b a) -> Maybe Nat;
ident_solve_impl = (\ mu (s, t) -> ident_solve_impla mu [] (s, (t, Zero_nat)));

redex_rps_decision ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Term b a)] ->
                 [(Term b a, Term b a)] -> Sum (Term b a, Term b a) ();
redex_rps_decision mu =
  let {
    mua = mk_subst mu;
    d = mk_subst_domain mu;
    v = v_incr_impl d;
    simp = simplify_mp mu v;
    ident = ident_solve_impl mua;
    w = w_impl d;
    main = redex_decision mua simp ident w;
  } in (\ xs ->
         catcha (forallm (\ tl -> check (not (main tl)) tl) xs)
           (\ x -> Inl (snd x)));

check_NF_iteration ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Term b a)] ->
                 [Term b a] -> Term b a -> Sum (Term b a, Term b a) ();
check_NF_iteration mu =
  let {
    dec = redex_rps_decision mu;
  } in (\ q t -> dec (map (\ a -> (t, a)) q));

check_qrstep_subst ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Sum (Term a b, Term a b) ()) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        Pos ->
                          (Term a b, Term a b) ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qrstep_subst cni nfs =
  let {
    main =
      check_prop_rstep nfs
        (\ t ->
          catcha (cni t)
            (\ _ ->
              Inl (shows_prec_term Zero_nat t .
                    shows_prec_list Zero_nat
                      [' ', 'm', 'u', ' ', '^', '^', ' ', 'i', ' ', 'i', 's',
                        ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ', 'Q', '-', 'n',
                        'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', ' ',
                        'f', 'o', 'r', ' ', 'a', 'l', 'l', ' ', 'i'])));
  } in (\ r p ra s t ->
         bindb (check (not (is_Var (fst ra)))
                 (shows_prec_list Zero_nat
                   ['l', 'o', 'o', 'p', ' ', 'c', 'h', 'e', 'c', 'k', ' ', 'r',
                     'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'l', 'h', 's', 's',
                     ' ', 't', 'o', ' ', 'b', 'e', ' ', 'n', 'o', 'n', '-', 'v',
                     'a', 'r', 'i', 'a', 'b', 'l', 'e']))
           (\ _ -> main r p ra s t));

check_rqrstep_subst ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Sum (Term a b, Term a b) ()) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        (Term a b, Term a b) ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rqrstep_subst cni nfs = (\ r -> check_qrstep_subst cni nfs r Empty);

check_qsteps_subst ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Sum (Term a b, Term a b) ()) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                            Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qsteps_subst cni nfs p r [] s u =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
         ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e'] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_nl .
            shows_string
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'] .
              shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_qsteps_subst cni nfs p ra ((uu, (r, (True, t))) : prts) s u =
  bindb (check_rqrstep_subst cni nfs p r s t)
    (\ _ -> check_qsteps_subst cni nfs p ra prts t u);
check_qsteps_subst cni nfs pa ra ((p, (r, (False, t))) : prts) s u =
  bindb (check_qrstep_subst cni nfs ra p r s t)
    (\ _ -> check_qsteps_subst cni nfs pa ra prts t u);

check_qrsteps_subst ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Term a b -> Sum (Term a b, Term a b) ()) ->
                    Bool ->
                      [(Term a b, Term a b)] ->
                        [(Pos, ((Term a b, Term a b), Term a b))] ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_qrsteps_subst cni nfs r prts s u =
  check_qsteps_subst cni nfs [] r
    (map (\ (p, (ra, t)) -> (p, (ra, (False, t)))) prts) s u;

check_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [Term a b] ->
                    Bool ->
                      (Term a b -> Bool) ->
                        Term a b ->
                          [(Pos, ((Term a b, Term a b), Term a b))] ->
                            [(b, Term a b)] ->
                              Ctxt a b ->
                                [(Term a b, Term a b)] ->
                                  Sum ([Char] -> [Char]) ();
check_loop q nfs nf s rseq sigma c r =
  bindb (check (not (null rseq))
          (shows_prec_list Zero_nat
            ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u', 'e',
              'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n',
              'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
    (\ _ ->
      (if null q
        then check_qrsteps (\ _ -> True) nfs r rseq s
               (ctxt_apply c (subst_apply_term s (mk_subst sigma)))
        else check_qrsteps_subst (check_NF_iteration sigma q) nfs r rseq s
               (ctxt_apply c (subst_apply_term s (mk_subst sigma)))));

check_dp_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [Term a b] ->
                    Bool ->
                      (Term a b -> Bool) ->
                        Term a b ->
                          [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                            [(b, Term a b)] ->
                              Ctxt a b ->
                                [(Term a b, Term a b)] ->
                                  [(Term a b, Term a b)] ->
                                    Sum ([Char] -> [Char]) ();
check_dp_loop q nfs nf s prseq sigma c p r =
  (if equal_ctxt c Hole
    then bindb (check (not (null prseq))
                 (shows_prec_list Zero_nat
                   ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u',
                     'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
                     ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
           (\ _ ->
             (if null q
               then check_qsteps nf nfs p r prseq s
                      (subst_apply_term s (mk_subst sigma))
               else check_qsteps_subst (check_NF_iteration sigma q) nfs p r
                      prseq s (subst_apply_term s (mk_subst sigma))))
    else check_loop q nfs nf s (map (\ (x, (y, (_, z))) -> (x, (y, z))) prseq)
           sigma c r);

check_instance ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_instance pa p =
  catcha
    (forallm
      (\ st ->
        check (any (instance_rule st) pa)
          (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] st .
            shows_prec_list Zero_nat
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', 'i', 'n',
                's', 't', 'a', 'n', 'c', 'e', ' ', 'o', 'f', ' ', 'a', 'n', 'y',
                ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'p', 'a', 'i',
                'r']))
      p)
    (\ x -> Inl (snd x));

switch_termination_proc ::
  forall a b.
    (Eq b,
      Showa b) => Dpp_ops_ext a b [Char] () ->
                    Join_info b -> a -> Sum ([Char] -> [Char]) a;
switch_termination_proc i joins_i dpp =
  let {
    p = pairsa i dpp;
    r = rulesd i dpp;
    q = qc i dpp;
  } in (case let {
               cp = critical_pairs_impl r r;
             } in bindb (catcha
                          (forallm
                            (\ (b, _) ->
                              check b
                                (shows_prec_list Zero_nat
                                  ['r', 'u', 'l', 'e', 's', ' ', 'a', 'r', 'e',
                                    ' ', 'n', 'o', 't', ' ', 'o', 'v', 'e', 'r',
                                    'l', 'a', 'y']))
                            cp)
                          (\ x -> Inl (snd x)))
                    (\ _ ->
                      bindb (check_wf_trs r)
                        (\ _ ->
                          bindb (check_critical_pairs r cp joins_i)
                            (\ _ ->
                              bindb (check (null (critical_pairs_impl p r))
                                      (shows_prec_list Zero_nat
['t', 'h', 'e', 'r', 'e', ' ', 'a', 'r', 'e', ' ', 'o', 'v', 'e', 'r', 'l', 'a',
  'p', 's', ' ', 'b', 'e', 't', 'w', 'e', 'e', 'n', ' ', 'P', ' ', 'a', 'n',
  'd', ' ', 'R']))
                                (\ _ ->
                                  catcha (check_NF_terms_subset (map fst r) q)
                                    (\ x ->
                                      Inl
(shows_prec_term Zero_nat x .
  shows_prec_list Zero_nat
    [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'i', 'n', ' ', 'n', 'o', 'r', 'm',
      'a', 'l', ' ', 'f', 'o', 'r', 'm', ' ', 'w', '.', 'r', '.', 't', '.', ' ',
      'R']))))))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mkc i (nfsc i dpp) False p [] [] [] r);
       });

check_dp_disproof ::
  forall a b.
    (Eq b,
      Showa b) => ([Char] -> [Char]) ->
                    Dpp_ops_ext a b [Char] () ->
                      a -> Dp_disproof b [Char] -> Sum ([Char] -> [Char]) ();
check_dp_disproof ia i dpp (DP_Loop s rseq sigma c) =
  debug (ia []) ['L', 'o', 'o', 'p']
    (let {
       p = pairsa i dpp;
       r = rulesd i dpp;
       nf = is_QNFc i dpp;
       nfs = nfsc i dpp;
       q = qc i dpp;
     } in catcha (check_dp_loop q nfs nf s rseq sigma c p r)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n',
                            ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ',
                            'l', 'o', 'o', 'p', ' ', 'f', 'o', 'r', ' ', 't',
                            'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i',
                            'n', 'g', ' ', 'D', 'P', '-', 'p', 'r', 'o', 'b',
                            'l', 'e', 'm', ' '] .
                          shows_nl .
                            shows_dpp (shows_prec Zero_nat)
                              (shows_prec_list Zero_nat) i dpp .
                              shows_nl . indent x)));
check_dp_disproof ia i dpp (DP_Nonloop prf) =
  debug (ia []) ['N', 'o', 'n', 'l', 'o', 'o', 'p']
    (catcha
      (let {
         p = pairsa i dpp;
         r = rulesd i dpp;
         _ = nfsc i dpp;
       } in bindb (check (null (qc i dpp))
                    (shows_prec_list Zero_nat
                      ['s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'f', 'o',
                        'r', ' ', 'n', 'o', 'n', '-', 'l', 'o', 'o', 'p', 's',
                        ' ', 'u', 'n', 's', 'u', 'p', 'p', 'o', 'r', 't', 'e',
                        'd']))
              (\ _ -> check_non_loop_prf r p prf))
      (\ x ->
        Inl (ia . shows_string
                    [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'c',
                      'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'n', 'o', 'n',
                      'l', 'o', 'o', 'p', ' ', 'f', 'o', 'r', ' ', 't', 'h',
                      'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                      ' ', 'D', 'P', '-', 'p', 'r', 'o', 'b', 'l', 'e', 'm',
                      ' '] .
                    shows_nl .
                      shows_dpp (shows_prec Zero_nat) (shows_prec_list Zero_nat)
                        i dpp .
                        shows_nl . indent x)));
check_dp_disproof ia i dpp (DP_Rule_Removal p r prf) =
  debug (ia []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (let {
       prm = (if is_none p then [] else list_diff (pairsa i dpp) (the p));
       rrm = (if is_none r then [] else list_diff (rulesd i dpp) (the r));
     } in catcha
            (check_dp_disproof (ia . shows_string ['.', '1']) i
              (delete_R_Rwc i (delete_P_Pwa i dpp prm prm) rrm rrm) prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'p', 'a',
                            'i', 'r', '/', 'r', 'u', 'l', 'e', ' ', 'r', 'e',
                            'm', 'o', 'v', 'a', 'l'] .
                          shows_nl . indent x)));
check_dp_disproof ia i dpp (DP_Q_Reduction q prf) =
  debug (ia []) ['Q', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n']
    (let {
       p = pairsa i dpp;
       r = rulesd i dpp;
       qa = qc i dpp;
       nfs = nfsc i dpp;
       f = funas_trs_list (p ++ r);
       rQ = filter (\ qb -> all (membera f) (funas_term_lista qb)) qa;
     } in bindb (catcha
                  (catcha (check_NF_terms_subset q rQ)
                    (\ x ->
                      Inl (shows_prec_list Zero_nat
                             ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                            shows_prec_term Zero_nat x .
                              shows_prec_list Zero_nat
                                [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                  'n', 'g', ' ', 'i', 'n', ' ', 'Q', '\'',
                                  ' '])))
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i',
                                  'n', ' ', 'r', 'e', 'd', 'u', 'c', 'i', 'n',
                                  'g', ' ', 't', 'h', 'e', ' ', 'i', 'n', 'n',
                                  'e', 'r', 'm', 'o', 's', 't', ' ', 'l', 'h',
                                  's', 's', ' ', 'i', 'n', ' ', 't', 'h', 'e',
                                  ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n',
                                  'g', ' ', 'D', 'P', '-', 'p', 'r', 'o', 'b',
                                  'l', 'e', 'm', ' '] .
                                shows_nl .
                                  shows_dpp (shows_prec Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl . indent x)))
            (\ _ ->
              catcha
                (check_dp_disproof (ia . shows_string ['.', '1']) i
                  (mkc i nfs False p [] q [] r) prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'Q',
                                ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                                'n'] .
                              shows_nl . indent x))));
check_dp_disproof ia i dpp (DP_Q_Increase q prf) =
  debug (ia []) ['Q', ' ', 'i', 'n', 'c', 'r', 'e', 'a', 's', 'e']
    (let {
       p = pairsa i dpp;
       r = rulesd i dpp;
       qa = qc i dpp;
       nfs = nfsc i dpp;
     } in catcha
            (check_dp_disproof (ia . shows_string ['.', '1']) i
              (mkc i nfs False p [] (list_union qa q) [] r) prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'Q', ' ',
                            'i', 'n', 'c', 'r', 'e', 'a', 's', 'e'] .
                          shows_nl . indent x)));
check_dp_disproof ia i dpp (DP_Termination_Switch joins_o prf) =
  debug (ia [])
    ['S', 'w', 'i', 't', 'c', 'h', ' ', 't', 'o', ' ', 'T', 'e', 'r', 'm', 'i',
      'n', 'a', 't', 'i', 'o', 'n']
    (bindb
      (catcha (switch_termination_proc i joins_o dpp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                        's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ', 't',
                        'o', ' ', 'f', 'u', 'l', 'l', ' ', 's', 't', 'r', 'a',
                        't', 'e', 'g', 'y', ' ', 'f', 'o', 'r', ' ', 't', 'h',
                        'e', ' ', 'D', 'P', '-', 'p', 'r', 'o', 'b', 'l', 'e',
                        'm', ' '] .
                      shows_nl .
                        shows_dpp (shows_prec Zero_nat)
                          (shows_prec_list Zero_nat) i dpp .
                          shows_nl . indent x)))
      (\ dppa ->
        catcha (check_dp_disproof (ia . shows_string ['.', '1']) i dppa prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 't', 'e', 'r', 'm',
                          'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 's', 'w', 'i',
                          't', 'c', 'h', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                          'o', 'r'] .
                        shows_nl . indent x))));
check_dp_disproof ia i dpp (DP_Instantiation p prf) =
  debug (ia [])
    ['I', 'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't', 'i', 'o', 'n']
    (let {
       pa = pairsa i dpp;
       q = qc i dpp;
       r = rulesd i dpp;
       nfs = nfsc i dpp;
     } in bindb (catcha
                  (bindb
                    (check (not nfs || null q)
                      (shows_prec_list Zero_nat
                        ['n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm',
                          ' ', 's', 'u', 'b', 's', 't', '.', ' ', 'c', 'u', 'r',
                          'r', 'e', 'n', 't', 'l', 'y', ' ', 'n', 'o', 't', ' ',
                          's', 'u', 'p', 'p', 'o', 'r', 't', 'e', 'd', ' ', 'f',
                          'o', 'r', ' ', 'i', 'n', 'n', 'e', 'r', 'm', 'o', 's',
                          't']))
                    (\ _ -> check_instance pa p))
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'i', 'n', 's', 't', 'a',
                                  'n', 't', 'i', 'a', 't', 'i', 'n', 'g', ' ',
                                  't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o',
                                  'w', 'i', 'n', 'g', ' ', 'D', 'P', '-', 'p',
                                  'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                                shows_nl .
                                  shows_dpp (shows_prec Zero_nat)
                                    (shows_prec_list Zero_nat) i dpp .
                                    shows_nl . indent x)))
            (\ _ ->
              catcha
                (check_dp_disproof (ia . shows_string ['.', '1']) i
                  (mkc i nfs False p [] q [] r) prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'i',
                                'n', 's', 't', 'a', 'n', 't', 'i', 'a', 't',
                                'i', 'o', 'n'] .
                              shows_nl . indent x))));

unsharp :: forall a b. Lab a b -> Lab a b;
unsharp (Sharp f) = f;
unsharp (Lab v va) = Lab v va;
unsharp (FunLab v va) = FunLab v va;
unsharp (UnLab v) = UnLab v;

check_dps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> a) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] -> Sum (Term a b, Term a b) ();
check_dps unshp r p =
  let {
    d = defs_list r;
  } in catcha
         (forallm
           (\ x ->
             (if let {
                   (ll, rr) = x;
                 } in not (is_Var rr) &&
                        not (membera d (the (root rr))) &&
                          any (\ (l, ra) ->
                                equal_term l (sharp_term unshp ll) &&
                                  (equal_term ra (sharp_term unshp rr) ||
                                    supt_impl ra (sharp_term unshp rr)))
                            r
               then Inr () else Inl x))
           p)
         (\ x -> Inl (snd x));

q_emptya :: forall a b c d. Tp_ops_ext a b c d -> a -> Bool;
q_emptya
  (Tp_ops_ext qreltrs q r rw rules q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs more)
  = q_empty;

applicable_rule_impl ::
  forall a b. (Term a b -> Bool) -> (Term a b, Term a b) -> Bool;
applicable_rule_impl nf r = all nf (args (fst r));

check_wwf_qtrs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_wwf_qtrs nf r =
  catcha
    (catcha
      (forallm
        (\ ra ->
          (if applicable_rule_impl nf ra
            then catcha
                   (bindb
                     (check (not (is_Var (fst ra)))
                       (shows_string
                         ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'l', 'e',
                           'f', 't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i',
                           'd', 'e', ' ', 'i', 'n']))
                     (\ _ ->
                       catcha
                         (check_subseteq (vars_term_lista (snd ra))
                           (vars_term_lista (fst ra)))
                         (\ x ->
                           Inl (shows_string
                                  ['f', 'r', 'e', 'e', ' ', 'v', 'a', 'r', 'i',
                                    'a', 'b', 'l', 'e', ' '] .
                                 shows_prec Zero_nat x .
                                   shows_string
                                     [' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h',
                                       't', '-', 'h', 'a', 'n', 'd', ' ', 's',
                                       'i', 'd', 'e', ' ', 'o', 'f']))))
                   (\ x ->
                     Inl (x . shows_string [' ', 'r', 'u', 'l', 'e', ' '] .
                                shows_rule (shows_prec Zero_nat)
                                  (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                  ra .
                                  shows_nl))
            else Inr ()))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (shows_string
             ['t', 'h', 'e', ' ', 'Q', '-', 'T', 'R', 'S', ' ', 'i', 's', ' ',
               'n', 'o', 't', ' ', 'w', 'e', 'a', 'k', 'l', 'y', ' ', 'w', 'e',
               'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd'] .
            shows_nl . x));

check_not_wwf_qtrs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => (Term a b -> Bool) ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_not_wwf_qtrs nf r =
  check (not (isOK (check_wwf_qtrs nf r)))
    (shows_string
       ['T', 'h', 'e', ' ', 'Q', '-', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'w',
         'e', 'l', 'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd'] .
      shows_nl);

check_trs_disproof ::
  forall a b c d.
    (Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    Tp_ops_ext a (Lab b c) [Char] () ->
                      Dpp_ops_ext d (Lab b c) [Char] () ->
                        a -> Trs_disproof (Lab b c) [Char] ->
                               Sum ([Char] -> [Char]) ();
check_trs_disproof ia i j tp (TRS_Loop s rseq sigma c) =
  debug (ia []) ['L', 'o', 'o', 'p']
    (let {
       r = rulesb i tp;
       nf = is_QNFa i tp;
       q = qa i tp;
       nfs = nfsa i tp;
     } in catcha (check_loop q nfs nf s rseq sigma c r)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h',
                            'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n',
                            'g', ' ', 'l', 'o', 'o', 'p', ' ', 'o', 'f', ' '] .
                          shows_nl .
                            shows_tp (shows_prec_lab Zero_nat)
                              (shows_prec_list Zero_nat) i tp .
                              shows_nl . indent x)));
check_trs_disproof ia i j tp (TRS_Nonloop prf) =
  debug (ia []) ['N', 'o', 'n', 'l', 'o', 'o', 'p']
    (catcha
      (let {
         r = rulesb i tp;
       } in bindb (check (null (qa i tp))
                    (shows_prec_list Zero_nat
                      ['s', 't', 'r', 'a', 't', 'e', 'g', 'y', ' ', 'f', 'o',
                        'r', ' ', 'n', 'o', 'n', '-', 'l', 'o', 'o', 'p', 's',
                        ' ', 'u', 'n', 's', 'u', 'p', 'p', 'o', 'r', 't', 'e',
                        'd']))
              (\ _ -> check_non_loop_prf r [] prf))
      (\ x ->
        Inl (shows_prec_list Zero_nat
               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 'c', 'h',
                 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'n', 'o', 'n', '-', 'l',
                 'o', 'o', 'p', 'i', 'n', 'g', ' ', 'r', 'e', 'd', 'u', 'c',
                 't', 'i', 'o', 'n', ' ', 'f', 'o', 'r', ' '] .
              shows_nl .
                shows_tp (shows_prec_lab Zero_nat) (shows_prec_list Zero_nat) i
                  tp .
                  indent x)));
check_trs_disproof ia i j tp TRS_Not_Well_Formed =
  debug (ia [])
    ['N', 'o', 't', ' ', 'W', 'e', 'l', 'l', '-', 'F', 'o', 'r', 'm', 'e', 'd']
    (bindb
      (check (null (qa i tp) || not (nfsa i tp))
        (shows_prec_list Zero_nat
          ['s', 't', 'r', 'a', 't', 'e', 'g', 'i', 'e', 's', ' ', 'a', 'n', 'd',
            ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', ' ',
            's', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i', 'o', 'n', 's',
            ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm']))
      (\ _ ->
        catcha (check_not_wwf_qtrs (is_QNFa i tp) (rulesb i tp))
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                          'n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o',
                          'r', 'm', 'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f'] .
                        shows_nl . x))));
check_trs_disproof ia i j tp (TRS_Rule_Removal r prf) =
  debug (ia []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (let {
       ra = rulesb i tp;
       rrm = list_diff ra r;
     } in catcha
            (check_trs_disproof (ia . shows_string ['.', '1']) i j
              (delete_R_Rwa i tp rrm rrm) prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u',
                            'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l'] .
                          shows_nl . indent x)));
check_trs_disproof ia i j tp (TRS_String_Reversal prf) =
  debug (ia [])
    ['S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l']
    (let {
       r = rulesb i tp;
     } in bindb (catcha
                  (bindb
                    (check (q_emptya i tp)
                      (shows_prec_list Zero_nat
                        ['Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm',
                          'p', 't', 'y']))
                    (\ _ ->
                      (case check_unary_signature r of {
                        Inl a -> Inl a;
                        Inr _ -> Inr (mka i False [] (map rev_rule r) []);
                      })))
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'a', 'p', 'p', 'l', 'y',
                                  'i', 'n', 'g', ' ', 's', 't', 'r', 'i', 'n',
                                  'g', ' ', 'r', 'e', 'v', 'e', 'r', 's', 'a',
                                  'l', ' ', 'o', 'n'] .
                                shows_nl .
                                  shows_tp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i tp .
                                    shows_nl . x)))
            (\ tpa ->
              catcha
                (check_trs_disproof (ia . shows_string ['.', '1']) i j tpa prf)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                                'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 's',
                                't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v',
                                'e', 'r', 's', 'a', 'l', ' ', 't', 'e', 'c',
                                'h', 'n', 'i', 'q', 'u', 'e'] .
                              shows_nl . indent x))));
check_trs_disproof ia i j tp (TRS_DP_Trans p prf) =
  debug (ia [])
    ['D', 'P', ' ', 'T', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i',
      'o', 'n']
    (let {
       r = rulesb i tp;
       q = qa i tp;
     } in bindb (catcha
                  (bindb
                    (check (null (qa i tp) || not (nfsa i tp))
                      (shows_prec_list Zero_nat
                        ['s', 't', 'r', 'a', 't', 'e', 'g', 'i', 'e', 's', ' ',
                          'a', 'n', 'd', ' ', 'n', 'o', 'r', 'm', 'a', 'l', ' ',
                          'f', 'o', 'r', 'm', ' ', 's', 'u', 'b', 's', 't', 'i',
                          't', 'u', 't', 'i', 'o', 'n', 's', ' ', 'p', 'r', 'o',
                          'b', 'l', 'e', 'm']))
                    (\ _ ->
                      catcha (check_dps unsharp r p)
                        (\ x ->
                          Inl (shows_prec_list Zero_nat
                                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', 'a', 't',
                                   'i', 'c', ' ', 'r', 'u', 'l', 'e', ':',
                                   ' '] .
                                shows_rule (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat)
                                  [' ', '-', '>', ' '] x))))
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i',
                                  'n', ' ', 'D', 'P', ' ', 't', 'r', 'a', 'n',
                                  's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o',
                                  'n', ' ', 'o', 'n', ' '] .
                                shows_nl .
                                  shows_tp (shows_prec_lab Zero_nat)
                                    (shows_prec_list Zero_nat) i tp .
                                    shows_nl . x)))
            (\ _ ->
              let {
                dpp = mkc j False False p [] q [] r;
              } in catcha
                     (check_dp_disproof (ia . shows_string ['.', '1']) j dpp
                       prf)
                     (\ x ->
                       Inl (ia . shows_string
                                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                     'e', 'l', 'o', 'w', ' ', 't', 'h', 'e',
                                     ' ', 'd', 'p', ' ', 't', 'r', 'a', 'n',
                                     's', 'f', 'o', 'r', 'm', 'a', 't', 'i',
                                     'o', 'n'] .
                                   shows_nl . indent x))));
check_trs_disproof ia i j tp (TRS_Q_Increase q prf) =
  debug (ia []) ['Q', ' ', 'i', 'n', 'c', 'r', 'e', 'a', 's', 'e']
    (let {
       r = rulesb i tp;
       qaa = qa i tp;
       tpa = mka i (nfsa i tp) (list_union qaa q) r [];
     } in catcha (check_trs_disproof (ia . shows_string ['.', '1']) i j tpa prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'Q', ' ',
                            'i', 'n', 'c', 'r', 'e', 'a', 's', 'e', ' ', 't',
                            'e', 'c', 'h', 'n', 'i', 'q', 'u', 'e'] .
                          shows_nl . indent x)));

less_eq_complexity_class :: Complexity_class -> Complexity_class -> Bool;
less_eq_complexity_class x y = less_eq_nat (degree x) (degree y);

shows_prec_complexity_class :: Nat -> Complexity_class -> [Char] -> [Char];
shows_prec_complexity_class d c =
  (if equal_nat (degree c) Zero_nat
    then shows_prec_list Zero_nat ['O', '(', '1', ')']
    else (if equal_nat (degree c) one_nat
           then shows_prec_list Zero_nat ['O', '(', 'n', ')']
           else shows_prec_list Zero_nat ['O', '(', 'n', '^'] .
                  shows_prec_nat Zero_nat (degree c) .
                    shows_prec_list Zero_nat [')']));

rule_shift_complexity_tt ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => Tp_ops_ext a b c () ->
                    Redtriple_ext b c () ->
                      [(Term b c, Term b c)] ->
                        Complexity_measure b c ->
                          Complexity_class -> a -> Sum ([Char] -> [Char]) a;
rule_shift_complexity_tt i rp rdelete cm cc tp =
  let {
    r = ra i tp;
    rw = rwa i tp;
    r2 = list_diff r rdelete;
  } in (case catcha
               (bindb (valid rp)
                 (\ _ ->
                   bindb (catcha (mono rp (rdelete ++ rw ++ r2))
                           (\ x ->
                             Inl (shows_string
                                    ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
                                      'w', 'i', 't', 'h', ' ', 'm', 'o', 'n',
                                      'o', 't', 'o', 'n', 'i', 'c', 'i', 't',
                                      'y', ' ', 'o', 'f', ' ', 's', 't', 'r',
                                      'i', 'c', 't', ' ', 'o', 'r', 'd', 'e',
                                      'r'] .
                                   shows_nl . x)))
                     (\ _ ->
                       bindb (catcha
                               (catcha (forallm (s rp) rdelete)
                                 (\ x -> Inl (snd x)))
                               (\ x ->
                                 Inl (shows_string
['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i',
  'e', 'n', 't', 'i', 'n', 'g', ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'T',
  'R', 'S'] .
                                       shows_nl . x)))
                         (\ _ ->
                           bindb (catcha
                                   (catcha (forallm (ns rp) (rw ++ r2))
                                     (\ x -> Inl (snd x)))
                                   (\ x ->
                                     Inl (shows_string
    ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
      'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'n', 'o', 'n', '-', 's', 't', 'r',
      'i', 'c', 't', ' ', 'T', 'R', 'S'] .
   shows_nl . x)))
                             (\ _ ->
                               (case cpx rp cm of {
                                 Nothing ->
                                   Inl (shows_prec_list Zero_nat
 ['n', 'o', ' ', 'c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', ' ', 'f',
   'o', 'u', 'n', 'd']);
                                 Just cca ->
                                   check (less_eq_complexity_class cca cc)
                                     (shows_prec_list Zero_nat
['c', 'o', 'u', 'l', 'd', ' ', 'o', 'n', 'l', 'y', ' ', 'd', 'e', 'r', 'i', 'v',
  'e', ' ', 'c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', ' '] .
                                       shows_prec_complexity_class Zero_nat
 cca);
                               }))))))
               (\ x ->
                 Inl (shows_string
                        ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'd',
                          'e', 'r', 'i', 'v', 'e', ' ', 't', 'h', 'e', ' ', 'i',
                          'n', 't', 'e', 'n', 'd', 'e', 'd', ' ', 'c', 'o', 'm',
                          'p', 'l', 'e', 'x', 'i', 't', 'y', ' '] .
                       shows_prec_complexity_class Zero_nat cc .
                         shows_prec_list Zero_nat
                           [' ', 'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ',
                             'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
                           shows_nl . desca rp . shows_nl . x))
         of {
         Inl a -> Inl a;
         Inr _ -> Inr (mka i (nfsa i tp) (qa i tp) r2 (list_union rw rdelete));
       });

check_complexity_proof ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    Tp_ops_ext a b c () ->
                      a -> Complexity_measure b c ->
                             Complexity_class ->
                               Complexity_proof b c ->
                                 Sum ([Char] -> [Char]) ();
check_complexity_proof ia i tp cm cc (Rule_Shift_Complexity redp rdelete prf) =
  debug (ia [])
    ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'C', 'o',
      'm', 'p', 'l', 'e', 'x', 'i', 't', 'y']
    (bindb
      (catcha (rule_shift_complexity_tt i (get_redtriple redp) rdelete cm cc tp)
        (\ x ->
          Inl (ia . shows_string
                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                        'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                        't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 's', 'h',
                        'i', 'f', 't', 'i', 'n', 'g', ' ', 't', 'e', 'c', 'h',
                        'n', 'i', 'q', 'u', 'e', ' ', 'o', 'n', ' '] .
                      shows_nl .
                        shows_tp (shows_prec Zero_nat) (shows_prec Zero_nat) i
                          tp .
                          shows_nl . x)))
      (\ tpa ->
        catcha
          (check_complexity_proof (ia . shows_string ['.', '1']) i tpa cm cc
            prf)
          (\ x ->
            Inl (ia . shows_string
                        [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                          'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
                          ' ', 's', 'h', 'i', 'f', 't', 'i', 'n', 'g', ' ', 'p',
                          'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                        shows_nl . indent x))));
check_complexity_proof ia i tp cm cc RisEmpty_Complexity =
  debug (ia [])
    ['R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y', ' ', 'f', 'o', 'r', ' ',
      'c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y']
    (check (null (ra i tp))
      (ia . shows_string
              [':', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm',
                'p', 't', 'y', ' ', 'i', 'n', ' '] .
              shows_nl .
                shows_tp (shows_prec Zero_nat) (shows_prec Zero_nat) i tp));

ident_prob_of_semp ::
  forall a b.
    (Eq a,
      Eq b) => (Ctxt a b,
                 (Term a b, (Ctxt a b, (Term a b, [(Term a b, Term a b)])))) ->
                 [(Term a b, Term a b)];
ident_prob_of_semp (d, (l, (c, (t, mp)))) = ident_prob_of_smp mp;

eident_prob_of_semp ::
  forall a b.
    (Eq a,
      Eq b) => (Ctxt a b,
                 (Term a b, (Ctxt a b, (Term a b, [(Term a b, Term a b)])))) ->
                 Maybe (Ctxt a b, (Term a b, (Ctxt a b, Term a b)));
eident_prob_of_semp (d, (l, (c, (t, mp)))) =
  bind (map_of (map (\ (x, y) -> (y, x)) mp) l)
    (\ si -> Just (d, (si, (c, t))));

simplify_emp_main_impl ::
  forall a b c d e f g.
    (Eq a, Eq b, Eq e,
      Eq f) => (a -> Term b a) ->
                 ([(Term b a, Term b c)] ->
                   [d] -> Maybe ([(Term e f, Term e f)], Nat)) ->
                   ((Term e f, Term e f) -> Maybe Nat) ->
                     [a] ->
                       (Ctxt b a,
                         (Term b c,
                           (Ctxt b a, (Term b a, [(Term b a, Term b c)])))) ->
                         [(Term b a, Term b c)] ->
                           Maybe (Maybe (Ctxt b a,
  (Term g c, (Ctxt b a, (Term b a, [(Term b a, Term b c)])))));
simplify_emp_main_impl mu simp ident v_incr egmp solved =
  (case egmp of {
    (d, (Var x, (c, (t, [])))) -> Just (Just (d, (Var x, (c, (t, solved)))));
    (Hole, (Fun g ls, (c, (t, [])))) ->
      (if gmatch_decision mu simp ident ((t, Fun g ls) : solved)
        then Just Nothing
        else (if equal_ctxt c Hole then Nothing
               else simplify_emp_main_impl mu simp ident v_incr
                      (c, (Fun g ls,
                            (subst_apply_ctxt c mu,
                              (subst_apply_term t mu, []))))
                      solved));
    (More f bef d aft, (Fun g ls, (c, (t, [])))) ->
      bind (guard (f == g))
        (\ _ ->
          bind (guard
                 (equal_nat (size_list ls)
                   (Suc (plus_nat (size_list bef) (size_list aft)))))
            (\ _ ->
              let {
                pairs_bef = zip bef (takea (size_list bef) ls);
                pairs_aft = zip aft (dropa (Suc (size_list bef)) ls);
              } in simplify_emp_main_impl mu simp ident v_incr
                     (d, (nth ls (size_list bef),
                           (c, (t, pairs_bef ++ pairs_aft))))
                     solved));
    (d, (l, (c, (t, (Var v, Var x) : mp)))) ->
      simplify_emp_main_impl mu simp ident v_incr (d, (l, (c, (t, mp))))
        ((Var v, Var x) : solved);
    (d, (l, (c, (t, (Var v, Fun f ls) : mp)))) ->
      bind (guard (membera v_incr v))
        (\ _ ->
          let {
            m = map (\ (s, a) -> (subst_apply_term s mu, a));
          } in simplify_emp_main_impl mu simp ident v_incr
                 (subst_apply_ctxt d mu,
                   (l, (subst_apply_ctxt c mu,
                         (subst_apply_term t mu, m ((Var v, Fun f ls) : mp)))))
                 (m solved));
    (d, (l, (c, (t, (Fun fa lista, Var x) : mp)))) ->
      simplify_emp_main_impl mu simp ident v_incr (d, (l, (c, (t, mp))))
        ((Fun fa lista, Var x) : solved);
    (d, (l, (c, (t, (Fun fa lista, Fun f ls) : mp)))) ->
      bind (guard (f == fa))
        (\ _ ->
          bind (zip_option lista ls)
            (\ pairs ->
              simplify_emp_main_impl mu simp ident v_incr
                (d, (l, (c, (t, pairs ++ mp)))) solved));
  });

simplify_emp_impl ::
  forall a b c d e f g.
    (Eq a, Eq b, Eq e,
      Eq f) => (a -> Term b a) ->
                 ([(Term b a, Term b c)] ->
                   [d] -> Maybe ([(Term e f, Term e f)], Nat)) ->
                   ((Term e f, Term e f) -> Maybe Nat) ->
                     [a] ->
                       (Ctxt b a, (Term b c, (Ctxt b a, Term b a))) ->
                         Sum (Ctxt b a,
                               (Term g c,
                                 (Ctxt b a,
                                   (Term b a, [(Term b a, Term b c)]))))
                           Bool;
simplify_emp_impl mu simp ident v_incr emp =
  let {
    (d, (l, (c, t))) = emp;
  } in (case simplify_emp_main_impl mu simp ident v_incr (d, (l, (c, (t, []))))
               []
         of {
         Nothing -> Inr False;
         Just Nothing -> Inr True;
         Just (Just a) -> Inl a;
       });

eident_prob_to_ident_prob ::
  forall a b.
    (Ctxt a b, (Term a b, (Ctxt a b, Term a b))) -> (Term a b, Term a b);
eident_prob_to_ident_prob (d, (si, (c, t))) = (ctxt_apply d t, si);

ident_prob_of_emp_impl ::
  forall a b c d e.
    (Eq a, Eq b, Eq d,
      Eq e) => (a -> Term b a) ->
                 ([(Term b a, Term b a)] ->
                   [c] -> Maybe ([(Term d e, Term d e)], Nat)) ->
                   ((Term d e, Term d e) -> Maybe Nat) ->
                     [a] ->
                       (Ctxt b a, (Term b a, (Ctxt b a, Term b a))) ->
                         Maybe [(Term b a, Term b a)];
ident_prob_of_emp_impl mu simp ident v_incr emp =
  (case simplify_emp_impl mu simp ident v_incr emp of {
    Inl semp ->
      Just (map eident_prob_to_ident_prob
              (option_to_list (eident_prob_of_semp semp)) ++
             ident_prob_of_semp semp);
    Inr True -> Just [];
    Inr False -> Nothing;
  });

ident_decision ::
  forall a b. (Eq a, Eq b) => (a -> Term b a) -> (Term b a, Term b a) -> Bool;
ident_decision ident ip =
  (case ident_solve_impl ident ip of {
    Nothing -> False;
    Just _ -> True;
  });

o_match_probs_impl ::
  forall a b. Term a b -> Pos -> Term a b -> [(Term a b, Term a b)];
o_match_probs_impl l q t = map (\ p -> (subt_at t p, l)) (prefix_list q);

o_ematch_probs_impl ::
  forall a b.
    (a -> Term b a) ->
      Ctxt b a ->
        Term b a ->
          Pos -> Term b a -> [(Ctxt b a, (Term b a, (Ctxt b a, Term b a)))];
o_ematch_probs_impl mu c l q t =
  map (\ p ->
        (subt_at_ctxt c p, (l, (subst_apply_ctxt c mu, subst_apply_term t mu))))
    (prefix_list (hole_pos c));

ostep_ctxt_subst_cond_impl ::
  forall a b c d e.
    (Eq a, Eq b, Eq d,
      Eq e) => ([(Term a b, Term a b)] ->
                 [c] -> Maybe ([(Term d e, Term d e)], Nat)) ->
                 ((Term d e, Term d e) -> Maybe Nat) ->
                   ((Ctxt a b, (Term a b, (Ctxt a b, Term a b))) ->
                     Maybe [(Term a b, Term a b)]) ->
                     (b -> Term a b) ->
                       Ctxt a b -> Term a b -> Pos -> Term a b -> Bool;
ostep_ctxt_subst_cond_impl simp ident ident_of_emp mu c l q t =
  not (any (match_decision mu simp ident) (o_match_probs_impl l q t) ||
        any (\ mp -> bex (set (ident_of_emp mp)) (all (ident_decision mu)))
          (o_ematch_probs_impl mu c l q t));

ostep_ctxt_subst_cond_decision ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Term b a)] ->
                 (Ctxt b a, (Term b a, (Pos, Term b a))) -> Bool;
ostep_ctxt_subst_cond_decision mu =
  let {
    mua = mk_subst mu;
    d = mk_subst_domain mu;
    v = v_incr_impl d;
    simp = simplify_mp mu v;
    ident = ident_solve_impl mua;
    ident_of_emp = ident_prob_of_emp_impl mua simp ident v;
    _ = w_impl d;
    main = ostep_ctxt_subst_cond_impl simp ident ident_of_emp mua;
  } in (\ (c, (l, (a, b))) -> main c l a b);

check_oloop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [Term a b] ->
                      Ctxt a b ->
                        [(b, Term a b)] ->
                          Term a b ->
                            [(Pos, ((Term a b, Term a b), Term a b))] ->
                              Sum ([Char] -> [Char]) ();
check_oloop r q c sigma t seqa =
  let {
    mu = mk_subst sigma;
  } in bindb (check (not (null seqa))
               (shows_prec_list Zero_nat
                 ['l', 'o', 'o', 'p', 'i', 'n', 'g', ' ', 'r', 'e', 'd', 'u',
                   'c', 't', 'i', 'o', 'n', ' ', 'm', 'u', 's', 't', ' ', 'n',
                   'o', 't', ' ', 'b', 'e', ' ', 'e', 'm', 'p', 't', 'y']))
         (\ _ ->
           bindb (check
                   (equal_term (let {
                                  (_, (_, ta)) = lasta seqa;
                                } in ta)
                     (ctxt_apply c (subst_apply_term t mu)))
                   (shows_prec_list Zero_nat
                     ['l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ', 'i',
                       'n', ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e', ' ',
                       'i', 's', ' ', 'n', 'o', 't', ' ', 'C', '[', 't', ' ',
                       's', 'i', 'g', 'm', 'a', ']']))
             (\ _ ->
               let {
                 seqb = zip (t : map (\ (_, (_, ta)) -> ta) seqa) seqa;
                 check_ostep = ostep_ctxt_subst_cond_decision sigma;
               } in bindb (catcha
                            (forallm
                              (\ (ta, (p, (ra, s))) ->
                                bindb (check_rstep r p ra ta s)
                                  (\ _ ->
                                    catcha
                                      (forallm
(\ l ->
  check (check_ostep (c, (l, (p, ta))))
    (shows_prec_list Zero_nat
       ['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', ' ', 'f', 'r', 'o', 'm',
         ' '] .
      shows_prec_term Zero_nat ta .
        shows_prec_list Zero_nat [' ', '-', '-', '>'] .
          shows_prec_pos Zero_nat p .
            shows_prec_list Zero_nat [' '] .
              shows_prec_term Zero_nat s .
                shows_prec_list Zero_nat
                  [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'r', 'e',
                    's', 'p', 'e', 'c', 't', ' ', 'i', 't', 'e', 'r', 'a', 't',
                    'e', 'd', ' ', 'o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't',
                    ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n']))
q)
                                      (\ x -> Inl (snd x))))
                              seqb)
                            (\ x -> Inl (snd x)))
                      (\ _ -> Inr ())));

check_otrs_disproof ::
  forall a b c.
    (Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    Tp_ops_ext a b c () ->
                      a -> Otrs_disproof b c -> Sum ([Char] -> [Char]) ();
check_otrs_disproof ia i tp (OTRS_Loop s rseq sigma c) =
  debug (ia []) ['L', 'o', 'o', 'p']
    (let {
       r = rulesb i tp;
       q = qa i tp;
     } in catcha (check_oloop r q c sigma s rseq)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h',
                            'e', 'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n',
                            'g', ' ', 'o', 'u', 't', 'e', 'r', 'm', 'o', 's',
                            't', ' ', 'l', 'o', 'o', 'p'] .
                          shows_nl . indent x)));
check_otrs_disproof ia i tp (OTRS_Rule_Removal r prf) =
  debug (ia []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (let {
       ra = rulesb i tp;
       rrm = list_diff ra r;
     } in catcha
            (check_otrs_disproof (ia . shows_string ['.', '1']) i
              (delete_R_Rwa i tp rrm rrm) prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u',
                            'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l'] .
                          shows_nl . indent x)));

check_rel_seq ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                        Term a b ->
                          Term a b -> Bool -> Sum ([Char] -> [Char]) ();
check_rel_seq r sa [] s u True =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
         ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e'] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_string
            (shows_nl
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm']) .
            shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_rel_seq r sa [] s u False =
  Inl (shows_prec_list Zero_nat
        ['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 's',
          't', 'r', 'i', 'c', 't', ' ', 's', 't', 'e', 'p', ' ', 'i', 'n', ' ',
          'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u', 'e', 'n',
          'c', 'e']);
check_rel_seq ra sa ((p, (r, (True, t))) : prts) s u b =
  bindb (check_qrstep (\ _ -> True) False ra p r s t)
    (\ _ -> check_rel_seq ra sa prts t u True);
check_rel_seq ra sa ((p, (r, (False, t))) : prts) s u b =
  bindb (check_qrstep (\ _ -> True) False sa p r s t)
    (\ _ -> check_rel_seq ra sa prts t u b);

check_rel_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b ->
                    [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                      [(b, Term a b)] ->
                        Ctxt a b ->
                          [(Term a b, Term a b)] ->
                            [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_rel_loop sa rseq sigma c r s =
  check_rel_seq r s rseq sa
    (ctxt_apply c (subst_apply_term sa (mk_subst sigma))) False;

check_wf_reltrs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                    Sum ([Char] -> [Char]) ();
check_wf_reltrs (r, s) =
  bindb (check_wf_trs r)
    (\ _ ->
      (if null r then Inr ()
        else catcha
               (forallm
                 (\ rule ->
                   catcha
                     (catcha
                       (check_subseteq (vars_term_lista (snd rule))
                         (vars_term_lista (fst rule)))
                       (\ x ->
                         Inl (shows_string
                                ['f', 'r', 'e', 'e', ' ', 'v', 'a', 'r', 'i',
                                  'a', 'b', 'l', 'e', ' '] .
                               shows_prec Zero_nat x .
                                 shows_string
                                   [' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't',
                                     '-', 'h', 'a', 'n', 'd', ' ', 's', 'i',
                                     'd', 'e', ' ', 'o', 'f'])))
                     (\ x ->
                       Inl (x . shows_string [' ', 'r', 'u', 'l', 'e', ' '] .
                                  shows_rule (shows_prec Zero_nat)
                                    (shows_prec Zero_nat) [' ', '-', '>', ' ']
                                    rule .
                                    shows_nl)))
                 s)
               (\ x -> Inl (snd x))));

check_not_wf_reltrs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                    Sum ([Char] -> [Char]) ();
check_not_wf_reltrs rs =
  check (not (isOK (check_wf_reltrs rs)))
    (shows_string
       ['T', 'h', 'e', ' ', 'T', 'R', 'S', 's', ' ', 'R', ' ', 'a', 'n', 'd',
         ' ', 'S', ' ', 'a', 'r', 'e', ' ', 'w', 'e', 'l', 'l', ' ', 'f', 'o',
         'r', 'm', 'e', 'd'] .
      shows_nl);

check_reltrs_disproof ::
  forall a b c d.
    (Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    Tp_ops_ext a (Lab b c) [Char] () ->
                      Dpp_ops_ext d (Lab b c) [Char] () ->
                        a -> Reltrs_disproof (Lab b c) [Char] ->
                               Sum ([Char] -> [Char]) ();
check_reltrs_disproof ia i j tp (Rel_Loop s rseq sigma c) =
  debug (ia []) ['L', 'o', 'o', 'p']
    (let {
       r = ra i tp;
       sa = rwa i tp;
     } in bindb (check (q_emptya i tp)
                  (shows_prec_list Zero_nat
                    ['Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p',
                      't', 'y']))
            (\ _ ->
              catcha (check_rel_loop s rseq sigma c r sa)
                (\ x ->
                  Inl (ia . shows_string
                              [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n',
                                ' ', 'l', 'o', 'o', 'p', ' ', 'p', 'r', 'o',
                                'o', 'f'] .
                              shows_nl . x))));
check_reltrs_disproof ia i j tp (Rel_R_Not_SN prf) =
  let {
    q = qa i tp;
    r = ra i tp;
    nfs = nfsa i tp;
    tpa = mka i nfs q r [];
  } in catcha (check_trs_disproof (ia . shows_string ['.', '1']) i j tpa prf)
         (\ x ->
           Inl (ia . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'n', 'o', 'n', 't',
                         'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ',
                         'p', 'r', 'o', 'o', 'f', ' ', 'f', 'o', 'r', ' ',
                         'R'] .
                       shows_nl . indent x));
check_reltrs_disproof ia i j tp Rel_Not_Well_Formed =
  debug (ia [])
    ['N', 'o', 't', ' ', 'W', 'e', 'l', 'l', '-', 'F', 'o', 'r', 'm', 'e', 'd']
    (let {
       r = ra i tp;
       s = rwa i tp;
     } in catcha
            (bindb
              (check (q_emptya i tp)
                (shows_prec_list Zero_nat
                  ['Q', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm', 'p',
                    't', 'y']))
              (\ _ -> check_not_wf_reltrs (r, s)))
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n',
                            ' ', 'n', 'o', 't', '-', 'w', 'e', 'l', 'l', '-',
                            'f', 'o', 'r', 'm', 'e', 'd', ' ', 'p', 'r', 'o',
                            'o', 'f'] .
                          shows_nl . x)));
check_reltrs_disproof ia i j tp (Rel_Rule_Removal r s prf) =
  debug (ia []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (let {
       rrm = (if is_none r then [] else list_diff (ra i tp) (the r));
       srm = (if is_none s then [] else list_diff (rwa i tp) (the s));
     } in catcha
            (check_reltrs_disproof (ia . shows_string ['.', '1']) i j
              (delete_R_Rwa i tp rrm srm) prf)
            (\ x ->
              Inl (ia . shows_string
                          [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                            'l', 'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u',
                            'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l'] .
                          shows_nl . indent x)));

check_subsumption_NF ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_subsumption_NF e r =
  catcha
    (forallm
      (\ ea ->
        catcha (check_join_NF r (fst ea) (snd ea))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'j', 'o',
                     'i', 'n', ' ', 'e', 'q', 'u', 'a', 't', 'i', 'o', 'n',
                     ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '=', ' '] ea .
                    shows_nl . x)))
      e)
    (\ x -> Inl (snd x));

check_estep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Pos ->
                      (Term a b, Term a b) ->
                        Bool ->
                          Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_estep e p rule l_to_r s t =
  bindb (check (any (eq_rule_mod_vars rule) e)
          (shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
             [' ', '-', '>', ' '] rule .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', 'e', 'q',
                'u', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f'] .
              shows_nl .
                shows_trs (shows_prec Zero_nat) (shows_prec Zero_nat)
                  ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', ' ', 's',
                    'y', 's', 't', 'e', 'm', ':']
                  [' ', '=', ' '] e .
                  shows_nl))
    (\ _ ->
      bindb (check (in_poss p s)
              (shows_prec_pos Zero_nat p .
                shows_string
                  [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'o',
                    's', 'i', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                  shows_prec_term Zero_nat s . shows_nl))
        (\ _ ->
          bindb (check (in_poss p t)
                  (shows_prec_pos Zero_nat p .
                    shows_string
                      [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p',
                        'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                      shows_prec_term Zero_nat t . shows_nl))
            (\ _ ->
              let {
                c = ctxt_of_pos_term p s;
                d = ctxt_of_pos_term p t;
                u = subt_at s p;
                v = subt_at t p;
                rrule = (if l_to_r then rule else (snd rule, fst rule));
                err = shows_string
                        ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                        shows_prec_term Zero_nat t .
                          shows_string
                            [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ',
                              'r', 'e', 's', 'u', 'l', 't', ' ', 'f', 'r', 'o',
                              'm', ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r',
                              ' ', 'a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                              'o', 'n', ' ', 'o', 'f', ' ', 't', 'e', 'r', 'm',
                              ' '] .
                            shows_prec_term Zero_nat s .
                              shows_string
                                [' ', 'u', 's', 'i', 'n', 'g', ' ', 'e', 'q',
                                  'u', 'a', 't', 'i', 'o', 'n', ' '] .
                                shows_nl .
                                  shows_rule (shows_prec Zero_nat)
                                    (shows_prec Zero_nat) [' ', '=', ' ']
                                    rrule .
                                    shows_string
                                      [' ', 'a', 't', ' ', 'p', 'o', 's', 'i',
't', 'i', 'o', 'n', ' '] .
                                      shows_prec_pos Zero_nat p . shows_nl;
              } in (case match_list u [(u, fst rrule), (v, snd rrule)] of {
                     Nothing -> Inl err;
                     Just _ -> check (equal_ctxt c d) err;
                   }))));

check_conversion ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                      Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_conversion e [] s u =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'c', 'o', 'n', 'v', 'e', 'r', 's',
         'i', 'o', 'n', ' '] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_nl .
            shows_string
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm'] .
              shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_conversion e ((p, (r, (l_to_r, t))) : c) s u =
  bindb (check_estep e p r l_to_r s t) (\ _ -> check_conversion e c t u);

check_subsumptions_guided ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [((Term a b, Term a b),
                       [(Pos, ((Term a b, Term a b), (Bool, Term a b)))])] ->
                      Sum ([Char] -> [Char]) ();
check_subsumptions_guided e [] = Inr ();
check_subsumptions_guided ea ((e, seqa) : convs) =
  bindb (catcha (check_conversion ea seqa (fst e) (snd e))
          (\ x ->
            Inl (shows_string
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n', ' ', 'c',
                     'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n', ' ', 'f', 'o',
                     'r', ' ', 'e', 'q', 'u', 'a', 't', 'i', 'o', 'n', ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '=', ' '] e .
                    shows_nl . x)))
    (\ _ -> check_subsumptions_guided (e : ea) convs);

check_subsumption_guided ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      [((Term a b, Term a b),
                         [(Pos, ((Term a b, Term a b), (Bool, Term a b)))])] ->
                        Sum ([Char] -> [Char]) ();
check_subsumption_guided ea e convs =
  bindb (catcha (check_subseteq ea (map fst convs))
          (\ x ->
            Inl (shows_string
                   ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                     'n', 'd', ' ', 'c', 'o', 'n', 'v', 'e', 'r', 's', 'i', 'o',
                     'n', ' ', 'f', 'o', 'r', ' ', 'e', 'q', 'u', 'a', 't', 'i',
                     'o', 'n', ' '] .
                  shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                    [' ', '=', ' '] x)))
    (\ _ -> check_subsumptions_guided e convs);

check_subsumption ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      Maybe [((Term a b, Term a b),
                               [(Pos, ((Term a b, Term a b),
(Bool, Term a b)))])] ->
                        Sum ([Char] -> [Char]) ();
check_subsumption e r convs_o =
  (case convs_o of {
    Nothing -> check_subsumption_NF e r;
    Just a -> check_subsumption_guided e r a;
  });

check_completion_proof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          [(Term (Lab b [Nat]) [Char],
                             Term (Lab b [Nat]) [Char])] ->
                            [(Term (Lab b [Nat]) [Char],
                               Term (Lab b [Nat]) [Char])] ->
                              Completion_proof b [Nat] [Char] ->
                                Sum ([Char] -> [Char]) ();
check_completion_proof a ia i j e r (SN_WCR_Eq joins_i prf conv1 conv2) =
  debug (ia []) ['S', 'N', '_', 'W', 'C', 'R', '_', 'E', 'q']
    (let {
       tp = mka i False [] r [];
     } in bindb (catcha
                  (check_trs_proof a (ia . shows_string ['.', '1']) i j tp prf)
                  (\ x ->
                    Inl (ia . shows_string
                                [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b',
                                  'e', 'l', 'o', 'w', ' ', 's', 't', 'r', 'o',
                                  'n', 'g', ' ', 'n', 'o', 'r', 'm', 'a', 'l',
                                  'i', 'z', 'a', 't', 'i', 'o', 'n', ' ', '+',
                                  ' ', 'w', 'c', 'r'] .
                                shows_nl . indent x)))
            (\ _ ->
              bindb (catcha (check_subsumption_guided r e conv1)
                      (\ x ->
                        Inl (ia . shows_prec_list Zero_nat
                                    [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
                                      'w', 'h', 'e', 'n', ' ', 's', 'h', 'o',
                                      'w', 'i', 'n', 'g', ' ', 't', 'h', 'a',
                                      't', ' ', 'r', 'e', 'w', 'r', 'i', 't',
                                      'e', ' ', 'r', 'e', 'l', 'a', 't', 'i',
                                      'o', 'n', ' ', 'c', 'a', 'n', ' ', 'b',
                                      'e', ' ', 's', 'i', 'm', 'u', 'l', 'a',
                                      't', 'e', 'd', ' ', 'b', 'y', ' ', 'e',
                                      'q', 'u', 'a', 't', 'i', 'o', 'n', 's'] .
                                    indent x)))
                (\ _ ->
                  bindb (catcha (check_subsumption e r conv2)
                          (\ x ->
                            Inl (ia . shows_prec_list Zero_nat
[':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'h', 'o',
  'w', 'i', 'n', 'g', ' ', 't', 'h', 'a', 't', ' ', 'e', 'q', 'u', 'a', 't',
  'i', 'o', 'n', 's', ' ', 'c', 'a', 'n', ' ', 'b', 'e', ' ', 's', 'i', 'm',
  'u', 'l', 'a', 't', 'e', 'd', ' ', 'b', 'y', ' ', 'r', 'e', 'w', 'r', 'i',
  't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm'] .
indent x)))
                    (\ _ ->
                      catcha
                        (check_critical_pairs r (critical_pairs_impl r r)
                          joins_i)
                        (\ x ->
                          Inl (ia . shows_prec_list Zero_nat
                                      [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ',
'w', 'h', 'e', 'n', ' ', 'p', 'r', 'o', 'v', 'i', 'n', 'g', ' ', 'l', 'o', 'c',
'a', 'l', ' ', 'c', 'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'o', 'f',
' '] .
                                      shows_tp (shows_prec_lab Zero_nat)
(shows_prec_list Zero_nat) i tp .
shows_nl . indent x))))));

eq_proof_lines ::
  forall a b.
    (Showa a,
      Showa b) => Eq_proof a b ->
                    Nat -> ([Char] -> [Char], (Nat, (Term a b, Term a b)));
eq_proof_lines (Refl s) i =
  ((shows_nl . shows_prec_nat Zero_nat (Suc i) . shows_string [':', ' ']) .
     shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat) [' ', '=', ' ']
       (s, s) .
       shows_string [' ', '[', 'r', 'e', 'f', 'l', ']'],
    (Suc i, (s, s)));
eq_proof_lines (Sym p) i =
  let {
    (s, (ia, (l, r))) = eq_proof_lines p i;
  } in (s . (shows_nl .
              shows_prec_nat Zero_nat (Suc ia) . shows_string [':', ' ']) .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '=', ' '] (r, l) .
                shows_string [' ', '[', 's', 'y', 'm', ' '] .
                  shows_prec_nat Zero_nat ia . shows_string [']'],
         (Suc ia, (r, l)));
eq_proof_lines (Trans p1 p2) i =
  let {
    (s1, (i1, (s, _))) = eq_proof_lines p1 i;
    (s2, (i2, (_, v))) = eq_proof_lines p2 i1;
  } in (s1 . s2 . (shows_nl .
                    shows_prec_nat Zero_nat (Suc i2) .
                      shows_string [':', ' ']) .
                    shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                      [' ', '=', ' '] (s, v) .
                      shows_string [' ', '[', 't', 'r', 'a', 'n', 's', ' '] .
                        shows_prec_nat Zero_nat i1 .
                          shows_string [',', ' '] .
                            shows_prec_nat Zero_nat i2 . shows_string [']'],
         (Suc i2, (s, v)));
eq_proof_lines (Assm (l, r) sigma) i =
  let {
    eq = (subst_apply_term l sigma, subst_apply_term r sigma);
  } in ((shows_nl . shows_prec_nat Zero_nat (Suc i) . shows_string [':', ' ']) .
          shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat) [' ', '=', ' ']
            eq .
            shows_string [' ', '[', 'a', 's', 's', 'm', ' '] .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '=', ' '] (l, r) .
                shows_string [']'],
         (Suc i, eq));
eq_proof_lines (Cong f ps) i =
  let {
    (s, (is, (ls, rs))) = eq_proofs_lines ps i;
    eq = (Fun f ls, Fun f rs);
    ia = lasta is;
    isa = butlast is;
  } in (s . (shows_nl .
              shows_prec_nat Zero_nat (Suc ia) . shows_string [':', ' ']) .
              shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                [' ', '=', ' '] eq .
                shows_list_gen (shows_prec_nat Zero_nat)
                  [' ', '[', 'c', 'o', 'n', 'g', ']']
                  [' ', '[', 'c', 'o', 'n', 'g', ' '] [',', ' '] [']'] isa,
         (Suc ia, eq));

eq_proofs_lines ::
  forall a b.
    (Showa a,
      Showa b) => [Eq_proof a b] ->
                    Nat ->
                      ([Char] -> [Char], ([Nat], ([Term a b], [Term a b])));
eq_proofs_lines [] i = (id, ([i], ([], [])));
eq_proofs_lines (p : ps) i =
  let {
    (s1, (ia, (l, r))) = eq_proof_lines p i;
    (s2, (is, (ls, rs))) = eq_proofs_lines ps ia;
  } in (s1 . s2, (ia : is, (l : ls, r : rs)));

check_proves ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Eq_proof a b -> Sum ([Char] -> [Char]) (Term a b, Term a b);
check_proves e (Refl s) = Inr (s, s);
check_proves e (Sym p) = bindb (check_proves e p) (\ (s, t) -> Inr (t, s));
check_proves e (Trans p1 p2) =
  bindb (check_proves e p1)
    (\ (s, t) ->
      bindb (check_proves e p2)
        (\ (ta, u) ->
          (if equal_term t ta then Inr (s, u)
            else Inl (shows_string
                        ['t', 'h', 'e', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'o',
                          'c', 'c', 'u', 'r', 's', ' ', 'i', 'n', ' ', 't', 'h',
                          'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                          ' ', 'p', 'a', 'r', 't', ' '] .
                       shows_nl .
                         fst (eq_proof_lines (Trans p1 p2) Zero_nat) .
                           shows_nl .
                             shows_nl .
                               shows_prec_term Zero_nat t .
                                 shows_string
                                   [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e',
                                     'q', 'u', 'a', 'l', ' ', 't', 'o', ' '] .
                                   shows_prec_term Zero_nat ta))));
check_proves e (Assm (l, r) sigma) =
  (if membera e (l, r)
    then Inr (subst_apply_term l sigma, subst_apply_term r sigma)
    else Inl (shows_string
                ['t', 'h', 'e', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'o', 'c',
                  'c', 'u', 'r', 's', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ',
                  'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'p', 'a',
                  'r', 't', ' '] .
               shows_nl .
                 fst (eq_proof_lines (Assm (l, r) sigma) Zero_nat) .
                   shows_nl .
                     shows_nl .
                       shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                         [' ', '=', ' '] (l, r) .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'i', 'n',
                             ' ', 't', 'h', 'e', ' '] .
                           shows_trs (shows_prec Zero_nat) (shows_prec Zero_nat)
                             ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l',
                               ' ', 's', 'y', 's', 't', 'e', 'm', ':']
                             [' ', '=', ' '] e));
check_proves e (Cong f ps) =
  bindb (mapMa (check_proves e) ps)
    (\ sts -> Inr (Fun f (map fst sts), Fun f (map snd sts)));

check_eq_proof ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Eq_proof a b ->
                      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_eq_proof e p eq =
  catcha
    (bindb (check_proves e p)
      (\ eqa ->
        (if eq == eqa then Inr ()
          else Inl (shows_string
                     ['t', 'h', 'e', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'd',
                       'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'f', 'i', 't',
                       ' ', 't', 'h', 'e', ' ', 'g', 'o', 'a', 'l']))))
    (\ x ->
      Inl (shows_string
             ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'a', 'n', ' ', 'e',
               'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 'e',
               'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'l', 'o', 'g',
               'i', 'c', ' ', 'p', 'r', 'o', 'o', 'f'] .
            shows_nl .
              fst (eq_proof_lines p Zero_nat) .
                shows_nl .
                  shows_nl .
                    shows_string
                      ['f', 'o', 'r', ' ', 'p', 'r', 'o', 'v', 'i', 'n', 'g',
                        ' ', 't', 'h', 'e', ' ', 'e', 'q', 'u', 'a', 't', 'i',
                        'o', 'n'] .
                      shows_nl .
                        shows_nl .
                          shows_rule (shows_prec Zero_nat) (shows_prec Zero_nat)
                            [' ', '=', ' '] eq .
                            shows_nl .
                              shows_nl .
                                shows_string
                                  ['u', 's', 'i', 'n', 'g', ' ', 't', 'h', 'e',
                                    ' '] .
                                  shows_trs (shows_prec Zero_nat)
                                    (shows_prec Zero_nat)
                                    ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n',
                                      'a', 'l', ' ', 's', 'y', 's', 't', 'e',
                                      'm', ':']
                                    [' ', '=', ' '] e .
                                    shows_nl . x));

check_equational_proof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          [(Term (Lab b [Nat]) [Char],
                             Term (Lab b [Nat]) [Char])] ->
                            (Term (Lab b [Nat]) [Char],
                              Term (Lab b [Nat]) [Char]) ->
                              Equational_proof b [Nat] [Char] ->
                                Sum ([Char] -> [Char]) ();
check_equational_proof a ia i j e eq (Equational_Proof_Tree p) =
  debug (ia [])
    ['E', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', '_', 'P', 'r', 'o', 'o',
      'f', '_', 'T', 'r', 'e', 'e']
    (check_eq_proof e p eq);
check_equational_proof a ia i j e eq (Conversion eseq) =
  debug (ia []) ['C', 'o', 'n', 'v', 'e', 'r', 's', 'i', 'o', 'n']
    (check_conversion e eseq (fst eq) (snd eq));
check_equational_proof a ia i j e eq (Completion_and_Normalization r p) =
  debug (ia [])
    ['C', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', '_', 'a', 'n', 'd', '_',
      'N', 'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n']
    (bindb (check_completion_proof a ia i j e r p)
      (\ _ ->
        let {
          s = fst eq;
          t = snd eq;
        } in (case (compute_rstep_NF r s, compute_rstep_NF r t) of {
               (Nothing, b) ->
                 Inl (shows_string
                        ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                          'c', 'o', 'm', 'p', 'u', 't', 'i', 'n', 'g', ' ', 'n',
                          'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', 's',
                          ' ', 'o', 'f', ' '] .
                       shows_term (shows_prec_lab Zero_nat)
                         (shows_prec_list Zero_nat) s .
                         shows_string [' ', 'a', 'n', 'd', ' '] .
                           shows_term (shows_prec_lab Zero_nat)
                             (shows_prec_list Zero_nat) t);
               (Just sa, Nothing) ->
                 Inl (shows_string
                        ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                          'c', 'o', 'm', 'p', 'u', 't', 'i', 'n', 'g', ' ', 'n',
                          'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', 's',
                          ' ', 'o', 'f', ' '] .
                       shows_term (shows_prec_lab Zero_nat)
                         (shows_prec_list Zero_nat) s .
                         shows_string [' ', 'a', 'n', 'd', ' '] .
                           shows_term (shows_prec_lab Zero_nat)
                             (shows_prec_list Zero_nat) t);
               (Just sa, Just ta) ->
                 (if equal_term sa ta then Inr ()
                   else Inl (shows_term (shows_prec_lab Zero_nat)
                               (shows_prec_list Zero_nat) s .
                              shows_string [' ', 'a', 'n', 'd', ' '] .
                                shows_term (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat) t .
                                  shows_string
                                    [' ', 'h', 'a', 'v', 'e', ' ', 'd', 'i',
                                      'f', 'f', 'e', 'r', 'e', 'n', 't', ' ',
                                      'n', 'o', 'r', 'm', 'a', 'l', ' ', 'f',
                                      'o', 'r', 'm', 's']));
             })));

check_equational_disproof ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    ([Char] -> [Char]) ->
                      Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                        Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                          [(Term (Lab b [Nat]) [Char],
                             Term (Lab b [Nat]) [Char])] ->
                            (Term (Lab b [Nat]) [Char],
                              Term (Lab b [Nat]) [Char]) ->
                              Equational_disproof b [Nat] [Char] ->
                                Sum ([Char] -> [Char]) ();
check_equational_disproof a ia i j e eq
  (Completion_and_Normalization_Different r p) =
  debug (ia [])
    ['C', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', '_', 'a', 'n', 'd', '_',
      'N', 'o', 'r', 'm', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n']
    (bindb (check_completion_proof a ia i j e r p)
      (\ _ ->
        let {
          s = fst eq;
          t = snd eq;
        } in (case (compute_rstep_NF r s, compute_rstep_NF r t) of {
               (Nothing, b) ->
                 Inl (shows_string
                        ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                          'c', 'o', 'm', 'p', 'u', 't', 'i', 'n', 'g', ' ', 'n',
                          'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', 's',
                          ' ', 'o', 'f', ' '] .
                       shows_term (shows_prec_lab Zero_nat)
                         (shows_prec_list Zero_nat) s .
                         shows_string [' ', 'a', 'n', 'd', ' '] .
                           shows_term (shows_prec_lab Zero_nat)
                             (shows_prec_list Zero_nat) t);
               (Just sa, Nothing) ->
                 Inl (shows_string
                        ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                          'c', 'o', 'm', 'p', 'u', 't', 'i', 'n', 'g', ' ', 'n',
                          'o', 'r', 'm', 'a', 'l', ' ', 'f', 'o', 'r', 'm', 's',
                          ' ', 'o', 'f', ' '] .
                       shows_term (shows_prec_lab Zero_nat)
                         (shows_prec_list Zero_nat) s .
                         shows_string [' ', 'a', 'n', 'd', ' '] .
                           shows_term (shows_prec_lab Zero_nat)
                             (shows_prec_list Zero_nat) t);
               (Just sa, Just ta) ->
                 (if not (equal_term sa ta) then Inr ()
                   else Inl (shows_term (shows_prec_lab Zero_nat)
                               (shows_prec_list Zero_nat) s .
                              shows_string [' ', 'a', 'n', 'd', ' '] .
                                shows_term (shows_prec_lab Zero_nat)
                                  (shows_prec_list Zero_nat) t .
                                  shows_string
                                    [' ', 'h', 'a', 'v', 'e', ' ', 's', 'a',
                                      'm', 'e', ' ', 'n', 'o', 'r', 'm', 'a',
                                      'l', ' ', 'f', 'o', 'r', 'm', ' '] .
                                    shows_term (shows_prec_lab Zero_nat)
                                      (shows_prec_list Zero_nat) sa));
             })));

certify_cert_problem ::
  forall a b c.
    (Eq b, Linorder b,
      Showa b) => Bool ->
                    Tp_ops_ext a (Lab b [Nat]) [Char] () ->
                      Dpp_ops_ext c (Lab b [Nat]) [Char] () ->
                        Cert_problem b [Nat] [Char] -> Cert_result;
certify_cert_problem a i j (TRS_Proof q r Nothing prf) =
  (case check_trs_proof a (shows_string ['1']) i j
          (mka i True (strategy_to_Q q r) r []) prf
    of {
    Inl err ->
      Error ((shows_string
                ['t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ',
                  'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c',
                  'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (TRS_Proof q r Nothing prf));
  });
certify_cert_problem a i j (TRS_Proof q r (Just s) prf) =
  (case check_trs_proof a (shows_string ['1']) i j
          (mka i True (strategy_to_Q q r) r s) prf
    of {
    Inl err ->
      Error ((shows_string
                ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 't', 'e', 'r',
                  'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o',
                  'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p',
                  't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (TRS_Proof q r (Just s) prf));
  });
certify_cert_problem a i j (Complexity_Proof q r s_o cm cc prf) =
  (case check_complexity_proof (shows_string ['1']) i
          (mka i True (strategy_to_Q q r) r (rel_rules_of s_o)) cm cc prf
    of {
    Inl err ->
      Error ((shows_string
                ['c', 'o', 'm', 'p', 'l', 'e', 'x', 'i', 't', 'y', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (Complexity_Proof q r s_o cm cc prf));
  });
certify_cert_problem a i j (TRS_Disproof q r prf) =
  (case check_trs_disproof (shows_string ['1']) i j
          (mka i False (strategy_to_Q q r) r []) prf
    of {
    Inl err ->
      Error ((shows_string
                ['n', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
                  'o', 'n', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't',
                  ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (TRS_Disproof q r prf));
  });
certify_cert_problem a i j (OTRS_Disproof q r prf) =
  (case check_otrs_disproof (shows_string ['1']) i
          (mka i False (ostrategy_to_Q q r) r []) prf
    of {
    Inl err ->
      Error ((shows_string
                ['o', 'u', 't', 'e', 'r', 'm', 'o', 's', 't', ' ', 'n', 'o',
                  'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n',
                  ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a',
                  'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (OTRS_Disproof q r prf));
  });
certify_cert_problem a i j (Rel_TRS_Disproof q r s prf) =
  (case check_reltrs_disproof (shows_string ['1']) i j
          (mka i False (strategy_to_Q q r) r s) prf
    of {
    Inl err ->
      Error ((shows_string
                ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'n', 'o', 'n',
                  't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ',
                  'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c',
                  'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (Rel_TRS_Disproof q r s prf));
  });
certify_cert_problem a i j (DP_Proof m p pw q r rw prf) =
  (case check_dp_proof_with_unlab a (shows_string ['1']) j
          (mkc j True m p pw (strategy_to_Q q (r ++ rw)) r rw) prf
    of {
    Inl err ->
      Error ((shows_string
                ['f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (DP_Proof m p pw q r rw prf));
  });
certify_cert_problem a i j (DP_Disproof m p q r prf) =
  (case check_dp_disproof (shows_string ['1']) j
          (mkc j False m p [] (strategy_to_Q q r) [] r) prf
    of {
    Inl err ->
      Error ((shows_string
                ['i', 'n', 'f', 'i', 'n', 'i', 't', 'e', 'n', 'e', 's', 's',
                  ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a',
                  'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (DP_Disproof m p q r prf));
  });
certify_cert_problem a i j (TRS_Confluence r prf) =
  (case check_cr_proof a (shows_string ['1']) i j r prf of {
    Inl err ->
      Error ((shows_string
                ['c', 'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (TRS_Confluence r prf));
  });
certify_cert_problem a i j (TRS_Non_Confluence r prf) =
  (case check_ncr_proof a (shows_string ['1']) i j r prf of {
    Inl err ->
      Error ((shows_string
                ['c', 'o', 'n', 'f', 'l', 'u', 'e', 'n', 'c', 'e', ' ', 'd',
                  'i', 's', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ',
                  'a', 'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (TRS_Non_Confluence r prf));
  });
certify_cert_problem a i j (Completion_Proof e r prf) =
  (case check_completion_proof a (shows_string ['1']) i j e r prf of {
    Inl err ->
      Error ((shows_string
                ['c', 'o', 'm', 'p', 'l', 'e', 't', 'i', 'o', 'n', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (Completion_Proof e r prf));
  });
certify_cert_problem a i j (Equational_Proof e eq prf) =
  (case check_equational_proof a (shows_string ['1']) i j e eq prf of {
    Inl err ->
      Error ((shows_string
                ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'p',
                  'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c',
                  'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (Equational_Proof e eq prf));
  });
certify_cert_problem a i j (Equational_Disproof e eq prf) =
  (case check_equational_disproof a (shows_string ['1']) i j e eq prf of {
    Inl err ->
      Error ((shows_string
                ['e', 'q', 'u', 'a', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'd',
                  'i', 's', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ',
                  'a', 'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ -> Certified (xml_tag (Equational_Disproof e eq prf));
  });

mk_impl ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Bool ->
                 [Term a b] ->
                   [(Term a b, Term a b)] ->
                     [(Term a b, Term a b)] ->
                       (Bool,
                         ([Term a b],
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 Rbt (a, Nat)
                                   [(Bool, (Term a b, Term a b))])))));
mk_impl nfs q r rw =
  let {
    (vr, ra) = partition (is_Var . fst) r;
    (vrw, rwa) = partition (is_Var . fst) rw;
  } in (nfs, (q, (nF_subset (map fst (r ++ rw)) q,
                   (vr, (vrw, insert_rules True ra
                                (insert_rules False rwa (rm_empty ())))))));

mk :: forall a b.
        (Eq a, Linorder a,
          Eq b) => Bool ->
                     [Term a b] ->
                       [(Term a b, Term a b)] ->
                         [(Term a b, Term a b)] -> Tp a b;
mk nfs q r rw = Tp (mk_impl nfs q r rw);

rw_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat)
                                 [(Bool, (Term a b, Term a b))]))))) ->
                       [(Term a b, Term a b)];
rw_impl (uu, (uv, (uw, (ux, (vRw, m))))) = vRw ++ rules_with not m;

rw :: forall a b. (Eq a, Linorder a) => Tp a b -> [(Term a b, Term a b)];
rw tp = rw_impl (impl_ofa tp);

nfs_impl ::
  forall a b.
    (Linorder a) => (Bool,
                      ([Term a b],
                        (Bool,
                          ([(Term a b, Term a b)],
                            ([(Term a b, Term a b)],
                              Rbt (a, Nat)
                                [(Bool, (Term a b, Term a b))]))))) ->
                      Bool;
nfs_impl (nfs, uu) = nfs;

nfs :: forall a b. (Linorder a) => Tp a b -> Bool;
nfs tp = nfs_impl (impl_ofa tp);

rules_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat)
                                 [(Bool, (Term a b, Term a b))]))))) ->
                       [(Term a b, Term a b)];
rules_impl (uu, (uv, (uw, (vR, (vRw, m))))) = vR ++ vRw ++ map snd (values m);

rulesa :: forall a b. (Eq a, Linorder a) => Tp a b -> [(Term a b, Term a b)];
rulesa tp = rules_impl (impl_ofa tp);

is_QNF_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Bool,
                 ([Term a b],
                   (Bool,
                     ([(Term a b, Term a b)],
                       ([(Term a b, Term a b)],
                         Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
                 Term a b -> Bool;
is_QNF_impl (uu, (q, uv)) = nF_terms_list q;

is_QNF :: forall a b. (Eq a, Linorder a, Eq b) => Tp a b -> Term a b -> Bool;
is_QNF tp = is_QNF_impl (impl_ofa tp);

q_empty_impl ::
  forall a b.
    (Bool,
      ([Term a b],
        (Bool,
          ([(Term a b, Term a b)],
            ([(Term a b, Term a b)],
              Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
      Bool;
q_empty_impl (uu, (q, uv)) = null q;

q_empty :: forall a b. (Linorder a) => Tp a b -> Bool;
q_empty tp = q_empty_impl (impl_ofa tp);

qreltrs_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat)
                                 [(Bool, (Term a b, Term a b))]))))) ->
                       (Bool,
                         (Set (Term a b),
                           (Set (Term a b, Term a b),
                             Set (Term a b, Term a b))));
qreltrs_impl (nfs, (q, (uu, (vR, (vRw, m))))) =
  (nfs, (Set q, (Set (vR ++ rules_with id m), Set (vRw ++ rules_with not m))));

qreltrs ::
  forall a b.
    (Eq a,
      Linorder a) => Tp a b ->
                       (Bool,
                         (Set (Term a b),
                           (Set (Term a b, Term a b),
                             Set (Term a b, Term a b))));
qreltrs tp = qreltrs_impl (impl_ofa tp);

rules_map_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat)
                                 [(Bool, (Term a b, Term a b))]))))) ->
                       (a, Nat) -> [(Term a b, Term a b)];
rules_map_impl (uu, (uv, (uw, (ux, (uy, m))))) fn =
  (case rm_lookup fn m of {
    Nothing -> [];
    Just a -> map snd a;
  });

rules_map ::
  forall a b.
    (Eq a, Linorder a) => Tp a b -> (a, Nat) -> [(Term a b, Term a b)];
rules_map tp = rules_map_impl (impl_ofa tp);

delete_R_Rw_impl ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => (Bool,
                 ([Term a b],
                   (Bool,
                     ([(Term a b, Term a b)],
                       ([(Term a b, Term a b)],
                         Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
                 [(Term a b, Term a b)] ->
                   [(Term a b, Term a b)] ->
                     (Bool,
                       ([Term a b],
                         (Bool,
                           ([(Term a b, Term a b)],
                             ([(Term a b, Term a b)],
                               Rbt (a, Nat) [(Bool, (Term a b, Term a b))])))));
delete_R_Rw_impl (nfs, (q, (uu, (vR, (vRw, m))))) r rw =
  let {
    (vr, ra) = partition (is_Var . fst) r;
    (vrw, rwa) = partition (is_Var . fst) rw;
    vra = list_diff vR vr;
    vrwa = list_diff vRw vrw;
    ma = delete_rules True ra (delete_rules False rwa m);
  } in (nfs, (q, (nF_subset (map fst (vra ++ vrwa ++ map snd (values ma))) q,
                   (vra, (vrwa, ma)))));

delete_R_Rw ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Tp a b ->
                 [(Term a b, Term a b)] -> [(Term a b, Term a b)] -> Tp a b;
delete_R_Rw tp r rw = Tp (delete_R_Rw_impl (impl_ofa tp) r rw);

split_rules_impl ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => (Bool,
                 ([Term a b],
                   (Bool,
                     ([(Term a b, Term a b)],
                       ([(Term a b, Term a b)],
                         Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_rules_impl tp rs = partition (membera rs) (rules_impl tp);

split_rules ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Tp a b ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_rules tp = split_rules_impl (impl_ofa tp);

nFQ_subset_NF_rules_impl ::
  forall a b.
    (Bool,
      ([Term a b],
        (Bool,
          ([(Term a b, Term a b)],
            ([(Term a b, Term a b)],
              Rbt (a, Nat) [(Bool, (Term a b, Term a b))]))))) ->
      Bool;
nFQ_subset_NF_rules_impl (uu, (uv, (b, uw))) = b;

nFQ_subset_NF_rules :: forall a b. (Linorder a) => Tp a b -> Bool;
nFQ_subset_NF_rules tp = nFQ_subset_NF_rules_impl (impl_ofa tp);

tp_rbt_impl ::
  forall a b. (Eq a, Linorder a, Eq b) => Tp_ops_ext (Tp a b) a b ();
tp_rbt_impl =
  Tp_ops_ext qreltrs q r rw rulesa q_empty is_QNF nFQ_subset_NF_rules rules_map
    delete_R_Rw split_rules mk nfs ();

m_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
m_impl (uu, (mi, uv)) = mi;

impl_ofb ::
  forall a b.
    (Linorder a) => Dpp a b ->
                      (Bool,
                        (Bool,
                          ([(Term a b, Term a b)],
                            ([(Term a b, Term a b)],
                              ([Term a b],
                                (Bool,
                                  (Bool,
                                    ([(Term a b, Term a b)],
                                      ([(Term a b, Term a b)],
(Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
  (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool)))))))))));
impl_ofb (Dpp x) = x;

m :: forall a b. (Linorder a) => Dpp a b -> Bool;
m d = m_impl (impl_ofb d);

p_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      [(Term a b, Term a b)];
p_impl (uu, (uv, (p, uw))) = p;

p :: forall a b. (Linorder a) => Dpp a b -> [(Term a b, Term a b)];
p d = p_impl (impl_ofb d);

q_impla ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      [Term a b];
q_impla (uu, (uv, (uw, (ux, (q, uy))))) = q;

qb :: forall a b. (Linorder a) => Dpp a b -> [Term a b];
qb d = q_impla (impl_ofb d);

r_impla ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       [(Term a b, Term a b)];
r_impla (uu, (uv, (uw, (ux, (uy, (uz, (va, (vr, (vb, (m, vc)))))))))) =
  vr ++ rules_with id m;

rb :: forall a b. (Eq a, Linorder a) => Dpp a b -> [(Term a b, Term a b)];
rb d = r_impla (impl_ofb d);

applicable_rule_implb ::
  forall a b. (Term a b -> Bool) -> (Term a b, Term a b) -> Bool;
applicable_rule_implb nf r = all nf (args (fst r));

wwf_qtrs_impl ::
  forall a b. (Eq b) => (Term a b -> Bool) -> [(Term a b, Term a b)] -> Bool;
wwf_qtrs_impl nf r =
  all (\ ra -> wf_rule ra || not (applicable_rule_implb nf ra)) r;

mk_impla ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Bool ->
                 Bool ->
                   [(Term a b, Term a b)] ->
                     [(Term a b, Term a b)] ->
                       [Term a b] ->
                         [(Term a b, Term a b)] ->
                           [(Term a b, Term a b)] ->
                             (Bool,
                               (Bool,
                                 ([(Term a b, Term a b)],
                                   ([(Term a b, Term a b)],
                                     ([Term a b],
                                       (Bool,
 (Bool,
   ([(Term a b, Term a b)],
     ([(Term a b, Term a b)],
       (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
         (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool)))))))))));
mk_impla nfs mi p pw q r rw =
  let {
    (vr, ra) = partition (is_Var . fst) r;
    (vrw, rwa) = partition (is_Var . fst) rw;
    rs = r ++ rw;
  } in (nfs, (mi, (p, (pw, (q, (nF_subset (map fst rs) q,
                                 (all (\ (_, rb) -> not (is_Var rb)) rs,
                                   (vr, (vrw,
  (insert_rules True ra (insert_rules False rwa (rm_empty ())),
    (insert_rules True (map (\ (x, y) -> (y, x)) r)
       (insert_rules False (map (\ (x, y) -> (y, x)) rw) (rm_empty ())),
      wwf_qtrs_impl (nF_terms_list q) rs)))))))))));

mkb ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Bool ->
                 Bool ->
                   [(Term a b, Term a b)] ->
                     [(Term a b, Term a b)] ->
                       [Term a b] ->
                         [(Term a b, Term a b)] ->
                           [(Term a b, Term a b)] -> Dpp a b;
mkb nfs mi p pw q r rw = Dpp (mk_impla nfs mi p pw q r rw);

pw_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      [(Term a b, Term a b)];
pw_impl (uu, (uv, (uw, (pw, ux)))) = pw;

pw :: forall a b. (Linorder a) => Dpp a b -> [(Term a b, Term a b)];
pw d = pw_impl (impl_ofb d);

rw_impla ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       [(Term a b, Term a b)];
rw_impla (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vrw, (m, vc)))))))))) =
  vrw ++ rules_with not m;

rwb :: forall a b. (Eq a, Linorder a) => Dpp a b -> [(Term a b, Term a b)];
rwb d = rw_impla (impl_ofb d);

nfs_impla ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
nfs_impla (nfs, uu) = nfs;

nfsb :: forall a b. (Linorder a) => Dpp a b -> Bool;
nfsb d = nfs_impla (impl_ofb d);

dpp_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       (Bool,
                         (Bool,
                           (Set (Term a b, Term a b),
                             (Set (Term a b, Term a b),
                               (Set (Term a b),
                                 (Set (Term a b, Term a b),
                                   Set (Term a b, Term a b)))))));
dpp_impl (nfs, (mi, (p, (pw, (q, (uu, (uv, (vr, (vrw, (m, uw)))))))))) =
  (nfs, (mi, (Set p,
               (Set pw,
                 (Set q,
                   (Set (vr ++ rules_with id m),
                     Set (vrw ++ rules_with not m)))))));

dppa ::
  forall a b.
    (Eq a,
      Linorder a) => Dpp a b ->
                       (Bool,
                         (Bool,
                           (Set (Term a b, Term a b),
                             (Set (Term a b, Term a b),
                               (Set (Term a b),
                                 (Set (Term a b, Term a b),
                                   Set (Term a b, Term a b)))))));
dppa d = dpp_impl (impl_ofb d);

pairs_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      [(Term a b, Term a b)];
pairs_impl (uu, (uv, (p, (pw, uw)))) = p ++ pw;

pairs :: forall a b. (Linorder a) => Dpp a b -> [(Term a b, Term a b)];
pairs d = pairs_impl (impl_ofb d);

rules_impla ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       [(Term a b, Term a b)];
rules_impla (uu, (uv, (uw, (ux, (uy, (uz, (va, (vr, (vrw, (m, vb)))))))))) =
  vr ++ vrw ++ map snd (values m);

rulesc :: forall a b. (Eq a, Linorder a) => Dpp a b -> [(Term a b, Term a b)];
rulesc d = rules_impla (impl_ofb d);

is_QNF_impla ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 Term a c -> Bool;
is_QNF_impla (uu, (uv, (uw, (ux, (q, uy))))) = nF_terms_list q;

is_QNFb :: forall a b. (Eq a, Linorder a, Eq b) => Dpp a b -> Term a b -> Bool;
is_QNFb d = is_QNF_impla (impl_ofb d);

q_empty_impla ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
q_empty_impla (uu, (uv, (uw, (ux, (q, uy))))) = null q;

q_emptyb :: forall a b. (Linorder a) => Dpp a b -> Bool;
q_emptyb d = q_empty_impla (impl_ofb d);

rules_map_impla ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       (a, Nat) -> [(Term a b, Term a b)];
rules_map_impla (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (m, vd)))))))))) fn
  = (case rm_lookup fn m of {
      Nothing -> [];
      Just a -> map snd a;
    });

rules_mapa ::
  forall a b.
    (Eq a, Linorder a) => Dpp a b -> (a, Nat) -> [(Term a b, Term a b)];
rules_mapa d = rules_map_impla (impl_ofb d);

wwf_rules_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
wwf_rules_impl
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (vd, (ve, wwf))))))))))) = wwf;

wwf_rules :: forall a b. (Linorder a) => Dpp a b -> Bool;
wwf_rules d = wwf_rules_impl (impl_ofb d);

delete_P_Pw_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   [(Term a b, Term a b)] ->
                     (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool)))))))))));
delete_P_Pw_impl (nfs, (mi, (p, (pw, rest)))) pd pwd =
  (nfs, (mi, (list_diff p pd, (list_diff pw pwd, rest))));

delete_P_Pw ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Dpp a b ->
                 [(Term a b, Term a b)] -> [(Term a b, Term a b)] -> Dpp a b;
delete_P_Pw d p pw = Dpp (delete_P_Pw_impl (impl_ofb d) p pw);

delete_R_Rw_impla ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   [(Term a b, Term a b)] ->
                     (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool)))))))))));
delete_R_Rw_impla d r rw =
  let {
    (nfs, (mi, (p, (pw, (q, (nfq, (nc, (vR, (vRw, (m, (rm, wwf))))))))))) = d;
    (vr, ra) = partition (is_Var . fst) r;
    (vrw, rwa) = partition (is_Var . fst) rw;
    vra = list_diff vR vr;
    vrwa = list_diff vRw vrw;
    ma = delete_rules True ra (delete_rules False rwa m);
    rma = delete_rules True (map (\ (x, y) -> (y, x)) r)
            (delete_rules False (map (\ (x, y) -> (y, x)) rw) rm);
    rs = vra ++ vrwa ++ map snd (values ma);
  } in (nfs, (mi, (p, (pw, (q, (nfq || nF_subset (map fst rs) q,
                                 (nc || all (\ (_, rb) -> not (is_Var rb)) rs,
                                   (vra, (vrwa,
   (ma, (rma, wwf || wwf_qtrs_impl (nF_terms_list q) rs)))))))))));

delete_R_Rwb ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Dpp a b ->
                 [(Term a b, Term a b)] -> [(Term a b, Term a b)] -> Dpp a b;
delete_R_Rwb d r rw = Dpp (delete_R_Rw_impla (impl_ofb d) r rw);

split_pairs_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_pairs_impl d ps = partition (membera ps) (pairs_impl d);

split_pairs ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Dpp a b ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_pairs d = split_pairs_impl (impl_ofb d);

split_rules_impla ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_rules_impla d rs = partition (membera rs) (rules_impla d);

split_rulesb ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Dpp a b ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
split_rulesb d = split_rules_impla (impl_ofb d);

replace_pair_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 (Term a b, Term a b) ->
                   [(Term a b, Term a b)] ->
                     (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool)))))))))));
replace_pair_impl (nfs, (mi, (p, (pw, rest)))) pair ps =
  (nfs, (mi, (replace_impl pair ps p, (replace_impl pair ps pw, rest))));

replace_pair ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => Dpp a b ->
                 (Term a b, Term a b) -> [(Term a b, Term a b)] -> Dpp a b;
replace_pair d pair ps = Dpp (replace_pair_impl (impl_ofb d) pair ps);

intersect_pairs_impl ::
  forall a b.
    (Eq a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   (Bool,
                     (Bool,
                       ([(Term a b, Term a b)],
                         ([(Term a b, Term a b)],
                           ([Term a b],
                             (Bool,
                               (Bool,
                                 ([(Term a b, Term a b)],
                                   ([(Term a b, Term a b)],
                                     (Rbt (a, Nat)
[(Bool, (Term a b, Term a b))],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 Bool)))))))))));
intersect_pairs_impl (nfs, (mi, (p, (pw, rest)))) ps =
  (nfs, (mi, (list_inter p ps, (list_inter pw ps, rest))));

intersect_pairs ::
  forall a b.
    (Eq a, Linorder a, Eq b) => Dpp a b -> [(Term a b, Term a b)] -> Dpp a b;
intersect_pairs d ps = Dpp (intersect_pairs_impl (impl_ofb d) ps);

intersect_rules_impl ::
  forall a b.
    (Eq a, Linorder a,
      Eq b) => (Bool,
                 (Bool,
                   ([(Term a b, Term a b)],
                     ([(Term a b, Term a b)],
                       ([Term a b],
                         (Bool,
                           (Bool,
                             ([(Term a b, Term a b)],
                               ([(Term a b, Term a b)],
                                 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                   (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                                     Bool))))))))))) ->
                 [(Term a b, Term a b)] ->
                   (Bool,
                     (Bool,
                       ([(Term a b, Term a b)],
                         ([(Term a b, Term a b)],
                           ([Term a b],
                             (Bool,
                               (Bool,
                                 ([(Term a b, Term a b)],
                                   ([(Term a b, Term a b)],
                                     (Rbt (a, Nat)
[(Bool, (Term a b, Term a b))],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 Bool)))))))))));
intersect_rules_impl d ri =
  let {
    (nfs, (mi, (p, (pw, (q, (nfq, (nc, (vR, (vRw, (m, (rm, wwf))))))))))) = d;
    (vri, ria) = partition (is_Var . fst) ri;
    vr = list_inter vR vri;
    vrw = list_inter vRw vri;
    ma = intersect_rules ria m;
    rma = intersect_rules (map (\ (x, y) -> (y, x)) ri) rm;
    rs = vr ++ vrw ++ map snd (values ma);
  } in (nfs, (mi, (p, (pw, (q, (nfq || nF_subset (map fst rs) q,
                                 (nc || all (\ (_, r) -> not (is_Var r)) rs,
                                   (vr, (vrw,
  (ma, (rma, wwf || wwf_qtrs_impl (nF_terms_list q) rs)))))))))));

intersect_rulesa ::
  forall a b.
    (Eq a, Linorder a, Eq b) => Dpp a b -> [(Term a b, Term a b)] -> Dpp a b;
intersect_rulesa d rs = Dpp (intersect_rules_impl (impl_ofb d) rs);

reverse_rules_map_impl ::
  forall a b.
    (Eq a,
      Linorder a) => (Bool,
                       (Bool,
                         ([(Term a b, Term a b)],
                           ([(Term a b, Term a b)],
                             ([Term a b],
                               (Bool,
                                 (Bool,
                                   ([(Term a b, Term a b)],
                                     ([(Term a b, Term a b)],
                                       (Rbt (a, Nat)
  [(Bool, (Term a b, Term a b))],
 (Rbt (a, Nat) [(Bool, (Term a b, Term a b))], Bool))))))))))) ->
                       (a, Nat) -> [(Term a b, Term a b)];
reverse_rules_map_impl
  (uu, (uv, (uw, (ux, (uy, (uz, (va, (vb, (vc, (vd, (m, ve))))))))))) fn =
  (case rm_lookup fn m of {
    Nothing -> [];
    Just a -> map snd a;
  });

reverse_rules_map ::
  forall a b.
    (Eq a, Linorder a) => Dpp a b -> (a, Nat) -> [(Term a b, Term a b)];
reverse_rules_map d = reverse_rules_map_impl (impl_ofb d);

rules_no_left_var_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
rules_no_left_var_impl (uu, (uv, (uw, (ux, (uy, (uz, (va, ([], ([], vb)))))))))
  = True;
rules_no_left_var_impl (v, (vb, (ve, (vg, (vi, (vk, (vm, (vq : vr, vp)))))))) =
  False;
rules_no_left_var_impl
  (v, (vb, (ve, (vg, (vi, (vk, (vm, (vo, (vs : vt, vr))))))))) = False;

rules_no_left_var :: forall a b. (Linorder a) => Dpp a b -> Bool;
rules_no_left_var d = rules_no_left_var_impl (impl_ofb d);

nFQ_subset_NF_rules_impla ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
nFQ_subset_NF_rules_impla (uu, (uv, (uw, (ux, (uy, (b, uz)))))) = b;

nFQ_subset_NF_rulesa :: forall a b. (Linorder a) => Dpp a b -> Bool;
nFQ_subset_NF_rulesa d = nFQ_subset_NF_rules_impla (impl_ofb d);

rules_non_collapsing_impl ::
  forall a b.
    (Bool,
      (Bool,
        ([(Term a b, Term a b)],
          ([(Term a b, Term a b)],
            ([Term a b],
              (Bool,
                (Bool,
                  ([(Term a b, Term a b)],
                    ([(Term a b, Term a b)],
                      (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                        (Rbt (a, Nat) [(Bool, (Term a b, Term a b))],
                          Bool))))))))))) ->
      Bool;
rules_non_collapsing_impl (uu, (uv, (uw, (ux, (uy, (uz, (nc, va))))))) = nc;

rules_non_collapsing :: forall a b. (Linorder a) => Dpp a b -> Bool;
rules_non_collapsing d = rules_non_collapsing_impl (impl_ofb d);

dpp_rbt_impl ::
  forall a b. (Eq a, Linorder a, Eq b) => Dpp_ops_ext (Dpp a b) a b ();
dpp_rbt_impl =
  Dpp_ops_ext dppa p pw pairs qb rb rwb rulesc q_emptyb rules_no_left_var
    rules_non_collapsing is_QNFb nFQ_subset_NF_rulesa rules_mapa
    reverse_rules_map intersect_pairs replace_pair intersect_rulesa delete_P_Pw
    delete_R_Rwb split_pairs split_rulesb mkb m nfsb wwf_rules ();

certify_proof :: Bool -> [Char] -> Parse_result [Char] Cert_result;
certify_proof a problemString =
  catchb
    (bindc (parse_cert_problem problemString)
      (\ cp ->
        returna
          (case extract_input_proof problemString of {
            Nothing ->
              Error ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                      'n', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'a', 'n',
                      'd', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ',
                      's', 't', 'r', 'i', 'n', 'g', ' ', 'f', 'o', 'r', ' ',
                      'c', 'o', 'm', 'p', 'a', 'r', 'i', 's', 'o', 'n', ' ',
                      'w', 'i', 't', 'h', ' ', 'p', 'a', 'r', 's', 'e', 'd',
                      ' ', 'r', 'e', 's', 'u', 'l', 't'];
            Just (the_input, (_, the_proof)) ->
              (if eqWhiteSpace
                    (shows_prec_xml Zero_nat (xml_cert_problem cp) []) the_input
                then (case certify_cert_problem a tp_rbt_impl dpp_rbt_impl cp of
                       {
                       Certified tag ->
                         (if debug ['0']
                               ['c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g',
                                 ' ', 't', 'a', 'g', 's']
                               starts_with the_proof tag
                           then Certified tag
                           else Error (['p', 'r', 'o', 'v', 'e', 'n', ' ', 'p',
 'r', 'o', 'p', 'e', 'r', 't', 'y', ' '] ++
tag ++
  [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e',
    's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 'p', 'r', 'o', 'o', 'f', ' ',
    'i', 'n', ' ', 'i', 'n', 'p', 'u', 't', ' ', 's', 't', 'r', 'i', 'n', 'g',
    ':', ' '] ++
    takea (nat 20) the_proof));
                       Unsupported aa -> Unsupported aa;
                       Error aa -> Error aa;
                     })
                else Error (concat
                             [['p', 'a', 'r', 's', 'e', 'd', ' ', 'p', 'r', 'o',
                                'b', 'l', 'e', 'm', ' ', 'd', 'o', 'e', 's',
                                ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o',
                                ' ', 'i', 'n', 'p', 'u', 't'],
                               [newline], ['i', 'n', 'p', 'u', 't', ':'],
                               [newline],
                               filter
                                 (\ c ->
                                   not (membera [' ', newline, '\t', '\r'] c))
                                 the_input,
                               [newline], [newline],
                               ['p', 'a', 'r', 's', 'e', 'd', ':'], [newline],
                               filter
                                 (\ c ->
                                   not (membera [' ', newline, '\t', '\r'] c))
                                 (shows_prec_xml Zero_nat (xml_cert_problem cp)
                                   []),
                               [newline]]));
          })))
    (\ err ->
      returna
        (Unsupported
          (['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e', ' ', 'p',
             'a', 'r', 's', 'i', 'n', 'g'] ++
            [newline] ++ err)));

}
